/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012-2014 StITPro (www.stitpro.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: PCircAdaptMainwindow.h 1174 2014-03-21 14:15:01Z eelco $
 *
 */

#ifndef _PCIRCADAPTMAINWINDOW_H
#define _PCIRCADAPTMAINWINDOW_H

#include "ucirculation.h"
#include "msimulatecirculation.h"

#include "psignalview.h"
#include "ppvview.h"
#include "pvalueview.h"
#include "psystemview.h"
#include "pparamview.h"

#include "PCircAdaptOptionsDialog.h"

//#include "PEcgSimUndoStack.h"
#include "ui_PCircAdaptMainwindow.h"

#include <QMainWindow>
#include <QLabel>
#include <QTimerEvent>
#include <QSignalMapper>
#include <QSplashScreen>

#ifndef QT_NO_DEBUG
#include <QUndoView>
#endif

namespace Peacs
{
    namespace CircAdapt
    {

        /** @class PCircAdaptMainWindow PCircAdaptMainWindow.h "CircAdaptMain\PCircAdaptMainWindow.h"
         *  @brief The main window of CircAdapt
         * All connections required to synchornize between the diffrent views on the CircAdaptData are
         * created in this class
         * @see PCircAdaptData
         */
        class PCircAdaptMainwindow : public QMainWindow {
            Q_OBJECT
            Q_DISABLE_COPY(PCircAdaptMainwindow)

        public:
            /** @brief Constructor
              * @param the filename of a case file. Used when the program is started with input arguments
              * @param parent The parent of this mainwindow
              */
            explicit PCircAdaptMainwindow( /*const QString &filename,*/ QWidget * parent = 0 );


            /** @brief Destructor of this main window
              */
            ~PCircAdaptMainwindow();


            /** @brief Returns the const reference CircAdapt data store object
              * This object is used throughout the whole application as the central data store
              *
              * @return The const UCirculation storage object
              */
            inline const UCirculation & getCircAdaptDataConst( void ) const { return mData; }


            /** @brief Returns the changeable reference CircAdapt data store object
              * This object is used throughout the whole application as the central data store
              *
              * @return The UCirculation storage object
              */
            inline UCirculation & getCircAdaptData( void ) { return mData; }


            /** @brief write the current state of the main window to the settings file
              */
            void writeSettings( void );


            /** @brief read the current state of the main window to the settings file
              */
            void readSettings( void );


            /** @brief load the first case found in the data directory (mDataDir)
              * @param the filename of a case file. Used when the program is started with input arguments
              */
            void loadDefaultCase(const QString &filename);


            /** @brief Creates and initializes all GUI items
              * This function should be called once after an instance of this class is created
              */
            void setupUi( void );


            /** @brief Starts the simulation thread
              */
            inline void startSimulation( void ) { if ( !mCirculation.isRunning() ) { mCirculation.start( /*QThread::LowestPriority*/ ); } }


            /** @brief Returns the undo stack of this application
              *
              * @return reference to the undo stack
              */
//            inline PEcgSimUndoStack& getUndoStack( void ) { return *mUndoStack; }


        public slots:

            /** @brief react to a close event
              * @param event the closing event
              */
            void closeEvent( QCloseEvent * event );


            /** @brief slot for the options dialog
              * Reads info from the settings file
              * For the meaning of the parameter @see POptionsDialog::mChangedItems
              *
              * @param changed indicated what options have changed. This is a 32 bit-flag value
              */
            void updateAllSettings( int changed );


//            /** @brief Slot to handle user changes at the source and create an undo command for
//              * it. A new {@link PSourceUndo} is created with the new and previous values of
//              * the source parameters {@link PSource}
//              *
//              * @param text the text to set with the undo command. This text must consist of a
//              * short string describing the command
//              *
//              * @see PSourceUndo
//              */


        signals:

            void update();

        private slots:

//            void on_actionDefault_open_triggered();

            /** @brief slot used for save a case file
              */
            void on_actionSave_as_triggered();

            /** @brief slot used to show the index help
              */
            void on_actionIndex_triggered();

            /** @brief slot used to show the help
              */
            void on_actionContext_Help_F1_triggered();

            /** @brief slot used to show the about box
              */
            void on_actionAbout_triggered();

            /** @brief event used in the movie mode to cretaed moving surface functions and timelines
              * @param event the current timer event
              */
            void timerEvent(QTimerEvent *event);


            /** @brief closes the application
              * Checks if the case file needs to be saved
              * @see mDirty
              */
            void on_actionExit_triggered();

            /** @brief opens the options dialog
              */
            void on_actionOptions_triggered();

            /** @brief copies the window with the mouse underneath to the clipboard
              */
            void on_actionCopy_triggered();

            /** @brief print TODO implement
              */
            void on_actionPrint_triggered();

            /** @brief Save the current case file
              * Checks if no errors and then stores the case file under the given
              * filename
              * @see mFilename
              */
            void on_actionSave_triggered();

            /** @brief Open the case file and fills all fields from
              * the ecgsim casefile (*.ecgsimcase)
              * @see mFilename
              */
            void on_action_Open_triggered();

            void openRecentFile();

            /** @brief Slot that can recieve when the circulation thread has started
              */
            void thread_started( void );

            void on_actionRun_triggered();

            void on_actionStore_current_triggered();

            void on_actionAutoscale_triggered();

            void on_actionReset_triggered();

            /** @brief Toggles the widgets visibility state
              * This function must be called when the user presses an action item
              * in the 'Windows' submenu in the 'View' main menu item. It toggles
              * the visibility state of the clicked item
              *
              * windowPtr the pointer to the windget to toggle the visibility state for
              */
            void toggleWindowVisibleState( QWidget * windowPtr );

            /** @brief Recieves that the state of the thread has changed
              * This signal is recieved when the thread machine is going to 'paused' state or
              * 'active' state
              *
              * @param state indicates the new state
              */
            void stateChanged( /*MSimulateCirculation::statetype*/ int state );

            /** @brief Recieves that the circulation model is (re)loaded
              * Loaded means eigther that a file is (re)loaded or that the
              * model is (re)initialized with default settings.
              *
              * @param type indicates the type of model loaded
              */
            void modelLoaded( /*MSimulateCirculation::modeltype*/ int type );

            /** @brief Recieves that the circulation model is saved to file
              */
            void modelSaved( void );


            /** @brief Recieves that a previous action failed
              * All actions send to the simulation thread are performed asynchronusly. The
              * status of the action can only be and must be reported after it was initiated.
              * This function informs about failures
              *
              * @param type indicates the type of action that failed
              */
            void actionFailed( /*MSimulateCirculation::failtype*/ int type );


            /** @brief Recieves that the circulation model and local data is initialized
              */
            void allInitialized( void );


            /** @brief Recieves that the initial heartbeats, to reach a more steady state, is are processed
              */
            void initialHeartbeatsDone( void );


            /** @brief slot that must be called when new data is recieved and ready to handled
              * This function will check if a new/initial (reset) case is loaded and will trigger
              * an autoscale call
              */
            void updateInfo( void );

        private:
            /** TODO Comment
              */
            void createWidgetInfo( void );

            /** @brief Build the menu and toolbars for this application as far as they are not defined
              * in the ui resource.
              * In this function some dynamic menu items are added and the special Qt undo and redo actions
              * are created and added
              */
            void buildMenu(void);


            /** @brief Makes the view refresh itself. All items in the view are refreshed
              * according the current information
              */
//            void refreshWindow( void );

            /** @brief open a case file with @param filename
              * if the case cannot be opened the current case is cleared
              */
            void open(const QString &filename);

            /** @brief Create the toolbar of the ECG window
              */
            void createMainToolBar(void);


            void setCurrentFile(const QString &fileName);

            void updateRecentFileActions();

            QString strippedName(const QString &fullFileName);

            /** @brief fils the undo stack with the initial values
              */
            void InitUndoStack( void );

            void initSpecs(const TCirculation &);



        private:

            QSplashScreen * mSplashScreenPtr;

            // Class variables section
            UCirculation mData;


            /** @brief the Qt generated ui items
              */
            Ui::PCircAdaptMainwindow mUi;

            /** @brief all case data
              */
            MSimulateCirculation mCirculation;

            /** @brief Directory where the resource data files can be found
              */
            QString mDataDir;

//            /** @brief The filename of the case
//              */
//            QString mFilename;

            /** @brief The directory from which the file in the case are loaded
              */
            QString mCurrentDir;

            /** @brief mDirty is true if any of the fields are changed and not
              * saved
              */
            bool mDirty;

            QSignalMapper * mWinSigMapper;

            /** @brief splitter for both the left and right splitter
              */
//            QSplitter *mMainSplitter;

            /** @brief splitter in which the heart and AP is shown
              */
//            QSplitter *mLeftSplitter;

            /** @brief splitter in which the thorax and ECG is shown
              */
//            QSplitter *mRightSplitter;

            /** @brief the timerId used to cretae the movie
              */
            int mTimerId;

            /** This member enables the linking of help QStrings with a certain widget. This
              * is used to create context senistive help              */
            QMap<QWidget*, QString> mWidgetInfo;

            /** The undo/redo stack used to contain all changes that can be undone/redone
              */
//            PEcgSimUndoStack* mUndoStack;

#ifndef QT_NO_DEBUG
//            QUndoView* mUndoView;
#endif

//            QLabel *mStatusLabel;

            /** @brief the number of files shown in the recent file list in the file menu
              */
            int mMaxRecentFiles;

            QList<QAction*> mRecentFileActs;

            QAction *mRecentFileSeparator;

            int mMaxUndoSteps;

            /** @brief flag that is initialy (on startup and after reset/load) false to indicate an autoscale is needed
              */
            bool mInitialScale;

            PPVView     * mPVView;

            PSignalView * mSignalView;

            PParamView  * mParamView;

//            PSystemView * mSystemView;

            PValueView  * mValuesView;
        };
    }
}

#endif // _PCIRCADAPTMAINWINDOW_H
