/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: PCircAdaptOptionsDialog.cpp 791 2013-07-18 14:44:15Z eelco $
 *
 */

#include <QIcon>
#include <QPixmap>
#include <QPushButton>
#include <QVBoxLayout>
#include <QColorDialog>
#include <QFontDialog>
#include <QLabel>
#include <QComboBox>
#include <QButtonGroup>
#include <QRadioButton>
#include <QButtonGroup>
#include <QEvent>

#include <QDebug>

#include "PCircAdaptOptionsDialog.h"
#include "pcircadapt.h"
#include "PCircAdaptDefines.h"
#include "PSettings.h"


using namespace Peacs::CircAdapt;
using namespace Peacs::Support::Widget;
//using namespace Peacs::Support;


const int heartOptions                              = 0;
const int thoraxOptions                             = 1;
const int APOptions                                 = 2;
const int ECGOptions                                = 3;

const int COLOR_BTN_PIXMAP_WIDTH                    = 100;
const int COLOR_BTN_PIXMAP_HEIGHT                   = 50;

const int BACKGROUNDCOLOR_INDEX                     = 0;
const int BACKGROUNDCOLOR_GRAPH_INDEX               = 1;

const char * POptionsDialog::PARAM_TAB_ID_GEN       = "General";
const char * POptionsDialog::PARAM_TAB_ID_VENTR     = "Ventricles";
const char * POptionsDialog::PARAM_TAB_ID_ATRIA     = "Atria";
const char * POptionsDialog::PARAM_TAB_ID_VALVES    = "Valves";
const char * POptionsDialog::PARAM_TAB_ID_VASC      = "Vasculature";
const char * POptionsDialog::PARAM_TAB_ID_SHUNT     = "Shunt";
const char * POptionsDialog::PARAM_TAB_ID_ADAPT     = "Adaptation";


POptionsDialog::POptionsDialog( QWidget * parent ) :
    QDialog( parent ),
    ui(),
    mGridType( GRIDTYPE_DEFAULT ),
    mSignalFonts(),
    mSignalColors(),
    mDirty( false ),
    mChangedItems( NOTHING_CHANGED )
{
    init();

    readCircAdaptSettings();
}


void POptionsDialog::changeEvent(QEvent *e)
{
    switch (e->type())
    {
    case QEvent::LanguageChange:
        ui.retranslateUi(this);
        break;
    default:
        break;
    }
}


//*****************************************************


void POptionsDialog::init( void )
{
    QPixmap backgroundPxMap( COLOR_BTN_PIXMAP_WIDTH, COLOR_BTN_PIXMAP_HEIGHT );

    ui.setupUi( this );

    ui.toolBox->setCurrentIndex( 0 );

    ui.buttonBox->button( QDialogButtonBox::Cancel )->setDefault( true );
    ui.buttonBox->button( QDialogButtonBox::Apply )->setEnabled( false );

    ui.cbGrid->addItem( tr( "No grid" ), int(NO_GRID) );
    ui.cbGrid->addItem( tr( "Basic grid" ), int(SHOW_GRID) );
    ui.cbGrid->addItem( tr( "Full grid" ), int(SHOW_FULL_GRID) );

    // Add the signal font items
    ui.signalsFontsCombo->addItem( tr("Labels"), 0 );
    ui.signalsFontsCombo->addItem( tr("Scale"),  1 );
    ui.signalsFontsCombo->addItem( tr("Text"),   2 );

    mSignalFonts.append( SIGNAL_FONT_LABEL_DEFAULT );
    mSignalFonts.append( SIGNAL_FONT_SCALE_DEFAULT );
    mSignalFonts.append( SIGNAL_FONT_TEXT_DEFAULT );

    // Initialyze the signal color information and buttons
    backgroundPxMap.fill( DEFAULTBACKGROUNDCOLOR );
    ui.signalBackgroundColorBtn->setIcon( QIcon( backgroundPxMap ) );
    mSignalColors.append( DEFAULTBACKGROUNDCOLOR );                     // i.e. BACKGROUNDCOLOR_INDEX
    ui.signalColorChangeBg->setId( ui.signalBackgroundColorBtn, 0 );    // i.e. BACKGROUNDCOLOR_INDEX
    backgroundPxMap.fill( DEFAULTBACKGROUNDCOLORGRAPH );
    ui.signalBackgroundGraphColorBtn->setIcon( QIcon( backgroundPxMap ) );
    mSignalColors.append( DEFAULTBACKGROUNDCOLORGRAPH );                    // i.e. BACKGROUNDCOLOR_2_INDEX
    ui.signalColorChangeBg->setId( ui.signalBackgroundGraphColorBtn, 1 );   // i.e. BACKGROUNDCOLOR_2_INDEX

    QList<QAbstractButton*> selectBtns = ui.paramTabSelectChkBg->buttons();
    for ( int index = 0; index < selectBtns.count(); ++index )
    {
        // ID 0 is not used because that would be the 'General' tab and that tab is always visible
        ui.paramTabSelectChkBg->setId( selectBtns.at( index ), index + 1 ); // Initialyze the button ID's
        selectBtns.at( index )->setChecked( true );
    }
    if ( 0 < ui.paramTabSelectRadioBg->buttons().count() )
    {
        selectBtns = ui.paramTabSelectRadioBg->buttons();

        selectBtns.at( 0 )->setChecked( true );

        // Initialyze the button ID's
        for ( int index = 0; index < selectBtns.count(); ++index )
        {
            ui.paramTabSelectRadioBg->setId( selectBtns.at( index ), index );
        }
    }
}


void POptionsDialog::restoreDefaults( void )
{
    if ( mGridType != GRIDTYPE_DEFAULT )
    {
        mGridType       =  GRIDTYPE_DEFAULT;

        mChangedItems   |= GRID_CHANGED;

        ui.cbGrid->setCurrentIndex( (int)mGridType );

        setDirty( true );
    }

    if ( (mSignalFonts[0] != SIGNAL_FONT_LABEL_DEFAULT) ||
         (mSignalFonts[1] != SIGNAL_FONT_SCALE_DEFAULT) ||
         (mSignalFonts[2] != SIGNAL_FONT_TEXT_DEFAULT) )
    {
        mSignalFonts[0]     = SIGNAL_FONT_LABEL_DEFAULT;
        mSignalFonts[1]     = SIGNAL_FONT_SCALE_DEFAULT;
        mSignalFonts[2]     = SIGNAL_FONT_TEXT_DEFAULT;

        mChangedItems       |= SIGNAL_FONT_CHANGED;

        setSignalsFontDisplayText();

        setDirty( true );
    }

    if ( (mSignalColors[BACKGROUNDCOLOR_INDEX]      != DEFAULTBACKGROUNDCOLOR) ||
         (mSignalColors[BACKGROUNDCOLOR_GRAPH_INDEX]    != DEFAULTBACKGROUNDCOLORGRAPH) )
    {
        QPixmap backgroundPxMap( COLOR_BTN_PIXMAP_WIDTH, COLOR_BTN_PIXMAP_HEIGHT );

        backgroundPxMap.fill( DEFAULTBACKGROUNDCOLOR );
        ui.signalBackgroundColorBtn->setIcon( QIcon( backgroundPxMap ) );
        mSignalColors[BACKGROUNDCOLOR_INDEX] = DEFAULTBACKGROUNDCOLOR;

        backgroundPxMap.fill( DEFAULTBACKGROUNDCOLORGRAPH );
        ui.signalBackgroundGraphColorBtn->setIcon( QIcon( backgroundPxMap ) );
        mSignalColors[BACKGROUNDCOLOR_GRAPH_INDEX] = DEFAULTBACKGROUNDCOLORGRAPH;

        mChangedItems       |= SIGNAL_COLOR_CHANGED;

        setDirty( true );
    }

    QList<QAbstractButton*> selectBtns = ui.paramTabSelectChkBg->buttons();
    for ( int index = 0; index < selectBtns.count(); ++index )
    {
        if ( !selectBtns.at( index )->isChecked() )
        {
            mChangedItems   |= PARAM_TABS_CHANGED;
            selectBtns.at( index )->setChecked( true );

            setDirty( true );
        }

        // There is one more radio button then there are check bockes. The general tab cannot be disabled
        ui.paramTabSelectRadioBg->buttons().at( index + 1 )->setEnabled( selectBtns.at( index )->isChecked() );
    }
    if ( 0 != ui.paramTabSelectRadioBg->checkedId() )
    {
        mChangedItems   |= PARAM_TABS_CHANGED;
        ui.paramTabSelectRadioBg->buttons().at( 0 )->setChecked( true );

        setDirty( true );
    }

    //writeCircAdaptSettings();
}


void POptionsDialog::writeCircAdaptSettings(void)
{
    Support::PSettings      pSetting;
    QSettings &             settings = pSetting.getSettings();
    QString                 paramTabSelectedIdnt( PARAM_TAB_ID_GEN );
    QList<QAbstractButton*> paramTabSelectBtns = ui.paramTabSelectChkBg->buttons();

    settings.setValue( GRIDTYPE, (int)mGridType );

    if ( mSignalFonts.count() == 3 )
    {
        settings.setValue( SIGNAL_FONT_LABEL, mSignalFonts[0] );
        settings.setValue( SIGNAL_FONT_SCALE, mSignalFonts[1] );
        settings.setValue( SIGNAL_FONT_TEXT,  mSignalFonts[2] );
    }

    switch ( ui.paramTabSelectRadioBg->checkedId() )
    {
    case 0:
        // Default value used
        break;
    case 1:
        paramTabSelectedIdnt = PARAM_TAB_ID_VENTR;
        break;
    case 2:
        paramTabSelectedIdnt = PARAM_TAB_ID_ATRIA;
        break;
    case 3:
        paramTabSelectedIdnt = PARAM_TAB_ID_VALVES;
        break;
    case 4:
        paramTabSelectedIdnt = PARAM_TAB_ID_VASC;
        break;
    case 5:
        paramTabSelectedIdnt = PARAM_TAB_ID_SHUNT;
        break;
    case 6:
        paramTabSelectedIdnt = PARAM_TAB_ID_ADAPT;
        break;
    default:
        Q_ASSERT( false );
        break;
    }

    // Store the signal solors
    settings.setValue( SIGNAL_COLOR_BG, mSignalColors[BACKGROUNDCOLOR_INDEX] );
    settings.setValue( SIGNAL_COLOR_BG_GRAPH, mSignalColors[BACKGROUNDCOLOR_GRAPH_INDEX] );

    // Store the parameter view default tab
    settings.setValue( PARAMETER_TABS_SELECTED, paramTabSelectedIdnt );

    Q_ASSERT( 6 == paramTabSelectBtns.count() );

    //settings.setValue( PARAMETER_TABS_GEN,      paramTabSelectBtns[0]->isChecked() ); Not used
    settings.setValue( PARAMETER_TABS_VENTR,    paramTabSelectBtns[0]->isChecked() );
    settings.setValue( PARAMETER_TABS_ATRIA,    paramTabSelectBtns[1]->isChecked() );
    settings.setValue( PARAMETER_TABS_VALVES,   paramTabSelectBtns[2]->isChecked() );
    settings.setValue( PARAMETER_TABS_VASC,     paramTabSelectBtns[3]->isChecked() );
    settings.setValue( PARAMETER_TABS_SHUNT,    paramTabSelectBtns[4]->isChecked() );
    settings.setValue( PARAMETER_TABS_ADAPT,    paramTabSelectBtns[5]->isChecked() );

    settings.sync();

    setDirty( false );
}


void POptionsDialog::readCircAdaptSettings(void)
{
    bool                    ok;
    Support::PSettings      pSetting;
    QSettings &             settings            = pSetting.getSettings();
    int                     tmpI                = settings.value( GRIDTYPE, GRIDTYPE_DEFAULT ).toInt( &ok );
    QString                 valStr;
    bool                    valB;
    QList<QAbstractButton*> paramTabSelectBtns  = ui.paramTabSelectChkBg->buttons();
    QList<QAbstractButton*> paramTabVisibleBtns = ui.paramTabSelectRadioBg->buttons();

    if ( ok )
    {
        if ( (int(NO_GRID) <= tmpI) && (int(SHOW_FULL_GRID) >= tmpI) )
        {
            mGridType = (Grid_type)tmpI;

            ui.cbGrid->setCurrentIndex( (int)mGridType );
        }
    }

    if ( mSignalFonts.count() == 3 )
    {
        mSignalFonts[0] = settings.value( SIGNAL_FONT_LABEL, SIGNAL_FONT_LABEL_DEFAULT ).toString();
        mSignalFonts[1] = settings.value( SIGNAL_FONT_SCALE, SIGNAL_FONT_SCALE_DEFAULT ).toString();
        mSignalFonts[2] = settings.value( SIGNAL_FONT_TEXT,  SIGNAL_FONT_TEXT_DEFAULT ).toString();

        setSignalsFontDisplayText();
    }

    // Retrieve the signal solors
    QPixmap backgroundPxMap( COLOR_BTN_PIXMAP_WIDTH, COLOR_BTN_PIXMAP_HEIGHT );

    mSignalColors[BACKGROUNDCOLOR_INDEX]    = settings.value( SIGNAL_COLOR_BG, DEFAULTBACKGROUNDCOLOR ).value<QColor>();
    backgroundPxMap.fill( mSignalColors[BACKGROUNDCOLOR_INDEX] );
    ui.signalBackgroundColorBtn->setIcon( QIcon( backgroundPxMap ) );
    mSignalColors[BACKGROUNDCOLOR_GRAPH_INDEX]  = settings.value( SIGNAL_COLOR_BG_GRAPH, DEFAULTBACKGROUNDCOLORGRAPH ).value<QColor>();
    backgroundPxMap.fill( mSignalColors[BACKGROUNDCOLOR_GRAPH_INDEX] );
    ui.signalBackgroundGraphColorBtn->setIcon( QIcon( backgroundPxMap ) );

    // Retrieve the parameter view default tab
    if ( settings.contains( PARAMETER_TABS_SELECTED ) )
    {
        Q_ASSERT( 7 == paramTabVisibleBtns.count() );

        valStr = settings.value( PARAMETER_TABS_SELECTED, PARAM_TAB_ID_GEN ).toString();
        if ( valStr == PARAM_TAB_ID_VENTR )
        {
            paramTabVisibleBtns.at( 1 )->setChecked( true );
        }
        else if ( valStr == PARAM_TAB_ID_ATRIA )
        {
            paramTabVisibleBtns.at( 2 )->setChecked( true );
        }
        else if ( valStr == PARAM_TAB_ID_VALVES )
        {
            paramTabVisibleBtns.at( 3 )->setChecked( true );
        }
        else if ( valStr == PARAM_TAB_ID_VASC )
        {
            paramTabVisibleBtns.at( 4 )->setChecked( true );
        }
        else if ( valStr == PARAM_TAB_ID_SHUNT )
        {
            paramTabVisibleBtns.at( 5 )->setChecked( true );
        }
        else if ( valStr == PARAM_TAB_ID_ADAPT )
        {
            paramTabVisibleBtns.at( 6 )->setChecked( true );
        }
        else
        {
            paramTabVisibleBtns.at( 0 )->setChecked( true );
        }
    }

    Q_ASSERT( 6 == paramTabSelectBtns.count() );

    //valB = settings.value( PARAMETER_TABS_GEN, true ).toBool();
    //paramTabSelectBtns[0]->setChecked( valB );
    valB = settings.value( PARAMETER_TABS_VENTR, true ).toBool();
    paramTabSelectBtns[0]->setChecked( valB );
    paramTabVisibleBtns.at( 1 )->setEnabled( valB );    // There is one more radio button then there are check bockes. The general tab cannot be disabled
    valB = settings.value( PARAMETER_TABS_ATRIA, true ).toBool();
    paramTabSelectBtns[1]->setChecked( valB );
    paramTabVisibleBtns.at( 2 )->setEnabled( valB );
    valB = settings.value( PARAMETER_TABS_VALVES, true ).toBool();
    paramTabSelectBtns[2]->setChecked( valB );
    paramTabVisibleBtns.at( 3 )->setEnabled( valB );
    valB = settings.value( PARAMETER_TABS_VASC, true ).toBool();
    paramTabSelectBtns[3]->setChecked( valB );
    paramTabVisibleBtns.at( 4 )->setEnabled( valB );
    valB = settings.value( PARAMETER_TABS_SHUNT, true ).toBool();
    paramTabSelectBtns[4]->setChecked( valB );
    paramTabVisibleBtns.at( 5 )->setEnabled( valB );
    valB = settings.value( PARAMETER_TABS_ADAPT, true ).toBool();
    paramTabSelectBtns[5]->setChecked( valB );
    paramTabVisibleBtns.at( 6 )->setEnabled( valB );

    setDirty( false );
}


void Peacs::CircAdapt::POptionsDialog::setDirty( bool dirty )
{
    if ( mDirty != dirty )
    {
        mDirty = dirty;
    }

    ui.buttonBox->button( QDialogButtonBox::Apply )->setEnabled( mDirty );
}


void POptionsDialog::setSignalsFontDisplayText( void )
{
    if ( mSignalFonts.count() == 3 )
    {
        QFont   fnt;
        int     displaySize = ui.signalsFontDisplay->font().pointSize();

        fnt.fromString( mSignalFonts[ ui.signalsFontsCombo->currentIndex() ] );

        int     newItemSize = fnt.pointSize();
        QString sizeText    = QString( tr( ", size: %1" ) ).arg( newItemSize );

        fnt.setPointSize( displaySize );

        ui.signalsFontDisplay->setFont( fnt );

        ui.signalsFontDisplay->setText( fnt.family() + sizeText );
    }
}


void POptionsDialog::on_buttonBox_clicked( QAbstractButton * button )
{
    switch ( ui.buttonBox->buttonRole( button ) )
    {
    case QDialogButtonBox::AcceptRole:
        accept();

        if ( mDirty )
        {
            writeCircAdaptSettings();

            emit( applySettings( mChangedItems ) );

            mChangedItems = NOTHING_CHANGED;
        }

        close();
        break;

    case QDialogButtonBox::ApplyRole:
        if ( mDirty )
        {
            writeCircAdaptSettings();

            emit( applySettings( mChangedItems ) );

            mChangedItems = NOTHING_CHANGED;
        }
        break;

    case QDialogButtonBox::ResetRole:
        restoreDefaults();

        writeCircAdaptSettings();

        emit( applySettings( mChangedItems ) );

        mChangedItems = NOTHING_CHANGED;
        break;

    default:
        reject();
        close();
        break;
    }
}


void POptionsDialog::on_cbGrid_currentIndexChanged( int /*index*/ )
{
    if ( mGridType != Grid_type(ui.cbGrid->currentIndex()) )
    {
        mGridType       = Grid_type( ui.cbGrid->currentIndex() );
        mChangedItems   |= GRID_CHANGED;

        setDirty( true );
    }
}


void POptionsDialog::on_signalsFontsCombo_currentIndexChanged( int /*index*/ )
{
    if ( mSignalFonts.count() == 3 )
    {
        setSignalsFontDisplayText();
    }
}


void POptionsDialog::on_SignalsFontSelect_clicked ( bool /*checked*/ )
{
    bool        ok;
    QFont       fnt;

    fnt.fromString( mSignalFonts[ ui.signalsFontsCombo->currentIndex() ] );

    fnt = QFontDialog::getFont( &ok, fnt, this );

    if ( ok && (mSignalFonts[ ui.signalsFontsCombo->currentIndex() ] != fnt.toString()) )
    {
        mSignalFonts[ ui.signalsFontsCombo->currentIndex() ] = fnt.toString();

        setSignalsFontDisplayText();

        mChangedItems |= SIGNAL_FONT_CHANGED;

        setDirty( true );

        // TODO Update the signal view
    }
}


void POptionsDialog::on_paramTabSelectChkBg_buttonClicked ( int id )
{
    QList<QAbstractButton*> selectBtns = ui.paramTabSelectRadioBg->buttons();

//    for ( int index = 0; index < selectBtns.count(); ++index )
//    {
//        selectBtns[ index ]->setEnabled( true );
//    }

    selectBtns[ id ]->setEnabled( ui.paramTabSelectChkBg->button( id )->isChecked() );

    // If the corresponding selection radio button is selected set it to the default first radion button
    if ( ui.paramTabSelectRadioBg->checkedId() == id )
    {
        selectBtns[ 0 ]->setChecked( true );
    }

    mChangedItems |= PARAM_TABS_CHANGED;

    setDirty( true );
}


void POptionsDialog::on_paramTabSelectRadioBg_buttonClicked ( int /*id*/ )
{
    mChangedItems |= PARAM_TABS_CHANGED;

    setDirty( true );
}


void POptionsDialog::on_signalColorChangeBg_buttonClicked ( int id )
{
    Q_ASSERT( (0 <= id) && (id < mSignalColors.size()) );

    // Make sure that the user can select the default colors (not on the Mac)
    QColorDialog::setCustomColor( 0, DEFAULTBACKGROUNDCOLOR.rgba() );
    QColorDialog::setCustomColor( 1, DEFAULTBACKGROUNDCOLORGRAPH.rgba() );

    QString captions[]  = { tr("background color"),
                            tr("graph background color") };
    QColor  color       = QColorDialog::getColor( mSignalColors.at( id ), this, tr("Choose ") + captions[id] );

    if ( color.isValid() )
    {
        QPixmap backgroundPxMap( COLOR_BTN_PIXMAP_WIDTH, COLOR_BTN_PIXMAP_HEIGHT );

        backgroundPxMap.fill( color );
        switch ( id )
        {
        case 0:
            ui.signalBackgroundColorBtn->setIcon( QIcon( backgroundPxMap ) );
            break;
        case 1:
            ui.signalBackgroundGraphColorBtn->setIcon( QIcon( backgroundPxMap ) );
            break;
        default:
            Q_ASSERT( false );
            break;
        }

        mSignalColors[id]   = color;

        mChangedItems       |= SIGNAL_COLOR_CHANGED;

        setDirty( true );
    }
}
