/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012-2014 StITPro (www.stitpro.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: pcircadapt.cpp 1176 2014-03-21 15:43:24Z eelco $
 *
 */


#include <QtCore>
#include <QFileOpenEvent>

#include "pcircadapt.h"
#include "PCircAdaptDefines.h"


using namespace Peacs::CircAdapt;


PCircAdapt::PCircAdapt( int & argc, char ** argv ) :
    QApplication( argc, argv ),
#ifdef Q_WS_MAC
    mEventListener(),
#endif
    mSettingObjPtr( NULL ),
    mMainWindowObjPtr( NULL )
{
    QString filename;

    // Required for the settings file
    QCoreApplication::setOrganizationName( "StITPro" );
    QCoreApplication::setOrganizationDomain( "stitpro.nl" );
    QCoreApplication::setApplicationName( "CircAdapt" );

#ifdef CIRCADAPT_VERSION_STR
    QCoreApplication::setApplicationVersion( CIRCADAPT_VERSION_STR );
#else
    #pragma message( "Unable to set the application version in: " __FILE__ )
#endif

    // All settings are now initialized. Now we can create the central settings object
    mSettingObjPtr = new Support::PSettings( CASE_FILE_EXT, this );

    if ( mSettingObjPtr )
    {
        if ( !mSettingObjPtr->getFileName().isEmpty() )
        {
            filename = mSettingObjPtr->getFileName();
        }

#ifdef Q_WS_MAC
        // Needed to get the filename to open under OSx when the file is to be opened due to
        // double clicking on it
        // Under OSx a window application that is opened by double clicking on a, for this application, registered file
        // the file name to open isn't added to the commandline. The application is told by the operating system to
        // open the file by sending a signal to the application. The eventlistener is installed to get the filename to
        // open. Before the event can be recieved the events must be processed.
        // The commandline is leading
        if ( mSettingObjPtr->getFileName().isEmpty() )
        {
            installEventFilter( &mEventListener );

            processEvents();    // The event could already be waiting to be handled. Do some processing to check

            if ( !mEventListener.fileName().isEmpty() )
            {
                filename = mEventListener.fileName();
            }
        }
#endif

        // Now create the main window
        mMainWindowObjPtr = new PCircAdaptMainwindow( /*filename*/ );

        if ( mMainWindowObjPtr )
        {
#ifdef Q_WS_MAC
            // Make sure that the event information can be handled propperly
            mEventListener.setMainWin( mMainWindowObjPtr );
#endif

            mMainWindowObjPtr->setupUi();

            mMainWindowObjPtr->loadDefaultCase( filename );

            mMainWindowObjPtr->show();

            processEvents();

            mMainWindowObjPtr->startSimulation();
        }
    }

    if ( !mSettingObjPtr || !mMainWindowObjPtr )
    {
        if ( mSettingObjPtr )
        {
            delete mSettingObjPtr;
            mSettingObjPtr = NULL;
        }

        throw "Unable to instantiate the application";
    }
}


PCircAdapt::~PCircAdapt( void )
{
#ifdef Q_WS_MAC
    removeEventFilter( &mEventListener );
#endif

    delete mSettingObjPtr;
    mSettingObjPtr = NULL;

    delete mMainWindowObjPtr;
    mMainWindowObjPtr = NULL;
}


bool PCircAdapt::notify( QObject * receiver, QEvent * event )
{
    try
    {
        return QApplication::notify( receiver, event );
    }
    catch( QtConcurrent::Exception & var )
    {
        qFatal( "Error %s sending event %i to object %s (%s)",
                var.what(),
                event->type(),
                qPrintable( receiver->objectName() ),
                (receiver->metaObject() ? receiver->metaObject()->className() : "[no QObject derived]") );
    }
    catch ( std::exception & e )
    {
        qFatal( "Error %s sending event %i to object %s (%s)",
                e.what(),
                event->type(),
                qPrintable( receiver->objectName() ),
                (receiver->metaObject() ? receiver->metaObject()->className() : "[no QObject derived]") );
    }
    catch( const char * str )
    {
        qFatal( "Error %s sending event %i to object %s (%s)",
                str,
                event->type(),
                qPrintable( receiver->objectName() ),
                (receiver->metaObject() ? receiver->metaObject()->className() : "[no QObject derived]") );
    }
    catch ( ... )
    {
        qFatal( "Error <unknown> sending event %i to object %s (%s)",
                event->type(),
                qPrintable( receiver->objectName() ),
                (receiver->metaObject() ? receiver->metaObject()->className() : "[no QObject derived]") );
    }

    // qFatal aborts, so this isn't really necessary
    // but you might continue if you use a different logging lib
    return false;
}


#ifdef Q_WS_MAC
bool PEventListener::eventFilter(QObject* obj, QEvent* event)
{
    if ( event->type() == QEvent::FileOpen )
    {
        QString filename = static_cast<QFileOpenEvent*>(event)->file();

        if ( mMainWindowPtr ) {
            // Application is running and the window is already showing. Open the file
            if ( QApplication::activeModalWidget() == NULL )
                mMainWindowPtr->loadDefaultCase( filename );
            else
                return true; // ignore file open events while a modal dialog is displayed
        }
        else {
            // Application is running but the window isn't showing yet. store the name of
            // the file so it can be given to the window class to open.
            mFileName = filename;
        }
    }
    else
    {
        // Give it on to the standard event processing
        QObject::eventFilter(obj, event);
    }

    return false;
}
#endif
