/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012-2014 StITPro (www.stitpro.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: pcircadapt.h 1176 2014-03-21 15:43:24Z eelco $
 *
 */


#ifndef __PCIRCADAPT_H
#define __PCIRCADAPT_H

#include <QApplication>

#include "PCircAdaptMainwindow.h"
#include "PSettings.h"


/** @namespace peacs All code is part of the peacs software
 *  @namespace CircAdapt CircAdapt specific classes are member if this namespace
 *  @namespace Support Peacs support classes with general functionality
 * */
namespace Peacs
{
    namespace CircAdapt
    {
    // The one and only global pointer to the ECGsim application object
    class PCircAdapt;
    #if defined( pAppPtr )
    #undef pAppPtr
    #endif
    #define pAppPtr ( static_cast<Peacs::CircAdapt::PCircAdapt *>( QCoreApplication::instance() ) )

#ifdef Q_WS_MAC
        /** @class PEventListener pecgsim.h "ECGsim\pecgsim.h"
          * @brief Event listener class needed to get the name of the double clicked file to open
          * under MAC OSx
          *
          * Under MAC OSx a window application that is opened by double clicking on a, for this
          * application, registered file, the file name to open isn't added to the commandline.
          * The application is told by the operating system to open the file by sending a signal
          * to the application. The eventlistener is installed to get the filename to open.
          *
          * @attention Only used when creating a window application for MAC OSx! Otherwise the
          * commandline is used
          */
        class PEventListener : public QObject
        {
            Q_OBJECT

        public:
            /** @brief The constructor
              *
              * @param pMainWindow A pointer to the main window. This window will be told what file to
              * open when a user double clicks on a registered file and the event is caught.
              */
            PEventListener() : QObject(), mFileName(), mMainWindowPtr( NULL ) { ; }

            /** @brief Returns mFileName
              *
              * Return the name of the file in case that the event was caught and the window isn't
              * showing yet. Otherwise an empty string is returned
              *
              * @return the name of the file to open or an empty string
              */
            inline const QString& fileName() { return mFileName; }

            /** @brief Sets mMainWindowPtr
              *
              * Sets the pointer to the window that must recieve the name of the file to open.
              */
            inline void setMainWin( PCircAdaptMainwindow* mainWin ) { mMainWindowPtr = mainWin; }

        private:

            /** The name of the file to open
              */
            QString             mFileName;

            /** A pointer to the window that must recieve the name of the file to open.
              */
            PCircAdaptMainwindow*  mMainWindowPtr;

        protected:

            /** @brief The event filter function
              *
              * This function is called by the framework when an event is caught. This function filters
              * for the {@link QEvent::FileOpen} event
              *
              * @param obj The object the event was for
              * @param event The event caught
              *
              * @return true if the event was handled, false if the event must be processed further
              */
            bool eventFilter( QObject * obj, QEvent * event );
        };
#endif

        /** @class PCircAdapt PCircAdapt.h "CircAdapt\PCircAdapt.h"
          * @brief The application. In this class the variables are set for the settings file
          */
        class PCircAdapt : public QApplication
        {
            Q_OBJECT

        public:

            /** @brief constructor
              * Here the application settings variables are set and the command-line is checked for loading
              * the setting from an ini-file
              *
              * @param argc number of input parameters
              * @param argv list of input parameters
              */
            PCircAdapt( int & argc, char ** argv );


            /** @brief Destructor
              */
            ~PCircAdapt( void );


            /** @brief Returns the one and only settings object for this application
              * Use this object to retrieve command line parameters specified during startup and
              * retrieve or set settings from a ini file or system registry (Windows and OSx)
              */
            Support::PSettings & settings( void ) { Q_ASSERT(NULL != mSettingObjPtr); return *mSettingObjPtr; }


            /** @brief Returns the one and only Qt settings object for this application
              * Use this object to retrieve or set settings from a ini file or system registry
              * (Windows and OSx)
              */
            QSettings & getSettings( void ) { Q_ASSERT(NULL != mSettingObjPtr); return mSettingObjPtr->getSettings(); }


            /** @brief Returns the const reference CircAdapt data store object
              * This object is used throughout the whole application as the central data store
              *
              * @return The const UCirculation storage object
              */
            inline const UCirculation & getCircAdaptDataConst( void ) const { Q_ASSERT(NULL != mMainWindowObjPtr); return mMainWindowObjPtr->getCircAdaptDataConst(); }


            /** @brief Returns the changeable reference CircAdapt data store object
              * This object is used throughout the whole application as the central data store
              *
              * @return The UCirculation storage object
              */
            inline UCirculation & getCircAdaptData( void ) { Q_ASSERT(NULL != mMainWindowObjPtr); return mMainWindowObjPtr->getCircAdaptData(); }


            /** @brief Sends event to receiver: receiver->event(event)
              * This function is overridden from the base implementation to ensure exception handling for all
              * events processed by this application
              *
              * @return Returns the value that is returned from the receiver's event handler.
              * @note Note that this function is called for all events sent to any object in any thread.
              * @see QCoreApplication::notify( QObject * , QEvent * )
              */
            bool notify( QObject * receiver, QEvent * event );


        private:

#ifdef Q_WS_MAC
            /** @brief Instantiation of the event listener
              */
            PEventListener mEventListener;
#endif


            /** @brief Contains the applications settings object
              * This object can be used to store or retrieve application settings. This
              * class holds a static pointer to a QSettings object that will be used to
              * read/store the application settings. Each instantiation of this class
              * returns the same QSettings object.
              */
            Support::PSettings * mSettingObjPtr;


            /** @brief Will contain a pointer to the main window object
              */
            PCircAdaptMainwindow * mMainWindowObjPtr;
        };
    }
}
#endif // __PCIRCADAPT_H
