/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: atria.cpp 789 2013-07-18 13:22:28Z peter $
 *
 */

#include <QDebug>

#include "defs.h"
#include "atria.h"
#include "tatria.h"



// keys for the xml file
const char * MYCHAMBERS_KEY           = "atrial_chambers";

const char * SPECIFICATION_ATRIA_KEY  = "atria_specification";
const char * STATE_ATRIA_KEY          = "atria_state";

const int FILE_ATRIA_VERSION_1  = 1;
const int FILE_ATRIA_VERSION_DEFAULT  = 1;

const char* FILE_ATRIA_NAME = "ATRIA";


const char* LEFT_ATRIA_NAME     = "left_atrium";
const char* RIGHT_ATRIA_NAME    = "right_atrium";

using namespace Peacs::CircAdapt;

// Constructor
SAtria::SAtria() :
    SCircElement("atria"),
    mLeftA(LEFT_ATRIA,  false, LEFT_ATRIA_NAME),
    mRigtA(RIGHT_ATRIA, false, RIGHT_ATRIA_NAME)
{

}


SAtria::SAtria(const QString &name):
    SCircElement(name),
    mLeftA(LEFT_ATRIA,  false, LEFT_ATRIA_NAME),
    mRigtA(RIGHT_ATRIA, false, RIGHT_ATRIA_NAME)
{

}


SAtria::SAtria(const SAtria & atria) :
    SCircElement(atria),
    mLeftA( atria.mLeftA),
    mRigtA( atria.mRigtA)
{

}


void SAtria::init()
{

}


void SAtria::copyUserSpecifications(const TAtria &atria)
{
    mLeftA.copyUserSpecifications( atria.getLeftAtrium() );
    mRigtA.copyUserSpecifications( atria.getRightAtrium() );
}


int SAtria::getDefaultVersion(void) const
{
    return FILE_ATRIA_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SAtria::getClassName(void) const
{
    return FILE_ATRIA_NAME;
}

const SAtria & SAtria::operator =(const SAtria &atria)
{
    if ( this != &atria)
    {
        SCircElement::operator =(atria);
        mLeftA      = atria.mLeftA;
        mRigtA      = atria.mRigtA;
    }
    return *this;
}

void SAtria::adjustForSackPressure(double pressure)
{
    mRigtA.addPressure(pressure);
    // In human: pericardium is half around left atrium
    mLeftA.addPressure(pressure*0.5);

}

// Method to initialize cavity pressure and cross-section
void SAtria::initAtrialMechanics (double tCycleAct, double minActivationTime)
{
    mLeftA.initCavityMechanics(tCycleAct,minActivationTime);
    mRigtA.initCavityMechanics(tCycleAct,minActivationTime);
}

// Method to update cavity pressure and cross-section
void SAtria::updateAtrialMechanics (double deltaT)
{
    mLeftA.updateCavityMechanics(deltaT);
    mRigtA.updateCavityMechanics(deltaT);
}


// Implement method to initialize adaptation variables
void SAtria::initAdaptation()
{
    mLeftA.initAdaptation();
    mRigtA.initAdaptation();
}


// Implement method to update adaptation variables
void SAtria::updateAdaptation()
{
    mLeftA.updateAdaptation();
    mRigtA.updateAdaptation();
}


// Method to perform adaptation
void SAtria::adapt (bool adaptWV, bool adaptSL, bool adaptWA,
                     int nrAdaptUpdates)
{
    mLeftA.adapt(adaptWV,adaptSL,adaptWA,nrAdaptUpdates);
    mRigtA.adapt(adaptWV,adaptSL,adaptWA,nrAdaptUpdates);
}



bool SAtria::writeXml(QDomNode &parentNode) const
{
    bool ok = false;
    QDomNode myNode = createXmlEntry(parentNode,getClassName(),getDefaultVersion(),getName());
    if ( !myNode.isNull())
    {
        ok &= writeSpecification(myNode);
        ok &= writeState(myNode);
        QDomNode chambersNode = createXmlEntry(myNode,MYCHAMBERS_KEY);
        ok &= mLeftA.writeXml(chambersNode);
        ok &= mRigtA.writeXml(chambersNode);
    }
    return ok;
}


bool SAtria::readXml(const QDomNode &node)
{
    bool ok = SCircElement::readXml(node);
    if ( ok )
    {
        QDomNodeList myChambers = node.toElement().elementsByTagName(MYCHAMBERS_KEY);
        if ( myChambers.size() == 1 )
        {
            myChambers = myChambers.at(0).childNodes();
            for (int i=0; i < myChambers.size(); i++ )
            {
                SChamber chamber;
                chamber.readXml( myChambers.at(i) );
                if ( chamber.getChamberType() == LEFT_ATRIA )
                {
                    mLeftA = chamber;
                }
                else if ( chamber.getChamberType() == RIGHT_ATRIA )
                {
                    mRigtA = chamber;
                }
                else
                {
                    ok = false;
                    qDebug().nospace() << "Do not know what the atrial chamber type is";
                }
            }
        }
        else
        {
            ok = false;
        }
    }
    return ok;
}



//bool Atria::writeState (QDomNode &parentNode) const
//{
//    QDomNode stateNode(createXmlEntry(parentNode,STATE_ATRIA_KEY));
//    CircElement::writeState(stateNode);

//    addValue(stateNode, STATE_VOLUME_MIDWALL,       mVm);

//    return !stateNode.isNull();
//}

///// Abstract method to read state from file
//bool Atria::readState(const QDomNode &node)
//{
//    QDomNodeList elements = node.toElement().elementsByTagName(STATE_ATRIA_KEY);
//    if ( elements.count() == 1 )
//    {
//        QDomElement stateNode = elements.at(0).toElement();
//        CircElement::readState(stateNode);
//        mVm     = readValue(stateNode, STATE_VOLUME_MIDWALL);
//        return true;
//    }
//    return false;
//}

