/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: chamber.cpp 789 2013-07-18 13:22:28Z peter $
 *
 */

#include <QDebug>

#include "math.h"
#include "defs.h"
#include "chamber.h"
#include "tchamber.h"

using namespace Peacs::CircAdapt;

// keys for the xml file
const char * MYPATCHES_KEY              = "chamber_patches";

const char * CHAMBER_TYPE             = "chamber_type";
const char * ADAPTATION_IN_EXERCISE     = "adaptation_in_exercise";

const char * SPECIFICATION_CHAMBER_KEY  = "chamber_specification";
const char * STATE_CHAMBER_KEY          = "chamber_state";

const char * STATE_VOLUME_MIDWALL       = "volume_midwall";
const char * STATE_CENTRALBULGE_DIST    = "centralbulgedist";
const char * STATE_MIDWALLRADIUS        = "midwallradius";
const char * STATE_WALLAREA             = "wallarea";
const char * STATE_CURVATURE            = "curvature";
const char * STATE_WALLTENSION          = "walltension";
const char * STATE_PTRANS               = "ptransmural";
const char * STATE_WALLSTIFFNESS        = "wallstiffness";
const char * STATE_Y_TRISEG             = "y_triseg";
const char * STATE_YPRIME_TRISEG        = "yprime_triseg";
const char * STATE_MEANY_TRISEG         = "meany_triseg";


const char * SPECIFICATION_REFERENCEAREA            = "reference_area";
const char * SPECIFICATION_NON_CONTRACTILE_AREA     = "non_contractile_area";
const char * SPECIFICATION_WALLVOLUME               = "wallvolume";

const double DEFAULT_REFERENCEAREA              = 0.0;
const double DEFAULT_NON_CONTRACTILE_AREA       = 0.0;
const double DEFAULT_WALLVOLUME                 = 0.0;

const int FILE_CHAMBER_VERSION_1  = 1;
const int FILE_CHAMBER_VERSION_DEFAULT  = 1;

const char* FILE_CHAMBER_NAME = "CHAMBER";

// Constructor

SChamber::SChamber(bool useTriSeg) :
    SCompliance(),
    mAmRef(0.0),
    mAmDead(0.0),
    mVWall(0.0),
    mVm(0.0),
    mXm(0.0),
    mYm(0.0),
    mSheetAreaMid(0.0),
    mCm(0.0),
    mT(0.0),
    mPtrans(0.0),
    mDTDAm(0.0),
    mY(0.0),
    mYDot(0.0),
    mLRef(0.0),
    mNrUpdates(0),
    mYSum(0.0),
    mEAmRef(0.0),
    mPatches(),
    mChamberType(UNDEFINED_TYPE),
    mAdaptToExercise(false),
    mUseTriSeg(useTriSeg)
{
    init();
}

//SChamber::SChamber(chamber_type type, bool adaptToExercise, bool useTriSeg) :
//    SCompliance(),
//    mAmRef(0.0),
//    mAmDead(0.0),
//    mVWall(0.0),
//    mVm(0.0),
//    mXm(0.0),
//    mYm(0.0),
//    mSheetAreaMid(0.0),
//    mCm(0.0),
//    mT(0.0),
//    mPtrans(0.0),
//    mDTDAm(0.0),
//    mY(0.0),
//    mYDot(0.0),
//    mLRef(0.0),
//    mNrUpdates(0),
//    mYSum(0.0),
//    mEAmRef(0.0),
//    mPatches(),
//    mChamberType(type),
//    mAdaptToExercise(adaptToExercise),
//    mUseTriSeg(useTriSeg)
//{
//    init();
//}

SChamber::SChamber(chamber_type type, bool adaptToExercise, const QString &name, bool useTriSeg) :
    SCompliance(name),
    mAmRef(0.0),
    mAmDead(0.0),
    mVWall(0.0),
    mVm(0.0),
    mXm(0.0),
    mYm(0.0),
    mSheetAreaMid(0.0),
    mCm(0.0),
    mT(0.0),
    mPtrans(0.0),
    mDTDAm(0.0),
    mY(0.0),
    mYDot(0.0),
    mLRef(0.0),
    mNrUpdates(0),
    mYSum(0.0),
    mEAmRef(0.0),
    mPatches(),
    mChamberType(type),
    mAdaptToExercise(adaptToExercise),
    mUseTriSeg(useTriSeg)
{
    init();
}


SChamber::SChamber(const SChamber & chamber):
    SCompliance(chamber),
    mAmRef(chamber.mAmRef),
    mAmDead(chamber.mAmDead),
    mVWall(chamber.mVWall),
    mVm(chamber.mVm),
    mXm(chamber.mXm),
    mYm(chamber.mYm),
    mSheetAreaMid(chamber.mSheetAreaMid),
    mCm(chamber.mCm),
    mT(chamber.mT),
    mPtrans(chamber.mPtrans),
    mDTDAm(chamber.mDTDAm),
    mY(chamber.mY),
    mYDot(chamber.mYDot),
    mLRef(chamber.mLRef),
    mNrUpdates(chamber.mNrUpdates),
    mYSum(chamber.mYSum),
    mEAmRef(chamber.mEAmRef),
    mPatches(chamber.mPatches),
    mChamberType(chamber.mChamberType),
    mAdaptToExercise(chamber.mAdaptToExercise),
    mUseTriSeg(chamber.mUseTriSeg)
{

}

SChamber::~SChamber()
{
    for ( int i=0; i < mPatches.size(); i++)
    {
        delete mPatches[i];
    }
}

void SChamber::init()
{
    SCompliance::init();

    if ( mChamberType == RIGHT_ATRIA )
    {
        setParRightAtrium();
    }
    else if ( mChamberType == LEFT_ATRIA )
    {
        setParLeftAtrium();
    }
    else if ( mChamberType == RIGHT_VENTRICLE )
    {
        setParRightVentricle();
    }
    else if ( mChamberType == SEPTUM )
    {
        setParSeptumVentricle();
    }
    else if ( mChamberType == LEFT_VENTRICLE )
    {
        setParLeftVentricle();
    }
    else
    {
        qDebug() << "undefined chamber";
    }
    addPatch();
}


void SChamber::copyUserSpecifications(const TChamber &chamber)
{
    SCompliance::copyUserSpecifications( chamber );
    mAmRef          = chamber.getAmRef();
    mAmDead         = chamber.getAmDead();
    mVWall          = chamber.getVwall();
    mAdaptToExercise= chamber.isAdapationOn();
    mUseTriSeg      = chamber.isTriSeg();
//    mEAmRef         = chamber.getEAmRef();

    for (int i=0; i < chamber.getPatches().size(); i++)
    {
        mPatches[i]->copyUserSpecifications( *chamber.getPatches().at(i) );
    }

}


bool SChamber::addPatch()
{
    mPatches.append( new SPatch(*this) );
    return true;
}


int SChamber::getDefaultVersion(void) const
{
    return FILE_CHAMBER_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SChamber::getClassName(void) const
{
    return FILE_CHAMBER_NAME;
}

const SChamber & SChamber::operator =(const SChamber &chamber)
{
    if ( this != &chamber)
    {
        SCompliance::operator =(chamber);

        mAmRef              = chamber.mAmRef;
        mAmDead             = chamber.mAmDead;
        mVWall              = chamber.mVWall;
        mVm                 = chamber.mVm;
        mXm                 = chamber.mXm;
        mYm                 = chamber.mYm;
        mSheetAreaMid       = chamber.mSheetAreaMid;
        mCm                 = chamber.mCm;
        mT                  = chamber.mT;
        mPtrans             = chamber.mPtrans;
        mDTDAm              = chamber.mDTDAm;
        mY                  = chamber.mY;
        mYDot               = chamber.mYDot;
        mLRef               = chamber.mLRef;
        mNrUpdates          = chamber.mNrUpdates;
        mYSum               = chamber.mYSum;
        mEAmRef             = chamber.mEAmRef;
        mChamberType        = chamber.mChamberType;
        mAdaptToExercise    = chamber.mAdaptToExercise;
        mUseTriSeg          = chamber.mUseTriSeg;
        for ( int i=0; i < mPatches.size(); i++)
        {
            delete mPatches[i];
        }

        mPatches.clear();

        for (int i=0; i < chamber.mPatches.size(); i++)
        {
            mPatches.append( new SPatch(*this) );
            *mPatches[i] = *chamber.mPatches[i];
        }
    }
    return *this;
}

// Method to initialize cavity pressure and cross-section
void SChamber::initCavityMechanics (double tCycleAct, double minActivationTime)
{
    // Initialize variables for TriSeg computation
    if (mNrUpdates > 0)
    {
        // Mean over Y during former beat
        mLRef = mYSum/mNrUpdates;
    }
    else
    {
        // Default value
        mLRef = LREFDEFAULT;
    }
    mNrUpdates = 0;     // Number of state updates during current beat
    mYSum      = 0.0;   // Sum over Y during current beat

    // Initialize sheet mechanics of all patches
    for ( int i = 0; i < mPatches.size(); i++)
    {
        // Initialize sheet mechanics of patch
        // Set actual cycle time
        mPatches[i]->initSheetMechanics(tCycleAct,minActivationTime);
    }
}


double SChamber::getMinActivationTime(double cycleTime) const
{
    double minTime = 1000;
    for ( int i = 0; i < mPatches.size(); i++)
    {
        minTime = qMin(minTime, mPatches[i]->getActivationTime(cycleTime));
    }
    return minTime;
}


// Private method to update sheet mechanics
void SChamber::updateSheetMechanics (double deltaT)
{
    // Distinguish between single patch and multiple patches
    if ( mPatches.size() == 1 )
    {

        // Update sheet mechanics
        // Area and curvature of sheet are equal to
        // area (Am) and curvature (Cm) of cavity
        mPatches.first()->updateSheetMechanics(mSheetAreaMid,mCm,deltaT);


        // Obtain wall tension from sheet
        mT = mPatches.first()->getWallTension();

        // Obtain transmural pressure from sheet
        mPtrans = mPatches.first()->getTransmuralPressure();

        // Obtain wall stiffness from sheet
        mDTDAm = mPatches.first()->getWallStiffness();
    }
    else
    {
        // Sheet is composed of multiple patches
        // Update sheet mechanics of all patches
        // DEBUG TO DO: Sheet = MultiPatch(Sheet) (multiple patches)
        qDebug().nospace() << "ERROR: Multiple patches not supported!\n";
//        throw Error(ERR_FUNCERROR);

        for ( int i= 0; i < mPatches.size(); i++)
        {
            // Update sheet mechanics of patch
            mPatches[i]->updateSheetMechanics(mSheetAreaMid,mCm,deltaT);
        }
    }
}


// Method to update cavity pressure and cross-section
void SChamber::updateCavityMechanics (double deltaT)
{
    // Declarations
    double  AmTot;     // Mid-wall surface area
    double  Len;       // Estimate of long-axis length

    // Update cavity volume
    mVolume += deltaT * mVDot;

    // Check cavity volume (should not be negative)
    mVolume = max(mVolume,0.0);

    // Update volume enclosed by midwall
    // VWall is wall volume
    mVm = mVolume + 0.5 * mVWall;

    // Update curvature = 1/radius of midwall
    mCm = pow((4.0 * M_PI)/(3.0 * mVm), 1.0/3.0);

    // Compute mid-wall surface area
    AmTot = (4.0 * M_PI)/(mCm * mCm);

    // Update area of wall laid as sheet
    // AmDead is non-contractile area (valve openings)
    mSheetAreaMid = AmTot - mAmDead;

    // Update sheet mechanics
    updateSheetMechanics(deltaT);

    // Estimate long-axis length of cavity
    Len = 2.0 * pow( mVm, 1.0 /3.0 );


    // Crude estimation of cavity cross-section
    mA = (mVolume + 0.1 * mVWall) / Len;

    // Initialize pressure at central cavity
    mP = mPtrans;

    // Initialize pressure at inlet and outlet
    mPin = mP;

    // Initialize d/dt Cavity volume
    mVDot = 0.0;

    // Initialize flow used for diameter adaptation
    mQRemod = 0.0;
}


// Method to update cavity hemodynamics for TriSeg computation
void SChamber::updateCavityHemodynamics()
{
    // Declarations
    double  Len;  // Estimate of long-axis length

    // Estimate long-axis length of cavity
    Len = 2.0 * pow( mVolume + mVWall, 1.0 /3.0 );

    // Crude estimation of cavity cross-section
    mA = (mVolume + 0.1*mVWall) / Len;

    // Impedance of compliance
    mZ = 0.0;

    // Initialize pressure at inlet and outlet
    mPin = mP;

    // Initialize d/dt Cavity volume
    mVDot = 0.0;

    // Initialize flow used for diameter adaptation
    mQRemod = 0.0;
}


// Implement method to initialize adaptation variables
void SChamber::initAdaptation()
{
    SCompliance::initAdaptation();
    // Initialize adaptation variables for all patches
    for ( int i= 0; i < mPatches.size(); i++)
    {
        // Initialize adaptation variables
        mPatches[i]->initAdaptation();
    }
}


// Implement method to update adaptation variables
void SChamber::updateAdaptation()
{
    SCompliance::updateAdaptation();
    // Update adaptation variables for all patches
    for ( int i= 0; i < mPatches.size(); i++)
    {
        // Update adaptation variables
        mPatches[i]->updateAdaptation();
    }
}


// Method to perform adaptation
void SChamber::adapt (bool adaptWV, bool adaptSL, bool adaptWA,
                     int nrAdaptUpdates)
{
    // Consistency check: check validity of adaptation variables
    if (nrAdaptUpdates == 0)
    {
        // Error: adaptation variables not valid
        qDebug().nospace() << "ERROR: Adaptation variables not valid in method call Chamber::adapt()";
//        throw Error(ERR_INVALIDCALL);
    }

    // Perform adaptation for all patches and
    // compute new reference area and wall volume
    mAmRef = 0.0;
    mVWall = 0.0;
    for ( int i= 0; i < mPatches.size(); i++)
    {
        // Perform adaptation for patch
        mPatches[i]->adapt(adaptWV,adaptSL,adaptWA,mEAmRef,nrAdaptUpdates);

        // Update reference area for patch
        mAmRef += mPatches[i]->getRefArea();

        // Update wall volume for patch
        mVWall += mPatches[i]->getWallVolume();
    }

    // Compute non-contractile area (Valve openings)
    // DEBUG QUESTION: HOW TO UPDATE AmDead ???
}


// Private method to compute Am, Cm and Xm from Vm and Ym
void SChamber::computeAmCmXm()
{
    // Declarations
    double SignVm;     // Sign of Vm
    double Qaux, Vaux; // Aux variables
    double x2;         // Xm^2
    double y2;         // Ym^2
    double x2py2;      // Xm^2 + Ym^2

    /*-----------------------------------------------+
 |  Solving 3rd order polynomial                  |
 |  Mathematica: Solve[x^3 + 3y^2x  - 2V == 0, x] |
 |  Q = (V + Sqrt(V^2 + y^6))^(1/3)               |
 |  x = Q - y^2/Q                                 |
 +-----------------------------------------------*/

    // Compute Xm from Vm
    SignVm  = sign(mVm);
    Vaux    = ( 3.0 / M_PI ) * fabs( mVm );
    Qaux    = pow( Vaux + sqrt( Vaux * Vaux + pow(mYm, 6.0)), 1.0 /3.0 );
    mXm     = SignVm * (Qaux - mYm * mYm / Qaux);

    // Compute midwall area Am and curvature Cm=1/rm
    x2     = mXm*mXm;
    y2     = mYm*mYm;
    x2py2  = x2 + y2;
    mSheetAreaMid     = M_PI * x2py2 - mAmDead;
    mCm     = 2.0 * mXm / x2py2;
}



bool SChamber::writeXml(QDomNode &parentNode) const
{
    bool ok = false;
    QDomNode myNode = createXmlEntry(parentNode,getClassName(),getDefaultVersion(),getName());
    if ( !myNode.isNull())
    {
        ok  = addValue(myNode,CHAMBER_TYPE          ,mChamberType);
        ok  = addValue(myNode,ADAPTATION_IN_EXERCISE,mAdaptToExercise);
        ok &= writeSpecification(myNode);
        ok &= writeState(myNode);
        QDomNode patchesNode = createXmlEntry(myNode,MYPATCHES_KEY);
        // Write state of all patches
        for (int i=0; i < mPatches.size(); i++)
        {
            ok &= mPatches[i]->writeXml( patchesNode );
        }
    }
    return ok;
}


bool SChamber::readXml(const QDomNode &node)
{
    mChamberType     = (chamber_type)readInt(node, CHAMBER_TYPE);
    mAdaptToExercise = readBool(node,         ADAPTATION_IN_EXERCISE);

    bool ok = SCircElement::readXml(node);
    if ( ok )
    {        
        QDomNodeList myPatches = node.toElement().elementsByTagName(MYPATCHES_KEY);
        if ( myPatches.size() == 1 )
        {
            // clear the patches first
            mPatches.clear();

            myPatches = myPatches.at(0).childNodes();
            for (int i=0; i < myPatches.size(); i++ )
            {
                mPatches.append( new SPatch ( *this ) );
                mPatches.last()->readXml( myPatches.at(i) );
            }
        }
        else
        {
            ok = false;
        }
    }
    return ok;
}



bool SChamber::writeState (QDomNode &parentNode) const
{
    QDomNode stateNode(createXmlEntry(parentNode,STATE_CHAMBER_KEY));
    SCompliance::writeState(stateNode);

    addValue(stateNode, STATE_VOLUME_MIDWALL,       mVm);
    addValue(stateNode, STATE_CENTRALBULGE_DIST,    mXm);
    addValue(stateNode, STATE_MIDWALLRADIUS,        mYm);
    addValue(stateNode, STATE_WALLAREA,             mSheetAreaMid);
    addValue(stateNode, STATE_CURVATURE,            mCm);
    addValue(stateNode, STATE_WALLTENSION,          mT);
    addValue(stateNode, STATE_PTRANS,               mPtrans);
    addValue(stateNode, STATE_WALLSTIFFNESS,        mDTDAm);
    addValue(stateNode, STATE_Y_TRISEG,             mY);
    addValue(stateNode, STATE_YPRIME_TRISEG,        mYDot);
    addValue(stateNode, STATE_MEANY_TRISEG,         mLRef);

    return !stateNode.isNull();
}

/// Abstract method to read state from file
bool SChamber::readState(const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(STATE_CHAMBER_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement stateNode = elements.at(0).toElement();
        SCompliance::readState(stateNode);
        mVm     = readValue(stateNode, STATE_VOLUME_MIDWALL);
        mXm     = readValue(stateNode, STATE_CENTRALBULGE_DIST);
        mYm     = readValue(stateNode, STATE_MIDWALLRADIUS);
        mSheetAreaMid     = readValue(stateNode, STATE_WALLAREA);
        mCm     = readValue(stateNode, STATE_CURVATURE);
        mT      = readValue(stateNode, STATE_WALLTENSION);
        mPtrans = readValue(stateNode, STATE_PTRANS);
        mDTDAm  = readValue(stateNode, STATE_WALLSTIFFNESS);
        mY      = readValue(stateNode, STATE_Y_TRISEG);
        mYDot   = readValue(stateNode, STATE_YPRIME_TRISEG);
        mLRef   = readValue(stateNode, STATE_MEANY_TRISEG);
        return true;
    }
    return false;
}


// Implement method to write specificatioin to file
bool SChamber::writeSpecification(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_CHAMBER_KEY));
    // Write specification of compliance
    bool ok = SCompliance::writeSpecification(specNode);

    ok &= addValue(specNode, SPECIFICATION_REFERENCEAREA,mAmRef);
    ok &= addValue(specNode, SPECIFICATION_NON_CONTRACTILE_AREA,mAmDead);
    ok &= addValue(specNode, SPECIFICATION_WALLVOLUME,mVWall);
    return ok;
}


// Implement method to read specification from file
bool SChamber::readSpecification(const QDomNode &parentNode)
{
    QDomNodeList elements = parentNode.toElement().elementsByTagName(SPECIFICATION_CHAMBER_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement specNode = elements.at(0).toElement();

        mAmRef      = readValue(specNode, SPECIFICATION_REFERENCEAREA,DEFAULT_REFERENCEAREA);
        mAmDead     = readValue(specNode, SPECIFICATION_NON_CONTRACTILE_AREA,DEFAULT_NON_CONTRACTILE_AREA);
        mVWall      = readValue(specNode, SPECIFICATION_WALLVOLUME,DEFAULT_WALLVOLUME);
        return true;
    }
    return false;
}




// Method to set default parameters for left atrium
void SChamber::setParLeftAtrium()
{
    // Set specification parameters
    mAmRef     = 0.0094;       // Reference area of wall laid as sheet
    mAmDead    = 0.0015;       // Non-contractile area (Valve openings)
    mVWall     = 1.2132e-005;  // Wall volume

    // Set simulation state variables for compliance
    mVolume      =   1.0606e-004;   // Volume of compliance
    mVDot   =   1.6003e-004;   // d/dt Volume
    mA      =   0.0011;        // Cross-section of compliance
    mP      = 734.9096;        // Pressure of compliance
    mZ      =   0.0;           // Impedance
    mPin    = 734.9096;        // Pressure at entrance
    mQRemod =   1.6003e-004;   // Through flow used for diameter adaptation

    // Set simulation state variables for chamber
    mVm     =   0.0;           // Volume enclosed by midwall
    mXm     =   0.0;           // Central bulge distance of midwall surface
    mYm     =   0.0;           // Radius of midwall at the junctional edge
    mSheetAreaMid     =   0.0097;        // Area of wall laid as sheet
    mCm     =  33.4295;        // Curvature = 1/radius of midwall
    mT      =  10.5538;        // Wall tension [N/m]
    mPtrans = 705.6142;        // Transmural pressure
    mDTDAm  =   9.6099e+003;   // dTension/dArea compliance (wall stiffness)
    mY      =   0.0;           // State variable for TriSeg computation
    mYDot   =   0.0;           // d/dt Y
    mLRef   =   0.0;           // Mean over Y during former beat

}


// Method to set default parameters for right atrium
void SChamber::setParRightAtrium()
{
    // Set specification parameters
    mAmRef     = 0.0088;       // Reference area of wall laid as sheet
    mAmDead    = 0.0015;       // Non-contractile area (Valve openings)
    mVWall     = 4.6010e-006;  // Wall volume

    // Set simulation state variables for compliance
    mVolume      =   8.8038e-005;   // Volume of compliance
    mVDot   =   2.0289e-004;   // d/dt Volume
    mA      =   9.8616e-004;   // Cross-section of compliance
    mP      =   230.2423;        // Pressure of compliance
    mZ      =   0.0;           // Impedance
    mPin    =   230.2423;        // Pressure at entrance
    mQRemod =   2.0289e-004;   // Through flow used for diameter adaptation

    // Set simulation state variables for chamber
    mVm     =   0.0;           // Volume enclosed by midwall
    mXm     =   0.0;           // Central bulge distance of midwall surface
    mYm     =   0.0;           // Radius of midwall at the junctional edge
    mSheetAreaMid     =   0.0082;        // Area of wall laid as sheet
    mCm     =  35.9257;        // Curvature = 1/radius of midwall
    mT      =   2.3890;        // Wall tension [N/m]
    mPtrans = 171.6516;        // Transmural pressure
    mDTDAm  =   4.0053e+003;   // dTension/dArea compliance (wall stiffness)
    mY      =   0.0;           // State variable for TriSeg computation
    mYDot   =   0.0;           // d/dt Y
    mLRef   =   0.0;           // Mean over Y during former beat
}


// Method to set default parameters for left ventricle
void SChamber::setParLeftVentricle ()
{
    // Set specification parameters

    if ( mUseTriSeg )
    {
        mAmRef     = 0.0095;       // Reference area of wall laid as sheet (TriSeg)
        mVWall     = 1.0359e-004;  // Wall volume (TriSeg)
    }
    else
    {
        mAmRef     = 0.0095 + 0.0041; // Reference area of wall laid as sheet (NOT TriSeg)
        mVWall     = 1.0359e-004 + 3.6798e-005;  // Wall volume (NOT TriSeg)
    }

    mAmDead    = 0.0015;       // Non-contractile area (Valve openings)

    // Set simulation state variables for compliance
    mVolume      =   1.1031e-004;   // Volume of compliance
    mVDot   =  -3.9084e-005;   // d/dt Volume
    mA      =   0.0010;        // Cross-section of compliance
    mP      = 844.2604;        // Pressure of compliance
    mZ      =   0.0;           // Impedance
    mPin    = 844.2604;        // Pressure at entrance
    mQRemod =  -3.9084e-005;   // Through flow used for diameter adaptation

    // Set simulation state variables for chamber
    mVm     =  -1.5411e-004;   // Volume enclosed by midwall
    mXm     =  -0.0493;        // Central bulge distance of midwall surface
    mYm     =   0.0343;        // Radius of midwall at the junctional edge
    mSheetAreaMid     =   0.0098;        // Area of wall laid as sheet
    mCm     = -27.3279;        // Curvature = 1/radius of midwall
    mT      =  14.3749;        // Wall tension [N/m]
    mPtrans =-785.6697;        // Transmural pressure
    mDTDAm  =   2.1649e+004;   // dTension/dArea compliance (wall stiffness)
    mY      =   0.0;           // State variable for TriSeg computation
    mYDot   =   0.0;           // d/dt Y
    mLRef   =   0.0;           // Mean over Y during former beat
}


// Method to set default parameters for septum of ventricle
void SChamber::setParSeptumVentricle ()
{
    // Set specification parameters
    mAmRef     = 0.0041;       // Reference area of wall laid as sheet
    mAmDead    = 0.0;          // Non-contractile area (Valve openings)

    if ( mUseTriSeg )
        mVWall     = 3.6798e-005;          // Wall volume (TriSeg)
    else
        mVWall     = 0.0;  // Wall volume (NOT TriSeg)

    // Set simulation state variables for compliance
    mVolume     =   2.6555e-005;   // Volume of compliance
    mVDot       =  -1.5395e-005;   // d/dt Volume
    mA          =   0.0;           // Cross-section of compliance
    mP          =   0.0;           // Pressure of compliance
    mZ          =   0.0;           // Impedance
    mPin        =   0.0;           // Pressure at entrance
    mQRemod     =   0.0;           // Through flow used for diameter adaptation

    // Set simulation state variables for chamber
    mVm         =   2.6401e-005;   // Volume enclosed by midwall
    mXm         =   0.0136;        // Central bulge distance of midwall surface
    mYm         =   0.0343;        // Radius of midwall at the junctional edge
    mSheetAreaMid         =   0.0043;        // Area of wall laid as sheet
    mCm         =  19.9584;        // Curvature = 1/radius of midwall
    mT          =  10.8100;        // Wall tension [N/m]
    mPtrans     = 431.4993;        // Transmural pressure
    mDTDAm      =   3.8108e+004;   // dTension/dArea compliance (wall stiffness)
    mY          =   0.0343;        // State variable for TriSeg computation
    mYDot       =  -3.8808e-004;   // d/dt Y
    mLRef       =   0.0329;        // Mean over Y during former beat

}


// Method to set default parameters for right ventricle
void SChamber::setParRightVentricle ()
{
    // Set specification parameters
    mAmRef     = 0.0116;       // Reference area of wall laid as sheet
    mAmDead    = 0.0015;       // Non-contractile area (Valve openings)
    mVWall     = 4.3367e-005;  // Wall volume

    // Set simulation state variables for compliance
    mVolume =   1.1535e-004;   // Volume of compliance
    mVDot   =  -3.2440e-006;   // d/dt Volume
    mA      =   0.0011;        // Cross-section of compliance
    mP      = 414.3753;        // Pressure of compliance
    mZ      =   0.0;           // Impedance
    mPin    = 414.3753;        // Pressure at entrance
    mQRemod =  -3.2440e-006;   // Through flow used for diameter adaptation

    // Set simulation state variables for chamber
    mVm     =   1.8183e-004;   // Volume enclosed by midwall
    mXm     =   0.0540;        // Central bulge distance of midwall surface
    mYm     =   0.0343;        // Radius of midwall at the junctional edge
    mSheetAreaMid     =   0.0113;        // Area of wall laid as sheet
    mCm     =  26.4037;        // Curvature = 1/radius of midwall
    mT      =   6.7374;        // Wall tension [N/m]
    mPtrans = 355.7847;        // Transmural pressure
    mDTDAm  =   8.4051e+003;   // dTension/dArea compliance (wall stiffness)
    mY      =   0.0;           // State variable for TriSeg computation
    mYDot   =   0.0;           // d/dt Y
    mLRef   =   0.0;           // Mean over Y during former beat
}


