/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: chamber.h 789 2013-07-18 13:22:28Z peter $
 *
 */

#ifndef _CHAMBER_H_
#define _CHAMBER_H_

#include "defs.h"
#include "patch.h"
#include "compliance.h"


using namespace Peacs::Support;

namespace Peacs
{

    namespace CircAdapt
    {
        class TChamber;

        /**
        Class for muscular chamber. Inherits from Compliance.
        */
        class SChamber: public SCompliance
        {

        public:
            SChamber(bool useTriSeg = false);

            /* @brief Default Constructor */
//            SChamber(chamber_type type, bool adaptToExercise, bool useTriseg = false);

            /* @brief Constructor */
            SChamber(chamber_type type, bool adaptToExercise, const QString &name, bool useTriseg = false);

            /* @brief Constructor */
            SChamber(const SChamber & chamber);

            virtual ~SChamber();

            virtual const SChamber& operator = (const SChamber &chamber);

            void init();


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TChamber &chamber);


            bool addPatch();


            inline chamber_type getChamberType() const { return mChamberType; }

            inline bool isAdapationOn() const { return mAdaptToExercise; }

            inline bool isVentricle() const { return mChamberType == RIGHT_VENTRICLE || mChamberType == SEPTUM || mChamberType == LEFT_VENTRICLE ; }

            inline bool isTriSeg() const { return mUseTriSeg; }



            inline double getAmRef() const { return mAmRef; }

            inline double getAmDead() const { return mAmDead; }

            inline double getVwall() const { return mVWall; }



            /// Method to get wall volume
            inline double getWallVolume() const { return mVWall; }

            inline double getLRef() const { return mLRef; }

            inline double getY() const { return mY; }

            inline double getAm() const { return mSheetAreaMid; }


            inline double getCm() const { return mCm; }

            inline double getDTDAm() const { return mDTDAm; }

            inline double getEAmRef() const { return mEAmRef; }

            inline double getPtrans() const { return mPtrans; }

            inline double getT() const { return mT; }

            inline double getVm() const { return mVm; }


            inline double getYdot() const { return mYDot; }

            inline double getXm() const { return mXm; }

            inline double getYm() const { return mYm; }

            inline void setEAmRef(double value) { mEAmRef = value; }

            inline void setYdot(double value) { mYDot = value;}

            inline void setYm(double value) { mYm = value;}

            inline void setVm(double value) { mVm = value;}

            inline void addY(double dY) { mY += dY; }

            inline void addVm(double dVm) { mVm += dVm; }

            inline void addYm(double dYm) { mYm += dYm; }


            const QList < SPatch * > &getPatches() const { return mPatches; }



            inline void addYsum(double dYsum) { mYSum += dYsum; }

            inline void addNrUpdates() { mNrUpdates++; }

            /// Method to estimate mitral/tricuspid leak signal
            inline double estimateLeakSignal (double leak_par) const    { return (mT/mSheetAreaMid)*(leak_par * mAmRef*mAmRef / mVWall); }

            /// Method to initialize cavity pressure and cross-section
            virtual void initCavityMechanics (double tCycleAct, double tCycleRef);

            /// Method to update cavity pressure and cross-section
            void updateCavityMechanics (double deltaT);

            /// Method to update cavity hemodynamics for TriSeg computation
            void updateCavityHemodynamics();

            /// Implement method to initialize adaptation variables
            void initAdaptation();

            /// Implement method to update adaptation variables
            void updateAdaptation();

            /// Method to perform adaptation
            virtual void adapt (bool adaptWV, bool adaptSL, bool adaptWA, int nrAdaptUpdates);

            void computeAmCmXm();

            /// Method to update sheet mechanics
            void updateSheetMechanics (double deltaT);

            /** @brief Return the minimal time of any of the patches inside this chamber
              *
              */
            double getMinActivationTime(double cycleTime) const;




            //==================== file functions ===================

            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;

            /// Virtual implementation of method to write specification to file
            virtual bool writeXml(QDomNode &parentNode) const;

            /// Virtual implementation of method to read specification from file
            virtual bool readXml(const QDomNode &parentNode);

            /// Abstract method to write specification to file
            virtual bool writeSpecification (QDomNode &parentNode) const;

            /// Abstract method to read specification from file
            virtual bool readSpecification (const QDomNode &parentNode);

            /// Abstract method to write state to file
            bool writeState (QDomNode &parentNode) const;

            /// Abstract method to read state from file
            bool readState (const QDomNode &node);

        private:

            /// Method to set default parameters for left atrium
            void setParLeftAtrium();

            /// Method to set default parameters for right atrium
            void setParRightAtrium();

            /// Method to set default parameters for left ventricle
            void setParLeftVentricle ();

            /// Method to set default parameters for septum of ventricle
            void setParSeptumVentricle ();

            /// Method to set default parameters for right ventricle
            void setParRightVentricle ();

        protected:
            // Specification parameters
            double mAmRef;     // Reference area of wall laid as sheet
            double mAmDead;    // Non-contractile area (Valve openings)
            double mVWall;     // Wall volume

            // Simulation state variables
            double mVm;             // Volume enclosed by midwall
            double mXm;             // Central bulge distance of midwall surface
            double mYm;             // Radius of midwall at the junctional edge
            double mSheetAreaMid;   // Area of wall laid as sheet
            double mCm;             // Curvature = 1/radius of midwall
            double mT;              // Wall tension [N/m]
            double mPtrans;         // Transmural pressure
            double mDTDAm;          // dTension/dArea compliance (wall stiffness)
            double mY;              // State variable Y for TriSeg computation
            double mYDot;           // d/dt Y for TriSeg computation
            double mLRef;           // Mean over Y during former beat

            // Variables for TriSeg computation
            int    mNrUpdates; // Number of state updates during current beat

            double mYSum;      // Sum over Y during current beat

            // Variables for adaptation
            double mEAmRef;    // Enlarge septum to anatomical size

            // List of pointers to patches
            // Patches form the sheet encapsulating the cavity
            QList <SPatch *> mPatches;

            chamber_type mChamberType;

            bool mAdaptToExercise;

            bool mUseTriSeg;

        };
    }
}
#endif
