/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: circadapt.cpp 776 2013-07-11 20:33:36Z eelco $
 *
 */

#include <iostream>
#include <fstream>
#include <string>
#include <cstdio>
#include <cmath>
#include "circsettings.h"       // Definitions and settings
#include "error.h"              // Error handling
#include "circulation.h"        // Circulation

// Circulation simulator main program
int main (int argc, char* argv[]) 
{
    // Declaration
    Circulation  *circulation;  // Circulation

    // Declaration of simulation variables
	int     heartbeat;          // Current heart beat 
    
    // Declaration of strings to construct file names
    string  inputfilename;                // Input circulation file name (C++)
    string  basefilename;                 // Output file base name (C++)
    char    c_basefilename[MAXNAMLEN];    // Output file base name (C)
	string  outputfilename;               // Output circulation file name (C++)
	string  specfilename;                 // Output specification file name (C++)
	string  configfilename;               // Output configuration file name (C++)
    string  statefilename;                // Output state file name (C++)
    char    c_statefilename[MAXNAMLEN];   // Output state file name (C)

    // Initialization
    circulation = NULL;   // Circulation

    // Determine file name for input circulation file
    if (argc > 1)
    {
        // Use first input argument to define filename
        inputfilename = argv[1];

        // Print name of circulation file
        cout << "Input circulation file: " << inputfilename << "\n";
    }
    else
    {
        // Obtain file name through user interaction
        cout << "Input circulation file: ";
        cin >> inputfilename;
    }

	// Default parameter settings
	if (inputfilename == "INIT")
	{
		cout << "CIRCADAPT INITIALIZATION\n";
		circulation = new Circulation();
		circulation->create();
		circulation->write(CIRCFILENAME);
		circulation->writeSpecification(CIRCSPECFILENAME);
		circulation->writeConfiguration(CIRCCONFIGFILENAME);
		circulation->writeState(CIRCSTATEFILENAME);
		circulation->print();
		delete circulation;
		circulation = NULL;
		cout << "CIRCADAPT INITIALIZATION FINISHED\n";
		exit(10);
	}

    // Create circulation and read circulation parameters from file
    try
    {
		// Create circulation and read circulation parameters
        circulation = new Circulation(inputfilename);
    }
    catch (Error error)
    {
        if (error.getCode() == ERR_READERR || error.getCode() == ERR_OPENERR)
        {
            // File does not exist
            // Obtain file name through user interaction
            cout << "File " << inputfilename << " does not exist\n";
            cout << "Input circulation file: ";
            cin >> inputfilename;

            // Read circulation file
            circulation = new Circulation();
            CALLM (circulation->read(inputfilename),
                   "Cannot read circulation parameters file");
        }
        else
        {
            QDebug().nospace() << "ERROR: Cannot read circulation parameters file\n";
            exit(10);
        }
    }

	// Print circulaton parameters on screen
	CALLM(circulation->print(),"Cannot print circulation paramters\n");

	// Initialize heart beat
	heartbeat = circulation->getCurrentHeartbeat();

    // Define base name for output files
    basefilename = string(circulation->getOutputfilename()); // C++ style
    sprintf_s (c_basefilename,"%s",basefilename.c_str());    // C style

	// Save circulation parameters
	outputfilename = string(basefilename);
    outputfilename.append(PAREXT);
    CALLM (circulation->write(outputfilename),
           "Cannot write circulation parameters file");

	// Save initial specification of circulation
	specfilename = string(basefilename);
	if (circulation->performAdaptation())
	{
		// Construct file name with "_init"
		specfilename.append("_init");
	}
	specfilename.append(SPCEXT);
	CALLM (circulation->writeSpecification(specfilename),
	       "Cannot write specification file");

	// Save configuration of circulation
    configfilename = string(basefilename);
    configfilename.append(CFGEXT);
    CALLM (circulation->writeConfiguration(configfilename),
           "Cannot write configuration file");

    // Save initial state of circulation
    sprintf_s (c_statefilename,"%s_%.3d",basefilename.c_str(),heartbeat);
	strcat_s (c_statefilename,SIMEXT);
    statefilename = string(c_statefilename);
    CALLM (circulation->writeState(statefilename),
           "Cannot write initial state to file");

	// Open monitor files
	CALLM (circulation->openMonitorFiles(basefilename),
	       "Cannot open monitor files");

	// Open adaptation files
	if (circulation->performAdaptation())
	{
		// Open adaption files
		CALLM (circulation->openAdaptationFiles(basefilename),
			   "Cannot open adaptation files");
	}

    // Simulate circulation
    while (heartbeat < circulation->getTotalHeartbeats())
    {
		// Print message on screen
		cout << "Simulating heart beat " << heartbeat + 1 << "\n";

        // Simulate one heart beat
		CALLM (circulation->simulate(),"Cannot simulate circulation");

		// Perform adaptation step after five heart beats
		if (heartbeat > 5 && circulation->performAdaptation())
		{
			// Perform adaptation step
		    CALLM (circulation->adapt(),"Cannot adapt circulation");
		}

		// Update heart beat
		heartbeat = circulation->getCurrentHeartbeat();

		// Save current state of circulation
        sprintf_s (c_statefilename,"%s_%.3d",basefilename.c_str(),heartbeat);
	    strcat_s (c_statefilename,SIMEXT);
        statefilename = string(c_statefilename);
        CALLM (circulation->writeState(statefilename),
               "Cannot write current state to file");
	}

	// Save final specification of circulation when adapted
	if (circulation->performAdaptation())
	{
		// Save final specification of circulation
        specfilename = string(basefilename);
		specfilename.append("_final");
        specfilename.append(SPCEXT);
        CALLM (circulation->writeSpecification(specfilename),
               "Cannot write specification file");
	}

	// Close monitor files
	CALLM (circulation->closeMonitorFiles(),
	       "Cannot close monitor files");

	// Close adaptation files
	if (circulation->performAdaptation())
	{
		// Close adaption files
		CALLM (circulation->closeAdaptationFiles(),
			   "Cannot close adaptation files");
	}

    // Destroy circulation
	delete circulation;
    circulation = NULL;

    // End of program
    exit(0);
}

