/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: circulation.cpp 789 2013-07-18 13:22:28Z peter $
 *
 */

#include <QDebug>

#include "defs.h"
#include "circsettings.h"

#include "tcirculation.h"
#include "circulation.h"


using namespace Peacs::CircAdapt;

// keys for the xml file
const char * MYCIRCULATIONPARTS_KEY             = "circulation_parts";

const char * MYCONFIGURATION_KEY                = "circulation_configuration";
const char * CONFIGURATION_ITEM                 = "cfg_item";
const char * CONFIGURATION_ITEM_PROXIMAL        = "cfg_item_proximal";
const char * CONFIGURATION_ITEM_DISTAL          = "cfg_item_distal";

const char * SPECIFICATION_CIRC_KEY             = "circulation_specification";
const char * SPECIFICATION_PULMDROP             = "pulmdrop";
const char * SPECIFICATION_VENOUS_RETURN        = "venous_return";
const char * SPECIFICATION_MAP                  = "mean_arterial_pressure";
const char * SPECIFICATION_PRESSURE_FLOW_CNTRL  = "pressure_flow_control";



const double DEFAULT_PULMDROP = 1500;
const double DEFAULT_VENOUS_RETURN = 8.5000e-005;
const double DEFAULT_MAP = 12200;

const char * STATE_CIRC_KEY                     = "circulation_state";


const int FILE_CIRCULATION_VERSION_1  = 1;
const int FILE_CIRCULATION_VERSION_DEFAULT  = 1;
const char* FILE_CIRCULATION_NAME = "CIRCULATION";



const char * AORTA_NAME = "aorta";
const char * PULMONARYVEINS_NAME = "pulmonary_veins";
const char * PULMONARYARTERY_NAME = "pulmonary_artery";
const char * SYSTEMICVEINS_NAME = "systemic_veins";




// Constructor: initialization only
SCirculation::SCirculation( void ) :
    SCircElement("Circulation"),
    mSim_timestep(CIRCSIMTIMESTEP),
    mAdapt_timestep(CIRCADAPTTIMESTEP),
    mPressureControl(false),
    mAdaptRest(false),
    mAdaptExercise(false),
    mAorta(STube::AORTA_TUBE,AORTA_NAME),
    mPulmonaryVeins(STube::PULMONARY_VEIN_TUBE,PULMONARYVEINS_NAME ),
    mPulmonaryArtery(STube::PULMONARY_ARTERY_TUBE,PULMONARYARTERY_NAME),
    mSystemicVeins(STube::SYSTEMIC_VEIN_TUBE,SYSTEMICVEINS_NAME),
    mHeart(mSystemicVeins, mPulmonaryArtery, mPulmonaryVeins, mAorta, TRISEG, mAdaptRest || mAdaptExercise ),
    mTubes(),
    mResistances(),
    mQ0(8.5000e-005),
    mP0(12200),
    mDropPulm(DEFAULT_PULMDROP),
    mAdtapStatus(TransitionAdaptation),
    mCircState(NULL)
{
    // Default initialization
//    create();

}


SCirculation::~SCirculation()
{
    if ( mCircState != NULL )
    {
        delete mCircState;
    }
}


const TCirculation & SCirculation::getCirculation() const
{
    return *mCircState;
}

// Private method for default initialization
// This method should be called by constructors only
void SCirculation::init()
{
//    mQ0       =     8.5000e-005; // Specific systemic flow
//    mP0        = 12200.0;         // Specific systemic pressure
    tCycle    =     0.85;        // Specific cycle time
    pResp     =     0.0;         // Pressure from respiration


}

void SCirculation::copyUserSpecifications(const TCirculation &circ)
{
    mPressureControl = circ.pressureControl();
    mAdaptRest       = circ.adaptForRest();
    mAdaptExercise   = circ.adaptForExercise();
    mQ0              = circ.venousReturn();
    mP0              = circ.meanArterialPressure();
    mDropPulm        = circ.getPulmDrop();

    mAorta.copyUserSpecifications( circ.getAorta() );
    mPulmonaryVeins.copyUserSpecifications( circ.getPulmonaryVeins() );
    mPulmonaryArtery.copyUserSpecifications( circ.getPulmonaryArtery() );
    mSystemicVeins.copyUserSpecifications( circ.getSystemicVein() );
    mHeart.copyUserSpecifications( circ.getHeart() );
    for ( int i=0; i < mTubes.size(); i++)
    {
        mTubes[i].copyUserSpecifications(  circ.getTubes().at(i) );
    }
    for ( int i=0; i < mResistances.size(); i++)
    {
        mResistances[i].copyUserSpecifications( circ.getResistances().at(i) );
    }
    mCircState->copySpecifications( *this );
}



/** @brief returns the default version number */
int SCirculation::getDefaultVersion(void) const
{
    return FILE_CIRCULATION_VERSION_DEFAULT;
}

/** @brief Returns the class name */
const char *SCirculation::getClassName(void) const
{
    return FILE_CIRCULATION_NAME;
}


STube* SCirculation::getTube(const QString &name)
{
    if ( mAorta.getName() == name )
    {
        return & mAorta;
    }
    else if ( mPulmonaryVeins.getName() == name )
    {
        return & mPulmonaryVeins;
    }
    else if ( mPulmonaryArtery.getName() == name )
    {
        return & mPulmonaryArtery;
    }
    else if ( mSystemicVeins.getName() == name )
    {
        return & mSystemicVeins;
    }
    else
    {
        for ( int i=0; i < mTubes.size(); i++)
        {
            if ( mTubes[i].getName() == name )
            {
                return & mTubes[i];
            }
        }
    }
    return NULL;
}

SResistance * SCirculation::getResistance(const QString & name)
{

    for ( int i=0; i < mResistances.size(); i++)
    {
        if ( mResistances[i].getName() == name )
        {
            return & mResistances[i];
        }
    }
    return NULL;
}


// Method to simulate one heart beat
void SCirculation::simulateHeartBeat( int cycleTimeInMs )
{
    // TODO: Make sure that the simulation cannot be called twice in the same time (2 threads)

    // Declarations
    Time    simTime     = 0;                            // Current simulation time
    //Time    cycle_time  = cycleTimeInMs * 1000;       // Cycle time (microseconds)
    Time    end_time    = cycleTimeInMs * 1000;         // End time of heart beat (microseconds)
    double  deltaT      = 0.000001 * mSim_timestep;     // Time step (seconds)
    int     t = 0;

    tCycle              = (double)cycleTimeInMs / 1000.0;
    // Initialize number of adaptation updates
    mNrAdaptUpdates     = 0;

    double minActivationTime = qMin( mHeart.getAtria().getLeftAtrium().getMinActivationTime(tCycle),
                               qMin( mHeart.getAtria().getRightAtrium().getMinActivationTime(tCycle),
                               qMin( mHeart.getVentricles().getLeftVentricle().getMinActivationTime(tCycle),
                               qMin( mHeart.getVentricles().getRightVentricle().getMinActivationTime(tCycle),
                                     mHeart.getVentricles().getSeptum().getMinActivationTime(tCycle) ) ) ) );

    mHeart.initMechanics( tCycle, minActivationTime); // rest tCycle TODO

    mHeart.initAdaptation();
    mAorta.initAdaptation();
    mPulmonaryVeins.initAdaptation();
    mPulmonaryArtery.initAdaptation();
    mSystemicVeins.initAdaptation();

    mCircState->resetData( ceil(cycleTimeInMs / ((double)CIRCMONTIMESTEP /1000.0)) );

    for ( int i = 0 ; i < mTubes.size() ; i++ )
    {
        mTubes[i].initAdaptation();
    }

    for ( int i = 0 ; i < mResistances.size() ; i++ )
    {
        mResistances[i].initAdaptation();
    }

    // Simulate one heart beat
    while (simTime <= end_time)
    {
        // Increase simulation time
        simTime += mSim_timestep;

        // Update cavity pressures and cross-sections for the heart
        mHeart.updateHeartMechanics(deltaT);

        mAorta.updateTubeDynamics(deltaT);
        mPulmonaryVeins.updateTubeDynamics(deltaT);
        mPulmonaryArtery.updateTubeDynamics(deltaT);
        mSystemicVeins.updateTubeDynamics(deltaT);
        // Update tube pressures, wave impedances and cross-sections
        for ( int i=0; i < mTubes.size(); i++)
        {
            // Update tube pressure, wave impedance and cross-section
            mTubes[i].updateTubeDynamics(deltaT);
        }

        updateSystemicFlow();

        updatePulmonaryFlow();

        mHeart.updateValves(deltaT, mPulmonaryVeins.getArea(),mSystemicVeins.getArea());

        // Update adaptation variables
        if ( simTime % mAdapt_timestep == 0 )
        {
            // Count number of adaptation updates
            mNrAdaptUpdates++;

            // Update adaptation variables for all compliances
            mHeart.updateAdaptation();
            mAorta.updateAdaptation();
            mPulmonaryVeins.updateAdaptation();
            mPulmonaryArtery.updateAdaptation();
            mSystemicVeins.updateAdaptation();
            for (int i=0; i < mTubes.size(); i++ )
            {
                mTubes[i].updateAdaptation();
            }
        }
        if ( simTime % CIRCMONTIMESTEP ==0 )
        {
            mCircState->addSample( *this, t );
            t++;
        }
    }
    adapt();
    mCircState->copySpecifications( *this );
}


// Method to perform one adaptation step
void SCirculation::adapt()
{
    // Consistency check: check validity of adaptation variables
    if (   mNrAdaptUpdates != 0 &&
         ( mAdaptExercise || mAdaptRest ) )
    {
        // Determine criterium for flow stationarity
        double ErrFlow = mHeart.updateErrorFlow();

        // Adaptation only when flow is stationary
        if ( ErrFlow < ERRFLOWCRITERIUM )
        {

//            qDebug() << " doadapt " << mNrAdaptUpdates << mHeart.getVentricles().getSeptum().getAmRef();
            // Adaptation for rest
            if ( mAdaptRest )
            {
                mAorta.adaptDiameter(mNrAdaptUpdates);
                mPulmonaryVeins.adaptDiameter(mNrAdaptUpdates);
                mPulmonaryArtery.adaptDiameter(mNrAdaptUpdates);
                mSystemicVeins.adaptDiameter(mNrAdaptUpdates);

                // Adapt tube diameters
                for ( int i =0; i < mTubes.size(); i++)
                {
                    mTubes[i].adaptDiameter(mNrAdaptUpdates);
                }
            }

            // Adaptation for exercise
            if ( mAdaptExercise )
            {
                // Adapt tube wall thickness wall volume of tube
                mAorta.adaptWallVolume(mNrAdaptUpdates);
                mPulmonaryVeins.adaptWallVolume(mNrAdaptUpdates);
                mPulmonaryArtery.adaptWallVolume(mNrAdaptUpdates);
                mSystemicVeins.adaptWallVolume(mNrAdaptUpdates);
                for ( int i = 0 ; i < mTubes.size(); i++)
                {
                    mTubes[i].adaptWallVolume(mNrAdaptUpdates);
                }
                // Adapt sheet properties of chambers
                mHeart.adapt(ADAPT_WV,ADAPT_SL,ADAPT_WA,mNrAdaptUpdates);
            }
        }
    }
}



void SCirculation::updateSystemicFlow()
{
    // Update systemic flow
    if (mPressureControl)
    {
        // Open circulation:
        // Adjust systematic peripheral resistance to control blood pressure
        // Blood pressure is controlled by opening the circulation at
        // the systemic venes (RVen)
        // p0 is actual systemic pressure
        // q0 is actual systemic flow
        for (int i = 0; i < mResistances.size(); i++ )
        {
            //TODO ALSO for the SYSTEMIC_CRANIAL ones??
            if ( mResistances[i].getResistanceType() == SResistance::SYSTEMIC )
            {
                mResistances[i].controlBloodPressure(mP0,mQ0);
            }
        }
    }
    else
    {
        // Closed circulation:
        // Update systemic flow and
        // volume derivatives in neighboring compliances
        for (int i = 0; i < mResistances.size(); i++ )
        {
            //TODO ALSO for the SYSTEMIC_CRANIAL ones??
            if ( mResistances[i].getResistanceType() == SResistance::SYSTEMIC )
            {
                mResistances[i].updateFlow();
            }
        }
    }

}


void SCirculation::updatePulmonaryFlow()
{
    double pNorm;           // Normalized pressure drop for pulmonary circ
    double pulm_flow;       // Pulmonary flow

    // Non-linear pressure-flow relation of pulmonary circulation
    // Pulmonary flow increases with power of two with pressure drop
    // pNorm is normalized pressure drop for pulmonary circulation
    // mDropPulm is pulmonary pressure drop
    pNorm = (mPulmonaryArtery.getPressure() - mPulmonaryVeins.getPressure()) / mDropPulm;

    // Compute pulmonary flow
    pulm_flow = pNorm * fabs(pNorm) * mQ0;

    // Set flow for pulmonary peripheral resistance and
    // update volume derivatives in neighboring compliances
    for (int i = 0; i < mResistances.size(); i++ )
    {
        //TODO ALSO for all the PULMONARY resistances??
        if ( mResistances[i].getResistanceType() == SResistance::PULMONARY )
        {
            mResistances[i].setFlow(pulm_flow);
        }
    }
}


// Method to create default circulation
void SCirculation::create()
{
    // Create default circulation
    mPressureControl    = false;
    mAdaptRest          = false;
    mAdaptExercise      = false;
    mAorta              = STube(STube::AORTA_TUBE, AORTA_NAME);
    mPulmonaryVeins   = STube(STube::PULMONARY_VEIN_TUBE,PULMONARYVEINS_NAME );
    mPulmonaryArtery    = STube(STube::PULMONARY_ARTERY_TUBE,PULMONARYARTERY_NAME);
    mSystemicVeins      = STube(STube::SYSTEMIC_VEIN_TUBE,SYSTEMICVEINS_NAME);
    mHeart              = SHeart(mSystemicVeins, mPulmonaryArtery, mPulmonaryVeins, mAorta, TRISEG, mAdaptRest || mAdaptExercise );
    mResistances.clear();
    mTubes.clear();
    // Systemic peripheral resistance (LRp)
    mResistances.append(SResistance(SResistance::SYSTEMIC,mAorta,mSystemicVeins,SYSTEMIC_R));
    // Pulmonary peripheral resistance (RRp)
    mResistances.append(SResistance(SResistance::PULMONARY,mPulmonaryArtery, mPulmonaryVeins,PULMONARY_R));

    init();

    if ( mCircState != NULL )
    {
        delete mCircState;
    }

    mCircState = new TCirculation( *this );
}

//kaj 1234567890 nu    illllllliiiiiiii nul luz   massa massa assa assa ataj



bool SCirculation::writeSpecification (QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_CIRC_KEY));
    // Write specification of compliance
    bool ok = SCircElement::writeSpecification(specNode);

    ok &= addValue(specNode, SPECIFICATION_PULMDROP,mDropPulm);
    ok &= addValue(specNode, SPECIFICATION_VENOUS_RETURN,mQ0);
    ok &= addValue(specNode, SPECIFICATION_MAP,mP0);
    ok &= addValue(specNode, SPECIFICATION_PRESSURE_FLOW_CNTRL,mPressureControl);

    return ok;
}

/// method to read specification from file
bool SCirculation::readSpecification (const QDomNode &parentNode)
{
    QDomNodeList elements = parentNode.toElement().elementsByTagName(SPECIFICATION_CIRC_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement specNode = elements.at(0).toElement();
        mDropPulm           = readValue(specNode, SPECIFICATION_PULMDROP,DEFAULT_PULMDROP);
        mQ0                 = readValue(specNode, SPECIFICATION_VENOUS_RETURN,DEFAULT_VENOUS_RETURN);
        mP0                 = readValue(specNode, SPECIFICATION_MAP,DEFAULT_MAP);
        mPressureControl    = readValue(specNode, SPECIFICATION_PRESSURE_FLOW_CNTRL,false);
        return true;
    }
    return false;
}


/// Virtual implementation of method to write specification to file
bool SCirculation::writeXml(QDomNode &parentNode) const
{
    bool        ok      = false;
    QDomNode    myNode  = createXmlEntry( parentNode, getClassName(), getDefaultVersion(), getName() );

    if ( !myNode.isNull())
    {
        ok &= writeSpecification(myNode);
        ok &= writeState(myNode);

        QDomNode circPartNode = createXmlEntry(myNode,MYCIRCULATIONPARTS_KEY);
        ok &= mHeart.writeXml(circPartNode);
        ok &= mAorta.writeXml(circPartNode);
        ok &= mPulmonaryVeins.writeXml(circPartNode);
        ok &= mPulmonaryArtery.writeXml(circPartNode);
        ok &= mSystemicVeins.writeXml(circPartNode);
        for (int i=0; i < mResistances.size(); i++ )
        {
            ok &= mResistances[i].writeXml(circPartNode);
        }
        for (int i=0; i < mTubes.size(); i++ )
        {
            ok &= mTubes[i].writeXml(circPartNode);
        }
        QDomNode circConfigNode = createXmlEntry(myNode,MYCONFIGURATION_KEY);
        for ( int i = 0; i < mResistances.size(); i++)
        {
            QDomNode thisNode = createXmlEntry(circConfigNode,CONFIGURATION_ITEM,mResistances[i].getDefaultVersion(), mResistances[i].getName());
            ok &= addValue(thisNode, CONFIGURATION_ITEM_PROXIMAL, mResistances[i].getProximal().getName());
            ok &= addValue(thisNode, CONFIGURATION_ITEM_DISTAL, mResistances[i].getDistal().getName());
        }
    }
    return ok;
}


/// Virtual implementation of method to read specification from file
bool SCirculation::readXml(const QDomNode &node)
{
    bool ok = SCircElement::readXml( node );
    int nRes = mResistances.size();
    if ( ok )
    {
        QDomNodeList myNodes = node.toElement().elementsByTagName(MYCIRCULATIONPARTS_KEY);
        if ( myNodes.size() == 1 )
        {
            QDomNode node = myNodes.at(0);
            myNodes = node.childNodes();
            SCompliance in,out;
            SResistance resistance(SResistance::UNDEFINED_RESISTANCE,in,out);
            STube tube(STube::UNDEFINED_TUBE);

            for ( int i = 0; i < myNodes.size(); i++)
            {
                if ( myNodes.at(i).toElement().tagName() == mHeart.getClassName())
                {
                    ok &= mHeart.readXml(myNodes.at(i));
                }
                else if ( myNodes.at(i).toElement().tagName() == resistance.getClassName() )
                {
                    ok &= resistance.readXml(myNodes.at(i));
                    if ( ok )
                    {
                        bool found = false;
                        for ( int i=0; i < mResistances.size();i++)
                        {
                            if ( mResistances[i].getResistanceType() == resistance.getResistanceType() )
                            {
                                mResistances[i] = resistance;
                                found = true;
                                break;
                            }
                        }
                        if ( !found )
                        {
                            mResistances.append(resistance);
                        }
                    }
                }
                else if ( myNodes.at(i).toElement().tagName() == tube.getClassName() )
                {
                    ok &= tube.readXml(myNodes.at(i));
                    if ( ok )
                    {
                        switch ( tube.getTubeType() )
                        {
                            case STube::AORTA_TUBE :
                                mAorta = tube;
                                break;
                            case STube::PULMONARY_VEIN_TUBE:
                                mPulmonaryVeins = tube;
                                break;
                            case STube::PULMONARY_ARTERY_TUBE:
                                mPulmonaryArtery = tube;
                                break;
                            case STube::SYSTEMIC_VEIN_TUBE:
                                mSystemicVeins = tube;
                                break;
                            case STube::OTHER_TUBE:
                                mTubes.append(tube);
                                break;
                            default:
                                qDebug().nospace() << "Could not determine the right Tuibe type";
                                ok = false;
                                break;
                        }
                    }
                }
                else
                {
                    ok = false;
                    qDebug().nospace() << "unknown circulation element ";
                }
            }
        }
        else
        {
            ok = false;
        }
        if ( ok )
        {

            QDomNodeList myNodes = node.toElement().elementsByTagName(MYCONFIGURATION_KEY);
            if ( myNodes.size() == 1 )
            {
                QDomNode node = myNodes.at(0);
                myNodes = node.childNodes();
                for ( int i = 0; i < myNodes.size(); i++)
                {
                    QDomNode itemNode = node.childNodes().at(i);
                    QString proximal = readString(itemNode,CONFIGURATION_ITEM_PROXIMAL);
                    QString distal = readString(itemNode,CONFIGURATION_ITEM_DISTAL);
                    SResistance *resPtr = getResistance( getNodeName(itemNode) );
                    STube *proxPtr = getTube(proximal);
                    STube *distPtr = getTube(distal);
                    if ( resPtr !=NULL && proxPtr != NULL && distPtr != NULL )
                    {
                        resPtr->setConnectedCompliance( * proxPtr, * distPtr);
                    }
                    else
                    {
                        ok = false;
                    }
                }
            }
        }
    }
    if ( !ok )
    {
        qDebug().nospace() << "file could not be loaded";
    }   
    return ok;
}



bool SCirculation::open(const QString &filename)
{
    bool    ok = false;
    QFile   file( filename );

    if ( !filename.isEmpty() && file.open( QIODevice::ReadOnly )  )
    {
        QDomDocument doc("CircAdapt");

        if ( doc.setContent( &file ) )
        {
            // Find the circulation item. Only one should be present
            QDomNodeList nodes = doc.elementsByTagName( getClassName() );

            if ( nodes.count() == 1 )
            {
                ok = readXml( nodes.at( 0 ) );
            }
        }

        file.close();

        if ( ok )
        {
            if ( mCircState != NULL )
            {
                delete mCircState;
            }

            init();

            mCircState      = new TCirculation( *this );
        }
        else
        {

            //TODO what??
        }

    }

    return ok;
}

bool SCirculation::save(const QString &filename)
{
    QFile   file( filename );
    bool    ok = false;

    if ( file.open( QIODevice::WriteOnly ) )
    {
        QDomDocument    doc( "CircAdapt" );
        QDomNode        root = doc.createElement( "CircAdapt" );

        doc.appendChild( root );

        writeXml( root );

        QString     xml = doc.toString();
        QTextStream ts( &file );

        ts << xml;

        file.close();

        ok = true;
    }
    return ok;
}
