/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: circulation.h 781 2013-07-14 17:01:20Z peter $
 *
 */

#ifndef _CIRCULATION_H_
#define _CIRCULATION_H_




#include "circsettings.h"
#include "compliance.h"
#include "heart.h"
#include "tube.h"
#include "flowduct.h"
#include "resistance.h"
#include "circelement.h"
#include "PCircAdaptDefines.h"

#include <QList>

namespace Peacs
{

    namespace CircAdapt
    {

        class TCirculation;

        /** @brief Class to simulate heart and circulation
          */
//        class SCirculation : public QObject, public SCircElement
        class SCirculation : public SCircElement
        {
//            Q_OBJECT

        public:

            /** @brief Default Constructor
              */
            SCirculation( void );

            /// Destructor
            virtual ~SCirculation();

            /** Simulate one heart beat
              */
            void simulateHeartBeat( int cyclTimeInMs );

            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TCirculation  &circ);

            inline const SHeart & getHeart() const { return mHeart; }

            inline const STube & getAorta() const { return mAorta; }

            inline const STube & getPulmonaryArtery() const { return mPulmonaryArtery; }

            inline const STube & getPulmonaryVeins() const { return mPulmonaryVeins; }

            inline const STube & getSystemicVein() const { return mSystemicVeins; }

            inline const QList <STube> & getTubes() const { return mTubes; }

            inline const QList <SResistance> & getResistances() const { return mResistances; }

            /// Perform adaptation
            inline bool performAdaptation() const { return mAdaptRest || mAdaptExercise; }

            /** @return the status of the pressure contral flag on/off
              */
            bool   pressureControl() const  { return  mPressureControl; }

            /** @brief set the status of the pressure contral flag on/off
              * @param control the new value
              */
            void setPressureControl( bool control ) { mPressureControl  = control; }


            /** @brief the status of the adapt for rest flag
              */
            bool   adaptForRest() const { return mAdaptRest; }

            /** @brief adapt with rest setting when true
              * @param control the new value
              */
            void setAdaptForRest(bool adapt) { mAdaptRest = adapt; }

            /** @brief the status of the adapt for rest flag
              */
            bool   adaptForExercise() const { return mAdaptExercise; }

            /** @brief venous return */
            double venousReturn() const { return mQ0; }


            /** @brief set the venous return */
            void setVenousReturn(double CO) { mQ0 = CO; }


            /** @brief venous return */
            double meanArterialPressure() const { return mP0; }

            /** @brief set the venous return */
            void setMeanArterialPressure(double mAP) { mP0 = mAP; }

            /** @brief adapt with exercise setting when true
              * @param control the new value
              */
            void setAdaptForExercise(bool adapt) { mAdaptExercise = adapt; }

            /** @brief pulm drop*/
            void setPulmDrop(double drop) { mDropPulm = drop; }

            double getPulmDrop() const { return mDropPulm; }

            void setAdaptStatus(AdaptStatusType adaptStatus) { mAdtapStatus = adaptStatus; }

            AdaptStatusType getAdaptStatus() const { return mAdtapStatus; }


             //==================== file functions ===================

            bool open(const QString &filename);

            bool save(const QString &filename);

            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;

            int getCycleTime() const { return qRound(tCycle * 1000.0); }

            /// Method to create default circulation
            void create();

            const TCirculation & getCirculation() const;


        private:

            /** Default initialization
              */
            void init();

            /** Perform one adaptation step
              */
            void adapt();

            /** Virtual implementation of method to write specification to file
              */
            bool writeXml(QDomNode &parentNode) const;

            /** Virtual implementation of method to read specification from file
              */
            bool readXml(const QDomNode &node);


            /// method to write specification to file
            bool writeSpecification (QDomNode &parentNode) const;

            /// method to read specification from file
            bool readSpecification (const QDomNode &parentNode);

            void updateSystemicFlow();

            void updatePulmonaryFlow();

            STube* getTube(const QString & name);

            SResistance* getResistance(const QString & name);


        private:

            // Simulation and monitor time steps
            Time   mSim_timestep;        // Simulation time step

            Time   mAdapt_timestep;      // Time step to update adaptation vars

//            Time   mMon_timestep;        // Monitor time step

            // Flags
            bool   mPressureControl;     // Control blood pressure when TRUE
            bool   mAdaptRest;           // Adaptation for rest when TRUE
            bool   mAdaptExercise;       // Adaptation for exercise when TRUE

            STube        mAorta;                // Lart

            STube        mPulmonaryVeins;       // LVen

            STube        mPulmonaryArtery;       // Rart

            STube        mSystemicVeins;         // RVen

            SHeart       mHeart;

            // List of pointers to all tubes
            QList <STube> mTubes;

            // List of pointers to all resistances
            QList <SResistance> mResistances;

            // Circulation control parameters
            double mQ0;                 // Actual systemic flow
            double mP0;                 // Actual systemic pressure
            double tCycle;              // Actual cycle time

            // Specification parameters
            double mDropPulm;           // Pulmonary pressure drop

            double pResp;               // Pressure from respiration

            // Variables for adaptation
            int    mNrAdaptUpdates;      // Number of adaptation updates

            AdaptStatusType mAdtapStatus;

            TCirculation *mCircState;
        };
    }
}
#endif
