/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: compliance.cpp 776 2013-07-11 20:33:36Z eelco $
 *
 */

#include <QString>

#include <cmath>
#include "defs.h"
#include "circsettings.h"
#include "tokens.h"
#include "compliance.h"
#include "tcompliance.h"

// keys for the xml file
const char * SPECIFICATION_COMPLIANCE_KEY = "compliance_specification";

const char * STATE_COMPLIANCE_KEY         = "compliance_state";

const char * STATE_VOLUME      = "volume";
const char * STATE_DDTVOLUME   = "volumeprime";
const char * STATE_AREA        = "area";
const char * STATE_P           = "pressure";
const char * STATE_PIN         = "pressure_in";
const char * STATE_IMPEDANCE   = "Impedance";
const char * STATE_FLOWREMOD   = "flowremod";



const int FILE_COMPLIANCE_SPECIFICATION_VERSION  = 1;
const int FILE_COMPLIANCE_STATE_VERSION  = 1;

const int FILE_COMPLIANCE_VERSION  = 1;

const char* FILE_COMPLIANCE_NAME = "COMPLIANCE";



using namespace Peacs::CircAdapt;
// Constructor

SCompliance::SCompliance(const QString &name):
    SCircElement(name),
    mVolume(0.0),
    mVDot(0.0),
    mA(0.0),
    mP(0.0),
    mZ(0.0),
    mPin(0.0),
    mQRemod(0.0)
{

}

SCompliance::SCompliance():
    SCircElement(),
    mVolume(0.0),
    mVDot(0.0),
    mA(0.0),
    mP(0.0),
    mZ(0.0),
    mPin(0.0),
    mQRemod(0.0)
{

}

SCompliance::SCompliance(const SCompliance &comp):
    SCircElement(comp),
    mVolume(comp.mVolume),
    mVDot(comp.mVDot),
    mA(comp.mA),
    mP(comp.mP),
    mZ(comp.mZ),
    mPin(comp.mPin),
    mQRemod(comp.mQRemod)
{

}


const SCompliance& SCompliance::operator = (const SCompliance &comp)
{
    if (this != &comp )
    {
        SCircElement::operator =(comp);

        mVolume = comp.mVolume;
        mVDot   = comp.mVDot;
        mA      = comp.mA;
        mP      = comp.mP;
        mZ      = comp.mZ;
        mPin    = comp.mPin;
        mQRemod = comp.mQRemod;
    }
    return *this;
}

void SCompliance::init()
{
    mVolume = 0.0;
    mVDot   = 0.0;
    mA      = 0.0;
    mP      = 0.0;
    mZ      = 0.0;
    mPin    = 0.0;
    mQRemod = 0.0;
}


void SCompliance::copyUserSpecifications(const TCompliance &comp)
{
    if ( comp.getInitialVolume()  != mVolume && comp.getInitialVolume() > 0 )
    {
        mVolume = comp.getInitialVolume();
    }
}


// Method to add incoming flow from valve
void SCompliance::addIncomingFlowFromValve (double flow)
{
       // Add incoming flow to d/dt Volume
    mVDot += flow;

    // Update through flow used for diameter adaptation
    mQRemod += flow;

    // Update pressure at entrance
    mPin = mP + mVDot*mZ;
}


// Method to subtract outgoing flow to valve
void SCompliance::subtractOutgoingFlowToValve (double flow)
{
    // Subtract outgoing flow from d/dt Volume
    mVDot -= flow;

    // Update through flow used for diameter adaptation
    mQRemod -= flow;

    // Update pressure at entrance
    mPin = mP + mVDot*mZ;
}



bool SCompliance::writeState (QDomNode &parentNode) const
{
    QDomNode stateNode(createXmlEntry(parentNode,STATE_COMPLIANCE_KEY));
    addValue(stateNode, STATE_VOLUME,       mVolume );
    addValue(stateNode, STATE_DDTVOLUME,    mVDot );
    addValue(stateNode, STATE_AREA,         mA );
    addValue(stateNode, STATE_P,            mP );
    addValue(stateNode, STATE_PIN,          mPin);
    addValue(stateNode, STATE_IMPEDANCE,    mZ);
    addValue(stateNode, STATE_FLOWREMOD,    mQRemod);
    return !stateNode.isNull();
}

/// Abstract method to read state from file
bool SCompliance::readState (const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(STATE_COMPLIANCE_KEY);
    if ( elements.count() == 1 )
    {
        QDomNode stateNode = elements.at(0);
        mVolume = readValue(stateNode, STATE_VOLUME);
        mVDot   = readValue(stateNode, STATE_DDTVOLUME);
        mA      = readValue(stateNode, STATE_AREA );
        mP      = readValue(stateNode, STATE_P );
        mPin    = readValue(stateNode, STATE_PIN );
        mZ      = readValue(stateNode, STATE_IMPEDANCE );
        mQRemod = readValue(stateNode, STATE_FLOWREMOD );
        return true;
    }
    return false;
}




