/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: compliance.h 776 2013-07-11 20:33:36Z eelco $
 *
 */


#ifndef _COMPLIANCE_H_
#define _COMPLIANCE_H_

#include <cmath>
#include "circsettings.h"
#include "circelement.h"



namespace Peacs
{
    using namespace Support;

    namespace CircAdapt
    {
        class TCompliance;

        /**
        Abstract class for compliant chamber or blood vessel.
        Inherits from CircElement.
        */
        class SCompliance: public SCircElement {

        public:
            /* @brief Default Constructor */
            SCompliance();

            /* @brief Constructor */
            SCompliance(const QString &name);

            SCompliance(const SCompliance &comp);

            virtual const SCompliance& operator = (const SCompliance &comp);

            void init();


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TCompliance &);


            /// Method to get cross-section
            inline double getArea() const { return mA; }

            /// Method to get volume of compliance
            inline double getVolume() const { return mVolume; }

            inline double getVolumeDot() const { return mVDot; }

            /// Method to get pressure of compliance
            inline const double & getPressure() const { return mP; }

            /// Method to get pressure at entrance
            inline double getPressureAtEntrance() const { return mPin; }

            inline void setVdot(double value) { mVDot = value; }

            inline void setPressure(double value) { mP = value; }

            /// Method to get pressure at entrance
            inline double getImpedance() const { return mZ; }

            inline double getQRemod() const { return mQRemod; }


            /// Method to add Volume and limit volume
            inline void addVolume (double value) { mVolume = qMax(0.0, mVolume + value); }

            /// Method to add pressure from pericardial sack
            inline void addPressure (double value) { mP += value; }

            /// Method to add incoming flow from peripheral resistance
            inline void addIncomingFlowFromResistance (double flow) { mVDot += flow; }

            /// Method to subtract outgoing flow to peripheral resistance
            inline void subtractOutgoingFlowToResistance (double flow) { mVDot -= flow; }

            /// Method to add incoming flow from valve
            void addIncomingFlowFromValve (double flow);

            /// Method to subtract outgoing flow to valve
            void subtractOutgoingFlowToValve (double flow);

            /// Virtual method to initialize adaptation variables
            virtual void initAdaptation() {}

            /// Virtual method to update adaptation variables
            virtual void updateAdaptation() {}


            //==================== file functions ===================

            /** @brief returns the default version number */
            int getDefaultVersion(void) const { return -1; }

            /** @brief Returns the class name */
            const char *getClassName(void) const {return "NOTVALID"; }

            /// Abstract method to write state to file
            bool writeState (QDomNode &parentNode) const;

            /// Abstract method to read state from file
            bool readState (const QDomNode &node);

        protected:
            // Simulation state variables
            double mVolume;       // Volume of compliance

            double mVDot;    // d/dt Volume

            double mA;       // Cross-section of compliance

            double mP;       // Pressure of compliance

            double mZ;       // Impedance

            double mPin;     // Pressure at entrance

            double mQRemod;  // Through flow used for diameter adaptation

        };
    }
}
#endif
