/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: flowduct.cpp 776 2013-07-11 20:33:36Z eelco $
 *
 */


/*---------------------------------------------------------------+
| Class: MFlowDuct                                                |
| Representation of a flow channel between compliances.          |
| Inherits from MCircElement.                                     |
| Valve and Resistance inherit from MFlowDuct.                    |
|                                                                |
| Nico Kuijpers and Theo Arts                                    |
| September 12, 2008                                             |
| Maastricht University                                          |
|                                                                |
| Sum of flow over heart beat added to determine stability       |
| criterion for adaptation.                                      |
| Nico Kuijpers                                                  |
| July 7, 2010                                                   |
| Maastricht University                                          |
+---------------------------------------------------------------*/




#include "compliance.h"
#include "circsettings.h"
#include "tokens.h"
#include "flowduct.h"

const char * SPECIFICATION_FLOWDUCT_KEY  = "flowduct_specification";

const char * STATE_FLOWDUCT_KEY          = "flowduct_state";
const char * STATE_FLOWDUCT_FLOW         = "flow";

const char * ADAPTATION_FLOWDUCT_KEY     = "flowduct_adaptation";
const char * ADAPTATION_FLOWSUM          = "flowduct_flowsum";

const int FILE_FLOWDUCT_VERSION_1  = 1;
const int FILE_FLOWDUCT_VERSION_DEFAULT  = 1;

const char* FILE_FLOWDUCT_NAME = "FLOWDUCT";

using namespace  Peacs::CircAdapt;

SFlowDuct::SFlowDuct(SCompliance &prox, SCompliance &dist):
    SCircElement(),
    mProximal(prox),
    mDistal(dist),
    mQ(0.0)
{

}


SFlowDuct::SFlowDuct (SCompliance &prox, SCompliance &dist, const QString &name):
    SCircElement(name),
    mProximal(prox),
    mDistal(dist),
    mQ(0.0)
{

}


SFlowDuct::SFlowDuct (const SFlowDuct & flowduct):
    SCircElement(flowduct),
    mProximal(flowduct.mProximal),
    mDistal(flowduct.mDistal),
    mQ(flowduct.mQ)
{

}

const SFlowDuct & SFlowDuct::operator = (const SFlowDuct &flowduct)
{
    if ( this != &flowduct )
    {
        mQ      = flowduct.mQ;
    }
    return *this;
}

void SFlowDuct::init()
{
    mQ = 0;
}


int SFlowDuct::getDefaultVersion(void) const
{
    return FILE_FLOWDUCT_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SFlowDuct::getClassName(void) const
{
    return FILE_FLOWDUCT_NAME;
}

bool SFlowDuct::isValid() const
{
    return mProximal.getDefaultVersion() > 0 && mDistal.getDefaultVersion() > 0;
}


void SFlowDuct::setConnectedCompliance(SCompliance & prox, SCompliance& dist)
{
    mProximal   = prox;
    mDistal     = dist;
}


// Virtual implementation of method to write state to file
bool SFlowDuct::writeState(QDomNode &parentNode) const
{
    QDomNode stateNode(createXmlEntry(parentNode,STATE_FLOWDUCT_KEY));
    return addValue(stateNode,STATE_FLOWDUCT_FLOW, mQ);
}


// Virtual implementation of method to read state from file
bool SFlowDuct::readState(const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(STATE_FLOWDUCT_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement stateNode = elements.at(0).toElement();
        mQ = readValue(stateNode,STATE_FLOWDUCT_FLOW);
        return true;
    }
    return false;

}


