/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: heart.cpp 789 2013-07-18 13:22:28Z peter $
 *
 */

#include <QDebug>

#include "defs.h"
#include "heart.h"
#include "theart.h"


// keys for the xml file
const char * MYHEARTPARTS_KEY           = "heart_parts";

const char * SPECIFICATION_HEART_KEY  = "heart_specification";
const char * STATE_HEART_KEY          = "heart_state";

const int FILE_HEART_VERSION_1  = 1;
const int FILE_HEART_VERSION_DEFAULT  = 1;

const char* FILE_HEART_NAME = "HEART";


const char* LEFT_HEART_NAME     = "heart_atria";
const char* RIGHT_HEART_NAME    = "heart_ventricles";

enum valve_type { UNDEFINED_VALVE=0,LEFT_ATRIUM_IN_LVen=1,MITRALVALVE_LAv=2,AORTA_LArt=3,VENA_CAVA_RVen=4,TRICUSPID_RAv=5,
                  RVOT_RArt=6, ASD=7,VSD=8,DUCT=9};


using namespace  Peacs::CircAdapt;

// Constructor
SHeart::SHeart(SCompliance &rightIn, SCompliance &rightOut, SCompliance &LeftIn, SCompliance &leftOut, ventricle_type type, bool adaptToExercise):

    SCircElement("heart"),
    mAtria("atria"),
    mVentricle(type, adaptToExercise,       "ventricles"),
    mPericardialSack(mAtria,mVentricle,     "pericardial_sack"),
    mValveLVen( SValve::LEFT_ATRIUM_IN_LVen, LeftIn,                      mAtria.leftAtrium(),         "Inlet_left_atrium"),
    mValveLAv(  SValve::MITRALVALVE_LAv,     mAtria.leftAtrium(),         mVentricle.leftVentricle(),  "Mitral_valve"),
    mValveLArt( SValve::AORTA_LArt,          mVentricle.leftVentricle(),  leftOut,                     "Aortic_valve" ),
    mValveRVen( SValve::VENA_CAVA_RVen,      rightIn ,                    mAtria.rightAtrium()      ,  "Inlet_right_atrium"),
    mValveRAv(  SValve::TRICUSPID_RAv,       mAtria.rightAtrium(),        mVentricle.rightVentricle(), "Tricuspid_valve"),
    mValveRArt( SValve::RVOT_RArt,           mVentricle.rightVentricle(), rightOut,                    "Pulmonary_valve"),
    mValveASD(  SValve::ASD,                 mAtria.leftAtrium(),         mAtria.rightAtrium(),        "ASD" ),
    mValveVSD(  SValve::VSD,                 mVentricle.leftVentricle(),  mVentricle.rightVentricle(), "VSD" ),
    mValveDUCT( SValve::DUCT,                leftOut,                     rightOut,                    "DUCTUS"),
    mRVenFlowPrev(0.0),
    mRArtFlowPrev(0.0),
    mLVenFlowPrev(0.0),
    mLArtFlowPrev(0.0),
    mMeanFlowPrev(0.0)
{

}



SHeart::SHeart(const SHeart & heart) :
    SCircElement(heart),
    mAtria(heart.mAtria),
    mVentricle(heart.mVentricle),
    mPericardialSack(heart.mPericardialSack),
    mValveLVen(heart.mValveLVen),            // Inlet left atrium
    mValveLAv(heart.mValveLAv),             // Mitral valve
    mValveLArt(heart.mValveLArt),            // Aortic valve
    mValveRVen(heart.mValveRVen),            // Inlet right atrium
    mValveRAv(heart.mValveRAv),             // Tricuspid valve
    mValveRArt(heart.mValveRArt),            // Pulmonary valve
    mValveASD(heart.mValveASD),            // Valve for ASD
    mValveVSD(heart.mValveVSD),           // Valve for VSD
    mValveDUCT(heart.mValveDUCT),            // Valve for DUCT))
    mRVenFlowPrev(heart.mRVenFlowPrev),
    mRArtFlowPrev(heart.mRArtFlowPrev),
    mLVenFlowPrev(heart.mLVenFlowPrev),
    mLArtFlowPrev(heart.mLArtFlowPrev),
    mMeanFlowPrev(heart.mMeanFlowPrev)
{

}


void SHeart::init()
{
    mRVenFlowPrev     = 0.0;    // Flow through inlet right atrium
    mRArtFlowPrev     = 0.0;    // Flow through pulmonary valve
    mLVenFlowPrev     = 0.0;    // Flow through inlet left atrium
    mLArtFlowPrev     = 0.0;    // Flow through aortic valve
    mMeanFlowPrev     = 0.0;    // Mean flow through valves
}


int SHeart::getDefaultVersion(void) const
{
    return FILE_HEART_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SHeart::getClassName(void) const
{
    return FILE_HEART_NAME;
}

const SHeart & SHeart::operator =(const SHeart &heart)
{
    if ( this != &heart)
    {
        SCircElement::operator =(heart);
        mAtria     = heart.mAtria;
        mVentricle = heart.mVentricle;
        mPericardialSack = heart.mPericardialSack;
        mValveLVen = heart.mValveLVen;            // Inlet left atrium
        mValveLAv  = heart.mValveLAv;             // Mitral valve
        mValveLArt = heart.mValveLArt;            // Aortic valve
        mValveRVen = heart.mValveRVen;            // Inlet right atrium
        mValveRAv  = heart.mValveRAv;             // Tricuspid valve
        mValveRArt = heart.mValveRArt;            // Pulmonary valve
        mValveASD  = heart.mValveASD;            // Valve for ASD
        mValveVSD  = heart.mValveVSD;           // Valve for VSD
        mValveDUCT = heart.mValveDUCT;            // Valve for DUCT))
        mRVenFlowPrev = heart.mRVenFlowPrev;
        mRArtFlowPrev = heart.mRArtFlowPrev;
        mLVenFlowPrev = heart.mLVenFlowPrev;
        mLArtFlowPrev = heart.mLArtFlowPrev;
        mMeanFlowPrev = heart.mMeanFlowPrev;
    }
    return *this;
}



void SHeart::copyUserSpecifications(const THeart &heart)
{
    mAtria.copyUserSpecifications(             heart.getAtria() );
    mVentricle.copyUserSpecifications(         heart.getVentricles() );
    mPericardialSack.copyUserSpecifications(   heart.getSack() );
    mValveLVen.copyUserSpecifications(         heart.getValveLVen() );
    mValveLAv.copyUserSpecifications(          heart.getValveLAv() );
    mValveLArt.copyUserSpecifications(         heart.getValveLArt() );
    mValveRVen.copyUserSpecifications(         heart.getValveRVen() );
    mValveRAv.copyUserSpecifications(          heart.getValveRAv() );
    mValveRArt.copyUserSpecifications(         heart.getValveRArt() );
    mValveASD.copyUserSpecifications(          heart.getValveASD() );
    mValveVSD.copyUserSpecifications(          heart.getValveVSD() );
    mValveDUCT.copyUserSpecifications(         heart.getValveDUCT() );
}

// Method to update mechanics for pericardial sack
void SHeart::updateSackMechanics()
{
    double PSack = mPericardialSack.updateVSackAndComputePSack( mAtria.getVolume() + mVentricle.getVolume());

    // Adjust cavity pressures for pericardial sack
    mAtria.adjustForSackPressure(PSack);
    mVentricle.adjustForSackPressure(PSack);
}



// Method to initialize cavity pressure and cross-section
void SHeart::initMechanics (double tCycleAct, double minActivationTime)
{
    mAtria.initAtrialMechanics(tCycleAct,minActivationTime);

    mVentricle.initMechanics(tCycleAct,minActivationTime);
}


// Method to update cavity pressure and cross-section
void SHeart::updateHeartMechanics(double deltaT)
{
    mAtria.updateAtrialMechanics(deltaT);

    mVentricle.updateVentricleMechanics(deltaT);

    updateSackMechanics();
}


void SHeart::updateValves(double deltaT, double pulmonaryVeinsArea,double systemVeinsArea)
{
    double dL;              // Mitral leak signal
    double dR;              // Tricuspid leak signal
    double passiveLv;       // Passive leak of mitral valve
    double passiveRv;       // Passive leak of tricuspid valve
    double mitral_leak;     // Leak of mitral valve
    double tricuspid_leak;  // Leak of tricuspid valve
    // Update flows around valves
    // Flow continuity
    // Time derivatives of volumes and boundary pressures pIn and pOut
    // Time derivatives of flows

    // Enlarged leak of atrioventricular valves during diastole
    // Estimate mitral leak signal
    dL = mVentricle.estimateLeftLeakSignal(MITRALLEAKPAR);

    // Passive leak of mitral valve
    passiveLv = 1.0 - tanh( dL * dL);

    // Set leak of mitral valve (mitral regurgitation)
    mitral_leak = qMax(MITRALLEAKFACTOR*passiveLv,MITRALLEAKMIN);
    mValveLAv.setLeak(mitral_leak);

    // Estimate tricuspid leak signal
    dR = mVentricle.estimateRightLeakSignal(TRICUSPIDLEAKPAR);

    // Passive leak of tricuspid valve
    passiveRv = 1.0 - tanh( dR * dR );

    // Set leak of tricuspid valve
    tricuspid_leak = qMax(TRICUSPIDLEAKFACTOR*passiveRv,TRICUSPIDLEAKMIN);
    mValveRAv.setLeak(tricuspid_leak);

    // Venous outflow areas in the atria vary with venous diameter
    // Set cross-sections of open valves for atria inlets
    mValveLVen.setOpenArea(pulmonaryVeinsArea);
    mValveRVen.setOpenArea(systemVeinsArea);

    // Compute volume derivatives for all valves
    mValveLVen.computeVolumeDerivatives(deltaT);
    mValveLAv.computeVolumeDerivatives(deltaT);
    mValveLArt.computeVolumeDerivatives(deltaT);
    mValveRVen.computeVolumeDerivatives(deltaT);
    mValveRAv.computeVolumeDerivatives(deltaT);
    mValveRArt.computeVolumeDerivatives(deltaT);
    mValveASD.computeVolumeDerivatives(deltaT);
    mValveVSD.computeVolumeDerivatives(deltaT);
    mValveDUCT.computeVolumeDerivatives(deltaT);


    mValveLVen.computeFlowDerivative();
    mValveLAv.computeFlowDerivative();
    mValveLArt.computeFlowDerivative();
    mValveRVen.computeFlowDerivative();
    mValveRAv.computeFlowDerivative();
    mValveRArt.computeFlowDerivative();
    mValveASD.computeFlowDerivative();
    mValveVSD.computeFlowDerivative();
    mValveDUCT.computeFlowDerivative();
}


// Implement method to initialize adaptation variables
void SHeart::initAdaptation()
{
    mAtria.initAdaptation();

    mVentricle.initAdaptation();

}


// Implement method to update adaptation variables
void SHeart::updateAdaptation()
{

    mAtria.updateAdaptation();

    mVentricle.updateAdaptation();

    mPericardialSack.updateAdaptation();


}


double SHeart::updateErrorFlow()
{
    // Declarations
    double RVenFlowCur;  // Current flow through inlet right atrium
    double RArtFlowCur;  // Current flow through pulmonary valve
    double LVenFlowCur;  // Current flow through inlet left atrium
    double LArtFlowCur;  // Current flow through aortic valve
    double MeanFlowCur;  // Mean flow for current heart beat
    double SumSquares;   // Sum of squares to compute standard deviation
    double StdFlow;      // Standard deviation of FlowCur - FlowPrev
    double ErrFlow;      // Criterium for flow stationarity

    // Obtain current flows through valves
    RVenFlowCur = mValveRVen.getFlow();  // Inlet right atrium
    RArtFlowCur = mValveRArt.getFlow();  // Pulmonary valve
    LVenFlowCur = mValveLVen.getFlow();  // Inlet left atrium
    LArtFlowCur = mValveLArt.getFlow();  // Aortic valve

    // Compute mean flow for current heart beat
    MeanFlowCur = (RVenFlowCur + RArtFlowCur + LVenFlowCur + LArtFlowCur) / 4.0;

    // Compute standard deviation of difference between FlowCur and FlowPrev
    // Standard deviation for vector x of length n is defined by
    // mean(x) = (SUM_i^n x_i) / n
    // std(x)  = sqrt((SUM_i^n (x_i - mean(x))^2) / (n-1))
    // Note the division by (n-1), which is equal to 3 in this case
    SumSquares =
            pow( ( RVenFlowCur - MeanFlowCur ) - ( mRVenFlowPrev - mMeanFlowPrev), 2.0) +
            pow( ( RArtFlowCur - MeanFlowCur ) - ( mRArtFlowPrev - mMeanFlowPrev), 2.0) +
            pow( ( LVenFlowCur - MeanFlowCur ) - ( mLVenFlowPrev - mMeanFlowPrev), 2.0) +
            pow( ( LArtFlowCur - MeanFlowCur ) - ( mLArtFlowPrev - mMeanFlowPrev), 2.0);
    StdFlow = sqrt(SumSquares / 3.0);

    // Store current flows for adaptation after next heart beat
    mRVenFlowPrev = RVenFlowCur;  // Inlet right atrium
    mRArtFlowPrev = RArtFlowCur;  // Pulmonary valve
    mLVenFlowPrev = LVenFlowCur;  // Inlet left atrium
    mLArtFlowPrev = LArtFlowCur;  // Aortic valve
    mMeanFlowPrev = MeanFlowCur;  // Mean flow through valves

    // Determine criterium for flow stationarity
    ErrFlow = StdFlow / MeanFlowCur;
    return ErrFlow;
}


// Method to perform adaptation
void SHeart::adapt (bool adaptWV, bool adaptSL, bool adaptWA,int nrAdaptUpdates)
{

    mAtria.adapt(adaptWV,adaptSL,adaptWA,nrAdaptUpdates);

    mVentricle.adapt(adaptWV,adaptSL,adaptWA,nrAdaptUpdates);

    mPericardialSack.adapt();
}



bool SHeart::writeXml(QDomNode &parentNode) const
{
    bool ok = false;
    QDomNode myNode = createXmlEntry(parentNode,getClassName(),getDefaultVersion(),getName());
    if ( !myNode.isNull())
    {
        ok &= writeSpecification(myNode);
        ok &= writeState(myNode);

        QDomNode chambersNode = createXmlEntry(myNode,MYHEARTPARTS_KEY);
        ok &= mAtria.writeXml(chambersNode);
        ok &= mVentricle.writeXml(chambersNode);
        ok &= mPericardialSack.writeXml(chambersNode);
        ok &= mValveLVen.writeXml(chambersNode);
        ok &= mValveLAv.writeXml(chambersNode);
        ok &= mValveLArt.writeXml(chambersNode);
        ok &= mValveRVen.writeXml(chambersNode);
        ok &= mValveRAv.writeXml(chambersNode);
        ok &= mValveRArt.writeXml(chambersNode);
        ok &= mValveASD.writeXml(chambersNode);
        ok &= mValveVSD.writeXml(chambersNode);
        ok &= mValveDUCT.writeXml(chambersNode);
    }
    return ok;
}


bool SHeart::readXml(const QDomNode &node)
{
    bool ok = SCircElement::readXml(node);
    if ( ok )
    {
        QDomNodeList myNodes = node.toElement().elementsByTagName(MYHEARTPARTS_KEY);
        if ( myNodes.size() == 1 )
        {
            QDomNode node = myNodes.at(0);
            myNodes = node.childNodes();
            for ( int i = 0; i < myNodes.size(); i++)
            {
                if ( myNodes.at(i).toElement().tagName() == mAtria.getClassName())
                {
                    ok &= mAtria.readXml(myNodes.at(i));
                }
                else if ( myNodes.at(i).toElement().tagName() == mVentricle.getClassName() )
                {
                    ok &= mVentricle.readXml(myNodes.at(i));
                }
                else if ( myNodes.at(i).toElement().tagName() == mPericardialSack.getClassName() )
                {
                    ok &= mPericardialSack.readXml(myNodes.at(i));
                }
                else if ( myNodes.at(i).toElement().tagName() == mValveLVen.getClassName() )
                {
                    SCompliance in,out;
                    SValve valve(SValve::UNDEFINED_VALVE,in,out);
                    valve.readXml(myNodes.at(i));
                    switch ( valve.getValveType() )
                    {
                    case SValve::LEFT_ATRIUM_IN_LVen:
                        mValveLVen = valve;
                        break;
                    case SValve::MITRALVALVE_LAv:
                        mValveLAv = valve;
                        break;
                    case SValve::AORTA_LArt:
                        mValveLArt = valve;
                        break;
                    case SValve::VENA_CAVA_RVen:
                        mValveRVen = valve;
                        break;
                    case SValve::TRICUSPID_RAv:
                        mValveRAv = valve;
                        break;
                    case SValve::RVOT_RArt:
                        mValveRArt = valve;
                        break;
                    case SValve::ASD:
                        mValveASD = valve;
                        break;
                    case SValve::VSD:
                        mValveVSD = valve;
                        break;
                    case SValve::DUCT:
                        mValveDUCT = valve;
                        break;
                    default:
                        qDebug().nospace() << "Valve was not recognized";
                        ok = false;
                        break;
                    }
                }
            }
        }
        else
        {
            ok = false;
        }
    }
    return ok;
}



//bool MHeart::writeState (QDomNode &parentNode) const
//{
//    QDomNode stateNode(createXmlEntry(parentNode,STATE_ATRIA_KEY));
//    MCircElement::writeState(stateNode);

//    addValue(stateNode, STATE_VOLUME_MIDWALL,       mVm);

//    return !stateNode.isNull();
//}

///// Abstract method to read state from file
//bool MHeart::readState(const QDomNode &node)
//{
//    QDomNodeList elements = node.toElement().elementsByTagName(STATE_ATRIA_KEY);
//    if ( elements.count() == 1 )
//    {
//        QDomElement stateNode = elements.at(0).toElement();
//        MCircElement::readState(stateNode);
//        mVm     = readValue(stateNode, STATE_VOLUME_MIDWALL);
//        return true;
//    }
//    return false;
//}

