/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: heart.h 789 2013-07-18 13:22:28Z peter $
 *
 */

#ifndef _HEART_H_
#define _HEART_H_

#include "defs.h"
#include "atria.h"
#include "ventricle.h"
#include "valve.h"
#include "compliance.h"
#include "sack.h"
#include "circelement.h"


using namespace Peacs::Support;

namespace Peacs
{

    namespace CircAdapt
    {

        class THeart;
        /**


        */
        class SHeart: public SCircElement
        {

        public:
            /** @brief Default Constructor */
            SHeart(SCompliance &RightIn,
                   SCompliance &rightOut,
                   SCompliance &LeftIn,
                   SCompliance &leftOut,
                   ventricle_type type = BISEG,
                   bool adaptToExercise=false );

            SHeart(const SHeart & SHeart);

            virtual ~SHeart() {;}


            virtual const SHeart& operator = (const SHeart &ventr);

            void init();

            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const THeart &heart);

            const SAtria &getAtria() const { return mAtria; }

            const SVentricle &getVentricles() const { return mVentricle; }


            const SSack  &getSack() const      { return mPericardialSack; }

            const SValve &getValveLVen() const { return mValveLVen; }           // Inlet left atrium
            const SValve &getValveLAv()  const { return mValveLAv;   }        // Mitral valve
            const SValve &getValveLArt() const { return mValveLArt;  }          // Aortic valve
            const SValve &getValveRVen() const { return mValveRVen; }            // Inlet right atrium
            const SValve &getValveRAv()  const { return mValveRAv; }            // Tricuspid valve
            const SValve &getValveRArt() const { return mValveRArt; }           // Pulmonary valve
            const SValve &getValveASD()  const { return mValveASD;  }           // Valve for ASD
            const SValve &getValveVSD()  const { return mValveVSD;  }           // Valve for VSD
            const SValve &getValveDUCT() const { return mValveDUCT; }            // Valve for DUCT

//            const SValve &getLeftAin() const      { return mValveLVen; }           // Inlet left atrium
//            const SValve &getMitralValve() const      { return mValveLAv;  }           // Mitral valve
//            const SValve &getAorticValve() const      { return mValveLArt; }           // Aortic valve
//            const SValve &getRightAin() const      { return mValveRVen; }           // Inlet right atrium
//            const SValve &getTricuspidValve() const      { return mValveRAv;  }           // Tricuspid valve
//            const SValve &getRVOTValve() const      { return mValveRArt; }           // Pulmonary valve
//            const SValve &getASD() const      { return mValveASD;  }           // Valve for ASD
//            const SValve &getVSD() const      { return mValveVSD;  }           // Valve for VSD
//            const SValve &getDUCT() const      { return mValveDUCT; }           // Valve for DUCT








            inline double getVolume() const { return mAtria.getVolume() + mVentricle.getVolume(); }

            void adjustForSackPressure(double pressure);

            /// Method to initialize cavity pressure and cross-section
            void initMechanics(double tCycleAct, double tCycleRef);

            void updateHeartMechanics (double deltaT);

            void updateValves(double deltatT, double pulmonaryVeinsArea,double systemVeinsArea);

            /// Implement method to initialize adaptation variables
            void initAdaptation();

            /// Implement method to update adaptation variables
            void updateAdaptation();

            /// Method to perform adaptation
            void adapt (bool adaptWV, bool adaptSL, bool adaptWA, int nrAdaptUpdates);

            double updateErrorFlow();

            //==================== file functions ===================

            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;

            /// Virtual implementation of method to write specification to file
            virtual bool writeXml(QDomNode &parentNode) const;

            /// Virtual implementation of method to read specification from file
            virtual bool readXml(const QDomNode &parentNode);

        private:
            void updateSackMechanics();


        private:

            SAtria       mAtria;

            SVentricle   mVentricle;

            SSack        mPericardialSack;

            SValve       mValveLVen;            // Inlet left atrium
            SValve       mValveLAv;             // Mitral valve
            SValve       mValveLArt;            // Aortic valve
            SValve       mValveRVen;            // Inlet right atrium
            SValve       mValveRAv;             // Tricuspid valve
            SValve       mValveRArt;            // Pulmonary valve
            SValve       mValveASD;             // Valve for ASD
            SValve       mValveVSD;             // Valve for VSD
            SValve       mValveDUCT;            // Valve for DUCT

            // Variables for adaptation
            double mRVenFlowPrev;        // Flow through inlet right atrium
            double mRArtFlowPrev;        // Flow through pulmonary valve
            double mLVenFlowPrev;        // Flow through inlet left atrium
            double mLArtFlowPrev;        // Flow through aortic valve
            double mMeanFlowPrev;        // Mean flow through valves

        };
    }
}
#endif
