/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id:
 *
 */

//#include <QDebug>

#include <ctime>
#include <QFileInfo>
#include "defs.h"
#include "circsettings.h"
#include "tokens.h"

#include "msimulatecirculation.h"

#include <QDebug>
using namespace Peacs::CircAdapt;


const int   FILE_MSIMULATECIRCULATION_VERSION_1         = 1;
const int   FILE_MSIMULATECIRCULATION_VERSION_DEFAULT   = 1;
const char* FILE_MSIMULATECIRCULATION_NAME              = "MSIMULATECIRCULATION";
const int   HEARTBEAT_INTERVAL_DEFAULT                  = 60000 / 70;
const int   INITIAL_RUN_COUNT                           = 21;


// Constructor: initialization only
MSimulateCirculation::MSimulateCirculation(/*const QString &filename*/) :
    QThread(),
    mMutex(),
    mPause( true ),     // We start paused
    mReset( false ),
    mSave( false ),
    mLoad( true ),      // We start loading a model (if specified) or initialize with default settings
    mAbort( false ),
    mAdaptMode ( false ),
    mNrInitialRuns( 0 ),
    mFilename( /*filename*/ ),
    mSaveFilename(),
    mUpdateSpec(),
    mUpdateSpecChanged( false ),
    mCirculation(),
    mHeartRateInterval( HEARTBEAT_INTERVAL_DEFAULT ),
    mRestCO(  8.5000e-005 ),
    mExerciseCO( 2.5500e-004 ),
    mRestHeartInterval( HEARTBEAT_INTERVAL_DEFAULT),
    mExerciseHeartInterval(60000/140),
    mTransitionBeatNr(100),
    mRestBeatNr(100),
    mExerciseBeatNr(100),
    mTransitionCnt(0),
    mExerciseCnt(100),
    mRestCnt(100),
    mDoRest(true),
    mDoExercise(true),
    mPrevHeartRate(mHeartRateInterval),
    mPrevCO(mRestCO),
    mPrevPressureControl(false),
    mAdaptionOn(false)
{
    // TODO For some reason we always need to create (initialize) the default curculation. This should be fixed
    // Caused by the reading of resistances
    mCirculation.create();
    mPrevPressureControl = mCirculation.pressureControl();
}


MSimulateCirculation::~MSimulateCirculation()
{
    if ( isRunning() )
    {
        if ( !mAbort )
        {
            mMutex.tryLock( 100 );  // try to lock for 100 ms
            mAbort = true;
            mMutex.unlock();
        }

//        wait( 200 );    // Wait for the thread to stop for 200 ms
    }
}


void MSimulateCirculation::run( void )
{
    statetype   state = initState;
    bool        abort( false );
    bool        pause( false );
    bool        reset( false );
    bool        save( false );
    bool        load( false );
    bool        adapt( false );

    clock_t     start;
    clock_t     stop;
    int         diff;
    int         sleep;

    // Just to be sure ...
    mNrInitialRuns      = 0;
    mHeartRateInterval  = HEARTBEAT_INTERVAL_DEFAULT;

    // Sleep for a while, so the main window (thread) can finish its initialization and display properly
    msleep( 100 );

    while ( !abort )
    {
        start = clock();

        mMutex.lock();
        // Check shared resources by geting the values
        abort               = mAbort;
        pause               = mPause;
        reset               = mReset;
        save                = mSave;
        load                = mLoad;
        adapt               = mAdaptMode;

        // Some, one-time actions, can be reset to prevent conflicts
        mReset              = false;
        mSave               = false;
        mLoad               = false;

        if ( mUpdateSpecChanged )
        {
            mCirculation.copyUserSpecifications( mUpdateSpec );

            emit( specsCopied() );

            mUpdateSpecChanged = false;
        }

        mMutex.unlock();

        if ( !abort )           // Abort means nothing to do any more
        {
            if ( load )         // Loading is more important then resetting or saving
            {
                doLoad();
            }
            else if ( reset )
            {
                doLoad();       // In this version the reset function is the same as the load
            }
            else if ( save )
            {
                doSave();
            }

            if ( pause )
            {
                if ( state != pauseState )
                {
                    state = pauseState;

                    emit( stateChanged( (int)state ) );
                }
            }
            else
            {
                if ( state != activeState )
                {
                    state = activeState;

                    emit( stateChanged( (int)state ) );
                }

                // Do the work
                doHeartbeat();

                doAdapt( adapt );
            }

            stop = clock();
            diff = double( stop - start ) / CLOCKS_PER_SEC * 1000;
            sleep = mHeartRateInterval - diff;

            //qDebug() << "time:" << diff << "- heartrate:" << mHeartRateInterval << "- sleeping:" << qBound( 1, sleep, mHeartRateInterval ) << "- old diff:" << difftime( stop, start );
            if ( !adapt )
            {
                msleep( qBound( 1, sleep, (int)mHeartRateInterval ) );
            }
            else
            {
                msleep( 1 );
            }
        }
    }
}


void MSimulateCirculation::stopAdaptCirc()
{
    mMutex.lock();

    mAdaptMode = false;

    mMutex.unlock();
}



void MSimulateCirculation::doHeartbeat( void )
{
    if ( mNrInitialRuns < INITIAL_RUN_COUNT )
    {
        TCirculation specInfo( mCirculation.getCirculation() );

        specInfo.copySpecifications( mCirculation );

        emit( initializedModel( specInfo ) );

        // Run the initial heartbeats to reach a steady state
        while ( mNrInitialRuns < INITIAL_RUN_COUNT )
        {
            mCirculation.simulateHeartBeat( mHeartRateInterval );

            mNrInitialRuns++;
        }

        emit( initialRunDone() );
    }
    else
    {
        // Steady state already reached, do the normal heartbeat stuff
        mCirculation.simulateHeartBeat( mHeartRateInterval );
    }

    TCirculation monInfo( mCirculation.getCirculation() );

    monInfo.setHeartbeatTime( mHeartRateInterval );

    emit( updateBeatInfo( monInfo ) );
}


void MSimulateCirculation::doAdapt( bool adapt )
{
    // maxTransNr indicates teh number of beats for transition, including stabilization
    // transCnt  is just the number to get from one state of the other

    mMutex.lock();

    if ( adapt )
    {
        mCirculation.setPressureControl( true );

        mAdaptionOn                     = true;

        if ( mTransitionCnt > 0 ) // transition
        {
            mCirculation.setAdaptStatus(TransitionAdaptation);
            mCirculation.setAdaptForRest( false );
            mCirculation.setAdaptForExercise( false );

            double delta = (mRestHeartInterval - mExerciseHeartInterval) / (double) mTransitionBeatNr;

            mTransitionCnt--;
            if ( mExerciseCnt > 0  ) //from rest to exercise
            {
                mHeartRateInterval      = qBound( mExerciseHeartInterval , mHeartRateInterval - delta, mRestHeartInterval);
                mCirculation.setVenousReturn( qBound( mRestCO, mCirculation.venousReturn() + (mExerciseCO - mRestCO) / mTransitionBeatNr, mExerciseCO ) );
            }
            else
            {
                mHeartRateInterval      = qBound( mExerciseHeartInterval , mHeartRateInterval + delta, mRestHeartInterval);
                mCirculation.setVenousReturn( qBound( mRestCO, mCirculation.venousReturn() - (mExerciseCO - mRestCO) / mTransitionBeatNr, mExerciseCO ) );
            }
        }
        else if ( mRestCnt > 0 )
        {
            mCirculation.setAdaptStatus(RestAdaptation);
            mCirculation.setAdaptForRest( mDoRest );
            mCirculation.setAdaptForExercise( false );
            mCirculation.setVenousReturn( mRestCO );
            mHeartRateInterval          = mRestHeartInterval;

            mRestCnt--;
            qDebug() << mRestCnt;

            if ( mRestCnt <= 0 )
            {
                mTransitionCnt  = mTransitionBeatNr;
                if ( mDoExercise )
                {
                    mExerciseCnt = mExerciseBeatNr;
                }
                else// if ( mDoRest )
                {
                    mRestCnt = mRestBeatNr;
                }
            }
        }
        else //if ( mExerciseCnt > 0 ) // exercise
        {
            mCirculation.setAdaptStatus(ExerciseAdapation);
            mCirculation.setAdaptForRest( false );
            mCirculation.setAdaptForExercise( mDoExercise );
            mHeartRateInterval          = mExerciseHeartInterval;
            mCirculation.setVenousReturn( mExerciseCO );

            mExerciseCnt--;
            if ( mExerciseCnt <= 0 )
            {
                mTransitionCnt  = mTransitionBeatNr;
                // make the switch?
                if ( mDoRest )
                {
                    mRestCnt        = mRestBeatNr;
                }
                else// if ( mDoExercise )
                {
                    mExerciseCnt    = mExerciseBeatNr;
                }
            }
        }

    }
    else
    {
        mCirculation.setAdaptStatus(NOADAPTATION);
        if ( mAdaptionOn )
        {
            mHeartRateInterval          = mPrevHeartRate;
            mCirculation.setVenousReturn( mPrevCO );
            mCirculation.setAdaptForRest( false );
            mCirculation.setAdaptForExercise( false );
            mCirculation.setPressureControl( mPrevPressureControl );
            mAdaptionOn                 = false;
        }
        else
        {
            mPrevCO                     = mCirculation.venousReturn();
            mPrevPressureControl        = mCirculation.pressureControl();
            mPrevHeartRate              = mHeartRateInterval;
        }
    }

    mMutex.unlock();
}


void MSimulateCirculation::adaptCirc(double restFlow,
                                     double exerciseFlow,
                                     double restHeartInterval,
                                     double exerciseHeartInterval,
                                     int nrTransitionbeats,
                                     int nrRestBeats,
                                     int nrExerciseBeats,
                                     bool dorest,
                                     bool doExercise)
{
    mMutex.lock();
    mRestCO = restFlow;
    mExerciseCO = exerciseFlow;
    mRestHeartInterval = restHeartInterval;
    mExerciseHeartInterval = exerciseHeartInterval;
    mAdaptMode = true;
    mTransitionBeatNr  = nrTransitionbeats;
    mRestBeatNr = nrRestBeats;
    mExerciseBeatNr = nrExerciseBeats;
    mTransitionCnt = mTransitionBeatNr;
    mDoRest = dorest;
    mDoExercise = doExercise;
    if ( !mDoRest )
    {
        mRestCnt = 0;
    }
    if ( !mDoExercise )
    {
        mDoExercise = 0;
    }

    mMutex.unlock();
}

void MSimulateCirculation::updateAdaptCirc(double restFlow,
                                           double exerciseFlow,
                                           double restHeartInterval,
                                           double exerciseHeartInterval,
                                           int nrTransitionbeats,
                                           int nrRestBeats,
                                           int nrExerciseBeats,
                                           bool dorest,
                                           bool doExercise)
{
    mMutex.lock();
    mRestCO = restFlow;
    mExerciseCO = exerciseFlow;
    mRestHeartInterval = restHeartInterval;
    mExerciseHeartInterval = exerciseHeartInterval;
    mAdaptMode = true;
    mTransitionBeatNr  = nrTransitionbeats;
    mRestBeatNr = nrRestBeats;
    mExerciseBeatNr = nrExerciseBeats;
    mDoRest = dorest;
    mDoExercise = doExercise;
    mMutex.unlock();
}


void MSimulateCirculation::doLoad( void )
{
    mMutex.lock();
    if ( mFilename.isEmpty() )
    {
        mCirculation.create();                  // Create a model with default settings

        mNrInitialRuns      = 0;
        mHeartRateInterval  = HEARTBEAT_INTERVAL_DEFAULT;

        emit( modelLoaded( (int)defaultmodel ) );

        doHeartbeat();                          // Make sure that the initial runs are performed
    }
    else
    {
        if ( mCirculation.open( mFilename ) )   // Load the new model
        {
            mNrInitialRuns      = 0;
            mHeartRateInterval  = HEARTBEAT_INTERVAL_DEFAULT;

            emit( modelLoaded( (int)loadedmodel ) );

            doHeartbeat();                      // Make sure that the initial runs are performed
        }
        else
        {
            mCirculation.create();                  // Create a model with default settings

            mFilename.clear();

            emit( actionFailed( (int)openFailed ) );
        }
    }
    mPrevHeartRate       = mHeartRateInterval;
    mPrevCO              = mCirculation.venousReturn();
    mPrevPressureControl = mCirculation.pressureControl();
    mAdaptMode           = false;
    mAdaptionOn          = false;
    mMutex.unlock();
}


void MSimulateCirculation::doSave( void )
{
    mMutex.lock();

    if ( !mSaveFilename.isEmpty() )
    {
        if ( mCirculation.save( mSaveFilename ) )
        {
            mFilename = mSaveFilename;

            emit( modelSaved() );
        }
        else
        {
            emit( actionFailed( (int)saveFailed ) );
        }

        mSaveFilename.clear();
    }

    mMutex.unlock();
}


QString MSimulateCirculation::getFilename()
{
    QString filename;

    mMutex.lock();

    filename = mFilename;

    mMutex.unlock();

    return filename;
}


bool MSimulateCirculation::isActive( void )
{
    bool toReturn = false;

    if ( isRunning() )          // If the thread is running ...
    {
        mMutex.lock();

        toReturn = !mPause;     // ... and not paused ...!

        mMutex.unlock();
    }

    return toReturn;
}


void MSimulateCirculation::startHeartbeat( void )
{
    mMutex.lock();

    mPause = false;

    mMutex.unlock();
}


void MSimulateCirculation::pauseHeartbeat( void )
{
    mMutex.lock();

    mPause = true;

    mMutex.unlock();
}


void MSimulateCirculation::reset(void)
{
    mMutex.lock();

    mReset = true;

    mMutex.unlock();
}


void MSimulateCirculation::abort( void )
{
    mMutex.lock();

    mAbort = true;

    mMutex.unlock();
}


void MSimulateCirculation::load( const QString & filename )
{
    mMutex.lock();


    mFilename   = filename;
    mLoad       = true;

    mMutex.unlock();
}


void MSimulateCirculation::save( const QString & filename )
{
    mMutex.lock();

    mSaveFilename   = filename;
    mSave           = true;

    mMutex.unlock();
}


void MSimulateCirculation::updateHeartInterval(int interval)
{
    mMutex.lock();

    mHeartRateInterval = interval;

    mMutex.unlock();
}


void MSimulateCirculation::updateSpecs( const TCirculation & spec )
{
    mMutex.lock();

    mUpdateSpec         = spec;
    mUpdateSpecChanged  = true;

    mMutex.unlock();
}


