/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id:
 *
 */

#ifndef __MSIMULATECIRCULATION_H_
#define __MSIMULATECIRCULATION_H_

#include <QObject>
#include <QThread>
#include <QMutex>
//#include <QTime>
//#include <QTimerEvent>


#include "tcirculation.h"
#include "circulation.h"


namespace Peacs
{

    namespace CircAdapt
            {

        /** @brief Class to simulate heart and circulation
          */
        class MSimulateCirculation : public QThread
        {
            Q_OBJECT

        public:

            typedef enum
            {
                defaultmodel,
                loadedmodel
            } modeltype;

            typedef enum
            {
                initState,
                pauseState,
                activeState
            } statetype;

            typedef enum
            {
                openFailed,
                saveFailed
            } failtype;


            /** @brief Default Constructor
              */
            MSimulateCirculation( /*const QString & filename*/ );


            /** Destructor
              */
            virtual ~MSimulateCirculation();


            /** @brief returns the name of the file currently used to simulate heartbeats
              * If an empty string is returned simulation is being performed with default
              * settings
              *
              * @return the name of the file
              */
            QString getFilename();


            /** @brief returns if the thread is currently simulating heartbeats
              *
              * @return true is simulating
              */
            bool isActive( void );


            /** @brief Tells the thread to perform the heartbeat calculations
              */
            void startHeartbeat( void );


            /** @brief Tells the thread to pause performing the heartbeat calculations
              */
            void pauseHeartbeat( void );


            /** @brief Tells the thread to reset all changes (load initial settings)
              */
            void reset( void );


            /** @brief Tells the thread to abort.
              * After this signal is handled the thread will stop and won't be able to
              * do the work
              */
            void abort( void );


            /** @brief Tells the thread to load a file
              *
              * @param filename the name of the new file to load
              */
            void load( const QString & filename );


            /** @brief Tells the thread to save the current simulation to the file
              *
              * @param filename the new filename to save to
              */
            void save( const QString & filename );


        public slots:

            /** @brief Recieves changes in the circulation specifications
              *
              * @param spec the changes specification
              */
            void updateSpecs( const TCirculation & spec );


            /** @brief Recieves changes in the heartrate interval time
              *
              * @param interval the new interval to use
              */
            void updateHeartInterval( int interval );


            /** @brief Signals the reciever (application) with adaptation parameters
              *
              * @param mon the new data
              */
            void adaptCirc(double restFlow = 8.5000e-005,
                           double exerciseFlow=2.5500e-004,
                           double restHeartInterval=60000 / 70,
                           double exerciseHeartInterval=60000 / 140,
                           int nrTransitionbeats = 100,
                           int nrRestBeats = 100,
                           int nrExerciseBeats = 100,
                           bool dorest = true,
                           bool doExercise = true);

            /** @brief Signals the reciever (application) with adaptation parameters
              *
              * @param mon the new data
              */
            void updateAdaptCirc(double restFlow,
                                 double exerciseFlow,
                                 double restHeartInterval,
                                 double exerciseHeartInterval,
                                 int nrTransitionbeats,
                                 int nrRestBeats,
                                 int nrExerciseBeats,
                                 bool dorest,
                                 bool doExercise);


            /** @brief Signals the reciever (application) with adaptation parameters
              *
              * @param mon the new data
              */
            void stopAdaptCirc();


        signals:

            /** @brief Signals the reciever (application) that the state of the thread has changed
              * This signal is send when the thread machine is going to 'paused' state or
              * 'active' state
              *
              * @param state indicates the new state
              */
            void stateChanged( /*statetype*/ int state );


            /** @brief Signals the reciever (application) that the model is (re)loaded
              * Loaded means eigther that a file is (re)loaded or that the
              * model is (re)initialized with default settings.
              *
              * @param type indicates the type of model loaded
              */
            void modelLoaded( /*modeltype*/ int type );


            /** @brief Signals the reciever (application) that a previous action failed
              * All actions are performed asynchronusly. The status of the action can only
              * be and must be reported after it was initiated. This function informs about
              * failures
              *
              * @param type indicates the type of action that failed
              */
            void actionFailed( /*failtype*/ int type );


            /** @brief Signals the reciever (application) that the model is saved to file
              * When the model is saved to file the current model filename (mFilename)
              * is changed to the new saved filename
              */
            void modelSaved( void );


            /** @brief Signals the reciever (application) with initial heartbeat data
              *
              * @param mon the initial data
              */
            void initializedModel( const TCirculation & mon );


            /** @brief Signals the reciever (application) that the initial heartbeats are processed
              * After initialization, a number of heartbeats are processed to reach an (almost)
              * steady state. These heartbeats are prcessed without any delay
              */
            void initialRunDone( void );


            /** @brief Signals the reciever (application) with new heartbeat data
              *
              * @param mon the new data
              */
            void updateBeatInfo( const TCirculation & mon );


            /** @brief Signals the reciever (application) with new heartbeat data
              *
              * @param mon the new data
              */
            void specsCopied();


        protected:

            /** @brief The worker function of the thread.
              * Alle work of the thread is done inside this function
              */
            void run( void );


        private:

            /** @brief performs the heartbeat simulation
              *
              * @param HeartRateInterval the heartrate interval time
              */
            void doHeartbeat( void );


            /** @brief performs the reset action of the model
              */
            void doReset( void );


            /** @brief performs the load action of the model
              */
            void doLoad( void );

            /** @brief performs the adapt action of the model
              */
            void doAdapt( bool adapt );


            /** @brief performs the save action of the model
              */
            void doSave( void );


            /** Member that handles the access control to the interface members
              */
            QMutex mMutex;


            /** Flag indicating that the thread must pasue the heartbear runs
              */
            bool mPause;


            /** Flag indicating that the thread must reset the current model
              */
            bool mReset;


            /** Flag indicating that the thread must save the current model
              */
            bool mSave;


            /** Flag indicating that the thread must load a new model
              */
            bool mLoad;


            /** Flag indicating that the thread must be aborted (stopped)
              */
            bool mAbort;


            /** Flag indicating that the thread must be aborted (stopped)
              */
            bool mAdaptMode;


            /** Counter that keeps track of the number of initializing steps performed
              * After a new model is loaded or the model is initialized with default
              * settings or after a reset this member will be set to 0 tho make the
              * initial runs be performed
              */
            int mNrInitialRuns;


            /** Used to temporary store the name of the file to load a new model from
              */
            QString mFilename;


            /** Used to temporary store the name of the file to store the model to
              */
            QString mSaveFilename;


            /** The temporary storage for the changed specifications
              * These specifications will be copied into the model before the next
              * heartbeat simulation sequence starts
              */
            TCirculation mUpdateSpec;


            /** Flag indicating that mUpdateSpec has changed
              * This flag means that the circulation must be updated before the
              * next heartbeat simulation starts
              */
            bool mUpdateSpecChanged;


            /** The circulation simulation class instantiation
              */
            SCirculation mCirculation;


            /** Holds the interval time for the heartbeats
              */
            int mHeartRateInterval;


            double mRestCO;

            double mExerciseCO;

            double mRestHeartInterval;

            double mExerciseHeartInterval;

            int mTransitionBeatNr;

            int mRestBeatNr;

            int mExerciseBeatNr;

            int mTransitionCnt;

            int mExerciseCnt;

            int mRestCnt;

            bool mDoRest;

            bool mDoExercise;

            double mPrevHeartRate;

            double mPrevCO;

            bool mPrevPressureControl;

            bool mAdaptionOn;
        };
    }
}

#endif

