/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: patch.cpp 784 2013-07-16 17:35:19Z peter $
 *
 */

#include <QDebug>

#include "math.h"
#include "defs.h"
#include "circsettings.h"
#include "patch.h"
#include "chamber.h"
#include "tpatch.h"




const char * MYSARCOMERE_KEY                        = "patch_sarcomere";
const char * PATCH_CHAMBER_TYPE                     = "patch_chamber_type";

const char * SPECIFICATION_PATCH_KEY                = "patch_specification";
const char * SPECIFICATION_REF_PATCH_AREA           = "reference_patch_area";
const char * SPECIFICATION_PATCH_VOLUME             = "patch_volume";

const char * STATE_PATCH_KEY                        = "patch_state";
const char * STATE_CURRENT_PATCH_AREA               = "patch_area";
const char * STATE_PATCH_CURVATURE                  = "patch_curvature";
const char * STATE_PATCH_WALLTENSION                = "patch_wall_tension";
const char * STATE_PATCH_PTRANS                     = "patch_ptrans";
const char * STATE_PATCH_WALLSTIFFNESS              = "patch_wallstiffness";
const char * STATE_PATCH_FIBERSTRAIN                = "patch_fiber_strain";
const char * STATE_PATCH_FIBERSTRESS                = "patch_fiberstress";
const char * STATE_PATCH_FIBERSTIFFNESS             = "patch_fiberstiffness";


const char * ADAPTATION_PATCH_KEY                   = "patch_adaptation";
const char * ADAPTATION_PATCH_AREA_SUM              = "patch_areasum";
const char * ADAPTATION_PATCH_CURVATURESUM          = "patch_curvature_sum";
const char * ADAPTATION_PATCH_FAC_STRAIN            = "patch_fac_strain";
const char * ADAPTATION_PATCH_FAC_SFPAS             = "patch_fac_sfpas";
const char * ADAPTATION_PATCH_FAC_LSACT             = "patch_fact_lsact";
const char * ADAPTATION_PATCH_FAC_SFPASACT          = "patch_fac_sfpasact";
const char * ADAPTATION_PATCH_FAC_WALL              = "patch_fac_wall";
const char * ADAPTATION_PATCH_FAC_Am                = "patch_fac_am";


const int FILE_PATCH_VERSION_1  = 1;
const int FILE_PATCH_VERSION_DEFAULT  = 1;

const char* FILE_PATCH_NAME = "PATCH";

using namespace  Peacs::CircAdapt;
// Constructor
SPatch::SPatch(const SChamber &parent):
    SCircElement(parent.getName()),
    mAmRef(0.0),
    mVWall(0.0),
    mPatchAreaMid(0.0),
    mCm(0.0),
    mT(0.0),
    mPTrans(0.0),
    mDTDAm(0.0),
    mEf(0.0),
    mAreaSum(0.0),
    mCurvatureSum(0.0),
    mFacStrain(1.0),
    mFacSfPasM(1.0),
    mFacLsAct(1.0),
    mFacSfPasAct(1.0),
    mFacVWall(1.0),
    mFacAm(1.0),
    mErrorWallVolume(0.0),
    mErrorWallArea(0.0),
    mParent(parent),
    mSarcomere(*this)

{
    init();
}


SPatch::SPatch(const QString &name, const SChamber & parent):
    SCircElement(name),
    mAmRef(0.0),
    mVWall(0.0),
    mPatchAreaMid(0.0),
    mCm(0.0),
    mT(0.0),
    mPTrans(0.0),
    mDTDAm(0.0),
    mEf(0.0),
    mAreaSum(0.0),
    mCurvatureSum(0.0),
    mFacStrain(1.0),
    mFacSfPasM(1.0),
    mFacLsAct(1.0),
    mFacSfPasAct(1.0),
    mFacVWall(1.0),
    mFacAm(1.0),
    mErrorWallVolume(0.0),
    mErrorWallArea(0.0),
    mParent(parent),
    mSarcomere(name,*this)
{
    init();
}


SPatch::SPatch(const SPatch &patch) :
    SCircElement(        patch),
    mAmRef(             patch.mAmRef),
    mVWall(             patch.mVWall),
    mPatchAreaMid(                patch.mPatchAreaMid),
    mCm(                patch.mCm),
    mT(                 patch.mT),
    mPTrans(            patch.mPTrans),
    mDTDAm(             patch.mDTDAm),
    mEf(                patch.mEf),
    mAreaSum(           patch.mAreaSum),
    mCurvatureSum(      patch.mCurvatureSum),
    mFacStrain(         patch.mFacStrain),
    mFacSfPasM(         patch.mFacSfPasM),
    mFacLsAct (         patch.mFacLsAct),
    mFacSfPasAct(       patch.mFacSfPasAct),
    mFacVWall (         patch.mFacVWall),
    mFacAm(             patch.mFacVWall),
    mErrorWallVolume(   patch.mErrorWallVolume),
    mErrorWallArea(     patch.mErrorWallArea),
    mParent(            patch.mParent),
    mSarcomere(         patch.mSarcomere)
{
    init();
}


const SPatch& SPatch::operator = (const SPatch &patch)
{
    if ( this != &patch)
    {
        SCircElement::operator =(patch);

        mAmRef          = patch.mAmRef;
        mVWall          = patch.mVWall;
        mPatchAreaMid   = patch.mPatchAreaMid;
        mCm             = patch.mCm;
        mT              = patch.mT;
        mPTrans         = patch.mPTrans;
        mDTDAm          = patch.mDTDAm;
        mEf             = patch.mEf;
        mAreaSum        = patch.mAreaSum;
        mCurvatureSum   = patch.mCurvatureSum;
        mFacStrain      = patch.mFacStrain;
        mFacSfPasM      = patch.mFacSfPasM;
        mFacLsAct       = patch.mFacLsAct;
        mFacSfPasAct    = patch.mFacSfPasAct;
        mFacVWall       = patch.mFacVWall;
        mFacAm          = patch.mFacVWall;
        mErrorWallVolume= patch.mErrorWallVolume;
        mErrorWallArea  = patch.mErrorWallArea;
        mSarcomere      = patch.mSarcomere;
//        const_cast <SChamber &> (mParent)  = patch.mParent;
    }
    return *this;
}


void SPatch::init()
{
    if ( mParent.getChamberType() == RIGHT_ATRIA )
    {
        setParRightAtrium();
    }
    else if ( mParent.getChamberType() == LEFT_ATRIA )
    {
        setParLeftAtrium();
    }
    else if ( mParent.getChamberType() == RIGHT_VENTRICLE )
    {
        setParRightVentricle();
    }
    else if ( mParent.getChamberType() == SEPTUM )
    {
        setParSeptumVentricle();
    }
    else if ( mParent.getChamberType() == LEFT_VENTRICLE )
    {
        setParLeftVentricle();
    }
    else
    {
        setParLeftVentricle();
        qDebug() << "undefined patch";
    }
    mSarcomere.init();    
}



void SPatch::copyUserSpecifications(const TPatch &patch)
{
    mAmRef = patch.getRefArea();
    mVWall = patch.getWallVolume();
    mSarcomere.copyUserSpecifications( patch.getSarcomere() );

}

int SPatch::getDefaultVersion(void) const
{
    return FILE_PATCH_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SPatch::getClassName(void) const
{
    return FILE_PATCH_NAME;
}


chamber_type SPatch::getChamberType() const
{
    return mParent.getChamberType();
}

bool SPatch::isAdapationOn() const
{
    return mParent.isAdapationOn();
}

bool SPatch::isVentricle() const
{
    return mParent.isVentricle();
}


// Method to update sheet mechanics
void SPatch::updateSheetMechanics (double area, double curvature, double deltaT)
{
    // Declarations
    double z, z2, z4;    // Ratio curvature / wall thickness
    double TDSf;         // Best work balance TdA/SfdEf
    double dT;		     // Time step in milliseconds
    Time   timestep;     // Time step in microseconds

    // Time step in microseconds
    timestep = (Time) (1000000.0 * deltaT);

    // Time step in milliseconds
    dT = deltaT * 1000.0;

    // Update current area of patch
    mPatchAreaMid = area;

    // Update curvature of patch
    mCm = curvature;

    // Compute ratio curvature / wall thickness
    // Dimensionless z for wall thickness to cavity radius
    // For thick-walled sphere: z = VWall / (2*Vm)
    z  = 1.5 * (mCm / mPatchAreaMid) * mVWall;
    z2 = z * z;
    z4 = z2 * z2;

    // Best work balance TdA/SfdEf
    // Polynomial approximate of ratio wall tension/fiber stress.
    // TDSf= 0.5 * VWall * (1+z2/3+z4/5+....)/Am
    TDSf = 0.5 * (1.0 + 0.27 * z2) * (mVWall / mPatchAreaMid);

    // Compute fiber strain
    mEf = 0.5 * log( mPatchAreaMid / mAmRef ) - z2 / 12.0  - 0.019 * z4;


    // Update fiber mechanics of sarcomere
    mSarcomere.updateFiberMechanics(mEf,deltaT);

    // Compute wall tension (N/m)
    mT = TDSf * mSarcomere.getFiberStress();
    // Compute transmural pressure
    mPTrans = 2.0 * mT * mCm;

    // Compute wall stiffness
    mDTDAm = 0.5 * ( TDSf * mSarcomere.getFiberStiffness() ) / mPatchAreaMid;
}


// Method to initialize adaptation variables
void SPatch::initAdaptation()
{
    // Initialize variables for adaptation
    mAreaSum       = 0.0;  // Sum of area of wall laid as sheet
    mCurvatureSum  = 0.0;  // Sum of curvature
    mFacStrain     = 1.0;  // Ratio target strain / actual strain
    mFacSfPasM     = 1.0;  // Ratio target SfPas  / actual SfPas
    mFacLsAct      = 1.0;  // Ratio target LsAct  / actual LsAct
    mFacSfPasAct   = 1.0;  // Ratio target SfPas  / actual SfPas
    mFacVWall      = 1.0;  // Factor for wall volume adaptation
    mFacAm         = 1.0;  // Factor for wall area adaptation

    // Initialize adaptation variables for sarcomere
    mSarcomere.initAdaptation();
}


// Method to update adaptation variables
void SPatch::updateAdaptation()
{
    // Update sum of area of wall laid as sheet
    mAreaSum += mPatchAreaMid;

    // Update sum of curvature
    mCurvatureSum += mCm;

    // Update adaptation variables for sarcomere
    mSarcomere.updateAdaptation();
}


// Method to perform adaptation
void SPatch::adapt (bool adaptWV, bool adaptSL, bool adaptWA, 
                   double EAmRef, int nrAdaptUpdates)
{
    // Declarations
    double areaMean;       // Mean of area of wall laid as sheet
    double curvatureMean;  // Mean of curvature

    // Compute mean of area of wall laid as sheet
    areaMean = mAreaSum / nrAdaptUpdates;

    // Compute mean of curvature
    curvatureMean = mCurvatureSum / nrAdaptUpdates;

    // Adapt sarcomere and compute adaptation factors
    mSarcomere.adapt(adaptSL,EAmRef,nrAdaptUpdates,
                          mFacStrain,mFacSfPasM,mFacLsAct,mFacSfPasAct);

    // Compute factors for adaptation
    mFacVWall = pow( mFacStrain,  0.25 ) * pow( mFacSfPasM,   -0.19 ) *
                pow( mFacLsAct,  -0.23 ) * pow( mFacSfPasAct,  0.13 );
    mFacAm    = pow( mFacStrain, -0.35 ) * pow( mFacSfPasM,   -0.14 ) *
                pow( mFacLsAct,   2.30 ) * pow( mFacSfPasAct, -1.16 );

    // Adapt wall volume
    if (adaptWV)
    {
        // Adapt wall volume
        // Hypertrophy by stretch of passive matrix,
        // preservation of fiber length
        mVWall = mVWall * mFacVWall;

        // Determine adaptation error for wall volume
        mErrorWallVolume = abs(log(mFacVWall));
    }
    // Adapt wall area
    if (adaptWA)
    {
        // Adapt wall area
        // Strain softening of passive matrix by creep
        mAmRef = mFacAm * mAmRef;

        // Determine adaptation error for wall area
        mErrorWallArea = abs(log(mFacAm));
    }
}


bool SPatch::writeXml(QDomNode &parentNode) const
{
    bool ok = false;
    QDomNode myNode = createXmlEntry(parentNode,getClassName(),getDefaultVersion(),getName());
    if ( !myNode.isNull())
    {
        ok  = writeSpecification(myNode);
        ok &= writeState(myNode);
        QDomNode sarcomereNode = createXmlEntry(myNode,MYSARCOMERE_KEY);
        mSarcomere.writeXml(sarcomereNode);
    }
    return ok;
}


bool SPatch::readXml(const QDomNode &node)
{
    bool ok = SCircElement::readXml(node);
    if ( ok )
    {
        QDomNodeList mySarcomere = node.toElement().elementsByTagName(MYSARCOMERE_KEY);
        if ( mySarcomere.size() == 1 )
        {
            mySarcomere = mySarcomere.at(0).childNodes();
            if ( mySarcomere.size() == 1 )
            {
                mSarcomere.readXml(mySarcomere.at(0));
            }
        }
        else
        {
            ok = false;
        }
    }
    return ok;
}



// Implement method to write specificatioin to file
bool SPatch::writeSpecification(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_PATCH_KEY));

    bool ok = true;
    ok &= addValue(specNode,SPECIFICATION_REF_PATCH_AREA,   mAmRef);
    ok &= addValue(specNode,SPECIFICATION_PATCH_VOLUME,     mVWall);
    return ok;
}


// Implement method to read specification from file
bool SPatch::readSpecification(const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(SPECIFICATION_PATCH_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement specNode = elements.at(0).toElement();
        mAmRef    = readValue(specNode,SPECIFICATION_REF_PATCH_AREA);
        mVWall    = readValue(specNode,SPECIFICATION_PATCH_VOLUME);
        return true;
    }
    return false;
}



// Implement method to write state to file
bool SPatch::writeState(QDomNode &parentNode) const
{
    QDomNode stateNode(createXmlEntry(parentNode,STATE_PATCH_KEY));
    bool ok = true;
    ok &= addValue(stateNode,STATE_CURRENT_PATCH_AREA,  mPatchAreaMid);
    ok &= addValue(stateNode,STATE_PATCH_CURVATURE,     mCm);
    ok &= addValue(stateNode,STATE_PATCH_WALLTENSION,   mT);
    ok &= addValue(stateNode,STATE_PATCH_PTRANS,        mPTrans);
    ok &= addValue(stateNode,STATE_PATCH_WALLSTIFFNESS, mDTDAm);
    return ok;
}


// Implement method to read state from file
bool SPatch::readState(const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(STATE_PATCH_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement stateNode = elements.at(0).toElement();
        mPatchAreaMid     = readValue(stateNode,STATE_CURRENT_PATCH_AREA);
        mCm     = readValue(stateNode,STATE_PATCH_CURVATURE);
        mT      = readValue(stateNode,STATE_PATCH_WALLTENSION);
        mPTrans = readValue(stateNode,STATE_PATCH_PTRANS);
        mDTDAm  = readValue(stateNode,STATE_PATCH_WALLSTIFFNESS);
        return true;
    }
    return false;
}





// Method to set default parameters for left atrium
void SPatch::setParLeftAtrium()
{
    // Set specification parameters
    mAmRef     = 0.0094;       // Area of patch for reference sarcomere length
    mVWall     = 1.2132e-005;  // Wall volume of patch

    // Set simulation state variables
    mPatchAreaMid     =   0.0097;        // Current area of patch
    mCm     =  33.4295;        // Curvature = 1/radius
    mT      =  10.5538;        // Wall tension [N/m]
    mPTrans = 705.6142;        // Transmural pressure
    mDTDAm  =   9.6099e+003;   // Wall stiffness
}


// Method to set default parameters for right atrium
void SPatch::setParRightAtrium()
{
    // Set specification parameters
    mAmRef     = 0.0088;       // Area of patch for reference sarcomere length
    mVWall     = 4.6010e-006;  // Wall volume of patch

    // Set simulation state variables
    mPatchAreaMid     =   0.0082;        // Current area of patch
    mCm     =  35.9257;        // Curvature = 1/radius
    mT      =   2.3890;        // Wall tension [N/m]
    mPTrans = 171.6516;        // Transmural pressure
    mDTDAm  =   4.0053e+003;   // Wall stiffness
}


// Method to set default parameters for left ventricle
void SPatch::setParLeftVentricle ()
{
    if ( mParent.isTriSeg() )
    {
        mAmRef     = 0.0095;       // Area of patch for reference sarcomere length (TriSeg)
        mVWall     = 1.0359e-004;  // Wall volume of patch (TriSeg)
    }
    else
    {
        mAmRef     = 0.0095 + 0.0041;   // Area of patch for reference sarcomere length (NOT TriSeg)
        mVWall     = 1.0359e-004 + 3.6798e-005;  // Wall volume of patch (NOT TriSeg)
    }

    // Set simulation state variables
    mPatchAreaMid     =   0.0098;        // Current area of patch
    mCm     = -27.3279;        // Curvature = 1/radius
    mT      =  14.3749;        // Wall tension [N/m]
    mPTrans =-785.6697;        // Transmural pressure
    mDTDAm  =   2.1649e+004;   // Wall stiffness
}


// Method to set default parameters for septum of ventricle
void SPatch::setParSeptumVentricle ()
{
    // Set specification parameters
    mAmRef     = 0.0041;       // Area of patch for reference sarcomere length
    if ( mParent.isTriSeg() )
    {
        mVWall     = 3.6798e-005;  // Wall volume of patch (TriSeg)
    }
    else
    {
        mVWall     = 0.0;          // Wall volume of patch (NOT TriSeg)
    }

    // Set simulation state variables
    mPatchAreaMid     =   0.0043;        // Current area of patch
    mCm     =  19.9584;        // Curvature = 1/radius
    mT      =  10.8100;        // Wall tension [N/m]
    mPTrans = 431.4993;        // Transmural pressure
    mDTDAm  =   3.8108e+004;   // Wall stiffness

}


// Method to set default parameters for right ventricle
void SPatch::setParRightVentricle ()
{
    // Set specification parameters
    mAmRef     = 0.0116;       // Area of patch for reference sarcomere length
    mVWall     = 4.3367e-005;  // Wall volume of patch

    // Set simulation state variables
    mPatchAreaMid     =   0.0113;        // Current area of patch
    mCm     =  26.4037;        // Curvature = 1/radius
    mT      =   6.7374;        // Wall tension [N/m]
    mPTrans = 355.7847;        // Transmural pressure
    mDTDAm  =   8.4051e+003;   // Wall stiffness

}
