/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: patch.h 789 2013-07-18 13:22:28Z peter $
 *
 */

#ifndef __SPATCH_H_
#define __SPATCH_H_

#include "PCircAdaptDefines.h"
#include "circsettings.h"
#include "circelement.h"
#include "sarcomere.h"


namespace Peacs
{

    namespace CircAdapt
    {
        /**
        Class for a patch of the sheet encapsulating a cavity.
        Inherits from CircElement.
        */

        class SChamber;
        class TPatch;

        class SPatch: public SCircElement {

        public:
            /** @brief default constructor added for debug purposes*/
            SPatch(const SChamber& parent);

            /* @brief Constructor */
            SPatch (const QString &name,const SChamber& parent);

            SPatch (const SPatch &patch);

            virtual ~SPatch() { ; }


            void init();


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TPatch &patch);

            /** @brief Return the minimal activation time of the sarcomere
              *
              */
            inline double getActivationTime(double cycleTime) const { return mSarcomere.getActivationTime(cycleTime); }

            chamber_type getChamberType() const;

            bool isAdapationOn() const;

            bool isVentricle() const;


            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;

            const SPatch& operator = (const SPatch &element);

            SSarcomere & sarcomere() { return mSarcomere; }

            const SSarcomere & getSarcomere() const { return mSarcomere; }

            /// Method to obtain wall tension (N/m)
            inline double getWallTension() const { return mT; }

            /// Method to obtain transmural pressure
            double getTransmuralPressure() const { return mPTrans; }

            /// Method to obtain wall stiffness
            double getWallStiffness() const { return mDTDAm; }

            /// Method to obtain wall volume
            double getWallVolume() const { return mVWall; }

            /// Method to obtain reference patch area
            double getArea() const { return mPatchAreaMid; }

            /// Method to obtain reference patch area
            double getRefArea() const { return mAmRef; }

            /// Method to obtain reference patch area
            double getCurvature() const { return mCm; }

            /// Method to obtain reference patch area
            double getFiberStrain() const { return mEf; }


            /// Method to initialize sheet mechanics
            inline void initSheetMechanics (double tCycleAct, double minActivationTime) { mSarcomere.initFiberMechanics(tCycleAct,minActivationTime); }

            /// Method to update sheet mechanics
            void updateSheetMechanics (double area, double curvature, double deltaT);

            /// Method to initialize adaptation variables
            virtual void initAdaptation();

            /// Method to update adaptation variables
            void updateAdaptation();

            /// Method to perform adaptation
            void adapt (bool adaptWV,
                        bool adaptSL,
                        bool adaptWA,
                        double EAmRef,
                        int nrAdaptUpdates);

            //==================== file functions ===================


            bool writeXml(QDomNode &parentNode) const;

            bool readXml(const QDomNode &node);

            /// Abstract method to write specification to file
            bool writeSpecification (QDomNode &parentNode) const;

            /// Abstract method to read specification from file
            bool readSpecification (const QDomNode &node);

            /// Abstract method to write state to file
            bool writeState (QDomNode &parentNode) const;

            /// Abstract method to read state from file
            bool readState (const QDomNode &node);


        private:
            /// Method to set default parameters for left atrium
            void setParLeftAtrium();

            /// Method to set default parameters for right atrium
            void setParRightAtrium();

            /// Method to set default parameters for left ventricle
            void setParLeftVentricle ();

            /// Method to set default parameters for septum of ventricle
            void setParSeptumVentricle ();

            /// Method to set default parameters for right ventricle
            void setParRightVentricle ();

        private:
            // Specification parameters
            double mAmRef;  // Area of patch for reference sarcomere length
            double mVWall;  // Wall volume of patch

            // Simulation state variables
            double mPatchAreaMid;     // Current area of patch
            double mCm;               // Curvature = 1/radius
            double mT;                // Wall tension (N/m)
            double mPTrans;           // Transmural pressure
            double mDTDAm;            // Wall stiffness

            // Simulation state variables of fiber
            double mEf;     // Fiber strain (natural strain)


            // Variables for adaptation
            double mAreaSum;       // Sum of area of wall laid as sheet
            double mCurvatureSum;  // Sum of curvature
            double mFacStrain;     // Ratio target strain / actual strain
            double mFacSfPasM;     // Ratio target SfPas  / actual SfPas
            double mFacLsAct;      // Ratio target LsAct  / actual LsAct
            double mFacSfPasAct;   // Ratio target SfPas  / actual SfPas
            double mFacVWall;      // Factor for wall volume adaptation
            double mFacAm;         // Factor for wall area adaptation

            // Adaptation error
            double mErrorWallVolume;  // Error for wall volume adaptation
            double mErrorWallArea;    // Error for wall area adaptation

            const SChamber& mParent;

            // Sarcomere describes mechanics of myocardium
            SSarcomere mSarcomere;
        };
    }
}

#endif

