/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: resistance.cpp 907 2013-11-01 16:31:42Z eelco $
 *
 */

#include "resistance.h"
#include "tresistance.h"


const char * SPECIFICATION_RESISTANCE_KEY  = "resistance_specification";
const char * SPECIFICATION_RESISTANCE      = "resistance";
const char * SPECIFICATION_RESISTANCE_TYPE = "resistance_type";


double DEFAULT_RESISTANCE                  = 1.3925e+008;

const char * STATE_RESISTANCE_KEY          = "resistance_state";

const char * ADAPTATION_RESISTANCE_KEY       = "resistance_adaptation";

const int FILE_RESISTANCE_VERSION_1  = 1;
const int FILE_RESISTANCE_VERSION_DEFAULT  = 1;

const char* FILE_RESISTANCE_NAME = "RESISTANCE";

using namespace  Peacs::CircAdapt;
// Constructor

SResistance::SResistance(resistance_type type, SCompliance &prox, SCompliance &dist):
    SFlowDuct(prox,dist),
    mR(0.0),
    mResistanceType(type)
{
    init();
}

SResistance::SResistance(resistance_type type, SCompliance &prox, SCompliance &dist, const QString &name):
    SFlowDuct(prox,dist,name),
    mR(0.0),
    mResistanceType(type)
{
    init();
}


SResistance::SResistance (const SResistance &res):
    SFlowDuct(res),
    mR(res.mR),
    mResistanceType(res.mResistanceType)
{

}

void SResistance::init()
{
    SFlowDuct::init();
    mR = 0;
    if ( mResistanceType == SYSTEMIC )
    {
        setParLRp();
    }
    else
    {
        setParRRp();
    }

}


const SResistance& SResistance::operator = (const SResistance &res)
{
    if ( this != &res )
    {
        SFlowDuct::operator =(res);
        mR = res.mR;
        mResistanceType = res.mResistanceType;
    }
    return *this;
}

void SResistance::copyUserSpecifications(const TResistance &res)
{
    SFlowDuct::copyUserSpecifications(res);
    mR = res.getResistance();
}


int SResistance::getDefaultVersion(void) const
{
    return FILE_RESISTANCE_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SResistance::getClassName(void) const
{
    return FILE_RESISTANCE_NAME;
}


// Method to control blood pressure
void SResistance::controlBloodPressure (double p0, double q0)
{
    // Set peripheral resistance
    mR = (p0 - mDistal.getPressure()) / q0;

    // Update flow
    mQ = (mProximal.getPressure() - mDistal.getPressure()) / mR;

    // Update volume derivatives in proximal compliance
    mProximal.subtractOutgoingFlowToResistance(mQ);

    // Update volume derivatives in distal compliance
    mDistal.addIncomingFlowFromResistance(q0);
}


// Method to set flow
void SResistance::setFlow (double flow)
{
    // Set flow
    mQ = flow;

    // Update volume derivatives in proximate compliance
    mProximal.subtractOutgoingFlowToResistance(mQ);

    // Update volume derivatives in distal compliance
    mDistal.addIncomingFlowFromResistance(mQ);
}


// Method to update flow
void SResistance::updateFlow()
{
    // Update flow
    mQ = (mProximal.getPressure() - mDistal.getPressure()) / mR;

    // Update volume derivatives in proximate compliance
    mProximal.subtractOutgoingFlowToResistance(mQ);

    // Update volume derivatives in distal compliance
    mDistal.addIncomingFlowFromResistance(mQ);
}


// Implement method to write specificatioin to file
bool SResistance::writeSpecification(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_RESISTANCE_KEY));
    // Write specification of compliance
    bool ok = SFlowDuct::writeSpecification(specNode);
    ok &= addValue(specNode, SPECIFICATION_RESISTANCE,mR);

    ok &= addValue(specNode, SPECIFICATION_RESISTANCE_TYPE,(int)mResistanceType);

    return ok;
}


// Implement method to read specification from file
bool SResistance::readSpecification(const QDomNode &parentNode)
{
    QDomNodeList elements = parentNode.toElement().elementsByTagName(SPECIFICATION_RESISTANCE_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement specNode = elements.at(0).toElement();

        mR              = readValue(specNode, SPECIFICATION_RESISTANCE,DEFAULT_RESISTANCE);
        mResistanceType = (resistance_type)(int)readValue(specNode, SPECIFICATION_RESISTANCE_TYPE,SYSTEMIC);
        return true;
    }
    return false;
}


// Method to set default parameters for systemic peripheral resistance (LRp)
void SResistance::setParLRp()
{
    // Set specification parameters
    mR = 1.3925e+008;     // Peripheral resistance

    // Set simulation state variables for flow duct
    mQ = 8.1608e-005;     // Flow
}


// Method to set default parameters for pulmonary peripheral resistance (RRp)
void SResistance::setParRRp ()
{
    // Set specification parameters
    mR = 2.9412e+007;     // Peripheral resistance

    // Set simulation state variables for flow duct
    mQ = 6.1686e-005;     // Flow
}




