/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: sack.cpp 776 2013-07-11 20:33:36Z eelco $
 *
 */

#include "sack.h"
#include "tsack.h"

using namespace  Peacs::CircAdapt;


const char * SPECIFICATION_SACK_KEY             = "sack_specification";
const char * SPECIFICATION_REFERENCE_VOLUME     = "sack_reference_volume";

const char * STATE_SACK_KEY                     = "sack_state";
const char * STATE_SACK_VOLUME                  = "sack_volume";
const char * STATE_SACK_PRESSURE                = "sack_pressure";


const int FILE_SACK_VERSION_1       = 1;
const int FILE_SACK_VERSION_DEFAULT = 1;

const char* FILE_SACK_NAME = "SACK";



// Definition of model constants for pericardial sack
const double SSack::pAdaptSack = SACKPRESSADAPT; // ????
const double SSack::kSack      = SACKSTIFFNESS;  // Stiffness of pericardial sack



SSack::SSack(SAtria &atria, SVentricle &ventricle):
    SCompliance(),
    mVrefSack(0.0),
    mPSack(0.0),
    mVSack(0.0),
    mPSackMax(0.0),
    mErrorSackVolume(0.0),
    mAtria(atria),
    mVentricle(ventricle)
{
    init();
}


// Constructor
SSack::SSack(SAtria &atria, SVentricle &ventricle, const QString name):
    SCompliance(name),
    mVrefSack(0.0),
    mPSack(0.0),
    mVSack(0.0),
    mPSackMax(0.0),
    mErrorSackVolume(0.0),
    mAtria(atria),
    mVentricle(ventricle)
{
    init();
}

// Constructor
SSack::SSack(const SSack &sack) :
    SCompliance(sack),
    mVrefSack( sack.mVrefSack ),
    mPSack( sack.mPSack ),
    mVSack( sack.mVSack),
    mPSackMax( sack.mPSackMax),
    mErrorSackVolume( sack.mErrorSackVolume ),
    mAtria(sack.mAtria),
    mVentricle(sack.mVentricle)
{

}

const SSack& SSack::operator =(const SSack & sack)
{
    if ( this  != &sack )
    {
        SCompliance::operator =(sack);
        mVrefSack           = sack.mVrefSack;
        mPSack              = sack.mPSack;
        mVSack              = sack.mVSack;
        mPSackMax           = sack.mPSackMax;
        mErrorSackVolume    = sack.mErrorSackVolume;
        mAtria              = sack.mAtria;
        mVentricle          = sack.mVentricle;
    }
    return *this;
}

void SSack::init()
{
    SCompliance::init();
    setParPericardium();
}


void SSack::copyUserSpecifications(const TSack &sack)
{
    mVrefSack = sack.getRefVolume();
}


/** @brief returns the default version number */
int SSack::getDefaultVersion(void) const
{
    return FILE_SACK_VERSION_DEFAULT;
}

/** @brief Returns the class name */
const char *SSack::getClassName(void) const
{
    return FILE_SACK_NAME;
}





// Implement method to initialize adaptation variables
void SSack::initAdaptation()
{
    // Initialize maximum pressure of pericardial sack
    mPSackMax = 0.0;
}


// Implement method to update adaptation variables
void SSack::updateAdaptation()
{
    // Update maximum pressure of pericardial sack
    mPSackMax = max(mPSackMax,mPSack);
}


// Method to perform adaptation
void SSack::adapt()
{
    // Declarations
    double Fac_pMax;  // Adaptation factor

    // Compute adaptation factor
    Fac_pMax = pAdaptSack / mPSackMax;

    // Adapt volume of pericardial sack
    mVrefSack = mVrefSack / pow( Fac_pMax, ( 0.3 / kSack ) );

    // Compute adaptation error for pericardial sack
    mErrorSackVolume = abs( log( Fac_pMax ) );
}


double SSack::updateVSackAndComputePSack(double heartVolume)
{
    // Declarations
    double VnormSack;  // Normalized volume of pericardial sack


    // Add wall volume of ventricular septum to sack volume
    // Add wall volume of ventricular septum
    mVSack = heartVolume;

    // Compute normalized volume of pericardial sack
    // VrefSack is constraint reference volume of pericardial sack
    VnormSack = mVSack / mVrefSack;

    // Compute pressure of pericardial sack
    // pAdaptSack is ???
    // kSack represents stiffness of pericardial sack
    mPSack = pAdaptSack * pow(VnormSack,kSack);

    return mPSack;
}


// Implement method to write specificatioin to file
bool SSack::writeSpecification(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_SACK_KEY));
    SCompliance::writeSpecification(specNode);
    addValue(specNode, SPECIFICATION_REFERENCE_VOLUME,       mVrefSack);
    return !specNode.isNull();
}


// Implement method to read specification from file
bool SSack::readSpecification(const QDomNode &parentNode)
{
    QDomNodeList elements = parentNode.toElement().elementsByTagName(SPECIFICATION_SACK_KEY);
    if ( elements.count() == 1 )
    {
        QDomNode specNode = elements.at(0);
        SCompliance::readSpecification(specNode);
        mVrefSack = readValue(specNode, SPECIFICATION_REFERENCE_VOLUME);
        return true;
    }
    return false;
}


// Implement method to write state to file
bool SSack::writeState(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,STATE_SACK_KEY));
    SCompliance::writeSpecification(specNode);
    addValue(specNode, STATE_SACK_VOLUME,       mVSack);
    addValue(specNode, STATE_SACK_PRESSURE,     mPSack);
    return !specNode.isNull();
}


// Implement method to read state from file
bool SSack::readState(const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(STATE_SACK_KEY);
    if ( elements.count() == 1 )
    {
        QDomNode stateNode = elements.at(0);
        SCompliance::readState(stateNode);
        mVSack = readValue(stateNode , STATE_SACK_VOLUME);
        mPSack = readValue(stateNode , STATE_SACK_PRESSURE);
        return true;
    }
    return false;
}




// Method to set default parameters for pericardium
void SSack::setParPericardium()
{
    // Set specification parameters
    mVrefSack = 7.6856e-004;  // Reference volume of sack

    // Set simulation state variables
    mPSack    = 58.5907;      // Pressure of sack
    mVSack    =  6.2024e-004; // Volume of sack
}




