/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: sack.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef _SACK_H_
#define _SACK_H_

#include "chamber.h"
#include "atria.h"
#include "ventricle.h"


namespace Peacs
{

    namespace CircAdapt
    {
        class TSack;
        /**
        Class for pericardial sack. Inherits from Compliance.
        */
        class SSack: public SCompliance {

        public:
            /* @brief Default Constructor */
            SSack(SAtria &atria,SVentricle &ventricle);

            /* @brief Constructor */
            SSack (SAtria &atria,SVentricle &ventricle,const QString name);

            SSack (const SSack & sack);

            /// Destructor
            ~SSack() { ;}

            void init();


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TSack &sack);


            virtual const SSack& operator = (const SSack &sack);


            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;


            /// Implement method to initialize adaptation variables
            void initAdaptation();

            /// Implement method to update adaptation variables
            void updateAdaptation();

            /// Method to perform adaptation
            void adapt();

            double updateVSackAndComputePSack(double heartVolume);


            /** @return the reference volume of the sack
              */
            inline double getRefVolume() const { return mVrefSack; }

            /** @@brief set the reference volume of the sack
              * @param vol the new volume
              */
            inline void setRefVolume(double vol) { mVrefSack = vol; }


            inline double getPSack() const { return mPSack; }

            inline double getVSack() const { return mVSack; }


            //==================== file functions ===================

            /// Abstract method to write specification to file
            virtual bool writeSpecification (QDomNode &parentNode) const;

            /// Abstract method to read specification from file
            virtual bool readSpecification (const QDomNode &parentNode);

            /// Abstract method to write state to file
            bool writeState (QDomNode &parentNode) const;

            /// Abstract method to read state from file
            bool readState (const QDomNode &node);

        private:

            /// Method to set default parameters for pericardium
            void setParPericardium();

        private:
            // Specification parameters
            double mVrefSack;     // Reference volume of sack

            // Simulation state variables
            double mPSack;        // Pressure of sack
            double mVSack;        // Volume of sack

            // Variables for adaptation
            double mPSackMax;     // Maximum pressure of sack

            // Adaptation error
            double mErrorSackVolume;

            SAtria &mAtria;

            SVentricle &mVentricle;

            // Model constants for pericardial sack
            static const double pAdaptSack;  // ????
            static const double kSack;       // Stiffness of pericardial sack
        };
    }
}

#endif
