/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: sarcomere.cpp 789 2013-07-18 13:22:28Z peter $
 *
 */

#include <QDebug>

#include "defs.h"
#include "sarcomere.h"
#include "tsarcomere.h"
#include "patch.h"

#include "pfunctions.h"

using namespace Peacs::Support;
using namespace  Peacs::CircAdapt;


const int NADAPTUPDATES = 3010;

const char * STATE_SARCOMERE_KEY                    = "sarcomere_state";
const char * STATE_ELAPSEDTIME                      = "elapsed_time";
const char * STATE_FIBERSTRAIN                      = "fiber_strain";
const char * STATE_CONTRACTION                      = "contraction";
const char * STATE_CONTRACTION_DOT                  = "contraction_dot";
const char * STATE_INTERNAL_SARCOMERE_LENGTH        = "internal_sarcomere_length";
const char * STATE_INTERNAL_SARCOMERE_LENGTH_DOT    = "internal_sarcomere_length_dot";
const char * STATE_SARCOMERE_LENGTH                 = "sarcomere_length";
const char * STATE_FIBERSTRESS                      = "fiber_stress";
const char * STATE_PASSIVEFIBERSTRESS               = "passive_fiber_stress";
const char * STATE_FIBERSTIFFNESS                   = "fiber_stiffness";

const char * SPECIFICATION_SARCOMERE_KEY            = "sarcomere_specification";
const char * SPECIFICATION_ACTIVATIONDELAY          = "activation_delay";
const char * SPECIFICATION_SARCOMERE_LENGTH         = "reference_sarcomere_length";
const char * SPECIFICATION_PASSIVE_LENGTH           = "reference_passive_length";
const char * SPECIFICATION_PASSIVE_LENGTH_DOT       = "reference_passive_length_dot";
const char * SPECIFICATION_ACTIVE_ISZERO_LENGTH     = "reference_length_zero_active";
const char * SPECIFICATION_PASSIVE_FIBERSTRESS      = "passive_fiber_stress";
const char * SPECIFICATION_ACTIVE_FIBERSTRESS       = "active_fiber_stress";
const char * SPECIFICATION_LENGTH_SERIES_ELEMENT    = "length_series_element";
const char * SPECIFICATION_VMAX                     = "sarcomere_vmax";
const char * SPECIFICATION_TIME_ACTIVE              = "sracomere_time_active";
const char * SPECIFICATION_TR                       = "sarcomere_tr";
const char * SPECIFICATION_TD                       = "sarcomere_td";
const char * SPECIFICATION_CRest                    = "sarcomere_crest";
const char * SPECIFICATION_ADAPTLSBE                = "sarcomere_adapt_lsbe";
const char * SPECIFICATION_ADAPTLSEE                = "sarcomere_adapt_lsee";
const char * SPECIFICATION_ADAPTSFPASMAX            = "sarcomere_adapt_sfpasmax";



double DEFAULT_ACTIVATIONDELAY          = 0.0;
double DEFAULT_SARCOMERE_LENGTH         = 0.0;
double DEFAULT_PASSIVE_LENGTH           = 0.0;
double DEFAULT_PASSIVE_LENGTH_DOT       = 0.0;
double DEFAULT_ACTIVE_ISZERO_LENGTH     = 0.0;
double DEFAULT_PASSIVE_FIBERSTRESS      = 0.0;
double DEFAULT_ACTIVE_FIBERSTRESS       = 0.0;
double DEFAULT_LENGTH_SERIES_ELEMENT    = 0.0;
double DEFAULT_VMAX                     = 0.0;
double DEFAULT_TIME_ACTIVE              = 0.0;
double DEFAULT_TR                       = 0.0;
double DEFAULT_TD                       = 0.0;
double DEFAULT_CRest                    = 0.0;
double DEFAULT_ADAPTLSBE                = 0.0;
double DEFAULT_ADAPTLSEE                = 0.0;
double DEFAULT_ADAPTSFPASMAX            = 0.0;

const int FILE_SARCOMERE_VERSION_1  = 1;
const int FILE_SARCOMERE_VERSION_DEFAULT  = 1;

const char* FILE_SARCOMERE_NAME = "SARCOMERE";


// Constructor
SSarcomere::SSarcomere(const SPatch &parent):
    SCircElement(parent.getName()),
    mActivationDelay(0.0),
    mActivationTime(0.0),
    mPrevTCycle(0.85),
    mLsRef(0.0),
    mLs0Pas(0.0),
    mdLsPas(0.0),
    mSfPas(0.0),
    mLsi0Act(0.0),
    mLenSeriesElement(0.0),
    mSfAct(0.0),
    mvMax(0.0),
    mTimeAct(0.0),
    mTR(0.0),
    mTD(0.0),
    mCRest(0.0),
    mAdapt_LsBe(0.0),
    mAdapt_LsEe(0.0),
    mAdapt_SfPasMax(0.0),
    mTCycle(0.0),
    mElapsedTime(0.0),
    mFiberStrain(0.0),
    mC(0.0),
    mCDot(0.0),
    mLsi(0.0),
    mLsiDot(0.0),
    mLs(0.0),
    mSfPasT(0.0),
    mSf(0.0),
    mDSfDEf(0.0),
    mLsMax(0.0),
    mLsMin(0.0),
    mSfActSum(0.0),
    mLsActSum(0.0),
    mSfPasTMax(0.0),
    mVectorIndex(0.0),
    mErrorSfPas(0.0),
    mLsiVector(NADAPTUPDATES),
    mSfActVector(NADAPTUPDATES),
    mSfPasTVector(NADAPTUPDATES),
    mParent(parent)
{
    init();
}

// Constructor
SSarcomere::SSarcomere(const QString &name, const SPatch &parent):
    SCircElement(name),
    mActivationDelay(0.0),
    mActivationTime(0.0),
    mPrevTCycle(0.85),
    mLsRef(0.0),
    mLs0Pas(0.0),
    mdLsPas(0.0),
    mSfPas(0.0),
    mLsi0Act(0.0),
    mLenSeriesElement(0.0),
    mSfAct(0.0),
    mvMax(0.0),
    mTimeAct(0.0),
    mTR(0.0),
    mTD(0.0),
    mCRest(0.0),
    mAdapt_LsBe(0.0),
    mAdapt_LsEe(0.0),
    mAdapt_SfPasMax(0.0),
    mTCycle(0.0),
    mElapsedTime(0.0),
    mFiberStrain(0.0),
    mC(0.0),
    mCDot(0.0),
    mLsi(0.0),
    mLsiDot(0.0),
    mLs(0.0),
    mSfPasT(0.0),
    mSf(0.0),
    mDSfDEf(0.0),
    mLsMax(0.0),
    mLsMin(0.0),
    mSfActSum(0.0),
    mLsActSum(0.0),
    mSfPasTMax(0.0),
    mVectorIndex(0.0),
    mErrorSfPas(0.0),
    mLsiVector(NADAPTUPDATES),
    mSfActVector(NADAPTUPDATES),
    mSfPasTVector(NADAPTUPDATES),
    mParent(parent)
{
    init();
}
SSarcomere::SSarcomere (const SSarcomere &sarco):
    SCircElement(sarco),
    mActivationDelay    ( sarco.mActivationDelay ),
    mActivationTime     ( sarco.mActivationTime ),
    mPrevTCycle         ( sarco.mPrevTCycle ),
    mLsRef              ( sarco.mLsRef ),
    mLs0Pas             ( sarco.mLs0Pas ),
    mdLsPas             ( sarco.mdLsPas ),
    mSfPas              ( sarco.mSfPas ),
    mLsi0Act            ( sarco.mLsi0Act ),
    mLenSeriesElement   ( sarco.mLenSeriesElement ),
    mSfAct              ( sarco.mSfAct ),
    mvMax               ( sarco.mvMax ),
    mTimeAct            ( sarco.mTimeAct ),
    mTR                 ( sarco.mTR ),
    mTD                 ( sarco.mTD ),
    mCRest              ( sarco.mCRest ),
    mAdapt_LsBe         ( sarco.mAdapt_LsBe ),
    mAdapt_LsEe         ( sarco.mAdapt_LsEe ),
    mAdapt_SfPasMax     ( sarco.mAdapt_SfPasMax ),
    mTCycle             ( sarco.mTCycle ),
    mElapsedTime        ( sarco.mElapsedTime ),
    mFiberStrain        ( sarco.mFiberStrain ),
    mC                  ( sarco.mC ),
    mCDot               ( sarco.mCDot ),
    mLsi                ( sarco.mLsi ),
    mLsiDot             ( sarco.mLsiDot ),
    mLs                 ( sarco.mLs ),
    mSfPasT             ( sarco.mSfPasT ),
    mSf                 ( sarco.mSf ),
    mDSfDEf             ( sarco.mDSfDEf ),
    mLsMax              ( sarco.mLsMax ),
    mLsMin              ( sarco.mLsMin ),
    mSfActSum           ( sarco.mSfActSum ),
    mLsActSum           ( sarco.mLsActSum ),
    mSfPasTMax          ( sarco.mSfPasTMax ),
    mVectorIndex        ( sarco.mVectorIndex ),
    mErrorSfPas         ( sarco.mErrorSfPas ),
    mLsiVector          ( sarco.mLsiVector ),
    mSfActVector        ( sarco.mSfActVector ),
    mSfPasTVector       ( sarco.mSfPasTVector ),
    mParent             ( sarco.mParent)
{

}

const SSarcomere& SSarcomere::operator = (const SSarcomere &sarco)
{
    if ( this != &sarco)
    {
        SCircElement::operator =(sarco);

        mActivationDelay    = sarco.mActivationDelay;
        mActivationTime     = sarco.mActivationTime;
        mPrevTCycle         = sarco.mPrevTCycle;
        mLsRef              = sarco.mLsRef;
        mLs0Pas             = sarco.mLs0Pas;
        mdLsPas             = sarco.mdLsPas;
        mSfPas              = sarco.mSfPas;
        mLsi0Act            = sarco.mLsi0Act;
        mLenSeriesElement   = sarco.mLenSeriesElement;
        mSfAct              = sarco.mSfAct;
        mvMax               = sarco.mvMax;
        mTimeAct            = sarco.mTimeAct;
        mTR                 = sarco.mTR;
        mTD                 = sarco.mTD;
        mCRest              = sarco.mCRest;
        mAdapt_LsBe         = sarco.mAdapt_LsBe;
        mAdapt_LsEe         = sarco.mAdapt_LsEe;
        mAdapt_SfPasMax     = sarco.mAdapt_SfPasMax;
        mTCycle             = sarco.mTCycle;
        mElapsedTime        = sarco.mElapsedTime;
        mFiberStrain        = sarco.mFiberStrain;
        mC                  = sarco.mC;
        mCDot               = sarco.mCDot;
        mLsi                = sarco.mLsi;
        mLsiDot             = sarco.mLsiDot;
        mLs                 = sarco.mLs;
        mSfPasT             = sarco.mSfPasT;
        mSf                 = sarco.mSf;
        mDSfDEf             = sarco.mDSfDEf;
        mLsMax              = sarco.mLsMax;
        mLsMin              = sarco.mLsMin;
        mSfActSum           = sarco.mSfActSum;
        mLsActSum           = sarco.mLsActSum;
        mSfPasTMax          = sarco.mSfPasTMax;
        mVectorIndex        = sarco.mVectorIndex;
        mErrorSfPas         = sarco.mErrorSfPas;
        mLsiVector          = sarco.mLsiVector;
        mSfActVector        = sarco.mSfActVector;
        mSfPasTVector       = sarco.mSfPasTVector;
    }
    return *this;
}

void SSarcomere::init()
{
    if ( mParent.getChamberType() == RIGHT_ATRIA )
    {
        setParRightAtrium();
    }
    else if ( mParent.getChamberType() == LEFT_ATRIA )
    {
        setParLeftAtrium();
    }
    else if ( mParent.getChamberType() == RIGHT_VENTRICLE )
    {
        setParRightVentricle();
    }
    else if ( mParent.getChamberType() == SEPTUM )
    {
        setParSeptumVentricle();
    }
    else if ( mParent.getChamberType() == LEFT_VENTRICLE )
    {
        setParLeftVentricle();
    }
    else
    {
        setParLeftVentricle();
        qDebug() << "undefined sarcomere";
    }
}

void SSarcomere::copyUserSpecifications(const TSarcomere &sarco)
{
    mActivationDelay    = sarco.getActivationdelay();
    mLsRef              = sarco.getLsRef();
    mLs0Pas             = sarco.getLs0Pas();
    mdLsPas             = sarco.getdLsPas();
    mSfPas              = sarco.getSfPas();
    mLsi0Act            = sarco.getLsi0Act();
    mLenSeriesElement   = sarco.getLenSeriesElement();
    mSfAct              = sarco.getSfAct();
    mvMax               = sarco.getvMax();
    mTimeAct            = sarco.getTimeAct();
    mTR                 = sarco.getTR();
    mTD                 = sarco.getTD();
    mCRest              = sarco.getCRest();
    mAdapt_LsBe         = sarco.getAdapt_LsBe();
    mAdapt_LsEe         = sarco.getAdapt_LsEe();
    mAdapt_SfPasMax     = sarco.getAdapt_SfPasMax();

}



void SSarcomere::setParent(const SPatch &parent)
{
    const_cast <SPatch &> (mParent) = parent;
}


int SSarcomere::getDefaultVersion(void) const
{
    return FILE_SARCOMERE_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SSarcomere::getClassName(void) const
{
    return FILE_SARCOMERE_NAME;
}


double SSarcomere::getActivationTime( double cycleTime) const
{
    if ( mParent.isVentricle() )
    {
        return AVDELAY_RELATIVEFROMRATE  * cycleTime + mActivationDelay;
    }
    else
    {
        return mActivationDelay;
    }

}


// Method to initialize fiber mechanics
void SSarcomere::initFiberMechanics(double tCycleAct, double minActivationTime)
{
    // Set actual cycle time (s)
    if ( mTCycle != 0 )
    {
        mPrevTCycle = mTCycle;
    }
    mTCycle = tCycleAct;

    // Set time scale of contraction pulse
    // TODO NEEDS TO BE CHECKED WITH JOOST (Peter)
    if ( mParent.isVentricle() )
    {
        // Determine contraction time for ventricle
        // See Timing.m
        // new according to Joost 12 July 2013
//        mTimeAct = 0.1 * tCycleRef + 0.4 * tCycleAct;
        mTimeAct = 0.085 + 0.4 * tCycleAct;
        if ( mActivationTime  > mPrevTCycle )
        {
            mActivationTime -= mPrevTCycle;
        }
        else
        {
            mActivationTime = AVDELAY_RELATIVEFROMRATE  * tCycleAct + mActivationDelay;
        }
        minActivationTime = qMax(0.0, minActivationTime);
    }
    else
    {
        // Determine contraction time for atrium
        // See Timing.m
        mTimeAct = 0.13;
        if ( minActivationTime >= 0 )
        {
            mActivationTime = mActivationDelay;
        }
        else
        {
            mActivationTime = -minActivationTime + mActivationDelay;
        }
    }

    // Initialize time since beginning of heart beat (s)
    mElapsedTime = 0.0;
}


// Method to update fiber mechanics
void SSarcomere::updateFiberMechanics (double fiber_strain, double deltaT)
{
    // Declarations
    // DEBUG TO DO: Add some explanation
    double tc;        // ??
    double L;         // Normalized sarcomere length
    double tA;        // Activation time
    double tR;        // Rise time
    double tD;        // Decay time
    double T;         // Tension ??
    double ft1;       // ??
    double FL;        // ??
    double SfIso;     // ??
    double kS;        // MSarcomere stiffness (titin?)
    double xEcm;      // Normalized ECM strain
    double xPos;      // ECM stress
    double xSarc;     // MSarcomere passive stress
    double DyDEf;     // ??
    double DSfPasDEf; // ??
    double LNormSe;   // Normalized length of Series Elastic element
    double NoBuckle;  // NoBuckle: Lsi follows Ls
    double x,y;       // Aux variables
    double m1,m2,m3;  // Aux variables
    double tanhx;     // ??

    // Update fiber strain
    mFiberStrain = fiber_strain;

    // Update time since beginning of heart beat (s)
    mElapsedTime += deltaT;

    // Update C
    mC += deltaT * mCDot;

    // Update internal sarcomere length (Lsi)
    mLsi += deltaT * mLsiDot;

    // Compute current time during contraction
    // TODO NEEDS TO BE CHECKED WITH JOOST (Peter)
    if ( mElapsedTime > mActivationTime )
    {
        // Active after excitation during current beat
        tc = mElapsedTime - mActivationTime;
    }
    else
    {
        // Not excited yet during current heart beat
        // Still active from former heart beat
        tc = mPrevTCycle + mElapsedTime - mActivationTime;
    }

    // Series elasticity and sarcomere shortening velocity
    // Ef should be defined by patch
    mLs = exp( mFiberStrain ) * mLsRef;

    // Compute normalized sarcomere length for active contraction
    L = qMax(mLsi - mLsi0Act,0.0001);

    // Compute activation time
    // Activation time lengthens with sarcomere length
    tA = mTimeAct*(0.74 + 0.3*L);

    // Compute rise time
    tR = 0.45 * mTR * mTimeAct;

    // Compute decay time (default 0.3)
    tD = 0.30 * mTD * mTimeAct;

    // Compute active tension
    T  = tc/tR;

    // Compute ft1 (What is ft1 ??)
    x   = qMin(8.0,qMax(0.0,T));
    ft1 = pow(x,3) * exp(-x) * (8-x)*(8-x) * 0.020 / tR;

    // Determine rise of contraction ("amount of Ca++ release")
    // Integral T^n exp(-T) = Gamma(n+1) = n!
    // FL regulates increase of contractility with sarcomere length
    FL    = tanh(4.0 * L*L);
    x     = (tc - tA)/tD;
    y     = qMin(3.0,qMax(-3.0,x));
    tanhx = 0.5 + 0.25*y - 9.2593e-3 * pow(y,3);

    // NOTE: Time confined approximation of 1/(1-e^x) function
    // Approximation of CDot = FL*ft1 - C/((1+exp(-x))*tD)
    // representing 1st order rise and decay of [Ca++]
    mCDot = FL*ft1 - (mC - mCRest) * tanhx/tD;

    // Compute SfIso (??)
    SfIso = mC * L * mSfAct;

    /*------------------------------------------------------
 // BEGIN OLD CODE
 //--- Passive stress and stiffness characteristics ---

 // Compute sarcomere stiffness (Titin?)
 kS  = Ls0Pas / dLsPas;
    aS  = 4.0 * kS*kS;
    a0S = 0.1; // Small linear sarcomere stiffness (can be <0)

 // Compute sarcomere strain EfS0
 // MSarcomere strain may be < 0 (pushing)
 EfS0 = log(Ls/Ls0Pas);

 // Compute sarcomere strain EfS
 if (Ls > Ls0Pas)
 {
  // Positive sarcomere stretch
  // MSarcomere strain EfS equals EfS0
  EfS = EfS0;
 }
 else
 {
  // Zero or negative sarcomere stretch
  // Zero sarcomere strain EfS
  EfS = 0.0;
 }

 // Compute passive stiffness of matrix
 kM        = 10 * kS;
 y         = aS * EfS*EfS + a0S * EfS0 +    exp(Ef*kM - 0.2*kM);
 DyDEf     = 2*aS*EfS     + a0S        + kM*exp(Ef*kM - 0.2*kM);
    SfPasT    = SfPas * y;
    DSfPasDEf = SfPas * DyDEf;
 // END OLD CODE
 ----------------------------------------------*/

    // BEGIN NEW CODE
    /*--- Passive ECM and sarcomere ---*/

    // Compute sarcomere stiffness (Titin?)
    kS   = mLs0Pas / mdLsPas;

    // Normalized ECM strain, may be < 0 (pushing)
    xEcm = kS * log( mLs / mLs0Pas);

    // ECM stress, buckling for negative strain
    if (mLs > mLs0Pas)
    {
        // Positive ECM stretch
        xPos = xEcm;
    }
    else
    {
        // Zero or negative ECM stretch
        xPos = 0.0;
    }

    // Compure sarcomere passive stress
    xSarc = log( mLs / mLsRef ) + 0.1;

    // Compute passive stiffness of matrix
    m1        = 5.0;
    m2        = 0.12;
    m3        = 1.0;
    y         = m2 * ( cosh( m1 * xPos ) - 1.0) + m3 * xSarc;
    DyDEf     = m2 * m1 * kS * sinh(m1*xPos) + m3;
    mSfPasT   = mSfPas * y;
    DSfPasDEf = mSfPas * DyDEf;
    // END NEW CODE

    /*--- Stress/Ls and stiffness ---*/

    // Compute normalized length of Series Elastic element
    LNormSe = qMax( -0.01,( mLs - mLsi ) / mLenSeriesElement );

    // NoBuckle: Lsi follows Ls
    NoBuckle = tanh(10.0 * mC + qMax(0.0,1e-4 * mSfPasT * mSfPasT));
    mLsiDot   = qMax(LNormSe - 1.0, NoBuckle * (LNormSe - 1.0)) * mvMax;
    mSf       = mSfPasT + SfIso * LNormSe;

    // Estimate fiber stiffness
    mDSfDEf = DSfPasDEf + SfIso * (mLs / mLenSeriesElement);
}


// Method to initialize adaptation variables
void SSarcomere::initAdaptation()
{
    // Initialize variables for adaptation
    mLsMax       =  0.0;  // Maximum sarcomere length
    mLsMin       = 10.0;  // Minimum sarcomere length
    mSfActSum    =  0.0;  // Sum of active stress
    mLsActSum    =  0.0;  // Sum of active stress-weighted sarc length
    mSfPasTMax   =  0.0;  // Maximum SfPasT
    mVectorIndex =  0;
}


// Method to update adaptation variables
void SSarcomere::updateAdaptation()
{
    // Declarations
    double SfActCur;  // Current active stress

    // Update maximum internal sarcomere length
    mLsMax = max(mLsMax, mLsi);

    // Update minimum internal sarcomere length
    mLsMin = min(mLsMin,mLsi);

    // Compute current active stress
    SfActCur = max(0.0,mSf - mSfPasT);

    // Compute sum of active stress
    mSfActSum += SfActCur;

    // Compute sum of active stress-weighted sarc length
    mLsActSum += SfActCur * mLsi;

    // Update maximum passive stress
    mSfPasTMax = max(mSfPasTMax,mSfPasT);

    // Store internal sarcomere length in adaptation vector
    mLsiVector[mVectorIndex] = mLsi;

    // Store active stress in adaptation vector
    mSfActVector[mVectorIndex] = SfActCur;

    // Store passive stress in adaptation vector
    mSfPasTVector[mVectorIndex] = mSfPasT;

    // Increase vector index
    if ( mVectorIndex + 1 < mLsiVector.N() )
    {
        mVectorIndex++;
    }
    else
    {
        qDebug() << "Te veel indexen nodig ... (" << mVectorIndex << "van" << NADAPTUPDATES << ")";
    }
}


// Method to perform adaptation and compute adaptation factors
void SSarcomere::adapt(bool adaptSL,
                       double EAmRef,
                       int     nrAdaptUpdates,
                       double &pFacStrain,
                       double &pFacSfPasM,
                       double &pFacLsAct,
                       double &pFacSfPasAct)
{
    // Declarations
    double SfActCur;      // Current value of SfAct (in vector)
    double LsiCur;        // Current value of Lsi (in vector)
    double SfPasTCur;     // Current value of SfPasT (in vector)
    double SfActLsiSum;   // Sum of SfActCur * (Lsi - LsAct)^2
    double SfPasActSum;   // Sum of (1 - (SfPasT / SfPasTMax)) * SfActCur
    double SfActMean;     // Mean active stress
    double LsActMean;     // Mean active stress-weighted sarc length
    double LsAct;         // Active stress-weighted sarc length
    double SfActLsiMean;  // Mean of SfActCur * (Lsi - LsAct)^2
    double DLs;           // Active stress-weighted strain (stdev Ls)
    double SfPasActMean;  // Mean of (1 - (SfPasT / SfPasTMax)) * SfActCur
    double SfPasAct;      // Active stress-weighted ECM stress
    double FacStrain;     // Ratio target strain / actual strain
    double FacSfPasM;     // Ratio target SfPas  / actual SfPas
    double FacSfPasAct;   // Ratio target SfPas  / actual SfPas
    double FacSfPas;      // Ratio target SfPas  / actual SfPas
    double FacLsAct;      // Ratio target LsAct  / actual LsAct
    double Tmax;          // Maximum passive stress (for safety)

    // Consistency check: vectorIndex should equal nrAdaptUpdates
    if (mVectorIndex != nrAdaptUpdates)
    {
        // Error: vectorIndex not equal to nrAdaptUpdates
        qDebug().nospace() << "ERROR: MSarcomere::adapt()   Adaptation vector index out of bound";
    }
    else
    {
        // Compute mean active stress
        SfActMean = mSfActSum / nrAdaptUpdates;

        // Compute mean of active stress-weighted sarc length
        LsActMean = mLsActSum / nrAdaptUpdates;

        // Compute active stress-weighted sarc length
        LsAct = LsActMean / SfActMean;

        // Compute sum of SfActCur * (Lsi - LsAct)^2
        // Compute sum of (1 - (SfPasT / SfPasTMax)) * SfActCur
        SfActLsiSum =  0.0;
        SfPasActSum =  0.0;
        for (int i = 0; i < nrAdaptUpdates; i++)
        {
            // Current values of SfAct, Lsi, SfPasT
            SfActCur  = mSfActVector[i];
            LsiCur    = mLsiVector[i];
            SfPasTCur = mSfPasTVector[i];

            // Compute sum of SfActCur * (Lsi - LsAct)^2
            SfActLsiSum += SfActCur * (LsiCur - LsAct)*(LsiCur - LsAct);

            // Compute sum of (1 - (SfPasT / SfPasTMax)) * SfActCur
            SfPasActSum += (1.0 - (SfPasTCur/mSfPasTMax)) * SfActCur;
        }

        // Compute mean of SfActCur * (Lsi - LsAct)^2
        SfActLsiMean = SfActLsiSum / nrAdaptUpdates;

        // Compute active stress-weighted strain = stdev Ls
        DLs = 3.0 * sqrt(SfActLsiMean/SfActMean);

        // Compute mean of (1 - (SfPasT / SfPasTMax)) * SfAct
        SfPasActMean = SfPasActSum / nrAdaptUpdates;

        // Compute active stress-weighted ECM stress
        SfPasAct = 1.6 * (SfPasActMean / SfActMean);

        // Compute ratios target value / actual value
        FacStrain   = (mAdapt_LsBe - mAdapt_LsEe) / DLs;
        Tmax        = mAdapt_SfPasMax * exp(-1.0 * EAmRef); // Max pass stress, for safety
        FacSfPasM   = Tmax / sqrt(mSfPasTMax * mSfPasTMax + 0.01 * Tmax * Tmax);
        FacLsAct    = (mAdapt_LsEe + 0.6 *(mAdapt_LsBe - mAdapt_LsEe)) / LsAct;
        FacSfPasAct = pow( SfPasAct, 0.0 ); // 0.2

        // Clipping of factors around 1 with range +/- 0.2
        FacStrain   = exp(0.2 *tanh( log( FacStrain   ) / 0.2));
        FacSfPasM   = exp(0.2 *tanh( log( FacSfPasM   ) / 0.2));
        FacLsAct    = exp(0.2 *tanh( log( FacLsAct    ) / 0.2));
        FacSfPasAct = exp(0.2 *tanh( log( FacSfPasAct ) / 0.2));

        // Calculate adaptation factor for sarcomere
        FacSfPas    = pow( FacStrain, -0.55) * pow( FacSfPasM  , 0.51) *
                      pow( FacLsAct , -4.40) * pow( FacSfPasAct,-0.6);

        // Adapt sarcomere
        if ( adaptSL )
        {
            // Shift active sarcomere range over passive matrix
            // Passive stress increases with same sarc length is equivalent with
            // decrease of sarcomere length with same passive stress
            mSfPas = mSfPas * FacSfPas;

            // Determine adaptation error for passive stress
            mErrorSfPas = fabs(log(FacSfPas));
        }

        // Return adaptation factors
        pFacStrain   = FacStrain;
        pFacSfPasM   = FacSfPasM;
        pFacLsAct    = FacLsAct;
        pFacSfPasAct = FacSfPasAct;
    }
}


// Implement method to write specificatioin to file
bool SSarcomere::writeSpecification(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_SARCOMERE_KEY));

    bool ok = true;
    ok &= addValue(specNode,SPECIFICATION_ACTIVATIONDELAY,      mActivationDelay);
    ok &= addValue(specNode,SPECIFICATION_SARCOMERE_LENGTH,     mLsRef );
    ok &= addValue(specNode,SPECIFICATION_PASSIVE_LENGTH,       mLs0Pas );
    ok &= addValue(specNode,SPECIFICATION_PASSIVE_LENGTH_DOT,   mdLsPas );
    ok &= addValue(specNode,SPECIFICATION_PASSIVE_FIBERSTRESS,  mSfPas );
    ok &= addValue(specNode,SPECIFICATION_ACTIVE_ISZERO_LENGTH, mLsi0Act );
    ok &= addValue(specNode,SPECIFICATION_LENGTH_SERIES_ELEMENT,mLenSeriesElement );
    ok &= addValue(specNode,SPECIFICATION_ACTIVE_FIBERSTRESS,   mSfAct );
    ok &= addValue(specNode,SPECIFICATION_VMAX,                 mvMax );
    ok &= addValue(specNode,SPECIFICATION_TIME_ACTIVE,          mTimeAct );
    ok &= addValue(specNode,SPECIFICATION_TR,                   mTR );
    ok &= addValue(specNode,SPECIFICATION_TD,                   mTD );
    ok &= addValue(specNode,SPECIFICATION_CRest,                mCRest);

    ok &= addValue(specNode,SPECIFICATION_ADAPTLSBE,            mAdapt_LsBe );
    ok &= addValue(specNode,SPECIFICATION_ADAPTLSEE,            mAdapt_LsEe );
    ok &= addValue(specNode,SPECIFICATION_ADAPTSFPASMAX,        mAdapt_SfPasMax );
    return ok;
}


// Implement method to read specification from file
bool SSarcomere::readSpecification (const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(SPECIFICATION_SARCOMERE_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement specNode = elements.at(0).toElement();
        mActivationDelay    = readValue(specNode,SPECIFICATION_ACTIVATIONDELAY,         DEFAULT_ACTIVATIONDELAY);
        mLsRef              = readValue(specNode,SPECIFICATION_SARCOMERE_LENGTH,        DEFAULT_SARCOMERE_LENGTH);
        mLs0Pas             = readValue(specNode,SPECIFICATION_PASSIVE_LENGTH,          DEFAULT_PASSIVE_LENGTH);
        mdLsPas             = readValue(specNode,SPECIFICATION_PASSIVE_LENGTH_DOT,      DEFAULT_PASSIVE_LENGTH_DOT);
        mSfPas              = readValue(specNode,SPECIFICATION_PASSIVE_FIBERSTRESS,     DEFAULT_PASSIVE_FIBERSTRESS);
        mLsi0Act            = readValue(specNode,SPECIFICATION_ACTIVE_ISZERO_LENGTH,    DEFAULT_ACTIVE_ISZERO_LENGTH);
        mLenSeriesElement   = readValue(specNode,SPECIFICATION_LENGTH_SERIES_ELEMENT,   DEFAULT_LENGTH_SERIES_ELEMENT);
        mSfAct              = readValue(specNode,SPECIFICATION_ACTIVE_FIBERSTRESS,      DEFAULT_ACTIVE_FIBERSTRESS);
        mvMax               = readValue(specNode,SPECIFICATION_VMAX,                    DEFAULT_VMAX);
        mTimeAct            = readValue(specNode,SPECIFICATION_TIME_ACTIVE,             DEFAULT_TIME_ACTIVE);
        mTR                 = readValue(specNode,SPECIFICATION_TR,                      DEFAULT_TR);
        mTD                 = readValue(specNode,SPECIFICATION_TD,                      DEFAULT_TD);
        mCRest              = readValue(specNode,SPECIFICATION_CRest,                   DEFAULT_CRest);
        mAdapt_LsBe         = readValue(specNode,SPECIFICATION_ADAPTLSBE,               DEFAULT_ADAPTLSBE);
        mAdapt_LsEe         = readValue(specNode,SPECIFICATION_ADAPTLSEE,               DEFAULT_ADAPTLSEE);
        mAdapt_SfPasMax     = readValue(specNode,SPECIFICATION_ADAPTSFPASMAX,           DEFAULT_ADAPTSFPASMAX);
        return true;
    }
    return false;
}



// Implement method to write state to file
bool SSarcomere::writeState(QDomNode &parentNode) const
{
    QDomNode stateNode(createXmlEntry(parentNode,STATE_SARCOMERE_KEY));

    bool ok = true;
    ok &= addValue(stateNode,STATE_ELAPSEDTIME,                     mElapsedTime);
    ok &= addValue(stateNode,STATE_FIBERSTRAIN,                     mFiberStrain);
    ok &= addValue(stateNode,STATE_CONTRACTION,                     mC);
    ok &= addValue(stateNode,STATE_CONTRACTION_DOT,                 mCDot);
    ok &= addValue(stateNode,STATE_INTERNAL_SARCOMERE_LENGTH,       mLsi);
    ok &= addValue(stateNode,STATE_INTERNAL_SARCOMERE_LENGTH_DOT,   mLsiDot);
    ok &= addValue(stateNode,STATE_SARCOMERE_LENGTH,                mLs);
    ok &= addValue(stateNode,STATE_PASSIVEFIBERSTRESS,              mSfPasT);
    ok &= addValue(stateNode,STATE_FIBERSTRESS,                     mSf);
    ok &= addValue(stateNode,STATE_FIBERSTIFFNESS,                  mDSfDEf);
    return ok;
}


// Implement method to read state from file
bool SSarcomere::readState (const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(STATE_SARCOMERE_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement stateNode = elements.at(0).toElement();
        mElapsedTime    = readValue(stateNode,STATE_ELAPSEDTIME);
        mFiberStrain    = readValue(stateNode,STATE_FIBERSTRAIN);
        mC              = readValue(stateNode,STATE_CONTRACTION);
        mCDot           = readValue(stateNode,STATE_CONTRACTION_DOT);
        mLsi            = readValue(stateNode,STATE_INTERNAL_SARCOMERE_LENGTH);
        mLsiDot         = readValue(stateNode,STATE_INTERNAL_SARCOMERE_LENGTH_DOT);
        mLs             = readValue(stateNode,STATE_SARCOMERE_LENGTH);
        mSfPasT         = readValue(stateNode,STATE_PASSIVEFIBERSTRESS);
        mSf             = readValue(stateNode,STATE_FIBERSTRESS);
        mDSfDEf         = readValue(stateNode,STATE_FIBERSTIFFNESS);
        return true;
    }
    return false;
}




// Method to set default parameters for left atrium
void SSarcomere::setParLeftAtrium()
{
    // Set specification parameters
    mActivationDelay  =  0.0200;   // Activation delay [s]
    mActivationTime   =  0.0000;
    // LsRef            =  2.0023;  // JOOST
    mLsRef            =  2.0;      // ParRef.mat and ParExc.mat
    mLs0Pas           =  1.8000;
    // dLsPas           =  0.6000;  // JOOST
    mdLsPas           =  1.0;      // ParRef.mat and ParExc.mat
    // SfPas            =  2.8495e+004; // JOOST
    mSfPas            =  7.5075e+004;  // ParExc.mat
    mLsi0Act          =  1.5100;   // MSarcomere length when active stress is zero
    mLenSeriesElement =  0.0400;   // Reference length of series elastic element
    mSfAct            =  8.4e+004;
    // vMax             = 14.0;   // JOOST
    mvMax             = 10.0;      // ParRef.mat and ParExc.mat
    // TimeAct          =  0.2000;   // Time scale of contraction pulse JOOST
    mTimeAct          =  0.1300;   // Time scale of contraction pulse ParRef.mat
    mTR               =  0.5000;
    mTD               =  0.5000;
    mCRest            =  0.0200;
    mAdapt_LsBe       =  2.2000;   // For adaptation
    // Adapt_LsEe       =  1.6500;   // For adaptation JOOST
    mAdapt_LsEe       =  1.7500;   // For adaptation
    mAdapt_SfPasMax   =  100000;   // For adaptation

    // Set simulation state variables
    mElapsedTime =  0.0;           // Time since beginning of heart beat (ms)
    mFiberStrain          =  0.0168;        // Fiber strain
    mC           =  0.0200;
    mCDot        = -1.9055e-008;
    mLsi         =  1.9928;        // Internal sarcomere length
    mLsiDot      =  1.1718;        // d/dt Internal sarcomere length
    mLs          =  2.0362;        // MSarcomere length
    mSfPasT      =  1.6057e+004;
    mSf          =  1.6936e+004;   // Fiber stress
    mDSfDEf      =  3.0055e+005;   // Fiber stiffness
}


// Method to set default parameters for right atrium
void SSarcomere::setParRightAtrium()
{
    // Set specification parameters
    mActivationDelay  =  0.0000;   // Activation delay [s]
    mActivationTime   =  0.2000;
    // LsRef            =  2.0018;  // JOOST
    mLsRef            =  2.0;      // ParRef.mat and ParExc.mat
    mLs0Pas           =  1.8000;
    // dLsPas           =  0.6000;  // JOOST
    mdLsPas           =  1.0;      // ParRef.mat and ParExc.mat
    // SfPas            =  3.6974e+004;  // JOOST
    mSfPas            =  9.5406e+004;  // ParExc.mat
    mLsi0Act          =  1.5100;   // MSarcomere length when active stress is zero
    mLenSeriesElement =  0.0400;   // Reference length of series elastic element
    mSfAct            =  8.4e+004;
    // vMax             = 14.0;     // JOOST
    mvMax             = 10.0;      // ParRef.mat and ParExc.mat
    // TimeAct          =  0.2000;   // Time scale of contraction pulse JOOST
    mTimeAct          =  0.1300;   // Time scale of contraction pulse ParRef.mat
    mTR               =  0.5000;
    mTD               =  0.5000;
    mCRest            =  0.0200;
    mAdapt_LsBe       =  2.2000;   // For adaptation
    // Adapt_LsEe       =  1.6500;   // For adaptation JOOST
    mAdapt_LsEe       =  1.7500;   // For adaptation
    mAdapt_SfPasMax   =  100000;   // For adaptation

    // Set simulation state variables
    mElapsedTime =  0.0;           // Time since beginning of heart beat (ms)
    mFiberStrain          = -0.0312;        // Fiber strain
    mC           =  0.0200;
    mCDot        =  0.0095;
    mLsi         =  1.8955;        // Internal sarcomere length
    mLsiDot      =  1.7319;        // d/dt Internal sarcomere length
    mLs          =  1.9404;        // MSarcomere length
    mSfPasT      =  7.8231e+003;
    mSf          =  8.5512e+003;   // Fiber stress
    mDSfDEf      =  2.3616e+005;   // Fiber stiffness
}


// Method to set default parameters for left ventricle
void SSarcomere::setParLeftVentricle ()
{
    // Set specification parameters
    mActivationDelay  =  0.0000;   // Activation delay [s]
    mActivationTime   =  0.2000;
    // LsRef            =  2.0002;  // JOOST
    mLsRef            =  2.0;      // ParRef.mat and ParExc.mat
    mLs0Pas           =  1.8000;
    mdLsPas           =  0.6000;
    // SfPas            =  3.8931e+003; // JOOST
    mSfPas            =  5.2958e+003;  // ParExc.mat
    mLsi0Act          =  1.5100;   // MSarcomere length when active stress is zero
    mLenSeriesElement =  0.0400;   // Reference length of series elastic element
    // SfAct            =  1.0e+005;  // JOOST
    mSfAct            =  1.2e+005; // ParRef.mat and ParExc.mat
    mvMax             =  7.0;
    if ( mParent.isAdapationOn() )
    {
        mTimeAct      =  0.2554;   // Time scale of contraction pulse
    }
    else
    {
        mTimeAct      =  0.4250;   // Time scale of contraction pulse
    }
    mTR               =  0.2500;
    mTD               =  0.2500;
    mCRest            =  0.0200;
    mAdapt_LsBe       =  2.2000;   // For adaptation
    mAdapt_LsEe       =  1.7500;   // For adaptation
    mAdapt_SfPasMax   =  10000;    // For adaptation

    // Set simulation state variables
    mElapsedTime =  0.0;           // Time since beginning of heart beat (ms)
    mFiberStrain          =  0.0030;        // Fiber strain
    mC           =  0.0201;
    mCDot        = -0.0032;
    mLsi         =  1.9669;        // Internal sarcomere length
    mLsiDot      = -0.1306;        // d/dt Internal sarcomere length
    mLs          =  2.0062;        // MSarcomere length
    mSfPasT      =  1.7006e+003;
    mSf          =  2.6019e+003;   // Fiber stress
    mDSfDEf      =  7.7165e+004;   // Fiber stiffness
}


// Method to set default parameters for septum of ventricle
void SSarcomere::setParSeptumVentricle ()
{
    // Set specification parameters
    mActivationDelay  =  0.0000;   // Activation delay [s]
    mActivationTime   =  0.2000;
    // LsRef            =  2.0007;  // JOOST
    mLsRef            =  2.0;      // ParRef.mat and ParExc.mat
    mLs0Pas           =  1.8000;
    mdLsPas           =  0.6000;
    // SfPas            =  3.0297e+003;  // JOOST
    mSfPas            =  5.6138e+003;  // ParExc.mat
    mLsi0Act          =  1.5100;   // MSarcomere length when active stress is zero
    mLenSeriesElement =  0.0400;   // Reference length of series elastic element
    // SfAct            =  1.0e+005;  // JOOST
    mSfAct            =  1.2e+005; // ParRef.mat and ParExc.mat
    mvMax             =  7.0;
    if ( mParent.isAdapationOn() )
    {
        mTimeAct      =  0.2554;   // Time scale of contraction pulse
    }
    else
    {
        mTimeAct      =  0.4250;   // Time scale of contraction pulse
    }
    mTR               =  0.2500;
    mTD               =  0.2500;
    mCRest            =  0.0200;
    mAdapt_LsBe       =  2.2000;   // For adaptation
    mAdapt_LsEe       =  1.7500;   // For adaptation
    // Adapt_SfPasMax   =  8.0e+003; // For adaptation JOOST
    mAdapt_SfPasMax   =  10000; // For adaptation

    // Set simulation state variables
    mElapsedTime =  0.0;           // Time since beginning of heart beat (ms)
    mFiberStrain          =  0.0112;        // Fiber strain
    mC           =  0.0201;
    mCDot        = -0.0034;
    mLsi         =  1.9843;        // Internal sarcomere length
    mLsiDot      = -0.1742;        // d/dt Internal sarcomere length
    mLs          =  2.0233;        // MSarcomere length
    mSfPasT      =  1.5374e+003;
    mSf          =  2.4673e+003;   // Fiber stress
    mDSfDEf      =  7.4364e+004;   // Fiber stiffness
}


// Method to set default parameters for right ventricle
void SSarcomere::setParRightVentricle ()
{
    // Set specification parameters
    mActivationDelay  =  0.0000;   // Activation delay [s]
    mActivationTime   =  0.2000;
    // LsRef            =  1.9980;  // JOOST
    mLsRef            =  2.0;      // ParRef.mat and ParExc.mat
    mLs0Pas           =  1.8000;
    mdLsPas           =  0.6000;
    // SfPas            =  8.4189e+003;  // JOOST
    mSfPas            =  6.4545e+003;  // ParExc.mat
    mLsi0Act          =  1.5100;   // MSarcomere length when active stress is zero
    mLenSeriesElement =  0.0400;   // Reference length of series elastic element
    // SfAct            =  1.0e+005;  // JOOST
    mSfAct            =  1.2e+005; // ParRef.mat and ParExc.mat
    mvMax             =  7.0;
    if ( mParent.isAdapationOn() )
    {
        mTimeAct      =  0.2554;   // Time scale of contraction pulse
    }
    else
    {
        mTimeAct      =  0.4250;   // Time scale of contraction pulse
    }
    mTR               =  0.2500;
    mTD               =  0.2500;
    mCRest            =  0.0200;
    mAdapt_LsBe       =  2.2000;   // For adaptation
    mAdapt_LsEe       =  1.7500;   // For adaptation
    // Adapt_SfPasMax   =  2.0e+004; // For adaptation JOOST
    mAdapt_SfPasMax   =  10000;    // For adaptation

    // Set simulation state variables
    mElapsedTime =  0.0;           // Time since beginning of heart beat (ms)
    mFiberStrain          = -0.0124;        // Fiber strain
    mC           =  0.0201;
    mCDot        = -0.0028;
    mLsi         =  1.9336;        // Internal sarcomere length
    mLsiDot      = -0.0433;        // d/dt Internal sarcomere length
    mLs          =  1.9734;        // MSarcomere length
    mSfPasT      =  2.6546e+003;
    mSf          =  3.5004e+003;   // Fiber stress
    mDSfDEf      =  9.9001e+004;   // Fiber stiffness
}

