/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: sarcomere.h 789 2013-07-18 13:22:28Z peter $
 *
 */

#ifndef __SSARCOMERE_H_
#define __SSARCOMERE_H_


#include "pvector.h"
#include "circelement.h"

namespace Peacs
{

    namespace CircAdapt
    {
        class SPatch;
        class TSarcomere;

        /**
        Class for mechanics of myocardium.
        Inherits from CircElement.
        */
        class SSarcomere: public SCircElement {

        public:
            /** @brief default constructor added for debug purposes*/
            SSarcomere(const SPatch & parent);

            /* @brief Constructor */
            SSarcomere (const QString &name, const SPatch & parent);

            SSarcomere (const SSarcomere &sarco);

            virtual ~SSarcomere() { ;}

            void init();

            /** @brief Return the minimal time of any of the patches inside this chamber
              *
              */
            double getActivationTime(double cycleTime) const;

            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TSarcomere &sarco);


            const SSarcomere& operator = (const SSarcomere &sarco);

            const SPatch &  getParent() const { return mParent;}

            void setParent(const SPatch &parent);


            /// Method to obtain fiber stress
            double getFiberStress() const { return mSf; }

            /// Method to obtain fiber stiffness
            double getFiberStiffness() const { return mDSfDEf; }

            /// Method to initialize fiber mechanics
            void initFiberMechanics (double tCycleAct, double tCycleRef);

            /// Method to update fiber mechanics
            void updateFiberMechanics (double fiber_strain, double deltaT);

            /// Method to initialize adaptation variables
            void initAdaptation();

            /// Method to update adaptation variables
            void updateAdaptation();

            /// Method to perform adaptation and compute adaptation factors
            void adapt(bool     adaptSL,
                       double   EAmRef,
                       int      nrAdaptUpdates,
                       double &pFacStrain,
                       double &pFacSfPasM,
                       double &pFacLsAct,
                       double &pFacSfPasAct);

            //==================== file functions ===================

            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;
            /// Abstract method to write specification to file
            bool writeSpecification (QDomNode &parentNode) const;

            /// Abstract method to read specification from file
            bool readSpecification (const QDomNode &node);

            /// Abstract method to write state to file
            bool writeState (QDomNode &parentNode) const;

            /// Abstract method to read state from file
            bool readState (const QDomNode &node);

            inline void setActiovationDelay(double delay) { mActivationDelay = delay; }

            //------------------------ memeber access function ------------

            inline double getActivationdelay() const   { return mActivationDelay; }
            inline double getLsRef()   const           { return mLsRef; }
            inline double getLs0Pas()  const           { return mLs0Pas; }
            inline double getdLsPas()  const           { return mdLsPas; }
            inline double getSfPas()   const           { return mSfPas; }
            inline double getLsi0Act() const           { return mLsi0Act;}         // Sarcomere length when active stress is zero
            inline double getLenSeriesElement() const  { return mLenSeriesElement;}// Reference length of series elastic element

            inline double getSfAct() const             { return mSfAct;}
//            void setSfAct(double sfact)         { mSfAct = sfact;}

            inline double getvMax() const              { return mvMax; }
            inline double getTimeAct() const           { return mTimeAct; }        // Time scale of contraction pulse
            inline double getTR() const                { return mTR; }
            inline double getTD() const                { return mTD; }
            inline double getCRest() const             { return mCRest; }
            inline double getAdapt_LsBe() const        { return mAdapt_LsBe;}      // For adaptation
            inline double getAdapt_LsEe() const        { return mAdapt_LsEe; }     // For adaptation
            inline double getAdapt_SfPasMax() const    { return mAdapt_SfPasMax; } // For adaptation

            // Simulation state variables
            inline double getFiberStrain() const       { return mFiberStrain; }    // Fiber strain
            inline double getC() const                 { return mC;}               // Contraction
            inline double getCDot() const              { return mCDot;}            // d/dt Contraction
            inline double getLsi() const               { return mLsi;  }           // Internal sarcomere length
            inline double getLsiDot() const            { return mLsiDot;}          // d/dt Internal sarcomere length
            inline double getLs() const                { return mLs;  }            // Sarcomere length
            inline double getSfPasT() const            { return mSfPasT; }         // Fiber passive stress
            inline double getSf() const                { return mSf; }             // Fiber stress
            inline double getDSfDEf() const            { return mDSfDEf; }         // Fiber stiffness

        private:
            /// Method to set default parameters for left atrium
            void setParLeftAtrium();

            /// Method to set default parameters for right atrium
            void setParRightAtrium();

            /// Method to set default parameters for left ventricle
            void setParLeftVentricle ();

            /// Method to set default parameters for septum of ventricle
            void setParSeptumVentricle ();

            /// Method to set default parameters for right ventricle
            void setParRightVentricle ();

        private:
            // Specification parameters
            // DEBUG TO DO: Add some explanation
            // DEBUG TO DO: List of activation times to replace ActivationDelay
            double mActivationDelay;
            double mActivationTime;
            double mPrevTCycle;
            double mLsRef;
            double mLs0Pas;
            double mdLsPas;
            double mSfPas;
            double mLsi0Act;          // Sarcomere length when active stress is zero
            double mLenSeriesElement; // Reference length of series elastic element
            double mSfAct;
            double mvMax;
            double mTimeAct;          // Time scale of contraction pulse
            double mTR;
            double mTD;
            double mCRest;
            double mAdapt_LsBe;       // For adaptation
            double mAdapt_LsEe;       // For adaptation
            double mAdapt_SfPasMax;   // For adaptation

            // Simulation state variables
            double mTCycle;      // Actual cycle time (s)
            double mElapsedTime; // Time since beginning of heart beat (s)
            double mFiberStrain; // Fiber strain
            double mC;           // Contraction
            double mCDot;        // d/dt Contraction
            double mLsi;         // Internal sarcomere length
            double mLsiDot;      // d/dt Internal sarcomere length
            double mLs;          // Sarcomere length
            double mSfPasT;      // Fiber passive stress
            double mSf;          // Fiber stress
            double mDSfDEf;      // Fiber stiffness

            // Variables for adaptation
            double mLsMax;        // Maximum sarcomere length
            double mLsMin;        // Minimum sarcomere length
            double mSfActSum;     // Sum of active stress
            double mLsActSum;     // Sum of active stress-weighted sarc length
            double mSfPasTMax;    // Maximum SfPasT
            int    mVectorIndex;  // Index for adaptation vectors

            // Adaptation error
            double mErrorSfPas;   // Error for passive stress adaptation

            // Adaptation vectors
            PVector mLsiVector;    // Store Lsi
            PVector mSfActVector;  // Store active stress
            PVector mSfPasTVector; // Store passive stress

            const SPatch & mParent;
        };
    }
}

#endif
