#include <QDebug>

#include "math.h"
#include "defs.h"
#include "circsettings.h"
#include "tube.h"
#include "ttube.h"

using namespace  Peacs::CircAdapt;


// Definition of model constants
const double STube::rhob = BLOODDENSITY; // Density of blood




const char * TUBE_TYPE                          = "tube_type";

const char * STATE_TUBEKEY                      = "tube_state";
const char * SPECIFICATION_TUBE_KEY             = "tube_specification";
const char * SPECIFICATION_STIFFNESS_EXPONENT   = "stiffness_exponent";
const char * SPECIFICATION_MEAN_REST_FLOW       = "mean_rest_flow";
const char * SPECIFICATION_MEAN_REST_PRESSURE   = "mean_rest_pressure";
const char * SPECIFICATION_TUBE_VESSEL_LENGTH   = "tube_vessel_length";
const char * SPECIFICATION_TUBE_PMAX            = "tube_pmax";
const char * SPECIFICATION_TUBE_AREA_WALL       = "tube_wall_area";
const char * SPECIFICATION_TUBE_MEAN_AREA       = "tube_mean_area";

const double DEFAULT_STIFFNESS_EXPONENT         = 10.0;
const double DEFAULT_MEAN_REST_FLOW             = 8.5e-5;
const double DEFAULT_MEAN_REST_PRESSURE         = 381.4712;
const double DEFAULT_TUBE_VESSEL_LENGTH         = 0.4001;
const double DEFAULT_TUBE_PMAX                  = 6.2551e3;
const double DEFAULT_TUBE_AREA_WALL             = 2.9968e-5;
const double DEFAULT_TUBE_MEAN_AREA             = 5.0e-4;

const char * ADAPT_TUBE_KEY                     = "tube_adaptation";
const char * ADAPT_MAX_WALL_STRESS              = "tube_max_wall_stress";
const char * ADAPT_MEAN_FLOW_VELOCITY           = "tube_mean_flow_velocity";
const char * ADAPT_SHOCKWAVE_VELOCITY           = "tube_shockwave_velocity";

//const double DEFAULT_MAX_WALL_STRESS            = "tube_max_wall_stress";
//const double DEFAULT_MEAN_FLOW_VELOCITY         = "tube_mean_flow_velocity";
//const double DEFAULT_SHOCKWAVE_VELOCITY         = "tube_shockwave_velocity";


const int FILE_TUBE_VERSION_1  = 1;
const int FILE_TUBE_VERSION_DEFAULT  = 1;

const char* FILE_TUBE_NAME = "TUBE";



// Constructor
STube::STube(tube_type type):
    SCompliance(),
    mK(0.0),
    mQ0(0.0),
    mP0(0.0),
    mVesselLength(0.0),
    mPmax(0.0),
    mAWall(0.0),
    mA0(0.0),
    mWallStressAdapt(0.0),
    mVFlowMeanAdapt(0.0),
    mVImpactAdapt(0.0),
    mQRemodSum(0.0),
    mQPosRemodSum(0.0),
    mQNegRemodSum(0.0),
    mAreaSum(0.0),
    mPressureSquareSum(0.0),
    mPMaxMax(0.0),
    mWallStressMax(0.0),
    mTubeType(type)
{
    init();
}


STube::STube(tube_type type, const QString &name):
    SCompliance(name),
    mK(0.0),
    mQ0(0.0),
    mP0(0.0),
    mVesselLength(0.0),
    mPmax(0.0),
    mAWall(0.0),
    mA0(0.0),
    mWallStressAdapt(0.0),
    mVFlowMeanAdapt(0.0),
    mVImpactAdapt(0.0),
    mQRemodSum(0.0),
    mQPosRemodSum(0.0),
    mQNegRemodSum(0.0),
    mAreaSum(0.0),
    mPressureSquareSum(0.0),
    mPMaxMax(0.0),
    mWallStressMax(0.0),
    mTubeType(type)
{
    init();
}

STube::STube(const STube &tube):
    SCompliance(tube),
    mK(tube.mK),
    mQ0(tube.mQ0),
    mP0(tube.mP0),
    mVesselLength(tube.mVesselLength),
    mPmax(tube.mPmax),
    mAWall(tube.mAWall),
    mA0(tube.mA0),
    mWallStressAdapt(tube.mWallStressAdapt),
    mVFlowMeanAdapt(tube.mVFlowMeanAdapt),
    mVImpactAdapt(tube.mVImpactAdapt),
    mQRemodSum(tube.mQRemodSum),
    mQPosRemodSum(tube.mQPosRemodSum),
    mQNegRemodSum(tube.mQNegRemodSum),
    mAreaSum(tube.mAreaSum),
    mPressureSquareSum(tube.mPressureSquareSum),
    mPMaxMax(tube.mPMaxMax),
    mWallStressMax(tube.mWallStressMax),
    mTubeType(tube.mTubeType)
{

}


const STube& STube::operator = (const STube &tube)
{
    if ( this != &tube )
    {
        SCompliance::operator =(tube);
        mK                  = tube.mK;
        mQ0                 = tube.mQ0;
        mP0                 = tube.mP0;
        mVesselLength       = tube.mVesselLength;
        mPmax               = tube.mPmax;
        mAWall              = tube.mAWall;
        mA0                 = tube.mA0;
        mWallStressAdapt    = tube.mWallStressAdapt;
        mVFlowMeanAdapt     = tube.mVFlowMeanAdapt;
        mVImpactAdapt       = tube.mVImpactAdapt;
        mQRemodSum          = tube.mQRemodSum;
        mQPosRemodSum       = tube.mQPosRemodSum;
        mQNegRemodSum       = tube.mQNegRemodSum;
        mAreaSum            = tube.mAreaSum;
        mPressureSquareSum  = tube.mPressureSquareSum;
        mPMaxMax            = tube.mPMaxMax;
        mWallStressMax      = tube.mWallStressMax;
        mTubeType           = tube.mTubeType;
    }
    return *this;
}

void STube::init()
{
    SCompliance::init();
    if ( mTubeType == AORTA_TUBE )
    {
        setParLArt();
    }
    else if ( mTubeType == PULMONARY_VEIN_TUBE )
    {
        setParLVen();
    }
    else if ( mTubeType == PULMONARY_ARTERY_TUBE )
    {
        setParRArt();
    }
    else
    {
        setParRVen();
    }
}

void STube::copyUserSpecifications(const TTube &tube)
{
    SCompliance::copyUserSpecifications(tube);
    mK                  = tube.stiffnessWall();
    mQ0                 = tube.meanRestFlow();
    mP0                 = tube.meanRestPressure();
    mVesselLength       = tube.meanVesselLength();
    mPmax               = tube.Pmax();
    mAWall              = tube.Awall();
    mA0                 = tube.A0();
    mWallStressAdapt    = tube.wallStressAdapt();
    mVFlowMeanAdapt     = tube.vFlowMeanAdapt();
    mVImpactAdapt       = tube.vImpactAdapt();
}

/** @brief returns the default version number */
int STube::getDefaultVersion(void) const
{
    return FILE_TUBE_VERSION_DEFAULT;
}

/** @brief Returns the class name */
const char *STube::getClassName(void) const
{
    return FILE_TUBE_NAME;
}






// Method to update tube dynamics
void STube::updateTubeDynamics (double deltaT)
{

    // Declarations
    double AWalld3;  // One-third of cross-sectional area of wall
    double aux;      // Help variable

    // Update volume of tube
    mVolume += deltaT * mVDot;

    // Compute one-third of cross-sectional area of wall
    AWalld3 = mAWall / 3.0;

    // Compute mean cross-section of tube (safe for A0 < 0)
    mA = max(0.001 * mA0, mVolume / mVesselLength );


    // Compute pressure (safe for collapse)
    aux = pow((mA + AWalld3)/(mA0 + AWalld3),(mK/3.0)-1.0);
    mP   = mP0 * (aux - 0.01*(mA0/mA));

    // Compute impedance
    // Avoid (too) high impedance for low pressure
    mZ = sqrt( rhob * (mK - 3.0) * ( max( 1.0 , mP) / 3.0 ) ) / ( mA + AWalld3 );

    // Initialize pressure at entrance
    mPin = mP;

    // Initialize d/dt Volume
    mVDot = 0.0;

    // Initialize through flow used for diameter adaptation
    mQRemod = 0.0;
}


// Implement method to initialize adaptation variables
void STube::initAdaptation()
{
    // Initialize variables for adaptation
    mQRemodSum          = 0.0;  // Sum of qRemod
    mQPosRemodSum       = 0.0;  // Sum of max(0.0,qRemod)
    mQNegRemodSum       = 0.0;  // Sum of max(0.0,-qRemod)
    mAreaSum            = 0.0;  // Sum of cross-sectional area
    mPressureSquareSum  = 0.0;  // Sum of square of pressure
    mPMaxMax            = 0.0;  // Maximum pMax
    mWallStressMax      = 0.0;  // Maximum wall stress
}


// Implement method to update adaptation variables
void STube::updateAdaptation()
{
    // Declarations
    double pMaxVal; // Current value of maximum pressure

    // Update sum of qRemod
    mQRemodSum += mQRemod;

    // Update sum of max(0.0,qRemod)
    mQPosRemodSum += max(0.0,mQRemod);

    // Update sum of max(0.0,-qRemod)
    mQNegRemodSum += max(0.0,-mQRemod);

    // Update sum of cross-sectional area
    mAreaSum += mA;

    // Update sum of square of pressure
    mPressureSquareSum += mP*mP;

    // Update maximum pMax
    // vImpactAdapt is shock wave velocity due to body movement
    pMaxVal = mPin + mZ * mA * mVImpactAdapt;
    mPMaxMax = max(mPMaxMax,pMaxVal);

    // Update maximum wall stress
    mWallStressMax = max(mWallStressMax,pMaxVal*(1.0 + 3.0*(mA/mAWall)));
}


// Method to adapt diameter
void STube::adaptDiameter(int nrAdaptUpdates)
{
    // Declarations
    double qPos;               // Mean of max(0.0,qRemod)
    double qNeg;               // Mean of max(0.0,-qRemod)
    double qMean;              // Mean of qRemod
    double qMedian;            // Median of qPos, qNeg and qMean
    double areaMean;           // Mean of cross-sectional area
    double vFlowMean;          // Mean flow velocity
    double Fac_vFlow;          // Factor to compute mean cross-section at rest
    double pressureSquareMean; // Mean of square of pressure
    double Fac_p0;             // Factor to compute mean pressure at rest

    // Consistency check: check validity of adaptation variables
    if (nrAdaptUpdates == 0)
    {
        // Error: adaptation variables not valid
        qDebug().nospace() << "ERROR: Adaptation variables not valid in method call MTube::adaptDiameter()";
    }
    else
    {
        // Compute qPos, qNeg and qMean
        qPos    = mQPosRemodSum   / (double)nrAdaptUpdates;
        qNeg    = mQNegRemodSum   / (double)nrAdaptUpdates;
        qMean   = fabs(mQRemodSum / (double)nrAdaptUpdates);

        // Compute median of qPos, qNeg and qMean
        // to determine size of entering blood vessel
        qMedian     = median(qPos,qNeg,qMean);

        // Compute mean of cross-sectional area
        areaMean    = mAreaSum / (double)nrAdaptUpdates;

        // Compute mean flow velocity
        vFlowMean   = qMedian / areaMean;

        // Adapt mean cross-section at rest
        Fac_vFlow   = mVFlowMeanAdapt / vFlowMean;
        mA0         = mA0 / pow( Fac_vFlow, 0.1 );

        // Compute mean of square of pressure
        pressureSquareMean = mPressureSquareSum / (double)nrAdaptUpdates;

        // Adapt mean pressure at rest
        Fac_p0      = mP0 / sqrt(pressureSquareMean);
        mP0         = pow(Fac_p0, 0.3) * sqrt(pressureSquareMean);
    }
}


// Method to adapt wall volume
void STube::adaptWallVolume(int nrAdaptUpdates)
{
    // Declarations
    double FacWallStress; // Wall stress factor

    // Consistency check: check validity of adaptation variables
    if (nrAdaptUpdates == 0)
    {
        // Error: adaptation variables not valid
        qDebug().nospace() << "ERROR: Adaptation variables not valid  in method call MTube::adaptWallVolume()";
    }
    else
    {
        // Compute wall stress factor
        FacWallStress = mWallStressAdapt / mWallStressMax;

        // Adapt cross-sectional area of wall
        mAWall = mAWall / sqrt(FacWallStress);

        // Adapt maximum pressure
        mPmax = mPmax * pow((mPMaxMax / mPmax),0.25);
    }
}



bool STube::writeXml(QDomNode &parentNode) const
{
    bool ok = false;
    QDomNode myNode = createXmlEntry(parentNode,getClassName(),getDefaultVersion(),getName());
    if ( !myNode.isNull())
    {
        ok  = addValue(myNode,TUBE_TYPE ,mTubeType);
        ok &= writeSpecification(myNode);
        ok &= writeState(myNode);
    }
    return ok;
}


bool STube::readXml(const QDomNode &node)
{
    mTubeType     = (tube_type)readInt(node, TUBE_TYPE);
    bool ok = SCompliance::readXml(node);
    return ok;
}


// Implement method to write specificatioin to file
bool STube::writeSpecification(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_TUBE_KEY));
    bool ok = SCompliance::writeSpecification(specNode);


    ok &= addValue(specNode,SPECIFICATION_STIFFNESS_EXPONENT,   mK);
    ok &= addValue(specNode,SPECIFICATION_MEAN_REST_FLOW,       mQ0);
    ok &= addValue(specNode,SPECIFICATION_MEAN_REST_PRESSURE,   mP0);
    ok &= addValue(specNode,SPECIFICATION_TUBE_VESSEL_LENGTH,   mVesselLength);
    ok &= addValue(specNode,SPECIFICATION_TUBE_PMAX,            mPmax);
    ok &= addValue(specNode,SPECIFICATION_TUBE_AREA_WALL,       mAWall);
    ok &= addValue(specNode,SPECIFICATION_TUBE_MEAN_AREA,       mA0);
    return ok;
}


// Implement method to read specification from file
bool STube::readSpecification (const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(SPECIFICATION_TUBE_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement specNode = elements.at(0).toElement();
        SCompliance::readSpecification(specNode);
        mK              = readValue(specNode,SPECIFICATION_STIFFNESS_EXPONENT,  DEFAULT_STIFFNESS_EXPONENT);
        mQ0             = readValue(specNode,SPECIFICATION_MEAN_REST_FLOW,      DEFAULT_MEAN_REST_FLOW);
        mP0             = readValue(specNode,SPECIFICATION_MEAN_REST_PRESSURE,  DEFAULT_MEAN_REST_PRESSURE);
        mVesselLength   = readValue(specNode,SPECIFICATION_TUBE_VESSEL_LENGTH,  DEFAULT_TUBE_VESSEL_LENGTH);
        mPmax           = readValue(specNode,SPECIFICATION_TUBE_PMAX,           DEFAULT_TUBE_PMAX);
        mAWall          = readValue(specNode,SPECIFICATION_TUBE_AREA_WALL,      DEFAULT_TUBE_AREA_WALL);
        mA0             = readValue(specNode,SPECIFICATION_TUBE_MEAN_AREA,      DEFAULT_TUBE_MEAN_AREA);
        return true;
    }
    return false;
}






// Method to set default parameters for aorta (LArt)
void STube::setParLArt()
{
    // Set specification parameters
    mK     =  8.0;          // Stiffness exponent of wall material
    mQ0    =  8.5000e-005;  // Mean flow at rest
    mP0    =  1.2298e+004;  // Mean pressure at rest
    mVesselLength   =  0.4001;       // Effective length of vessel (compliance)
    mPmax  =  3.1856e+004;  // Maximum pressure
    mAWall =  1.1636e-004;  // Cross-sectional area of wall
    mA0    =  5.0007e-004;  // Mean cross-section at rest

    // Set specification parameters for adaptation
    mWallStressAdapt = 5.0e+005;  // Maximum wall stress
    mVFlowMeanAdapt  = 0.1700;    // Mean flow velocity
    mVImpactAdapt    = 3.0;       // Shock wave velocity due to body movement

    // Set simulation state variables for compliance
    mVolume      =  1.9538e-004;   // Volume of compliance
    mVDot   = -8.1000e-005;   // d/dt Volume
    mA      =  4.8830e-004;   // Cross-section of compliance
    mP      =  1.1728e+004;   // Pressure of compliance
    mZ      =  8.5949e+006;   // Impedance
    mPin    =  1.1032e+004;   // Pressure at entrance
    mQRemod =  6.0834e-007;   // Through flow used for diameter adaptation
}


// Method to set default parameters for pulmonary venes (LVen)
void STube::setParLVen()
{
    // Set specification parameters
    mK     =  10.0;          // Stiffness exponent of wall material
    mQ0    =   8.5000e-005;  // Mean flow at rest
    mP0    = 805.3027;       // Mean pressure at rest
    mVesselLength   =   0.2001;       // Effective length of vessel (compliance)
    mPmax  =   9.1520e+003;  // Maximum pressure
    mAWall =   4.2827e-005;  // Cross-sectional area of wall
    mA0    =   5.0362e-004;  // Mean cross-section at rest

    // Set specification parameters for adaptation
    mWallStressAdapt = 5.0e+005;  // Maximum wall stress
    mVFlowMeanAdapt  = 0.1700;    // Mean flow velocity
    mVImpactAdapt    = 3.0;       // Shock wave velocity due to body movement

    // Set simulation state variables for compliance
    mVolume      =   9.3671e-005;  // Volume of compliance
    mVDot   =  -5.9875e-005;  // d/dt Volume
    mA      =   4.6823e-004;  // Cross-section of compliance
    mP      = 674.0319;       // Pressure of compliance
    mZ      =   2.6633e+006;  // Impedance
    mPin    = 514.5642;       // Pressure at entrance
    mQRemod =  -1.2156e-004;  // Through flow used for diameter adaptation
}


// Method to set default parameters for pulmonary artery (RArt)
void STube::setParRArt()
{
    // Set specification parameters
    mK     =   8.0;          // Stiffness exponent of wall material
    mQ0    =   8.5000e-005;  // Mean flow at rest
    mP0    =   3.2594e+003;  // Mean pressure at rest
    mVesselLength   =   0.2001;       // Effective length of vessel (compliance)
    mPmax  =   1.4347e+004;  // Maximum pressure
    mAWall =   5.9533e-005;  // Cross-sectional area of wall
    mA0    =   5.0612e-004;  // Mean cross-section at rest

    // Set specification parameters for adaptation
    mWallStressAdapt = 5.0e+005;  // Maximum wall stress
    mVFlowMeanAdapt  = 0.1700;    // Mean flow velocity
    mVImpactAdapt    = 3.0;       // Shock wave velocity due to body movement

    // Set simulation state variables for compliance
    mVolume      =  9.2891e-005;  // Volume of compliance
    mVDot   = -6.1016e-005;  // d/dt Volume
    mA      =  4.6433e-004;  // Cross-section of compliance
    mP      =  2.8038e+003;  // Pressure of compliance
    mZ      =  4.5750e+006;  // Impedance
    mPin    =  2.5246e+003;  // Pressure at entrance
    mQRemod =  6.7001e-007;  // Through flow used for diameter adaptation
}


// Method to set default parameters for systemic venes (RVen)
void STube::setParRVen()
{
    // Set specification parameters
    mK     =  10.0;          // Stiffness exponent of wall material
    mQ0    =   8.5000e-005;  // Mean flow at rest
    mP0    = 381.4712;       // Mean pressure at rest
    mVesselLength   =   0.4001;       // Effective length of vessel (compliance)
    mPmax  =   6.2551e+003;  // Maximum pressure
    mAWall =   2.9968e-005;  // Cross-sectional area of wall
    mA0    =   5.0027e-004;  // Mean cross-section at rest

    // Set specification parameters for adaptation
    mWallStressAdapt = 5.0e+005;  // Maximum wall stress
    mVFlowMeanAdapt  = 0.1700;    // Mean flow velocity
    mVImpactAdapt    = 3.0;       // Shock wave velocity due to body movement

    // Set simulation state variables for compliance
    mVolume      =   1.9705e-004;  // Volume of compliance
    mVDot   =  -1.1531e-004;  // d/dt Volume
    mA      =   4.9250e-004;  // Cross-section of compliance
    mP      = 364.1710;       // Pressure of compliance
    mZ      =   1.8798e+006;  // Impedance
    mPin    = 147.4118;       // Pressure at entrance
    mQRemod =  -2.0031e-004;  // Through flow used for diameter adaptation
}




