/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: valve.cpp 776 2013-07-11 20:33:36Z eelco $
 *
 */
#include "defs.h"
#include "circsettings.h"
#include "tokens.h"
#include "valve.h"
#include "tvalve.h"
#include "tube.h"


using namespace  Peacs::CircAdapt;

// Definition of model constants
const double SValve::rhob    = BLOODDENSITY; // Density of blood
const double SValve::pMargin = VALVEPMARGIN; // Threshold for valve closure

// Definition of velocity threshold for valve opening
const double SValve::vMargin = VALVEVMARGINFACTOR * sqrt(abs(2*pMargin)/rhob);



const char * VALVE_TYPE                         = "valve_type";

const char * STATE_VALVE_KEY                    = "valve_state";
const char * STATE_AREA_OPEN_VALVE              = "area_open_valve";
const char * STATE_AREA_CLOSED_VALVE            = "area_closed_valve";
const char * STATE_QDOT                         = "flow_dot";
const char * STATE_CHANNELLENGTH_WITH_INERTIA   = "Channel_length_with_inertia";

const char * SPECIFICATION_VALVE_KEY            = "valve_specification";
const char * SPECIFICATION_VALVE_LENGTH         = "valve_length";
const char * SPECIFICATION_VALVE_MAX_AOPEN      = "valve_maxAopen";
const char * SPECIFICATION_VALVE_MIN_ALEAK      = "valve_minAleak";



const int FILE_VALVE_VERSION_1  = 1;
const int FILE_VALVE_VERSION_DEFAULT  = 1;

const char* FILE_VALVE_NAME = "VALVE";


SValve::SValve(valve_type type, SCompliance &prox, SCompliance &dist):
    SFlowDuct(prox,dist),
    mLen(0.0),    
    mAOpenMax(DEFAULT_VALVE_MAX_AOPEN),
    mALeakMin(DEFAULT_VALVE_MIN_ALEAK),
    mAOpen(0.0),
    mALeak(0.0),
    mQDot(0.0),
    mL(0.0),
    mValveType(type)
{
    init();
}


SValve::SValve(valve_type type, SCompliance &prox, SCompliance &dist, const QString &name):
    SFlowDuct(prox,dist,name),
    mLen(0.0),
    mAOpenMax(DEFAULT_VALVE_MAX_AOPEN),
    mALeakMin(DEFAULT_VALVE_MIN_ALEAK),
    mAOpen(DEFAULT_VALVE_MAX_AOPEN),
    mALeak(DEFAULT_VALVE_MIN_ALEAK),
    mQDot(0.0),
    mL(0.0),
    mValveType(type)
{
    init();
}



SValve::SValve(const SValve &valve):
    SFlowDuct(valve),
    mLen( valve.mLen ),
    mAOpenMax(valve.mAOpenMax),
    mALeakMin(valve.mALeakMin),
    mAOpen( valve.mAOpen ),
    mALeak( valve.mALeak ),
    mQDot(  valve.mQDot ),
    mL( valve.mL ),
    mValveType( valve.mValveType )
{

}

const SValve& SValve::operator = (const SValve &valve)
{
    if ( this != &valve)
    {
        SFlowDuct::operator =(valve);
        mLen        = valve.mLen;
        mAOpenMax   = valve.mAOpenMax;
        mALeakMin   = valve.mALeakMin;
        mAOpen      = valve.mAOpen;
        mALeak      = valve.mALeak;
        mQDot       = valve.mQDot;
        mL          = valve.mL;
        mValveType  = valve.mValveType;
    }
    return *this;
}

void SValve::init()
{
    SFlowDuct::init();
    // Do not call parent init!!!
    if ( mValveType == LEFT_ATRIUM_IN_LVen )
    {
        setParValveLVen ();
    }
    else if ( mValveType == MITRALVALVE_LAv )
    {
        setParValveLAv ();
    }
    else if ( mValveType == AORTA_LArt )
    {
        setParValveLArt ();
    }
    else if ( mValveType == VENA_CAVA_RVen )
    {
        setParValveRVen ();
    }
    else if ( mValveType ==  TRICUSPID_RAv )
    {
        setParValveRAv ();
    }
    else if ( mValveType == RVOT_RArt )
    {
        setParValveRArt ();
    }
    else if ( mValveType == ASD )
    {
        setParValveASD ();
    }
    else if ( mValveType == VSD )
    {
        setParValveVSD ();
    }
    else // this is automatically the default
    {
        setParValveDUCT ();
    }
    // TODO give default values
    mAOpenMax = mAOpen;
    mALeakMin = mALeak;
}


void SValve::copyUserSpecifications(const TValve &valve)
{
    SFlowDuct::copyUserSpecifications( valve );
    mLen        = valve.getLength();
    if ( valve.getAreaOpenMax() < DEFAULT_VALVE_MIN_ALEAK )
    {
        if ( mValveType == ASD )
        {
            setParValveASD ();
        }
        else if ( mValveType == VSD )
        {
            setParValveVSD ();
        }
        else // this is automatically the default
        {
            setParValveDUCT ();
        }
    }
    else if ( (mAOpenMax - valve.getAreaOpenMax())/mAOpenMax > 0.25)// <  / 4.0  )//< DEFAULT_VALVE_MIN_ALEAK )
    {
        mQ /= 1e5;
        mQDot /= 1e5;
    }

    mAOpenMax   = qMax(valve.getAreaOpenMax(),DEFAULT_VALVE_MIN_ALEAK);
    mALeakMin   = qMax(valve.getAreaLeakMin(), DEFAULT_VALVE_MIN_ALEAK);

    mAOpen      = mAOpenMax;
    mALeak      = mALeakMin;
}


int SValve::getDefaultVersion(void) const
{
    return FILE_VALVE_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SValve::getClassName(void) const
{
    return FILE_VALVE_NAME;
}



// Method to set leak of valve
void SValve::setLeak (double leak)
{
    if ( mValveType == MITRALVALVE_LAv ||
         mValveType == TRICUSPID_RAv )
    {
        // Adjust cross-section of closed valve
        mALeak = qMax(mALeakMin, leak * mAOpen);
    }
}


// Method to set cross-section of open valve
void SValve::setOpenArea (double area)
{
    // Set cross-section of open valve
    if ( mValveType == LEFT_ATRIUM_IN_LVen )
    {
        mAOpen = area;
    }
}

// Method to compute volume derivatives in neighboring compliances
void SValve::computeVolumeDerivatives (double deltaT)
{
    // Update flow
    mQ += deltaT * mQDot;

    // Update volume derivatives in proximal compliance
    mProximal.subtractOutgoingFlowToValve(mQ);

    // Update volume derivatives in distal compliance
    mDistal.addIncomingFlowFromValve(mQ);
}


// Method to compute flow derivative from pressure drop
void SValve::computeFlowDerivative()
{
    // Declarations
    double Dp;       // Pressure drop over valve
    double DpB;      // Bernouilli pressure drop
    double AFw;      // Forward flow channel area
    double ABw;      // Backward flow channel area
    double AValve;   // Opening area of valve
    double AProx;    // Proximal orifice area
    double ADist;    // Distal orifice area
    double vProx;    // Proximal blood flow velocity
    double vValve;   // Blood flow velocity within valve
    double vDist;    // Distal blood flow velocity

    // Compute pressure drop over valve
    Dp = mProximal.getPressureAtEntrance() - mDistal.getPressureAtEntrance();

    // Numerical safety for valve opening around flow = 0
    // pMargin is pressure threshold for valve closure (10 Nm-2)
    // vMargin is velocity threshold for valve opening

    // Define forward flow channel area
    AFw = 0.0;
    if (mQ > mAOpen*vMargin || Dp > -pMargin)
    {
        AFw = mAOpen;
    }

    // Define backward flow channel area
    ABw = 0.0;
    if (mQ < -mALeak*vMargin || Dp < pMargin)
    {
        ABw = mALeak;
    }

    // Define opening area
    AValve = max(AFw,ABw);

    // Define proximal and distal orifice areas
    AProx = mProximal.getArea();
    ADist = mDistal.getArea();

    // Compute blood flow velocities
    vProx  = mQ / AProx;
    vValve = mQ / AValve;
    vDist  = mQ / ADist;

    // Narrow outflow area widens with forward flow
    // Compute Bernouilli pressure drop
    DpB = 0.0;
    if (mQ > mAOpen*vMargin)
    {
        DpB = (0.5 * rhob) * max(vValve*vValve - vProx*vProx,0.0);
    }
    if (mQ < -mALeak*vMargin)
    {
        // DEBUG - BUG FIX times (0.5 * rhob)
        DpB -= (0.5 * rhob) * max(vValve*vValve - vDist*vDist,0.0);
    }

    // Update the length of the channel with inertia
    // Proximal and distal series inertial effects are added to valve inertia
    mL = rhob * (mLen/AValve + 0.5*(1/sqrt(AProx) + 1/sqrt(ADist)));

    // Update flow derivative
    mQDot = (Dp - DpB)/mL;
}


void SValve::adaptWallVolumeOfConnctedTube(int nrAdaptUpdates)
{
    STube t(STube::UNDEFINED_TUBE);
    if ( mProximal.getClassName() == t.getClassName() )
    {
        dynamic_cast <STube*> (&mProximal)->adaptWallVolume(nrAdaptUpdates);
    }

    if ( mDistal.getClassName() == t.getClassName() )
    {
        dynamic_cast <STube*> (&mDistal)->adaptWallVolume(nrAdaptUpdates);
    }
}














bool SValve::writeXml(QDomNode &parentNode) const
{
    bool ok = false;
    QDomNode myNode = createXmlEntry(parentNode,getClassName(),getDefaultVersion(),getName());
    if ( !myNode.isNull())
    {
        ok  = addValue(myNode,VALVE_TYPE          ,mValveType);
        ok &= writeSpecification(myNode);
        ok &= writeState(myNode);
    }
    return ok;
}


bool SValve::readXml(const QDomNode &node)
{
    mValveType     = (valve_type)readInt(node, VALVE_TYPE);
    bool ok = SFlowDuct::readXml(node);
    return ok;
}




// Implement method to write specificatioin to file
bool SValve::writeSpecification(QDomNode &parentNode) const
{
    QDomNode specNode(createXmlEntry(parentNode,SPECIFICATION_VALVE_KEY));

    bool ok = SFlowDuct::writeSpecification(specNode);
    ok &= addValue(specNode,SPECIFICATION_VALVE_LENGTH,      mLen);
    ok &= addValue(specNode,SPECIFICATION_VALVE_MAX_AOPEN,   mAOpenMax);
    ok &= addValue(specNode,SPECIFICATION_VALVE_MIN_ALEAK,   mALeakMin);
    return ok;
}


// Implement method to read specification from file
bool SValve::readSpecification (const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(SPECIFICATION_VALVE_KEY);
    if ( elements.count() == 1 )
    {
        QDomElement specNode = elements.at(0).toElement();
        SFlowDuct::readSpecification(specNode);
        mLen        = readValue(specNode,SPECIFICATION_VALVE_LENGTH,    DEFAULT_VALVE_LENGTH);
        mAOpenMax   = readValue(specNode,SPECIFICATION_VALVE_MAX_AOPEN, DEFAULT_VALVE_MAX_AOPEN);
        mALeakMin   = readValue(specNode,SPECIFICATION_VALVE_MIN_ALEAK, DEFAULT_VALVE_MIN_ALEAK);
        return true;
    }
    return false;
}



// Implement method to write state to file
bool SValve::writeState(QDomNode &parentNode) const
{
    QDomNode stateNode(createXmlEntry(parentNode,STATE_VALVE_KEY));

    bool ok = SFlowDuct::writeState(stateNode);
    ok &= addValue(stateNode,STATE_AREA_OPEN_VALVE,             mAOpen);
    ok &= addValue(stateNode,STATE_AREA_CLOSED_VALVE,           mALeak);
    ok &= addValue(stateNode,STATE_QDOT,                        mQDot);
    ok &= addValue(stateNode,STATE_CHANNELLENGTH_WITH_INERTIA,  mL);
    return ok;
}


// Implement method to read state from file
bool SValve::readState (const QDomNode &node)
{
    QDomNodeList elements = node.toElement().elementsByTagName(STATE_VALVE_KEY);

    if ( elements.count() == 1 )
    {
        QDomElement stateNode = elements.at(0).toElement();
        SFlowDuct::readState(stateNode);
        mAOpen  = readValue(stateNode, STATE_AREA_OPEN_VALVE);
        mALeak  = readValue(stateNode, STATE_AREA_CLOSED_VALVE);
        mQDot   = readValue(stateNode, STATE_QDOT);
        mL      = readValue(stateNode, STATE_CHANNELLENGTH_WITH_INERTIA);

        return true;
    }
    return false;
}



// Method to set default parameters for inlet left atrium (ValveLVen)
void SValve::setParValveLVen()
{
    // Set specification parameters
    mLen   = 0.0201;       // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = 1.2156e-004;  // Flow

    // Set simulation state variables of valve
    mAOpen = 4.6823e-004;  // Cross-section of open valve
    mALeak = 5.0000e-004;  // Cross-section of closed valve
    mQDot  = -0.0027;      // d/dt Flow
    mL     = 8.2266e+004;  // Channel length with inertia
}


// Method to set default parameters for mitral valve (ValveLAv)
void SValve::setParValveLAv ()
{
    // Set specification parameters
    mLen   = 0.0285;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = -3.8474e-005;  // Flow

    // Set simulation state variables of valve
    mAOpen =  6.0000e-004;  // Cross-section of open valve
    mALeak =  DEFAULT_VALVE_MIN_ALEAK;//1.2394e-004;  // Cross-section of closed valve
    mQDot  = -2.1774e-004;  // d/dt Flow
    mL     =  2.7338e+005;  // Channel length with inertia
}


// Method to set default parameters for aortic valve (ValveLArt)
void SValve::setParValveLArt ()
{
    // Set specification parameters
    mLen   = 0.0201;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = 6.1035e-007;   // Flow

    // Set simulation state variables of valve
    mAOpen =  4.0000e-004;  // Cross-section of open valve
    mALeak =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of closed valve
    mQDot  = -2.4105e-007;  // d/dt Flow
    mL     =  4.2262e+010;  // Channel length with inertia
}


// Method to set default parameters for inlet right atrium (ValveRVen)
void SValve::setParValveRVen ()
{
    // Set specification parameters
    mLen   = 0.0201;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = 2.0031e-004;   // Flow

    // Set simulation state variables of valve
    mAOpen =  4.9250e-004;  // Cross-section of open valve
    mALeak =  5.0000e-004;  // Cross-section of closed valve
    mQDot  = -0.0010;       // d/dt Flow
    mL     =  8.2637e+004;  // Channel length with inertia
}


// Method to set default parameters for tricuspid valve (ValveRAv)
void SValve::setParValveRAv ()
{
    // Set specification parameters
    mLen   = 0.0285;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = -2.5765e-006;  // Flow

    // Set simulation state variables of valve
    mAOpen =  6.0000e-004;  // Cross-section of open valve
    mALeak =  DEFAULT_VALVE_MIN_ALEAK;//6.2523e-005;  // Cross-section of closed valve
    mQDot  = -3.5897e-004;  // d/dt Flow
    mL     =  5.1047e+005;  // Channel length with inertia
}


// Method to set default parameters for pulmonary valve (ValveRArt)
void SValve::setParValveRArt ()
{
    // Set specification parameters
    mLen   = 0.0201;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = 6.6800e-007;   // Flow

    // Set simulation state variables of valve
    mAOpen =  4.0000e-004;  // Cross-section of open valve
    mALeak =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of closed valve
    mQDot  = -4.9933e-008;  // d/dt Flow
    mL     =  4.2262e+010;  // Channel length with inertia

}


// Method to set default parameters for valve for ASD (ValveASD)
void SValve::setParValveASD ()
{
    // Set specification parameters
    mLen   = 0.0201;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = 4.5396e-010;   // Flow

    // Set simulation state variables of valve
    mAOpen =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of open valve
    mALeak =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of closed valve
    mQDot  =  1.7011e-009;  // d/dt Flow
    mL     =  4.2262e+010;  // Channel length with inertia
}


// Method to set default parameters for valve for VSD (ValveVSD)
void SValve::setParValveVSD ()
{
    // Set specification parameters
    mLen   = 0.0201;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = 4.5974e-010;   // Flow

    // Set simulation state variables of valve
    mAOpen =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of open valve
    mALeak =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of closed valve
    mQDot  = -3.3072e-010;  // d/dt Flow
    mL     =  4.2262e+010;  // Channel length with inertia

}


// Method to set default parameters for valve for DUCT (ValveDUCT)
void SValve::setParValveDUCT ()
{
    // Set specification parameters
    mLen   = 0.0201;        // Length DEBUG: What is Len??

    // Set simulation state variables of flow duct
    mQ     = 2.0154e-009;   // Flow

    // Set simulation state variables of valve
    mAOpen =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of open valve
    mALeak =  DEFAULT_VALVE_MIN_ALEAK;  // Cross-section of closed valve
    mQDot  = -5.5056e-010;  // d/dt Flow
    mL     =  4.2262e+010;  // Channel length with inertia

}
