/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: valve.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef _VALVE_H_
#define _VALVE_H_

#include "flowduct.h"

namespace Peacs
{
    namespace CircAdapt
    {
    class TValve;
        /**
        Class for valve. Inherits from FlowDuct.
        */
        class SValve: public SFlowDuct {

        public:
            enum valve_type { UNDEFINED_VALVE       = 0,
                              LEFT_ATRIUM_IN_LVen   = 1,
                              MITRALVALVE_LAv       = 2,
                              AORTA_LArt            = 3,
                              VENA_CAVA_RVen        = 4,
                              TRICUSPID_RAv         = 5,
                              RVOT_RArt             = 6,
                              ASD                   = 7,
                              VSD                   = 8,
                              DUCT                  = 9 };

            /* @brief Default Constructor */
            SValve(valve_type type, SCompliance &prox, SCompliance &dist);

            SValve (valve_type type, SCompliance &prox, SCompliance &dist, const QString &name);

            SValve (const SValve & valve);

            virtual ~SValve() {;}

            const SValve& operator = (const SValve &valve);

            void init();

            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TValve &chamber);


            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;

            /** @returns the valve type*/
            valve_type getValveType() const { return mValveType; }


            double getLength() const { return mLen; }

            double getAreaOpenMax() const {return mAOpenMax; }

            double getAreaLeakMin() const { return mALeakMin; }


            // Simulation state variables
            double getAreaOpen() const  { return mAOpen; }     // Cross-section of open valve

            double getALeak() const     { return mALeak; }      // Cross-section of closed valve

            double getQdot() const      { return mQDot;  }       // d/dt Flow

            double getLwithInertia() const         { return mL;     }          // Channel length with inertia




            /// Implement method to initialize adaptation variables
            void initAdaptation() { ;}

            /// Method to set leak of valve
            void setLeak (double leak);

            /// Method to set cross-section of open valve
            void setOpenArea (double area);

            /// Method to compute volume derivatives in neighboring chambers/tubes
            void computeVolumeDerivatives (double deltaT);

            /// Method to compute flow derivative from pressure drop
            void computeFlowDerivative();

            void adaptWallVolumeOfConnctedTube(int nrAdaptUpdatesnrAdaptUpdates);





            //==================== file functions ===================

            bool writeXml(QDomNode &parentNode) const;

            bool readXml(const QDomNode &node);

            /// Abstract method to write specification to file
            virtual bool writeSpecification (QDomNode &parentNode) const;

            /// Abstract method to read specification from file
            virtual bool readSpecification (const QDomNode &parentNode);

            /// Abstract method to write state to file
            bool writeState (QDomNode &parentNode) const;

            /// Abstract method to read state from file
            bool readState (const QDomNode &node);

        private:
            /// Method to set default parameters for inlet left atrium (ValveLVen)
            void setParValveLVen ();

            /// Method to set default parameters for mitral valve (ValveLAv)
            void setParValveLAv ();

            /// Method to set default parameters for aortic valve (ValveLArt)
            void setParValveLArt ();

            /// Method to set default parameters for inlet right atrium (ValveRVen)
            void setParValveRVen ();

            /// Method to set default parameters for tricuspid valve (ValveRAv)
            void setParValveRAv ();

            /// Method to set default parameters for pulmonary valve (ValveRArt)
            void setParValveRArt ();

            /// Method to set default parameters for valve for ASD (ValveASD)
            void setParValveASD ();

            /// Method to set default parameters for valve for VSD (ValveVSD)
            void setParValveVSD ();

            /// Method to set default parameters for valve for DUCT (ValveDUCT)
            void setParValveDUCT ();

        private:
            // Specification parameters
            double mLen;        // Length DEBUG: What is Len??
            double mAOpenMax;
            double mALeakMin;

            // Simulation state variables
            double mAOpen;      // Cross-section of open valve
            double mALeak;      // Cross-section of closed valve
            double mQDot;       // d/dt Flow
            double mL;          // Channel length with inertia

            valve_type mValveType;

            // Model constants
            static const double rhob;     // Density of blood
            static const double pMargin;  // Threshold for valve closure
            static const double vMargin;  // Velocity threshold for valve opening
        };
    }
}

#endif
