/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: ventricle.cpp 789 2013-07-18 13:22:28Z peter $
 *
 */

#include <QDebug>

#include "math.h"

#include "defs.h"
#include "ventricle.h"
#include "tventricle.h"


// keys for the xml file
const char * MYVENTRICLES_KEY             = "ventricular_chambers";

const char * VENTRICLE_TYPE               = "ventricule_type";

const char * SPECIFICATION_VENTRICLE_KEY  = "ventricle_specification";
const char * STATE_VENTRICLE_KEY          = "ventricle_state";

const int FILE_VENTRICLE_VERSION_1  = 1;
const int FILE_VENTRICLE_VERSION_DEFAULT  = 1;
const char* FILE_VENTRICLE_NAME = "VENTRICLE";


const char* LEFT_VENTRICLE_NAME     = "left_ventricle";
const char* SEPTUM_VENTRICLE_NAME   = "septum_ventricle";
const char* RIGHT_VENTRICLE_NAME    = "right_ventricle";

using namespace Peacs::CircAdapt;

// Constructor
SVentricle::SVentricle(ventricle_type type, bool adaptToExercise) :
    SCircElement("ventricle"),
    mLeftV(LEFT_VENTRICLE,  adaptToExercise, LEFT_VENTRICLE_NAME,   type == TRISEG),
    mSeptV(SEPTUM,          adaptToExercise, SEPTUM_VENTRICLE_NAME, type == TRISEG),
    mRigtV(RIGHT_VENTRICLE, adaptToExercise, RIGHT_VENTRICLE_NAME,  type == TRISEG),
    mkL1(0.0),
    mkL2(0.0),
    mkS1(0.0),
    mkS2(0.0),
    mkR1(0.0),
    mkR2(0.0),
    mTensionSum(0.0),
    mVentrType(type)
{

}


SVentricle::SVentricle(ventricle_type type, bool adaptToExercise, const QString &name):
    SCircElement(name),
    mLeftV(LEFT_VENTRICLE,  adaptToExercise, LEFT_VENTRICLE_NAME,   type == TRISEG),
    mSeptV(SEPTUM,          adaptToExercise, SEPTUM_VENTRICLE_NAME, type == TRISEG),
    mRigtV(RIGHT_VENTRICLE, adaptToExercise, RIGHT_VENTRICLE_NAME,  type == TRISEG),
    mkL1(0.0),
    mkL2(0.0),
    mkS1(0.0),
    mkS2(0.0),
    mkR1(0.0),
    mkR2(0.0),
    mTensionSum(0.0),
    mVentrType(type)
{

}

SVentricle::SVentricle(const SVentricle & ventricle) :
    SCircElement(ventricle),
    mLeftV( ventricle.mLeftV),
    mSeptV( ventricle.mSeptV),
    mRigtV( ventricle.mRigtV),
    mkL1( ventricle.mkL1),
    mkL2( ventricle.mkL2),
    mkS1( ventricle.mkS1),
    mkS2( ventricle.mkS2),
    mkR1( ventricle.mkR1),
    mkR2( ventricle.mkR2),
    mTensionSum( ventricle.mTensionSum),
    mVentrType( ventricle.mVentrType)
{

}


void SVentricle::init()
{

}





void SVentricle::copyUserSpecifications(const TVentricle &ventr)
{
    mLeftV.copyUserSpecifications( ventr.getLeftVentricle() );
    mRigtV.copyUserSpecifications( ventr.getRightVentricle() );
    mSeptV.copyUserSpecifications( ventr.getSeptum() );
}

int SVentricle::getDefaultVersion(void) const
{
    return FILE_VENTRICLE_VERSION_DEFAULT;
}


/** @brief Returns the class name */
const char *SVentricle::getClassName(void) const
{
    return FILE_VENTRICLE_NAME;
}

const SVentricle & SVentricle::operator =(const SVentricle &ventricle)
{
    if ( this != &ventricle)
    {
        SCircElement::operator =(ventricle);
        mLeftV      = ventricle.mLeftV;
        mSeptV      = ventricle.mSeptV;
        mRigtV      = ventricle.mRigtV;
        mkL1        = ventricle.mkL1;
        mkL2        = ventricle.mkL2;
        mkS1        = ventricle.mkS1;
        mkS2        = ventricle.mkS2;
        mkR1        = ventricle.mkR1;
        mkR2        = ventricle.mkR2;
        mTensionSum = ventricle.mTensionSum;
        mVentrType  = ventricle.mVentrType;
    }
    return *this;
}

double SVentricle::getVolume() const
{
    if ( mVentrType ==BISEG )
    {
        return mRigtV.getVolume() + mRigtV.getWallVolume() +
               mLeftV.getVolume() + mLeftV.getWallVolume();
    }
    else
    {
        return mRigtV.getVolume() + mRigtV.getWallVolume() +
               mLeftV.getVolume() + mLeftV.getWallVolume() +
                                    mSeptV.getWallVolume() ;
    }
}

void SVentricle::adjustForSackPressure(double pressure)
{
    mRigtV.addPressure(pressure);
    mLeftV.addPressure(pressure);
}

// Method to initialize cavity pressure and cross-section
void SVentricle::initMechanics (double tCycleAct, double minActivationTime)
{
    mLeftV.initCavityMechanics(tCycleAct,minActivationTime);
    mSeptV.initCavityMechanics(tCycleAct,minActivationTime);
    mRigtV.initCavityMechanics(tCycleAct,minActivationTime);
}


// Method to update cavity pressure and cross-section
void SVentricle::updateCavityMechanics (double deltaT)
{
    mLeftV.updateCavityMechanics(deltaT);
    mSeptV.updateCavityMechanics(deltaT);
    mRigtV.updateCavityMechanics(deltaT);
}


// Implement method to initialize adaptation variables
void SVentricle::initAdaptation()
{
    mLeftV.initAdaptation();
    mRigtV.initAdaptation();
    mSeptV.initAdaptation();
    initAdaptJunction();
}


// Implement method to update adaptation variables
void SVentricle::updateAdaptation()
{
    mLeftV.updateAdaptation();
    mRigtV.updateAdaptation();
    mSeptV.updateAdaptation();
    updateAdaptJunction();
}


// Method to perform adaptation
void SVentricle::adapt (bool adaptWV,
                        bool adaptSL,
                        bool adaptWA,
                        int nrAdaptUpdates)
{
    adaptJunction(nrAdaptUpdates);
    mLeftV.adapt(adaptWV,adaptSL,adaptWA,nrAdaptUpdates);
    mRigtV.adapt(adaptWV,adaptSL,adaptWA,nrAdaptUpdates);
    mSeptV.adapt(adaptWV,adaptSL,adaptWA,nrAdaptUpdates);
}

// Static private method to estimate Tx and Ty using linear approximation
void SVentricle::estimateTxTy (double Vm, double Ym, double AmOld, double AmDead,
                            double TOld, double DTDAm, double &Tx, double &Ty)
{
 /*---------------------------------------------------+
 | INPUT:                                             |
 | Vm      Volume enclosed by midwall                 |
 | Ym      Radius of midwall at the junctional edge   |
 | AmOld   "Old" area of wall laid as sheet           |
 | AmDead  Non-contractile area (Valve openings)      |
 | TOld    "Old" wall tension                         |
 | DTDAm   dTension/dArea compliance (wall stiffness) |
    |                                                    |
 | OUTPUT:                                            |
 | Tx      x-component of wall tension T              |
 | Ty      y-component of wall tension T              |
 +---------------------------------------------------*/

    // Declarations
    double SignVm;      // Sign of Vm
    double V;           // Volume of cavity
    double Xm;          // Central bulge distance of midwall surface
    double Am;          // "New" area of wall laid as sheet
    double Cm;          // Curvature = 1/radius of midwall
    double T;           // "New" wall tension
    double Qaux;        // Aux variable
    double x2;          // Xm^2
    double y2;          // Ym^2
    double x2py2;       // Xm^2 + Ym^2
    double SinA;        // (2*Xm*Ym) / (Xm^2 + Ym^2)
    double CosA;        // (-Xm^2 + Ym^2) / (Xm^2 + Ym^2)

/*-----------------------------------------------+
 |  Solving 3rd order polynomial                  |
 |  Mathematica: Solve[x^3 + 3y^2x  - 2V == 0, x] |
 |  Q = (V + Sqrt(V^2 + y^6))^(1/3)               |
 |  x = Q - y^2/Q                                 |
 +-----------------------------------------------*/

    // Compute Xm from Vm and Ym
    SignVm = sign(Vm);
    V      = (3.0 / M_PI) * fabs(Vm);
    Qaux   = pow(V + sqrt( V * V + pow( Ym, 6.0)), 1.0 /3.0 );
    Xm     = SignVm * (Qaux - Ym*Ym/Qaux);

    // Compute midwall area Am and curvature Cm=1/rm
    x2    = Xm*Xm;
    y2    = Ym*Ym;
    x2py2 = x2 + y2;
    Am    = M_PI * x2py2 - AmDead;
    Cm    = 2.0 * Xm / x2py2;

    // Estimate tension T using linear approximation
    T = TOld + (Am - AmOld) * DTDAm;

    // Compute Tx and Ty from tension T
    SinA  = Ym * Cm;
    CosA  = (y2 - x2)/x2py2;
    Tx   = T * SinA;
    Ty   = T * CosA;

}


// Static method to update cavity pressures and cross-sections for ventricles
void SVentricle::updateVentricleMechanics(double deltaT)
{
    if ( mVentrType == BISEG )
    {
        mRigtV.updateCavityMechanics(deltaT);
        mLeftV.updateCavityMechanics(deltaT);
    }
    else
    {
        // Declarations
        double Tau;       // Tau for TriSeg computation
        double Y;         // Septum Y
//        double V;         // Septum volume
        double LRef;      // Mean value of Y during former heart beat
        double DV;        // Differential for V
        double DY;        // Differential for Y
        double FacV;      // Factor for DV
        double FacY;      // Factor for DY
        double SV_Vm_old; // Store Vm of SV
        double SV_Ym_old; // Store Ym of SV
        int    nr_iter;   // Number of iterations

        // Declaration of x- and y-components of tension T
        double LVTx0,  LVTx1,  LVTx2,  LVTy0,  LVTy1,  LVTy2;
        double SVTx0,  SVTx1,  SVTx2,  SVTy0,  SVTy1,  SVTy2;
        double RVTx0,  RVTx1,  RVTx2,  RVTy0,  RVTy1,  RVTy2;
        double SumTx0, SumTx1, SumTx2, SumTy0, SumTy1, SumTy2;

        // Declaration of Jacobian matrix
        double D11, D12, D21, D22, Det;

        // Define Tau for TriSeg computation
        Tau = SEPTUMVENTRICLETAU;

        // Update cavity volumes
        mLeftV.addVolume( deltaT * mLeftV.getVolumeDot() );
//        mSeptV.addVolume( deltaT * mSeptV.getVolumeDot() );
        mRigtV.addVolume( deltaT * mRigtV.getVolumeDot() );


        // Mean value of Y during former beat
        LRef = mSeptV.getLRef();

        // Update state variable Y for septum
        mSeptV.addY( deltaT * mSeptV.getYdot());

        // Update sum over Y for septum
        mSeptV.addYsum( mSeptV.getY() );

        // Update number of state updates for septum
        mSeptV.addNrUpdates();

        // Define radius of midwall at the junctional edge
        Y      = mSeptV.getY();
        mLeftV.setYm( Y );
        mSeptV.setYm( Y );
        mRigtV.setYm( Y );

        // Compute volume enclosed by midwall
//        V      = mSeptV.getVm();
//        mSeptV.setVm( V );


        mLeftV.setVm( mSeptV.getVm() - mLeftV.getVolume() - 0.5*(mLeftV.getVwall() + mSeptV.getVwall() ) );
        mRigtV.setVm( mSeptV.getVm() + mRigtV.getVolume() + 0.5*(mSeptV.getVwall() + mRigtV.getVwall() ) );

        // Compute Am, Cm and Xm from Vm and Ym
        mLeftV.computeAmCmXm();
        mSeptV.computeAmCmXm();
        mRigtV.computeAmCmXm();

        // Update sheet mechanics (deltaT)
        mLeftV.updateSheetMechanics(deltaT);
        mSeptV.updateSheetMechanics(deltaT);
        mRigtV.updateSheetMechanics(deltaT);

        // Compute differentials for V and Y
        DV = 0.04 * pow(LRef,3.0);
        DY = 0.02 * LRef;

        // Store Vm and Ym of SV
        SV_Vm_old = mSeptV.getVm();
        SV_Ym_old = mSeptV.getYm();

        // Start iterating until sum of Tx and sum of Ty <= 1.0
        // Maximum number of iterations is 10
        // DEBUG: NOTE: NO ITERATIONS IN MATLAB IMPLEMENTATION
        // ITERATIONS MAY BE LEFT OUT
        nr_iter = 0;
        do
        {
            // Estimate x- and y-components Tx and Ty of tension T
            // for (Vm,Ym) (Vm+DV,Ym) and (Vm,Ym+DY)
            // for LV, SV, and RV wall segment using linear approximation
            estimateTxTy(mLeftV.getVm(),      mLeftV.getYm(),      mLeftV.getAm(), mLeftV.getAmDead(), mLeftV.getT(), mLeftV.getDTDAm(), LVTx0, LVTy0);
            estimateTxTy(mSeptV.getVm(),      mSeptV.getYm(),      mSeptV.getAm(), mSeptV.getAmDead(), mSeptV.getT(), mSeptV.getDTDAm(), SVTx0, SVTy0);
            estimateTxTy(mRigtV.getVm(),      mRigtV.getYm(),      mRigtV.getAm(), mRigtV.getAmDead(), mRigtV.getT(), mRigtV.getDTDAm(), RVTx0, RVTy0);
            estimateTxTy(mLeftV.getVm() + DV, mLeftV.getYm(),      mLeftV.getAm(), mLeftV.getAmDead(), mLeftV.getT(), mLeftV.getDTDAm(), LVTx1, LVTy1);
            estimateTxTy(mSeptV.getVm() + DV, mSeptV.getYm(),      mSeptV.getAm(), mSeptV.getAmDead(), mSeptV.getT(), mSeptV.getDTDAm(), SVTx1, SVTy1);
            estimateTxTy(mRigtV.getVm() + DV, mRigtV.getYm(),      mRigtV.getAm(), mRigtV.getAmDead(), mRigtV.getT(), mRigtV.getDTDAm(), RVTx1, RVTy1);
            estimateTxTy(mLeftV.getVm(),      mLeftV.getYm() + DY, mLeftV.getAm(), mLeftV.getAmDead(), mLeftV.getT(), mLeftV.getDTDAm(), LVTx2, LVTy2);
            estimateTxTy(mSeptV.getVm(),      mSeptV.getYm() + DY, mSeptV.getAm(), mSeptV.getAmDead(), mSeptV.getT(), mSeptV.getDTDAm(), SVTx2, SVTy2);
            estimateTxTy(mRigtV.getVm(),      mRigtV.getYm() + DY, mRigtV.getAm(), mRigtV.getAmDead(), mRigtV.getT(), mRigtV.getDTDAm(), RVTx2, RVTy2);

            // Compute sum of Tx and Ty over LV, SV, and RV wall segment
            // for (Vm,Ym) (Vm+DV,Ym) and (Vm,Ym+DY)
            SumTx0 = LVTx0 + SVTx0 + RVTx0;
            SumTy0 = LVTy0 + SVTy0 + RVTy0;
            SumTx1 = LVTx1 + SVTx1 + RVTx1;
            SumTy1 = LVTy1 + SVTy1 + RVTy1;
            SumTx2 = LVTx2 + SVTx2 + RVTx2;
            SumTy2 = LVTy2 + SVTy2 + RVTy2;


            // Construct Jacobian matrix
            D11 = SumTx1 - SumTx0;
            D21 = SumTy1 - SumTy0;
            D12 = SumTx2 - SumTx0;
            D22 = SumTy2 - SumTy0;

            // Compute determinant of Jacobian matrix
            Det = qMax( D11 * D22 - D12 * D21, 0.1*(fabs( D11 * D22) + fabs( D12 * D21)));

            // Estimate FacV and FacY to obtain zero sum of Tx and Ty
            FacV = (-D22 * SumTx0 + D12 * SumTy0) / Det;
            FacY = (+D21 * SumTx0 - D11 * SumTy0) / Det;

            // Update Vm for all wall segments
            mLeftV.addVm( FacV * DV);
            mSeptV.addVm( FacV * DV);
            mRigtV.addVm( FacV * DV);


            // Update Ym for all wall segments
            mLeftV.addYm( FacY * DY);
            mSeptV.addYm( FacY * DY);
            mRigtV.addYm( FacY * DY);

            // Compute Am, Cm and Xm from Vm and Ym
            mLeftV.computeAmCmXm();
            mSeptV.computeAmCmXm();
            mRigtV.computeAmCmXm();

            // Update sheet mechanics with deltaT = 0.0
            mLeftV.updateSheetMechanics(0.0);
            mSeptV.updateSheetMechanics(0.0);
            mRigtV.updateSheetMechanics(0.0);

            // Count number of iterations
            nr_iter++;
        }
        while ( ( fabs( SumTx0 ) >= 1.0 || fabs( SumTy0 ) >= 1.0 ) &&
               nr_iter < 30 );
//        while ( ( fabs( SumTx0 ) >= 1.0 || fabs( SumTy0 ) >= 1.0 ) &&
//        while ( nr_iter < 20 );

        // Obtain cavity pressure from transmural pressure
        mLeftV.setPressure(-1.0 * mLeftV.getPtrans() );
        mRigtV.setPressure(mRigtV.getPtrans());

        // Set VDot and YDot for septum
        mSeptV.setVdot( (mSeptV.getVm() - SV_Vm_old) / Tau );
        mSeptV.setYdot( (mSeptV.getYm() - SV_Ym_old) / Tau );

        // Compute hemodynamics of the cavities
        mLeftV.updateCavityHemodynamics();
        mRigtV.updateCavityHemodynamics();
    }
}


// Static method to initialize adaptation variables for sheets at junctions
void SVentricle::initAdaptJunction()
{
    // Initialize tension direction vector of left ventricle
    mkL1 = 0.0;
    mkL2 = 0.0;

    // Initialize tension direction vector of septum
    mkS1 = 0.0;
    mkS2 = 0.0;

    // Initialize tension direction vector of right ventricle
    mkR1 = 0.0;
    mkR2 = 0.0;

    // Initialize sum of tensions at junction
    mTensionSum = 0.0;
}


// Static method to update adaptation variables for sheets at junctions
void SVentricle::updateAdaptJunction()
{
    if ( mVentrType == TRISEG )
    {
        // Compute T3
        double T3 = pow( mLeftV.getT() * mSeptV.getT() * mRigtV.getT(), 1.0 /3.0 );

        // Compute tension direction vector of left ventricle
        double x = mLeftV.getXm();
        double y = mLeftV.getYm();
        double c = mLeftV.getCm();
        mkL1 += T3 * y * c;
        mkL2 += T3 * -( y * y - x * x ) / ( y * y + x * x );

        // Compute tension direction vector of septum
        x = mSeptV.getXm();
        y = mSeptV.getYm();
        c = mSeptV.getCm();
        mkS1 += T3 * y * c;
        mkS2 += T3 * -( y * y - x * x ) / ( y * y + x * x );

        // Compute tension direction vector of right ventricle
        x = mRigtV.getXm();
        y = mRigtV.getYm();
        c = mRigtV.getCm();
        mkR1 += T3 * y * c;
        mkR2 += T3 * -( y * y - x * x ) / ( y * y + x * x );

        // Update sum of tensions at junction
        mTensionSum += T3;
    }
}


// Static method to adapt sheets at junctions
void SVentricle::adaptJunction(int nrAdaptUpdates)
{

    // Declarations
    double TensionMean;
    double MeankL1, MeankL2;
    double MeankS1, MeankS2;
    double MeankR1, MeankR2;
    double mL1, mL2;
    double mS1, mS2;
    double mR1, mR2;
    double cL, cS, cR;
    double kTot1, kTot2;
    double dL, dS, dR;
    double hL, hS, hR, h0;

    // Consistency check: check validity of adaptation variables
    //TODO remove???
    if ( nrAdaptUpdates == 0 )
    {
        // Error: adaptation variables not valid
        qDebug().nospace() << "ERROR: Adaptation variables not valid in method call MVentricle::adaptJunction()";
    }
    else
    {
        // Compute mean of tensions at junction
        TensionMean = mTensionSum / nrAdaptUpdates;

        // Compute mean tension vecors kL, kS, kR
        MeankL1 = mkL1 / nrAdaptUpdates;
        MeankL2 = mkL2 / nrAdaptUpdates;
        MeankS1 = mkS1 / nrAdaptUpdates;
        MeankS2 = mkS2 / nrAdaptUpdates;
        MeankR1 = mkR1 / nrAdaptUpdates;
        MeankR2 = mkR2 / nrAdaptUpdates;

        // Normalize mean vectors kL, kS, kR
        mL1 = MeankL1 / TensionMean;
        mL2 = MeankL2 / TensionMean;
        mS1 = MeankS1 / TensionMean;
        mS2 = MeankS2 / TensionMean;
        mR1 = MeankR1 / TensionMean;
        mR2 = MeankR2 / TensionMean;

        // Compute outer products of unit vectors (cosine)
        cL    = mS1*mR1 + mS2*mR2;
        cS    = mR1*mL1 + mR2*mL2;
        cR    = mL1*mS1 + mL2*mS2;
        kTot1 = mL1*cL + mS1*cS + mR1*cR;
        kTot2 = mL2*cL + mS2*cS + mR2*cR;

        // Movement of junction along wall surfaces
        dL = kTot1*mL1 + kTot2*mL2;
        dS = kTot1*mS1 + kTot2*mS2;
        dR = kTot1*mR1 + kTot2*mR2;

        // AmRef extension ~ 1/AmRef
        hL = 1.0 / mLeftV.getAmRef();
        hS = 1.0 / mSeptV.getAmRef();
        hR = 1.0 / mRigtV.getAmRef();
        h0 = hL + hS + hR;

        // Enlarge septum to anatomical size
        mLeftV.setEAmRef( dL * (hL / h0) );
        mSeptV.setEAmRef( dS * (hS / h0) );
        mRigtV.setEAmRef( dR * (hR / h0) );
    }
}

bool SVentricle::writeXml(QDomNode &parentNode) const
{
    bool ok = false;
    QDomNode myNode = createXmlEntry(parentNode,getClassName(),getDefaultVersion(),getName());
    if ( !myNode.isNull())
    {
        ok  = addValue(myNode,VENTRICLE_TYPE          ,mVentrType);
        ok &= writeSpecification(myNode);
        ok &= writeState(myNode);
        QDomNode chambersNode = createXmlEntry(myNode,MYVENTRICLES_KEY);
        ok &= mLeftV.writeXml(chambersNode);
        ok &= mSeptV.writeXml(chambersNode);
        ok &= mRigtV.writeXml(chambersNode);
    }
    return ok;
}


bool SVentricle::readXml(const QDomNode &node)
{
    mVentrType = (ventricle_type)readInt(node, VENTRICLE_TYPE);
    bool ok = SCircElement::readXml(node);
    if ( ok )
    {        
        QDomNodeList myChambers = node.toElement().elementsByTagName(MYVENTRICLES_KEY);
        if ( myChambers.size() == 1 )
        {
            myChambers = myChambers.at(0).childNodes();
            for (int i=0; i < myChambers.size(); i++ )
            {
                SChamber chamber;
                chamber.readXml( myChambers.at(i) );
                if ( chamber.getChamberType() == LEFT_VENTRICLE )
                {
                    mLeftV = chamber;
                }
                else if ( chamber.getChamberType() == SEPTUM )
                {
                    mSeptV = chamber;
                }
                else if ( chamber.getChamberType() == RIGHT_VENTRICLE )
                {
                    mRigtV = chamber;
                }
                else
                {
                    ok = false;
                    qDebug().nospace() << "Do not know what the ventricular chamber type is";
                }
            }
        }
        else
        {
            ok = false;
        }
    }
    return ok;
}



//bool MVentricle::writeState (QDomNode &parentNode) const
//{
//    QDomNode stateNode(createXmlEntry(parentNode,STATE_VENTRICLE_KEY));
//    MCircElement::writeState(stateNode);

//    addValue(stateNode, STATE_VOLUME_MIDWALL,       mVm);

//    return !stateNode.isNull();
//}

///// Abstract method to read state from file
//bool MVentricle::readState(const QDomNode &node)
//{
//    QDomNodeList elements = node.toElement().elementsByTagName(STATE_VENTRICLE_KEY);
//    if ( elements.count() == 1 )
//    {
//        QDomElement stateNode = elements.at(0).toElement();
//        MCircElement::readState(stateNode);
//        mVm     = readValue(stateNode, STATE_VOLUME_MIDWALL);
//        return true;
//    }
//    return false;
//}

