/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: ventricle.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef _VENTRICLE_H_
#define _VENTRICLE_H_

#include "defs.h"
#include "chamber.h"
#include "circelement.h"
#include "PCircAdaptDefines.h"


using namespace Peacs::Support;
namespace Peacs
{

    namespace CircAdapt
    {

        class TVentricle;
        /**
        Class for muscular chamber. Inherits from MCompliance.
        */
        class SVentricle: public SCircElement
        {

        public:

            /* @brief Default Constructor */
            SVentricle(ventricle_type type, bool adaptToExercise);

            /* @brief Constructor */
            SVentricle(ventricle_type type, bool adaptToExercise, const QString &name);

            SVentricle(const SVentricle & ventricle);

            virtual ~SVentricle() {;}


            virtual const SVentricle& operator = (const SVentricle &ventr);

            void init();

            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const TVentricle &ventr);


            SCompliance & rightVentricle() { return mRigtV; }

            SCompliance & leftVentricle() { return mLeftV; }

            const SChamber & getRightVentricle() const { return mRigtV; }

            const SChamber & getLeftVentricle() const { return mLeftV; }

            const SChamber & getSeptum() const { return mSeptV; }


            ventricle_type getType() const { return mVentrType; }

            double getVolume() const;

            void adjustForSackPressure(double pressure);


            /// Method to initialize cavity pressure and cross-section
            void initMechanics (double tCycleAct, double tCycleRef);

            void updateCavityMechanics (double deltaT);

            double estimateLeftLeakSignal(double leakpar) const { return mLeftV.estimateLeakSignal(leakpar); }

            double estimateRightLeakSignal(double leakpar) const { return mRigtV.estimateLeakSignal(leakpar); }

            /// Implement method to initialize adaptation variables
            void initAdaptation();

            /// Implement method to update adaptation variables
            void updateAdaptation();

            /// Method to perform adaptation
            void adapt (bool adaptWV, bool adaptSL, bool adaptWA, int nrAdaptUpdates);



            /// Static method to update cavity pressures and cross-sections for ventricles
            void updateVentricleMechanics (double deltaT);

            /// Static method to initialize adaption variables for sheets at junctions
            void initAdaptJunction ();

            /// Static method to update adaptation variables for sheets at junctions
            void updateAdaptJunction ();

            /// Static method to adapt sheets at junctions
            void adaptJunction (int nrAdaptUpdates);


            /// Method to compute Am, Cm and Xm from Vm and Ym
            void computeAmCmXm();

            //==================== file functions ===================

            /** @brief returns the default version number */
            int getDefaultVersion(void) const;

            /** @brief Returns the class name */
            const char *getClassName(void) const;

            /// Virtual implementation of method to write specification to file
            virtual bool writeXml(QDomNode &parentNode) const;

            /// Virtual implementation of method to read specification from file
            virtual bool readXml(const QDomNode &parentNode);

        //    /// Abstract method to write state to file
        //    bool writeState (QDomNode &parentNode) const;

        //    /// Abstract method to read state from file
        //    bool readState (const QDomNode &node);


        private:
            /// Static method to estimate Tx and Ty using linear approximation
            void estimateTxTy (double mVm, double mYm, double AmOld, double mAmDead,
                                      double TOld, double mDTDAm, double &Tx, double &Ty);

        private:
            SChamber mLeftV;
            SChamber mSeptV;
            SChamber mRigtV;

            // variables for adaptation of sheets at junction
            // Tension direction vector left ventricle
            double mkL1;
            double mkL2;

            // Tension direction vector septum
            double mkS1;
            double mkS2;

            // Tension direction vector right ventricle
            double mkR1;
            double mkR2;

            // Sum of tensions at junction
            double mTensionSum;

            ventricle_type mVentrType;
        };
    }
}

#endif
