/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pfile.h 574 2012-12-05 21:11:44Z eelco $
 *
 */

#ifndef _PFILE_H
#define _PFILE_H

#include <QDataStream>
#include <QString>
#include <QStringList>
#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {

        /** @class PFile pfile.h "file/pfile.h"
          * @brief Mainly a virtual function class
          */
        class PEACSSUPPORTSHARED_EXPORT PFile
        {

        protected:

            /** @brief Flag that indicated if the default version must be used
              * Flag that indicated if the default version of the class reading or writing must be used
              * or the specified version. Used in the function writeHead(...)
              */
            static const int USE_DEFUALT_VERSION = -1;


            /** @brief flag that indicated that the read file (class) header is wrong or was failed to read
              * used in readHead(...)
              */
            static const int READ_HEADER_FAILED = -1;


            /** @brief Constructor used by inherted classes.
              */
            PFile( void );


            /** @brief Copy constructor used by inherted classes.
              * @param source the source to copy from
              */
            PFile( const PFile& source );


            /** @brief destructor
              */
            virtual ~PFile( void ) {}


            /** @brief Asignment operator implementation
              *  @param source the source to copy from
              */
            virtual const PFile & operator= ( const PFile & source );


            /** @brief Pure virtual function that requests the default class
              * data version to write to the case file
              */
            virtual int getDefaultVersion( void ) const = 0;


            /** @brief Reads the class identifier name and the class data
              * version number from the stream. This function must be
              * called before the read(...) function is called or must be
              * the first function called when reading from stream
              * @param stream The (file) data stream to read from
              * @param className the name of the class that is expected to be read from the file. If
              * empty (default) the returned value of getClassName() is used instead
              * @return the version number read from the file if found, otherwise READ_HEADER_FAILED
              * @see virtual bool read(QDataStream &)=0;
              */
            int readHead( QDataStream & stream, const QString & className = QString() );


            /** @brief Writes the class identifier name and the class data
              * version number to the stream. This function must be called
              * before the write(...) function is called or must be
              * the first function called when writimg to stream
              * @see bool read(QDataStream &)=0;
              * @param stream  The (file) data stream to read from
              * @param className the name of the class that is to be written to the file. If
              * empty (default) the returned value of getClassName() is used instead
              * @param versionNr The version number to use when writing the header (default USE_DEFUALT_VERSION)
              * @return true if the header is written
              */
            bool writeHead( QDataStream& stream,
                            const QString &className = QString(),
                            int versionNr = USE_DEFUALT_VERSION ) const;


        public:

            /** @brief Pure virtual function Read all avaliable for a specific
              * class from a file. Before calling this function the readHead(...)
              * function must have been called or must be the first function
              * called when reading from stream
              *
              * @see bool readHead(QDataStream& stream);
              * @see virtual bool write(QDatastream &ds)const =0;
              * @param stream the (file) data stream to read from
              */
            virtual bool read( QDataStream & stream ) = 0;


            /** @brief Pure virtual function write all avaliable class data
              * to a file. This function must be called
              * before the write(...) function is called or must be
              * the first function called when writimg to stream
              * @see virtual bool read(QDatastream &ds)=0;
              * @param stream the (file) data stream to write to
              */
            virtual bool write( QDataStream &stream ) const = 0;


            /**
              * @brief Checks the status of the stream for neing OK
              * @param extraOk extra flag to check and use in result
              * @param stream the stream to check and return the new flag value
              * @return true if the status of the stream is OK and the extraOk
              * flag is true, otherwise false
              */
            inline bool isOk( bool & extraOk, const QDataStream & stream ) const {
                return extraOk = extraOk && (stream.status() == QDataStream::Ok);
            }


            /** @brief Checks the status of the stream for neing OK
              * @param stream the stream to check
              * @return true if the status of the stream is OK, otherwise false
              */
            inline bool isOk( const QDataStream& stream ) const {
                return stream.status() == QDataStream::Ok;
            }


            /** @brief Returns the class name as contained by this class, i.e.
              * not the classname returnd by the virtual function
              * {@link getClassName()} although these names should be equal
              */
            const QString & getLoadedClassName( void ) const;


        protected:

            /** @brief Returns the class name
              */
            virtual const char *getClassName( void ) const = 0;


        private:

            /** @brief The name should be the name of the class that
              * inherites from PFile
              */
            QString mLoadedClassName;
        };
    }
}

#endif // _PFILE_H
