/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: P3DVector.cpp 1196 2014-04-15 10:20:27Z eelco $
 *
 */

#include "P3DVector.h"
#include "math.h"
#include "pCoordinateSystem.h"

#ifdef Q_CC_MSVC
#pragma warning( disable : 4351 )
#endif


using namespace Peacs::Support;


P3DVector::P3DVector( void ) :
    mVector()
{
    mVector[X] = 0;
    mVector[Y] = 0;
    mVector[Z] = 0;
}


P3DVector::P3DVector( const P3DVector &vector ) :
    mVector()
{
    mVector[X] = vector.mVector[X];
    mVector[Y] = vector.mVector[Y];
    mVector[Z] = vector.mVector[Z];
}


P3DVector::P3DVector( double x, double y, double z ) :
    mVector()
{
    mVector[X] = x;
    mVector[Y] = y;
    mVector[Z] = z;
}


const P3DVector &P3DVector::operator= ( const P3DVector &vector )
{
    if ( this != &vector )
    {
        mVector[X] = vector.mVector[X];
        mVector[Y] = vector.mVector[Y];
        mVector[Z] = vector.mVector[Z];
    }

    return *this;
}


bool P3DVector::operator==  (const P3DVector &vector) const
{
    return (mVector[X] == vector.mVector[X] &&
            mVector[Y] == vector.mVector[Y] &&
            mVector[Z] == vector.mVector[Z]);
}

bool P3DVector::operator!=  (const P3DVector &vector) const
{
    return ( mVector[X] != vector.mVector[X] ||
             mVector[Y] != vector.mVector[Y] ||
             mVector[Z] != vector.mVector[Z] );
}


const P3DVector &P3DVector::operator+= (const P3DVector &vector)
{
    mVector[X] += vector.mVector[X];
    mVector[Y] += vector.mVector[Y];
    mVector[Z] += vector.mVector[Z];

    return *this;
}


const P3DVector &P3DVector::operator-= (const P3DVector &vector)
{
    mVector[X] -= vector.mVector[X];
    mVector[Y] -= vector.mVector[Y];
    mVector[Z] -= vector.mVector[Z];

    return *this;
}


const P3DVector &P3DVector::operator*= (const P3DVector &vector)
{
    mVector[X] *= vector.mVector[X];
    mVector[Y] *= vector.mVector[Y];
    mVector[Z] *= vector.mVector[Z];

    return *this;
}


P3DVector P3DVector::operator+ (const P3DVector &vector) const
{
    return P3DVector( mVector[X] + vector.mVector[X],
                      mVector[Y] + vector.mVector[Y],
                      mVector[Z] + vector.mVector[Z] );
}


P3DVector P3DVector::operator- (const P3DVector &vector) const
{
    return P3DVector( mVector[X] - vector.mVector[X],
                      mVector[Y] - vector.mVector[Y],
                      mVector[Z] - vector.mVector[Z] );
}


P3DVector P3DVector::operator* (const P3DVector &vector) const
{
    return P3DVector( mVector[X] * vector.mVector[X],
                      mVector[Y] * vector.mVector[Y],
                      mVector[Z] * vector.mVector[Z] );
}


P3DVector P3DVector::operator* (double factor) const
{
    return P3DVector( mVector[X] * factor,
                      mVector[Y] * factor,
                      mVector[Z] * factor );
}


P3DVector P3DVector::operator/ (double factor) const
{
    if ( factor != 0 )
    {
        return P3DVector( mVector[X] / factor,
                          mVector[Y] / factor,
                          mVector[Z] / factor );
    }

    return P3DVector( *this );
}


const P3DVector &P3DVector::operator*= (double factor)
{
    mVector[X] *= factor;
    mVector[Y] *= factor;
    mVector[Z] *= factor;

    return *this;
}


const P3DVector &P3DVector::operator/= (double factor)
{
    if ( factor != 0.0 )
    {
        mVector[X] /= factor;
        mVector[Y] /= factor;
        mVector[Z] /= factor;
    }

    return *this;
}


const P3DVector &P3DVector::operator/= (const P3DVector & vector)
{
    mVector[X] /= vector.x();
    mVector[Y] /= vector.y();
    mVector[Z] /= vector.z();

    return *this;
}


P3DVector P3DVector::operator/ (const P3DVector & vector)
{
    P3DVector vec( *this );

    vec /= vector;

    return vec;
}


P3DVector  P3DVector::cross(const P3DVector &vector) const
{
    return P3DVector( mVector[Y] * vector.mVector[Z] - mVector[Z] * vector.mVector[Y],
                      mVector[Z] * vector.mVector[X] - mVector[X] * vector.mVector[Z],
                      mVector[X] * vector.mVector[Y] - mVector[Y] * vector.mVector[X] );
}


double  P3DVector::dot(const P3DVector &vector) const
{
    return mVector[X] * vector.mVector[X] +
           mVector[Y] * vector.mVector[Y] +
           mVector[Z] * vector.mVector[Z];
}


void P3DVector::rotate(double phi, double theta, double gamma,const P3DVector &rotPnt)
{
    PCoordinateSystem sys( phi, theta, gamma );

    if ( rotPnt == P3DVector( 0, 0, 0 ) )
    {
        sys.toLocalSystem( *this );
    }
    else
    {
        *this -= rotPnt;

        sys.toLocalSystem( *this );

        *this += rotPnt;
    }
}


void P3DVector::rotate(const P3DVector &angles, const P3DVector &rotPnt)
{
    PCoordinateSystem sys( angles.x(), angles.y(), angles.z() );

    if ( rotPnt == P3DVector( 0, 0, 0 ) )
    {
        sys.toLocalSystem( *this );
    }
    else
    {
        *this -= rotPnt;

        sys.toLocalSystem( *this );

        *this += rotPnt;
    }
}


void P3DVector::rotate(const P3DVector &axis,double angle,const P3DVector &rotPnt)
{
    PCoordinateSystem sys;

    sys.createLocalSystemWithGivenAxisAndAngle( axis, angle );

    *this -= rotPnt;

    sys.toLocalSystem( *this );

    *this += rotPnt;
}


P3DVector P3DVector::rotated(double phi, double theta, double gamma,const P3DVector &rotPnt)const
{
    PCoordinateSystem   sys( phi, theta, gamma );
    P3DVector           rotVer;

    if ( rotPnt == P3DVector( 0, 0, 0 ) )
    {
        rotVer = sys.getToLocalSystem( *this );
    }
    else
    {
        rotVer = *this - rotPnt;

        sys.toLocalSystem( rotVer );

        rotVer += rotPnt;

    }

    return rotVer;
}


P3DVector P3DVector::rotated(const P3DVector &angles, const P3DVector &rotPnt)const
{
    PCoordinateSystem   sys( angles.x(), angles.y(), angles.z() );
    P3DVector           rotVer;

    if ( rotPnt == P3DVector( 0, 0, 0 ) )
    {
        rotVer = sys.getToLocalSystem( *this );
    }
    else
    {
        rotVer = sys.getToLocalSystem( *this - rotPnt ) + rotPnt;
    }

    return rotVer;
}


P3DVector P3DVector::rotashsc(const P3DVector &rotation, const P3DVector & center,const P3DVector & shift, const P3DVector & scale) const
{
    P3DVector ver( *this );
    ver -= center;
    ver.rotate(rotation);
    ver *= scale;
    ver += (shift + center);
    return ver;//( ver.rotated( rotation, center ) + shift) * scale;
}


void P3DVector::limitPrecision(int precision)
{
    double prec = pow( 10.0, (double)precision );

    mVector[X] = qRound( mVector[X] * prec ) / prec;
    mVector[Y] = qRound( mVector[Y] * prec ) / prec;
    mVector[Z] = qRound( mVector[Z] * prec ) / prec;
}


double P3DVector::length(void) const
{
    return ( sqrt( sqlength() ) );
}


double P3DVector::sqlength(void) const
{
    return mVector[X] * mVector[X] +
           mVector[Y] * mVector[Y] +
           mVector[Z] * mVector[Z];
}


P3DVector P3DVector::normalized(void) const
{
    P3DVector   v( *this );
    double      l = length();

    if ( l != 0 )
    {
        v /= l;
    }
    else
    {
        v.toZeros();
        qDebug( "P3DVector::normalized() devide by zero" );
    }

    return v;
}


const P3DVector& P3DVector::normate(void)
{
    double l = length();

    if ( l != 0 )
    {
        mVector[X] /= l;
        mVector[Y] /= l;
        mVector[Z] /= l;
    }
    else
    {
        toZeros();
        qDebug( "P3DVector::normate() devide by zero" );
    }

    return *this;
}


P3DVector P3DVector::normalVector( void ) const
{
    if ( !isZero() )
    {
        if ( mVector[X] == 0 )
        {
            return P3DVector( 1.0, 0.0, 0.0 );
        }
        else if ( mVector[Y] == 0 )
        {
            return P3DVector( 0.0, 1.0, 0.0 );
        }
        else if ( mVector[Z] == 0 )
        {
            return P3DVector( 0.0, 0.0, 1.0 );
        }
        else
        {
            P3DVector nm( 1.0, 0.0, -1.0 * mVector[X] / mVector[Z] );

            return nm.normalized();
        }
    }
    else
    {
        qDebug( "P3DVector::normalvector() vector is zero" );
        return P3DVector( 0, 0, 0 );
    }
}


void P3DVector::toZeros( void )
{
    mVector[X] = 0.0;
    mVector[Y] = 0.0;
    mVector[Z] = 0.0;
}


QDataStream &operator<< (QDataStream &stream, const P3DVector &vec)
{
    stream << (double)vec[P3DVector::X];
    stream << (double)vec[P3DVector::Y];
    stream << (double)vec[P3DVector::Z];

    return stream;
}


QDataStream &operator>> (QDataStream &stream, P3DVector &vec)
{
    stream >> vec[P3DVector::X];
    stream >> vec[P3DVector::Y];
    stream >> vec[P3DVector::Z];

    return stream;
}


QTextStream &operator<< (QTextStream &stream, const P3DVector &vec)
{
    stream << vec[P3DVector::X] << vec[P3DVector::Y] << vec[P3DVector::Z];

    return stream;
}


QTextStream &operator >> (QTextStream &stream, P3DVector &vec)
{
    stream >> vec[P3DVector::X] >> vec[P3DVector::Y] >> vec[P3DVector::Z];

    return stream;
}
