/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: P3DVector.h 428 2012-03-21 11:04:31Z peter $
 *
 */

#ifndef __P3DVECTOR_H
#define __P3DVECTOR_H

#include <QMetaType>
#include <QTextStream>
#include <QDataStream>

#include "PeacsSupport_global.h"

namespace Peacs
{
    namespace Support
    {
        /** @class P3DVector p3dvector.h "support/p3dvector.h"
          * @brief This class describes a vector or point in 3D Space
          */
        class PEACSSUPPORTSHARED_EXPORT P3DVector
        {

        public:
            /** @brief identified the number of coordinates the vector holds
              */
            static const int VECTOR_ITEM_COUNT = 3;


            /** @brief type definition of the number of items in P3DVector.
              */
            enum xyz_t { X = 0,
                         Y = 1,
                         Z = 2 };


            /** @brief Default constructor.
              * Initialze all coordinates to 0
              */
            P3DVector( void );


            /** @brief Copy constructor.
              * @see operator=(const P3DVector &vector)
              * @param vector the source to initizlize with
              */
            P3DVector( const P3DVector &vector );


            /** @brief Initializer constructor.
              * @param x x position of the vector
              * @param y y position of the vector
              * @param z z position of the vector
              */
            P3DVector( double x, double y, double z );


            /** @brief Checks if the vector is equal to (0, 0, 0)
              * @return true if zero
              */
            inline bool isZero( void ) const { return mVector[X] == 0 && mVector[Y] == 0 && mVector[Z] == 0; }


            /** @brief assignment operator.
              * copy the coordinates from vector
              * @param vector the one that is copied
              */
            const P3DVector &operator= (const P3DVector &vector);


            /** @brief Returns the item at index position index as a modifiable
              * reference.
              * @param index a valid index position in the list
              * (i.e. 0 <= index < VECTOR_ITEM_COUNT)
              */
            inline double& operator[] (xyz_t index) { return mVector[index]; }


            /** @brief This is an member function, provided for
              * convenience.
              * @param index a valid index position in the list
              * (i.e. 0 <= index < VECTOR_ITEM_COUNT)
              */
            inline double operator[] (xyz_t index) const { return mVector[index]; }


            /** @brief In thi sway the data can be accessed directly,
              * required for instance in open gl interace
              */
            inline const double *data(void) const {return mVector; }


            /** @brief equal operator.
              * true if all coordinates match between vector and this one
              * @param vector the one that is used to check
              * @see operator!=(const P3DVector &vector)
              */
            bool operator==  (const P3DVector &vector) const;


            /** @brief unequal operator.
              * true if one of the coordinates does not match between vector
              * and this one
              * @param vector the one that is used to check
              * @see operator==(const P3DVector &vector)
              */
            bool operator!=  (const P3DVector &vector) const;


            /** @brief += operator.
              * add the input vector to this one and returns the object
              * @param vector the one that is added
              * @see operator+(const P3DVector &vector)
              */
            const P3DVector &operator+= (const P3DVector &vector);


            /** @brief -= operator.
              * subtract the vector from this one and returns the object
              * @param vector the one that is subtracted
              * @see operator-(const P3DVector &vector)
              */
            const P3DVector &operator-= (const P3DVector &vector);


            /** @brief *= product operator.
              * take the product from the input vector and this one
              * and returns the object
              * @param vector the one that is used in the product
              * @see operator*(const P3DVector &vector)
              */
            const P3DVector &operator*= (const P3DVector &vector);


            /** @brief + operator.
              * Adds the vector to this vector and returns the object
              * @param vector the one that is added
              * @see P3DVector operator+=(const P3DVector &vector)
              */
            P3DVector  operator+  (const P3DVector &vector) const;


            /** @brief -operator.
              * subtract the input vector from this one vector and returns
              * this vector.
              * @param vector the one that is subtracted
              * @see &operator-=(const P3DVector &vector)
              */
            P3DVector  operator-  (const P3DVector &vector) const;


            /** @brief returns product of vector and this one and
              * returns this vector.
              * @param vector that is used in the product
              */
            P3DVector  operator*  (const P3DVector &vector) const;


            /** @brief multiplies the vector by factor and returns this vector.
              * @param factor Multiplication factor
              * @see   operator*=(const double &factor)
              */
            P3DVector  operator*  (double factor) const;


            /** @brief returns a vector(copied from this object)
              * devided by factor.
              * @param factor Devision factor
              * @see   operator/=(const double &factor)
              */
            P3DVector  operator/  (double factor) const;


            /** @brief multiplies the vectro by factor and returns the object
              * @param factor Multiplication factor
              * @see   operator*(const double &factor)
              */
            const P3DVector &operator*= (double factor);


            /** @brief Devides the vector by factor and returns this object.
              * @param factor Devision factor
              *	@see   &operator/(const double &factor)
              */
            const P3DVector &operator/= (double factor);


            /** @brief Devides the vector by factor and returns this object.
              * @param vector Devision vector per dimension
              * @see   &operator/(const double &factor)
              */
            const P3DVector &operator/= (const P3DVector &vector);


            /** @brief Devides the vector by factor and returns this object.
              * @param vector Devision vector per dimension
              * @see   &operator/(const double &factor)
              */
            P3DVector operator/ (const P3DVector &vector);


            /** @brief returns cross product of this vector and vector.
              * @param vector that is used in the cross product
              */
            P3DVector  cross(const P3DVector &vector)const;


            /** @brief returns dot product of both vectors.
              * @param vector that is used in the dot product
              */
            double  dot(const P3DVector &vector)const;


            /** @brief rotate vertices.
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param angles    phi,theta,gamma
              * @param rotPnt    the point around which the geometry is rotated
              */
            void rotate(const P3DVector &angles, const P3DVector &rotPnt=P3DVector());


            /** @brief rotate vertices.
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param phi       rotation around  ?? axes
              * @param theta     rotation around  ?? axes
              * @param gamma     rotation around  ?? axes
              * @param rotPnt    the point aroudn which the geometry is rotated
              */
            void rotate(double phi, double theta, double gamma,const P3DVector &rotPnt=P3DVector());


            /** @brief copies the vertices and rotate these copied vertices
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param angles    phi,theta,gamma
              * @param rotPnt    the point around which the geometry is rotated
              */
            P3DVector rotated(const P3DVector &angles, const P3DVector &rotPnt=P3DVector()) const;


            /** @brief copies the vertices and rotate these copied vertices
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param phi       rotation around  ?? axes
              * @param theta     rotation around  ?? axes
              * @param gamma     rotation around  ?? axes
              * @param rotPnt    the point aroudn which the geometry is rotated
              */
            P3DVector rotated(double phi, double theta, double gamma,const P3DVector &rotPnt=P3DVector()) const;


            /** @brief TODO
              * @param axis
              * @param angle
              * @param rotPnt
              */

            void rotate(const P3DVector &axis, double angle,const P3DVector &rotPnt = P3DVector());

            /** @brief TODO
              * @param rotation
              * @param center
              * @param shift
              * @param scale
              * @return
              */
            P3DVector rotashsc( const P3DVector & rotation,
                                const P3DVector & center,
                                const P3DVector & shift,
                                const P3DVector & v ) const;


            /** @brief Limit the precisoon of the vertices
              * @param precision the number of digits
              */
            void limitPrecision( int precision = 3 );


            /** @brief returns length of this vector.
              */
            double length( void ) const;


            /** @brief returns squared length of the vector.
              */
            double sqlength( void ) const;


            /** @brief returns the normated vector.
              */
            P3DVector normalized( void )const;


            /** @brief normalized this vector
              */
            const P3DVector & normate(void);


            /** @brief normalVector returns a vector perpendicular to this vector assuming
              * it starts at (0,0,0)
              */
            P3DVector normalVector( void ) const ;


            /** @brief returns the x coordinate.
              */
            inline double &x( void ) { return mVector[X]; }


            /** @brief returns the y coordinate.
              */
            inline double &y( void ) { return mVector[Y]; }


            /** @brief returns the z coordinate.
              */
            inline double &z( void ) { return mVector[Z]; }



            /** @brief returns the x coordinate.
              */
            inline double x( void ) const { return mVector[X]; }


            /** @brief returns the y coordinate.
              */
            inline double y( void ) const { return mVector[Y]; }


            /** @brief returns the z coordinate.
              */
            inline double z( void ) const{ return mVector[Z]; }


            /** @brief clears the content of the vector to 0 0 0(zero)
              */
            void toZeros( void );


        private:

            /** @brief 3D coordinate (x,y,z).
              */
            double  mVector[VECTOR_ITEM_COUNT];
        };
    }
}

/** @brief Streams the 3 coordinates to a datastream
  * @param vec The vecor that is streamed
  * @returns stream
  * @see P3DVector
  */
QDataStream &operator<< ( QDataStream &stream, const Peacs::Support::P3DVector &vec );


/** @brief Streams the 3 coordinates from a datastream
  * @param vec The vecor that is streamed
  * @returns stream
  * @see P3DVector
  */
QDataStream &operator>> ( QDataStream &stream, Peacs::Support::P3DVector &vec );


/** @brief Streams the 3 coordinates to a textstream
  * @param vec The vecor that is streamed
  * @returns stream
  * @see P3DVector
  */
QTextStream &operator<< ( QTextStream &stream, const Peacs::Support::P3DVector &vec );


/** @brief Streams the 3 coordinates from a textstream
  * @param vec The vecor that is streamed
  * @returns stream
  * @see P3DVector
  */
QTextStream &operator>> ( QTextStream &stream, Peacs::Support::P3DVector &vec );

// This makes it possible for TTubeState values to be stored in QVariant
// objects and retrieved later
Q_DECLARE_METATYPE( Peacs::Support::P3DVector );

#endif //__P3DVECTOR_H

