/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: P3DVertices.cpp 1196 2014-04-15 10:20:27Z eelco $
 *
 */

#include "P3DVertices.h"
#include "pfunctions.h"
#include "pCoordinateSystem.h"

#include <QFile>
#include <QDebug>


using namespace Peacs::Support;


P3DVertices::P3DVertices() :
    mVertices()
{
}


P3DVertices::P3DVertices(int nVer) :
    mVertices( qMax( 0, nVer ) )
{
    if ( nVer < 0 )
    {
        qDebug( "P3DVertices constructor n <0" );
    }
}


P3DVertices::P3DVertices(int nVer,const P3DVector & init) :
    mVertices( nVer, init )
{
}


P3DVertices::P3DVertices(const P3DVertices &vertices) :
    mVertices( vertices.mVertices )
{
}


P3DVertices::P3DVertices(const QList <P3DVector>& vertices) :
    mVertices( vertices.toVector() )
{
}


P3DVertices::P3DVertices(const QVector <P3DVector>& vertices) :
    mVertices( vertices )
{
}


P3DVertices::~P3DVertices()
{
}


void P3DVertices::clear()
{
    mVertices.clear();
}


void P3DVertices::resizeAndToZero(int newsize)
{
    mVertices.resize( newsize );

    toZeros();
}


void P3DVertices::addVertices(const P3DVertices &addVertices)
{
    mVertices += addVertices.mVertices;
}


P3DVertices P3DVertices::select( int i1, int i2 )
{
    P3DVertices ret;

    if ( i1 <= i2 &&
         i1 < mVertices.size() &&
         i2 < mVertices.size() )
    {
        int j = 0;

        ret.resizeAndToZero( i2 - i1 + 1 );

        for ( int i = i1; i <= i2; i++ )
        {
            ret[j] = mVertices.at( i );

            j++;
        }
    }

    return ret;
}


const P3DVertices& P3DVertices::operator = (const P3DVertices& vertices)
{
    if ( this != &vertices )
    {
        mVertices = vertices.mVertices;
    }

    return *this;
}


int P3DVertices::addVertex(const P3DVector& addVertex, bool docheck)
{
    int index = -1;

    if ( !docheck )
    {
        mVertices.append( addVertex );
    }
    else
    {
        index = findVertex( addVertex );

        if (  index < 0 )
        {
            index = mVertices.size();

            mVertices.append( addVertex );
        }
    }

    return index;
}


bool P3DVertices::insertVertex(int i, const P3DVector& addVertex, bool docheck )
{
    if ( docheck )
    {
        if ( !mVertices.contains( addVertex ) )
        {
            mVertices.insert( i, addVertex );

            return true;
        }
    }
    else
    {
        mVertices.insert( i, addVertex );

        return true;
    }

    return false;
}


void P3DVertices::removeVertex(int vertex)
{
    if ( vertex >= 0 &&
         vertex < size() )
    {
        mVertices.remove( vertex );
    }
    else
    {
        qDebug() << "vertex not removed";
    }
}


bool P3DVertices::operator ==(const P3DVertices &other)const
{
    if ( this == &other )
    {
        return true;
    }
    else if ( other.mVertices.size() == mVertices.size() )
    {
        for ( int i = 0; i < size(); i++ )
        {
            if ( mVertices.at( i ) != other.at( i ) )
            {
                return false;
            }
        }

        return true;
    }

    return false;
}


bool P3DVertices::operator !=(const P3DVertices &other)const
{
    if ( this == &other )
    {
        return false;
    }
    else if ( other.mVertices.size() == mVertices.size() )
    {
        for ( int i = 0; i < size(); i++ )
        {
            if ( mVertices.at( i ) != other.at( i ) )
            {
                return true;
            }
        }
        return false;
    }
    return true;
}


P3DVertices P3DVertices::operator + ( const P3DVertices &other)const
{
    P3DVertices vertices( *this );

    vertices += other;

    return vertices;
}


P3DVertices P3DVertices::operator - ( const P3DVertices &other)const
{
    P3DVertices vertices( *this );

    vertices -= other;

    return vertices;
}


const P3DVertices &P3DVertices::operator += ( const P3DVertices &other)
{
    if ( isEmpty() )
    {
        *this = other;
    }
    else if ( other.isEmpty() )
    {
        // Nothing to do
    }
    else if ( size() == other.size())
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            mVertices[i] += other.at( i );
        }
    }
    else
    {
        qDebug("P3DVertices += operator: wrong sizes");
    }

    return *this;
}


const P3DVertices &P3DVertices::operator -= ( const P3DVertices &other)
{
    if ( isEmpty() )
    {
        *this = other;
        *this *= P3DVector( -1.0, -1.0, -1.0 );
    }
    else if ( other.isEmpty() )
    {
        // Nothing to do
    }
    else if ( size() == other.size() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            mVertices[i] -= other.at( i );
        }
    }
    else
    {
        qDebug("P3DVertices -= operator: wrong sizes");
    }

    return *this;
}


P3DVertices P3DVertices::operator+(const P3DVector &other)const
{
    P3DVertices vertices( *this );

    vertices += other;

    return vertices;
}


P3DVertices P3DVertices::operator-(const P3DVector &other)const
{
    P3DVertices vertices( *this );

    vertices -= other;

    return vertices;
}


const P3DVertices &P3DVertices::operator+=(const P3DVector &other)
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i] += other;
    }

    return *this;
}


const P3DVertices &P3DVertices::operator-=(const P3DVector &other)
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i] -= other;
    }

    return *this;
}


P3DVertices P3DVertices::operator*(const P3DVector &other)const
{
    P3DVertices vertices( *this );

    vertices *= other;

    return vertices;
}


const P3DVertices &P3DVertices::operator*=(const P3DVector &other)
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i] *= other;
    }
    return *this;
}


P3DVertices P3DVertices::operator * ( double fact ) const
{
    P3DVertices vertices( *this );

    vertices *= fact;

    return vertices;
}


const P3DVertices &P3DVertices::operator *= ( double fact )
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i] *= fact;
    }

    return *this;
}


const P3DVertices &P3DVertices::operator /= ( double fact )
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i] /= fact;
    }

    return *this;
}


P3DVertices P3DVertices::operator / ( double fact ) const
{
    P3DVertices vertices( *this );

    vertices /= fact;

    return vertices;
}


const P3DVertices &P3DVertices::operator /= ( const PVector & fact )
{
    if ( fact.N() == mVertices.size() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            mVertices[i] /= fact.at( i );
        }
    }

    return *this;
}


const P3DVertices &P3DVertices::operator *= ( const PVector & fact )
{
    if ( fact.N() == mVertices.size() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            mVertices[i] *= fact.at( i );
        }
    }

    return *this;
}


P3DVertices P3DVertices::operator / ( const PVector & fact ) const
{
    P3DVertices vertices( *this );

    vertices /= fact;

    return vertices;
}


P3DVertices P3DVertices::operator * ( const PVector & fact) const
{
    P3DVertices vertices( *this );

    vertices *= fact;

    return vertices;
}


P3DVertices P3DVertices::cross(const P3DVector &vector)const
{
    P3DVertices crossProd( mVertices.size() );

    for ( int i = 0; i < mVertices.size(); i++ )
    {
        crossProd[i] = mVertices.at( i ).cross( vector );
    }

    return crossProd;
}


P3DVertices P3DVertices::cross(const P3DVertices &vectors)const
{
    if ( vectors.N() == mVertices.size() )
    {
        P3DVertices crossProd( mVertices.size() );

        for ( int i = 0; i < mVertices.size(); i++ )
        {
            crossProd[i] = mVertices.at( i ).cross( vectors.at( i ) );
        }

        return crossProd;
    }
    else
    {
        qDebug( "P3DVertices::cross() vectors has the wrong size" );
    }

    return P3DVertices();
}


/** @brief returns dot product of both vectors.
 * 	@param vector that is used in the dot product        */
PVector P3DVertices::dot( const P3DVector &vector ) const
{
    PVector dots( mVertices.size(), 0.0 );

    for ( int i = 0; i < mVertices.size(); i++ )
    {
        dots[i] = mVertices.at( i ).dot( vector );
    }

    return dots;
}


/** @brief returns dot product of both vectors.
 * 	@param vector that is used in the dot product        */
PVector P3DVertices::dot( const P3DVertices  &vector ) const
{
    PVector dots( mVertices.size(), 0.0 );

    if ( vector.mVertices.size() == mVertices.size() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            dots[i] = mVertices.at( i ).dot( vector.mVertices.at( i ) );
        }
    }
    else
    {
        qDebug( "Error in P3DVertices::dot( P3DVertices )" );
    }

    return dots;
}


void P3DVertices::toZeros(void)
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i].toZeros();
    }
}


P3DVector P3DVertices::getMean(void) const
{
    P3DVector retVal;

    if ( !mVertices.isEmpty() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            retVal += mVertices.at( i );
        }

        retVal /= (double)mVertices.size();
    }

    return retVal;
}


P3DVector P3DVertices::getSum(void) const
{
    P3DVector retVal;

    for ( int i = 0; i < mVertices.size(); i++ )
    {
        retVal += mVertices.at( i );
    }

    return retVal;
}


P3DVertices P3DVertices::getSub(const QList<int> &iVertices) const
{
    P3DVertices set;

    if ( !isEmpty() )
    {
        set.resizeAndToZero( iVertices.size() );

        for (int i = 0; i < iVertices.size(); i++ )
        {
            if ( iVertices.at( i ) >= 0 &&
                 iVertices.at( i ) < mVertices.size() )
            {
                set[i] = mVertices.at( iVertices.at( i ) );
            }
            else
            {
                qDebug( "P3DVertices::getSub(): index out of range" );
            }
        }
    }

    return set;
}


P3DVertices P3DVertices::getSub(const PIVector &iVertices) const
{
    P3DVertices set;

    if ( !isEmpty() )
    {
        set.resizeAndToZero( iVertices.N() );

        for ( int i = 0; i < iVertices.N(); i++ )
        {
            if ( iVertices.at( i ) >= 0 &&
                 iVertices.at( i ) < mVertices.size() )
            {
                set[i] = mVertices.at( iVertices.at( i ) );
            }
            else
            {
                qDebug( "P3DVertices::getSub(): index out of range" );
            }
        }
    }

    return set;
}


P3DVector P3DVertices::meanOf(const QList <int> &iVertices) const
{
    P3DVector meanVect;

    for ( int i = 0; i < iVertices.size(); i++ )
    {
        meanVect += mVertices.at( iVertices.at( i ) );
    }

    return meanVect / double( iVertices.size() );
}


const P3DVertices &P3DVertices::relativeMax()
{
    double maxL = 0;

    for ( int i = 0; i < mVertices.size(); i++ )
    {
        maxL = qMax( maxL, at( i ).length() );
    }

    if ( maxL > 0 )
    {
        *this /= maxL;
    }

    return *this;
}


const P3DVertices &P3DVertices::normate()
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i].normate();
    }

    return *this;
}


P3DVertices P3DVertices::normalized()
{
    P3DVertices vers( *this );

    vers.normate();

    return vers;
}


PVector P3DVertices::norm()const

{
    PVector norms( mVertices.size() );

    for ( int i = 0; i < mVertices.size(); i++ )
    {
        norms[i] = mVertices.at( i ).length();
    }

    return norms;
}


double P3DVertices::getMaxNorm() const
{
    double maxLen = 0;

    for ( int i = 0; i < mVertices.size(); i++ )
    {
        maxLen = qMax( maxLen, mVertices.at( i ).length() );
    }

    return maxLen;
}


double P3DVertices::getMeanNorm() const
{
    double meanLen = 0;

    if ( !mVertices.isEmpty() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            meanLen += mVertices.at( i ).length();
        }

        meanLen /= double( mVertices.size() );
    }

    return meanLen;
}


PVector P3DVertices::sqNorm(void) const
{
    PVector norms( mVertices.size() );

    for ( int i = 0; i < mVertices.size(); i++ )
    {
        norms[i] = mVertices.at( i ).sqlength();
    }

    return norms;
}


P3DVertices P3DVertices::deltaVector(void) const
{
    P3DVertices vec( mVertices.size() );

    if ( !vec.isEmpty() )
    {
        for ( int i = 1; i < mVertices.size(); i++ )
        {
            vec[i] = mVertices.at( i ) - mVertices.at( i - 1 );
        }
    }

    return vec;
}


PVector P3DVertices::deltaLength(void) const
{
    PVector vec( mVertices.size() );

    if ( !vec.isEmpty() )
    {
        for ( int i = 1; i < mVertices.size(); i++ )
        {
            vec[i] = ( mVertices.at( i ) - mVertices.at( i - 1 ) ).length();
        }
    }

    return vec;
}


PVector P3DVertices::cumLength(void) const
{
    PVector vec( mVertices.size() );

    if ( !vec.isEmpty() )
    {
        double length = 0;

        for ( int i = 1; i < mVertices.size(); i++ )
        {
            length += ( mVertices.at( i ) - mVertices.at( i - 1 ) ).length();

            vec[i] = length;
        }
    }

    return vec;
}


void P3DVertices::rotate( double phi, double theta, double gamma, const P3DVector &rotPnt )
{
    PCoordinateSystem sys( phi, theta, gamma );

    if ( rotPnt.isZero() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            sys.toLocalSystem( mVertices[i] );
        }
    }
    else
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            mVertices[i] -= rotPnt;

            sys.toLocalSystem( mVertices[i] );

            mVertices[i] += rotPnt;
        }
    }
}


void P3DVertices::rotate(const P3DVector &angles, const P3DVector &rotPnt)
{
    PCoordinateSystem sys( angles.x(), angles.y(), angles.z() );

    if ( rotPnt.isZero() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            sys.toLocalSystem( mVertices[i] );
        }
    }
    else
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            mVertices[i] -= rotPnt;

            sys.toLocalSystem( mVertices[i] );

            mVertices[i] += rotPnt;
        }
    }
}


void P3DVertices::rotate(const P3DVector &axis, double angle, const P3DVector &rotPnt)
{
    PCoordinateSystem sys;

    sys.createLocalSystemWithGivenAxisAndAngle( axis, angle );

    if ( rotPnt.isZero() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            sys.toLocalSystem( mVertices[i] );
        }
    }
    else
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            mVertices[i] -= rotPnt;

            sys.toLocalSystem( mVertices[i] );

            mVertices[i] += rotPnt;
        }
    }
}


P3DVertices P3DVertices::rotated(const P3DVector &axis, double angle, const P3DVector &rotPnt) const
{
    PCoordinateSystem   sys;
    P3DVertices         rotVer( mVertices );

    sys.createLocalSystemWithGivenAxisAndAngle( axis, angle );

    if ( rotPnt.isZero() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            rotVer[i] = sys.getToLocalSystem( mVertices.at( i ) );
        }
    }
    else
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            rotVer[i] -= rotPnt;

            sys.toLocalSystem( rotVer[i] );

            rotVer[i] += rotPnt;
        }
    }

    return rotVer;
}


P3DVertices P3DVertices::rotated(double phi, double theta, double gamma, const P3DVector &rotPnt)const
{
    PCoordinateSystem   sys( phi, theta, gamma );
    P3DVertices         rotVer( mVertices.size() );

    if ( rotPnt.isZero() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            rotVer[i] = sys.getToLocalSystem( mVertices.at( i ) );
        }
    }
    else
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            rotVer[i] = mVertices.at( i ) - rotPnt;

            sys.toLocalSystem( rotVer[i] );
            rotVer[i] += rotPnt;
        }
    }

    return rotVer;
}


P3DVertices P3DVertices::rotated(const P3DVector &angles, const P3DVector &rotPnt) const
{
    PCoordinateSystem   sys( angles.x(), angles.y(), angles.z() );
    P3DVertices         rotVer( mVertices.size() );

    if ( rotPnt.isZero() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            rotVer[i] = sys.getToLocalSystem( mVertices.at( i ) );
        }
    }
    else
    {
        for ( int i = 0; i < size(); i++ )
        {
            rotVer[i] = sys.getToLocalSystem( mVertices.at( i ) - rotPnt ) + rotPnt;
        }
    }

    return rotVer;
}


P3DVertices P3DVertices::rotashsc(const P3DVector & rotation, const P3DVector &center, const P3DVector & shift, const P3DVector & scale) const
{
    P3DVertices ver( *this );
    //    return ( ver.rotated( rotation, center ) + shift );
    ver -= center;
    ver.rotate(rotation);
    ver *= scale;
    ver += ( shift + center);
    return ver;//( ver.rotated( rotation, center ) + shift) * scale;
}

P3DVertices P3DVertices::scaleKeepCenter(const P3DVector &scale, const P3DVector &center) const
{
    P3DVertices ver( *this );

    ver -= center;

    ver *= scale;

    ver += center;

    return ver;
}

void P3DVertices::scaledKeepCenter(const P3DVector &scale, const P3DVector &center)
{
    for ( int i=0; i <  mVertices.size(); i++)
    {
        mVertices[i] -= center;
        mVertices[i] *= scale;
        mVertices[i] += center;
    }
}


void P3DVertices::reorder(int firstIndex)
{
    P3DVertices ver( mVertices );
    int         k = 0;

    for ( int i = firstIndex; i < mVertices.size(); i++ )
    {
        ver[k] = mVertices[i];

        k++;
    }

    for ( int i = 0; i < firstIndex; i++ )
    {
        ver[k] = mVertices[i];

        k++;
    }

    mVertices = ver.mVertices;
}


void P3DVertices::reorder(const PIVector &iVertices)
{
    P3DVertices ver( mVertices );

    if ( iVertices.N() == mVertices.size() &&
         iVertices.getMin() >= 0  &&
         iVertices.getMax() < mVertices.size() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            ver[i] = mVertices[ iVertices[i] ];
        }
    }
    else
    {
        qDebug() << "vertices not reordred";
    }

    mVertices = ver.mVertices;
}


bool P3DVertices::save(const QString &filename, bool asText) const
{
    int     i;
    bool    ok = false;
    QFile   file( filename );
    if ( file.open( QIODevice::WriteOnly ) )
    {
        if ( asText )
        {
            QTextStream textStream( &file );

            textStream.setRealNumberPrecision( 6 );     // enough for geometry data
            textStream << size() << "\t" << 4;          // 4 indicates the number of columns
            textStream << " \n";

            for ( i = 0; i < size() && (textStream.status() == QTextStream::Ok); i++ )
            {
                textStream << ( i + 1 );                // Store the index
                textStream << "\t";
                textStream << mVertices.at( i ).x();
                textStream << "\t";
                textStream << mVertices.at( i ).y();
                textStream << "\t";
                textStream << mVertices.at( i ).z();
                textStream << "\n";
            }
            if ( textStream.status() == QTextStream::Ok )
            {
                ok = true;
            }
        }
        else
        {
            QDataStream dataStream( &file );
//            dataStream.setByteOrder( QDataStream::LittleEndian ); // default used in matlab
            dataStream.setFloatingPointPrecision( QDataStream::SinglePrecision );

            dataStream << (qint32)mVertices.size();
            for ( i = 0; i < mVertices.size() && (dataStream.status() == QDataStream::Ok); i++ )
            {
                dataStream << mVertices.at( i ).x();
                dataStream << mVertices.at( i ).y();
                dataStream << mVertices.at( i ).z();
            }
            if ( dataStream.status() == QDataStream::Ok )
            {
                ok = true;
            }
        }
    }
    file.close();
    return ok;
}


bool P3DVertices::load(const QString &filename, bool asText)
{
    int     i;
    bool    ok = false;
    QFile   file( filename );

    if ( file.open( QIODevice::ReadOnly ) )
    {
        if ( asText )
        {
            QTextStream textStream( &file );
            int         n;
            int         tmp;

            textStream >> n >> tmp ;        // Read the size and the number of columns

            resizeAndToZero( n );

            for ( i = 0; i < n && (textStream.status() == QTextStream::Ok); i++ )
            {
                textStream >> tmp;          // Read the index
                textStream >> mVertices[i];
            }

            if ( textStream.status() == QTextStream::Ok )
            {
                ok = true;
            }
        }
        else
        {
            QDataStream dataStream( &file );
            qint32         n;
//            dataStream.setByteOrder( QDataStream::LittleEndian ); // default used in matlab
            dataStream.setFloatingPointPrecision( QDataStream::SinglePrecision );
            dataStream >> n;        // Read the size
            if ( n > file.size() / 8 ) // probably text file
            {
                ok = false;
            }
            else
            {
                resizeAndToZero( n );

                for ( i = 0; i < n && (dataStream.status() == QDataStream::Ok); i++ )
                {
                    dataStream >> mVertices[i];
                }

                if ( dataStream.status() == QDataStream::Ok )
                {
                    ok = true;
                }
            }
        }
    }
    file.close();
    return ok;
}


void P3DVertices::limitPrecision(int precision)
{
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i].limitPrecision( precision );
    }
}


/* plindist(p, l1, l2, cp, lab): distance beween point p and
   line segment l1-l2. In cp the point on l1-l2 closed to
   p is returned; in lab the fraction of the distance
   l1-cp in respect to l1-l2 is returned. */
P3DVector P3DVertices::vertexOnLine(const P3DVector &p) const
{
    P3DVector   ret;
    double      minlen = HUGE_VAL;

    for ( int i = 0; i < mVertices.size() - 1; i++ )
    {
        P3DVector   pOnLine( verOnLine( p,
                                        mVertices.at( i ),
                                        mVertices.at( i + 1 ) ) );
        double      len = (p - pOnLine).length();

        if ( len < minlen )
        {
            minlen  = len;
            ret     = pOnLine;
        }
    }

    return ret;
}


P3DVector P3DVertices::vertexOnLine(const P3DVector &p, P3DVector & gradient) const
{
    P3DVector   ret;
    double      minlen = HUGE_VAL;

    for ( int i = 0; i < mVertices.size() - 1; i++ )
    {
        P3DVector   pOnLine( verOnLine( p,
                                        mVertices.at( i ),
                                        mVertices.at( i + 1 ) ) );
        double      len = (p - pOnLine).length();

        if ( len < minlen )
        {
            minlen      = len;
            ret         = pOnLine;
            gradient    = ( mVertices.at( i + 1 ) - mVertices.at( i ) ).normalized();
        }
    }

    return ret;
}


P3DVector P3DVertices::verOnLine(const P3DVector &p,const P3DVector &v1,const P3DVector &v2) const
{
    P3DVector   ret( v1 );
    P3DVector   s( v2 - v1 );
    double      sq = s.sqlength();

    if ( sq != 0 )
    {
        double alpha = (p - v1).dot( s ) / sq;

        alpha = qBound( 0.0, alpha, 1.0 );

        ret += (s * alpha);
    }

    return ret;
}


double P3DVertices::totLength() const
{
    double totL = 0;

    for ( int i = 0; i < mVertices.size() - 1; i++ )
    {
        totL += ( mVertices.at( i ) - mVertices.at( i + 1 ) ).length();
    }

    return totL;
}


void P3DVertices::changeXYaxis(void)//change the XY axis: y=-x and x = y
{
    // The orientation of the tri files diffrent from the MRI data.
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i] = P3DVector( mVertices.at( i ).y(),
                                  -mVertices.at( i ).x(),
                                  mVertices.at( i ).z() );
    }
}


void P3DVertices::changeYXaxis(void) //change the XY axis: y=-x and x = y
{
    // The orientation of the tri files diffrent from the MRI data.
    for ( int i = 0; i < mVertices.size(); i++ )
    {
        mVertices[i] = P3DVector( -mVertices.at( i ).y(),
                                  mVertices.at( i ).x(),
                                  mVertices.at( i ).z() );
    }
}


#ifndef QT_NO_DEBUG

const P3DVector &P3DVertices::at(int index) const
{
    Q_ASSERT( !mVertices.isEmpty() );                       // Test just to be sure
    Q_ASSERT( index >= 0 && index < mVertices.size() );     // Test just to be sure

    return mVertices.at( index );
}


const P3DVector &P3DVertices::last() const
{
    Q_ASSERT( !mVertices.isEmpty() );  // Test just to be sure

    return mVertices.last();
}


const P3DVector &P3DVertices::first() const
{
    Q_ASSERT( !mVertices.isEmpty() );  // Test just to be sure

    return mVertices.first();
}


P3DVector &P3DVertices::operator [] (int index)
{
    Q_ASSERT( !mVertices.isEmpty() );  // Test just to be sure
    Q_ASSERT( index >= 0 && index < mVertices.size() );  // Test just to be sure

    return mVertices[index];
}

void P3DVertices::setVertex(int index,const P3DVector & pos)
{
    Q_ASSERT( !mVertices.isEmpty() );  // Test just to be sure
    Q_ASSERT( index>=0 && index< mVertices.size() );  // Test just to be sure

    mVertices[index] = pos;
}

#endif



