/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: P3DVertices.h 1196 2014-04-15 10:20:27Z eelco $
 *
 */

#ifndef __P3DVERTICES_H__
#define __P3DVERTICES_H__


#include "pvector.h"
#include "P3DVector.h"

#include "PeacsSupport_global.h"
#include <QVector>
#include <QList>

namespace Peacs
{
    namespace Support
    {

        /** @class P3DVertices p3dvertices.h "support/p3dvertices.h"
          * @brief This class describes a list of vertices in 3D Space
          * Several operators can be applied to scale, rotate  and translate
          * the vertices.
          */
        class PEACSSUPPORTSHARED_EXPORT P3DVertices
        {
        public:
            /** @brief constructor, constructs an emtpy list
              */
            P3DVertices();


            /** @brief constructor, constructs a list with nVer vertices
              * All vertices are initialized with the default P3Dvector constructor,
              * i.e. all x,yz coordinates are set to zero.
              * @param nVer the number of vertices
              */
            P3DVertices(int nVer);


            /** @brief constructor, constructs a list with nVer vertices
              * All vertices are initialized with the default P3Dvector constructor,
              * i.e. all x,yz coordinates are set to zero.
              * @param nVer the number of vertices
              */
            P3DVertices(int nVer,const P3DVector & init);


            /** @brief copy constructor, copies the given vertices
              * @param vertices the to be copied vertices the number of vertices
              */
            P3DVertices(const P3DVertices& vertices);

            /** @brief copy constructor, copies the given vertices
              * @param vertices the to be copied vertices the number of vertices
              */
            P3DVertices(const QList <P3DVector>& vertices);


            /** @brief copy constructor, copies the given vertices
              * @param vertices the to be copied vertices the number of vertices
              */
            P3DVertices(const QVector <P3DVector>& vertices);


            /** @brief destructor, deletes the pointer to the vertices
              */
            virtual ~P3DVertices();


            /** @brief append the given vertices to the end of the
              * vertices list
              * @param vertices the vertices that are appended
              */
            void addVertices(const P3DVertices & addVertices);


            /** @brief append the given vertex to the end of the vertices list
              * @param vertex the vertex that are appended
              * @param docheck Must check if the vertex already exists (default false)
              */
            int addVertex(const P3DVector& addVertex, bool docheck = false);


            /** @brief append the given vertex to the end of the vertices list
              * @param vertex the vertex that are appended
              */
            void append(const P3DVector& addVertex) { mVertices.append(addVertex); }


            /** @brief append the given vertex to the end of the vertices list
              * @param i Inserts vertex at index position i in the vector. If i is 0,
              * the value is prepended to the vector. If i is size(), the value is
              * appended to the vector.
              * A check is executed to make sure the vertex does not already exist
              */
            bool insertVertex(int i, const P3DVector& addVertex, bool docheck = false);


            /** @brief remove vertex at vertex index ivertex
              * @param ivertex th eidenc of the vertex that must be removed
              */
            void removeVertex(int ivertex);


            /** @returns true if no vertices are in the list, otherwise false
              */
            inline bool isEmpty(void) const { return mVertices.isEmpty(); }


            /** @returns the number of vertices in the list
              */
            inline int size() const { return mVertices.size(); }


            /** @returns the number of vertices in the list
              */
            inline int N() const { return mVertices.size(); }


            /** @brief copy operator
              * @param vertices thevertices that are copied
              */
            const P3DVertices& operator= ( const P3DVertices &vertices);


            /** @brief equal operator
              * @returns false if the number of vertices is unequal or if one of
              * the vertices is unqual, otherise returns true
              * @param other the vertices that are compared
              */
            bool operator== ( const P3DVertices &other)const;


            /** @brief unequal operator
              * @returns true if the number of vertices is unequal or if one of
              * the vertices is unqual, otherise returns false
              * @param other the vertices that are compared
              */
            bool operator!= ( const P3DVertices &other)const;


            /** @brief add operator
              * @returns a new list of vertices , copied from this list to which other
              * is added. @var other and @var this vertices should be the same length,
              * otherwise an exception is thrown. If other is empty the copied vertices
              * are returned. If this object is empty, other is copied and returned
              * @param other the vertices that are added
              */
            P3DVertices operator+ ( const P3DVertices &other)const;


            /** @brief subtraction operator
              * @returns a new list of vertices , copied from this list to which other
              * is subtracted. @var other and @var this vertices should be the same length,
              * otherwise an exception is thrown. If other is empty the copied vertices
              * are returned. If this object is empty, other is copied and returned
              * @param other the vertices that are subtracted
              */
            P3DVertices operator- ( const P3DVertices &other)const;


            /** @brief add operator
              * @returns this list of vertices to which other is added. @var other
              * and @var this vertices should be the same length,
              * otherwise an exception is thrown. If other is empty the copied vertices
              * are returned. If this object is empty, other is copied and returned
              * @param other the 3D coordinates of a vertex that are added
              */
            const P3DVertices &operator+= ( const P3DVertices &other);


            /** @brief subtraction operator
              * @returns this list of vertices to which other is subtracted. @var other
              * and @var this vertices should be the same length,
              * otherwise an exception is thrown. If other is empty the copied vertices
              * are returned. If this object is empty, other is copied and returned
              * @param other the 3D coordinates of a vertex that are subtracted
              */
            const P3DVertices &operator-= ( const P3DVertices &other);


            /** @brief add operator
              * @returns a new list of vertices, copied from this list to which other
              * is added.
              * @param other the 3D coordinates of a vertex that are subtracted
              */
            P3DVertices operator+ ( const P3DVector &other) const;


            /** @brief subtraction operator
              * @returns a new list of vertices, copied from this list to which other
              * is subtracted.
              * @param other the 3D coordinates of a vertex that are subtracted
              */
            P3DVertices operator- ( const P3DVector &other) const;


            /** @brief add operator
              * @returns this list of vertices to which other is added.
              * @param other the 3D coordinates of a vertex that are subtracted
              */
            const P3DVertices &operator+= ( const P3DVector &other);


            /** @brief subtraction operator
              * @returns this list of vertices from which other is subtracted.
              * @param other the 3D coordinates of a vertex that are subtracted
              */
            const P3DVertices &operator-= ( const P3DVector &other);


            /** @brief multiplication operator
              * @returns Copy this list of vertices and multiply it by other,
              * i.e. x * other.x etc
              * @param other the 3D coordinates of a vertex that are used to multiply
              */
            P3DVertices operator* ( const P3DVector &other) const;


            /** @brief multiplication operator
              * @returns this list of vertices multiplied by a other, i.e. x * other.x etc
              * @param other the 3D coordinates of a vertex that are used to multiply
              */
            const P3DVertices &operator*= ( const P3DVector &other);


            /** @brief multiplication operator
              * @returns this list of vertices multiplied by a factor, i.e. x * fact etc
              * @param fact the multiplivcation factor
              */
            const P3DVertices &operator*= ( double fact);


            /** @brief multiplication operator
              * @returns this list is copied and then of vertices multiplied by a factor
              * , i.e. x * fact etc
              * @param fact the multiplication factor
              */
            P3DVertices operator* ( double fact) const;


            /** @brief devison operator
              * @returns a copy of this list of vertices devided by a factor
              * , i.e. x / fact etc
              * @param fact the devison factor
              */
            const P3DVertices &operator/= ( double fact);


            /** @brief devison operator
              * @returns this list is copied and then of vertices devided by a factor
              * , i.e. x / fact etc
              * @param fact the devison factor
              */
            P3DVertices operator / ( double fact) const;

            /** @brief multiplication operator
              * @returns this list of vertices multiplied by @var fact. @var fact
              * and @var this vertices should be the same length,
              * otherwise this obejct is returned as was
              * @param fact the multiplication factor per vertex
              */
            const P3DVertices &operator*= ( const PVector & fact);


            /** @brief devison operator
              * @returns this list of vertices devided by @var fact. @var fact
              * and @var this vertices should be the same length,
              * otherwise this object is returned as was
              * @param fact the division factor per vertex
              */
            const P3DVertices &operator/= ( const PVector & fact);


            /** @brief devison operator
              * @returns copy of this list of vertices devided by @var fact. @var fact
              * and @var this vertices should be the same length,
              * otherwise this object is returned as was
              * @param fact the division factor per vertex
              */
            P3DVertices operator/ ( const PVector & fact)const;


            /** @brief multiplication operator
              * @returns copy of this list of vertices multiplied by @var fact. @var fact
              * and @var this vertices should be the same length,
              * otherwise this obejct is returned as was
              * @param fact the multiplication factor per vertex
              */
            P3DVertices operator* ( const PVector & fact)const;


            /** @brief normalizes all vector and returns this
              */
            const P3DVertices &normate();


            /** @brief copies thsi vector, and returns the normalized vectors
              */
            P3DVertices normalized();


            /** @brief devides all coordinated by the maximum length
              */
            const P3DVertices &relativeMax();


            /** @brief returns cross product of this vector and vector.
              * @param vector that is used in the cross product
              */
            P3DVertices cross(const P3DVector &vector)const;


            /** @brief returns cross product of this vector and vector.
              *	@param vector that is used in the cross product
              */
            P3DVertices cross(const P3DVertices &vectors)const;


            /** @brief returns dot product of both vectors.
              *	@param vector that is used in the dot product
              */
            PVector dot(const P3DVector &vector)const;


            /** @brief returns dot product of both vectors.
              *	@param ver that is used in the dot product
              */
            PVector dot(const P3DVertices &ver)const;


            /** @returns the projection point on the line, assuming all vertices from a line
              * @param p the point to be projected
              */
            P3DVector vertexOnLine(const P3DVector &p) const;


            /** @returns the projection point on the line, assuming all vertices from a line
               * @param p the point to be projected
               * @returs the normalized gradient in the found point
               */
            P3DVector vertexOnLine(const P3DVector &p, P3DVector & gradient) const;


            /** @returns the total length of all distances of each vertex to (0,0,0)
              */
            double totLength() const;


#ifndef QT_NO_DEBUG

            /** @brief Returns the vertex at index position @var index as a constant
              * reference. @var index must be a valid index position in the vecor
              * (i.e. no range check)
              * @param index postion
              */
            const P3DVector &at(int index) const;


            /** @brief Returns the last vertex as a constant reference.
              */
            const P3DVector &last() const;


            /** @brief Returns the first vertex as a constant reference.
              */
            const P3DVector &first() const;


            /** @brief Returns the vertex at index position @var index as a modifiable
              * reference. @var index must be a valid index position in the vecor
              * (i.e. no range check)
              * @param index postion
              */
            P3DVector &operator [] (int index);


            /** @brief set the vertex at @variable i to @variable pos
              */
            void setVertex(int i,const P3DVector & pos);

#else

            /** @brief Returns the vertex at index position @var index as a constant
              * reference. @var index must be a valid index position in the vecor
              * (i.e. no range check)
              * @param index postion
              */
            inline const P3DVector &at(int index) const { return mVertices.at(index); }


            /** @brief Returns the last vertex as a constant reference.
              */
            inline const P3DVector &last() const { return mVertices.last(); }


            /** @brief Returns the first vertex as a constant reference.
              */
            inline const P3DVector &first() const { return mVertices.first(); }


            /** @brief Returns the vertex at index position @var index as a modifiable
              * reference. @var index must be a valid index position in the vecor
              * (i.e. no range check)
              * @param index postion
              */
            inline P3DVector &operator [] (int index) { return mVertices[index]; }


            /** @brief set the vertex at @variable i to @variable pos
              */
            inline void setVertex(int i,const P3DVector & pos) {    mVertices[i] = pos;}

#endif


            /** @returns the mean of the vertices indicated by the QList of indices
              * @param iVertices the list of indeces for which the mean postion is computed
              */
            P3DVector meanOf(const QList <int> &iVertices) const;


            /** @returns a subset of the vertices indicated by the QList of indices
              * @param iVertices the list of indeces
              */
            P3DVertices getSub(const QList <int> &iVertices) const;


            /** @returns a subset of the vertices indicated by the QList of indices
              * @param iVertices the list of indeces
              */
            P3DVertices getSub(const PIVector &iVertices) const;


            /** @brief reorder the vertices starting from the firstIndex
              */
            void reorder(int firstIndex);


            /** @brief reorder the vertices according to the given indices
              */
            void reorder(const PIVector &iVertices);


            /** @returns the length of each vector in the list of vertices
              */
            PVector norm(void) const;


            /** @returns the maximum length of each vector in the list of vertices
              */
            double getMaxNorm(void) const;


            /** @returns the maximum length of each vector in the list of vertices
              */
            double getMeanNorm(void) const;


            /** @brief returns squared length of the vector.
              */
            PVector sqNorm( void ) const;


            /** @brief returns length between each next vertex.
              * This assumes that the poinst forme a continuous line
              */
            P3DVertices deltaVector( void ) const;


            /** @brief returns length between each next vertex .
              * This assumes that the poinst forme a continuous line
              */
            PVector deltaLength( void ) const;


            /** @brief returns cumullative length between each next vertex .
              * This assumes that the poinst forme a continuous line
              */
            PVector cumLength( void ) const;


            /** @brief rotate vertices.
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param angles    phi,theta,gamma
              * @param rotPnt    the point around which the geometry is rotated
              */
            void rotate( const P3DVector &angles, const P3DVector &rotPnt = P3DVector() );


            /** @brief rotate vertices.
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param phi       rotation around  ?? axes
              * @param theta     rotation around  ?? axes
              * @param gamma     rotation around  ?? axes
              * @param rotPnt    the point aroudn which the geometry is rotated
              */
            void rotate( double phi, double theta, double gamma, const P3DVector &rotPnt = P3DVector() );


            /** @brief rotate with the constucted rotation matrix based on the axis and angle
              * @param axis
              */
            void rotate( const P3DVector &axis, double angle, const P3DVector &rotPnt = P3DVector() );


            /** @brief rotate with the constucted rotation matrix based on the axis and angle
              * @param axis
              */
            P3DVertices rotated( const P3DVector &axis, double angle, const P3DVector &rotPnt = P3DVector() ) const;


            /** @brief copies the vertices and rotate these copied vertices
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param angles    phi,theta,gamma
              * @param rotPnt    the point around which the geometry is rotated
              */
            P3DVertices rotated( const P3DVector &angles, const P3DVector &rotPnt = P3DVector() ) const;


            /** @brief copies the vertices and rotate these copied vertices
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param phi       rotation around  ?? axes
              * @param theta     rotation around  ?? axes
              * @param gamma     rotation around  ?? axes
              * @param rotPnt    the point aroudn which the geometry is rotated
              */
            P3DVertices rotated( double phi, double theta, double gamma, const P3DVector &rotPnt = P3DVector() ) const;


            /** @brief TODO
              * @param rotation
              * @param center
              * @param shift
              * @param scale
              * @return
              */
            P3DVertices rotashsc( const P3DVector & rotation,
                                  const P3DVector & center,
                                  const P3DVector & shift,
                                  const P3DVector & scale ) const;

            /** @brief scale relative to the center
              * @param center the center of the vertices
              */
            P3DVertices scaleKeepCenter(const P3DVector &scale, const P3DVector & center) const;

            /** @brief scale relative to the center
              * @param center the center of the vertices
              */
            void scaledKeepCenter(const P3DVector &scale, const P3DVector & center);


            /** @brief save the vertices to an asci file with the given filename
              * @param filename
              */
            bool save(const QString &filename,bool asText = true)const;


            /** @brief TODO
              * @param filename
              */
            bool load(const QString &filename, bool asText = true);


            /** @brief clears all data
              */
            void clear(void);


            /** @brief alloactes size number of vertices. All vertices ccordinates
              * are set to zeros
              *
              * @param size the new size of the vertices
              */
            void resizeAndToZero( int size );


            /**
              * if i1 > i2 or out of range an empty list is returned
              * @param i1 lowest number from which is selected
              * @param i2 highsest number from which is selected
              * @returns a number of vertices ranging from (i2 - i1 + 1)
              */
            P3DVertices select( int i1, int i2 );


            /** @returns the index of the vertex with the given vertex position.
              * If not found -1 is returned
              * @param vertex the postion looked for in the list of vertices
              */
            inline int findVertex(const P3DVector & vertex) const { return mVertices.indexOf(vertex); }


            /** @returns true if the vertex is foudn in the list of vertices.
              * If not found false is returned
              * @param vertex the postion looked for in the list of vertices
              */
            inline bool contains(const P3DVector & vertex) const { return mVertices.contains(vertex); }


            /** @returns the mean of x,y and z coordinate of all vertices
              * If the members are not valid the mean coordinate is recomputed
              */
            P3DVector getMean( void ) const;


            /** @returns the sum of x,y and z coordinate of all vertices
              * If the members are not valid the mean coordinate is recomputed
              */
            P3DVector getSum( void ) const;


            /** @brief clears the content of the all vertices to 0 0 0(zero)
              */
            void toZeros( void );


            /** @brief Limit the precisoon of the vertices
              * @param precision the number of digits
              */
            void limitPrecision( int precision = 3 );


            /** @brief change the XY axis: y=-x and x = y
              * Created because the definition in MRI coordinates is diffrent from the one used at mbfys
              * Nijmgen, the Netherlands
              */
            void changeXYaxis( void );


            /** @brief change the XY axis: y=-x and x = y
              * Created because the definition in MRI coordinates is diffrent from the one used at mbfys
              * Nijmgen, the Netherlands
              */
            void changeYXaxis( void );


        private:

            /** @brief TODO
              */
            P3DVector verOnLine( const P3DVector &p, const P3DVector &v1, const P3DVector &v2 ) const;


            /** @brief The vertices
              */
            QVector<P3DVector> mVertices;
        };
    }
}
#endif // __P3DVERTICES_H
