/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PIVector.cpp 1011 2014-01-18 13:05:20Z peter $
 *
 */


#include <QTextStream>
#include <QFile>
#include <QtAlgorithms>
#include <QDebug>
#include <float.h>
#include "PIVector.h"
#include "pvector.h"


using namespace Peacs::Support;


const int FILE_PIVECTOR_VERSION_1           = 1;
const int FILE_PIVECTOR_VERSION_DEFAULT     = 1;

const char* FILE_PIVECTOR_NAME              = "PIVector";



PIVector::PIVector( void ) :
    PFile(),
    mVec()
{
}


PIVector::PIVector( int n, int val ) :
    PFile(),
    mVec( n, val )
{
//    if ( n==0)
//    {
//        qDebug("PIVector: size = 0");
//    }
}


PIVector::PIVector( int n ) :
    PFile(),
    mVec( n, 0 )
{
}


PIVector::PIVector( const PIVector &v ) :
    PFile( v ),
    mVec( v.mVec )
{
}


PIVector::PIVector( const QVector<int> v ) :
    PFile(),
    mVec( v )
{
}

PIVector::PIVector( const QList<int> v ) :
    PFile(),
    mVec( v.toVector() )
{
}


PIVector::PIVector( const PVector & vec ) :
    PFile(),
    mVec( vec.N() )
{
    for ( int i = 0; i < vec.N(); i++ )
    {
        mVec[i] = (int)vec[i];
    }
}


PIVector::~PIVector( void )
{
}


int PIVector::getDefaultVersion( void ) const
{
    return FILE_PIVECTOR_VERSION_DEFAULT;
}


const char *PIVector::getClassName( void ) const
{
    return FILE_PIVECTOR_NAME;
}


bool PIVector::operator== ( const PIVector &vec ) const
{
    bool equal = false;

    if ( mVec.size() == vec.mVec.size() )
    {
        int i = 0;

        equal = true;   // Assume equal

        while ( i < mVec.size()  &&
                equal )
        {
            equal = (mVec[i] == vec.mVec[i]);

            i++;
        }
    }

    return equal;
}


bool PIVector::operator!= ( const PIVector &vec ) const
{
    bool unequal = false;

    if ( !mVec.size() != vec.mVec.size() )
    {
        int i = 0;

        while ( (i < mVec.size()) && !unequal )
        {
            unequal = (mVec[i] != vec.mVec[i]);

            i++;
        }
    }

    return unequal;
}


PIVector PIVector::operator+ ( const int f ) const
{
    PIVector v( *this );

    for ( int i = 0; i < v.mVec.size(); i++ )
    {
        v.mVec[i] += f;
    }

    return v;
}


PIVector PIVector::operator- ( const int f ) const
{
    PIVector v( *this );

    for ( int i = 0; i < v.mVec.size(); i++ )
    {
        v.mVec[i] -= f;
    }

    return v;
}


PIVector PIVector::operator* ( const int f ) const
{
    PIVector v( *this );

    for ( int i = 0 ; i < v.mVec.size(); i++ )
    {
        v.mVec[i] *= f;
    }

    return v;
}


PIVector PIVector::operator/ ( const int f ) const
{
    PIVector v( *this );

    for ( int i = 0; i < v.mVec.size(); i++ )
    {
        v.mVec[i] /= f;
    }

    return v;
}


const PIVector &PIVector::operator+= ( const int f )
{
    for ( int i = 0; i < mVec.size(); i++ )
    {
        mVec[i] += f;
    }

    return *this;
}


const PIVector &PIVector::operator-= ( const int f )
{
    for ( int i = 0; i < mVec.size(); i++ )
    {
        mVec[i] -= f;
    }

    return *this;
}


const PIVector &PIVector::operator*= ( const int f )
{
    for ( int i = 0; i < mVec.size(); i++ )
    {
        mVec[i] *= f;
    }

    return *this;
}


const PIVector &PIVector::operator/= ( int f )
{
    for ( int i = 0; i < mVec.size(); i++ )
    {
        mVec[i] /= f;
    }

    return *this;
}


const PIVector &PIVector::operator= ( int f )
{
    // Don't call base class asignment operator. Version info must always be
    // the last version
    mVec.fill( f );

    return *this;
}


PIVector PIVector::operator+ ( const PIVector &v2 ) const
{
    PIVector v( *this );

    if ( !v2.isEmpty() )
    {
        if ( !isEmpty() )
        {
            if ( v2.mVec.size() != v.mVec.size() )
            {
                qDebug("PIVector operator +: vectors are not the same size");
            }
            else
            {
                for ( int i = 0; i < v.mVec.size(); i++ )
                {
                    v.mVec[i] += v2.mVec[i];
                }
            }
        }
        else
        {
            v = v2;
        }
    }

    return v;
}


PIVector PIVector::operator- ( const PIVector &v2 ) const
{
    PIVector v( *this );

    if ( !v2.isEmpty() )
    {
        if ( !v.isEmpty()  )
        {
            if ( v2.mVec.size() != v.mVec.size() )
            {
                qDebug("PIVector operator -: vectors are not the same size");
            }
            else
            {
                for ( int i = 0; i < v.mVec.size(); i++ )
                {
                    v.mVec[i] -= v2.mVec[i];
                }
            }
        }
        else
        {
            v -= v2;
        }
    }

    return v;
}


PIVector PIVector::operator* ( const PIVector &v2 ) const
{
    PIVector v( *this );

    if ( v2.isEmpty() || isEmpty() )
    {
        v.resize( qMax( v.mVec.size(), v2.mVec.size() ) );

        for ( int i = 0; i < v.mVec.size(); i++ )
        {
            v.mVec[i] = 0;
        }
    }
    else
    {
        if ( v2.mVec.size() != v.mVec.size() )
        {
            qDebug("PIVector operator *: vectors are not the same size");
        }
        else
        {
            for ( int i = 0; i < v.mVec.size(); i++ )
            {
                v.mVec[i] *= v2.mVec[i];
            }
        }
    }

    return v;
}


PIVector PIVector::operator/ ( const PIVector &v2 ) const
{
    PIVector v( *this );

    if ( v2.isEmpty() )
    {
        qDebug("PIVector operator /: illegal to devide by 0  (v2 is empty)");
    }
    else if ( v.isEmpty() )
    {
        v.resize( v2.mVec.size() );

        for ( int i = 0; i < v.mVec.size(); i++ )
        {
            v.mVec[i] = 0;
        }
    }
    else if ( v2.mVec.size() != v.mVec.size() )
    {
        qDebug("PIVector operator /: vectors are not the same size");
    }
    else
    {
        for ( int i = 0; i < mVec.size(); i++ )
        {
            v.mVec[i] /= v2.mVec[i];
        }
    }

    return v;
}


const PIVector &PIVector::operator+= ( const PIVector &v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            *this = v;
        }
        else
        {
            if ( v.mVec.size() != mVec.size() )
            {
                qDebug("PIVector operator +=: vectors are not the same size");
            }
            else
            {
                for ( int i = 0; i < mVec.size(); i++ )
                {
                    mVec[i] += v.mVec[i];
                }
            }
        }
    }

    return *this;
}


const PIVector &PIVector::operator-= ( const PIVector &v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            resize( v.mVec.size() );

            for ( int i = 0; i < mVec.size(); i++ )
            {
                mVec[i] = -v.mVec[i];
            }
        }
        else if ( v.mVec.size() != mVec.size() )
        {
            qDebug("PIVector operator -=: vectors are not the same size");
        }
        else
        {
            for ( int i = 0; i < mVec.size(); i++ )
            {
                mVec[i] -= v.mVec[i];
            }
        }
    }

    return *this;
}


const PIVector &PIVector::operator*= ( const PIVector &v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            *this = v;

            qDebug("PIVector operator *=: this vector was empty");
        }
        else
        {
            if ( v.mVec.size() != mVec.size() )
            {
                qDebug("PIVector operator *=: vectors are not the same size");
            }
            else
            {
                for ( int i = 0; i < mVec.size(); i++ )
                {
                    mVec[i] *= v.mVec[i];
                }
            }
        }
    }

    return *this;
}


const PIVector &PIVector::operator/= ( const PIVector &v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            resize( v.mVec.size() );

            for ( int i = 0; i < mVec.size(); i++ )
            {
                mVec[i] = 0;
            }
        }
        else
        {
            if ( v.mVec.size() != mVec.size() )
            {
                qDebug("PIVector operator /=: vectors are not the same size");
            }
            else
            {
                for ( int i = 0; i < mVec.size(); i++ )
                {
                    mVec[i] /= v.mVec[i];
                }
            }
        }
    }
    else
    {
        qDebug("PIVector operator /=: illegal to devide by 0 (v is empty)");
    }

    return *this;
}


const PIVector &PIVector::operator*= ( const PMatrix &m )
{
    if ( m.Nr() == mVec.size() )
    {
        int         i;
        int         ir;
        int         k;
        double      val;
        PIVector    vec( *this );

        resize( m.Nc() );

        for ( i = 0; i < m.Nc(); i++ )
        {
            val = 0;

            for ( k = 0, ir = i; k < m.Nr(); k++, ir += m.Nc()  )
            {
                val += vec[k] * m[ir];
            }

            mVec[i] = val;
        }
    }
    else
    {
        qDebug("PIVector operator *=: vectors are not the same size");
    }
    return *this;

}


PIVector PIVector::operator* ( const PMatrix &m ) const
{
    PIVector v( m.Nc(), 0 );

    if ( m.Nr() == mVec.size() )
    {
        int i;
        int k;
        int ir;
        int val;

        for ( i = 0; i < m.Nc(); i++ )
        {
            val = 0;

            for ( k = 0, ir = i; k < m.Nr(); k++, ir += m.Nc()  )
            {
                val += mVec[k] * m[ir];
            }

            v.mVec[i] = val;
        }
    }
    else
    {
        qDebug("PIVector operator*: vector and rows matrix are not the same size");
    }

    return v;
}


const PIVector &PIVector::operator= ( const PIVector &v )
{
    if ( this != &v )
    {
        mVec = v.mVec;
    }

    return *this;
}


PIVector PIVector::operator!= ( int val ) const
{
    PIVector ret( mVec.size(), 0 );

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( mVec[i] != val )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PIVector::operator== ( int val ) const
{
    PIVector ret( mVec.size(), 0 );

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( mVec[i] == val )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PIVector::operator> ( int f ) const
{
    PIVector ret( mVec.size(), 0 );

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( mVec[i] > f )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PIVector::operator< ( int f ) const
{
    PIVector ret( mVec.size(), 0 );

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( mVec[i] < f )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PIVector::operator>= ( int f ) const
{
    PIVector ret( mVec.size(), 0 );

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( mVec[i] >= f )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PIVector::operator<= ( int f ) const
{
    PIVector ret( mVec.size(), 0 );

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( mVec[i] <= f )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


double PIVector::sum( void ) const
{
    double ret = 0;

    for ( int i = 0; i < mVec.size(); i++ )
    {
        ret += mVec[i];
    }

    return ret;
}


const PIVector &PIVector::mul( const PMatrix &m, const PIVector &v2 )
{
    if ( m.Nc() == v2.mVec.size() )
    {
        int i;
        int k;
        int val;

        mVec.resize( m.Nr() );

        if ( !mVec.isEmpty() )
        {
            for ( i = 0; i < m.Nr(); i++ )
            {
                val = 0;

                for ( k = 0; k < m.Nr(); k++ )
                {
                    val += m.value( i, k ) * v2.mVec[k];
                }

                mVec[i] = val;
            }
        }
    }
    else
    {
        qDebug("PIVector mul(matrix,vector): not the same size");
    }

    return *this;
}


const PIVector &PIVector::mul( const PIVector &v2, const PMatrix &m )
{
    if ( m.Nr() == v2.mVec.size() )
    {
        int i;
        int k;
        int val;

        resize( m.Nc() );

        for ( i = 0; i < m.Nc(); i++ )
        {
            val = 0;

            for ( k = 0; k < m.Nr(); k++ )
            {
                val += m.value( k, i ) * v2.mVec[k];
            }

            mVec[i] = val;
        }
    }
    else
    {
        qDebug("PIVector mul(vector,matrix): not the same size");
    }

    return *this;
}


void PIVector::normalize( void )
{
    if ( !mVec.isEmpty() )
    {
        int i;
        int min = mVec[0];
        int max = mVec[0];

        for ( i = 1; i < mVec.size(); i++ )
        {
            if ( mVec[i] < min)
            {
                min = mVec[i];
            }

            if ( max < mVec[i] )
            {
                max = mVec[i];
            }
        }

        if ( max - min != 0 )
        {
            for ( i = 0; i < mVec.size(); i++ )
            {
                mVec[i] = (mVec[i] - min) / (max - min);
            }
        }
    }
}


PIVector PIVector::normalized( void ) const
{
    PIVector v( *this );

    if ( !isEmpty() )
    {
        int max = getMax();
        int min = getMin();

        if ( max - min != 0 )
        {
            for ( int i = 0; i < v.mVec.size(); i++ )
            {
                v.mVec[i] = (v.mVec[i] - min) / (max - min);
            }
        }
    }

    return v;
}


int PIVector::getMin( void ) const
{
    int min = 0;

    if ( !isEmpty() )
    {
        int i;

        min = mVec[0];

        for ( i = 1; i < mVec.size(); i++ )
        {
            if ( mVec[i] < min )
            {
                min = mVec[i];
            }
        }
    }

    return min;
}


int PIVector::getMax( void ) const
{
    int max = 0;

    if ( !isEmpty() )
    {
        max = mVec[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( mVec[i] > max )
            {
                max = mVec[i];
            }
        }
    }

    return max;
}


int PIVector::getMinI( void ) const
{
    int     minI    = 0;
    double  min     = 0;

    if ( !isEmpty() )
    {
        min = mVec[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( mVec[i] < min )
            {
                minI    = i;
                min     = mVec[i];
            }
        }
    }

    return minI;
}


int PIVector::getMaxI( void ) const
{
    int max;
    int maxI = 0;

    if ( !isEmpty() )
    {
        max = mVec[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( max > mVec[i] )
            {
                max     = mVec[i];
                maxI    = i;
            }
        }
    }

    return maxI;
}


int PIVector::getMean( void ) const
{
    double mean = 0;

    if ( !isEmpty() )
    {
        for ( int i = 0; i < mVec.size(); i++ )
        {
            mean += mVec[i];
        }

        mean /= (double)mVec.size();
    }

    return mean;
}


int PIVector::getStd( void ) const
{
    int stdev = 0;

    if ( !isEmpty() )
    {
        int     i;
        double  mean    = 0;
        double  sqSum   = 0;
        int     tmp;

        for ( i = 0; i < mVec.size(); i++ )
        {
            mean += mVec[i];
        }

        mean /= (double)mVec.size();

        for ( i = 0; i < mVec.size(); i++ )
        {
            tmp = (mVec[i] - mean);
            sqSum += (tmp * tmp);
        }

        stdev = (int)sqrt( sqSum / ((double)mVec.size()) );
    }

    return stdev;
}


int PIVector::getRange( void ) const
{
    int min    = 0;
    int max    = 0;

    if ( !isEmpty() )
    {
        int i;

        min    = mVec[0];
        max    = mVec[0];

        for ( i = 1; i < mVec.size(); i++ )
        {
            min = qMin( min, mVec[i] );
            max = qMax( max, mVec[i] );
        }
    }

    return max - min;
}


bool PIVector::save( QString filename, unsigned int precision ) const
{
    int     ir;
    bool    ok = false;
    QFile   f( filename );

    if ( f.open( QIODevice::WriteOnly ) )
    {
        QTextStream ts( &f );

        ts.setRealNumberPrecision( precision );

        ts << mVec.size()  << " 1  \n";

        for ( ir = 0; ir < mVec.size(); ir++ )
        {
            ts << mVec[ir] << "\n";
        }

        ts << "\n";

        f.close();

        ok = true;
    }

    return ok;
}


bool PIVector::load( QString filename )
{
    bool    ok = false;
    int     i;
    int     n;
    int     nc;

    if ( !filename.isEmpty() )
    {
        QFile f( filename );

        if ( f.open( QIODevice::ReadOnly ) )
        {
            QTextStream ts( &f );

            ts >> n >> nc;

            if ( ts.status() == QTextStream::Ok &&
                 n > 0 &&
                 nc == 1 )
            {
                resize( n );

                if ( !isEmpty() )
                {
                    ok = true;

                    for ( i = 0; i < n; i++ )
                    {
                        ts >> mVec[i];

                        if ( ts.status() != QTextStream::Ok )
                        {
                            ok = false;

                            break;
                        }
                    }
                }
            }
        }

        if ( !ok )
        {
            clear();
        }

        f.close();
    }

    return ok;
}


bool PIVector::read( QDataStream &dataStream )
{
    bool    ok      = false;
    int     version = readHead( dataStream, PIVector::getClassName() );

    if ( version != READ_HEADER_FAILED )
    {
        qint32 n;

        switch ( version )
        {
        case FILE_PIVECTOR_VERSION_1:
            clear();

            dataStream >> n;

            if ( isOk( dataStream ) )
            {
                if ( n > 0 )
                {
                    resize( (int)n );

                    if ( !isEmpty() )
                    {
                        ok = true;

                        for ( int i = 0; i < mVec.size() && ok; i++ )
                        {
                            dataStream >> mVec[i];

                            ok = isOk( dataStream );
                        }
                    }
                }
                else
                {
                    ok = true; // Nothing more to read
                }
            }

            break;

        default:
            break;
        }
    }

    return ok;
}


bool PIVector::write( QDataStream &dataStream ) const
{
    bool ok = writeHead( dataStream, PIVector::getClassName(), PIVector::getDefaultVersion() );

    if ( ok )
    {
        switch ( PIVector::getDefaultVersion() )
        {
        case FILE_PIVECTOR_VERSION_1:
            {
                qint32 n = (qint32)mVec.size();

                dataStream << n;

                if ( isOk( ok, dataStream ) && !isEmpty() )
                {
                    for ( int i = 0; i < mVec.size() && isOk( dataStream ); i++ )
                    {
                        dataStream << mVec[i];
                    }
                }
                break;
            }

        default:
            ok = false;
            break;
        }
    }

    return isOk( ok, dataStream );
}


void PIVector::reorder( int firstIndex )
{
    PIVector    vec( mVec.size() );
    int         k = 0;
    int         i;

    for ( i = firstIndex; i < vec.N(); i++ )
    {
        vec[k] = mVec[i];

        k++;
    }

    for ( i = 0; i < firstIndex; i++ )
    {
        vec[k] = mVec[i];

        k++;
    }

    mVec = vec.mVec;
}


void PIVector::reorder( const PIVector &iVertices )
{
    PIVector vec( mVec );

    if ( iVertices.N() == mVec.size() &&
         iVertices.getMin() >= 0  &&
         iVertices.getMax() < mVec.size() )
    {
        for ( int i = 0; i < mVec.size(); i++ )
        {
            vec[i] = mVec[ iVertices[i] ];
        }
    }
    else
    {
        qDebug() << "vertices not reordred";
    }

    mVec = vec.mVec;
}


void PIVector::sort( bool asscending )
{
    if ( asscending )
    {
        qSort( mVec.begin(), mVec.end() );
    }
    else
    {
        qSort( mVec.begin(), mVec.end(), qGreater<int>() );
    }
}


#ifndef QT_NO_DEBUG

int &PIVector::operator[] ( int i )
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure
    Q_ASSERT( i >= 0 && i < mVec.size() );  // Test just to be sure

    return mVec[i];
}


const int &PIVector::operator[] ( int i ) const
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure
    Q_ASSERT( i >= 0 && i < mVec.size() );  // Test just to be sure

    return mVec[i];
}


const int &PIVector::at( int i ) const
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure
    Q_ASSERT( i >= 0 && i < mVec.size() );  // Test just to be sure

    return mVec[i];
}


int PIVector::first( void ) const
{
    Q_ASSERT( !mVec.isEmpty() );  // Test just to be sure

    return mVec.first();
}


int PIVector::last( void ) const
{
    Q_ASSERT( !mVec.isEmpty() );  // Test just to be sure

    return mVec.last();
}

#endif
