/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PIVector.h 1196 2014-04-15 10:20:27Z eelco $
 *
 */

#ifndef __PIVECTOR_H
#define __PIVECTOR_H

#include <QVector>

#include "pfile.h"
#include "pmatrix.h"

#include "PeacsSupport_global.h"

#include "math.h"


namespace Peacs
{
    namespace Support
    {
        //class PMatrix;

        class PVector;

        /** @class PIVector PIVector.h "support/PIVector.h"
          * @brief This class hold a vector (1 demensional array)
          * and its functions
          */
        class PEACSSUPPORTSHARED_EXPORT PIVector : public PFile
        {

        public:

            /** @brief Default constructor.
              * Create a empty vector
              */
            PIVector( void );


            /** @brief Copy constructor
              * @see operator=(const PIVector &vector)
              */
            PIVector( const PIVector &vector );


            /** @brief Constructor
              *  Initializes the vector (size n) with val)
              *  @param val the value assigned to all elements
              *  @param n  size of the vector
              */
            PIVector( int n, int val );


            /** @brief Constructor
              * No initialization is done
              * @param n size of the new vector
              */
            PIVector( int n );


            /** @brief Initializer constructor
              * Initializes this class with the presented information where the
              * values are casted to integers
              *
              * @param vec the vector to intialize with
              */
            PIVector( const PVector & vec );


            /** @brief Initializer constructor
              * Initializes this class with the presented vector
              *
              * @param v the vector to intialize with
              */
            PIVector( const QVector<int> v );


            /** @brief Initializer constructor
              * Initializes this class with the presented vector
              *
              * @param v the vector to intialize with
              */
            PIVector( const QList<int> v );



            /** @brief destructor.
              */
            virtual ~PIVector( void );


            /** @brief Returns the minimum value of all elements in the vector
              * This one is slower because it computes the minimum value every time
              */
            int getMin( void ) const;


            /** @brief The index posistion of the minimum in the vector
              */
            int getMinI( void ) const;


            /** @brief Returns the maximum value of all elements in the vector
              * This one is slower because it computes the maximum value every time
              * @see max()
              */
            int getMax( void ) const;


            /** @brief The index posistion of the maximum in the vector
              */
            int getMaxI( void ) const;


            /** @brief The mean of the vector
              */
            int getMean( void ) const;


            /** @brief The standard deviation of the vector
              */
            int getStd( void ) const;


            /** @brief The difference between max and min of the vector
              */
            int getRange( void ) const;


            /** @brief Returns the number of elements of the vector.
              */
            inline int N( void ) const { return mVec.size(); }


            /** @brief Returns the number of elements of the vector.
              */
            inline int count( void ) const { return mVec.size(); }


            /** @brief Returns the number of elements of the vector.
              */
            inline int size( void ) const { return mVec.size(); }


            /** @brief Returns true if the vector is empty.
              */
            inline bool isEmpty( void ) const { return mVec.isEmpty(); }


            /** @returns true if the vector contains the given val
              * @param val the value looked for
              * @return true if found
              */
            inline bool contains( int val ) const { return mVec.contains( val ); }


            /** @brief Gets the first index position where 'val' is located
              * Returns the index position of the first occurrence of value in the vector,
              * searching forward from position 0.
              *
              * @param val the value to search for
              * @return returns -1 if no item matched, otherwise its position in the vector
              */
            inline int indexOf (int val ) const { return mVec.indexOf( val ); }


            /** @brief Inserts value at the beginning of the vector
              * @param value the value to prepend
              */
            inline void prepend( int value ) { mVec.prepend( value ); }


            /** @brief Inserts value at the end of the vector
              * @param value the value to append
              */
            inline void append( int value ) { mVec.append( value ); }


            /** @brief Inserts value at index position 'before' of the vector
              * Inserts value at index position 'before' in the vector. If 'before'
              * is 0, the value is prepended to the vector. If 'before' is size(),
              * the value is appended to the vector
              */
            inline void insert( int before, int value )  { mVec.insert( before, value ); }


            /** @brief Appends the vector vec at the end of the vector.
              */
            inline void append( const PIVector & vec ) { mVec += vec.mVec; }


            /** @brief sets the size of the vector to the size mn
              * If size is greater than the current size, elements are added to the end; the new elements are initialized
              * with a default-constructed value. If size is less than the current size, elements are removed from the end.
              *
              * @param mn The new size of the vector
              */
            inline void resize( int mn ) { mVec.resize( mn ); }


            /** @brief Removes the value at index.
              * @param index the index of the element to remove
              */
            inline void remove( int index ) { mVec.remove( index ); }


            /** @brief Removes count elements from within of the vector, starting at index position pos.
              * @param pos The starting position in the vector from where to remove the elements
              * @param count The number of elements to remove
              */
            inline void remove( int pos, int count ) { mVec.remove( pos, count ); }


            /** @brief Clear the vector (empty)
              */
            inline void clear( void ) { mVec.clear(); }


            /** @brief Set all values in the vector to zero
              */
            inline void toZero( void ) { mVec.fill( 0 ); }


            /** @brief returns the const data pointer
              */
            inline const int *data( void ) const { return mVec.data(); }


            /** @brief returns the data pointer
              */
            inline int *data( void ) { return mVec.data(); }

            /** @brief Returns a QList object with the data contained in this QVector
              */
            inline QList <int> toList() const { return mVec.toList(); }


            /** @brief comparisson operator.
              * compares all elements from vector (vec) with this object
              *
              * @param vec the vetor being compared to
              *
              * @returns true if the data is the same
              */
            bool operator== ( const PIVector &vec ) const;


            /** @brief comparisson operator.
              * compares all elements from vector (vec) with this object
              *
              * @param vec the vetor being compared to
              * @returns true if the data is not the same
              */
            bool operator!= ( const PIVector &vec ) const;


            /** @brief reorder the vertices according to the given indices
              */
            void reorder( const PIVector &iVertices );


            /** @brief reorder the vertices starting from the firstIndex
              */
            void reorder( int firstIndex );


            /** @brief Sorts the items in the vector
              * @param asscending
              */
            void sort( bool asscending = true );


#ifndef QT_NO_DEBUG

            /** @brief Returns the item at index position i as a modifiable
              * reference. i must be a valid index position in the vecor
              * (i.e. no range check)
              * @param i index postion
              * @see int &operator[](int i) const;
              */
            int &operator[] ( int i );


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see int &operator[](int i);
              */
            const int &operator[] ( int i ) const;


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see int &operator[](int i);
              */
            const int &at( int i ) const;


            /** @brief This is an overloaded member function, provided for
              * convenience.
              */
            int first( void ) const;


            /** @brief This is an overloaded member function, provided for
              * convenience.
              */
            int last( void ) const;

#else

            /** @brief Returns the item at index position i as a modifiable
              * reference. i must be a valid index position in the vecor
              * (i.e. no range check)
              * @param i index postion
              * @see int &operator[](int i) const;
              */
            inline int &operator[] ( int i ) { return mVec[i]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see int &operator[](int i);
              */
            inline const int &operator[] ( int i ) const { return mVec[i]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see int &operator[](int i);
              */
            inline const int &at( int i ) const { return mVec[i]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              */
            inline int first( void ) const { return mVec.first(); }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              */
            inline int last( void ) const { return mVec.last(); }

#endif


            /** @brief Copies this vector and adds a constant (f) to all
              * elements of the vector.
              * @param f factor added
              * @see operator+=(const int f)
              */
            PIVector operator+ ( int f ) const;


            /** @brief Copies this vector and subracts a constant (f) from all
              * elements of this new vector.
              * @param f factor subtracted
              * @see operator-=(const int f)
              */
            PIVector operator- ( int f ) const;


            /** @brief Copies this vector and multiplies all elements with a
              * constant (f) of this new vector.
              * @param f multiplying factor
              * @see operator*=(const int f)
              */
            PIVector operator* ( int f ) const;


            /** @brief Copies this vector and divides all elements by a
              * constant (f) of this new vector.
              * @param f deviding factor
              * @see operator/=(const int f)
              */
            PIVector operator/ ( int f ) const;


            /** @brief Subtracts to all elements of the vector a constant (f)
              * and returns the vector.
              * @param f factor subtracted
              * @see operator+(const int f)
              */
            const PIVector& operator-= ( int f );


            /** @brief  Adds to all elements of the vector a constant (f) and
              * retruns the vector.
              * @param f factor added
              * @see operator+(const int f)
              */
            const PIVector& operator+= ( int f );


            /** @brief  Multiplies all elements by a constant (f) of the
              * vector and returns the vector.
              * @param f multiplying factor
              * @see operator*(const int f)
              */
            const PIVector& operator*= ( int f );


            /** @brief  Divides all elements by a constant (f) of the vector
              * and returns the vector.
              * @param f deviding factor
              * @see operator/(const int f)
              */
            const PIVector& operator/= ( int f );


            /** @brief Copy operator.
              * All elements of the vector are set to value f
              * @param f value of all elements
              */
            const PIVector& operator= ( int f );


            /** @brief Copies vector, adds element wise both vectors and returns the new vector.
              * if both vectors differ in size an exception is raiseds
              * @see operator+=(const PIVector &v)
              */
            PIVector operator+ ( const PIVector &v ) const;


            /** @brief Copies vector, subtracts element wise both vectors and
              * returns the new vector.
              * \n if both vectors differ in size an exception is raised
              * @param v  input vector
              * @see operator-=(const PIVector &v)
              */
            PIVector operator- ( const PIVector &v ) const;


            /** @brief Copies vector, multiplies element wise both vectors and
              * returns the new vector.
              * if both vectors differ in size an exception is raised
              * @param v  input vector
              * @see operator*=(const PIVector &v)
              */
            PIVector operator* ( const PIVector &v ) const;


            /** @brief Copies vector, divides element wise both vectors and
              * returns the new vector.
              * if both vectors differ in size an exception is raised
              * @param v  input vector
              * @see operator/=(const PIVector &v)
              */
            PIVector operator/ ( const PIVector &v ) const;


            /** @brief Adds element wise both vectors and returns this vector.
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator+(const PIVector &v)
              */
            const PIVector& operator+= ( const PIVector &v );


            /** @brief Subtracts element wise both vectors and returns this
              * vector.
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator-(const PIVector &v)
              */
            const PIVector& operator-= ( const PIVector &v );


            /** @brief Multiplies element wise both vectors and returns the
              * vector.
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator*(const PIVector &v)
              */
            const PIVector& operator*= ( const PIVector &v );


            /** @brief Divides element wise both vectors and returns the vector.
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator/(const PIVector &v)
              */
            const PIVector& operator/= ( const PIVector &v );


            /** @brief Copies the vector, multiplies the vector with a vector:
              * vec=vec*M, and returns the vector
              * if the number of rows is unequal to the length of the vector
              * an exception is raised
              * @returns this object
              * @param m input vector
              * @see operator*=(const PIVector &m)
              * @see mul(const PIVector &v, const PIVector &m)
              * @see mul(const PIVector &m, const PIVector &v)
              */
            PIVector operator* ( const PMatrix &m ) const;


            /** @brief Multiplies the vector with a vector: vec=vec*M
              * if the number of rows is unequalto the length of the evector
              * an exception is raised
              * @returns this object
              * @param m  input vector
              * @see operator*(const PIVector &m)
              * @see mul(const PIVector &v, const PIVector &m)
              * @see mul(const PIVector &m, const PIVector &v)
              */
            const PIVector &operator*= ( const PMatrix &m );


            /** @brief copy operator.
              * copy all elements from vector (vec)
              * @returns this object
              * @param vec the vetor being copied
              */
            const PIVector& operator= ( const PIVector &vec );


            /** @brief comparison operator. If values if == val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator== ( int val ) const;


            /** @brief comparison operator. If values if != val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator!= ( int val ) const;


            /** @brief comparison operator. If values if >= val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator>= ( int val ) const;


            /** @brief comparison operator. If values <= val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator<= ( int val ) const;


            /** @brief comparison operator. If values if > val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator> ( int val ) const;


            /** @brief comparison operator. If values , val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator< ( int val ) const;


            /** @returns sum of all items in the vector
              */
            double sum( void ) const;


            /** @brief Multiplies the matrix with a vector: vec=M*vec and
              * copies the resulting vector in this object.
              * if the number of rows is unequal to the length of the vector
              * an exception is raised
              * @param v input vector v
              * @param m input vector m
              * @returns this object
              * @see operator*(const PIVector &m)
              * @see mul(const PIVector &v, const PIVector &m)
              * @see mul(const PIVector &m, const PIVector &v)
              */
            const PIVector &mul( const PMatrix &m, const PIVector &v );


            /** @brief Multiplies the vector (v) with a vector (m): vec=vec*M
              * if the number of rows is unequalto the length of the vector
              * an exception is raised
              * @returns this object
              * @param v input vector v
              * @param m input vector m
              * @see operator*(const PIVector &m)
              * @see mul(const PIVector &v, const PIVector &m)
              * @see mul(const PIVector &m, const PIVector &v)
              */
            const PIVector &mul( const PIVector &v, const PMatrix &m );


            /** @brief Normalize the vector
              */
            void normalize( void );


            /** @brief return a normalized copy of this vector
              */
            PIVector normalized( void ) const;


            /** @brief Reads vector data from file.
              *	@see virtual bool PFile::read(QDatastream &ds);
              */
            bool read( QDataStream &ds );


            /** @brief Write vector data to file.
              * @see virtual bool read(QDatastream &ds) = 0;
              */
            bool write( QDataStream &ds ) const;


            /** @brief load the vector from a file.
              * @see save(QString filename, unsigned int precision)
              */
            bool load( QString filename );


            /** @brief Save the vector to a text file with given precision.
              * @param filename the name of the file to save
              * @param precision the precision to use
              * @see load(QString filename)
              */
            bool save( QString filename, unsigned int precision = 8 ) const;


        protected:

            /** @brief Returns the class name
              * @see PFile
              */
            const char *getClassName( void ) const;


            /** @brief Implementation of a pure virtual function (PFile)that
              *  requests the default class data version to write to the case
              * file
              */
            int getDefaultVersion( void ) const;


        private:

            /** @brief The Vector
              */
            QVector <int> mVec;
        };
    }
}

#endif //PIVector_H
