/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2013 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id$
 *
 */


#include "PSettings.h"

#include <QCoreApplication>
#include <QFileInfo>
#include <QDir>

#ifndef QT_NO_DEBUG
    #include <QDebug>
#endif


using namespace Peacs::Support;


// Initialize static members
QString     PSettings::mFilename;
QSettings * PSettings::mSettingObjPtr   = NULL;
int         PSettings::mRefCount        = 0;



PSettings::PSettings( const QString & fileExt, QObject * parent ) :
    mMutex( QMutex::Recursive ),
    mUseInitSetting( false ),
    mIniFilename()
{
    mMutex.lock();

    mRefCount++;

    if ( mSettingObjPtr == NULL )
    {
        QStringList args( QCoreApplication::arguments() );

        if ( args.size() > 0 )
        {
            QFileInfo fileInfo;

            for ( int i = 0; i < args.count(); i++ )
            {
                QByteArray argL1 = args.at( i ).toLatin1();

                if ( (argL1 == "-ini") || (argL1 == "/ini") )
                {
                    mUseInitSetting = true;
                }
                else if ( (argL1 == "-inifile") || (argL1 == "/inifile") )
                {
                    // There should be an extra option with the filename after this one
                    if ( args.count() > (i + 1) )
                    {
                        i++;    // Go to the next part of the option

                        fileInfo.setFile( args.at( i ).toLatin1() );

                        if ( fileInfo.exists() )
                        {
                            mUseInitSetting = true;
                            mIniFilename    = fileInfo.filePath();
                        }
#ifndef QT_NO_DEBUG
                        else
                        {
                            qDebug() << "The ini file specified at the command prompt does not exist:" << fileInfo.filePath() <<
                                        "(" << __FILE__ << __LINE__ << ")";
                        }
#endif
                    }
#ifndef QT_NO_DEBUG
                    else
                    {
                        qWarning() << "No ini file specified at the command prompt:" << __FILE__ << __LINE__;
                    }
#endif
                }
                else
                {
                    // There could be a filename specified at the command line. It must exist
                    fileInfo.setFile( argL1 );

                    if ( fileInfo.exists() )
                    {
                        if ( fileExt.isEmpty() )
                        {
                            mFilename = fileInfo.filePath();
                        }
                        else if ( fileInfo.suffix().contains( fileExt, Qt::CaseInsensitive) )
                        {
                            mFilename = fileInfo.filePath();
                        }
                    }
                }
            }
        }

        if ( mUseInitSetting && mIniFilename.isEmpty() )
        {
            // First look in the application directory
#ifdef Q_WS_MACX
            QFileInfo inifile( QCoreApplication::applicationDirPath() + "/../../../" + QCoreApplication::applicationName() + ".ini" );
#else
            QFileInfo inifile( QCoreApplication::applicationDirPath() + "/" + QCoreApplication::applicationName() + ".ini" );
#endif

            if ( inifile.exists() )
            {
                mIniFilename = inifile.filePath();
            }
            else
            {
                // Now use the one in the users home directory
#ifndef QT_NO_DEBUG
                inifile.setFile( QDir::homePath() +  + "/.CircAdapt/" + QCoreApplication::applicationName() + ".ini" );

                if ( !inifile.exists() )
                {
                    qDebug() << "The ini option was specified at the command prompt but the file does not exist (yet). The file will be created in the $HOME/.CircAdapt directory:" << __FILE__ << __LINE__;
                }
#endif
                mIniFilename = QDir::toNativeSeparators( QDir::homePath() +  + "/.CircAdapt/" + QCoreApplication::applicationName() + ".ini" );
            }
        }

        if ( mUseInitSetting && !mIniFilename.isEmpty() )
        {
            mSettingObjPtr = new QSettings( mIniFilename, QSettings::IniFormat, parent );
        }
        else
        {
            mSettingObjPtr = new QSettings( parent );
        }
    }

    mMutex.unlock();
}


PSettings::~PSettings( void )
{
    mMutex.lock();

    mRefCount--;

    if ( 0 == mRefCount )
    {
        delete mSettingObjPtr;

        mSettingObjPtr = NULL;
    }

    mMutex.unlock();
}


#ifndef QT_NO_DEBUG
QSettings & PSettings::getSettings( void )
{
    Q_ASSERT( mSettingObjPtr != NULL );

    return *mSettingObjPtr;
}
#endif
