/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2013 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id$
 *
 */

#ifndef PSETTINGS_H
#define PSETTINGS_H

#include <QObject>
#include <QSettings>
#include <QMutex>

#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PSettings PSettings.h "support\PSettings.h"
          * @brief This class represents a configurable wrapper around the Qt QSettings class
          *
          * This class can let the QSettings object it contains retrieve the settings from the registry (Windows)
          * or from an ini-file. It does this by retrieving the command-line settings from the global application
          * object static functions (QCoreApplication) and parse it for the specific key-words.
          */
        class PEACSSUPPORTSHARED_EXPORT PSettings
        {
            Q_DISABLE_COPY( PSettings )

        public:

            /** @brief Constructor
              *
              * @param fileExt An optional extension for a filename that can be given to the commandline
              * @param parent The parent object given to the QSettings object during construction
              */
            PSettings( const QString & fileExt = QString(), QObject * parent = 0 );


            /** @brief Destructor
              */
            ~PSettings( void );


            /** @brief Returns a QSettings object instantiated according the given command-line arguments
              *
              * @return A QSettings object to store/load the application settings
              * @note This function is not thread safe. The object returned should not be stored for later
              * use
              */
#ifndef QT_NO_DEBUG
            QSettings & getSettings( void );
#else
            inline QSettings & getSettings( void ) { return *mSettingObjPtr; }
#endif

            /** @brief Returns the file name specified at the command line
              * This filename can be specified at the command line without options. Eventually
              * this name can be checked against a given extension during construction of this class
              *
              * @return mFilename
              * @note This function is not thread safe. The object returned should not be stored for later
              * use
              */
            const QString & getFileName( void ) const { return mFilename; }

        private:


            /** Member that handles the access control to this classes static data objects
              * This mutex is only used during construction and destruction.
              */
            QMutex mMutex;


            /** @brief If true all settings will be loaded from or stored in a ini-file
              * This setting can be specified at teh command-line by giving '-ini' => means true
              * If both '-ini' and '-inifile <filename>' are specified the '-inifile <filename>'
              * option overrules
              * If only '-ini' is speceified the application first looks for a '-applicationname-.ini' file
              * in the directory where the application is located and then in the users home
              * directory in the sub-folder '.-applicationname-'. If no file is found, default settings
              * will be used when loading and new settings will be stored in the file. This
              * file will then be located in the users home directory in the sub-folder '.-applicationname-'.
              *
              * '-applicationname-' is found by using the 'QCoreApplication::applicationName'
              */
            bool mUseInitSetting;


            /** @brief Contains an optional filename where the settings must be loaded from
              * This filename can be specified at the command-line by giving '-inifile <filename>'
              */
            QString mIniFilename;


            /** @brief Contains an optional filename where the settings must be loaded from
              * This filename can be specified at the command-line with no extra options, by
              * giving '<filename>'.
              * Optionally the extension of this filename can be checked with a extension given
              * during construction of this class object
              */
            static QString mFilename;


            /** @brief Contains the pointer to the applications settings object
              * This object must be used to store or retrieve application settings
              */
            static QSettings * mSettingObjPtr;


            /** @brief Keeps track of the number of instances that instantiate this class
              * For each new instantiation of this class this counter is increased by one. During
              * destruction of the object the counter is decreased. When reached zero, the
              * QSettings object is released
              */
            static int mRefCount;
        };
    }
}

#endif // PSETTINGS_H
