/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PStatistics.h 616 2013-02-04 19:20:57Z peter $
 *
 */

#ifndef __PSTATISTICS_H
#define __PSTATISTICS_H


#include "pvector.h"
#include "math.h"

#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PStatistics PStatistics.h "support/PStatistics.h"
          * @brief This class computes the statistical values of a PVector class
          */
        class PEACSSUPPORTSHARED_EXPORT PStatistics
        {

        public:

            /** @brief enumeration that define the calculations that can be performed on the PVector
              */
            typedef enum
            {
                doMin               = 0x00000001,
                doMax               = 0x00000002,
                doMean              = 0x00000004,
                doStd               = 0x00000008,
                doMedian            = 0x00000010,
                doSum               = 0x00000020,
                doMinMax            = doMin | doMax,
                doMinMaxMean        = doMin | doMax | doMean,
                doMinMaxMeanStd     = doMin | doMax | doMean | doStd,
                doMinMaxSum         = doMin | doMax | doSum,
                doMinMaxMeanSum     = doMin | doMax | doMean | doSum,
                doMinMaxMeanStdSum  = doMin | doMax | doMean | doStd | doSum,
                doDefault           = doMinMaxMeanStdSum
            } doStatisticsTypes;


            /** @brief Initializer constructor.
              * @param vector
              * @see operator=(const PStatistics &vector)
              */
            PStatistics( const PVector &vector,
                         doStatisticsTypes doStatistics /*= doDefault*/ );


            /** @brief destructor.
              */
            virtual ~PStatistics( void ) {}


            /** @brief Returns the minimum value of all elements in the
              * vector.    This one is slower because it computes the minimum value every time
              */
            inline double getMin( void ) const { return mMin; }


            /** @brief The index posistion of the minimum in the vector
              */
            inline int getMinI( void ) const { return mMinI; }


            /** @brief Returns the maximum value of all elements in the vector
              */
            inline double getMax( void ) const { return mMax; }


            /** @brief The index posistion of the maximum in the vector
              */
            inline int getMaxI( void ) const { return mMaxI; }


            /** @brief The mean of the vector
              */
            inline double getMean( void ) const { return mMean; }


            /** @brief The mean of the vector
              */
            inline double getMedian( void ) const { return mMedian; }


            /** @brief The standard deviation of the vector
              */
            inline double getStd( void ) const { return mStd; }


            /** @brief The diffrence between max and min of the vector
              */
            inline double getRange( void ) const { return mMax - mMin; }

            /** @brief The sum of the data
              */
            inline double getSum( void ) const { return mSum; }


            /** @brief recompute all statistical values
              */
            inline void updateStatistics( doStatisticsTypes doStatistics = doDefault ) { compStatistics( doStatistics ); }


        private:

            /** @brief Determine the minimum and maximum value in the vector
              * @see min()
              * @see max()
              */
            void compStatistics( doStatisticsTypes doStatistics );


            /** @brief The minimum value in the vector
              * Value is computed by  getMinMax(void)
              * @see getMinMax(void)
              */
            double mMin;


            /** @brief The maximum value in the vector
              * Value is computed by  getMinMax(void)
              * @see getMinMax(void)
              */
            double mMax;


            /** @brief The index minimum value in the vector
              * Value is computed by  getMinMax(void)
              * @see getMinMax(void)
              */
            int mMinI;


            /** @brief The index of the maximum value in the vector
              * Value is computed by  getMinMax(void)
              * @see getMinMax(void)
              */
            int mMaxI;


            /** @brief the mean of the data
              */
            double mMean;


            /** @brief the standard deviation of the data
              */
            double mStd;


            /** @brief the median of the data
              */
            double mMedian;

            /** @brief the sum of the data
              */
            double mSum;



            /** @brief The Vector
              */
            const PVector & mVec;
        };
    }
}

#endif //__PSTATISTICS_H
