/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pCoordinateSystem.cpp 1196 2014-04-15 10:20:27Z eelco $
 *
 */

#include "pCoordinateSystem.h"
#include "pfunctions.h"

#include "math.h"


using namespace Peacs::Support;


//=== Constructors ============================================================

PCoordinateSystem::PCoordinateSystem (void)
{
    mAxes[X] = P3DVector( 1, 0, 0 );
    mAxes[Y] = P3DVector( 0, 1, 0 );
    mAxes[Z] = P3DVector( 0, 0, 1 );
}


PCoordinateSystem::PCoordinateSystem (const PCoordinateSystem &coorSys)
{
    mAxes[X] = coorSys.mAxes[X];
    mAxes[Y] = coorSys.mAxes[Y];
    mAxes[Z] = coorSys.mAxes[Z];
}


PCoordinateSystem::PCoordinateSystem (const P3DVector &x, const P3DVector &y, const P3DVector &z)
{
    mAxes[X] = x;
    mAxes[Y] = y;
    mAxes[Z] = z;
}


PCoordinateSystem::PCoordinateSystem (double phi, double theta, double gamma)
{
    createEulerZYZRotationMatrix( phi, theta, gamma );
}


PCoordinateSystem::~PCoordinateSystem()
{
}


//=== Operators ==============================================================

bool PCoordinateSystem::operator== (const PCoordinateSystem &coorSys) const
{
    return ( mAxes[X] == coorSys.mAxes[X] &&
             mAxes[Y] == coorSys.mAxes[Y] &&
             mAxes[Z] == coorSys.mAxes[Z] );
}


bool PCoordinateSystem::operator!= (const PCoordinateSystem &coorSys) const
{
    return ( mAxes[X] != coorSys.mAxes[X] ||
             mAxes[Y] != coorSys.mAxes[Y] ||
             mAxes[Z] != coorSys.mAxes[Z] );
}


const PCoordinateSystem & PCoordinateSystem::operator= (const PCoordinateSystem &coorSys)
{
    if ( this != &coorSys )
    {
        mAxes[X] = coorSys.mAxes[X];
        mAxes[Y] = coorSys.mAxes[Y];
        mAxes[Z] = coorSys.mAxes[Z];
    }

    return *this;
}


double PCoordinateSystem::determinant(void) const
{
    return deter( mAxes[X], mAxes[Y], mAxes[Z] );
}


void PCoordinateSystem::clear( void )
{
    // @brief clears the content of the vector to 0 (zero)
    mAxes[X] = P3DVector( 1, 0, 0 );
    mAxes[Y] = P3DVector( 0, 1, 0 );
    mAxes[Z] = P3DVector( 0, 0, 1 );
}


void PCoordinateSystem::createLocalSystemWithGivenZaxis(const P3DVector &zAxis)
{
    double len1  =  zAxis.length();
    double len2  =  sqrt( zAxis.x() * zAxis.x() +
                          zAxis.y() * zAxis.y() );
    double sint  = -zAxis.z() / len1;
    double cost  =  len2      / len1;
    double sinf  =  zAxis.y() / len2;
    double cosf  =  zAxis.x() / len2;

    mAxes[X].x() =  cosf * cost;
    mAxes[X].y() = -sinf;
    mAxes[X].z() =  cosf * sint;

    mAxes[Y].x() =  sinf * cost;
    mAxes[Y].y() =  cosf;
    mAxes[Y].z() =  sinf * sint;

    mAxes[Z].x() = -sint;
    mAxes[Z].y() =  0;
    mAxes[Z].z() =  cost;
}


void PCoordinateSystem::createLocalSystemWithGivenAxisAndAngle(const P3DVector &axis, double angle)
{
    P3DVector a( axis );

    a.normate();

    double x    = a.x();
    double y    = a.y();
    double z    = a.z();
    double ac   = cos( angle );
    double ac1  = 1 - ac;
    double as   = sin( angle );

    mAxes[X].x()= x * x * ac1 + ac;
    mAxes[X].y()= x * y * ac1 - z * as;
    mAxes[X].z()= x * z * ac1 + y * as;

    mAxes[Y].x()= y * x * ac1 + z * as;
    mAxes[Y].y()= y * y * ac1 + ac;
    mAxes[Y].z()= y * z * ac1 - x * as;

    mAxes[Z].x()= z * x * ac1 - y * as;
    mAxes[Z].y()= z * y * ac1 + x * as;
    mAxes[Z].z()= z * z * ac1 + ac;
}


void PCoordinateSystem::createEulerZYZRotationMatrix(double phi, double theta, double gamma)
{
    //ZYZ rotation matrix; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
    double sinf = sin( phi );
    double cosf = cos( phi );
    double sint = sin( theta );
    double cost = cos( theta );
    double sing = sin( gamma );
    double cosg = cos( gamma );

    mAxes[X].x()=  cosf * cost * cosg - sinf * sing;
    mAxes[X].y()= -cosf * cost * sing - sinf * cosg;
    mAxes[X].z()=  cosf * sint;

    mAxes[Y].x()=  sinf * cost * cosg + cosf * sing;
    mAxes[Y].y()= -sinf * cost * sing + cosf * cosg;
    mAxes[Y].z()=  sinf * sint;

    mAxes[Z].x()= -sint * cosg;
    mAxes[Z].y()=  sint * sing;
    mAxes[Z].z()=  cost;
}


P3DVector PCoordinateSystem::getToLocalSystem(const P3DVector &vect) const
{
    return P3DVector( vect.dot( mAxes[X] ),
                      vect.dot( mAxes[Y] ),
                      vect.dot( mAxes[Z] ) );
}


void PCoordinateSystem::toLocalSystem(P3DVector &vector) const
{
    P3DVector vect( vector );

    vector.x() = vect.dot( mAxes[X] );
    vector.y() = vect.dot( mAxes[Y] );
    vector.z() = vect.dot( mAxes[Z] );
}


void PCoordinateSystem::toLocalSystem(P3DVertices &vect) const
{
    for ( int i = 0; i < vect.size(); i++ )
    {
        toLocalSystem( vect[i] );
    }
}


P3DVertices PCoordinateSystem::getToLocalSystem(const P3DVertices &vers) const
{
    P3DVertices rotver( vers );

    toLocalSystem( rotver );

    return rotver;
}


void PCoordinateSystem::rotateSystemFromEulerVectorAndAngle(const P3DVector& axis, double angle)
{
    if ( angle != 0 )
    {
        P3DVector u = axis.normalized();

        P3DVector rotmat[3] = { u * u.x(),
                                u * u.y(),
                                u * u.z() };
        P3DVector c[3] = { (P3DVector( 1, 0, 0 ) - rotmat[0]) * cos( angle ),
                           (P3DVector( 0, 1, 0 ) - rotmat[1]) * cos( angle ),
                           (P3DVector( 0, 0, 1 ) - rotmat[2]) * cos( angle ) };
        P3DVector s[3];

        s[0].x() =  0;
        s[0].y() = -u.z();
        s[0].z() =  u.y();

        s[1].x() =  u.z();
        s[1].y() =  0;
        s[1].z() = -u.x();

        s[2].x() = -u.y();
        s[2].y() =  u.x();
        s[2].z() =  0;

        rotmat[0] += c[0] + ( s[0] * sin( angle ) );
        rotmat[1] += c[1] + ( s[1] * sin( angle ) );
        rotmat[2] += c[2] + ( s[2] * sin( angle ) );

        mAxes[X] = P3DVector(rotmat[0].dot(mAxes[X]),
                             rotmat[1].dot(mAxes[X]),
                             rotmat[2].dot(mAxes[X]));

        mAxes[Y] = P3DVector(rotmat[0].dot(mAxes[Y]),
                             rotmat[1].dot(mAxes[Y]),
                             rotmat[2].dot(mAxes[Y]));

        mAxes[Z] = P3DVector(rotmat[0].dot(mAxes[Z]),
                             rotmat[1].dot(mAxes[Z]),
                             rotmat[2].dot(mAxes[Z]));

        mAxes[X].normate();
        mAxes[Y].normate();
        mAxes[Z].normate();
    }
}


void PCoordinateSystem::rotateSystem(PCoordinateSystem & system) const
{
    system[X].x() = mAxes[X].dot( system.mAxes[X] );
    system[X].y() = mAxes[Y].dot( system.mAxes[X] );
    system[X].z() = mAxes[Z].dot( system.mAxes[X] );

    system[Y].x() = mAxes[X].dot( system.mAxes[Y] );
    system[Y].y() = mAxes[Y].dot( system.mAxes[Y] );
    system[Y].z() = mAxes[Z].dot( system.mAxes[Y] );

    system[Z].x() = mAxes[X].dot( system.mAxes[Z] );
    system[Z].y() = mAxes[Y].dot( system.mAxes[Z] );
    system[Z].z() = mAxes[Z].dot( system.mAxes[Z] );
}


void PCoordinateSystem::restrictAngle(double &angle)
{
    if ( angle >=  M_PI )
    {
        angle -= 2.0 * M_PI;
    }
    else if (angle <= -M_PI )
    {
        angle += 2.0 * M_PI;
    }
}

void PCoordinateSystem::compThetaPhiGamma(double &phi, double &theta, double &gamma) const
{
    double      len = sqrt( mAxes[Z].x() * mAxes[Z].x() +
                            mAxes[Z].y() * mAxes[Z].y() );
    P3DVector   xasFirstRot;

    gamma = 0;

    if ( len != 0 )
    {
        phi = atan2( mAxes[Z].y(), mAxes[Z].x() );

        restrictAngle( phi );
    }
    else if ( mAxes[X].y() != 0 )		      // special case, pure rotation around z
    {
        phi = atan2( mAxes[X].y(), mAxes[X].x() ) / 2.0; // distributed between mPhi and mGamma!

        restrictAngle( phi );
    }
    else
    {
        phi = 0;
    }

    if ( fabs(phi) <= M_PI_4 ||
         fabs(phi) >= 0.75 * M_PI )
    {
        theta = atan2( mAxes[Z].x() / cos( phi ),
                       mAxes[Z].z() );
    }
    else
    {
        theta = atan2( mAxes[Z].y() / sin( phi ),
                       mAxes[Z].z() );
    }

    if ( mAxes[X].z() != 0 ||
         mAxes[Y].z() != 0 )
    {
        gamma = -atan2( mAxes[Y].z(),
                        mAxes[X].z() );
    }

    PCoordinateSystem tempSys( phi, theta, gamma );

    xasFirstRot = tempSys.getToLocalSystem( P3DVector( 1, 0, 0 ) );

    if ( mAxes[X].dot(xasFirstRot) < 0 )
    {
        phi += M_PI;

        restrictAngle( phi );

        if ( fabs(phi) <= M_PI_4 ||
             fabs(phi) >= 0.75 * M_PI )
        {
            theta = atan2( mAxes[Z].x() / cos( phi ),
                           mAxes[Z].z() );
        }
        else
        {
            theta = atan2( mAxes[Z].y() / sin( phi ),
                           mAxes[Z].z() );
        }

        if ( mAxes[X].z() != 0 ||
             mAxes[Y].z() != 0 )
        {
            gamma = -atan2( mAxes[Y].z(),
                            mAxes[X].z() );
        }
    }
}
