/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pCoordinateSystem.h 415 2012-02-26 20:49:06Z eelco $
 *
 */

#ifndef _PCOORDINATESYSTEM_H
#define _PCOORDINATESYSTEM_H


#include <QTextStream>
#include <QDataStream>

#include "P3DVector.h"
#include "P3DVertices.h"

#include "PeacsSupport_global.h"

namespace Peacs
{
    namespace Support
    {
        /** @class PCoordinateSystem PCoordinateSystem.h "support/PCoordinateSystem.h"
          * @brief This class describes  a 3D coordinate system. Support function to create rotated
          * coordinate systems are supplied.
          */
        class PEACSSUPPORTSHARED_EXPORT PCoordinateSystem
        {

        public:

            /** @brief identified the number of coordinates the coordSys holds
              */
            static const int AXES_ITEM_COUNT = 3;


            /** @brief type definition of the number of items in PCoordinateSystem.        */
            enum xyz_t { X = 0,
                         Y = 1,
                         Z = 2 };


            /** @brief Default constructor.
              * Initialze all coordinates to 0
              */
            PCoordinateSystem (void);


            /** @brief Copy constructor.
              * @see operator=(const PCoordinateSystem &coorSys)
              * @param coorSys the source to initizlize with
              */
            PCoordinateSystem (const PCoordinateSystem &coorSys);


            /** @brief Initializer constructor.
              * @param x x axis
              * @param y y axis
              * @param z z axis
              */
            PCoordinateSystem (const P3DVector &x, const P3DVector &y, const P3DVector &z);


            /** @brief rotate coordinate system constructor.
              * The system is based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param phi       rotation around  ?? axes
              * @param theta     rotation around  ?? axes
              * @param gamma     rotation around  ?? axes
              */
            PCoordinateSystem (double phi, double theta, double gamma);


            /** @brief Destuctor
              */
            virtual ~PCoordinateSystem();


            /** @brief assignment operator.
              * copy the coordinates from coorSys
              *  @param coorSys the one that is copied
              */
            const PCoordinateSystem &operator= (const PCoordinateSystem &coorSys);


            /** @brief Returns the item at index position index as a modifiable
              * reference.
              * @param index a valid index position in the list
              * (i.e. 0 <= index < VECTOR_ITEM_COUNT)
              */
            inline P3DVector& operator[] (xyz_t index) { return mAxes[index]; }


            /** @brief This is an member function, provided for
              * convenience.
              * @param index a valid index position in the list
              * (i.e. 0 <= index < VECTOR_ITEM_COUNT)
              */
            inline const P3DVector &operator[] (xyz_t index) const { return mAxes[index]; }


            /** @brief equal operator.
              * true if all coordinates match between coorSys and this one
              *	@param coorSys the one that is used to check
              * @see operator!=(const PCoordinateSystem &coorSys)
              */
            bool operator==(const PCoordinateSystem &coorSys) const;


            /** @brief unequal operator.
              * true if one of the coordinates does not match between coorSys
              * and this one
              * @param coorSys the one that is used to check
              * @see operator==(const PCoordinateSystem &coorSys)
              */
            bool operator!=(const PCoordinateSystem &coorSys) const;


            /** @brief Returns the determinant of mAxes
              * @returns the determinant of 3 vectors of the coordinate system
              */
            double determinant( void ) const;


            /** @brief returns the x coordinate
              */
            inline const P3DVector &x(void) const { return mAxes[X]; }


            /** @brief returns the y coordinate
              */
            inline const P3DVector &y(void) const { return mAxes[Y]; }


            /** @brief returns the z coordinate
              */
            inline const P3DVector &z(void) const { return mAxes[Z]; }


            /** @brief clears the content of the coorSys to 0 (zero)
              */
            void clear( void );


            /** @brief cretae a local coordinate system based on a given Z axes
              * @param zAxis the given z axis
              */
            void createLocalSystemWithGivenZaxis(const P3DVector &zAxis);


            /** @brief cretae a local coordinate system based on a given Z axes
              * @param Axis the given axis
              * @param angle Rotation agle around the axis
              */
            void createLocalSystemWithGivenAxisAndAngle(const P3DVector &axis, double angle);


            /** @brief create
              * The system is based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param phi       rotation around  ?? axes
              * @param theta     rotation around  ?? axes
              * @param gamma     rotation around  ?? axes
              */
            void createEulerZYZRotationMatrix(double phi, double theta, double gamma);


            /** @brief rotate vector into this system
              * @param vect the vector to be rotated
              * @return the rotated vector
              */
            P3DVector getToLocalSystem(const P3DVector &vect) const;


            /** @brief rotate vector into this system
              * @returns the rotated vector
              */
            void toLocalSystem(P3DVector &vect) const;


            /** @brief rotate vector into this system
              * @param vect the vector to be rotated
              * @return the rotated vector
              */
            P3DVertices getToLocalSystem(const P3DVertices &vers) const;


            /** @brief rotate vertices into this system
              * @returns the rotated vector
              */
            void toLocalSystem(P3DVertices &vect) const;


            /** Create a retated Coordinate system based on the euler rataion vector and ist angle
              * @param axis euler rotation vector
              * @param angle euler angle
              */
            void rotateSystemFromEulerVectorAndAngle(const P3DVector &axis, double angle);


            /** @brief rotate the Coordinate system system to the system of this object
              * @param system th system taht is rotated
              */
            void rotateSystem(PCoordinateSystem & system) const;


            /** @brief compute the euler angles of the coordinate system
              * @param phi the computed angel around the ?? axis
              * @param theta the computed angel around the ?? axis
              * @param gamma the computed angel around the ?? axis
              */
            void compThetaPhiGamma(double &phi, double &theta, double &gamma) const;


        private:

            /** @brief limite the rotation of the given angle between +pi and -pi
              * @param angle
              */
            static void restrictAngle(double &angle);

            /** @brief 3D coordinate (x, y, z)
              */
            P3DVector mAxes[AXES_ITEM_COUNT];
        };
    }
}


/** @brief Streams the 3 coordinates to a datastream
  * @param vec The vecor that is streamed
  * @returns stream
  * @see PCoordinateSystem
  */
QDataStream &operator<< (QDataStream &stream, const Peacs::Support::PCoordinateSystem &vec);


/** @brief Streams the 3 coordinates from a datastream
  * @param vec The vecor that is streamed
  * @returns stream
  * @see PCoordinateSystem
  */
QDataStream &operator>> (QDataStream &stream, Peacs::Support::PCoordinateSystem &vec);


/** @brief Streams the 3 coordinates to a textstream
  * @param vec The vecor that is streamed
  * @returns stream
  *  @see PCoordinateSystem
  */
QTextStream &operator<< (QTextStream &stream, const Peacs::Support::PCoordinateSystem &vec);


/** @brief Streams the 3 coordinates from a textstream
  * @param vec The vecor that is streamed
  * @returns stream
  * @see PCoordinateSystem
  */
QTextStream &operator>> (QTextStream &stream, Peacs::Support::PCoordinateSystem &vec);


#endif //_PCOORDINATESYSTEM_H
