/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pbox.h 691 2013-04-12 13:41:25Z peter $
 *
 */

#ifndef PBOX_H
#define PBOX_H

#include "P3DVertices.h"

#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {
        /** \class PBox pbox.h "support/pbox.h"
          * @brief This class computes a box that can be drawn around a number of vertices
          */
        class PEACSSUPPORTSHARED_EXPORT PBox
        {
        public:

            /** @brief Constructor
              * @param vertices the vertices representing the geometry to be contained by this box
              */
            PBox( const P3DVertices & vertices );


            /** @brief Checks if the box contains valid geometry coordinate information
              * @returns true if a valid box can be created around the vertices
              */
            bool isValid() const { return !mVertices.isEmpty() &&
                                          mMaxBox.x() > mMinBox.x() &&
                                          mMaxBox.y() > mMinBox.y() &&
                                          mMaxBox.z() > mMinBox.z(); }


            /** @brief update the derived member variables
              */
            void update();


            /** @brief Returns mMinBox
              * @returns the min x, y and z coordinate of all vertices (the minBox)
              */
            inline const P3DVector &getMinBox() const { return mMinBox; }


            /** @brief Returns mMaxBox
              * @returns the max x, y and z coordinate of all vertices (the maxBox)
              */
            inline const P3DVector &getMaxBox() const { return mMaxBox; }


            /** @brief Returns the difference between mMaxBox and mMinBox
              * @returns the range in the x, y and z coordinate of all vertices
              */
            inline P3DVector getBoxRange() const { return mMaxBox - mMinBox; }


            /** @brief Returns the center value between mMAxBox and mMinBox
              * @returns (getMaxBox() + getMinBox()) / 2.0
              */
            inline P3DVector getBoxCenter() const { return (mMaxBox + mMinBox) / 2;}

            /** @brief Returns the upper left corner of the box
              */
            inline P3DVector getMaxXYZ() const { return mMaxBox; }

            /** @brief Returns the lower right corner of the box
              */
            inline P3DVector getMinXYZ() const { return mMinBox; }

            /** @brief Returns the lower right corner of the box
              */
            inline P3DVector getMaxXMinYZ() const { return P3DVector(mMaxBox.x(), mMinBox.y(),mMinBox.z()); }

            inline P3DVector getMaxXYMinZ() const { return P3DVector(mMaxBox.x(), mMaxBox.y(),mMinBox.z()); }

            inline P3DVector getMaxYMinXZ() const { return P3DVector(mMinBox.x(), mMaxBox.y(),mMinBox.z()); }

            inline P3DVector getMaxZMinXY() const { return P3DVector(mMinBox.x(), mMinBox.y(),mMaxBox.z()); }

            inline P3DVector getMaxXZMinY() const { return P3DVector(mMaxBox.x(), mMinBox.y(),mMaxBox.z()); }

            inline P3DVector getMaxYZMinX() const { return P3DVector(mMinBox.x(), mMaxBox.y(),mMaxBox.z()); }


            /** @brief Returns true if pos is inside the box around the vertices
              * @param pos the postion of a vertex that is checked
              * @see mMaxBox
              * @see mMinBox
              */
            bool insideBox( const P3DVector &pos ) const;


        private:

            /** @brief Disabled default constructor
              */
            PBox();


            /** @brief The vertices representing the geometry contained by this box
              */
            const P3DVertices &mVertices;


            /** @brief max coordinate of the box around the geometry
              */
            P3DVector mMaxBox;


            /** @brief min coordinate of the box around the geometry
              */
            P3DVector mMinBox;
        };
    }
}
#endif // PBOX_H
