/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pedge.h 415 2012-02-26 20:49:06Z eelco $
 *
 */

#ifndef PEDGE_H
#define PEDGE_H

#include "PeacsSupport_global.h"

namespace Peacs
{
    namespace Support
    {
        /** @class PEdge pedge.h "support/pedge.h"
          * @brief This class holds an edge
          * An edge decribes the connection between two vertices. The two verices
          * are listed by their index number
          * @see PGeometry
          * @see P3DVector
          */
        class PEACSSUPPORTSHARED_EXPORT PEdge
        {

        public:


            /** @brief identified the number of indexes the edge holds
              */
            static const int EDGE_ITEM_COUNT = 2;


            /** @brief type definition of the number of items in PEdge.
              */
            enum edges_t { A = 0,
                           B = 1 };


            /** @brief default constructor
              */
            PEdge();


            /** @brief Contructor with the edge index a and b
              * @param a edge index A
              *	@param b edge index B
              */
            PEdge(int a, int b);


            /** @brief Copy constructor
              *  @param edge is copied into this object
              */
            PEdge(const PEdge &edge);


            /** @brief copy operator.
              * edge is copied into this object and this object is returned
              * @param edge is copied into this object
              */
            const PEdge &operator= ( const PEdge& edge );


            /** @brief equal operator.
              * true if both indices are the same otherwise false
              * @param edge that is compered to
              */
            bool operator== ( const PEdge& edge ) const;


            /** @brief unequal operator.
              * false if all indices are the same
              * @param edge that is compered to
              */
            bool operator!= ( const PEdge& edge ) const;


            /** @brief copies the given edge and changes the direction of the edge by exchanging the indices
              * @param edge is copied and inverted into this object
              */
            const PEdge &invert( const PEdge& edge );


            /** @brief changes the direction of the edge by exchanging the indices
              * @return a reference to this (inverted) edge object
              */
            const PEdge &invert( void );


            /** @brief changes the direction of the edge by exchanging the indices
              * @return an inverted copy of this edge object
              */
            PEdge inverted( void ) const;


            /** /brief returns first const edge index
              */
            inline const int &a( void ) const { return mEdge[A]; }


            /** /brief returns first (non const) edge index
              */
            inline int &a(void) { return mEdge[A]; }

            /** /brief returns second const edge index
              */
            inline const int &b(void) const { return mEdge[B]; }


            /** /brief returns second (non const) edge index
              */
            inline int &b(void) { return mEdge[B]; }


#ifndef QT_NO_DEBUG

            /** @brief Returns this index to of either index A or B indicated by index
              * @param index of the edge vertex
              * @see operator [](int i) const
              */
            int & operator[] ( edges_t index );


            /** @brief Returns const index to of either index A or B indicated by index
              * @param index of the edge vertex
              * @see operator [](int i)
              */
            const int & operator[] ( edges_t index ) const;
#else

            /** @brief Returns this index to of either index A or B indicated by index
              * @param index of the edge vertex
              * @see operator [](int i) const
              */
            inline int & operator[] ( edges_t index ) { return mEdge[index]; }


            /** @brief Returns const index to of either index A or B indicated by index
              * @param index of the edge vertex
              * @see operator [](int i)
              */
            inline const int & operator[] ( edges_t index ) const { return mEdge[index]; }
#endif

        private:

            /** @brief the edge
              */
            int mEdge[EDGE_ITEM_COUNT];
        };
    }
}

#endif // PEDGE_H
