/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pface.cpp 1118 2014-02-26 16:09:13Z peter $
 *
 */

#include "pface.h"
#include "math.h"
#include "stddef.h"

#include <QDebug>

#ifdef Q_CC_MSVC
#pragma warning( disable : 4351 )
#endif


using namespace Peacs::Support;


PFace::PFace(const PGeometry &myGeom) :
    mFace(),
    mMyGeom(myGeom),
    mNormal(),
    mArea( 0 )
//    mGravityCenter(),
//    mDeter( 0 )
{
    mFace[P1] = -1;
    mFace[P2] = -1;
    mFace[P3] = -1;

    updateMembers();
}


PFace::PFace(const PFace &f) :
    mFace(),
    mMyGeom(f.mMyGeom),
    mNormal( f.mNormal ),
    mArea( f.mArea )/*,
    mGravityCenter( f.mGravityCenter ),
    mDeter( f.mDeter )*/
{
    mFace[P1] = f.mFace[P1];
    mFace[P2] = f.mFace[P2];
    mFace[P3] = f.mFace[P3];
}


PFace::PFace( int i1, int i2, int i3,
              const PGeometry &myGeom ):
    mFace(),
    mMyGeom(myGeom),
    mNormal(),
    mArea( 0 )/*,
    mGravityCenter(),
    mDeter( 0 )*/
{
    mFace[P1] = i1;
    mFace[P2] = i2;
    mFace[P3] = i3;

    updateMembers();
}


const PFace &PFace::operator= ( const PFace& f )
{
    if ( this != & f )
    {
        mFace[P1]       = f.mFace[P1];
        mFace[P2]       = f.mFace[P2];
        mFace[P3]       = f.mFace[P3];
//        mDeter          = f.mDeter;
        mNormal         = f.mNormal;
//        mGravityCenter  = f.mGravityCenter;
        mArea           = f.mArea;
        const_cast <PGeometry &> (mMyGeom)         = f.mMyGeom;
    }

    return *this;
}


void PFace::setData(int i0, int i1, int i2)
{
    if ( i0 >= 0 && i0 < mMyGeom.nVertices()  &&
         i1 >= 0 && i1 < mMyGeom.nVertices()  &&
         i2 >= 0 && i2 < mMyGeom.nVertices()  )
        {
        mFace[P1]       = i0;
        mFace[P2]       = i1;
        mFace[P3]       = i2;
        updateMembers();
    }
 }


void PFace::changeVertexNumber(int index1,int index2)
{
    if  ( mFace[P1] == index1 )
        mFace[P1] = index2;
    else if  ( mFace[P1] == index2 )
        mFace[P1] = index1;

    if  ( mFace[P2] == index1 )
        mFace[P2] = index2;
    else if  ( mFace[P2] == index2 )
        mFace[P2] = index1;

    if  ( mFace[P3] == index1 )
        mFace[P3] = index2;
    else if  ( mFace[P3] == index2 )
        mFace[P3] = index1;
}


bool PFace::operator== ( const PFace& f ) const
{
    return ( mFace[P1] == f.mFace[P1] &&
             mFace[P2] == f.mFace[P2] &&
             mFace[P3] == f.mFace[P3] );
}


bool PFace::operator!= ( const PFace& f ) const
{
    return ( mFace[P1] != f.mFace[P1] ||
             mFace[P2] != f.mFace[P2] ||
             mFace[P3] != f.mFace[P3] );
}


const P3DVector &PFace::vertex( facecount_t i ) const
{
    return mMyGeom.vertex(mFace[i]);
}



#ifndef QT_NO_DEBUG

const P3DVector & PFace::vertex0(void) const
{
    Q_ASSERT( mMyGeom.nVertices() != 0 );  // Test just to be sure
    Q_ASSERT( mFace[P1] >= 0 && mFace[P1] < mMyGeom.nVertices() );  // Test just to be sure
    return mMyGeom.vertex(mFace[P1]);
}

const P3DVector & PFace::vertex1(void) const
{
    Q_ASSERT( mMyGeom.nVertices() != 0 );  // Test just to be sure
    Q_ASSERT( mFace[P2] >= 0 && mFace[P2] < mMyGeom.nVertices() );  // Test just to be sure

    return mMyGeom.vertex(mFace[P2]);
}


const P3DVector & PFace::vertex2(void) const
{
    Q_ASSERT( mMyGeom.nVertices() != 0 );  // Test just to be sure
    Q_ASSERT( mFace[P3] >= 0 && mFace[P3] < mMyGeom.nVertices() );  // Test just to be sure

    return mMyGeom.vertex(mFace[P3]);
}

#endif

void PFace::invertNormal()
{
    int                 i   = mFace[1];

    mFace[1] = mFace[2];
    mFace[2] = i;

    updateMembers();
}


bool PFace::contains(int vertex) const
{
    if  ( mFace[P1] == vertex ||
          mFace[P2] == vertex ||
          mFace[P3] == vertex )
    {
        return true;
    }
    else
    {
        return false;
    }
}

bool PFace::contains(const P3DVector &vertex) const
{

    if  ( vertex0() == vertex ||
          vertex1() == vertex ||
          vertex2() == vertex )
    {
        return true;
    }
    else
    {
        return false;
    }
}

bool PFace::inFaceSpace(const P3DVector &pos) const
{
    // TODO Use normalysed distances
    if ( (vertex0() - pos).length() < 30.0 ||
         (vertex1() - pos).length() < 30.0 ||
         (vertex2() - pos).length() < 30.0  )
    {
        return true;
    }
    else
    {
        return false;
    }
}


double PFace::angle0(void) const
{
    P3DVector tmpVector1 = (vertex1() - vertex0()).normalized();
    P3DVector tmpVector2 = (vertex2() - vertex0()).normalized();

    return acos( tmpVector1.dot( tmpVector2 ) ) * 180.0 / M_PI;
}


double PFace::angle1(void) const
{
    P3DVector tmpVector1 = (vertex0() - vertex1()).normalized();
    P3DVector tmpVector2 = (vertex2() - vertex1()).normalized();

    return acos( tmpVector1.dot( tmpVector2 ) ) * 180.0 / M_PI;
}


double PFace::angle2(void) const
{
    P3DVector tmpVector1 = (vertex0() - vertex2()).normalized();
    P3DVector tmpVector2 = (vertex1() - vertex2()).normalized();

    return acos( tmpVector1.dot( tmpVector2 ) ) * 180.0 / M_PI;
}


double PFace::minAngle(void) const
{
    return qMin( angle0(),
                 qMin( angle1(),
                       angle2() ) );
}


double PFace::maxAngle(void) const
{
    return qMax( angle0(),
                 qMax( angle1(),
                       angle2() ) );
}


double PFace::minEdgeLength() const
{
    return qMin( (vertex0() - vertex1()).length(),
           qMin( (vertex1() - vertex2()).length(),
                 (vertex0() - vertex2()).length() ) );
}


double PFace::maxEdgeLength() const
{
    return qMax( (vertex0() - vertex1()).length(),
           qMax( (vertex1() - vertex2()).length(),
                 (vertex0() - vertex2()).length() ) );

}



double PFace::solidAngle(const P3DVector &observationPnt) const
{
    return compSolidAngle( vertex0(), vertex1(), vertex2(), observationPnt );
}


int PFace::projectVerOnFace(const P3DVector &ver, P3DVector &verOnFace) const
{
    double lambda,mu,d;
    int result = projectVerOnFace(ver,lambda,mu,d);
    verOnFace = pointOnFace(lambda, mu);
    return result;
}


int PFace::projectVerOnFace(const P3DVector &ver, double &lambda, double &mu, double &dif) const
{
    // Project the vertex on a face
    int result = 0;

    P3DVector   s1( vertex1() - vertex0() );
    P3DVector   s2( vertex2() - vertex0() );
    P3DVector   n;
    P3DVector   h;
    h.x() = 1.0;
    n.x() = deter(h,s1,s2);
    h.x() = 0.0;
    h.y() = 1.0;
    n.y() = deter(h,s1,s2);
    h.y() = 0.0;
    h.z() = 1.0;
    n.z() = deter(h,s1,s2);

    double tmp = n.sqlength();

    if ( tmp != 0 )
    {
        /* To scale n such that is in a similar range as s1 and s2 and
         * consequently diff is in the same range as lambda and mu */
        tmp = sqrt( sqrt( tmp ) );
        n /= tmp;

        P3DVector   sav( ver - vertex0() );
        double      det = deter( s1, s2, n );
        if ( det != 0.0 )
        {
            lambda = deter( sav, s2,  n ) / det;
            mu     = deter( s1,  sav, n ) / det;
            dif    = deter( s1,  s2, sav ) / det;
            if ( det < 0 )
                result = 1;
            else
                result = -1;
        }
        else
        {
            qDebug() << "point on ver error 2";
        }
    }
    else
    {
        qDebug() << "point on ver error";
    }
    return result;
}


double PFace::vertex2FaceDist(const P3DVector &ver, P3DVector &verOnFace, double &lab, double &mu) const
{
    double d = HUGE_VAL;
    lab = 0;
    mu = 0;
    double diff=0;

    if ( projectVerOnFace( ver, lab, mu, diff ) != 0 )
    {
        if ( lab >= 0 &&
             mu >= 0 &&
             (lab + mu) <= 1 )
        {
            verOnFace = vertex0() * ( 1 - lab - mu ) +
                        vertex1() * lab +
                        vertex2() * mu;
            d = ( ver - verOnFace ).length();
        }
        else if ( lab < 0 )
        {
            lab = 0;
            edgeDist( ver, vertex0(), vertex2(), verOnFace, mu );
            d = ( ver - verOnFace ).length();
        }
        else if ( mu < 0 )
        {
            mu = 0;
            edgeDist( ver, vertex0(), vertex1(), verOnFace, lab );
            d = ( ver - verOnFace ).length();
        }
        else
        {
            edgeDist( ver, vertex1(), vertex2(), verOnFace, mu );
            d = ( ver - verOnFace ).length();
            lab = 1.0 - mu;
        }
    }
    return d;
}


//ptridist
double PFace::vertex2FaceDist(const P3DVector &ver,P3DVector &verOnFace) const
{
    double lab = 0;
    double mu = 0;
    return vertex2FaceDist(ver,verOnFace,lab,mu);
}


double PFace::edgeDist(const P3DVector &p,
                       const P3DVector &l1,
                       const P3DVector &l2,
                       P3DVector &verOnline,
                       double    &alpha) const
{
    // plindist(p, l1, l2, cp, lab): distance beween point p and
    // line segment l1-l2. In cp the point on l1-l2 closed to
    // p is returned; in lab the fraction of the distance
    // l1-cp in respect to l1-l2 is returned.

    P3DVector   s( l2 - l1 );
    double      sq = s.sqlength();

    if ( sq == 0 )
    {
        verOnline   = l1;
        alpha       = 0;
    }
    else
    {
        alpha = qBound(0.0, (p - l1).dot(s) / sq, 1.0);
        verOnline   = l1 + s * alpha;
    }
    return ( p - verOnline ).length();
}


void PFace::updateMembers(void)
{
//    P3DVector r1 = vertex1() - vertex0();
//    P3DVector r2 = vertex2() - vertex0();

    mNormal         = ( vertex2() - vertex0() ).cross( vertex1() - vertex0() );
    mArea           = mNormal.length() / 2.0;
    mNormal.normate();
//    mGravityCenter  =  vertex0();
//    mGravityCenter  += vertex1();
//    mGravityCenter  += vertex2();
//    mGravityCenter  /= 3.0;
//    mDeter          = deter( vertex0(), vertex1(), vertex2() );
}



int PFace::facePlaneCrossing(const PPlane &crossPlane) const
{
    double tmp1 = vertex0().dot( crossPlane.vector() ) + crossPlane.pos();
    double tmp2 = vertex1().dot( crossPlane.vector() ) + crossPlane.pos();
    double tmp3 = vertex2().dot( crossPlane.vector() ) + crossPlane.pos();

    if ( tmp1 > 0 && tmp2 > 0 && tmp3 > 0 )  // the cross plane does not cross the face
    {
        return 0;
    }
    else if ( tmp1 < 0 && tmp2 < 0 && tmp3 < 0 ) // the cross plane does not cross the face
    {
        return -1;
    }
    else if ( tmp1 == 0 && tmp2 == 0 && tmp3 == 0 )
    {
        return 2;
    }
    else
    {
        return 1;
    }
}



int PFace::crossSection(const PPlane &crossPlane, P3DVector& v1, P3DVector &v2, PEdge &edge1,PEdge &edge2) const
{
    int returnVal = 0;

    double tmp1 = vertex0().dot( crossPlane.vector() ) + crossPlane.pos();
    double tmp2 = vertex1().dot( crossPlane.vector() ) + crossPlane.pos();
    double tmp3 = vertex2().dot( crossPlane.vector() ) + crossPlane.pos();
    double a1;
    double a2;
    double b2;

    if ( ( tmp1 > 0  && tmp2 > 0  && tmp3 > 0  ) || // the cross plane does not cross the face
         ( tmp1 < 0  && tmp2 < 0  && tmp3 < 0  ) || // the cross plane does not cross the face
         ( tmp1 == 0 && tmp2 == 0 && tmp3 == 0 ) )  // face lies in the cross plane
    {
        // In case the face lies in the cross plane other triangles should be found to determine the contour
        returnVal = 0;
    }
    else  if ( (tmp1 > 0 && tmp2 > 0 && tmp3 < 0) ||
               (tmp1 < 0 && tmp2 < 0 && tmp3 > 0) )
    {
        // I1 and I2 are on the same side so I1-I3 and I2-I3 are the crossing edges
        a1              = tmp1 / ( crossPlane.vector().dot( vertex0() - vertex2()) );
        b2              = tmp2 / ( crossPlane.vector().dot( vertex1() - vertex2()) );
        a2              = 1 - b2;
        v1              = (vertex0()) * (1 - a1) + (vertex2()) * a1 ;
        v2              = (vertex1()) * a2 + (vertex2()) * b2;
        edge1[PEdge::A] = mFace[P3];
        edge1[PEdge::B] = mFace[P1]; // v1 is located on edge 3 (I3-I1)
        edge2[PEdge::A] = mFace[P2];
        edge2[PEdge::B] = mFace[P3]; // v2 is located on edge 2 (I2-I3)

        returnVal       = 2; // two edges are crossed
    }
    else if ( (tmp1 > 0 && tmp2 < 0 && tmp3 > 0) ||
              (tmp1 < 0 && tmp2 > 0 && tmp3 < 0) )
    {
        // I1 and I3 are on the same side so I1-I2 and I2-I3 are the crossing edges
        a1              = tmp1 / ( crossPlane.vector().dot( vertex0() - vertex1()) );
        b2              = tmp2 / ( crossPlane.vector().dot( vertex1() - vertex2()) );
        a2              = 1 - b2;
        v1              = (vertex0()) * (1 - a1) + (vertex1()) * a1;
        v2              = (vertex1()) * a2 + (vertex2()) * b2;
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P2]; // v1 is located on edge 2 (I1-I2)
        edge2[PEdge::A] = mFace[P2];
        edge2[PEdge::B] = mFace[P3]; // v2 is located on edge 2 (I2-I3)

        returnVal       = 2; // two edges are crossed
    }
    else  if ( (tmp1 < 0 && tmp2 > 0 && tmp3 > 0 ) ||
               (tmp1 > 0 && tmp2 < 0 && tmp3 < 0 ) )
    {
        // I2 and I3 are on the same side so I1-I2 and I1-I3 are the crossing edges
        a1              = tmp1 / ( crossPlane.vector().dot( vertex0() - vertex1()) );
        a2              = tmp1 / ( crossPlane.vector().dot( vertex0() - vertex2()) );
        v1              = (vertex0()) * (1 - a1) + (vertex1()) * a1 ;
        v2              = (vertex0()) * (1 - a2) + (vertex2()) * a2;
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P2]; // v1 is located on edge 2 (I1-I2)
        edge2[PEdge::A] = mFace[P3];
        edge2[PEdge::B] = mFace[P1]; // v2 is located on edge 3 (I1-I3)

        returnVal       = 2; // two edges are crossed
    }
    else if ( tmp1 == 0 && tmp2 == 0 )
    {
        v1              = vertex0();
        v2              = vertex1();
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P2]; // edge 1 touched the cross plane (I1-I2)

        returnVal       = -1; // one edge is used
    }
    else if (tmp1 == 0 && tmp3 == 0 )
    {
        v1              = vertex0();
        v2              = vertex2();
        edge1[PEdge::A] = mFace[P3];
        edge1[PEdge::B] = mFace[P1]; // edge 1 touched the cross plane (I1-I3)

        returnVal       = -1; // one edge is used
    }
    else if ( tmp2 == 0 && tmp3 == 0 )
    {
        v1              = vertex1();
        v2              = vertex2();
        edge1[PEdge::A] = mFace[P2];
        edge1[PEdge::B] = mFace[P3]; // edge 1 touched the cross plane (I2-I3)

        returnVal       = -1; // one edge is used
    }
    else if ( tmp1 == 0 &&
              ( (tmp2 > 0 && tmp3 < 0) ||
                (tmp2 < 0 && tmp3 > 0) ) )
    {
        b2              = tmp2 / ( crossPlane.vector().dot( vertex1() - vertex2() ) );
        a2              = 1 - b2;
        v1              = vertex0();
        v2              = (vertex1()) * a2 + (vertex2()) * b2;
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P1]; // v1 vertex 1
        edge2[PEdge::A] = mFace[P2];
        edge2[PEdge::B] = mFace[P3]; // v2 is located on edge 2 (I2-I3)

        returnVal       = -2; // the face is touched on one vertex and crosses one edge
    }
    else if ( tmp2 == 0 &&
              ( (tmp1 > 0 && tmp3 < 0) ||
                (tmp1 < 0 && tmp3 > 0) ) )
    {
        a2              = tmp1 / ( crossPlane.vector().dot( vertex0() - vertex2() ) );
        v1              = vertex1();
        v2              = (vertex0()) * (1 - a2) + (vertex2()) * a2;
        edge1[PEdge::A] = mFace[P2];
        edge1[PEdge::B] = mFace[P2]; // v1 vertex 2
        edge2[PEdge::A] = mFace[P3];
        edge2[PEdge::B] = mFace[P1]; // v2 is located on edge 2 (I2-I3)

        returnVal       = -2; // the face is touched on one vertex and crosses one edge
    }
    else if ( tmp3 == 0 &&
              ( (tmp1 > 0 && tmp2 < 0) ||
                (tmp1 < 0 && tmp2 > 0) ) )
    {
        a2              = tmp1 / ( crossPlane.vector().dot( vertex0() - vertex1()) );
        v1              = vertex2();
        v2              = (vertex0()) * (1 - a2) + (vertex1()) * a2;
        edge1[PEdge::A] = mFace[P3];
        edge1[PEdge::B] = mFace[P3]; // v1 vertex 3
        edge2[PEdge::A] = mFace[P1];
        edge2[PEdge::B] = mFace[P2]; // v2 is located on edge 2 (I1-I2)

        returnVal       = -2; // the face is touched on one vertex and crosses one edge
    }
    else
    {
        returnVal = 0; // the face is touched on one vertex only; not interpreted as a crossection
    }

    return returnVal;
}


int PFace::crossFaceFunctionVal( double curVal,
                                 double tmp0, double tmp1, double tmp2,
                                 P3DVector &v1, P3DVector &v2,
                                 PEdge &edge1, PEdge &edge2 ) const
{
    int returnVal = 0;

    double a12 = (curVal - tmp0) / (tmp1 - tmp0);
    double a13 = (curVal - tmp0) / (tmp2 - tmp0);
    double a23 = (curVal - tmp1) / (tmp2 - tmp1);

    if ( ( tmp0 >  curVal && tmp1  > curVal && tmp2  > curVal ) || // the cross plane does not cross the face
         ( tmp0 <  curVal && tmp1  < curVal && tmp2  < curVal ) || // the cross plane does not cross the face
         ( tmp0 == curVal && tmp1 == curVal && tmp2 == curVal ) ) // face lies in the function plane
    {
        returnVal = 0;
        /* In case the face lies in the cross plane other triangles should be found to determine the contour */
    }
    else  if ( (tmp0 > curVal && tmp1 > curVal && tmp2 < curVal) ||
               (tmp0 < curVal && tmp1 < curVal && tmp2 > curVal) )
    {
        // I1 and I2 are on the same side so I1-I3 and I2-I3 are the crossing edges
        v1              = (vertex0()) + (vertex2() - vertex0()) * a13;
        v2              = (vertex1()) + (vertex2() - vertex1()) * a23;
        edge1[PEdge::A] = mFace[P3];
        edge1[PEdge::B] = mFace[P1]; // v1 is located on edge 3 (I0-I2)
        edge2[PEdge::A] = mFace[P2];
        edge2[PEdge::B] = mFace[P3]; // v2 is located on edge 2 (I1-I2)

        returnVal       = 2; // two edges are crossed
    }
    else  if ( (tmp0 > curVal && tmp1 < curVal && tmp2 > curVal) ||
               (tmp0 < curVal && tmp1 > curVal && tmp2 < curVal) )
    {
        // I1 and I3 are on the same side so I1-I2 and I2-I3 are the crossing edges
        v1              = (vertex0()) + (vertex1() - vertex0()) * a12;
        v2              = (vertex1()) + (vertex2() - vertex1()) * a23;
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P2]; // v1 is located on edge 2 (I0-I1)
        edge2[PEdge::A] = mFace[P2];
        edge2[PEdge::B] = mFace[P3]; // v2 is located on edge 2 (I1-I2)

        returnVal       = 2; // two edges are crossed
    }
    else  if ( (tmp0 < curVal && tmp1 > curVal && tmp2 > curVal ) ||
               (tmp0 > curVal && tmp1 < curVal && tmp2 < curVal ) )
    {
        // I2 and I3 are on the same side so I1-I2 and I1-I3 are the crossing edges
        v1              = (vertex0()) + (vertex1() - vertex0()) * a12 ;
        v2              = (vertex0()) + (vertex2() - vertex0()) * a13 ;
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P2]; // v1 is located on edge 2 (I0-I1)
        edge2[PEdge::A] = mFace[P3];
        edge2[PEdge::B] = mFace[P1]; // v2 is located on edge 3 (I0-I2)

        returnVal       = 2; // two edges are crossed
    }
    else if ( tmp0 == curVal && tmp1 == curVal )
    {
        v1              = vertex0();
        v2              = vertex1();
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P2]; // edge 1 touched the cross plane (I0-I1)

        returnVal       = -1; // one edge is used
    }
    else if (tmp0 == curVal && tmp2 == curVal )
    {
        v1              = vertex0();
        v2              = vertex2();
        edge1[PEdge::A] = mFace[P3];
        edge1[PEdge::B] = mFace[P1]; // edge 1 touched the cross plane (I0-I2)

        returnVal       = -1; // one edge is used
    }
    else if ( tmp1 == curVal && tmp2 == curVal )
    {
        v1              = vertex1();
        v2              = vertex2();
        edge1[PEdge::A] = mFace[P2];
        edge1[PEdge::B] = mFace[P3]; // edge 1 touched the cross plane (I1-I2)

        returnVal       = -1; // one edge is used
    }
    else if ( tmp0 == curVal &&
              ( (tmp1 > curVal && tmp2 < curVal) ||
                (tmp1 < curVal && tmp2 > curVal) ) )
    {
        v1              = vertex0();
        v2              = (vertex1()) + (vertex2() - vertex1()) * a23 ;
        edge1[PEdge::A] = mFace[P1];
        edge1[PEdge::B] = mFace[P1]; // v1 vertex 1
        edge2[PEdge::A] = mFace[P2];
        edge2[PEdge::B] = mFace[P3]; // v2 is located on edge 2 (I1-I2)

        returnVal       = -2; // the face is touched on one vertex and crosses one edge
    }
    else if ( tmp1 == curVal &&
              ( (tmp0 > curVal && tmp2 < curVal) ||
                (tmp0 < curVal && tmp2 > curVal) ) )
    {
        v1              = vertex1();
        v2              = (vertex0()) + (vertex2() - vertex0()) * a13 ;
        edge1[PEdge::A] = mFace[P2];
        edge1[PEdge::B] = mFace[P2]; // v1 vertex 2
        edge2[PEdge::A] = mFace[P3];
        edge2[PEdge::B] = mFace[P1]; // v2 is located on edge 2 (I0-I2)

        returnVal       = -2; // the face is touched on one vertex and crosses one edge
    }
    else if ( tmp2 == curVal &&
              ( (tmp0 > curVal && tmp1 < curVal) ||
                (tmp0 < curVal && tmp1 > curVal) ) )
    {
        v1              = vertex2();
        v2              = (vertex0()) + (vertex1() - vertex0()) * a12;
        edge1[PEdge::A] = mFace[P3];
        edge1[PEdge::B] = mFace[P3]; // v1 vertex 3
        edge2[PEdge::A] = mFace[P1];
        edge2[PEdge::B] = mFace[P2]; // v2 is located on edge 2 (I0-I1)

        returnVal       = -2; // the face is touched on one vertex and crosses one edge
    }
    else
    {
        returnVal = 0; // the face is touched on one vertex only; not interpreted as a crossection
    }

    return returnVal;
}


P3DVector PFace::pointOnFace(double lambda, double mu) const
{
    return  P3DVector( vertex0( ) +
                       ( (vertex1() - vertex0()) * lambda +
                         (vertex2() - vertex0()) * mu ) );
}


int PFace::linetrisect(const P3DVector& p1, const P3DVector& p2, double& alpha, double& lambda, double& mu) const
{
#ifndef QT_NO_DEBUG
    if ( p1 == p2 )
    {
        qDebug("PFace::linetrisect p1 ==p2");
    }
#endif

    int retval = 0;

    if ( p1 != vertex0() && p1 != vertex1() && p1 != vertex2() &&
         p2 != vertex0() && p2 != vertex1() && p2 != vertex2() )
    {
        P3DVector   a1( vertex1() - vertex0() );
        P3DVector   a2( vertex2() - vertex0() );
        P3DVector   a3( p1 - p2 );
        double      det0 = deter( a1, a2, a3 );

        if ( fabs( det0 ) > EPS )
        {
            P3DVector   sav( p1 - vertex0() );
            double      det1 = deter( sav, a2,  a3  );
            double      det2 = deter( a1,  sav, a3  );
            double      det3 = deter( a1,  a2,  sav );

            lambda  = det1 / det0; // =fraction along edge 1
            mu      = det2 / det0; // =fraction along edge 2
            alpha   = det3 / det0; // =fraction along the line from l1 to l2

            if ( lambda >= 0.0 && lambda <= 1.0 &&
                 mu >=0.0 && mu <= 1.0 &&
                 lambda + mu <= 1.0 )
            {
                if ( alpha >= 0.0 && alpha <= 1.0 )
                {
                    retval = 1;
                }
            }

            if ( det0 > 0 )
            {
                retval = -1 * retval;
            }
        }
    }

    return retval;
}


bool PFace::linetrisect(const P3DVector& p1, const P3DVector& p2) const
{
    double  alpha;
    double  lambda;
    double  mu;
    bool    retval = false;

    if ( p1 != vertex0() && p1 != vertex1() && p1 != vertex2() &&
         p2 != vertex0() && p2 != vertex1() && p2 != vertex2() )
    {
        P3DVector   a1( vertex1() - vertex0() );
        P3DVector   a2( vertex2() - vertex0() );
        P3DVector   a3( p1 - p2 );
        double      det0 = deter( a1, a2, a3 );

        if ( fabs( det0 ) > EPS )
        {
            P3DVector   sav( p1 - vertex0() );
            double      det1 = deter( sav, a2,  a3  );
            double      det2 = deter( a1,  sav, a3  );
            double      det3 = deter( a1,  a2,  sav );

            lambda  = det1 / det0; // =fraction along edge 1
            mu      = det2 / det0; // =fraction along edge 2
            alpha   = det3 / det0; // =fraction along the line from l1 to l2

            if ( lambda >= 0.0 && lambda <= 1.0 &&
                 mu >=0.0 && mu <= 1.0 &&
                 lambda + mu <= 1.0 &&
                 alpha >= 0.0 && alpha <= 1.0 )
            {
                retval = true;
            }
        }
    }

    return retval;
}
