/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pface.h 1196 2014-04-15 10:20:27Z eelco $
 *
 */

#ifndef __PFACE_H
#define __PFACE_H

#include "P3DVector.h"
#include "pedge.h"
#include "pplane.h"
#include "pgeometry.h"
#include "pfunctions.h"

#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {

        class PGeometry;

        /** @class PFace pface.h "support/pface.h"
          * @brief This class hold a face and its functions
          * A face is triangle, part of a triangulated surafce. A face contains
          * contains the indexes to the vertices of a geometry an its
          * coordinates. Based on these member the determinant, solidagle, the
          * projection onto the triangle etc.... can be computed.
          * @see PGeometry
          */
        class PEACSSUPPORTSHARED_EXPORT PFace
        {

        public:


            /** @brief identifies the number of vertices within the face
              */
            static const int FACE_ITEM_COUNT = 3;


            /** @brief type definition of the number of items in PFace.
              */
            enum facecount_t { P1 = 0,
                               P2 = 1,
                               P3 = 2 };


            /** @brief Default constructor.
              * Initialze all vertices to NULL and the indexes to 0
              * @see mVertex0
              * @see mVertex1
              * @see mVertex2
              * @see mFace
              */
            PFace(const PGeometry &myGeom);


            /** @brief Copy constructor.
              * @see operator=(const PFace &face)
              */
            PFace( const PFace &f );


            /** @brief Assignment constructor.
              * set the 3 vertices of a face and its indexes.
              * Notice that the pointers should be set to
              * @param i1  index of vertex1 (v1)
              * @param i2  index of vertex2 (v2)
              * @param i3  index of vertex3 (v3)
              * @param v1  vertex1 (see i1)
              * @param v2  vertex2 (see i2)
              * @param v3  vertex3 (see i3)
              * @see PFace(PFace &f)
              */
            PFace( int i1, int i2, int i3,
                   const PGeometry &myGeom );


            /** @brief copy operator.
              * copy the indices and vertices from face
              * the face is reparented!!
              * @param face the face that is copied
              */
            const PFace &operator= ( const PFace& face );


            /** @brief copy operator.
              * copy only the date, the face ios not reparented
              * @param face the face that is copied
              */
            void setData(int i0, int i1, int i2);


            /** @brief ecchanges the new and old index numbers so.
              * index1 = index2 and index2 = index1;
              * @param index1
              * @param index2
              */
            void changeVertexNumber(int index1,int index2);


            /** @brief Returns the vertex index number at index position i.
              * @param i index number
              */
            inline int &operator[] ( facecount_t i ) { return mFace[i]; }


            /** @brief This is an overloaded member function,
              * provided for convenience.
              * @param i index number
              */
            inline const int &operator[] ( facecount_t i ) const { return mFace[i]; }


            /** @brief equal operator.
              * true if all indices are the same
              * @param face the one used to compare with
              * @see operator !=(PFace& f) const
              * @see mFace
              */
            bool operator== ( const PFace& face ) const;


            /** @brief unequal operator.
              * @returns true if one of the indices does not match
              * @param face: the one used to compare with
              * @see operator ==(PFace& f) const
              * @see mFace
              */
            bool operator!= ( const PFace& face ) const;


            /** @brief true if one one of the index number is equal to vertex
              * otherwise false
              * @param vertex: the number to be checked
              * @see mFace
              */
            bool contains( int vertex ) const;


            /** @brief true if one one of the index number is equal to vertex
              * otherwise false
              * @param vertex: the vertex to be checked
              * @see mFace
              */
            bool contains( const P3DVector & vertex) const;


            /** @brief returns the according vertex index
              */
            inline int vertexIndex( facecount_t i ) const { return mFace[i]; }


            /** @brief returns the index of the first vertex (clockwise)
              */
            inline int I0(void) const { return mFace[P1]; }


            /** @brief returns the index of the second vertex (clockwise)
              */
            inline int I1(void) const { return mFace[P2]; }


            /** @brief returns the index of the third vertex (clockwise)
              */
            inline int I2(void) const { return mFace[P3]; }


            /** @brief returns the normal of the triangle
              * @see calcNormal()
              */
            inline const P3DVector &normal(void) const { return mNormal; }


            /** @brief change the direction of the face to flip the normal direction
              */
            void invertNormal();


            /** @brief returns the center of gravity
              */
            inline P3DVector gravityCenter(void) const { return ( vertex0() + vertex1() + vertex2() ) / 3.0; }

#ifndef QT_NO_DEBUG


            /** @brief returns vertex 0
              */
            const P3DVector &vertex0(void) const;


            /** @brief returns vertex 1
              */
            const P3DVector &vertex1(void) const;


            /** @brief returns vertex 2
              */
            const P3DVector &vertex2(void) const;

#else
            /** @brief returns vertex 0
              */
            inline const P3DVector &vertex0(void) const { return mMyGeom.vertex(mFace[P1]); }


            /** @brief returns vertex 1
              */
            inline const P3DVector &vertex1(void) const { return mMyGeom.vertex(mFace[P2]); }


            /** @brief returns vertex 2
              */
            inline const P3DVector &vertex2(void) const { return mMyGeom.vertex(mFace[P3]); }


#endif


            /** @returns the angle at vertex 0
              */
            double angle0(void) const;


            /** @returns the angle at vertex 1
              */
            double angle1(void) const;


            /** @returns the angle at vertex 0
              */
            double angle2(void) const;


            /** @returns the minimum angle of the three angels of the face
              */
            double minAngle(void) const;


            /** @returns the maximum angle of the three angels of the face
              */
            double maxAngle(void) const;


            /** @returns the minimum length of the three edges of the face
              */
            double minEdgeLength(void) const;


            /** @returns the maximum length of the three edges of the face
              */
            double maxEdgeLength(void) const;



            /** @brief returns vertex with index i
              * @param i index of the vertex\
              */
            const P3DVector &vertex(facecount_t i) const;


            /** @brief returns the area of the triangle
              */
            inline double area(void) const { return mArea; }


            /** @brief returns the determinant of the triangle
              */
            double faceDeter(void) const { return deter( vertex0(), vertex1(), vertex2() ); }


            /** @brief returns true if pos is inside the box surounding the face
              * @param pos the pos that is check
              */
            bool inFaceSpace(const P3DVector &pos) const;


            /** @brief returns the angle of the triangle from the observation
              * point
              * @param observationPnt The observation point from which the solid angle is
              * computed
              */
            double solidAngle(const P3DVector &observationPnt) const;


            /** @brief Determine if the face crosse the plane, lies before or behind
              * @param crossPlane the plane against which this is checked
              * @returns 0 if the triangle lies before the plane
              *          1 if the triangle crosses the plane
              *          2 if the triangle lies in the plane
              *         -1 if the triangle lies behidn the plane
              */
            int facePlaneCrossing(const PPlane &crossPlane) const;


            /** @brief Computes the crosssection points of the face with a plane (crossPlane)
              * @param crossPlane The crossectional plane
              * @param v1 first crosssection point
              * @param v2 second crosssection point
              * @param edge1 the edge that is crossed and contains v1
              * @param edge2 the edge that is crossed and contains v2
              * @returns 0 if no crossection is found, all vertices lie before the plane
              *         -1 if no crossection is found, all vertices lie behind the plane
              *          1 if the plane crosses
              *          2 if the face lies in the plane
              */
            int crossSection(const PPlane &crossPlane, P3DVector& v1, P3DVector &v2, PEdge &edge1, PEdge &edge2) const;

            /** @brief TODO
              *
              */
            inline bool hasEdge(const PEdge& edge) const {
                                                            return ( edge.a() == mFace[P1] && edge.b() == mFace[P2]) ||
                                                                   ( edge.a() == mFace[P2] && edge.b() == mFace[P3]) ||
                                                                   ( edge.a() == mFace[P3] && edge.b() == mFace[P1]);
                                                         }



            /** @brief TODO
              * @param curVal
              * @param val1
              * @param val2
              * @param val3
              * @param v1
              * @param v2
              * @param edge1
              * @param edge2
              */
            int crossFaceFunctionVal( double curVal, double val1, double val2, double val3,
                                      P3DVector& v1, P3DVector &v2, PEdge &edge1, PEdge &edge2 ) const;


            /** @brief Determine the lambda and mu paramters for point r projected on the face
              * @param ver    point to  be projected
              * @param lambda ratio along the the first edge (mVertices[1] - mVertices[0])
              *              to the projection point
              * @param mu     ratio along the the first edge (mVertices[1] - mVertices[0])
              *              to the projection point
              * @param dif    The distance from the surface in the normal direction
              */
            int projectVerOnFace(const P3DVector &ver, double &lambda, double &mu, double &dif) const; //lmoutr


            /** @brief Computes the minimal distance betweeen point r and the face
             * @param ver    point to  be projected
             * @param verOnFace projection point
             * @param lambda ratio along the the first edge
             *              (mVertices[1]-mVertices[0]) to the projection point
             * @param mu     ratio along the the first edge
             *              (mVertices[1]-mVertices[0]) to the projection point
             * @return 1 if the vertex is projected from the outside 0 is not else -1
             * @see projectVerOnFace(const P3DVector &ver,double &lambda, double &mu);        */
            int projectVerOnFace(const P3DVector &ver, P3DVector &verOnFace) const;


            /** @brief Determine the distance for point r projected on the face
                @param ver    point to be projected
                @param verOnFace projection point
             *  @returns the distance        */
            double vertex2FaceDist(const P3DVector &ver, P3DVector &verOnFace) const;


            /** @brief Determine the distance for point r projected on the face
              * @param ver    point to be projected
              * @param verOnFace projection point
              * @param lambda ratio along the the first edge (mVertices[1]-mVertices[0]) to the projection point
              * @param mu     ratio along the the first edge (mVertices[1]-mVertices[0]) to the projection point
              * @returns the distance
              */
            double vertex2FaceDist(const P3DVector &ver, P3DVector &verOnFace, double &lambda, double &mu) const;


            /** @brief compute the point on the vertex given the lambda (along edge I0-I1) and
              * mu (along edge I0-I2)
              * @param lambda factor between 0 -1
              * @param mu factor between 0-1
              * @returns the computed point
              */
            P3DVector pointOnFace(double lambda, double mu) const;


            /** @brief distance beween point p and line segment l1-l2.
              *      In cp the point on l1-l2 closed to p is returned;
              *      in lab the fraction of the distance
              *      l1-cp in respect to l1-l2 is returned.
              * @param p point to be projected on l1-l2
              * @param l1 point on line
              * @param l2 point on line
              * @param verOnline 3D coordinate of the projection point
              * @param alpha  fraction of line from point to projection point on the line
              */
            double edgeDist(const P3DVector &p, const P3DVector &l1, const P3DVector &l2, P3DVector &verOnline, double &alpha) const;


            /**  @brief Detect wether the line segment between p1 and p2 and the face intersect.
              *  If so, the fraction
              *  from p1 to p2 of the section point is returned as alfa,
              * and the fractions from r1 to r2 and r3 are returned
              * as lab and mu. In any case, parallel is set.
              * If either the line is degenerate, the function returns
              * -1; if the triangle is, it returns -2.
              * @param p1 point on line
              * @param p2 point on line
              * @param alpha  fraction of line from point to projection point on the line
              * @param lambda ratio along the the first edge
              *              (mVertices[1]-mVertices[0]) to the projection point
              * @param mu     ratio along the the first edge
              *              (mVertices[1]-mVertices[0]) to the projection point
              * @returns 0 if the triangle lies before the plane
              *          1 if the triangle crosses the plane
              *          2 if the triangle lies in the plane
              *         -1 if the triangle lies behidn the plane
              */
            int linetrisect(const P3DVector &p1, const P3DVector &p2, double &alpha, double &lambda, double &mu) const;


            /** @brief TODO
              * @param p1
              * @param p2
              * @return true/false
              */
            bool linetrisect(const P3DVector &p1, const P3DVector &p2) const;


            /** @brief calculate the normal of the triangle
              * @see normal();
              */
            void updateMembers(void);

        private:

            /** @brief The face indices decribing the triangle
              */
            int mFace[FACE_ITEM_COUNT];


            /** @brief The first vertex of the triangle
              */
            const PGeometry &mMyGeom;


            /** @brief The normal (length 1) of the triangle
              */
            P3DVector mNormal;


            /** @brief The area of the triangle triangle
              */
            double mArea;


//            /** @brief The center of gravity of the face
//              */
//            P3DVector mGravityCenter;


//            /** @brief The determinant of the triangle
//              */
//            double mDeter;
        };
    }
}

#endif //__PFACE_H
