/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pfunctions.cpp 1222 2014-05-30 13:38:08Z peter $
 *
 */

#include <QDebug>
#include <QApplication>

#include "math.h"
#include "pfunctions.h"
#include "PIVector.h"
#include "pedge.h"
#include "pbox.h"
#include "pface.h"

using namespace Peacs::Support;


double epsilon(void)
{
    double e = 1.0;
    double d = 2.0;

    for ( ; d > 1.0; e /= 2 )
    {
        d = e + 1;
    }

    return 4.0 * e;
}


const double Peacs::Support::EPS = epsilon();


PEACSSUPPORTSHARED_EXPORT int icyc( int k, int n )
{
    int a = k % n;

    if ( a < 0 )
    {
        a = a + n;
    }

    return a;
}


PEACSSUPPORTSHARED_EXPORT double Peacs::Support::deter( const P3DVector &r1,
                                                        const P3DVector &r2,
                                                        const P3DVector &r3 )
{
    return r1.x() * ( r2.y() * r3.z() - r2.z() * r3.y() ) +
           r1.y() * ( r2.z() * r3.x() - r2.x() * r3.z() ) +
           r1.z() * ( r2.x() * r3.y() - r2.y() * r3.x() );
}


PEACSSUPPORTSHARED_EXPORT double Peacs::Support::compSolidAngle( const P3DVector &vertex0,
                                                                 const P3DVector &vertex1,
                                                                 const P3DVector &vertex2,
                                                                 const P3DVector &observationPnt )
{
    // Compute the solid angle of the face from the obesrvation point
    // @see van Oosterom, A. and J. Strackee. The Solid Angle of a Plane Triangle.
    // IEEE Transactions on Biomedical Engineering 30: 125-126, 1983
    // If one of the 3 vertices is zero the return value is 0.
    P3DVector   r1( vertex0 - observationPnt );
    P3DVector   r2( vertex1 - observationPnt );
    P3DVector   r3( vertex2 - observationPnt );
    double      solidAngle  = 0;
    double      nom         = r1.dot( r2.cross( r3 ) );

    if ( fabs( nom ) > EPS )
    {
        double n1   = r1.length();
        double n2   = r2.length();
        double n3   = r3.length();
        double den  = n1 * n2 * n3 +
                      r1.dot( r2 ) * n3 +
                      r2.dot( r3 ) * n1 +
                      r3.dot( r1 ) * n2;

        solidAngle = -2.0 * atan2( nom, den );
    }

    return solidAngle;
}


PEACSSUPPORTSHARED_EXPORT PMatrix Peacs::Support::lowpassma( const PMatrix &M, int window )
{
    if ( window  > 1 )
    {
        PMatrix LOW( M.Nr(), M.Nc() );
        int     winb        = floor(window / 2.0);
        int     k0          = 0;
        int     k1          = 0;
        int     kF          = 0;
        double  sum0        = 0;
        double  sum1        = 0;
        double   winWidth   = window;

        const double* mapData = M.data();      // Get the matrix memory
        double* lowData = LOW.data();          // Get the matrix memory

        for ( int sigI = 0, rowI = 0; sigI < M.Nr(); sigI++, rowI += M.Nc() )
        {
            sum0 = 0;
            sum1 = 0;
            const double* row    = &mapData[ rowI ];
            double*       lowrow = &lowData[ rowI ];

            sum1 = row[0] *(double)winb;

            kF = 0;

            for ( k1 = 0; k1 < winb && k1 < M.Nc(); k1++ )
            {
                sum1 += row[k1];
            }

            for ( k0 = 0; k0 < winb && k1 < M.Nc(); k0++, k1++, kF++ )
            {
                lowrow[kF]  = (sum1 - sum0) / winWidth;

                sum0 += row[0];
                sum1 += row[k1];
            }

            for ( k0 = 0; k1 < M.Nc(); k0++, k1++,kF++ )
            {
                lowrow[kF]  = (sum1 - sum0) / winWidth;

                sum0 += row[k0];
                sum1 += row[k1];
            }

            for ( ; kF < M.Nc(); k0++,kF++ )
            {
                sum0 += row[k0];
                sum1 += row[M.Nc() - 1];
                lowrow[kF]  = (sum1 - sum0) / winWidth;
            }
        }

        return LOW;
    }
    else
    {
        return M;
    }
}


PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::lowpassma( const PVector &V, int window )
{
    PVector LOW( V.N() );
    int     dWindow = window / 2 + 1;
    int     k0;
    int     k1;
    double  sum0;
    double  sum1;
    double  winWidth = window;

    for ( int sigI = 0; sigI < V.N(); sigI++ )
    {
        sum0 = 0;
        sum1 = 0;

        for ( k1 = 0; k1 < dWindow; k1++ )
        {
            sum1 += V.at( 0 );
        }

        for ( ; k1 < window && k1 < V.N(); k1++ )
        {
            sum1 += V.at( k1 );
        }

        for ( k0 = 0; k1 < V.N(); k0++, k1++ )
        {
            LOW[k0] = (sum1 - sum0) / winWidth;

            sum0 += V.at( k0 );
            sum1 += V.at( k1 );
        }

        for ( ; k0 < V.N(); k0++ )
        {
            sum0 = sum0 + V.at( k0 );
            sum1 = sum1 + V.at( V.N() - 1 );

            LOW[k0] = (sum1 - sum0) / winWidth;
        }
    }

    return LOW;
}


PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::rms( const PMatrix &M )
{
    int     i;
    int     ir;
    PVector rmsVec( M.Nc(), 0.0 );
    double  sqsum;

    for ( i = 0; i < rmsVec.N(); i++ )
    {
        sqsum = 0;

        for ( ir = 0; ir < M.Nr() * M.Nc(); ir += M.Nc() )
        {
            sqsum += M.at( ir + i ) * M.at( ir + i );
        }

        rmsVec[i] = sqrt( sqsum / (double)M.Nr() );
    }

    return rmsVec;
}


PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::stdCurve( const PMatrix &M )
{
    int     i;
    int     ir;
    PVector rmsVec( M.Nc(), 0.0 );
    PVector meanVec( M.Nc(), 0.0 );
    double  sqsum;
    double  sum;

    if ( M.Nr() >= 2 )
    {
        for ( i = 0; i < meanVec.N(); i++ )
        {
            sum = 0;

            for ( ir = 0; ir < M.Nr() * M.Nc(); ir += M.Nc() )
            {
                sum += M.at( ir + i );
            }

            meanVec[i] = sum / (double)M.Nr();
        }


        for ( i = 0; i < rmsVec.N(); i++ )
        {
            sqsum = 0;

            for ( ir = 0; ir < M.Nr() * M.Nc(); ir += M.Nc() )
            {
                double value = M.at( ir + i ) - meanVec[i];
                sqsum += value * value;
            }

            rmsVec[i] = sqrt( sqsum / (double)M.Nr() );
        }
    }
    else if ( M.Nr() == 1 )
    {
        for ( i = 0; i < rmsVec.N(); i++ )
        {
            rmsVec[i] = fabs( M.at( i ) );
        }
    }

    return rmsVec;
}

PEACSSUPPORTSHARED_EXPORT PMatrix Peacs::Support::diffRows( const PMatrix &M )
{
    PMatrix DIF( M.Nr(), M.Nc() );

    const double* mapData = M.data();      // Get the matrix memory
    double* difData = DIF.data();          // Get the matrix memory

    for ( int sigI = 0, rowI = 0; sigI < M.Nr(); sigI++, rowI += M.Nc() )
    {
        const double* row    = &mapData[ rowI ];
        double*       difrow = &difData[ rowI ];
        for ( int i=1; i < M.Nc(); i++)
        {
            difrow[i] =  row[i] - row[i-1];
        }
    }
    return DIF;
}



PEACSSUPPORTSHARED_EXPORT PMatrix Peacs::Support::ACFilter( const PMatrix &M )
{
    PMatrix ACFiltered( M );

    if ( ACFiltered.Nc() < 1000 )
    {
        int     rowI;
        int     sigI;
        int     i;
        double  sum;

        for ( sigI = 0, rowI = 0; sigI < ACFiltered.Nr(); sigI++, rowI+=ACFiltered.Nc() )
        {
            sum = 0;

            for ( i = 0; i < ACFiltered.Nc(); i++ )
            {
                sum += ACFiltered[rowI + i];
            }

            /* A heart beat of 60 beats per minute is assumed reuslting an an 1000 ms interval to avarage over*/
            sum += ACFiltered[rowI + ACFiltered.Nc()-1] *
                    (1000 - ACFiltered.Nc());
            sum /= qMax( 1000.0, (double)ACFiltered.Nc() );

            for ( i = 0; i < ACFiltered.Nc(); i++ )
            {
                ACFiltered[rowI + i] -= sum;
            }
        }
    }
    else
    {
        // TODO Peter
        int     rowI;
        int     sigI;
        int     i;
        double  sum;

        for ( sigI = 0, rowI = 0; sigI < ACFiltered.Nr(); sigI++, rowI += ACFiltered.Nc() )
        {
            sum = 0;

            for ( i = 0; i < ACFiltered.Nc(); i++ )
            {
                sum += ACFiltered[rowI + i];
            }

            /* A heart beat of 60 beats per minute is assumed reuslting an an 1000 ms interval to avarage over
             * In case the signals contains more than 1 sec of data the baseline is avaraged over more samples*/
            sum += ACFiltered[rowI + ACFiltered.Nc()-1] *
                    (1000 - ACFiltered.Nc());
            sum /= qMax( 1000.0, (double)ACFiltered.Nc() );

            for ( i = 0; i < ACFiltered.Nc(); i++ )
            {
                ACFiltered[rowI + i] -= sum;
            }
        }
    }

    return ACFiltered;
}



PEACSSUPPORTSHARED_EXPORT PMatrix Peacs::Support::baselineCorrection( const PMatrix &M,
                                                                      int begin,
                                                                      int end,
                                                                      int window )
{
    int     tbegin  = qBound( 0, begin, M.Nc() - 1 );
    int     tend    = end;
    int     twindow = qMax( window, 1 );
    int     k       = 0;
    double  slope;
    double  beginValue;
    double  endValue;
    PMatrix BASE( M );

    if ( end < 0 )
    {
        tend = M.Nc() - 1;
    }
    tend = qBound( 0, tend, M.Nc() - 1 );

    double * mapData = BASE.data();

    for ( int sigI = 0, rowI = 0;
          sigI < M.Nr();
          sigI++, rowI = rowI + M.Nc() )
    {
        if ( M.at( rowI + tbegin ) != 0 && M.at( rowI + tend ) != 0 )
        {
            beginValue  = 0;
            endValue    = 0;

            for ( k = 0; k < twindow; k++ )
            {
                beginValue  += M.at( rowI + k );
                endValue    += M.at( rowI + tend - k );
            }

            beginValue  /= twindow;
            endValue    /= twindow;

            slope = (endValue - beginValue) / (tend - tbegin);

            double * rowData = &mapData[rowI];
            for ( k = 0; k < M.Nc(); k++ )
            {
                rowData[k] -= (beginValue + slope * (k - tbegin));
            }
        }
    }
    return BASE;
}


PEACSSUPPORTSHARED_EXPORT PMatrix Peacs::Support::baselineCorrection( const PMatrix &M,
                                                                      const QList <int> &baselinePoints,
                                                                      int window )
{
    int     twindow = qMax( window, 1 );
    PMatrix BASE( M );


    double * mapData = BASE.data();
    const double * orgData = M.data();

    for ( int sigI = 0, rowI = 0;
          sigI < M.Nr();
          sigI++, rowI = rowI + M.Nc() )
    {
        double * rowData = &mapData[ rowI ];
        const double * orgRow=   &orgData[ rowI ];
        for ( int i=0 ; i < baselinePoints.size()-1; i++)
        {

            int     tbegin  = qBound( 0, baselinePoints[i], M.Nc() - 1 );
            int     tend    = qBound( 0, baselinePoints[i+1], M.Nc() - 1 );

            double beginValue  = 0;
            double endValue    = 0;
            for ( int k = 0; k < twindow; k++ )
            {
                beginValue  += orgRow[ tbegin + k ];
                endValue    += orgRow[ tend - k ];
            }
            beginValue  /= (double)twindow;
            endValue    /= (double)twindow;

            double slope = (endValue - beginValue) / double(tend - tbegin);

            for ( int k = tbegin + 1; k < tend; k++ )
            {
                rowData[k] -= (beginValue + slope * double (k - tbegin));
            }
            rowData[tbegin] = 0;
        }
    }
    return BASE;
}



PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::baselineCorrection( const PVector &V,
                                                                      int begin,
                                                                      int end,
                                                                      int window )
{
    int     tbegin  = qBound( 0, begin, V.N() - 1 );
    int     tend    = end;
    int     twindow = qMax( window, 1 );
    int     k;
    double  slope;
    double  beginValue  = 0;
    double  endValue    = 0;
    PVector BASE( V );

    if ( end < 0 )
    {
        tend = V.N() - 1;
    }
    tend = qBound( 0, tend, V.N() - 1 );

    for ( k = 0; k < twindow; k++ )
    {
        beginValue  += V.at( k );
        endValue    += V.at( tend - k );
    }

    beginValue  /= twindow;
    endValue    /= twindow;

    slope = (endValue - beginValue) / (tend - tbegin);

    for ( k = 0; k < V.N(); k++ )
    {
        BASE[k] -= (beginValue + slope * (k - tbegin));
    }

    return BASE;
}


/**************************************************************************************/
PEACSSUPPORTSHARED_EXPORT double Peacs::Support::calcScaleStep( double & minIn,
                                                                double & maxIn,
                                                                double isoLineStep,
                                                                bool up )
{
    double minVal       = minIn;
    double maxVal       = maxIn;
    double correction   = 1.0;
    double tstvalmin    = fabs(minVal * correction);
    double tstvalmax    = fabs(maxVal * correction);

    Q_ASSERT( minIn < maxIn );

    // When the max or min values are between 0 and 1 the step calculation suffers more from round-up error. To
    // prevent this we change these values to higher values during the calculation. Afterwards, the output is
    // corrected again for this temporary correction
    while ( (correction < maxIn - minIn) &&  // Limmit, just to prevent endless looping due to round-up errors in doubles
            ( ( (tstvalmin < 1.0) && (tstvalmin > 0.0) ) ||
              ( (tstvalmax < 1.0) && (tstvalmax > 0.0) ) ) )
    {
        correction  *= 10.0;
        tstvalmin   = fabs(minVal * correction);
        tstvalmax   = fabs(maxVal * correction);
    }

    minVal *= correction;
    maxVal *= correction;

    double step    = 1.0;
    double size    = maxVal - minVal;
    double minstep = 0.5 * size / (isoLineStep * correction);
    double maxstep = 2.0 * minstep;
    int    nstep   = qMax( 2, (int)size );  // The minimum of steps should be 2

    if ( nstep <= minstep )
    {
        while ( nstep <= minstep )
        {
            step /= 2.0;
            nstep = int(size / step);

            if ( nstep > minstep)
            {
                break;
            }

            step /= 2.500;

            nstep = int( size / step);

            if ( nstep > minstep )
            {
                break;
            }

            step /= 2.0;

            nstep = int( size / step );

            if ( nstep > minstep )
            {
                break;
            }
        }
    }
    else
    {
        while ( nstep > maxstep)
        {
            step *= 2.0;

            nstep = int( size / step );

            if ( nstep <= maxstep )
            {
                break;
            }

            step *= 2.5;

            nstep = int( size / step );

            if ( nstep <= maxstep )
            {
                break;
            }

            step *= 2.0;

            nstep = int( size / step );

            if ( nstep <= maxstep )
            {
                break;
            }
        }
    }

    if ( minVal == -maxVal )
    {
        if ( up )
        {
            minVal = floor( minVal / step ) * step;
        }
        else
        {
            minVal = ceil( minVal / step ) * step;
        }
        maxVal = -minVal;
    }
    else
    {
        if ( up )
        {
            minVal = floor( minVal / step ) * step;
            maxVal = ceil( maxVal / step ) * step;
        }
        else
        {
            minVal = ceil( minVal / step ) * step;
            maxVal = floor( maxVal / step ) * step;
        }
    }

    if ( minVal == maxVal )
    {
        maxVal = minVal + (2.0 * step);
    }

    minIn = minVal / correction;
    maxIn = maxVal / correction;

    return step / correction;
}


PEACSSUPPORTSHARED_EXPORT void Peacs::Support::computeScale( const PVector & centerDist,
                                                             double radius,
                                                             double slope,
                                                             PVector & scale )
{
    double val;
    double constDist = qMax( 0.0, radius - TRANSITIONZONE );

    if ( centerDist.N() != scale.N() )
    {
        scale = PVector( centerDist.N(), 0.0 );
    }

    for ( int i = 0; i < scale.N(); i++ )
    {
        if ( centerDist[i] < constDist ) // not equal otherwise the normalization does not work
        {
            val = 1.0 - 2.0 / (1.0 + exp( slope *(constDist - radius) ));
        }
        else
        {
            val = 1.0 - 2.0 / (1.0 + exp( slope * (centerDist[i] - radius) ));
        }

        if ( centerDist[i] <= radius && val > scale[i] )
        {
            scale[i] = val;
        }
    }

    scale.normalize();
}


PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::computelinearScale( const PVector & centerDist, double radius )
{
    PVector scale( centerDist.N(), 0.0 );
    double  R = qMax( radius, 1.0 );

    for ( int i = 0; i < scale.N(); i++ )
    {
        if ( centerDist[i] < radius ) // not equal otherwise the normalization does not work
        {
            scale[i] = 1.0 - centerDist[i] / R;
        }
    }

    return scale;
}


PEACSSUPPORTSHARED_EXPORT PMatrix Peacs::Support::shortestPathAlgo( const PMatrix &adj, bool directional, QProgressDialog * progressDialog)
{
    PMatrix         D( adj );
    int             nEdge  = adj.nonZeros();
    QVector <int>   edgeA( nEdge );
    QVector <int>   edgeB( nEdge );
    PVector edgeLength( nEdge );
    double          maxLength = 0;
    int             k = 0;

    const double * Adata( adj.data() );


    // create a list of all the edges
    for ( int i = 0; i < adj.Nr(); i++ )
    {
        const double * rowA( &Adata[ i * adj.Nc() ] );

        for (int j = 0; j < adj.Nc(); j++ )
        {
            if ( rowA[ j ] != 0 )
            {
                edgeA[k] = i;
                edgeB[k] = j;
                edgeLength[k] = rowA[ j ];

                k++;

                maxLength += rowA[j];
            }
        }
    }

    int             start;
    QVector <bool>  fixed( adj.Nr(), false );

    QVector <int>   keepEdgeA( edgeA );
    QVector <int>   keepEdgeB( edgeB );
    PVector         keepEdgeLength(edgeLength);

    double * Ddata( D.data() );

    for ( start = 0; start < D.Nr(); start++ )
    {
        int nfix = 0;
        PVector         distance( D.Nc(), maxLength );


        if ( !directional )
        {
            for ( int i = 0; i < fixed.size(); i++ )
            {
                if ( i <= start )
                {
                    fixed[i] = true;
                    distance [ i ] = D(i, start);
                    nfix++;
                }
                else
                {
                    fixed[i] = false;
                }
            }
        }
        else
        {
            fixed.fill(false);
            fixed[start] = true;
        }

        distance[start] = 0;
        int npnt = distance.N();

        /* calculate all paths between startpnt and not fixed vertices*/
        while ( nfix < npnt )
        {
            double length = maxLength - 1.0;

            k = 0;

            for ( register int i = 0; i < nEdge; i++ )
            {
                if ( !fixed.at( edgeB[i] ) )
                {
                    if ( fixed.at( edgeA[i] ) )
                    {
                        distance[edgeB[i]]  = qMin( distance[ edgeB[i] ],
                                                    distance[ edgeA[i] ] + edgeLength.at( i ) );
                        length              = qMin( length, distance[ edgeB[i] ] );
                    }
                }
            }
            nfix    = 0;

            if ( length < maxLength - 2.0 )
            {
                for ( int i = 0; i < npnt; i++ )
                {
                    if ( distance[i] <= length )
                    {
                        fixed[i] = true;

                        nfix++;
                    }
                }
            }
            else
            {
                // incomplete graph, some parts are not connected
                for ( int i = 0; i < npnt; i++ )
                {
                    if ( distance[i] > maxLength -1.0 )
                    {
                        distance[i] = -1.0;
                        nfix++;
                    }
                }
                break;
            }
        }

        double *rowD( &Ddata[ start * D.Nc() ] );

        for ( int i = 0; i < distance.size(); i++ )
        {
            rowD[ i ] =  distance[i];
        }

        if ( !directional )
        {
            int prevNEdge =nEdge;
            nEdge = 0;

            for ( int i = 0; i < prevNEdge; i++ )
            {
                if (  keepEdgeA[i] >= start ||
                      keepEdgeB[i] >= start )
                {
                    edgeA[nEdge] = keepEdgeA[i];
                    edgeB[nEdge] = keepEdgeB[i];
                    edgeLength[nEdge] = keepEdgeLength[i];
                    nEdge++;
                }
            }
            keepEdgeA = edgeA;
            keepEdgeB = edgeB;
            keepEdgeLength = edgeLength;
        }

        if ( progressDialog )
        {
            progressDialog->setValue( (100 * start) / npnt );
            QApplication::processEvents();
            if ( progressDialog->wasCanceled() )
            {
                D.clear();

                break;
            }
        }
    }

    return D;
}




PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::shortestDist( const PGeometry &geom, int start )
{
    if ( geom.isEmpty() || start < 0 )
    {
        qDebug().nospace() << "shortestDist: invalid paramters" << start;

        return PVector();
    }

    int             k = 0;
    int             npnt = geom.nVertices();
    double          maxLength = 0;
    QVector <int>   edgeA( geom.nFaces() * 3 );         // create a list of all the edges
    QVector <int>   edgeB( geom.nFaces() * 3 );
    QVector <double> edgeLength( geom.nFaces() * 3 );

    for (int t = 0; t < geom.nFaces(); t++ )
    {
        edgeA[k]        = geom.face( t ).I0();
        edgeB[k]        = geom.face( t ).I1();
        edgeLength[k]   = ( geom.vertex( geom.face( t ).I0() ) - geom.vertex( geom.face( t ).I1() ) ).length();
        maxLength       += edgeLength[k];

        k++;

        edgeA[k]        = geom.face( t ).I1();
        edgeB[k]        = geom.face( t ).I2();
        edgeLength[k]   = ( geom.vertex( geom.face( t ).I1() ) - geom.vertex( geom.face( t ).I2() ) ).length();
        maxLength       += edgeLength[k];

        k++;

        edgeA[k]        = geom.face( t ).I2();
        edgeB[k]        = geom.face( t ).I0();
        edgeLength[k]   = ( geom.vertex( geom.face( t ).I2() ) - geom.vertex( geom.face( t ).I0() ) ).length();
        maxLength       += edgeLength[k];

        k++;
    }

    PVector         distance( npnt, maxLength );
    QVector <bool>  fixed( npnt, false );
    int             nfix    = 1;
    int             nEdge   = edgeA.size();

    fixed[start]    = true;
    distance[start] = 0;

    /* calculate all paths between startpnt and vertices*/
    while ( nfix < npnt )
    {
        double length = maxLength - 1.0;

        k = 0;

        for ( register int i = 0; i < nEdge; i++ )
        {
            if ( !fixed.at( edgeB[i] ) )
            {
                if ( fixed.at( edgeA[i] ) )
                {
                    distance[edgeB[i]]  = qMin( distance[ edgeB[i] ],
                                                distance[ edgeA[i] ] + edgeLength.at( i ) );
                    length              = qMin( length, distance[ edgeB[i] ] );
                }

                if ( k != i )
                {
                    edgeA[k] = edgeA[i];
                    edgeB[k] = edgeB[i];
                    edgeLength[k] = edgeLength[i];
                }

                k++;
            }
        }
        nEdge   = k;
        nfix    = 0;

        if ( length != maxLength - 1.0 )
        {
            for ( int i = 0; i < npnt; i++ )
            {
                if ( distance[i] <= length )
                {
                    fixed[i] = true;

                    nfix++;
                }
            }
        }
        else
        {
            for ( int i = 0; i < npnt; i++ )
            {
                if ( distance[i] > maxLength -1.0 )
                {
                    distance[i] = -1.0;
                    nfix++;
                }
            }
            break;
        }
    }

    return distance;
}


PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::fociPaths( const PMatrix &adj,
                                                             const QList <int> &foci,
                                                             const QList <double> &initTimes )
{
    if ( foci.isEmpty() )
    {
        qDebug() << "fociPaths: no foci defined";
    }

    int             npnt = adj.Nr();
    QVector <int>   edgeA( adj.nonZeros() / 2 );
    QVector <int>   edgeB( adj.nonZeros() / 2 );
    double          maxLength = 0;
    int             k = 0;

    /* create a list of all the edges*/
    for ( int i = 0, ir = 0; i < adj.Nr(); i++, ir += adj.Nc() )
    {
        for ( int j = i; j < adj.Nc(); j++ )
        {
            if ( adj.at( ir + j ) > 0.0 )
            {
                if ( k > edgeA.size() )
                {
                    edgeA.append( i );
                    edgeB.append( j );
                }
                else
                {
                    edgeA[k] = i;
                    edgeB[k] = j;
                }

                k++;

                maxLength = maxLength + adj.at( ir + j );
            }
        }
    }

    int             nfix    = 1;
    int             nEdges  = k;
    QVector <bool>  fixed( npnt, false );
    PVector         distance( npnt, maxLength );

    for ( int i = 0 ; i < foci.size(); i++ )
    {
        distance[ foci[i] ] = initTimes[i];
    }

    // only fix the initial focus. The distance of the other foci has been set but can be overruled by another focus
    fixed[ distance.getMinI() ] = true;

    /* calculate all paths between startpnt and vertices*/
    while ( nfix < npnt )
    {
        double length = maxLength - 1.0;

        k = 0;

        for ( register int i = 0; i < nEdges; i++ )
        {
            if ( !fixed.at( edgeB[i] ) )
            {
                if ( fixed.at( edgeA[i] ) )
                {
                    // Only the first node is fixed, but the distance at other nodes might be shorter
                    // because they are a focus too.
                    double thisDist = qMin(distance[ edgeB[i] ], distance[ edgeA[i] ] + adj[ edgeA[i] + edgeB[i] * npnt ]);
                    if ( thisDist < length )
                    {
                        length                  = thisDist;
                        distance[ edgeB[i] ]    = thisDist;
                    }

//                    distance[edgeB[i]] = qMin( distance[ edgeB[i] ],
//                                               distance[ edgeA[i] ] + adj[ edgeA[i] + edgeB[i] * npnt ] );
//                    length = qMin( length, distance[ edgeB[i] ] );
                }

                if ( k != i )
                {
                    edgeA[k] = edgeA[i];
                    edgeB[k] = edgeB[i];
                }

                k++;
            }
            else if ( !fixed.at( edgeA[i] ) )
            {
                if ( fixed.at( edgeB[i] ) )
                {
                    // Only the first node is fixed, but the distance at other nodes might be shorter
                    // because they are a focus too.
                    double thisDist = qMin( distance[ edgeA[i] ], distance[ edgeB[i] ] + adj[ edgeB[i] + edgeA[i] * npnt ] );
                    if ( thisDist < length )
                    {
                        length                  = thisDist;
                        distance[ edgeA[i] ]    = thisDist;
                    }
                }

                if ( k != i )
                {
                    edgeA[k] = edgeA[i];
                    edgeB[k] = edgeB[i];
                }
                k++;
            }
        }

        nfix    = 0;
        nEdges  = k;
        if ( length != maxLength - 1.0 )
        {
            for ( int i = 0; i < npnt; i++ )
            {
                if ( distance[i] <= length )
                {
                    fixed[i] = true;

                    nfix++;
                }
            }
        }
        else
        {
            for ( int i = 0; i < npnt; i++ )
            {
                if ( distance[i] > maxLength -1.0 )
                {
                    distance[i] = -1.0;
                    nfix++;
                }
            }
            break;
        }
    }

    return distance;
}


PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::fociPathsDirectional( const PMatrix &adj,
                                                                        const QList <int> &foci,
                                                                        const QList <double> &initTimes )
{
    if ( foci.isEmpty() )
    {
        qDebug() << "fociPaths: no foci defined";
    }

    int             npnt        = adj.Nr();
    int             k           = 0;
    double          maxLength   = 0;
    QVector <int>   edgeA( adj.nonZeros() );
    QVector <int>   edgeB( adj.nonZeros() );

    /* create a list of all the edges*/
    for ( int i = 0, ir = 0; i < adj.Nr(); i++, ir += adj.Nc() )
    {
        for ( int j = 0; j < adj.Nc(); j++ )
        {
            if ( adj.at( ir + j ) > 0.0 )
            {
                edgeA[k]    = i;
                edgeB[k]    = j;
                maxLength   = maxLength + adj.at( ir + j );

                k++;
            }
        }
    }

    PVector distance( npnt, maxLength );

    for ( int i = 0 ; i < foci.size(); i++ )
    {
        distance[ foci[i] ] = initTimes[i];
    }

    int             nfix    = 1;
    int             nEdges  = k;
    QVector <bool>  fixed( npnt, false );

    // only fix the initial focus. The distance of the other foci has been set but can be overruled by another focus
    fixed[ distance.getMinI() ] = true;

    /* calculate all paths between startpnt and vertices*/
    while ( nfix < npnt )
    {
        double length = maxLength - 1.0;

        k = 0;

        for ( register int i = 0; i < nEdges; i++ )
        {
            if ( !fixed.at( edgeB[i] ) )
            {
                if ( fixed.at( edgeA[i] ) )
                {
                    // Only the first node is fixed, but the distance at other nodes might be shorter
                    // because they are a focus too.
                    double thisDist = qMin(distance[ edgeB[i] ], distance[ edgeA[i] ] + adj[ edgeA[i] + edgeB[i] * npnt ]);
                    if ( thisDist < length )
                    {
                        length                  = thisDist;
                        distance[ edgeB[i] ]    = thisDist;
                    }

//                    distance[edgeB[i]] = qMin( distance[ edgeB[i] ],
//                                               distance[ edgeA[i] ] + adj[ edgeA[i] + edgeB[i] * npnt ] );
//                    length = qMin( length, distance[ edgeB[i] ] );



                }

                if ( k != i )
                {
                    edgeA[k] = edgeA[i];
                    edgeB[k] = edgeB[i];
                }

                k++;
            }
        }

        nEdges  = k;
        nfix    = 0;

        if ( length != maxLength - 1.0 )
        {
            for ( int i = 0; i < npnt; i++ )
            {
                if ( distance[i] <= length )
                {
                    fixed[i] = true;

                    nfix++;
                }
            }
        }
        else
        {
            for ( int i = 0; i < npnt; i++ )
            {
                if ( distance[i] > maxLength -1.0 )
                {
                    distance[i] = -1.0;
                    nfix++;
                }
            }
            break;
        }
    }

    return distance;
}

//PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::fociPathsDirectional( const PMatrix & adj,
//                                                                        const QList <int> & foci,
//                                                                        const QList<double> & initTimes )
//{
//    Q_ASSERT( foci.size() == initTimes.size() );

//    if ( foci.isEmpty() )
//    {
//        qDebug() << "fociPaths: no foci defined";
//    }

//    int             i;
//    int             ir;
//    int             npnt        = adj.Nr();
//    int             nfix        = 1;
//    int             nEdges      = 0;
//    int             k           = 0;
//    double          maxLength   = 0;
//    int             zerosCount  = adj.nonZeros();
//    bool *          fixedPtr    = 0;
//    int *           edgeAPtr    = new int[ zerosCount ];
//    int *           edgeBPtr    = new int[ zerosCount ];
//    const double *  adjData     = adj.data();

//    /* create a list of all the edges*/
//    for ( i = 0, ir = 0; i < adj.Nr(); i++, ir += adj.Nc() )
//    {
//        for ( int j = 0; j < adj.Nc(); j++ )
//        {
//            if ( adjData[ ir + j ] > 0.0 )
//            {
//                edgeAPtr[k] = i;
//                edgeBPtr[k] = j;
//                maxLength   += adjData[ ir + j ];

//                k++;
//            }
//        }
//    }

//    PVector     distance( npnt, maxLength );
//    double *    distancePtr = distance.data();

//    for ( i = 0 ; i < foci.size(); i++ )
//    {
//        distancePtr[ foci[i] ] = initTimes[i];
//    }

//    nEdges  = k;
//    fixedPtr = new bool[ npnt ];

//    for ( i = 0; i < npnt; ++i )
//    {
//        fixedPtr[i] = false;
//    }

//    // only fix the initial focus. The distance of the other foci has been set but can be overruled by another focus
//    fixedPtr[ distance.getMinI() ] = true;

//    /* calculate all paths between startpnt and vertices*/
//    while ( nfix < npnt )
//    {
//        double length = maxLength - 1.0;

//        k = 0;

//        for ( i = 0; i < nEdges; i++ )
//        {
//            if ( !fixedPtr[ edgeBPtr[i] ] )
//            {
//                if ( fixedPtr[ edgeAPtr[i] ] )
//                {
//                    // Only the first node is fixed, but the distance at other nodes might be shorter
//                    // because they are a focus too.
//                    double thisDist = qMin(distancePtr[ edgeBPtr[i] ],
//                                           distancePtr[ edgeAPtr[i] ] + adjData[ edgeAPtr[i] + edgeBPtr[i] * npnt ]);
//                    if ( thisDist < length )
//                    {
//                        length                       = thisDist;
//                        distancePtr[ edgeBPtr[ i ] ] = thisDist;
//                    }

////                    distancePtr[ edgeBPtr[ i ] ] = qMin( distancePtr[ edgeBPtr[i] ],
////                                                         distancePtr[ edgeAPtr[i] ] + adjData[ edgeAPtr[i] + edgeBPtr[i] * npnt ] );
////                    length = qMin( length, distancePtr[ edgeBPtr[i] ] );
//                }

//                if ( k != i )
//                {
//                    edgeAPtr[k] = edgeAPtr[i];
//                    edgeBPtr[k] = edgeBPtr[i];
//                }

//                k++;
//            }
//        }

//        nEdges  = k;
//        nfix    = 0;

//        if ( length != maxLength - 1.0 )
//        {
//            for ( i = 0; i < npnt; i++ )
//            {
//                if ( distancePtr[i] <= length )
//                {
//                    fixedPtr[ i ] = true;

//                    nfix++;
//                }
//            }
//        }
//        else
//        {
//            for ( i = 0; i < npnt; i++ )
//            {
//                if ( distancePtr[i] > maxLength -1.0 )
//                {
//                    distancePtr[i] = -1.0;
//                    nfix++;
//                }
//            }
//            break;
//        }
//    }

//    delete [] fixedPtr;
//    delete [] edgeAPtr;
//    delete [] edgeBPtr;

//    return distance;
//}


PEACSSUPPORTSHARED_EXPORT QList <int> Peacs::Support::getRoute( const PGeometry &geom,
                                                                int start,
                                                                int stop )
{
    if ( start < 0 || stop < 0 || geom.isEmpty() )
    {
        qDebug().nospace() << "getRoute (geometry): invalid paramters: " << start << stop;

        return QList <int> ();
    }

    int                 k           = 0;
    int                 npnt        = geom.nVertices();
    double              maxLength   = 0;
    QVector <int>       edgeA( geom.nFaces() * 3 );   // create a list of all the edges
    QVector <int>       edgeB( geom.nFaces() * 3 );
    QVector <double>    edgeLength( geom.nFaces() * 3 );
    QList <int>         route;

    for ( int t = 0; t < geom.nFaces(); t++ )
    {
        edgeA[k]        = geom.face( t ).I0();
        edgeB[k]        = geom.face( t ).I1();
        edgeLength[k]   = ( geom.vertex( geom.face( t ).I0() ) - geom.vertex( geom.face( t ).I1() ) ).length();
        maxLength       += edgeLength[k];

        k++;

        edgeA[k]        = geom.face( t ).I1();
        edgeB[k]        = geom.face( t ).I2();
        edgeLength[k]   = ( geom.vertex( geom.face( t ).I1() ) - geom.vertex( geom.face( t ).I2() ) ).length();
        maxLength       += edgeLength[k];

        k++;

        edgeA[k]        = geom.face( t ).I2();
        edgeB[k]        = geom.face( t ).I0();
        edgeLength[k]   = ( geom.vertex( geom.face( t ).I2() ) - geom.vertex( geom.face( t ).I0() ) ).length();
        maxLength       += edgeLength[k];

        k++;
    }

    int             nfix    = 1;
    int             nEdge   = edgeA.size();
    PVector         distance( npnt, maxLength );
    PIVector        fromVertex( npnt );
    QVector <bool>  fixed( npnt, false );

    fixed[start]    = true;
    distance[start] = 0;

    /* calculate all paths between startpnt and vertices*/
    while ( nfix < npnt )
    {
        double length = maxLength - 1.0;

        k = 0;

        for ( register int i = 0; i < nEdge ; i++ )
        {
            if ( !fixed.at( edgeB[i] ) )
            {
                if ( fixed.at( edgeA[i] ) )
                {
                    double tmpLength = distance[edgeA[i]] + edgeLength[i];

                    if ( tmpLength <= distance[edgeB[i]] )
                    {
                        distance[ edgeB[i] ]    = tmpLength;
                        fromVertex[ edgeB[i] ]  = edgeA[i];
                    }

                    length = qMin( length, distance[ edgeB[i] ] );
                }

                if ( k != i )
                {
                    edgeA[k]        = edgeA[i];
                    edgeB[k]        = edgeB[i];
                    edgeLength[k]   = edgeLength[i];
                }

                k++;
            }
        }

        nEdge   = k;
        nfix    = 0;

        for ( int i = 0; i < npnt; i++ )
        {
            if ( distance[i] <= length )
            {
                fixed[i] = true;

                nfix++;
            }
        }

        if ( fixed[stop] )
        {
            break;
        }
    }

    route.append( stop );

    k = 0;

    while ( route.first() != start &&
            k < geom.nVertices() )
    {
        route.prepend( fromVertex[ route.first() ] );

        k++;
    }

    return route;
}


PEACSSUPPORTSHARED_EXPORT QList <int> Peacs::Support::getRoute( const PMatrix &adj,
                                                                int start,
                                                                int stop )
{
    if ( start < 0 || stop < 0 || adj.isEmpty() )
    {
        qDebug().nospace() << "getRoute(adj): invalid paramters: " << start << stop;

        return QList <int> ();
    }

    int             k = 0;
    int             npnt        = adj.Nc();
    double          maxLength   = 0;
    int             ne          = adj.nonZeros();    // create a list of all the edges
    QVector <int>   edgeA( ne );
    QVector <int>   edgeB( ne );
    QList <int>     route;

    /* create a list of all the edges*/
    for ( int i = 0, ir = 0; i < adj.Nr(); i++, ir += adj.Nc() )
    {
        for ( int j = 0; j < adj.Nc(); j++ )
        {
            if ( adj.at( ir + j ) > 0 )
            {
                edgeA[k]    = i;
                edgeB[k]    = j;
                maxLength   += adj.at( ir + j );

                k++;
            }
        }
    }

    int             nfix    = 1;
    int             nEdge   = edgeA.size();
    PVector         distance( npnt, maxLength );
    PIVector        fromVertex( npnt );
    QVector <bool>  fixed( npnt, false );

    fixed[start]    = true;
    distance[start] = 0;

    /* calculate all paths between startpnt and vertices*/
    while ( nfix < npnt )
    {
        double length = maxLength - 1.0;

        k = 0;

        for ( register int i = 0; i < nEdge; i++ )
        {
            if ( !fixed.at( edgeB[i] ) )
            {
                if ( fixed.at( edgeA[i] ) )
                {
                    double tmpLength = distance[ edgeA[i] ] + adj.at( edgeA[i] + edgeB[i] * npnt );

                    if ( tmpLength <= distance[ edgeB[i] ] )
                    {
                        distance[ edgeB[i] ]    = tmpLength;
                        fromVertex[ edgeB[i] ]  = edgeA[i];
                    }

                    length = qMin( length, distance[ edgeB[i] ] );
                }

                if ( k != i )
                {
                    edgeA[k] = edgeA[i];
                    edgeB[k] = edgeB[i];
                }

                k++;
            }
        }

        nEdge   = k;
        nfix    = 0;

        for ( int i = 0; i < npnt; i++ )
        {
            if ( distance[i] <= length )
            {
                fixed[i] = true;

                nfix++;
            }
        }

        if ( fixed[stop] )
        {
            break;
        }
    }

    route.append( stop );

    k = 0;

    while ( route.first() != start &&
            k < npnt )
    {
        route.prepend( fromVertex[ route.first() ] );

        k++;
    }

    return route;
}


PEACSSUPPORTSHARED_EXPORT P3DVector Peacs::Support::pMax( const P3DVertices &ver )
{
    P3DVector val( -HUGE_VAL, -HUGE_VAL, -HUGE_VAL );

    for ( int i = 0; i < ver.size(); i++ )
    {
        val.x() = qMax( val.x(), ver.at( i ).x() );
        val.y() = qMax( val.y(), ver.at( i ).y() );
        val.z() = qMax( val.z(), ver.at( i ).z() );
    }

    return val;
}


PEACSSUPPORTSHARED_EXPORT P3DVector Peacs::Support::pMin( const P3DVertices &ver )
{
    P3DVector val( HUGE_VAL, HUGE_VAL, HUGE_VAL);

    for ( int i = 0; i < ver.size(); i++ )
    {
        val.x() = qMin( val.x(), ver.at( i ).x() );
        val.y() = qMin( val.y(), ver.at( i ).y() );
        val.z() = qMin( val.z(), ver.at( i ).z() );
    }

    return val;
}


PEACSSUPPORTSHARED_EXPORT P3DVector Peacs::Support::pMax( const P3DVector &ver1, const P3DVector &ver2 )
{
    return P3DVector( qMax( ver1.x(), ver2.x() ),
                      qMax( ver1.y(), ver2.y() ),
                      qMax( ver1.z(), ver2.z() ) );
}


PEACSSUPPORTSHARED_EXPORT P3DVector Peacs::Support::pMin( const P3DVector &ver1, const P3DVector &ver2 )
{
    return P3DVector( qMin( ver1.x(), ver2.x() ),
                      qMin( ver1.y(), ver2.y() ),
                      qMin( ver1.z(), ver2.z() ) );
}



PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::pMax( const PVector &v1, const PVector &v2 )
{
    PVector val( v1 );
    if ( v1.N() == v2.N() )
    {
        for ( int i = 0; i < val.N(); i++ )
        {
            val[i] = qMax( val[i], v2[i] );
        }
    }
    return val;
}


PEACSSUPPORTSHARED_EXPORT PVector Peacs::Support::pMin( const PVector &v1, const PVector &v2 )
{
    PVector val( v1 );
    if ( v1.N() == v2.N() )
    {
        for ( int i = 0; i < val.N(); i++ )
        {
            val[i] = qMin( val[i], v2[i] );
        }
    }
    return val;
}


PEACSSUPPORTSHARED_EXPORT double Peacs::Support::pearsonCorrelation(const PMatrix &X,const PMatrix &Y)
{
    //        (C) Copr. 1986-92 Numerical Recipes Software .)*^'K5.){2p49.
    double correlation = 0.0;
    if ( X.Nc() == Y.Nc() &&
         X.Nr() == Y.Nr() )
    {
        int n = X.Nc() * X.Nr();
        double meanX = 0.0;
        double meanY = 0.0;
        for ( int i= 0; i < n; i++)
        {
            meanX += X[i];
            meanY += Y[i];
        }
        meanX /= (double)n;
        meanY /= (double)n;

        double syy=0.0,sxy=0.0,sxx=0.0;
        for ( int i= 0; i < n; i++)
        {
            double xt = X[i] - meanX;
            double yt = Y[i] - meanY;
            sxx += xt * xt;
            syy += yt * yt;
            sxy += xt * yt;
        }
        correlation = sxy / sqrt( sxx * syy );
    }
    return correlation;
}

PEACSSUPPORTSHARED_EXPORT double Peacs::Support::meanRowPearsonCorrelation(const PMatrix &X,const PMatrix &Y)
{
    //        (C) Copr. 1986-92 Numerical Recipes Software .)*^'K5.){2p49.
    double correlation = 0.0;
    if ( X.Nc() == Y.Nc() &&
         X.Nr() == Y.Nr() )
    {
        const double * XData = X.data();
        const double * YData = Y.data();

        for ( int ir=0; ir < X.Nr(); ir++)
        {
            const double * XrowData = &XData[ir*X.Nc()];
            const double * YrowData = &YData[ir*X.Nc()];


            int n = X.Nc();
            double meanX = 0.0;
            double meanY = 0.0;
            for ( int i= 0; i < X.Nc(); i++)
            {
                meanX += XrowData[i];
                meanY += YrowData[i];
            }
            meanX /= (double)n;
            meanY /= (double)n;

            double syy=0.0,sxy=0.0,sxx=0.0;
            for ( int i= 0; i < n; i++)
            {
                double xt = XrowData[i] - meanX;
                double yt = YrowData[i] - meanY;
                sxx += xt * xt;
                syy += yt * yt;
                sxy += xt * yt;
            }
            correlation += (sxy / sqrt( sxx * syy ));
        }
        correlation /= (double)X.Nr();
    }
    return correlation;
}


