/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pfunctions.h 1222 2014-05-30 13:38:08Z peter $
 *
 */


#ifndef __PFUNCTIONS_H_
#define __PFUNCTIONS_H_

#include <QList>
#include <QProgressDialog>

#include "pmatrix.h"
#include "pvector.h"
#include "PIVector.h"
#include "pgeometry.h"
#include "P3DVector.h"
#include "P3DVertices.h"
#include "pCoordinateSystem.h"

#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {

        /** @brief the maximal width of the transition zone (150 mm)
          */
        const double TRANSITIONZONE = 150.0;//mm


        /** @brief This represents the calculated epsilon value
          * This is a system dependent epsilon determination because it is calculated during startup
          */
        extern const double EPS;


        /** @brief system dependent epssilon determination
          */
        //PEACSSUPPORTSHARED_EXPORT double epsilon(void);


        /** @brief cyclic counter
          */
        PEACSSUPPORTSHARED_EXPORT int icyc( int k, int n );

        /** @brief check is value is not a number
          */
//        inline PEACSSUPPORTSHARED_EXPORT bool isnan(double value) { return value != value; }


        /** @brief returns the determinant of 3 vectors
          * @param r1 vector used to compute the determinant
          * @param r2 vector used to compute the determinant
          * @param r3 vector used to compute the determinant
          */
        PEACSSUPPORTSHARED_EXPORT double deter( const P3DVector &r1, const P3DVector &r2, const P3DVector &r3 );


        /** @brief returns the angle of the triangle from the observation
          * point
          * @param observationPnt The observation point from which the solid angle is
          * computed
          */
        PEACSSUPPORTSHARED_EXPORT double compSolidAngle( const P3DVector &vertex0,
                                                         const P3DVector &vertex1,
                                                         const P3DVector &vertex2,
                                                         const P3DVector &observationPnt );


        /** @brief lowpass Moving Average filter applied to rows of matrix M;
          * Finite impulse response; no phase shift zero output for frequencies at
          * multiples of f = win * sampling_interval for highpass:
          * use PSI = PSI - lowpassma( M, win ) (matlab code)
          * 2005-02-17; V. Jacquemet / A. van Oosterom
          * 2008-02-12; AvO winb=round( changed into:
          *     winb = floor( LOW = M;
          *     if win < 2, return, end
          *     trans = 0;
          *
          * @param M
          * @param win
          * @return Matrix
          */
        PEACSSUPPORTSHARED_EXPORT PMatrix lowpassma( const PMatrix &M, int win );


        /** @brief lowpass Moving Average filter applied to rows of vector V;
          *
          * @param V
          * @param win
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector lowpassma( const PVector &V, int win );


        /** @brief TODO
          *
          * @param M Matrix
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector rms( const PMatrix &M );


        /** @brief TODO
          *
          * @param M Matrix
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PMatrix diffRows( const PMatrix &M );




        /** @brief TODO
          *
          * @param M Matrix
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector stdCurve( const PMatrix &M );


        /** @brief TODO
          *
          * @param M Matrix
          * @return Matrix
          */
        PEACSSUPPORTSHARED_EXPORT PMatrix ACFilter( const PMatrix &M );


        /** @brief TODO
          *
          * @param M Matrix
          * @param begin (default -1)
          * @param end (default -1)
          * @param window (default -1)
          * @return Matrix
          */
        PEACSSUPPORTSHARED_EXPORT PMatrix baselineCorrection( const PMatrix &M, int begin = 0, int end = -1, int window = -1 );


        /** @brief TODO
          *
          * @param M Matrix
          * @param begin (default -1)
          * @param end (default -1)
          * @param window (default -1)
          * @return Matrix
          */
        PEACSSUPPORTSHARED_EXPORT PMatrix baselineCorrection( const PMatrix &M, const QList <int> &baselinePoints, int window = 1 );



        /** @brief TODO
          *
          * @param V Vector
          * @param begin (default -1)
          * @param end (default -1)
          * @param window (default -1)
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector baselineCorrection(const PVector &V,int begin=-1,int end=-1,int window=-1);


        /** @brief Compute the scale steps for a given range
          *
          * @param minIn The minimal scale value to use. Returns a corrected, nicer minimal value
          * @param maxIn The maximal scale value to use. Returns a corrected, nicer maximal value
          * @param isoLineStep Gives the number of steps wanted
          * @param up flag that indicates how the functions floor() and ceil() are to be used. Default this value is true.
          *           This parameter is usefull when handling user input to change the scale. When the scale is to be made smaller
          *           this value should be set to false
          */
        PEACSSUPPORTSHARED_EXPORT double calcScaleStep( double & minIn, double & maxIn, double isoLineStep, bool up = true );


        /** @brief Compute the adaptation function transtions zone in a selected area
          * The computation depends on the distance to the center of the zone, the radius of the zone and the slope :
          * val = 1-2/(1+exp(slope*(centerDist[i]-radius))); The functions works only for the width of the
          * transition zone (defined above)
          * The final step is that the function is normalized as a consequence the slope can deviate for radii less
          * than the one used
          * Note that the user adapts the slope for a maximal distance of 50 mm. In the case the radius
          * of a zone (maximal distance) is less the adaptation function will start to approximate the step function.
          *
          * @param centerDist The distance between the center and the rest of the nodes
          * @param radius if the centerdist values are larger than radius, the scale function is 0.
          *         Between 0 and radius the values are scaled linearly
          * @param slope
          * @returns scale Values between 0-1.
          * @see TRANSITIONZONE
          */
        PEACSSUPPORTSHARED_EXPORT void computeScale( const PVector & centerDist, double radius, double slope, PVector & scale );


        /** @brief compute the adaptation function transtions zone in a selected area.
          *
          * @param centerDist The distance between the center and the rest of the nodes
          * @param radius if the centerdist values are larger than radius,, the scale function is 0.
          *        Between 0 and radius the values are scaled linearly
          * @returns scale: Values between 0-1.
          */
        PEACSSUPPORTSHARED_EXPORT PVector computelinearScale( const PVector & centerDist, double radius );


        /** @brief TODO
          * @param ver the vertices from which the maximum is determined
          * @returns the maximum of the vertices for each dimension
          */
        PEACSSUPPORTSHARED_EXPORT P3DVector pMax( const P3DVertices &ver );


        /** @brief TODO
          *
          * @param ver1
          * @param ver2
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT P3DVector pMax( const P3DVector &ver1, const P3DVector &ver2 );


        /** @brief TODO
          * @param ver the vertices from which the minimum is determined
          * @returns the minimum of the vertices for each dimension
          */
        PEACSSUPPORTSHARED_EXPORT P3DVector pMin( const P3DVertices &ver );


        /** @brief TODO
          *
          * @param ver1
          * @param ver2
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT P3DVector pMin( const P3DVector &ver1, const P3DVector &ver2 );


        /** @brief TODO
          *
          * @param v1
          * @param v2
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector pMax( const PVector &v1, const PVector &v2 );


        /** @brief TODO
          *
          * @param v1
          * @param v2
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector pMin( const PVector &v1, const PVector &v2 );


        /** @brief Computes the distances from the given foci based on the supplied adjacency matrix.
          * @param adj ajacency matrix ( entry with value zero means that the
          *        vertices are not connceted
          * @param  foci The shortest paths are computed from the foci
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector fociPaths( const PMatrix &adj, const QList <int> &foci, const QList <double> &initTimes );


        /** @brief Computes the distances from the given foci based on the supplied adjacency matrix.
          *
          * @param adj Ajacency matrix ( entry with value zero means that the vertices are not connceted)
          * @param foci The shortest paths are computed from the foci
          * @param initTimes The initial times to use for the given foci
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector fociPathsDirectional( const PMatrix & adj,
                                                                const QList <int> & foci,
                                                                const QList <double> & initTimes );


        /** @brief Computes the distances from the given startpoint based on the supplied geometry.
          * @param geom the geometrye use dto compute the first order ajacency matrix
          * @param   foci The shortest paths are computed from the foci
          * @return Vector
          */
        PEACSSUPPORTSHARED_EXPORT PVector shortestDist( const PGeometry &geom, int start );


        /** @brief
          */
        PEACSSUPPORTSHARED_EXPORT PMatrix shortestPathAlgo( const PMatrix &adj, bool directional = false, QProgressDialog * progressDialog = NULL);


        /** @brief Computes the shortest route from start to end point  given by index) based first order
          * adjacency matrix computed from the fiven geometry.
          * @param geom the geometrye use dto compute the first order ajacency matrix
          * @param start index of the start vertex
          * @param stop index of the start vertex
          * @return List with integer values
          */
        PEACSSUPPORTSHARED_EXPORT QList <int> getRoute( const PGeometry &geom, int start, int stop );


        /** @brief Computes the shortest route from start to end point  given by index) based on the
          *        supplied adjacency matrix.
          * @param adj ajacency matrix ( entry with value zero means that the
          *        vertices are not connceted
          * @param start index of the start vertex
          * @param stop index of the start vertex
          * @return List with integer values
          */
        PEACSSUPPORTSHARED_EXPORT QList <int> getRoute( const PMatrix &adj, int start, int stop );


        /** @brief
          */
        PEACSSUPPORTSHARED_EXPORT double pearsonCorrelation(const PMatrix &X,const PMatrix &Y);


        PEACSSUPPORTSHARED_EXPORT double meanRowPearsonCorrelation(const PMatrix &X,const PMatrix &Y);

//        PEACSSUPPORTSHARED_EXPORT PMatrix interpolSignals(const PMatrix &X,const PIVector &time, int interval=1);

        /** @brief
          */
        PEACSSUPPORTSHARED_EXPORT PMatrix interpolSignals(const PMatrix & X, const PIVector & time, int maxt, int interval = 1 );
    }
}

#endif // __PFUNCTIONS_H
