/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pgeometry.cpp 1190 2014-04-09 15:32:47Z peter $
 *
 */

#include <QDebug>
#include <QFile>
#include "pgeometry.h"
#include "pfunctions.h"
#include "pbox.h"
#include "pface.h"

using namespace Peacs::Support;


const int FILE_GEOMETRY_VERSION_1 = 1;
const int FILE_GEOMETRY_VERSION_DEFAULT  = FILE_GEOMETRY_VERSION_1;

const char* FILE_GEOMETRY_NAME = "PGeometry";


const int MAXVERFORFACEMATRIX = 5000;

PGeometry::PGeometry( void ) :
    PFile(),
    mVertices(),
    mFaces(),
    mVerticesNrm(),
    mVerticesArea(),
    mPrecision(4),
    mFaceEdgeConnectMatrix(),
    mNeighbors(),
    mEdgeList()
{
}


PGeometry::PGeometry( QString fn ) :
    PFile(),
    mVertices(),
    mFaces(),
    mVerticesNrm(),
    mVerticesArea(),
    mPrecision(4),
    mFaceEdgeConnectMatrix(),
    mNeighbors(),
    mEdgeList()
{
    load( fn );
}


PGeometry::PGeometry( const PGeometry& geom ) :
    PFile( geom ),
    mVertices( geom.mVertices ),
    mFaces( ),
    mVerticesNrm( geom.mVerticesNrm ),
    mVerticesArea( geom.mVerticesArea ),
    mPrecision(geom.mPrecision),
    mFaceEdgeConnectMatrix( geom.mFaceEdgeConnectMatrix ),
    mNeighbors( geom.mNeighbors ),
    mEdgeList( geom.mEdgeList )
{
    // Don't use the copy constructor of mFaces because this member should have pointers to vertices that
    // belong to the list of vertices contained by this class ant not to the vertices from the original
    // (geom) class
    // (Note that the vertices are copied)
    for ( int i = 0; i < geom.mFaces.size(); i++ )
    {
        mFaces.append( PFace( geom.mFaces[i][PFace::P1],
                              geom.mFaces[i][PFace::P2],
                              geom.mFaces[i][PFace::P3],
                              *this ) ) ;
    }
}


PGeometry::PGeometry(const P3DVertices &vertices, const QVector<int> &faceIndex ) :
    PFile(),
    mVertices(),
    mFaces(),
    mVerticesNrm(),
    mVerticesArea(),
    mPrecision(4),
    mFaceEdgeConnectMatrix(),
    mNeighbors(),
    mEdgeList()
{
    if ( faceIndex.size() % PFace::FACE_ITEM_COUNT == 0 )
    {
        int k = 0;
        int i0;
        int i1;
        int i2;

        mVertices = vertices;

        checkScale();

        for (int i = 0;i < faceIndex.size() / PFace::FACE_ITEM_COUNT; i++ )
        {
            i0          = faceIndex[k];
            k++;

            i1          = faceIndex[k];
            k++;

            i2          = faceIndex[k];
            k++;

            mFaces.append(PFace( i0, i1, i2, *this ));
        }
        update();
    }
}

PGeometry::PGeometry(const P3DVertices &vertices, const int *&triangleIndices, int numTriangles) :
    PFile(),
    mVertices(vertices),
    mFaces(),
    mVerticesNrm(),
    mVerticesArea(),
    mPrecision(20),
    mFaceEdgeConnectMatrix(),
    mNeighbors(),
    mEdgeList()
{
    mVertices.limitPrecision(10);
    PIVector indices(vertices.N());
    for (int i=0;i<indices.N(); i++)
    {
        indices[i] = i;
        for (int j = 0; j < i; j++)
        {
            if ( mVertices[i] == mVertices[j] )
            {
                indices[i] = j;
                break;
            }
        }
    }
    int maxI = 0;
    for (int i=0;i<indices.N(); i++)
    {
        maxI = qMax(maxI,indices[i]);
    }
    mVertices.resizeAndToZero( maxI + 1 );

    mFaces.reserve(numTriangles);
    for (int i = 0; i < numTriangles; i++ )
    {
        mFaces.append(PFace( indices[ triangleIndices[i * 3 + 0] ] ,
                             indices[ triangleIndices[i * 3 + 1] ] ,
                             indices[ triangleIndices[i * 3 + 2] ] , *this ));
    }
    update();
}



PGeometry::~PGeometry( void )
{
}


void PGeometry::clear( void )
{
    mFaceEdgeConnectMatrix.clear();
    mVertices.clear();
    mFaces.clear();
    mVerticesNrm.clear();
    mVerticesArea.clear();
    mEdgeList.clear();
    mNeighbors.clear();
}


int PGeometry::getDefaultVersion( void ) const
{
    return FILE_GEOMETRY_VERSION_DEFAULT;
}


const char *PGeometry::getClassName( void ) const
{
    return FILE_GEOMETRY_NAME;
}


bool PGeometry::isEmpty( void ) const
{
    return mVertices.isEmpty() || mFaces.isEmpty();
}


void PGeometry::update( void )
{
    updateMembers();
    checkFaceEdgeConnectmatrix();
    computeVertexNeighbors();
}


void PGeometry::geometryChanged( bool vertexChanged )
{
    if ( vertexChanged )
    {
        updateMembers();
    }
    else
    {
        mFaceEdgeConnectMatrix.clear();
        mNeighbors.clear();
        mEdgeList.clear();

        update();
    }
}


const PGeometry& PGeometry::operator= ( const PGeometry& geom )
                                     {
    if ( this != &geom )
    {
        mVerticesNrm            = geom.mVerticesNrm;
        mVerticesArea           = geom.mVerticesArea;
        mPrecision              = geom.mPrecision;
        mVertices               = geom.mVertices;
        mFaceEdgeConnectMatrix  = geom.mFaceEdgeConnectMatrix;
        mEdgeList               = geom.mEdgeList;
        mNeighbors              = geom.mNeighbors;

        // See the comment in the copy constructor why the code below is needed
        mFaces.clear();
        for ( int i = 0; i < geom.mFaces.size(); i++ )
        {
            mFaces.append( PFace( geom.mFaces[i][PFace::P1],
                                  geom.mFaces[i][PFace::P2],
                                  geom.mFaces[i][PFace::P3],
                                  *this ) );
        }
    }
    return *this;
}


const PGeometry & PGeometry::operator += ( const PGeometry &addgeom )
{
    if ( this != &addgeom )
    {
        for ( int i = 0; i < addgeom.nFaces(); i++ )
        {
            addFace( addgeom.face( i ).vertex0(),
                     addgeom.face( i ).vertex1(),
                     addgeom.face( i ).vertex2(),
                     false );
        }

        update();
    }

    return *this;
}


void PGeometry::setVertices( const PGeometry &geom )
{
    if ( this != &geom && mVertices.size() == geom.nVertices() )
    {
        mVertices = geom.mVertices;

        geometryChanged( true );
    }
}


void PGeometry::setVertices( const P3DVertices &vertices )
{
    if ( mVertices.size() == vertices.size() )
    {
        mVertices = vertices;

        limitPrecision();

        geometryChanged( true );
    }
}

void PGeometry::limitPrecision( )
{
    mVertices.limitPrecision( mPrecision );
}

bool PGeometry::checkLimitPrecision()
{
    P3DVertices ver(mVertices);
    ver.limitPrecision(mPrecision);
    bool changed = (ver != mVertices);

    PGeometry geom;
    for ( int i=0; i < mFaces.size(); i++ )
    {
        geom.addFace(ver.at(mFaces[i].I0()),ver.at(mFaces[i].I1()),ver.at(mFaces[i].I2()),false);
    }
    if ( geom.nVertices() != mVertices.N() )
    {
        qDebug() << "refined geometry due to double vertices";
        geom.geometryChanged(false);
        *this = geom;
        changed = true;
    }
    return changed;
}



P3DVertices PGeometry::smoothedNormals( void ) const
{
    P3DVertices nr( mVerticesNrm );

    for ( int i = 0; i < mVerticesNrm.size(); i++ )
    {
        QList <int> neigh( neighbors( i ) );

        nr[i] += mVerticesNrm.meanOf( neigh );

        nr[i].normate();
    }

    return nr;
}


bool PGeometry::operator!= ( const PGeometry &geom ) const
{
    bool different = false;

    if ( geom.nVertices() == mVertices.size() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            if ( mVertices.at( i ) != geom.mVertices.at( i ) )
            {
                different = true;

                break;
            }
        }
    }
    else
    {
        different = true;
    }

    return different;
}


bool PGeometry::operator== ( const PGeometry &geom ) const
{
    bool theSame = true;

    if ( geom.nVertices() == mVertices.size() )
    {
        for ( int i = 0; i < mVertices.size(); i++ )
        {
            if ( mVertices.at( i ) != geom.mVertices.at( i ) )
            {
                theSame = false;

                break;
            }
        }
    }
    else
    {
        theSame = false;
    }

    return theSame;
}



int PGeometry::findFace(const PFace & face) const
{
    int retVal =-1;
    for ( int i=0; i < mFaces.size(); i++)
    {
        if ( mFaces[i].contains( face.vertex0() ) &&
             mFaces[i].contains( face.vertex1() ) &&
             mFaces[i].contains( face.vertex2() ) )
        {
            retVal = i;
            break;
        }        
    }
    return retVal;
}


int PGeometry::findEdge(const PFace & face) const
{
    int retVal =-1;
    for ( int i=0; i < mFaces.size(); i++)
    {
        if ( mFaces[i].contains( face.vertex0() ) +
             mFaces[i].contains( face.vertex1() ) +
             mFaces[i].contains( face.vertex2() ) == 2 )
        {
            retVal = i;
            break;
        }
    }
    return retVal;
}
int PGeometry::findEdge(const P3DVector &v1, const P3DVector &v2, const P3DVector &v3)const
{
    int retVal =-1;
    for ( int i=0; i < mFaces.size(); i++)
    {
        if ( mFaces[i].contains( v1 ) +
             mFaces[i].contains( v2 ) +
             mFaces[i].contains( v3 ) == 2 )
        {
            retVal = i;
            break;
        }
    }
    return retVal;
}


int PGeometry::findFace(int i0, int i1) const
{
    int retVal =-1;
    for ( int i=0; i < mFaces.size(); i++)
    {
        if ( ( mFaces[i].I0() == i0 && mFaces[i].I1() ==i1 ) ||
             ( mFaces[i].I1() == i0 && mFaces[i].I2() ==i1 ) ||
             ( mFaces[i].I2() == i0 && mFaces[i].I0() ==i1 ) )
        {
            retVal = i;
            break;
        }
    }
    return retVal;
}


int PGeometry::findFace(const PEdge & edge, const QVector <bool>  skip) const
{
    int retVal =-1;
    if ( skip.size() != mFaces.size() )
    {
        for ( int i=0; i < mFaces.size(); i++)
        {
            if ( !skip[i] && mFaces[i].hasEdge(edge) )
            {
                retVal = i;
                break;
            }
        }
    }
    else
    {
        for ( int i=0; i < mFaces.size(); i++)
        {
            if ( mFaces[i].hasEdge(edge) )
            {
                retVal = i;
                break;
            }
        }
    }
    return retVal;
}


int PGeometry::findFace(int i0, int i1, int & otherVertexIndex) const
{
    int retVal =-1;
    otherVertexIndex = -1;
    for ( int i=0; i < mFaces.size(); i++)
    {
        if (  mFaces[i].I0() == i0 && mFaces[i].I1() ==i1 )
        {
            otherVertexIndex = mFaces[i].I2();
            retVal = i;
            break;
        }
        else if( mFaces[i].I1() == i0 && mFaces[i].I2() ==i1 )
        {
            otherVertexIndex = mFaces[i].I0();
            retVal = i;
            break;
        }
        else if ( mFaces[i].I2() == i0 && mFaces[i].I0() ==i1 )
        {
            otherVertexIndex = mFaces[i].I1();
            retVal = i;
            break;
        }
    }
    return retVal;
}



double PGeometry::solidAngle( const P3DVector &observationPnt ) const
{
    double solida = 0;

    for ( int j = 0; j < mFaces.size(); j++ )
    {
        solida += mFaces.at( j ).solidAngle( observationPnt );
    }

    return solida;
}


bool PGeometry::load( const QString &FileName )
{
    /* All face index number range between 1 and the number of vertices (matlab
     * based) In C+ the index number range from 0 to  number of vertices -1
     * */
    bool ok = false;

    if ( !FileName.isEmpty() )
    {
        QFile file( FileName );

        if ( file.open( QIODevice::ReadOnly ) )
        {
            int         nr;
            QTextStream textStream( &file );

            clear();

            textStream >> nr;

            if ( (textStream.status() == QTextStream::Ok) && (nr > 0) )
            {
                int tmpIndex;
                int ir;
                int i0;
                int i1;
                int i2;

                mVertices.resizeAndToZero( nr );

                // read vertices
                for ( ir = 0; (ir < nr) && (textStream.status() == QTextStream::Ok); ir++ )
                {
                    textStream >> tmpIndex >> mVertices[ir];
                }

                checkScale();

                if ( (ir == nr) && (textStream.status() == QTextStream::Ok) )
                {
                    textStream >> nr;

                    if ( textStream.status() == QTextStream::Ok &&  nr > 0 )
                    {
                        for ( ir = 0; (ir < nr) && (textStream.status() == QTextStream::Ok); ir++ )
                        {
                            textStream >> tmpIndex >> i0 >> i1 >> i2;

                            i0--;
                            i1--;
                            i2--;

                            mFaces.append( PFace( i0, i1, i2, *this ) ) ;
                        }

                        ok = (ir == nr);
                    }
                }
            }

            file.close();
        }

        if ( !ok )
        {
            // during reading an error might have occured, clear therefore
            clear();
            qDebug().nospace() << "PGeometry::load( "<< FileName << " ) could not be read";

        }
        else
        {
            update();
        }
    }
    else
    {
        qDebug().nospace() << "PGeometry::load filename is empty ";
    }

    return ok;
}


bool PGeometry::save( const QString &FileName ) const
{
    int     i;
    bool    ok = false;
    QFile   file( FileName );

    if ( file.open( QIODevice::WriteOnly ) )
    {
        QTextStream textStream( &file );

        textStream.setRealNumberPrecision( 6 );     // enough for geometry data

        textStream << mVertices.size() << " \n";

        for ( i = 0; i < mVertices.size() && (textStream.status() == QTextStream::Ok); i++ )
        {
            textStream << (i + 1);
            textStream << "\t";
            textStream << vertex( i ).x();
            textStream << "\t";
            textStream << vertex( i ).y();
            textStream << "\t";
            textStream << vertex( i ).z();
            textStream << "\n";
        }

        if ( textStream.status() == QTextStream::Ok )
        {
            textStream << mFaces.size() << " \n";

            for ( i = 0; i < mFaces.size() && (textStream.status() == QTextStream::Ok); i++ )
            {
                textStream << (i + 1);
                textStream << "\t";
                textStream << face( i )[PFace::P1] + 1;
                textStream << "\t";
                textStream << face( i )[PFace::P2] + 1;
                textStream << "\t";
                textStream << face( i )[PFace::P3] + 1;
                textStream << "\n";
            }
        }

        ok = ( textStream.status() == QTextStream::Ok );

        file.close();
    }

    return ok;
}

/** @brief stl file format handling
 *  facet normal 0.0 0.0 1.0
 *      outer loop
 *          vertex  1.0  1.0  0.0
 *          vertex -1.0  1.0  0.0
 *          vertex  0.0 -1.0  0.0
 *      endloop
 *  endfacet
 * Binary STL files consist of a 80 byte header line that can be
 * interpreted as a comment string. The following 4 bytes interpreted
 * as a long integer give the total number of facets. What follows is
 * a normal and 3 vertices for each facet, each coordinate represented
 * as a 4 byte floating point number (12 bytes in all). There is a 2
 * byte spacer between each facet. The result is that each facet is represented
 * by 50 bytes, 12 for the normal, 36 for the 3 vertices, and 2 for the spacer.
 * @param str the name of the file
 */
bool PGeometry::saveToSTL(const QString &filename)
{
    QFile   f( filename );

    if ( !f.open( QIODevice::WriteOnly ) )
    {
        return false;
    }

    QDataStream dataStream( &f );

    dataStream.setByteOrder( QDataStream::LittleEndian );
    dataStream.setFloatingPointPrecision( QDataStream::SinglePrecision );
    char header[80];
    char spacer[2];
    spacer[0] = 0;
    spacer[1] = 0;
    qint32 length = mFaces.size();

    dataStream << header;
    dataStream << length;

    for (int i=0; i < mFaces.size() && dataStream.status() == QDataStream::Ok; i++)
    {
        dataStream << mFaces[i].normal();
        dataStream << mFaces[i].vertex0();
        dataStream << mFaces[i].vertex1();
        dataStream << mFaces[i].vertex2();
        dataStream << spacer;
    }
    f.close();

    return dataStream.status() == QDataStream::Ok;
}


bool PGeometry::loadFromSTL(const QString &filename)
{
    if ( !filename.isEmpty() )
    {
        QFile f( filename);

        if ( f.open( QIODevice::ReadOnly ) )
        {
            clear();
            qint16      bc;
            qint32      nFaces;
            QDataStream dataStream( &f );
            dataStream.setFloatingPointPrecision( QDataStream::SinglePrecision );

            char buf[80];
            dataStream.readRawData(buf,80);
            dataStream >> nFaces;
            for (int i = 0; dataStream.status() == QDataStream::Ok && i < nFaces ; i++ )
            {
                float x,y,z;
                dataStream >> x >> y >> z;
                dataStream >> x >> y >> z;
                P3DVector v1(x,y,z);
                dataStream >> x >> y >> z;
                P3DVector v2(x,y,z);
                dataStream >> x >> y >> z;
                P3DVector v3(x,y,z);
                addFace(v1,v2,v3,false);
                dataStream >> bc;
            }
            bool ok = dataStream.status() == QDataStream::Ok;
            f.close();
            if ( !ok )
            {
                // during reading an error might have occured, clear therefore
                clear();
                qDebug().nospace() << "PGeometry::load( "<< filename << " ) could not be read";

            }
            else
            {
                update();
            }
            return ok;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
}

bool PGeometry::loadFromXml(const QString &str)
{
    Q_UNUSED( str )
    // TODO
    //QTextStream textStream( &str );

//    QString;
//    textStream >> nr;
//    if ( (textStream.status() == QTextStream::Ok) && (nr > 0) )
//    {
//        int tmpIndex,ir,i0,i1,i2;

//        mVertices.resize(nr);

//        // read vertices
//        for (ir=0; (ir < nr) && (textStream.status() == QTextStream::Ok); ir++)
//        {
//            textStream >> tmpIndex >> mVertices[ir];
//        }
//        checkScale();
//        if ( (ir == nr) && (textStream.status() == QTextStream::Ok) )
//        {
//            textStream >> nr;
//            mFaces.resize(nr);
//            if ( textStream.status()==QTextStream::Ok &&  nr>0 )
//            {
//                for (ir=0; (ir < nr) && (textStream.status()==QTextStream::Ok); ir++)
//                {
//                    textStream >> tmpIndex >> i0 >> i1 >> i2;
//                    i0--;
//                    i1--;
//                    i2--;
//                    mFaces[ir]=PFace(i0, i1, i2,
//                                     &mVertices[i0],
//                                     &mVertices[i1],
//                                     &mVertices[i2]);
//                }
//                if (ir == nr)
//                {
//                    ok=true;
//                }
//            }
//        }
    return false;
}


bool PGeometry::saveToXmL( QTextStream &textStream )
{
    Q_UNUSED( textStream )

    // TODO
    return false;
}


bool PGeometry::read( QDataStream &ds )
{
    PGeometry::clear();

    bool    ok      = false;
    int     version = readHead( ds, PGeometry::getClassName() );

    if ( version != READ_HEADER_FAILED )
    {
        switch ( version )
        {
        case FILE_GEOMETRY_VERSION_1:
            ok = readGeometry( ds );
            break;

        default:
            break;
        }
    }

    return ok;
}


bool PGeometry::readGeometry( QDataStream &dataStream )
{
    bool    ok = false;
    qint32  ir;
    qint32  nr;
    qint32  version;
    qint32  i0;
    qint32  i1;
    qint32  i2;

    /* The version information is read here also so derived classes can
     * de-stream this class information with the correct version as well */
    dataStream >> version;//FILE_GEOMETRY_VERSION_DEFAULT;

    if ( isOk( dataStream ) &&
         version > 0 &&
         version <= FILE_GEOMETRY_VERSION_DEFAULT )
    {
        switch ( version )
        {
        case FILE_GEOMETRY_VERSION_1 :
            dataStream >> nr;

            if ( isOk( dataStream ) && (nr == 0) )
            {
                ok = true;
            }
            else if ( isOk( dataStream ) && (nr > 0) )
            {
                // read vertices
                mVertices.resizeAndToZero( nr );

                for ( ir = 0; (ir < nr) && isOk( dataStream ); ir++ )
                {
                    dataStream >> mVertices[ir];
                }

                if ( isOk( dataStream ) )
                {
                    dataStream >> nr;

                    if ( isOk( dataStream ) &&  nr > 0 )
                    {
                        ok = true;

                        for ( ir = 0; (ir < nr) && isOk( dataStream ); ir++ )
                        {
                            dataStream >> i0 >> i1 >> i2;

                            if ( isOk( dataStream ) )
                            {
                                mFaces.append( PFace( i0, i1, i2, *this ) );
                            }
                            else
                            {
                                ok = false;

                                break;
                            }
                        }

                        if ( ok )
                        {
                            checkScale();
                        }
                    }
                    else
                    {
                        ok = false;
                    }
                }
            }
            break;

        default:
            break;
        }
    }

    if ( ok )
    {
        update();
    }

    return ok;
}


bool PGeometry::write( QDataStream &ds )const
{
    bool ok = writeHead( ds, PGeometry::getClassName(), PGeometry::getDefaultVersion() );

    if ( ok )
    {
        switch ( PGeometry::getDefaultVersion() )
        {
        case FILE_GEOMETRY_VERSION_1:
            ok = writeGeometry( ds );
            break;

        default:
            ok = false;
            break;
        }
    }

    return ok;
}


bool PGeometry::writeGeometry( QDataStream &dataStream )const
{
    int i;

    // Write the PGeometry version here also so derived classes can be de-streamed with
    // the correct version
    dataStream << FILE_GEOMETRY_VERSION_DEFAULT;

    if ( isOk( dataStream ) )
    {
        dataStream << (qint32)mVertices.size();

        for ( i = 0; i < mVertices.size() && dataStream.status() == QDataStream::Ok; i++ )
        {
            dataStream << vertex(i).x();
            dataStream << vertex(i).y();
            dataStream << vertex(i).z();
        }

        if ( dataStream.status() == QDataStream::Ok )
        {
            dataStream << (qint32)mFaces.size();

            for ( i = 0; i < mFaces.size() && dataStream.status() == QDataStream::Ok; i++ )
            {
                dataStream << (qint32)face(i)[PFace::P1];
                dataStream << (qint32)face(i)[PFace::P2];
                dataStream << (qint32)face(i)[PFace::P3];
            }
        }
    }

    return ( dataStream.status() == QDataStream::Ok );
}


void PGeometry::updateMembers( void )
{
    mVerticesNrm.resizeAndToZero( mVertices.size() );
    mVerticesArea.resize( mVertices.size() );

    for ( int i = 0; i < mFaces.size(); i++ )
    {
        mFaces[i].updateMembers();
        P3DVector normal    = mFaces[i].normal();//(mFaces[i].normal() * mFaces[i].area()) / 3.0;
//        P3DVector normal1    = (mFaces[i].vertex2() - mFaces[i].vertex0()).cross(mFaces[i].vertex1() - mFaces[i].vertex0())/6.0;
        mVerticesNrm [mFaces[i].I0()]   += normal;
        mVerticesNrm [mFaces[i].I1()]   += normal;
        mVerticesNrm [mFaces[i].I2()]   += normal;

        double surf = mFaces[i].area() / 3.0;
        mVerticesArea[mFaces[i].I0()]   += surf;
        mVerticesArea[mFaces[i].I1()]   += surf;
        mVerticesArea[mFaces[i].I2()]   += surf;
    }

    for ( int i = 0; i < mVertices.size(); i++ )
    {
//        mVerticesArea[i] = mVerticesNrm[i].length();
        mVerticesNrm[i].normate();
    }
}

int PGeometry::findVertexWithPrecision(const P3DVector &vertex) const
{
    double prec = getPrecisionValue();
    for (int i=0; i < mVertices.N(); i++ )
    {
        if ( (mVertices.at(i) - vertex).length() < prec )
        {
            return i;
        }
    }
    return -1;
}


void PGeometry::addVertex( const P3DVector& vertex, bool doUpdate )
{
    insertVertex( mVertices.size(), vertex, doUpdate );
}




void PGeometry::insertVertex( int i, const P3DVector& vertex ,bool doUpdate )
{
    if ( mVertices.insertVertex( i, vertex ) )
    {
        for ( int j = 0; j < mFaces.size(); j++ )
        {
            int i0 = mFaces[j].I0();
            int i1 = mFaces[j].I1();
            int i2 = mFaces[j].I2();

            if ( i <= i0)
            {
                i0++;
            }

            if ( i <= i1 )
            {
                i1++;
            }

            if ( i <= i2 )
            {
                i2++;
            }

            mFaces[j].setData( i0, i1, i2 );
        }
    }

    mNeighbors.clear();
    mFaceEdgeConnectMatrix.clear();
    mEdgeList.clear();

    if ( doUpdate )
    {
        geometryChanged( false );
    }
}


void PGeometry::removeVertex( int i, bool doUpdate )
{
    mVertices.removeVertex( i );

    int j = 0;
    while ( j < mFaces.size() )
    {
        if ( mFaces[j].contains( i ) )
        {
            mFaces.removeAt( j );
        }
        else
        {
            j++;
        }
    }

    for ( int j = 0; j < mFaces.size(); j++ )
    {
        if ( mFaces[j].I0() > i )
        {
            mFaces[j][PFace::P1] -= 1;
        }
        if ( mFaces[j].I1() > i )
        {
            mFaces[j][PFace::P2] -= 1;
        }
        if ( mFaces[j].I2() > i )
        {
            mFaces[j][PFace::P3] -= 1;
        }
    }
    if ( doUpdate )
    {
        geometryChanged( false );
    }
}

void PGeometry::removeUnusedVertices( bool doUpdate )
{
    PIVector isUsed(nVertices(),0);
    for (int i=0; i < nFaces(); i++)
    {
        isUsed[mFaces[i].I0()]++;
        isUsed[mFaces[i].I1()]++;
        isUsed[mFaces[i].I2()]++;
    }
    for ( int i = isUsed.N()-1; i >=0 ; i--)
    {
        if ( !isUsed[i] )
        {
            mVertices.removeVertex( i );
//            qDebug() << "removed  " << i;
            for ( int j = 0; j < mFaces.size(); j++ )
            {
                if ( mFaces[j].I0() > i )
                {
                    mFaces[j][PFace::P1] -= 1;
                }
                if ( mFaces[j].I1() > i )
                {
                    mFaces[j][PFace::P2] -= 1;
                }
                if ( mFaces[j].I2() > i )
                {
                    mFaces[j][PFace::P3] -= 1;
                }
            }
        }
    }
    if ( doUpdate )
    {
        geometryChanged( false );
    }
}



void PGeometry::updateFace( int face,int i0,int i1,int i2 )
{

    if ( face >= 0 && face < mFaces.size()    &&
         i0   >= 0 && i0   < mVertices.size() &&
         i1   >= 0 && i1   < mVertices.size() &&
         i2   >= 0 && i2   < mVertices.size() )
    {
        mFaces[face].setData( i0, i1, i2 );
    }
    else
    {
        qDebug("Face not updated");
    }

    geometryChanged( false );
}

void PGeometry::removeFaces( QList<int> &faces, bool doUpdate )
{
    if ( !faces.isEmpty() )
    {
        QList <int>remList(faces);
        qSort(remList);
        for (int j = remList.size()-1; j >=0; j--)
        {
            removeFace(remList[j],doUpdate);
        }
    }
}



void PGeometry::removeFace( int face,bool doUpdate )
{
    if ( face >=0  && face < mFaces.size() )
    {
        mFaces.removeAt( face );
        removeUnusedVertices(false);
        if ( doUpdate )
        {
            geometryChanged( false );
        }
    }
    else
    {
        qDebug("Face not removed");
    }
}



void PGeometry::addFace( const P3DVector &v0, const P3DVector &v1, const P3DVector &v2, bool doUpdate )
{
    int i0 = mVertices.addVertex( v0, true );
    int i1 = mVertices.addVertex( v1, true );
    int i2 = mVertices.addVertex( v2, true );

    if ( i0 == i1 || i0 == i2 || i1 == i2 )
    {
        qDebug().nospace() << " no face added " << i0 << " " << i1 << " " << i2;
    }
    else
    {
        mFaces.append( PFace(i0, i1, i2, *this) );

        mFaceEdgeConnectMatrix.clear();
        mNeighbors.clear();
        mEdgeList.clear();

        if ( doUpdate )
        {
            geometryChanged( false );
        }
    }
}


void PGeometry::changeVertexNumber(int index1,int index2)
{
    if ( index1 >=0 && index1 < mVertices.N() &&
         index2 >=0 && index2 < mVertices.N())
    {
        P3DVector cop(mVertices[index2]);
        mVertices[index2] = mVertices[index1];
        mVertices[index1] = cop;
        for (int i=0;i < mFaces.size(); i++)
        {
            mFaces[i].changeVertexNumber(index1, index2);
        }
    }
}

void PGeometry::checkScale( void )
{
    if ( mVertices.size() > 0 )
    {
        P3DVector minCoor;
        P3DVector maxCoor;
        P3DVector delta;

        minmax( minCoor, maxCoor );

        delta = maxCoor - minCoor;

        if ( delta.x() < 1 && delta.y() && delta.z() < 1 ) // less than a meter if expressed in meters
        {
            for ( int ir = 0; ir < mVertices.size(); ir++ )
            {
                mVertices[ir] *= 1000.0;
            }
        }
    }
}


//pntontotri
P3DVector PGeometry::findProjectionPointAdjacentToVertex( const P3DVector &ver, int &iMinFace , int iClostestVertex ) const
{
    P3DVector onGeom;

    if ( iClostestVertex < 0 )
    {
        onGeom = pointOnGeometry(ver);
    }
    else
    {
        P3DVector verOnFace;
        double dist;
        double mindist = HUGE_VAL;
        onGeom = mVertices.at(iClostestVertex);
        for ( int i = 0; i < mFaces.size(); i++ )
        {
            if ( mFaces[i].contains( iClostestVertex ) )
            {
                dist = mFaces[i].vertex2FaceDist( ver, verOnFace );
                // it is assumed that the user knowes if the vertex is part of the geometry (dist >0)
                if ( dist < mindist )
                {
                    mindist  = dist;
                    iMinFace = i;
                    onGeom   = verOnFace;
                }
            }
        }
    }
    return onGeom;
}


P3DVector PGeometry::pointfromOnesideOnGeometry( const P3DVector &point, bool outside ) const
{
    P3DVector   projectionPnt;
    int         i;
    bool        found       = false;
    double      normalAngel = 0.707;    // == cos( 45 degr. )

    if ( !outside )
    {
        normalAngel *= -1.0;
    }

    for ( i = 0; i < mVertices.size(); i++ )
    {
        // if the distance between point end the point on the surface is
        // less than 0.01 mm it is assumed to be on the surface
        if ( ( mVertices.at( i ) - point ).length() < 0.01 &&
             ( point - mVertices.at( i ) ).normalized().dot( mVerticesNrm.at( i ) ) > normalAngel )
        {
            projectionPnt   = mVertices.at( i );
            found           = true;

            break;
        }
    }

    if ( !found )
    {
        P3DVector   verOnFace;
        double      dist;
        double      mindist = HUGE_VAL;

        for ( int i = 0; i < mFaces.size(); i++ )
        {
            if ( mFaces[i].projectVerOnFace( point, verOnFace ) > 0 )
            {
                dist = ( verOnFace - point ).length();

                if ( ( point - verOnFace ).normalized().dot( mFaces[i].normal() ) > normalAngel &&
                      dist < mindist )
                {
                    mindist         = dist;
                    projectionPnt   = verOnFace;
                }
            }
        }
    }
    return projectionPnt;
}


P3DVector PGeometry::pointOnGeometry( const P3DVector &point ) const
{
    P3DVector   projectionPnt;
    P3DVector   verOnFace;
    double      dist;
    double      mindist = HUGE_VAL;

    for ( int i = 0; i < mFaces.size(); i++ )
    {
        dist = mFaces[i].vertex2FaceDist( point, verOnFace);

        if ( dist < mindist )
        {
            mindist         = dist;
            projectionPnt   = verOnFace;
        }
    }
    if ( mindist == HUGE_VAL && !isEmpty() )
    {
        qDebug() << "pointOnGeometry: no projection point found";
    }
    return projectionPnt;
}


bool PGeometry::pointOnGeometry( const P3DVector &point,int &vertexIndex,
                                 int &faceIndex, double &lambda, double &mu ) const
{
    vertexIndex = -1;
    faceIndex   = -1;
    lambda      = 0;
    mu          = 0;

    P3DVector   verOnFace;
    double      dist;
    double      mindist = HUGE_VAL;
    double      lab;
    double      u;
    P3DVector projectionPnt;

    for ( int i = 0; i < mFaces.size(); i++ )
    {
        dist = mFaces[i].vertex2FaceDist( point, verOnFace, lab, u );

        if ( dist < mindist )
        {
            faceIndex   = i;
            mindist     = dist;
            lambda      = lab;
            mu          = u;
            projectionPnt = verOnFace;
        }
    }

    if ( faceIndex >=0 )
    {
        double d0 = (mFaces[faceIndex].vertex0() - projectionPnt).length();
        double d1 = (mFaces[faceIndex].vertex1() - projectionPnt).length();
        double d2 = (mFaces[faceIndex].vertex2() - projectionPnt).length();
        if ( d0 < d1 && d0 < d2)
        {
            vertexIndex = mFaces[faceIndex].I0();
        }
        else if ( d1 < d2)
        {
            vertexIndex = mFaces[faceIndex].I1();
        }
        else
        {
            vertexIndex = mFaces[faceIndex].I2();
        }
    }
    else if ( !isEmpty() )
    {
        qDebug() << "pointOnGeometry: no projection point found";
    }
    return faceIndex >= 0;
}


bool PGeometry::pointOnGeometry( const P3DVector &point, int &vertexIndex, int &faceIndex, P3DVector &projectionPnt ) const
{
    vertexIndex = -1;
    faceIndex   = -1;

    P3DVector   verOnFace;
    double      dist;
    double      mindist = HUGE_VAL;

    for ( int i = 0; i < mFaces.size(); i++ )
    {
        dist = mFaces[i].vertex2FaceDist( point, verOnFace );
        if ( dist < mindist )
        {
            faceIndex       = i;
            mindist         = dist;
            projectionPnt   = verOnFace;
        }
    }
    if ( faceIndex >=0 )
    {
        double d0 = (mFaces[faceIndex].vertex0() - projectionPnt).length();
        double d1 = (mFaces[faceIndex].vertex1() - projectionPnt).length();
        double d2 = (mFaces[faceIndex].vertex2() - projectionPnt).length();
        if ( d0 < d1 && d0 < d2)
        {
            vertexIndex = mFaces[faceIndex].I0();
        }
        else if ( d1 < d2)
        {
            vertexIndex = mFaces[faceIndex].I1();
        }
        else
        {
            vertexIndex = mFaces[faceIndex].I2();
        }
    }
    else if ( !isEmpty() )
    {
        qDebug() << "pointOnGeometry: no projection point found";
    }
    return faceIndex >= 0;
}

int PGeometry::vertexIndexOnGeometry(const P3DVector &point) const
{
    int vertexIndex = -1;
    int faceIndex   = -1;
    P3DVector projectionPnt;

    P3DVector   verOnFace;
    double      dist;
    double      mindist = HUGE_VAL;

    for ( int i = 0; i < mFaces.size(); i++ )
    {
        dist = mFaces[i].vertex2FaceDist( point, verOnFace );
        if ( dist < mindist )
        {
            faceIndex       = i;
            mindist         = dist;
            projectionPnt   = verOnFace;
        }
    }
    if ( faceIndex >=0 )
    {
        double d0 = (mFaces[faceIndex].vertex0() - projectionPnt).length();
        double d1 = (mFaces[faceIndex].vertex1() - projectionPnt).length();
        double d2 = (mFaces[faceIndex].vertex2() - projectionPnt).length();
        if ( d0 < d1 && d0 < d2)
        {
            vertexIndex = mFaces[faceIndex].I0();
        }
        else if ( d1 < d2)
        {
            vertexIndex = mFaces[faceIndex].I1();
        }
        else
        {
            vertexIndex = mFaces[faceIndex].I2();
        }
    }
    else
    {
        qDebug() << "vertexIndexOnGeometry: no projection point found";
    }
    return vertexIndex;
}




QStringList PGeometry::checkData( void ) const
{
    QStringList errors;

    // TODO future tricheck functionality could generate output here
//    for ( int i = 0; i < mFaces.size(); i++ )
//    {
//        {
//            errors.append( QString( "One of the vertices of a face %1 is NULL " ).arg( i ) );

//            break;
//        }
//    }

    return errors;
}


void PGeometry::minmax( P3DVector &minCoor, P3DVector &maxCoor )const
{
    PBox box( mVertices );

    minCoor = box.getMinBox();
    maxCoor = box.getMaxBox();
}


P3DVector PGeometry::boxCenter( void ) const
{
    PBox box( mVertices );

    return box.getBoxCenter();
}


bool PGeometry::insideBox( const P3DVector &pos ) const
{
    PBox box( mVertices );

    return box.insideBox( pos );
}


void PGeometry::translate( const P3DVector &trans )
{
    mVertices += trans;

    updateMembers();
}

void PGeometry::translate( const P3DVertices &trans )
{
    mVertices += trans;

    updateMembers();
}


P3DVertices PGeometry::checkedTranslation(const P3DVertices &trans )
{
    bool transOk = false;
    P3DVertices checkedTrans(trans);

    while ( transOk )
    {
        transOk = true;
        checkedTrans.limitPrecision(mPrecision);
        double prec = getPrecisionValue() * 10.0;
        P3DVertices newVer(mVertices + checkedTrans);

        for ( int i = 0; i < newVer.N(); i++ )
        {
            for (int j = i+1; j < newVer.N(); j++ )
            {
                if ( newVer[i] == newVer[j] && !checkedTrans[i].isZero() )
                {
                    checkedTrans[i] /= 2.1; // to make sure a new x,y,z is computed
                    if ( checkedTrans[i].length() < prec  )
                    {
                        checkedTrans[i].toZeros();
                    }
                    transOk = false;
                    qDebug() << "translation refined in checkedTranslation";
                    break;
                }
            }
        }
    }
    mVertices += checkedTrans;
    updateMembers();
    return checkedTrans;
}


void PGeometry::scale( const P3DVector &scale )
{
    mVertices *= scale;

    updateMembers();
}

void PGeometry::scaleKeepCenter(const P3DVector &scale, const P3DVector & center)
{
    mVertices.scaledKeepCenter(scale,center);
}


void PGeometry::computeVertexNeighbors( void )
{
    mNeighbors.clear();

    int i;
    int j;
    int iN = 0;
    int N  = mVertices.size();

    mNeighbors.resize( mVertices.size() );

    for ( i = 0; i < N; i++, iN += N )
    {
        int neighbor        = 0;
        for ( j = 0; j < mFaces.size(); j++ ) // find next neighbor
        {
            if ( mFaces[j].contains( i ) )
            {
                if ( mFaces[j].I0() == i )
                {
                    neighbor = mFaces[j].I1();
                    break;
                }
                else
                {
                    neighbor = mFaces[j].I0();
                    break;
                }
            }
        }
        int startNeighbor   = neighbor;


        do
        {
            mNeighbors[i].append( neighbor );

            for ( j = 0; j < mFaces.size(); j++ ) // find next neighbor
            {
                if ( mFaces[j].contains( i ) )
                {
                    if ( mFaces[j].I0() == i && mFaces[j].I1() == neighbor )
                    {
                        neighbor = mFaces[j].I2();

                        break;
                    }
                    else if ( mFaces[j].I1() == i && mFaces[j].I2() == neighbor )
                    {
                        neighbor = mFaces[j].I0();

                        break;
                    }
                    else if ( mFaces[j].I2() == i && mFaces[j].I0() == neighbor )
                    {
                        neighbor = mFaces[j].I1();

                        break;
                    }
                }
            }

            if ( j == mFaces.size() )
            {
                // This can only happen if the geometry is not a closed surface
                mNeighbors[i].append( i );

                break;
            }
        }
        while ( neighbor != startNeighbor );
    }
}


bool PGeometry::checkFaceEdgeConnectmatrix( void )
{
    // in a consitent geometry the edges are oposite in direction for adjacent faces
    bool consistent = true;

    if ( !isEmpty() &&
         mVertices.size() <= MAXVERFORFACEMATRIX &&
         (mFaceEdgeConnectMatrix.isEmpty() ||
          mFaceEdgeConnectMatrix.size() != mVertices.size() * mVertices.size() ) )
    {
        int N = mVertices.size();

        mFaceEdgeConnectMatrix.fill( -1, N * N );

        for ( int i = 0; i < mFaces.size(); i++ )
        {
            if ( mFaceEdgeConnectMatrix[ mFaces[i].I0() * N + mFaces[i].I1() ] == -1 )
            {
                mFaceEdgeConnectMatrix[ mFaces[i].I0() * N + mFaces[i].I1() ] = i;
            }
            else
            {
                consistent = false;

                break;
            }

            if ( mFaceEdgeConnectMatrix[ mFaces[i].I1() * N + mFaces[i].I2() ] == -1 )
            {
                mFaceEdgeConnectMatrix[ mFaces[i].I1() * N + mFaces[i].I2() ] = i;
            }
            else
            {
                consistent = false;

                break;
            }

            if ( mFaceEdgeConnectMatrix[ mFaces[i].I2() * N + mFaces[i].I0() ] == -1 )
            {
                mFaceEdgeConnectMatrix[ mFaces[i].I2() * N + mFaces[i].I0() ] = i;
            }
            else
            {
                consistent = false;

                break;
            }
        }

        int i;
        int j;
        int ic;

        mEdgeList.clear();

        if ( consistent )
        {
            /* create a list of all the edges*/
            for ( i = 0, ic = 0; i < N; i++, ic += N )
            {
                for ( j = i; j < N; j++ )
                {
                    if ( mFaceEdgeConnectMatrix[ic + j] > 0 )
                    {
                        mEdgeList.append( PEdge( i, j ) );
                    }
                }
            }
        }
    }
    else
    {
        mFaceEdgeConnectMatrix.clear();
    }

    return consistent;
}





int PGeometry::crossSection(const PPlane &crossPlane,
                                  P3DVertices &contourPoints,
                                  PIVector &contourIndex) const
{
    int             nDone       = 0;
    int             faceIndex   = 0;
    int             nEdges;
    PEdge           edge1;
    PEdge           edge2;
    PEdge           endEdge;
    PEdge           nextEdge;
    P3DVector       v1;
    P3DVector       v2;
    P3DVector       beginVertex;
    bool            startContour = true;
    int             searchVertex = -1;
    QVector <bool>  done( mFaces.size(), false );

    QVector <PPlane::PlaneSide_type> planeSide(crossPlane.planeSide(mVertices));

    for ( int i =0; i < mFaces.size(); i++)
    {
        if ( planeSide[mFaces[i].I0()] == planeSide[mFaces[i].I1()] &&
             planeSide[mFaces[i].I0()] == planeSide[mFaces[i].I2()] &&
             planeSide[mFaces[i].I0()] != PPlane::ONPLANE &&
             planeSide[mFaces[i].I1()] != PPlane::ONPLANE &&
             planeSide[mFaces[i].I2()] != PPlane::ONPLANE )
        {
            done[i] = true;
            nDone++;
        }
    }

    contourPoints.clear();
    contourIndex.clear();

    while ( nDone < mFaces.size() )
    {
        if ( !done[faceIndex] )
        {
            done[faceIndex] = true;

            nDone++;

            nEdges = mFaces[faceIndex].crossSection( crossPlane, v1, v2, edge1, edge2 );

            switch ( nEdges )
            {
            case  0: // no crossection
                faceIndex++;
                break;
            case  2: // two edges are crossed
                if ( startContour ) // new contour must be started
                {
                    contourIndex.append(contourPoints.size()*-1);
                    contourPoints.append(v1);
                    contourIndex.append(contourPoints.size());
                    contourPoints.append(v2);

                    endEdge         = edge1;    // in a consitent geometry the edges are oposite in direction for adjacent faces
                    nextEdge.invert( edge2 );

                    beginVertex     = v1;
                    if ( !mFaceEdgeConnectMatrix.isEmpty() )
                    {
                        faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                    }
                    else
                    {
                        faceIndex       = findFace( nextEdge, done);
                    }
                    startContour    = false;
                }
                else
                {
                    /* only the second vertex needs to be added, because the first one is added
                     * by the previous face */
                    contourIndex.append( contourPoints.size() );

                    if ( nextEdge == edge1 )
                    {
                        contourPoints.append( v2 );
                        nextEdge.invert( edge2 );
                    }
                    else
                    {
                        contourPoints.append( v1 );
                        nextEdge.invert( edge1 );
                    }

                    if ( endEdge == nextEdge )
                    {
                        contourIndex.append( contourPoints.size() );
                        contourPoints.append( beginVertex );
                        faceIndex       = 0; // start looking for a new contour
                        startContour    = true;
                    }
                    else
                    {
                        if ( !mFaceEdgeConnectMatrix.isEmpty() )
                        {
                            faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                        }
                        else
                        {
                            faceIndex       = findFace( nextEdge, done);
                        }
                    }
                }
                break;

            case  -2: // the plane crosses the face through one vertex and the oposite edge
                if ( startContour )
                {
                    contourIndex.append( contourPoints.size() * -1 );
                    contourPoints.append( v2 );
                    contourIndex.append( contourPoints.size() );
                    contourPoints.append( v1 );

                    endEdge         = edge2; // the edge is selected as the endpoint
                    searchVertex    = edge1.a();
                    beginVertex     = v2;
                    startContour    = false;
                }
                else
                {
                    // next an edge crossing or an vertex has to be found
                    contourIndex.append( contourPoints.size() );

                    if ( searchVertex >= 0 ) // we looked for a vertex so the next will be an edge
                    {
                        contourPoints.append( v2 ); // v1 was the vertex we were looking for so v2 is added

                        searchVertex    = -1;
                        nextEdge.invert( edge2 );

                        if ( endEdge ==  nextEdge )
                        {
                            contourIndex.append( contourPoints.size() );
                            contourPoints.append( beginVertex );
                            faceIndex       = 0;
                            startContour    = true;
                        }
                        else
                        {
                            if ( !mFaceEdgeConnectMatrix.isEmpty() )
                            {
                                faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                            }
                            else
                            {
                                faceIndex       = findFace( nextEdge, done);
                            }
                        }
                    }
                    else // edge is found so the next one will be an vertex
                    {
                        searchVertex = edge1.a();

                        contourPoints.append( v1 ); // v2 was the edge we were looking for so v1 is added

                        if ( searchVertex == endEdge.a() )
                        {
                            faceIndex       = 0;
                            startContour    = true;

                            contourIndex.append( contourPoints.size() );
                            contourPoints.append( beginVertex );
                            searchVertex    = -1;
                        }
                        else
                        {
                            nextEdge[PEdge::A]  = searchVertex;
                            nextEdge[PEdge::B]  = searchVertex;
                        }
                    }
                }
                break;

            case  -1: // one edge touches the plane
                if ( !mFaceEdgeConnectMatrix.isEmpty() )
                {
                    done[ mFaceEdgeConnectMatrix[ edge1.a() * mVertices.size() + edge1.b() ] ] = true;
                    done[ mFaceEdgeConnectMatrix[ edge1.b() * mVertices.size() + edge1.a() ] ] = true;
                }
                else
                {
                    done[ findFace( edge1, done ) ] = true;
                    done[ findFace( edge1.inverted() , done ) ] = true;
                }
                nDone++;

                if ( startContour )
                {
                    contourIndex.append( contourPoints.size() * -1 );
                    contourPoints.append( v1 );
                    contourIndex.append( contourPoints.size() );
                    contourPoints.append( v2 );

                    endEdge[PEdge::A]   = edge1.a();
                    endEdge[PEdge::B]   = edge1.a();
                    searchVertex        = edge1.b();
                    nextEdge[PEdge::A]  = searchVertex;
                    nextEdge[PEdge::B]  = searchVertex;
                    beginVertex         = v1;
                    startContour        = false;
                }
                else
                {
                    contourIndex.append( contourPoints.size() );

                    if ( searchVertex == edge1.a() ) // we looked for a vertex and so will the next
                    {
                        contourPoints.append( v2 ); // v1 was the vertex we were looking for so v2 is added
                        searchVertex        = edge1.b();
                        nextEdge[PEdge::A]  = searchVertex;
                        nextEdge[PEdge::B]  = searchVertex;
                    }
                    else // we looked for a vertex and so will the next
                    {
                        contourPoints.append( v1 ); // v2 was the vertex we were looking for so v1 is added
                        searchVertex        = edge1.a();
                        nextEdge[PEdge::A]  = searchVertex;
                        nextEdge[PEdge::B]  = searchVertex;
                    }

                    if ( endEdge.a() == nextEdge.a() )
                    {
                        faceIndex       = 0;
                        startContour    = true;
                        searchVertex    = -1;

                        contourIndex.append( contourPoints.size() );
                        contourPoints.append( beginVertex );
                    }
                }
                break;
            default:
                qDebug().nospace() << "PGeometry::crossSection: unknown value for nEdges ["<< nEdges << "]";
                break;
            }
        }
        else
        {
            faceIndex++;
        }

        if ( faceIndex >= mFaces.size() )
        {
            faceIndex = 0;
        }

        if ( searchVertex >= 0 )
        {
            int irow = searchVertex * mVertices.size();
            int I1;
            int I2;

            for ( int j = 0; j < mVertices.size(); j++ )
            {
                if ( !mFaceEdgeConnectMatrix.isEmpty() )
                {
                    I1 = mFaceEdgeConnectMatrix[irow + j];
                    I2 = mFaceEdgeConnectMatrix[ j * mVertices.size() + searchVertex ];
                }
                else
                {
                    I1 = findFace(PEdge(searchVertex,j),done);
                    I2 = findFace(PEdge(j,searchVertex),done);
                }

                if ( I1 >= 0 && done[I1] == false )
                {
                    faceIndex = I1;

                    break;
                }
                if ( I2 >= 0 && done[I2] == false )
                {
                    faceIndex = I2;

                    break;
                }
            }
        }
    }

    return contourPoints.size();
}





int PGeometry::crossSection( const PPlane &crossPlane,
                             P3DVertices &contourPoints,
                             PIVector &contourIndex,
                             PVector &contourValues,
                             const PVector &functionvals ) const
{
    int             nDone       = 0;
    int             faceIndex   = 0;
    int             nEdges;
    PEdge           edge1;
    PEdge           edge2;
    PEdge           endEdge;
    PEdge           nextEdge;
    P3DVector       v1;
    P3DVector       v2;
    P3DVector       beginVertex;
    double          beginValue = 0;
    bool            startContour = true;
    int             searchVertex = -1;
    QVector <bool>  done( mFaces.size(), false );


    QVector <PPlane::PlaneSide_type> planeSide(crossPlane.planeSide(mVertices));

    for ( int i =0; i < mFaces.size(); i++)
    {
        if ( planeSide[mFaces[i].I0()] == planeSide[mFaces[i].I1()] &&
             planeSide[mFaces[i].I0()] == planeSide[mFaces[i].I2()] &&
             planeSide[mFaces[i].I0()] != PPlane::ONPLANE &&
             planeSide[mFaces[i].I1()] != PPlane::ONPLANE &&
             planeSide[mFaces[i].I2()] != PPlane::ONPLANE )
        {
            done[i] = true;
            nDone++;
        }
    }

    contourValues.clear();

    contourPoints.clear();
    contourIndex.clear();

    while ( nDone < mFaces.size() )
    {
        if ( !done[faceIndex] )
        {
            done[faceIndex] = true;

            nDone++;

            nEdges = mFaces[ faceIndex ].crossSection( crossPlane, v1, v2, edge1, edge2 );

            switch ( nEdges )
            {
            case  0: // no crossection
                faceIndex++;
                break;

            case  2: // two edges are crossed
                if ( startContour ) // new contour must be started
                {
                    contourIndex.append( contourPoints.size() * -1 );
                    contourPoints.append( v1 );
                    contourIndex.append( contourPoints.size() );
                    contourPoints.append( v2 );

                    if ( !functionvals.isEmpty() )
                    {
                        double fraction = ( v1 - mVertices.at( edge1.a() ) ).length() /
                                          ( mVertices.at( edge1.b() ) - mVertices.at( edge1.a() ) ).length();

                        contourValues.append( functionvals[ edge1.a() ] +
                                              fraction * ( functionvals[ edge1.b() ] - functionvals[ edge1.a() ] ) );

                        beginValue  = contourValues.last();
                        fraction    = ( v2 - mVertices.at( edge2.a() ) ).length() /
                                      ( mVertices.at( edge2.b() ) - mVertices.at( edge2.a() ) ).length();

                        contourValues.append( functionvals[ edge2.a() ] +
                                              fraction * ( functionvals[ edge2.b() ] - functionvals[ edge2.a() ] ) );
                    }

                    // in a consitent geometry the edges are oposite in direction for adjacent faces
                    endEdge         = edge1;
                    nextEdge.invert( edge2 );
                    beginVertex     = v1;
                    if ( !mFaceEdgeConnectMatrix.isEmpty() )
                    {
                        faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                    }
                    else
                    {
                        faceIndex       = findFace( nextEdge, done);
                    }
                    startContour    = false;
                }
                else
                {
                    /* only the second vertex needs to be added, because the first one is added
                     * by the previous face */

                    if ( nextEdge == edge1 )
                    {
                        if ( v2 != beginVertex )
                        {
                            contourIndex.append( contourPoints.size() );
                            contourPoints.append( v2 );

                            if ( !functionvals.isEmpty() )
                            {
                                double fraction = ( v2 - mVertices.at( edge2.a() ) ).length() /
                                                  ( mVertices.at( edge2.b() ) - mVertices.at( edge2.a() ) ).length();

                                contourValues.append( functionvals[ edge2.a() ] +
                                                      fraction * ( functionvals[ edge2.b() ] - functionvals[ edge2.a() ] ) );
                            }
                        }

                        nextEdge.invert( edge2 );
                    }
                    else
                    {
                        if ( v1 != beginVertex )
                        {
                            if ( !functionvals.isEmpty() )
                            {
                                double fraction = ( v1 - mVertices.at( edge1.a() ) ).length() /
                                                  ( mVertices.at( edge1.b() ) - mVertices.at( edge1.a() ) ).length();

                                contourValues.append( functionvals[ edge1.a() ] +
                                                      fraction * ( functionvals[ edge1.b() ] - functionvals[ edge1.a() ] ) );
                            }

                            contourIndex.append( contourPoints.size() );
                            contourPoints.append( v1 );
                        }

                        nextEdge.invert( edge1 );
                    }

                    if ( endEdge == nextEdge )
                    {
                        contourIndex.append( contourPoints.size() );
                        contourPoints.append( beginVertex );

                        if ( !functionvals.isEmpty() )
                        {
                            contourValues.append( beginValue );
                        }

                        faceIndex       = 0; // start looking for a new contour
                        startContour    = true;
                    }
                    else
                    {
                        if ( !mFaceEdgeConnectMatrix.isEmpty() )
                        {
                            faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                        }
                        else
                        {
                            faceIndex       = findFace( nextEdge, done);
                        }
                    }
                }
                break;

            case  -2: // the plane crosses the face through one vertex and the oposite edge
                if ( startContour )
                {
                    contourIndex.append( contourPoints.size() * -1 );
                    contourPoints.append( v2 );
                    contourIndex.append( contourPoints.size() );
                    contourPoints.append( v1 );

                    if ( !functionvals.isEmpty() )
                    {
                        double fraction = ( v2 - mVertices.at(edge2.a()) ).length() /
                                          ( mVertices.at( edge2.b() ) - mVertices.at( edge2.a() ) ).length();

                        contourValues.append( functionvals[edge2.a()] +
                                              fraction * ( functionvals[ edge2.b() ] - functionvals[ edge2.a() ] ) );

                        beginValue = contourValues.last();

                        contourValues.append( functionvals[ edge1.a() ] );
                    }

                    endEdge         = edge2;        // the edge is selected as the endpoint
                    searchVertex    = edge1.a();
                    beginVertex     = v2;
                    startContour    = false;
                }
                else
                {
                    // next an edge crossing or an vertex has to be found
                    contourIndex.append( contourPoints.size() );

                    if ( searchVertex >= 0 ) // we looked for a vertex so the next will be an edge
                    {
                        contourPoints.append( v2 ); // v1 was the vertex we were looking for so v2 is added

                        if ( !functionvals.isEmpty() )
                        {
                            double fraction = ( v2 - mVertices.at( edge2.a() ) ).length() /
                                              ( mVertices.at( edge2.b() ) - mVertices.at( edge2.a() ) ).length();

                            contourValues.append( functionvals[ edge2.a() ] +
                                                  fraction * ( functionvals[ edge2.b() ] - functionvals[ edge2.a() ] ) );
                        }

                        searchVertex    = -1;
                        nextEdge.invert( edge2 );

                        if ( endEdge ==  nextEdge )
                        {
                            contourIndex.append( contourPoints.size() );
                            contourPoints.append( beginVertex );

                            if ( !functionvals.isEmpty() )
                            {
                                contourValues.append( beginValue );
                            }

                            faceIndex       = 0;
                            startContour    = true;
                        }
                        else
                        {
                            if ( !mFaceEdgeConnectMatrix.isEmpty() )
                            {
                                faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                            }
                            else
                            {
                                faceIndex       = findFace( nextEdge, done);
                            }
                        }
                    }
                    else // edge is found so the next one will be an vertex
                    {
                        searchVertex    = edge1.a();
                        contourPoints.append( v1 ); // v2 was the edge we were looking for so v1 is added

                        if ( !functionvals.isEmpty() )
                        {
                            double fraction = ( v1 - mVertices.at( edge1.a() ) ).length() /
                                              ( mVertices.at( edge1.b() ) - mVertices.at( edge1.a() ) ).length();

                            contourValues.append( functionvals[ edge1.a() ] +
                                                  fraction * ( functionvals[ edge1.b() ] - functionvals[ edge1.a() ] ) );
                        }

                        if ( searchVertex == endEdge.a() )
                        {
                            faceIndex       = 0;
                            startContour    = true;

                            contourIndex.append( contourPoints.size() );
                            contourPoints.append( beginVertex );

                            if ( !functionvals.isEmpty() )
                            {
                                contourValues.append( beginValue );
                            }

                            searchVertex = -1;
                        }
                        else
                        {
                            nextEdge[PEdge::A] = searchVertex;
                            nextEdge[PEdge::B] = searchVertex;
                        }
                    }
                }
                break;

            case  -1: // one edge touches the plane
                if ( !mFaceEdgeConnectMatrix.isEmpty() )
                {
                    done[ mFaceEdgeConnectMatrix[ edge1.a() * mVertices.size() + edge1.b() ] ] = true;
                    done[ mFaceEdgeConnectMatrix[ edge1.b() * mVertices.size() + edge1.a() ] ] = true;
                }
                else
                {
                    done[ findFace(edge1, done ) ] = true;
                    done[ findFace(edge1.inverted(), done ) ] = true;
                }

                nDone++;

                if ( startContour )
                {
                    contourIndex.append( contourPoints.size() * -1 );
                    contourPoints.append( v1 );
                    contourIndex.append( contourPoints.size() );
                    contourPoints.append( v2 );

                    if ( !functionvals.isEmpty() )
                    {
                        contourValues.append( functionvals[ edge1.a() ] );

                        beginValue = contourValues.last();

                        contourValues.append( functionvals[ edge2.a() ] );
                    }

                    endEdge[PEdge::A]   = edge1.a();
                    endEdge[PEdge::B]   = edge1.a();
                    searchVertex        = edge1.b();
                    nextEdge[PEdge::A]  = searchVertex;
                    nextEdge[PEdge::B]  = searchVertex;
                    beginVertex         = v1;
                    startContour        = false;
                }
                else
                {
                    contourIndex.append( contourPoints.size() );

                    if ( searchVertex == edge1.a() ) // we looked for a vertex and so will the next
                    {
                        contourPoints.append( v2 ); // v1 was the vertex we were looking for so v2 is added

                        if ( !functionvals.isEmpty( ))
                        {
                            contourValues.append( functionvals[ edge2.a() ] );
                        }

                        searchVertex        = edge1.b();
                        nextEdge[PEdge::A]  = searchVertex;
                        nextEdge[PEdge::B]  = searchVertex;
                    }
                    else // we looked for a vertex and so will the next
                    {
                        contourPoints.append( v1 ); // v2 was the vertex we were looking for so v1 is added
                        if ( !functionvals.isEmpty() )
                        {
                            contourValues.append( functionvals[ edge1.a() ] );
                        }

                        searchVertex        = edge1.a();
                        nextEdge[PEdge::A]  = searchVertex;
                        nextEdge[PEdge::B]  = searchVertex;
                    }

                    if ( endEdge.a() == nextEdge.a() )
                    {
                        faceIndex       = 0;
                        startContour    = true;
                        searchVertex    = -1;

                        contourIndex.append( contourPoints.size() );
                        contourPoints.append( beginVertex );

                        if ( !functionvals.isEmpty() )
                        {
                            contourValues.append( beginValue );
                        }
                    }
                }
                break;

            default:
                qDebug().nospace() << "PGeometry::crossSection2: unknown value for nEdges ["<< nEdges << "]";
                break;
            }
        }
        else
        {
            faceIndex++;
        }

        if ( faceIndex >= mFaces.size() )
        {
            faceIndex = 0;
        }

        if ( searchVertex >= 0 )
        {
            int irow = searchVertex * mVertices.size();
            int I1;
            int I2;

            for ( int j = 0; j < mVertices.size(); j++ )
            {
                if ( !mFaceEdgeConnectMatrix.isEmpty() )
                {
                    I1 = mFaceEdgeConnectMatrix[irow + j];
                    I2 = mFaceEdgeConnectMatrix[ j * mVertices.size() + searchVertex ];
                }
                else
                {
                    I1 = findFace(PEdge(searchVertex,j),done);
                    I2 = findFace(PEdge(j,searchVertex),done);
                }

                if ( I1 >= 0 && done[I1] == false )
                {
                    faceIndex = I1;

                    break;
                }

                if ( I2 >= 0 && done[I2] == false )
                {
                    faceIndex = I2;

                    break;
                }
            }
        }
    }
    return contourPoints.size();
}


int PGeometry::crossFunctionVal( const PVector &functionval,
                                 double curfunctionVal,
                                 P3DVertices &contourPoints,
                                 QList<int> &contourIndex ) const
{
    int             nDone       = 0;
    int             faceIndex   = 0;
    int             nEdges;
    PEdge           edge1;
    PEdge           edge2;
    PEdge           endEdge;
    PEdge           nextEdge;
    P3DVector       v1;
    P3DVector       v2;
    P3DVector       beginVertex;
    bool            startContour = true;
    int             searchVertex = -1;
    QVector <bool>  done( mFaces.size(), false );

//    QVector <PPlane::PlaneSide_type> planeSide(crossPlane.planeSide(mVertices));

//    for ( int i =0; i < mFaces.size(); i++)
//    {
//        if ( planeSide[mFaces[i].I0()] == planeSide[mFaces[i].I1()] &&
//             planeSide[mFaces[i].I0()] == planeSide[mFaces[i].I2()] )
//        {
//            done[i] = true;
//            nDone++;
//        }
//    }
//    if ( !mFaceEdgeConnectMatrix.isEmpty() )//checkFaceEdgeConnectmatrix() )
//    if ( checkFaceEdgeConnectmatrix() )
//    {
        contourPoints.clear();
        contourIndex.clear();

        while ( nDone < mFaces.size() )
        {
            if ( !done[faceIndex] )
            {
                done[faceIndex] = true;

                nDone++;

                nEdges = mFaces[ faceIndex ].crossFaceFunctionVal( curfunctionVal,
                                                                   functionval.at( mFaces[ faceIndex ].I0() ),
                                                                   functionval.at( mFaces[ faceIndex ].I1() ),
                                                                   functionval.at( mFaces[ faceIndex ].I2() ),
                                                                   v1, v2, edge1, edge2 );

                switch ( nEdges )
                {
                case 0: // no crossection
                    faceIndex++;
                    break;

                case 2: // two edges are crossed
                    if ( startContour ) // new contour must be started
                    {
                        contourIndex.append( contourPoints.size() * -1 );
                        contourPoints.addVertex( v1, false );
                        contourIndex.append( contourPoints.size() );
                        contourPoints.addVertex( v2, false );

                        // in a consitent geometry the edges are oposite in direction for adjacent faces
                        endEdge         = edge1;
                        nextEdge.invert( edge2 );
                        beginVertex     = v1;
                        if ( !mFaceEdgeConnectMatrix.isEmpty() )
                        {
                            faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                        }
                        else
                        {
                            faceIndex       = findFace( nextEdge, done);
                        }
                        startContour    = false;
                    }
                    else
                    {
                        /* only the second vertex needs to be added, because the first one is added
                         * by the previous face */
                        contourIndex.append( contourPoints.size() );

                        if ( nextEdge == edge1 )
                        {
                            contourPoints.addVertex( v2, false );

                            nextEdge.invert( edge2 );
                        }
                        else
                        {
                            contourPoints.addVertex( v1, false );

                            nextEdge.invert( edge1 );
                        }

                        if ( endEdge == nextEdge )
                        {
                            contourIndex.append( contourPoints.size() );
                            contourPoints.addVertex( beginVertex, false );

                            faceIndex       = 0; // start looking for a new contour
                            startContour    = true;
                        }
                        else
                        {
                            if ( !mFaceEdgeConnectMatrix.isEmpty() )
                            {
                                faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                            }
                            else
                            {
                                faceIndex       = findFace( nextEdge, done);
                            }
                        }
                    }
                    break;

                case -2: // the plane crosses the face through one vertex and the oposite edge
                    if ( startContour )
                    {
                        contourIndex.append( contourPoints.size() * -1 );
                        contourPoints.addVertex( v2, false );
                        contourIndex.append( contourPoints.size() );
                        contourPoints.addVertex( v1, false );

                        endEdge         = edge2; // the edge is selected as the endpoint
                        searchVertex    = edge1.a();
                        beginVertex     = v2;
                        startContour    = false;
                    }
                    else
                    {
                        // next an edge crossing or an vertex has to be found
                        contourIndex.append( contourPoints.size() );

                        if ( searchVertex >= 0 ) // we looked for a vertex so the next will be an edge
                        {
                            contourPoints.addVertex( v2, false ); // v1 was the vertex we were looking for so v2 is added

                            searchVertex = -1;

                            nextEdge.invert( edge2 );

                            if ( endEdge ==  nextEdge )
                            {
                                contourIndex.append( contourPoints.size() );
                                contourPoints.addVertex( beginVertex, false );

                                faceIndex       = 0;
                                startContour    = true;
                            }
                            else
                            {
                                if ( !mFaceEdgeConnectMatrix.isEmpty() )
                                {
                                    faceIndex       = mFaceEdgeConnectMatrix[ nextEdge.a() * mVertices.size() + nextEdge.b() ];
                                }
                                else
                                {
                                    faceIndex       = findFace( nextEdge, done);
                                }
                            }
                        }
                        else // edge is found so the next one will be an vertex
                        {
                            searchVertex = edge1.a();

                            contourPoints.addVertex( v1, false ); // v2 was the edge we were looking for so v1 is added

                            if ( searchVertex == endEdge.a() )
                            {
                                faceIndex       = 0;
                                startContour    = true;
                                contourIndex.append( contourPoints.size() );
                                contourPoints.addVertex( beginVertex, false );
                                searchVertex    = -1;
                            }
                            else
                            {
                                nextEdge[PEdge::A] = searchVertex;
                                nextEdge[PEdge::B] = searchVertex;
                            }
                        }
                    }
                    break;

                case  -1: // one edge touches the plane
                    if (!mFaceEdgeConnectMatrix.isEmpty() )
                    {
                        done[ mFaceEdgeConnectMatrix[ edge1.a() * mVertices.size() + edge1.b() ] ] = true;
                        done[ mFaceEdgeConnectMatrix[ edge1.b() * mVertices.size() + edge1.a() ] ] = true;
                    }
                    else
                    {
                        done[ findFace(edge1,done) ] = true;
                        done[ findFace(edge1.inverted(),done) ] = true;
                    }
                    nDone++;

                    if ( startContour )
                    {
                        contourIndex.append( contourPoints.size() * -1 );
                        contourPoints.addVertex( v1, false );
                        contourIndex.append( contourPoints.size() );
                        contourPoints.addVertex( v2, false );

                        endEdge[PEdge::A]   = edge1.a();
                        endEdge[PEdge::B]   = edge1.a();
                        searchVertex        = edge1.b();
                        nextEdge[PEdge::A]  = searchVertex;
                        nextEdge[PEdge::B]  = searchVertex;
                        beginVertex         = v1;
                        startContour        = false;
                    }
                    else
                    {
                        contourIndex.append( contourPoints.size() );

                        if ( searchVertex == edge1.a() ) // we looked for a vertex and so will the next
                        {
                            contourPoints.addVertex( v2, false ); // v1 was the vertex we were looking for so v2 is added

                            searchVertex        = edge1.b();
                            nextEdge[PEdge::A]  = searchVertex;
                            nextEdge[PEdge::B]  = searchVertex;
                        }
                        else // we looked for a vertex and so will the next
                        {
                            contourPoints.addVertex( v1, false ); // v2 was the vertex we were looking for so v1 is added

                            searchVertex        = edge1.a();
                            nextEdge[PEdge::A]  = searchVertex;
                            nextEdge[PEdge::B]  = searchVertex;
                        }

                        if ( endEdge.a() == nextEdge.a() )
                        {
                            faceIndex       = 0;
                            startContour    = true;
                            searchVertex    = -1;

                            contourIndex.append( contourPoints.size() );
                            contourPoints.addVertex( beginVertex, false );
                        }
                    }
                    break;

                default:
                    qDebug().nospace() << "PGeometry::crossFunctionVal: unknown value for nEdges ["<< nEdges << "]";
                    break;
                }
            }
            else
            {
                faceIndex++;
            }

            if (faceIndex >= mFaces.size() )
            {
                faceIndex = 0;
            }

            if ( searchVertex >= 0 )
            {
                int irow = searchVertex*mVertices.size();
                int I1;
                int I2;

                for ( int j = 0; j < mVertices.size(); j++ )
                {
                    if ( !mFaceEdgeConnectMatrix.isEmpty() )
                    {
                        I1 = mFaceEdgeConnectMatrix[irow + j];
                        I2 = mFaceEdgeConnectMatrix[ j * mVertices.size() + searchVertex ];
                    }
                    else
                    {
                        I1 = findFace(PEdge(searchVertex,j),done);
                        I2 = findFace(PEdge(j,searchVertex),done);
                    }

                    if ( I1 >= 0 &&
                         !done[I1] )
                    {
                        faceIndex = I1;

                        break;
                    }

                    if ( I2 >= 0 &&
                         !done[I2] )
                    {
                        faceIndex = I2;

                        break;
                    }
                }
            }
        }
//    }
//    else
//    {
//        qDebug().nospace() << "PGeometry::crossFunctionVal: inconsitent geometry";
//    }

    return contourPoints.size();
}


bool PGeometry::lineGeomSect( const P3DVector& p1, const P3DVector& p2, P3DVector& crossPoint ) const
{
//    int     hit;
//    double  alpha;
//    double  lambda;
//    double  mu;
//    double  minAlpha = HUGE_VAL;
    int vertex,face;
    crossPoint = linetrisect(p1,p2,vertex,face);
    return vertex >=0;
//    for ( int i = 0; i < mFaces.size(); i++ )
//    {
//        hit = mFaces[i].linetrisect( p1, p2, alpha, lambda, mu );

//        if ( hit != 0 && alpha <= minAlpha )
//        {
//            minAlpha    = alpha;
//            crossPoint  = mFaces[i].pointOnFace( lambda, mu );
//        }
//    }

//    return  minAlpha != HUGE_VAL;
}


bool PGeometry::linetrisect( const P3DVector& p1, const P3DVector& p2 ) const
{
    for ( int i = 0; i < mFaces.size(); i++ )
    {
        if ( mFaces[i].linetrisect( p1, p2 ) )
        {
            return true;
        }
    }

    return  false;
}


bool PGeometry::linetrisect( const P3DVector& p1, const P3DVector& p2, int& faceIndex ) const
{
    for ( int i = 0; i < mFaces.size(); i++ )
    {
        if ( mFaces[i].linetrisect( p1, p2 ) )
        {
            faceIndex = i;

            return true;
        }
    }

    return  false;
}


bool PGeometry::selfsect( int i1, int i2, int& faceIndex ) const
{
    for ( int i = 0; i < mFaces.size(); i++ )
    {
        if ( !mFaces[i].contains( i1 ) &&
             !mFaces[i].contains( i2 ) &&
              mFaces[i].inFaceSpace( mVertices.at( i1 ) ) &&
              mFaces[i].inFaceSpace( mVertices.at( i2 ) ) )
        {
            if ( mFaces[i].linetrisect( mVertices.at( i1 ), mVertices.at( i2 ) ) )
            {
                faceIndex = i;

                return true;
            }
        }
    }

    return  false;
}


P3DVector PGeometry::linetrisect( const P3DVector& p1, const P3DVector& p2, int& closestVertex, int& faceIndex ) const
{
    int         hit;
    double      minAlpha    = 10000000;
    double      alpha;
    double      lambda;
    double      mu;
    double      keeplambda  = 0;
    double      keepmu      = 0;
    P3DVector   retVect;

    closestVertex   = -1;
    faceIndex       = -1;

    for ( int i = 0; i < mFaces.size(); i++ )
    {
//        if ( p1 != mFaces[i].vertex0() && p1 != mFaces[i].vertex1() && p1 != mFaces[i].vertex2() &&
//             p2 != mFaces[i].vertex0() && p2 != mFaces[i].vertex1() && p2 != mFaces[i].vertex2() )
//        {
            hit = mFaces[i].linetrisect( p1, p2, alpha, lambda, mu );

            if ( hit != 0 && alpha <= minAlpha )
            {
                faceIndex   = i;
                minAlpha    = alpha;
                keeplambda  = lambda;
                keepmu      = mu;
            }
//        }
    }

    if ( faceIndex >= 0 )
    {
        retVect = mFaces[faceIndex].pointOnFace( keeplambda, keepmu );

        if ( ( mFaces[faceIndex].vertex0() - retVect ).length() < ( mFaces[faceIndex].vertex1() - retVect ).length() )
        {
            if ( ( mFaces[faceIndex].vertex0() - retVect ).length() < ( mFaces[faceIndex].vertex2() - retVect ).length() )
            {
                closestVertex = mFaces[faceIndex].I0();
            }
            else
            {
                closestVertex = mFaces[faceIndex].I2();
            }
        }
        else if ( ( mFaces[faceIndex].vertex1() - retVect ).length() < ( mFaces[faceIndex].vertex2() - retVect ).length() )
        {
            closestVertex = mFaces[faceIndex].I1();
        }
        else
        {
            closestVertex = mFaces[faceIndex].I2();
        }
    }

    return retVect;
}


P3DVector PGeometry::linetrisect( const P3DVector& p1,
                                  const P3DVector& p2,
                                  int& closestVertex,
                                  int& faceIndex,
                                  const PPlane &crossPlane ) const
{
    int         hit;
    double      minAlpha    = 10000000;
    double      alpha;
    double      lambda;
    double      mu;
    double      keeplambda  = 0;
    double      keepmu      = 0;
    P3DVector   retVect;

    closestVertex   = -1;
    faceIndex       = -1;

    for ( int i = 0; i < mFaces.size(); i++ )
    {
//        if ( p1 != mFaces[i].vertex0() && p1 != mFaces[i].vertex1() && p1 != mFaces[i].vertex2() &&
//             p2 != mFaces[i].vertex0() && p2 != mFaces[i].vertex1() && p2 != mFaces[i].vertex2() )
//        {
            double tmp1 = mFaces[i].vertex0().dot( crossPlane.vector() ) + crossPlane.pos();
            double tmp2 = mFaces[i].vertex1().dot( crossPlane.vector() ) + crossPlane.pos();
            double tmp3 = mFaces[i].vertex2().dot( crossPlane.vector() ) + crossPlane.pos();

            if ( tmp1 >= 0 && tmp2 >= 0 && tmp3 >= 0 )
            {
                hit = mFaces[i].linetrisect( p1, p2, alpha, lambda, mu );

                if ( hit != 0 && alpha <= minAlpha )
                {
                    faceIndex   = i;
                    minAlpha    = alpha;
                    keeplambda  = lambda;
                    keepmu      = mu;
                }
            }
//        }
    }

    if ( faceIndex >= 0 )
    {
        retVect = mFaces[faceIndex].pointOnFace( keeplambda, keepmu );

        if ( ( mFaces[faceIndex].vertex0() - retVect ).length() < ( mFaces[faceIndex].vertex1() - retVect ).length() )
        {
            if ( ( mFaces[faceIndex].vertex0() - retVect ).length() < ( mFaces[faceIndex].vertex2() - retVect ).length() )
            {
                closestVertex = mFaces[faceIndex].I0();
            }
            else
            {
                closestVertex = mFaces[faceIndex].I2();
            }
        }
        else if ( ( mFaces[faceIndex].vertex1() - retVect ).length() < ( mFaces[faceIndex].vertex2() - retVect ).length() )
        {
            closestVertex = mFaces[faceIndex].I1();
        }
        else
        {
            closestVertex = mFaces[faceIndex].I2();
        }
    }

    return  retVect;
}


int PGeometry::linetrisect( const P3DVector& p1,
                            const P3DVector& p2,
                            double& keeplambda,
                            double& keepmu,
                            const PPlane &crossPlane ) const
{
    int         hit;
    double      minAlpha    = 10000000;
    double      alpha;
    double      lambda;
    double      mu;
    keeplambda  = 0;
    keepmu      = 0;

    int faceIndex       = -1;

    for ( int i = 0; i < mFaces.size(); i++ )
    {
        double tmp1 = mFaces[i].vertex0().dot( crossPlane.vector() ) + crossPlane.pos();
        double tmp2 = mFaces[i].vertex1().dot( crossPlane.vector() ) + crossPlane.pos();
        double tmp3 = mFaces[i].vertex2().dot( crossPlane.vector() ) + crossPlane.pos();

        if ( tmp1 >= 0 && tmp2 >= 0 && tmp3 >= 0 )
        {
            hit = mFaces[i].linetrisect( p1, p2, alpha, lambda, mu );

            if ( hit != 0 && alpha <= minAlpha )
            {
                faceIndex   = i;
                minAlpha    = alpha;
                keeplambda  = lambda;
                keepmu      = mu;
            }
        }
    }
    return  faceIndex;
}

bool  PGeometry::geometryPlaneCrossing( const PPlane &crossPlane ) const
{
    bool retVal = false;

    for ( int i = 0; i < nFaces(); i++ )
    {
        if  ( mFaces[i].facePlaneCrossing( crossPlane ) == 1 )
        {
            retVal = true;

            break;
        }
    }

    return retVal;
}


void PGeometry::rotate( const P3DVector &axis, double angle, const P3DVector &rotPnt )
{
    mVertices.rotate( axis, angle, rotPnt );

    updateMembers();
}


void PGeometry::rotate( double phi, double theta, double gamma, const P3DVector &rotPnt )
{
    mVertices.rotate( phi, theta, gamma, rotPnt );

    updateMembers();
}


void PGeometry::rotate( const P3DVector &angles, const P3DVector &rotPnt )
{
    mVertices.rotate( angles, rotPnt );

    updateMembers();
}


void PGeometry::invertNormals( void )
{
    for ( int i = 0; i < mFaces.size(); i++ )
    {
        mFaces[i].invertNormal();
    }

    update();
}


void PGeometry::changeXYaxis( void )//change the XY axis: y=-x and x = y
{
    mVertices.changeXYaxis();

    updateMembers();
}


void PGeometry::changeYXaxis( void ) //change the XY axis: y=-x and x = y
{
    mVertices.changeYXaxis();

    updateMembers();
}


double PGeometry::area( void ) const
{
    double a = 0;

    for ( int i = 0 ; i < mFaces.size(); i++ )
    {
        a += mFaces.at( i ).area();
    }

    return a;
}


PVector PGeometry::wallThickness( double minWallThickness ) const
{
    int         faceI;
    int         closestVertex;
    P3DVector   sectPnt;
    PVector     wallThick( mVertices.size(), 50.0 );

    if ( !mNeighbors.isEmpty() )
    {
        for ( int i = 0; i < wallThick.N(); i++ )
        {
            sectPnt = linetrisect( mVertices.at( i ),
                                   mVertices.at( i ) - ( mVerticesNrm.at( i ) * 50.0 ) ,
                                   closestVertex,
                                   faceI );

            if ( faceI >= 0 &&
                 !mNeighbors.at( i ).contains( mFaces.at( faceI ).I0() ) &&
                 !mNeighbors.at( i ).contains( mFaces.at( faceI ).I1() ) &&
                 !mNeighbors.at( i ).contains( mFaces.at( faceI ).I2() ) &&
                  mVerticesNrm.at( i ).dot( faceNormal( faceI ) ) < 0 )
            {
                wallThick[i] = qMax( minWallThickness, (mVertices.at( i ) - sectPnt).length() );

                if ( mVerticesNrm.at(i).dot( sectPnt - mVertices.at( i ) ) > 0 )
                {
                    wallThick[i] *= -1.0;
                }
            }
        }
    }

    return wallThick;
}

int PGeometry::isConnected( int i, int j ) const
{
//    Q_ASSERT( !mFaceEdgeConnectMatrix.isEmpty() );  // Test just to be sure
    Q_ASSERT( i >= 0 && i < mVertices.size() );  // Test just to be sure
    Q_ASSERT( j >= 0 && j < mVertices.size() );  // Test just to be sure
    if ( !mFaceEdgeConnectMatrix.isEmpty() )
    {
        return mFaceEdgeConnectMatrix[ i * mVertices.size() + j ] ;
    }
    else
    {
        return findFace(i,j);
    }

//
}


#ifndef QT_NO_DEBUG


const QList<int> &PGeometry::neighbors( int vertex ) const
{
    Q_ASSERT( !mNeighbors.isEmpty() );  // Test just to be sure
    Q_ASSERT( vertex >= 0 && vertex < mNeighbors.size() );  // Test just to be sure

    return mNeighbors.at( vertex );
}


const P3DVector& PGeometry::faceNormal( int Index ) const
{
    Q_ASSERT( !mFaces.isEmpty() );  // Test just to be sure
    Q_ASSERT( Index >= 0 && Index < mFaces.size() );  // Test just to be sure

    return mFaces[Index].normal();
}


const PFace& PGeometry::face( int Index ) const
{
    Q_ASSERT( !mFaces.isEmpty() );  // Test just to be sure
    Q_ASSERT( Index >= 0 && Index < mFaces.size() );  // Test just to be sure

    return mFaces[Index];
}


PFace& PGeometry::rface( int Index )
{
    Q_ASSERT( !mFaces.isEmpty() );  // Test just to be sure
    Q_ASSERT( Index >= 0 && Index < mFaces.size() );  // Test just to be sure

    return mFaces[Index];
}


#else

const P3DVector& PGeometry::faceNormal( int Index ) const
{
    return mFaces[Index].normal();
}

const PFace& PGeometry::face( int Index ) const
{
    return mFaces[Index];
}

PFace& PGeometry::rface( int Index )
{
    return mFaces[Index];
}

#endif

