/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pgeometry.h 1152 2014-03-08 17:26:41Z peter $
 *
 */

#ifndef __PGEOMETRY_H
#define __PGEOMETRY_H


#include "P3DVertices.h"
#include "PIVector.h"
#include "pfile.h"
#include "pedge.h"
#include "pplane.h"

#include "PeacsSupport_global.h"

namespace Peacs
{
    namespace Support
    {
        class PFace;

        /** @class PGeometry PGeometry.h "support/pgeometry.h"
          * @brief This class describes a triangulated surface
          */
        class PEACSSUPPORTSHARED_EXPORT PGeometry : public PFile
        {

        public:

            /** @brief Default constructor.
              * Initialze all members to 0
              */
            PGeometry( void );


            /** @brief Constructor.
              * If the file could not be read the default constructor is used.
              * @param filename is the name of the triagulation file
              * @see load(QString FileName)
              */
            PGeometry( QString filename );


            /** @brief Constructor.
              * Copies the given geometry s2
              * @param s2 the geometry that is copied
              * @see operator = (const PGeometry &geom)
              */
            PGeometry( const PGeometry& s2 );


            /** @brief Constructor.
              * Constructs a geometry from a (consitstent) ste of vertices and face described by
              * vertex indeces
              * @param  vertices a list of vertices
              * @param faceIndex the list of vertex indices to build the faces
              */
            PGeometry( const P3DVertices &vertices, const QVector<int> &faceIndex );


            /** @brief Constructor.
              * Constructs a geometry from a (consitstent) ste of vertices and face described by
              * vertex indeces
              * @param  vertices a list of vertices
              * @param faceIndex the list of vertex indices to build the faces
              */
            PGeometry( const P3DVertices &vertices, const int * &triangleIndices, int numTriangles);


            /** @brief Destructor.
              * Frees all memory
              */
            virtual ~PGeometry( void );


            /** @brief Initialize all derived members
              */
            void update( void );


            /** @brief compute normals
              * Compute all normals
              * @see mVerticesNrm
              * @see const P3DVector& vertexNormal(int Index)
              * @see const P3DVector& faceNorm(int Index)
              */
            void updateMembers( void );


            /** @brief recompute all relevant members due to a movement of one of the vertices,
              * removal of a vertex / face
              * @param vertexChanged if true only the vertex has been changed its position
              */
            virtual void geometryChanged( bool vertexChanged );


            /** @brief virtual function that checks the data
              * returns an empty Stringlist if all data is correct
              * otherwise errormessages are created
              * @see PFile
              */
            virtual QStringList checkData( void ) const;


            /**  @brief copy operator.
              *  Copies the vertices and the faces
              * IMPORTANT: All faces hold a pointer to its vertices. When compieing these
              * have to be determined again for the copied vertices
              * @param geom the copied geometry
              */
            virtual const PGeometry& operator= ( const PGeometry &geom );


            /**  @brief copy operator for the vertices only.
              *  Copies the vertices if the number of  vertices is equal
              * note that the geometry is not initialized.
              * @param geom the geometry from which the vertices are copied
              */
            void setVertices( const PGeometry &geom );


            /** @brief copy operator for the vertices only.
              * Copies the vertices if the number of  vertices is equal
              * note that the geometry is not initialized.
              * @param vertices the vertices from which the vertices are copied
              */
            void setVertices( const P3DVertices &vertices);


            /** @brief equal operator
              * @returns true iff all 3d coordinates of both
              * geometries are the same ortherwise false
              * @param geom the geometry that is compared
              */
            virtual bool operator== ( const PGeometry &geom ) const;


            /** @brief unequal operator.
              * @returns true iff one of the 3d coordinates of both
              * geometries differ ortherwise false
              * @param geom the geometry that is compared
              */
            virtual bool operator!= ( const PGeometry &geom ) const;


            /** @brief add operator
              * @param addgeom the geometry to add
              * @return this object
              */
            const PGeometry & operator += ( const PGeometry &addgeom );


            /** @brief Returns the Number of Faces of the Geometry
              */
            inline int nFaces( void ) const { return mFaces.size(); }


            /** @brief Returns the Number of Vertices of the Geometry
              */
            inline int nVertices( void ) const { return mVertices.size(); }


            /** @brief Returns vertex at location Index.
              * @param Index the location of the vertex
              */
            inline const P3DVector& vertex( int Index ) const { return mVertices.at( Index ); }


            /** @brief Returns vertex at location Index.
              * @param Index the location of the vertex
              */
            inline P3DVector& rvertex( int Index ) { return mVertices[Index]; }


            /** @brief Returns the reference to this vertex at location Index.
              * WARNING No updates are passed on using the PSignalInterface The calling function has to take care
              * @param Index the location of the vertex
              */
            inline void setVertex( int Index, const P3DVector &pos ) { mVertices[Index] = pos; }


            /** @brief Returns the normal at vertex at location Index.
              * @param Index the location of the vertex area
              */
            inline const P3DVector& vertexNormal( int Index ) const { return mVerticesNrm.at( Index ); }


            /** @brief Returns the area at vertex at location Index.
              * @param Index the location of the vertex area
              */
            inline double vertexArea( int Index ) const { return mVerticesArea[Index]; }


#ifndef QT_NO_DEBUG
            /** @returns a list of vertices connecte to node vertex
              * @param vertx the index of the node
              */
            const QList<int> &neighbors( int vertex ) const;


            /** @brief Returns the normal of the face at location Index.
              * If out of range throws an exception
              * @param Index the location of the face normal
              */
            const P3DVector& faceNormal( int Index ) const;


            /** @brief Returns face at location Index.
              * If out of range throws an exception
              * @param Index the location of the face
              */
            const PFace& face( int Index ) const;


            /** @brief Returns face at location Index.
              * If out of range throws an exception
              * @param Index the location of the face
              */
            PFace& rface( int Index );

#else

            /** @returns a list of vertices connecte to node vertex
              * @param vertx the index of the node
              */
            inline const QList<int> &neighbors( int vertex ) const { return mNeighbors.at( vertex ); }


            /** @brief Returns the normal of the face at location Index.
              * If out of range throws an exception
              * @param Index the location of the face normal
              */
//            inline const P3DVector& faceNormal( int Index ) const { return mFaces[Index].normal(); }
            const P3DVector& faceNormal( int Index ) const;


            /** @brief Returns face at location Index.
              * If out of range throws an exception
              * @param Index the location of the face
              */
//            inline const PFace& face( int Index ) const { return mFaces[Index]; }
            const PFace& face( int Index ) const;


            /** @brief Returns face at location Index.
              * If out of range throws an exception
              * @param Index the location of the face
              */
            PFace& rface( int Index );


#endif

            /** @brief returns the face index if the two vertices are connected otherwise -1
              */
            int isConnected( int i, int j ) const;


            /** @brief Returns the refernce to the list of vertices
              */
            inline const P3DVertices& getVertices( void ) const { return mVertices; }


            /** @brief Returns the normalsat all vertices
              */
            inline const P3DVertices& vertexNormals( void ) const { return mVerticesNrm; }


            /** @brief Returns the area at all vertices
              */
            inline const PVector &verticesArea( void ) const { return mVerticesArea; }


            /** @brief Limit the precision of the vertices
              * @param precision the number of digits (default 1)
              */
            void limitPrecision( );

            /** @brief Limit the precision of the vertices, in some cases this might lead to a reduction in triangles
              * because vertices coincide
              * @param precision the number of digits (default 1)
              * @returns true if the geometry has been adapted else false
              */
            bool checkLimitPrecision();


            /** @brief get the number of digits (precision)
              */
            inline double getPrecisionValue() const { return pow( 10.0, (double)mPrecision*-1.0); }

            /** @brief get the number of digits (precision)
              */
            inline int getPrecision() const { return mPrecision; }


            /** @brief set the number of digits (precision)
              */
            inline bool setPrecision(int precision) { mPrecision = qMax(0,precision); return checkLimitPrecision(); }


            /** @brief Returns the normals smoothe by avaring the neigbors
              */
            P3DVertices smoothedNormals() const;


            /** @brief determine if the given vertex is part of this geometry
              * @param vertex the vertex that is looked for
              * @returns the index of if the vertex is foudn, otherwise-1
              */
            inline int findVertex( const P3DVector & vertex ) const { return mVertices.findVertex( vertex ); }


            /** @brief determine if the given vertex is part of this geometry
              * @param vertex the vertex that is looked for
              * @returns the index of if the vertex is foudn, otherwise-1
              */
            int findVertexWithPrecision( const P3DVector & vertex ) const;


            /** @brief Adds a vertex to the end of the vertex list
              * @param vertex The vertex to be added
              * @param doUpdate if true update the geometry. If false do not forget to update afterwards
              */
            void addVertex( const P3DVector& vertex, bool doUpdate );


            /** @brief Adds a vertex to the end of the vertex list
              * @param vertex The vertex to be added
              * @param doUpdate if true update the geometry. If false do not forget to update afterwards
              */
            void insertVertex( int i, const P3DVector& vertex, bool doUpdate );


            /** @brief removes the vertex at index i from the vertex list also deletes the connected faces
              * @param i vertex at index to be deleted
              * @param doUpdate if true update the geometry. If false do not forget to update afterwards
              */
            void removeVertex( int i, bool doUpdate );


            /** @brief removes all vertices not used by any face
              * @param doUpdate if true update the geometry. If false do not forget to update afterwards
              */
            void removeUnusedVertices( bool doUpdate );


            /** @brief Adds a face to the end of the face list
              *  @param v0 The first vertex of the vertices of face to be added
              *  @param v1 The second vertex of the vertices of face to be added
              *  @param v2 The third vertex of the vertices of face to be added
              *  @param doUpdate if true update the geometry. If false do not forget to update afterwards
              */
            void addFace( const P3DVector &v0, const P3DVector &v1, const P3DVector &v2, bool doUpdate );


            void changeVertexNumber(int oldIndex,int newIndex);


            /** @brief updates face with the given verttex indices
              *  @param i0 The first vertex index of the vertices of face to be added
              *  @param i1 The second vertex index of the vertices of face to be added
              *  @param i2 The third vertex index of the vertices of face to be added
              */
            void updateFace( int face, int i0, int i1, int i2 );


            /** @brief removes a face
              *  @param face The index of the face to be removed
              *  @param doUpdate if true update the geometry. If false do not forget to update afterwards
              */
            void removeFace( int face, bool doUpdate );


            /** @brief removes faces
              *  @param faces The list of indexes of the faces to be removed
              *  @param doUpdate if true update the geometry. If false do not forget to update afterwards
              */
            void removeFaces( QList<int> &faces, bool doUpdate );


            /** @brief Check if the vertices are given in m or mm
              * If the maximum x value is smaller than 1, it is anticipated
              * that the coordinates are given in meters. If so, all
              * coordinates are multiplied by 1000.
              */
            void checkScale( void );


            /** @brief compute the neighbors of each node
              */
            void computeVertexNeighbors( void );


            /** @brief Computes and edge list from the connect Vertex matrix, all edges are listed only once
              */
            inline const QList <PEdge> &getEdgeList( void ) const { return mEdgeList; }


            /** @brief find the face with the same vertices as the given face
              * @param face of  an other mesh
              * @returns the index of face that contains the same vertices*/
            int findFace(const PFace & face) const;

            /** @brief find the face with at least two of the same vertices as the given face
              * @param face of  an other mesh
              * @returns the index of face that contains edge*/
            int findEdge(const PFace & face) const;

            /** @brief find the face with at least two of the same vertices as the given face
              * @param face of  an other mesh
              * @returns the index of face that contains edge*/
            int findEdge(const P3DVector & v1,const P3DVector & v2,const P3DVector & v3) const;



            /** @brief find the face with the given edge from i0 to i1
              * @param i0 first index
              * @param i1 second index
              * @returns the index of face that contains this edge*/
            int findFace(int i0, int i1) const;


            /** @brief find the face with the given edge from i0 to i1
              * @param i0 first index
              * @param i1 second index
              * @returns the index of face that contains this edge*/
            int findFace(const PEdge & edge, const QVector <bool>  skip) const;



            /** @brief find the face with the given edge from i0 to i1
              * @param i0 first index
              * @param i1 second index
              * @returns the index of face that contains this edge*/
            int findFace(int i0, int i1, int & otherVertexIndex) const;


            /** @brief checks if the geometry is empty
              * @returns true if the geometry is empty.
              */
            bool isEmpty( void ) const;


            /** @brief find all intersections of linesegment running from locations (vectors)
              * p1 to p2 in 3D space with all triangles of a triangulated surface
              * @param p1 the first point on a line
              * @param p2 the second point on a line
              * @param closestVertex the closest vertx of the geomtry
              * @param faceIndex of the face with which the lines intersects
              * @returns the point where the line inetersects
              */
            P3DVector linetrisect( const P3DVector& p1, const P3DVector& p2, int& closestVertex, int& faceIndex ) const;


            /** @brief find all intersections of linesegment running from locations (vectors)
              * p1 to p2 in 3D space with all triangles of a triangulated surface
              * @param p1 the first point on a line
              * @param p2 the second point on a line
              * @param closestVertex the closest vertx of the geomtry
              * @param faceIndex of the face with which the lines intersects
              * @param crossPlane if a face/vertex lies before the crossplane it should not be selected
              * @returns the point where the line inetersects
              */
            P3DVector linetrisect( const P3DVector& p1, const P3DVector& p2, int& closestVertex, int& faceIndex, const PPlane &crossPlane ) const;


            /** @brief find all intersections of linesegment running from locations (vectors)
              * p1 to p2 in 3D space with all triangles of a triangulated surface
              * @param p1 the first point on a line
              * @param p2 the second point on a line
              * @param lambda ratio along the the first edge
              *              (mVertices[1]-mVertices[0]) to the projection point
              * @param mu     ratio along the the first edge
              *              (mVertices[1]-mVertices[0]) to the projection point
              * @param crossPlane if a face/vertex lies before the crossplane it should not be selected
              * @returns the faceIndex of the face with which the lines intersects, if not found -1 is returned
              */
            int linetrisect( const P3DVector& p1, const P3DVector& p2, double& lambda, double& mu, const PPlane &crossPlane ) const;



            /** @brief find all intersections of linesegment running from locations (vectors)
              * p1 to p2 in 3D space with all triangles of a triangulated surface Note that if
              * p1 or p2 coincides with one of the vertices this function still returns false!!
              * @param p1 the first point on a line
              * @param p2 the second point on a line
              * @returns if the point intersects
              */
            bool linetrisect( const P3DVector& p1, const P3DVector& p2 ) const;


            /** @brief find all intersections of linesegment running from locations (vectors)
              * p1 to p2 in 3D space with all triangles of a triangulated surface Note that if
              * p1 or p2 coincides with one of the vertices this function still returns false!!
              * @param p1 the first point on a line
              * @param p2 the second point on a line
              * @param faceIndex the first face found of the face with which the line intersects
              * @returns if the point intersects
              */
            bool linetrisect( const P3DVector& p1, const P3DVector& p2, int& faceIndex ) const;


            /** @brief find all intersections of linesegment running from locations (vectors)
              * p1 to p2 in 3D space with all triangles of a triangulated surface Note that if
              * p1 or p2 coincides with one of the vertices this function still returns false!!
              * @param p1 the first point on a line
              * @param p2 the second point on a line
              * @param crossPoint the crossPint in the first face found with which the line intersects
              * @returns if the point intersects
              */
            bool lineGeomSect( const P3DVector& p1, const P3DVector& p2, P3DVector& crossPoint ) const;


            /** @brief find all intersections of linesegment running from locations (vectors)
              * mVertices[i1] to mVertices[i1] with all triangles of a triangulated surface.
              * @param i1 the first index of a vertex on a line
              * @param i2 the second index of a vertex on a line
              * @param faceIndex the first face found of the face with which the line intersects
              * @returns if the point intersects
              */
            bool selfsect( int i1, int i2, int& faceIndex ) const;


            /** @brief Determine if the gerometry crosse the plane
              * @param crossPlane the plane against which this is checked
              * @returns false if the geometry does not cross the plane otherwise true
              * @see int PFace::facePlaneCrossing(const PPlane &crossPlane) const
              */
            bool geometryPlaneCrossing( const PPlane &crossPlane ) const;


            /** @brief returns the angle of the triangulated geometry from
              * the observation point
              * @see double PFace::solidAngle(const P3DVector &observationPnt)
              */
            double solidAngle( const P3DVector &observationPnt ) const;


            /** @brief projects the given point on the surface of the
              * geometry when the point is on the outside.
              * @param point the to be projected point
              * @param outside if true the normals should point in the same direaction otherwise oposite
              * @returns projectionPnt the the point projected on the geometry
              */
            P3DVector pointfromOnesideOnGeometry( const P3DVector &point, bool outside ) const;


            /** @brief projects the given point on the surface of  the
              * geometry.
              * @param point the to be projected point
              * @returns projectionPnt the the point projected on the geometry
              */
            P3DVector pointOnGeometry( const P3DVector &point ) const;


            /** @brief determine if the given point lies on the surface of  the
              * geometry. returns true if the point is located on the surface
              * @param point the given point
              * @param vertexIndex if the point coincides with one of the
              * vertices of the geometry otherwise -1
              * @param faceIndex the index of the triangle in which the
              * point is located
              * @param projectionPnt the the point projecte don the geometry
              */
            bool pointOnGeometry( const P3DVector &point, int &vertexIndex,
                                  int &faceIndex, P3DVector &projectionPnt ) const;


            /** @brief determine if the given point lies on the surface of  the
              * geometry. returns true if the point is located on the surface
              * @param point the given point
              * @param vertexIndex if the point coincides with one of the
              * vertices of the geometry otherwise -1
              * @param faceIndex the index of the triangle in which the
              * point is located
              * @param projectionPnt the the point projecte don the geometry
              */
            int vertexIndexOnGeometry( const P3DVector &point) const;



            /** @brief determine if the given point lies on the surface of  the
              * geometry. returns true if the point is located on the surface
              * @param point the given point
              * @param vertexIndex if the point coincides with one of the
              * vertices of the geometry otherwise -1
              * @param faceIndex the index of the triangle in which the
              * point is located
              * @param lambda ratio along the the first edge
              *              (mVertices[1]-mVertices[0]) to the projection point
              * @param mu     ratio along the the first edge
              *              (mVertices[1]-mVertices[0]) to the projection point
              */
            bool pointOnGeometry( const P3DVector &point, int &vertexIndex, int &faceIndex,
                                  double &lambda, double &mu) const;


            /** @brief the center is the average of all vertices
              */
            inline P3DVector center( void ) const { return mVertices.getMean(); }


            /** @brief the minimum and maxiumum of all vertices
              *  @param minCoor reference to the minimum  x,y and z coordinates
              *  @param maxCoor reference to the maximumx,y and z coordinates
              */
            void minmax( P3DVector &minCoor, P3DVector &maxCoor ) const;


            /** @brief Determine the maxima and minima in x,y and z direction and take the average of these
              * extream as the center
              * @see minmax(P3DVector &minCoor, P3DVector &maxCoor)
              */
            P3DVector boxCenter( void ) const;


            /** @brief Returns true if pos is inside the box aroudn the geometry
              * @param pos the postion of a vertex that is checked
              */
            bool insideBox( const P3DVector &pos ) const;


            /** @brief returns the surface area of the geometry
              */
            double area( void ) const;


            /** @brief TODO
              * @param minWallThickness (default 0.0)
              * @return Vector
              */
            PVector wallThickness( double minWallThickness = 0.0 ) const;


            /** @brief translate the geometry by thegiven vector
              * @param trans The translation vectro
              */
            void translate( const P3DVector &trans );


            /** @brief translate the geometry by thegiven vector
              * @param trans The translation per position
              */
            void translate( const P3DVertices &trans );


            /** @brief translate the geometry by thegiven vector
              * @param trans The translation per position
              */
            P3DVertices checkedTranslation( const P3DVertices &trans );



            /** @brief scale the geometry by the given scale factors in the x, y and z direction
              * @param scale the scaling factors in x,y and z direction
              */
            void scale( const P3DVector &scale );


            /** @brief overloaded function
              * scale the geometry by the given scale factors in the x, y and z direction
              * @param scale the scaling factors in x,y and z direction
              * @see void scale( const P3DVector &scale )
              */
            void scale( double scal ) { scale(P3DVector(scal,scal,scal)); }


            /** @brief scale relative to the center
              * @param center the center of the vertices
              */
            void scaleKeepCenter(const P3DVector &scale, const P3DVector & center);


            /** @brief rotate geometry.
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param phi       rotation around  ?? axes
              * @param theta     rotation around  ?? axes
              * @param gamma     rotation around  ?? axes
              * @param rotPnt    the point aroudn which the geometry is rotated
              */
            void rotate( double phi, double theta, double gamma, const P3DVector &rotPnt = P3DVector( 0, 0, 0 ) );


            /** @brief rotate geometry.
              * The used the PCoordinateSystem to rotate based on the euler angles phi, theta, gamma
              * The used rotation is the ZYZ system
              * see also ; wikipedia.org: Euler_angles; gamma is angle 1, theta angle 2 and phi angle 3
              * @param angles    phi,theta,gamma
              * @param rotPnt    the point around which the geometry is rotated
              */
            void rotate( const P3DVector &angles, const P3DVector &rotPnt = P3DVector( 0, 0, 0 ) );


            /** @brief TODO
              * @param axis
              * @param angle
              * @param rotPnt
              */
            void rotate( const P3DVector &axis, double angle, const P3DVector &rotPnt = P3DVector( 0, 0, 0 ) );


            /** @brief change the XY axis: y=-x and x = y
              * Created because the definition in MRI coordinates is diffrent from the one used at mbfys
              * Nijmgen, the Netherlands
              */
            void changeXYaxis( void );


            /** @brief change the XY axis: y=-x and x = y
              * Created because the definition in MRI coordinates is diffrent from the one used at mbfys
              * Nijmgen, the Netherlands
              */
            void changeYXaxis( void );


            /** @brief change the direction of the faces to flip the normal direction
              */
            void invertNormals( void );


            /** @brief TODO
              * @param crossPlane
              * @param contourPoints
              * @param contourIndex
              */
            int crossSection( const PPlane &crossPlane,
                              P3DVertices &contourPoints,
                              PIVector &contourIndex ) const;


            /** @brief Computes the crossection with a plane
              * @param crossPlane
              * @param contourPoints
              * @param contourIndex
              * @param contourValues
              * @param functionvals
              */
            int crossSection( const PPlane &crossPlane,
                              P3DVertices &contourPoints,
                              PIVector &contourIndex,
                              PVector &contourValues,
                              const PVector &functionvals = PVector() ) const;


            /** @brief Computes the crossection given function values
              * @param functionval
              * @param curfunctionVal
              * @param contourPoints
              * @param contourIndex
              */
            int crossFunctionVal( const PVector &functionval,
                                  double curfunctionVal,
                                  P3DVertices &contourPoints,
                                  QList<int> &contourIndex ) const;


            /** @brief Reads geometery data from file.
              * @param ds the stream from which the data is read
              * @see virtual bool PFile::read(QDatastream &ds);
              */
            virtual bool read( QDataStream &ds );


            /** @brief Writes geometery data to file.
              *  @param ds the stream from which the data is read
              *  @see virtual bool read(QDatastream &ds) = 0;
              */
            virtual bool write( QDataStream &ds )const;


            /** @brief load triangulation file
              * If the file could not be read the geometry is cleared
              * 	\n\n
              * The format of the triangulation file is:\n
              * # vertices (N<SUB>v</SUB>)\n
              * 1  x<SUB>1</SUB>  y<SUB>1</SUB>   z<SUB>1</SUB>\n
              * .... .... ...... ......\n
              * .... .... ...... ......\n
              * N<SUB>v</SUB>   x<SUB>n</SUB>	y<SUB>n</SUB>   z<SUB>n</SUB>\n
              * # Faces (N<SUB>f</SUB>)\n
              * 1   index<SUB>1</SUB>	index<SUB>1</SUB>   index<SUB>1</SUB>\n
              * .... ......... ........... ........\n
              * .... ......... ........... ........\n
              * N<SUB>f</SUB>   index<SUB>n</SUB>
              * index<SUB>n</SUB>   index<SUB>n</SUB>\n
              * @param filename is the name of the triagulation file
              */
            bool load( const QString &filename );


            /** @brief save triangulation file
              * @param filename is the name of the triagulation file
              * Format of the triangulation file
              * @see load(QString FileName)
              */
            bool save( const QString &filename ) const;


            /** @brief stl file format handling
             *  facet normal 0.0 0.0 1.0
             *      outer loop
             *          vertex  1.0  1.0  0.0
             *          vertex -1.0  1.0  0.0
             *          vertex  0.0 -1.0  0.0
             *      endloop
             *  endfacet
             * Binary STL files consist of a 80 byte header line that can be
             * interpreted as a comment string. The following 4 bytes interpreted
             * as a long integer give the total number of facets. What follows is
             * a normal and 3 vertices for each facet, each coordinate represented
             * as a 4 byte floating point number (12 bytes in all). There is a 2
             * byte spacer between each facet. The result is that each facet is represented
             * by 50 bytes, 12 for the normal, 36 for the 3 vertices, and 2 for the spacer.
             * @param str the name of the file
             */
             bool saveToSTL( const QString &filename);

             bool loadFromSTL( const  QString &str );


            /** @brief TODO (nothing yet)
              * @param str the name of the file
              */
            bool loadFromXml(const  QString &str );


            /** @brief TODO (nothing yet)
              * @param stream the filestream to read from
              */
            bool saveToXmL( QTextStream &stream );


            /** @brief determine the postion of a point on the surface of
              * 'geometry closest to 'ver'
              * @param ver  The point from which the projection is reconstructed
              * @param iMinFace The index of the face on which the projection
              * point is located
              * @param iClostestVertex: The index of a vertex close to the point
              * where the shortest distance should be found.
              * Only faces with this index are taken into account.
              * Default all faces are used to determine theshoretst distance.
              */
            P3DVector findProjectionPointAdjacentToVertex( const P3DVector &ver,
                                                           int  &iMinFace ,
                                                            int iClostestVertex = -1 ) const;


            /** @brief Update the FaceEdgeConnectMatrix
              * The matrix must be updated when the traingulation of the geometry has changed
              * @see mFaceEdgeConnectMatrix
              */
            bool checkFaceEdgeConnectmatrix( void );


            /** @brief clears all data fields
              */
            virtual void clear( void );


        protected:

            /** @brief Returns the class name
              * @see PFile
              */
            virtual const char *getClassName( void ) const;


            /** @brief update the mVerticesList and mFaceIndexList if necesarry
              */
            void updateLists( void );


            /** @brief Implementation of a pure virtual function (PFile)that
              * requests the default class data version to write to the case file
              * @see PFile
              */
            virtual int getDefaultVersion( void ) const;


            /** @brief Reads geometery data from stream.
              * This function is created to make it also accessible for
              * subclasses
              * @param ds the stream from which the data is read
              * @see bool read(QDataStream &ds);
              */
            bool readGeometry( QDataStream &ds );


            /** @brief Writes geometery data to file.
              *  This function is created to make it also accessible for
              *  subclasses
              *  @param ds the stream to which the data is writen
              *  @see  bool write(QDataStream &ds)const;
              */
            bool writeGeometry( QDataStream &ds )const;


            /** @brief list of vertices
              *	@see bool load(QString FileName)
              * @see P3DVector
              */
            P3DVertices mVertices;


            /** @brief list of mFaces
              * @see PFace
              */
            QList <PFace> mFaces;


        private:


            /** @brief normalized normals of at the vertices
              * This normal is the average of the normals of the surrounding
              * triangles
              */
            P3DVertices mVerticesNrm;


            /** @brief The area around the vertex ( 1/3 of the surrounding faces
              */
            PVector mVerticesArea;


            /** @brief the precision of the stored vertices */
            int mPrecision;


            /** @brief represents an adajcency matrix in vector form.
              * The nonzero adjacency matrix elements contains the adajcent face indexes
              */
            QVector <int> mFaceEdgeConnectMatrix;


            /** @brief Contains a loop per vertex of the direct neighbors ordered clockwise
              * (viewed from outside) around vertex
              */
            QVector < QList<int> > mNeighbors;


            /** @brief Contains a list of all edges of the geometry
              */
            QList <PEdge> mEdgeList;
       };
    }
}

#endif // _PGEOMETRY_H

