/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pgraphgeometry.cpp 1137 2014-03-01 21:49:09Z peter $
 *
 */


#include <QString>
#include <QDebug>

#include "pgraphgeometry.h"
#include "pfunctions.h"
#include "math.h"
#include "pface.h"

const int FILE_GRAPHGEOMETRY_VERSION_1          = 1;
const int FILE_GRAPHGEOMETRY_VERSION_DEFAULT    = FILE_GRAPHGEOMETRY_VERSION_1;

const char* FILE_GRAPHGEOMETRY_NAME             = "PGraphGeometry";

const double NO_COORDINATE                      = HUGE_VAL;
const double MAX_TRANSMURAL_DIST                = 50;


using namespace Peacs::Support;


PGraphGeometry::PGraphGeometry( void ) :
    PGeometry(),
    mAnisotropyRatio( DEFAULT_ANISOTROPY ),
    mNeighAdjacency(),
    mSurfAdjacency(),
    m3dAdjacency(),
    mAnisAdjacency(),
    mNeighDistance(),
    mSurfDistance(),
    mAnisDistance(),
    m3dDistance(),
    mProgressDialog( NULL ),
    mNeighType( SECONDORDER )
{
}


PGraphGeometry::PGraphGeometry( const QString &fn ) :
    PGeometry( fn ),
    mAnisotropyRatio( DEFAULT_ANISOTROPY ),
    mNeighAdjacency(),
    mSurfAdjacency(),
    m3dAdjacency(),
    mAnisAdjacency(),
    mNeighDistance(),
    mSurfDistance(),
    mAnisDistance(),
    m3dDistance(),
    mProgressDialog( NULL ),
    mNeighType( SECONDORDER )
{
}


PGraphGeometry::PGraphGeometry( const PGeometry& geom ) :
    PGeometry( geom ),
    mAnisotropyRatio( DEFAULT_ANISOTROPY ),
    mNeighAdjacency(),
    mSurfAdjacency(),
    m3dAdjacency(),
    mAnisAdjacency(),
    mNeighDistance(),
    mSurfDistance(),
    mAnisDistance(),
    m3dDistance(),
    mProgressDialog( NULL ),
    mNeighType( SECONDORDER )
{
}


PGraphGeometry::PGraphGeometry( const PGraphGeometry& geom ) :
    PGeometry( geom ),
    mAnisotropyRatio( DEFAULT_ANISOTROPY ),
    mNeighAdjacency( geom.mNeighAdjacency ),
    mSurfAdjacency( geom.mSurfAdjacency ),
    m3dAdjacency( geom.m3dAdjacency ),
    mAnisAdjacency( geom.mAnisAdjacency ),
    mNeighDistance( geom.mNeighDistance ),
    mSurfDistance( geom.mSurfDistance ),
    mAnisDistance( geom.mAnisDistance ),
    m3dDistance( geom.m3dDistance ),
    mProgressDialog( NULL ),
    mNeighType( geom.mNeighType )
{
}


int PGraphGeometry::getDefaultVersion( void ) const
{
    return FILE_GRAPHGEOMETRY_VERSION_DEFAULT;
}


const char *PGraphGeometry::getClassName( void ) const
{
    return FILE_GRAPHGEOMETRY_NAME;
}


bool PGraphGeometry::initGraph( const NeigborOrder_type neightype,
                                double limitFactor,
                                double maxLengthFact )
{
    bool ok = !initNeeded();

    if ( !ok )
    {
        if ( mProgressDialog == NULL )
        {
            mProgressDialog = new QProgressDialog( "graph computations", "cancel" ,0, 100 );
        }
        mNeighAdjacency = calcNeighAdjacency();

        mProgressDialog->open();
        mProgressDialog->setLabelText( "Compute order neighbors matrix" );

        mNeighDistance  = shortestPath( mNeighAdjacency );

        if ( !mProgressDialog->wasCanceled() )
        {
            mSurfAdjacency  = calcSurfAdjacency( neightype );
            mNeighType      = neightype;

            mProgressDialog->setLabelText( "Compute surface distance matrix" );

            mSurfDistance   = shortestPath( mSurfAdjacency );

            if ( !mProgressDialog->wasCanceled() )
            {
                mProgressDialog->setLabelText( "Compute 3D adjacency matrix" );

                m3dAdjacency = calc3dAdjacency( mSurfAdjacency,
                                                mNeighDistance,
                                                mSurfDistance,
                                                limitFactor,
                                                maxLengthFact );

                if ( !mProgressDialog->wasCanceled() )
                {
                    mProgressDialog->setLabelText( "Compute 3D distance matrix" );

                    m3dDistance = shortestPath( m3dAdjacency );

                    if ( !mProgressDialog->wasCanceled() )
                    {
                        mAnisAdjacency = calcAnisotropicAdjacency( m3dAdjacency,
                                                                   mSurfAdjacency,
                                                                   mAnisotropyRatio );

                        mProgressDialog->setLabelText( "Compute anisotropic distance matrix" );

                        mAnisDistance = shortestPath( mAnisAdjacency );
                    }
                }
            }
        }

        ok = !mProgressDialog->wasCanceled();

        if ( mProgressDialog != NULL )
        {
            mProgressDialog->close();
            delete mProgressDialog;
            mProgressDialog = NULL;
        }
    }

    return ok;
}


void PGraphGeometry::initNeighbors( void )
{
    mNeighAdjacency = calcNeighAdjacency();
    mNeighDistance  = shortestPath( mNeighAdjacency );
}


void PGraphGeometry::initSurfaceMatrices( const NeigborOrder_type neightype )
{
    mSurfAdjacency  = calcSurfAdjacency( neightype );
    mNeighType      = neightype;
    mSurfDistance   = shortestPath( mSurfAdjacency );
}

void PGraphGeometry::initSurfaceAdjacency( const NeigborOrder_type neightype )
{
    mSurfAdjacency  = calcSurfAdjacency( neightype );
    mNeighType      = neightype;
}

PGraphGeometry::~PGraphGeometry( void )
{
    // Just to be shore
    if ( mProgressDialog != NULL )
    {
        mProgressDialog->close();
        delete mProgressDialog;
        mProgressDialog = NULL;
    }
}


bool PGraphGeometry::initNeeded( void ) const
{
    return !isEmpty() &&
            ( mSurfAdjacency.Nr() != nVertices() ||
              mSurfAdjacency.Nc() != nVertices() ||
              mSurfDistance.Nr()  != nVertices() ||
              mSurfDistance.Nc()  != nVertices() ||
              m3dAdjacency.Nr()   != nVertices() ||
              m3dAdjacency.Nc()   != nVertices() ||
              m3dDistance.Nr()    != nVertices() ||
              m3dDistance.Nc()    != nVertices() ||
              mAnisAdjacency.Nr() != nVertices() ||
              mAnisAdjacency.Nc() != nVertices() ||
              mAnisDistance.Nc()  != nVertices() ||
              mAnisDistance.Nr()  != nVertices() );
}


const PGraphGeometry& PGraphGeometry::operator= ( const PGraphGeometry &geom2 )
                                                {
    if ( this != &geom2 )
    {
        PGeometry::operator= ( geom2 );

        mAnisotropyRatio    = geom2.mAnisotropyRatio;
        mNeighAdjacency     = geom2.mNeighAdjacency;
        mSurfAdjacency      = geom2.mSurfAdjacency;
        m3dAdjacency        = geom2.m3dAdjacency;
        mAnisAdjacency      = geom2.mAnisAdjacency;
        mNeighDistance      = geom2.mNeighDistance;
        mSurfDistance       = geom2.mSurfDistance;
        mAnisDistance       = geom2.mAnisDistance;
        m3dDistance         = geom2.m3dDistance;
        mNeighType          = geom2.mNeighType;
    }

    return *this;
}


PMatrix PGraphGeometry::calcNeighAdjacency( void ) const
{
    PMatrix adj( nVertices(), nVertices() );

    if (  !isEmpty() )
    {
        //constructs the first order adjacency matrix
        for ( int t = 0; t < nFaces(); t++ )
        {
            adj[ face(t).I0() * adj.Nc() + face(t).I1() ] = 1.0;
            adj[ face(t).I1() * adj.Nc() + face(t).I0() ] = 1.0;
            adj[ face(t).I1() * adj.Nc() + face(t).I2() ] = 1.0;
            adj[ face(t).I2() * adj.Nc() + face(t).I1() ] = 1.0;
            adj[ face(t).I2() * adj.Nc() + face(t).I0() ] = 1.0;
            adj[ face(t).I0() * adj.Nc() + face(t).I2() ] = 1.0;
        }
    }

    return adj;
}


PMatrix PGraphGeometry::calcSurfAdjacency( const NeigborOrder_type neightype ) const
{
    PMatrix adj( nVertices(), nVertices() );

    if ( !isEmpty() )
    {
        //constructs the second order adjacency matrix
        for ( int t = 0; t < nFaces(); t++ )
        {
            int i0      = face(t).I0();
            int i1      = face(t).I1();
            int i2      = face(t).I2();
            double len  = ( vertex( i0 ) - vertex( i1 ) ).length();

            adj[ i0 * adj.Nc() + i1 ]   = len;
            adj[ i1 * adj.Nc() + i0 ]   = len;

            len                         = ( vertex( i1 ) - vertex( i2 ) ).length();
            adj[ i1 * adj.Nc() + i2 ]   = len;
            adj[ i2 * adj.Nc() + i1 ]   = len;

            len                         = ( vertex( i2 ) - vertex( i0 ) ).length();
            adj[ i2 * adj.Nc() + i0 ]   = len;
            adj[ i0 * adj.Nc() + i2 ]   = len;
        }

        //SECOND ORDER NEIGHBORS
        if ( neightype == SECONDORDER )
        {
            int     i;
            int     ic;
            int     kc;
            int     k;
            PMatrix DD = adj;

            for ( i = 0, ic = 0; i < adj.Nc(); i++, ic += adj.Nc() )
            {
                for ( int j = 0; j < nVertices(); j++ )
                {
                    if ( DD[ic + j] > 0 )
                    {
                        for ( k = 0, kc = 0; k < nVertices(); k++, kc += adj.Nc() )
                        {
                            if ( DD[kc + j] > 0 && DD[ic + k] == 0 && k != i )
                            {
                                double len = getLength( adj, i, j, k );

                                if ( len > 0 &&
                                     ( len < adj[ic + k] ||
                                       adj[ic + k] == 0 ) )
                                {
                                    adj[ ic + k ] = len;
                                    adj[ kc + i ] = len;
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    return adj;
}


PMatrix PGraphGeometry::calc3dAdjacency( const PMatrix &adjSurf,
                                         const PMatrix &neighOrder,
                                         const PMatrix &surfDistance,
                                         double limitFactor,
                                         double maxLengthFact ) const
{
    PMatrix adj3d( adjSurf );

    if ( !isEmpty() &&
         neighOrder.Nc()    == adjSurf.Nc() &&
         neighOrder.Nr()    == adjSurf.Nr() &&
         surfDistance.Nc()  == adjSurf.Nc() &&
         surfDistance.Nr()  == adjSurf.Nr() )
    {
        double avgLen   = 0;
        double nLen     = 0;

        if ( maxLengthFact > 0 )
        {
            for ( int t = 0; t < nFaces(); t++ )
            {
                avgLen += (vertex(face(t).I0()) - vertex(face(t).I1())).length();
                avgLen += (vertex(face(t).I1()) - vertex(face(t).I2())).length();
                avgLen += (vertex(face(t).I2()) - vertex(face(t).I0())).length();
                nLen += 3.0;
            }

            avgLen /= nLen;
        }

        int     i;
        int     ic;
        int     j;
        int     jc;
        double  len;

        // determine transmural edges
        for ( i = 0, ic = 0; i < adj3d.Nc(); i++, ic += adj3d.Nc() )
        {
            for ( j = 0, jc = 0; j < adj3d.Nc(); j++, jc += adj3d.Nc() )
            {
                if ( adj3d[ ic + j ] == 0 )
                {
                    len = ( vertex( i ) - vertex( j ) ).length();

                    /* if transmural edge and the distance over the wall is not similar
                     * this last option is to make the 3d adjacency matrix sparse
                     * Further more the connections through the cancave should go over the wall*/
                    if ( maxLengthFact <= 0 ||
                         len < avgLen * maxLengthFact )
                    {
                        if ( limitFactor <= 0 ||
                             ( neighOrder[ ic + j ] >= 2 &&
                               len / surfDistance[ ic + j ] < limitFactor ) )
                        {
                            if ( transmuralEdge( i, j )  )
                            {
                                adj3d[ ic + j ] = len;
                                adj3d[ jc + i ] = len;
                            }
                        }
                    }
                }
            }

            if ( mProgressDialog )
            {
                mProgressDialog->setValue( (100 * i) / nVertices() );

                if ( mProgressDialog->wasCanceled() )
                {
                    adj3d.clear();
                    break;
                }
            }
        }
    }

    return adj3d;
}

PMatrix PGraphGeometry::calcAnisotropicAdjacency(const PMatrix &adj3d, const PMatrix &adjSurf, double anis) const
{
    PMatrix anisAdj( adj3d );

    if ( !isEmpty() &&
         adj3d.Nc()     == nVertices() &&
         adj3d.Nr()     == nVertices() &&
         adjSurf.Nc()   == nVertices() &&
         adjSurf.Nr()   == nVertices() &&
         anis           != 0 &&
         anis           != 1 )
    {
        QVector <P3DVector> otherSides( nVertices() );
        int                 i;
        int                 j;
        int                 ic;
        int                 jc;
        int                 iMinVer;
        double              minDist = HUGE_VAL;
        double              d;
        double              h;
        double              len;

        // modify the transmural connctions
        for ( i = 0, ic = 0; i < nVertices(); i++, ic += anisAdj.Nc() )
        {
            iMinVer = -1;
            minDist = HUGE_VAL;

            // find the node on the oposite wall with the shoretst distance
            for ( j = 0; j < nVertices(); j++ )
            {
                // The edge should be a transmural one, so no connection over the surface
                if ( anisAdj[ ic + j ] > 0 &&
                     adjSurf[ ic + j ] == 0 &&
                     anisAdj[ ic + j ] < minDist )
                {
                    minDist = anisAdj[ ic + j ];
                    iMinVer = j;
                }
            }

            if ( iMinVer >= 0 )
            {
                otherSides[i] = findProjectionPointAdjacentToVertex( mVertices.at( i ), iMinVer );
            }
            else
            {
                otherSides[i] = P3DVector( NO_COORDINATE, NO_COORDINATE, NO_COORDINATE );
            }
        }

        for ( i = 0, ic = 0; i < nVertices(); i++, ic += anisAdj.Nc() )
        {
            for ( j = 0, jc = 0; j < nVertices(); j++, jc += anisAdj.Nc() )
            {
                if ( anisAdj[ ic + j ] > 0 &&
                     adjSurf[ ic + j ] == 0 &&
                     otherSides[i].x() != NO_COORDINATE &&
                     otherSides[j].x() != NO_COORDINATE )
                {
                    h = anis * 0.5 *
                            ( ( mVertices.at( i ) - otherSides.at( i ) ).length() +
                              ( mVertices.at( j ) - otherSides.at( j ) ).length() );

                    // factor 1.1 is to compentate for using a straigt line
                    // instead for a line over the (curved ) surface
                    d = 1.1 * ( ( mVertices.at( i ) - otherSides.at( j ) ).length() +
                                ( mVertices.at( j ) - otherSides.at( i ) ).length() );

                    len = sqrt( h * h + d * d );

                    if ( len > anisAdj[ ic + j ] )
                    {
                        anisAdj[ ic + j ] = len;
                        anisAdj[ jc + i ] = len;
                    }
                }
            }
        }
    }

    return anisAdj;
}

PMatrix PGraphGeometry::calcAnisotropicAdjacencyFibres(const PMatrix &adj3d,
                                                       const PMatrix &adjSurf,
                                                       const P3DVector &longAxisIn,
                                                       const PIVector & endoNode,
                                                       double anis) const
{
    PMatrix anisAdj( adj3d );
    P3DVector longAxis(longAxisIn.normalized());

    if ( !isEmpty() &&
         adj3d.Nc()     == nVertices() &&
         adj3d.Nr()     == nVertices() &&
         adjSurf.Nc()   == nVertices() &&
         adjSurf.Nr()   == nVertices() &&
         anis           != 0 &&
         anis           != 1 )
    {
        QVector <P3DVector> otherSides( nVertices() );
        int                 i;
        int                 j;
        int                 ic;
        int                 jc;
        int                 iMinVer;
        double              minDist = HUGE_VAL;
        double              d;
        double              h;
        double              len;

        // modify the transmural connctions
        for ( i = 0, ic = 0; i < nVertices(); i++, ic += anisAdj.Nc() )
        {
            iMinVer = -1;
            minDist = HUGE_VAL;

            // find the node on the oposite wall with the shoretst distance
            for ( j = 0; j < nVertices(); j++ )
            {
                // The edge should be a transmural one, so no connection over the surface
                if ( anisAdj[ ic + j ] > 0 &&
                     adjSurf[ ic + j ] == 0 &&
                     anisAdj[ ic + j ] < minDist )
                {
                    minDist = anisAdj[ ic + j ];
                    iMinVer = j;
                }
            }

            if ( iMinVer >= 0 )
            {
                otherSides[i] = findProjectionPointAdjacentToVertex( mVertices.at( i ), iMinVer );
            }
            else
            {
                otherSides[i] = P3DVector( NO_COORDINATE, NO_COORDINATE, NO_COORDINATE );
            }
        }

        for ( i = 0, ic = 0; i < nVertices(); i++, ic += anisAdj.Nc() )
        {
            for ( j = 0, jc = 0; j < nVertices(); j++, jc += anisAdj.Nc() )
            {
                if ( anisAdj[ ic + j ] > 0 &&
                     adjSurf[ ic + j ] == 0 &&
                     otherSides[i].x() != NO_COORDINATE &&
                     otherSides[j].x() != NO_COORDINATE )
                {
                    h = anis * 0.5 *
                            ( ( mVertices.at( i ) - otherSides.at( i ) ).length() +
                              ( mVertices.at( j ) - otherSides.at( j ) ).length() );

                    // factor 1.1 is to compentate for using a straigt line
                    // instead for a line over the (curved ) surface
                    d = 1.1 * ( ( mVertices.at( i ) - otherSides.at( j ) ).length() +
                                ( mVertices.at( j ) - otherSides.at( i ) ).length() );

                    len = sqrt( h * h + d * d );

                    if ( len > anisAdj[ ic + j ] )
                    {
                        anisAdj[ ic + j ] = len;
                        anisAdj[ jc + i ] = len;
                    }
                }
                else if ( adjSurf[ ic + j ] > 0 )// handle surface distances
                {
                    PPlane locPlane( (vertexNormal(i) + vertexNormal(j) ), P3DVector());
                    P3DVector projLongAxis( locPlane.projection(longAxis) );
                    if ( projLongAxis.length() > 0.1 )
                    {
                        projLongAxis.normate();
                        P3DVector vector( locPlane.projection(vertex(i) - vertex(j)).normalized());
                        double len = anisAdj[ ic + j ];
                        double angle = acos( vector.dot(projLongAxis) );
                        if ( endoNode[i] )
                            angle -= M_PI / 3.0;
                        else
                            angle += M_PI / 3.0;

                        double x = cos(angle) * len;
                        double y = sin(angle) * len * anis;
                        len = sqrt(x * x + y * y);
                        qDebug() << anisAdj[ ic + j ] << len << x << y << angle;
                        anisAdj[ ic + j ] = len;
                        anisAdj[ jc + i ] = len;
                    }
                }
            }
        }
    }

    return anisAdj;
}


double PGraphGeometry::getLength( const PMatrix & adj, int i1, int i2, int j2 ) const
{
    int         i;
    int         j;
    int         i1c;
    int         i3;
    int         j2c;
    int         k;
    int         secondOrderTriangle;
    P3DVector   vm;
    double      alpha;
    double      Dleft;
    double      Dright;
    double      xleft;
    double      xright;
    double      Dm      = 0;
    double      dist    = 0;

    for ( i = 0; i < nFaces(); i++ )
    {
        if ( mFaces[i].contains( i1 ) &&  mFaces[i].contains( i2 ) )
        {
            if ( ( mFaces[i][PFace::P1] == i1 && mFaces[i][PFace::P2] == i2 ) ||
                 ( mFaces[i][PFace::P2] == i1 && mFaces[i][PFace::P1] == i2 ) )
            {
                i3 = mFaces[i][PFace::P3];
            }
            else if ( ( mFaces[i][PFace::P2] == i1 && mFaces[i][PFace::P3] == i2 ) ||
                      ( mFaces[i][PFace::P3] == i1 && mFaces[i][PFace::P2] == i2 ) )
            {
                i3 = mFaces[i][PFace::P1];
            }
            else
            {
                i3 = mFaces[i][PFace::P2];
            }

            secondOrderTriangle = -1;

            for ( j = 0; j < nFaces(); j++ )
            {
                if ( j != i && mFaces[j].contains( i2 ) &&
                     mFaces[j].contains( i3 ) &&
                     mFaces[j].contains( j2 ) )
                {
                    secondOrderTriangle = j;

                    break;
                }
            }

            if ( secondOrderTriangle > 0 )
            {
                xleft   = 0;
                xright  = 1;
                i1c     = i1 * adj.Nc();
                j2c     = j2 * adj.Nc();
                Dleft   = adj[ i1c + i2 ] + adj[ j2c + i2 ];
                Dright  = adj[ i1c + i3 ] + adj[ j2c + i3 ];

                for ( k = 0; k < 20; k++ )
                {
                    alpha   = (xleft + xright) / 2.0;
                    vm      = mVertices.at( i2 ) + ( mVertices.at( i3 ) - mVertices.at( i2 ) ) * alpha;
                    Dm      = ( vm - mVertices.at( i1 ) ).length() + ( vm - mVertices.at( j2 ) ).length();

                    if ( (Dleft - Dm) > (Dright - Dm) )
                    {
                        xleft = alpha;
                        Dleft = Dm;
                    }
                    else
                    {
                        xright = alpha;
                        Dright = Dm;
                    }
                }

                if ( dist == 0 )
                {
                    dist = Dm;
                }
                else if ( Dm < dist)
                {
                    dist = Dm;
                }
            }
        }
    }

    return dist;
}


bool PGraphGeometry::transmuralEdge( int j1,int j2 ) const
{
    // true if line connecting node j1 to node j2 of the triangulated
    // surface specified by (mVertices and mFaces) lies entirely in
    // its interior; else false

    P3DVector   a1;
    P3DVector   a2;
    P3DVector   a3( vertex( j1 ) - vertex( j2 ) );
    P3DVector   sav;
    P3DVector   position;
    double      det;
    double      lambda;
    double      mu;
    double      alphas[100];
    double      deltas[100];
    double      alpha;
    double      tmin;
    double      eps             = EPS;
    int         i;
    int         j;
    int         na;
    int         ns              = 0;
    bool        isTransmural    = true;
    bool        maybe           = true;

    alphas[0] = 0;  // The first value must be zero

    /* identify intersections in between node1 and node2*/
    for ( i = 0; i < nFaces(); i++ )
    {
        if ( face( i ).contains( j1 ) )
        {
            if ( face( i ).contains( j2 ) )
            {
                isTransmural    = false;
                maybe           = false;
            }
            else
            {
                continue;
            }
        }
        else if ( face( i ).contains( j2 ) )
        {
            continue;
        }

        a1  = vertex( face( i ).I1() ) - vertex( face( i ).I0() );
        a2  = vertex( face( i ).I2() ) - vertex( face( i ).I0() );
        det = deter( a1, a2, a3 );

        if ( fabs( det ) > eps )
        {
            // no paralel detected; solution via Cramer's rule
            sav     = vertex( j1 ) - vertex( face( i )[PFace::P1] );
            lambda  = deter( sav, a2, a3 );
            lambda  = lambda / det;

            if ( lambda >= 0 && lambda <= 1 )
            {
                mu = deter( a1, sav, a3 ) / det;

                if ( mu >= 0 && mu <= 1 && (lambda + mu) <= 1 )
                {
                    alpha = deter( a1, a2, sav ) / det;

                    if ( alpha > 0 && alpha < 1 )
                    {
                        ns++;

                        alphas[ns] = alpha;
                    }
                }
            }
        }
    }

    if ( isTransmural && maybe )
    {
        alphas[++ns] = 1;
        ns++;

        deltas[0]   = 0;
        na          = 1;

        // sort for increasing values of alpha
        for ( i = 0; i < ns; i++ )
        {
            tmin = 1;

            for ( j = 0; j < ns; j++ )
            {
                if (alphas[j] > deltas[ na - 1 ] && alphas[j] < tmin )
                {
                    tmin = alphas[j];
                }
            }

            if ( tmin == 1 )
            {
                break;
            }

            deltas[na] = tmin;

            na++;
        }

        deltas[na] = 1.0;

        na++;

        // test if positions halfway the intersections are interior or exterior
        // if any of these is exterior, nodes j1 and j2 cannot be connected
        isTransmural = true;

        for ( i = 1; i < na; i++ )
        {
            position = (vertex( j2 ) - vertex( j1 )) * (deltas[i] + deltas[ i - 1 ]) / 2.0 + vertex( j1 );

            if ( fabs( solidAngle( position ) ) < 0.1 )
            {
                isTransmural = false;

                break;
            }
        }
    }

    return isTransmural;
}


PVector PGraphGeometry::getOtherSurfaceDist( int iVertex ) const
{
    double  mindist     = HUGE_VAL;
    int     otherVertex = 0;
    int     ic          = iVertex * mSurfDistance.Nc();

    for ( int i = 0; i < m3dDistance.Nc(); i++ )
    {
        if ( m3dDistance[ ic + i ] < mindist &&
             m3dDistance[ ic + i ] < mSurfDistance[ ic + i ] &&
             i != iVertex )
        {
            mindist     = m3dDistance[ ic + i ];
            otherVertex = i;
        }
    }

    return mSurfDistance.row( otherVertex );
    //    This might be more accurate, but it also give a moving center when switching wall sides
    //    P3DVector p=verOntoGeometry(mVertices[iVertex], iMinFace,otherVertex);
    //    otherDist= mSurfDistance.row(mFaces[iMinFace].I0())+(p-mVertices[mFaces[iMinFace].I0()]).length();
    //    v1       = mSurfDistance.row(mFaces[iMinFace].I1())+(p-mVertices[mFaces[iMinFace].I1()]).length();
    //    v2       = mSurfDistance.row(mFaces[iMinFace].I2())+(p-mVertices[mFaces[iMinFace].I2()]).length();
    //    for (i=0;i<otherDist.N();i++)
    //    {
    //        otherDist[i]= qMin(otherDist[i],qMin(v1[1],v2[i]));
    //    }
}


PVector PGraphGeometry::getTransmuralDist( int iVertex ) const
{
    PVector v0          = m3dDistance.row( iVertex );
    double  mindist     = HUGE_VAL;
    int     otherVertex = 0;
    int     ic          = iVertex * mSurfDistance.Nc();

    for ( int i = 0; i < m3dDistance.Nc(); i++ )
    {
        if ( m3dDistance[ ic + i ] < mindist &&
             m3dDistance[ ic + i ] < mSurfDistance[ ic + i ] &&
             i != iVertex )
        {
            mindist     = m3dDistance[ ic + i ];
            otherVertex = i;
        }
    }

    PVector v1 = mSurfDistance.row( otherVertex );

    for ( int i = 0; i < v0.N(); i++ )
    {
        v0[i] = qMin( v0[i], v1[i] );
    }

    return v0;
}


void PGraphGeometry::clear( void )
{
    PGeometry::clear();
    mNeighAdjacency.clear();
    mNeighDistance.clear();
    mSurfAdjacency.clear();
    mSurfDistance.clear();
    m3dAdjacency.clear();
    m3dDistance.clear();
    mAnisAdjacency.clear();
    mAnisDistance.clear();
}


void PGraphGeometry::geometryChanged( bool vertexChanged )
{
    PGeometry::geometryChanged( vertexChanged );

    if ( !vertexChanged )
    {
        // only required if a face is removed or added
        mNeighAdjacency.clear();
        mNeighDistance.clear();
    }

    // clear all other matrices
    mSurfAdjacency.clear();
    mSurfDistance.clear();
    m3dAdjacency.clear();
    m3dDistance.clear();
    mAnisAdjacency.clear();
    mAnisDistance.clear();
}


bool PGraphGeometry::saveGraphMatrices( const QString &filename ) const
{
    // only save if the matrices are computed
    if ( mAnisDistance.isEmpty() )
    {
        return PGeometry::save(       filename + ".tri"       );
    }
    else
    {
        return PGeometry::save(       filename + ".tri"       ) &&
                mSurfAdjacency.save(  filename + ".adj2d"     ) &&
                mSurfDistance.save(   filename + ".dst2d"     ) &&
                m3dAdjacency.save(    filename + ".adj3d"     ) &&
                m3dDistance.save(     filename + ".dst3d"     ) &&
                mAnisAdjacency.save(  filename + ".adjanis"   ) &&
                mAnisDistance.save(   filename + ".dstanis"   ) &&
                mNeighAdjacency.save( filename + ".adjneigh"  ) &&
                mNeighDistance.save(  filename + ".ordneigh"  );
    }
}


bool PGraphGeometry::loadGraphMatrices( const QString &filename )
{
    PGeometry::load(        filename + ".tri"       );
    mSurfAdjacency.load(    filename + ".adj2d"     );
    mSurfDistance.load(     filename + ".dst2d"     );
    m3dAdjacency.load(      filename + ".adj3d"     );
    m3dDistance.load(       filename + ".dst3d"     );
    mAnisAdjacency.load(    filename + ".adjanis"   );
    mAnisDistance.load(     filename + ".dstanis"   );
    mNeighAdjacency.load(   filename + ".adjneigh"  );
    mNeighDistance.load(    filename + ".ordneigh"  );

    return !isEmpty();
}





bool PGraphGeometry::read( QDataStream &ds )
{
    bool    ok      = false;
    int     version = readHead( ds, PGraphGeometry::getClassName() );

    clear();

    if ( version != READ_HEADER_FAILED )
    {
        switch ( version )
        {
        case FILE_GRAPHGEOMETRY_VERSION_1:
            ds >> mAnisotropyRatio;

            if ( isOk( ds ) )
            {
                ok = readGeometry( ds ) &&
                     mSurfAdjacency.read( ds ) &&
                     mSurfDistance.read( ds ) &&
                     m3dAdjacency.read( ds ) &&
                     m3dDistance.read( ds ) &&
                     mAnisAdjacency.read( ds ) &&
                     mAnisDistance.read( ds );
            }
            break;

        default:
            break;
        }
    }

    if ( ok )
    {
        initGraph();    // checks if all matrices are filled otherwise these are computed
    }

    return ok;
}


bool PGraphGeometry::write( QDataStream &ds ) const
{
    const_cast<PGraphGeometry&>( *this ).initGraph();   // checks if all matrices are filled otherwise these are computed

    bool ok = writeHead( ds, PGraphGeometry::getClassName(), PGraphGeometry::getDefaultVersion() );

    if ( ok )
    {
        switch ( PGraphGeometry::getDefaultVersion() )
        {
        case FILE_GRAPHGEOMETRY_VERSION_1:
            ds << mAnisotropyRatio;

            if ( isOk( ds ) )
            {
                ok = writeGeometry( ds ) &
                     mSurfAdjacency.write( ds ) &
                     mSurfDistance.write( ds ) &
                     m3dAdjacency.write( ds ) &
                     m3dDistance.write( ds ) &
                     mAnisAdjacency.write( ds ) &
                     mAnisDistance.write( ds );
            }
            break;

        default:
            ok = false;
            break;
        }
    }

    return ok;
}


QStringList PGraphGeometry::checkData( void ) const
{
    QStringList errors = PGeometry::checkData();

//    if ( initNeeded() )
//    {
//        errors.append(QString("Some matrices still need to be computed"));
//    }

    return errors;
}


PMatrix PGraphGeometry::shortestPath( const PMatrix &adj ) const
{
    return shortestPathAlgo(adj,false,mProgressDialog);

//    PMatrix         distance( adj );
//    QVector <int>   edgeA( adj.nonZeros() / 2 );
//    QVector <int>   edgeB( adj.nonZeros() / 2 );
//    double          maxLength = 0;
//    int             k = 0;

//    // create a list of all the edges
//    for ( int i = 0, ir = 0; i < adj.Nr(); i++, ir += adj.Nc() )
//    {
//        for (int j = i; j < adj.Nc(); j++ )
//        {
//            if ( adj.at( ir + j ) != 0 )
//            {
//                edgeA[k] = i;
//                edgeB[k] = j;

//                k++;

//                maxLength = maxLength + adj.at( ir + j );
//            }
//        }
//    }

//    for (int i = 0; i < distance.Nr() * distance.Nc(); i++ )
//    {
//        if ( distance[i] == 0 )
//        {
//            distance[i] = maxLength;
//        }
//    }

//    int             start;
//    int             istart;
//    QVector <bool>  fixed( adj.Nr(), false );

//    for ( start = 0, istart = 0; start < distance.Nr(); start++, istart += distance.Nc() )
//    {
//        int nfix = 0;

//        for ( int i = 0; i < fixed.size(); i++ )
//        {
//            if ( i <= start )
//            {
//                fixed[i] = true;

//                nfix++;
//            }
//            else
//            {
//                fixed[i] = false;
//            }
//        }

//        distance.value( start, start ) = 0;

//        /* calculate all paths between startpnt and not fixed vertices*/
//        QVector <int>   eA( edgeA );
//        QVector <int>   eB( edgeB );
//        register int    i;

//        while ( nfix < distance.Nr() )
//        {
//            double          length  = maxLength - 0.01;
//            register int    nE      = 0;

//            for ( i = 0; i < eA.size(); i++ )
//            {
//                int astart = istart + eA[i];
//                int bstart = istart + eB[i];

//                if ( !fixed[ eB[i] ] )
//                {
//                    if(  fixed[ eA[i] ] )
//                    {
//                        distance[bstart] = qMin( distance[bstart],
//                                                 distance[astart] + adj.value( eA[i], eB[i] ) );

//                        length = qMin( length, distance[bstart] );
//                    }

//                    eA[nE] = eA[i];
//                    eB[nE] = eB[i];

//                    nE++;
//                }
//                else if( !fixed[ eA[i] ] ) // b is fixed
//                {
//                    distance[astart] = qMin( distance[astart],
//                                             distance[bstart] + adj.value( eB[i], eA[i] ) );

//                    length = qMin( length, distance[astart] );
//                    eA[nE] = eA[i];
//                    eB[nE] = eB[i];

//                    nE++;
//                }
//            }

//            eA.resize( nE );
//            eB.resize( nE );

//            int prevNfix = nfix;
//            nfix = start;

//            for ( i = start; i < distance.Nr(); i++ )
//            {
//                if ( distance[ istart + i ]  <= length )
//                {
//                    distance.value( i, start ) = distance[ istart + i ];

//                    fixed[i] = true;

//                    nfix++;
//                }
//            }
//            if ( nfix == prevNfix )
//            {
//                qDebug() << "the graph is not consistent";
//                break;
//            }
//        }

//        register int nEt = 0;

//        for ( i = 0; i < edgeA.size(); i++ )
//        {
//            if (  edgeA[i] >= start ||
//                  edgeB[i] >= start )
//            {
//                edgeA[nEt] = edgeA[i];
//                edgeB[nEt] = edgeB[i];

//                nEt++;
//            }
//        }

//        edgeA.resize( nEt );
//        edgeB.resize( nEt );

//        if ( mProgressDialog )
//        {
//            mProgressDialog->setValue( (100 * start) / nVertices() );

//            if ( mProgressDialog->wasCanceled() )
//            {
//                distance.clear();

//                break;
//            }
//        }
//    }

//    return distance;
}


PMatrix PGraphGeometry::shortestPathDirectedGraph( const PMatrix &adj ) const
{
    return shortestPathAlgo(adj,true,mProgressDialog);


//    PMatrix         distance( adj );
//    QVector <int>   edgeA( adj.nonZeros() );
//    QVector <int>   edgeB( adj.nonZeros() );
//    double          maxLength   = 0;
//    int             k           = 0;

//    // create a list of all the edges
//    for ( int i = 0, ir = 0; i < adj.Nr(); i++, ir += adj.Nc() )
//    {
//        for ( int j = 0; j < adj.Nc(); j++ )
//        {
//            if ( adj.at( ir + j ) != 0 )
//            {
//                edgeA[k]    = i;
//                edgeB[k]    = j;
//                maxLength   = maxLength + adj.at( ir + j );

//                k++;
//            }
//        }
//    }

//    for ( int i = 0; i < distance.Nr() * distance.Nc(); i++ )
//    {
//        if ( distance[i] == 0 )
//        {
//            distance[i] = maxLength;
//        }
//    }

//    int             start   = 0;
//    int             istart  = 0;
//    int             nfix    = 0;
//    QVector <bool>  fixed( adj.Nr(), false );

//    for ( ; start < distance.Nr(); start++, istart += distance.Nc() )
//    {
//        nfix = 0;

//        fixed.fill( false );
//        fixed[start] = true;

//        distance.value( start, start ) = 0;

//        /* calculate all paths between startpnt and not fixed vertices*/
//        QVector <int> eA( edgeA );
//        QVector <int> eB( edgeB );

//        while ( nfix < distance.Nr() )
//        {
//            double  length  = maxLength - 0.01;
//            int     nE      = 0;

//            for ( int i = 0; i < eA.size(); i++ )
//            {
//                int astart = istart + eA[i];
//                int bstart = istart + eB[i];

//                if ( !fixed[ eB[i] ] )
//                {
//                    if(  fixed[ eA[i] ] )
//                    {
//                        distance[bstart] = qMin( distance[bstart],
//                                                 distance[astart] + adj.value( eA[i], eB[i] ) );

//                        length = qMin( length, distance[bstart] );
//                    }

//                    eA[nE] = eA[i];
//                    eB[nE] = eB[i];

//                    nE++;
//                }
//            }

//            eA.resize( nE );
//            eB.resize( nE );

//            nfix = 0;

//            for ( register int i = 0; i < distance.Nr(); i++ )
//            {
//                if ( distance[ istart + i ] <= length )
//                {
//                    distance.value( i, start ) = distance[ istart + i ];

//                    fixed[i] = true;

//                    nfix++;
//                }
//            }
//        }

//        if ( mProgressDialog )
//        {
//            mProgressDialog->setValue( (100 * start) / nVertices() );

//            if ( mProgressDialog->wasCanceled() )
//            {
//                distance.clear();

//                break;
//            }
//        }
//    }

//    return distance;
}
