/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pgraphgeometry.h 1118 2014-02-26 16:09:13Z peter $
 *
 */


#ifndef __PGRAPHGEOMETRY_H
#define __PGRAPHGEOMETRY_H

#include <QProgressDialog>

#include "pgeometry.h"
#include "pmatrix.h"

#include "PeacsSupport_global.h"

namespace Peacs
{
    namespace Support
    {
        const double DEFAULT_ANISOTROPY = 2.5;

        /** @class PGraphGeometry pgraphgeometry.h "support/pgraphgeometry.h"
          * @brief A geometry class with shortest path algorithms includes
          * This class inherites from the PGeometry class.
          * The triangulated surface is treated as fully connceted graph for
          * which different adjacency matrices can be cretaed. Based on these
          * adjacency matrices the distance matrix is computed by means of the
          * shortest path algorithm
          * @see PGeometry
          */
        class PEACSSUPPORTSHARED_EXPORT PGraphGeometry : public PGeometry
        {

        public:

            /** @brief TODO
              */
            enum NeigborOrder_type { FIRSTORDER,
                                     SECONDORDER };


            /** @brief Default constructor.
              *  Initialze all members, create an empty geometry
              */
            PGraphGeometry( void );


            /** @brief Constructor.
              * @param filename is the name of the triagulation file
              * If the file could not be read the default constructor is used.
              * Additionally the adjacency and distance matrices are read from
              * file. If no such file is availbel, they need to be computed
              * @see load(QString FileName)
              */
            PGraphGeometry( const QString &filename );


            /** @brief Constructor.
              * Copies the given geometry
              * @param geom2 The given geometry
              */
            PGraphGeometry( const PGeometry& geom2 );


            /** @brief copy Constructor.
              * Copies the given geometry and matrices
              * @param geom The given geometry
              */
            PGraphGeometry( const PGraphGeometry& geom );


            /** @brief Destructor.
              * @see ~PGeometry(void)
              */
            virtual ~PGraphGeometry( void );


            /** @brief make sure all matrices are computed
              * const PMatrix &getSurfAdjacency(void);
              * @see getSurfaceDist(void)
              * @see get3dAdjacency(void)
              * @see get3dDist(void)
              * @see getAnisDist(double anis=0)
              * @see getAnisotropicAdjacency(double anis=0)
              * @param limitFactor if the distance between 2 nodes over the
              * surface devided by the euclidian distance is less than the limit factor
              * the connection is added to the 3D adjacency. This introduces errors,
              * but makes the shortest path algorithm much faster
              * @param maxLengthFact if the euclidian distance is more than maxLengthFact
              * times the avarege of the edge lengths over the surface negative values indicate no
              * limitation i.e. all tranmural connections are added
              */
            bool initGraph( const NeigborOrder_type neightype = FIRSTORDER, double limitFactor = -1.0, double maxLengthFact = -1.0 );


            /** @brief recomputes the neighbor adjacency and distance matrix
              */
            void initNeighbors( void );


            /** @brief recomputes the surface adjacency and distance matrix with the given neightype
              * @param neightype
              */
            void initSurfaceMatrices( const NeigborOrder_type neightype );

            /** @brief recomputes the surface adjacency matrix with the given neightype
              * @param neightype
              */
            void initSurfaceAdjacency( const NeigborOrder_type neightype );


            /** @brief returns true if one of the adjacency or distance matrices has an inappropriate size
              * Used in derived classes
              */
            bool initNeeded( void ) const;


            /** @brief copy operator.
              * Copies the vertices and the faces and all the matricies
              * @param geom2 the copied geometry
              */
            const PGraphGeometry& operator= ( const PGraphGeometry &geom2 );


            /** @brief returns the order of all neighbors matrix, only connections over the wall.
              */
            const PMatrix &getNeighborOrderAdjacency( void ) const { return mNeighAdjacency; }

            /** @brief returns the order of all neighbors matrix, only connections over the wall.
              */
            PMatrix &getNeighborOrderAdjacency( void ) { return mNeighAdjacency; }



            /** @brief returns the order of all neighbors matrix, only connections over the wall.
              */
            const PMatrix &getNeighborOrder( void ) const { return mNeighDistance; }


            /** @brief returns the order of all neighbors matrix, only connections over the wall.
              */
            PMatrix &getNeighborOrder( void ) { return mNeighDistance; }



            /** @brief returns the surface adjacency matrix, only connections over the wall.
              */
            inline const PMatrix &getSurfAdjacency( void ) const { return mSurfAdjacency; }


            /** @brief returns the distance matrix with distances over the surafce only.
              */
            inline const PMatrix &getSurfaceDist( void ) const { return mSurfDistance; }


            /** @brief returns the 3d isotropic adjacency matrix, including
              * connections through the wall. If not already available, the
              * matrix is computed. Determines which edges are transmural
              * conncetions, i.e. lie completly within the closed surface
              * @see 3dAdjacency(void);
              */
            inline const PMatrix &get3dAdjacency( void ) const { return m3dAdjacency; }


            /** @brief returns the 3d isotropic distance matrix.
              * If not already available, the matrix is computed
              * @see mAnisotropyRatio
              */
            inline const PMatrix &get3dDist( void ) const { return m3dDistance; }


            /** @brief returns the 3d anisotropic distance matrix.
              * If not already available, the matrix is computed
              * @see mAnisotropyRatio
              * @see getAnisotropicAdjacency(const double &anis=0)/
              */
            inline const PMatrix &getAnisDist( void ) const { return mAnisDistance; }


            /** @brief Computes the adjacency matrix with transmural
              * anisotropic times slower propagtion velocity.
              * If no anisotropic ratio is supplied the current anisotropic is
              * used.
              * @see mAnisotropyRatio
              */
            inline const PMatrix &getAnisotropicAdjacency( void ) const { return mAnisAdjacency; }


            /** @brief Computes the distance over the surface from a given
              * vertex to all other nodes
              * @param iVertex is the index number of the vertex
              * @see mSurfDistance
              * @see PGeometry::mVertices
              */
            inline PVector getSurfaceDist( int iVertex ) const { return mSurfDistance.row(iVertex); }


            /** @brief Computes the distance over the surface from a vertex
              * located on the oposite wall of the given vertex to all other
              * nodes the selected vertex is projected perpendicular on the other
              * wall. The distances are computed from this point
              * @param iVertex is the index number of the vertex
              * @see mSurfDistance
              * @see PGeometry::mVertices
              */
            PVector getOtherSurfaceDist( int iVertex ) const;


            /** @brief This is a combination of the functions @see PVector
              *  getSurfaceDist(const int &iVertex) and @see PVector
              * getOtherSurfaceNodes(const int &iVertex);
              * Instead of using the 3d distance matrix the two oposite regions
              * are now similar in size
              * @param iVertex is the index number of the vertex
              * @see mSurfDistance
              * @see PGeometry::mVertices
              */
            PVector getTransmuralDist( int iVertex ) const;


            /** export all matrices and geometry data
              * The extension of the files are added to the filename
              * The extensions are
              * geometry data            : .tri
              * surface adajency         : .adj2d
              * surface distance         : .dst2d
              * 3d adajency              : .adj3d
              * 3d distnace              : .dst3d
              * anisotropic 3d adajency  : .adjanis
              * anisotropic 3d distance  : .dstanis
              * @param filename base filename
              */
            bool saveGraphMatrices( const QString &filename )const;


            /** @brief load triangulation file
              * overloaded function to make sure that if the heometry is not the
              * same the distance and adjecency matrices are cleared
              * @param filename is the name of the triagulation file
              * @see PGeometry::load(const QString &filename)
              */
            bool loadGraphMatrices( const QString &filename );


            /** @brief read all avaliable geometry and graph data from a stream.
              * @param ds the sream used to read the data
              *  @see write(QDatastream &ds);
              */
            virtual bool read( QDataStream &ds );


            /** @brief Write all avaliable geometry and graph data from a file.
              * @param ds the sream used to write the data
              * @see read(QDatastream &ds);
              */
            virtual bool write( QDataStream &ds ) const;


            /** @brief clears all data fields
              */
            virtual void clear( void );


            /** @brief recompute all relevant members due to a movement of one of the vertices,
              * removal of a vertex/ face
              * @param vertexChanged if true only the vertex has been changed its position
              */
            void geometryChanged( bool vertexChanged = true );


            /** @brief computes the neighbor adjacency matrix, only connections
              * over the wall. If not already available, the matrix is computed
              */
            PMatrix calcNeighAdjacency( void ) const;


            /** @brief computes the isotropic adjacency matrix, only connections
              * over the wall. If not already available, the matrix is computed
              * @param neighType include only the fisrt order neighbors or also the second order
              * @param force if force is true recompute matrix
              */
            PMatrix calcSurfAdjacency( const NeigborOrder_type neightype = SECONDORDER ) const;


            /** @returns the true if the @var mNeighType == SECONDORDER otherwise false
              */
            inline bool isSecondOrderUsed( void ) const { return mNeighType == SECONDORDER; }


            /** @brief computes the 3d isotropic adjacency matrix, including
              * connections through the wall. If not already available, the
              * matrix is computed. Determines which edges are transmural
              * conncetions, i.e. lie completly within the closed surface
              * @param limitFactor if the distance between 2 nodes over the
              * surface devided by the euclidian distance is less than the limit factor
              * the connection is added to the 3D adjacency. This introduces errors,
              * but makes the shortest path algorithm much faster
              * @param maxLengthFact if the euclidian distance is more than maxLengthFact
              * times the avarege of the edge lengths over the surface negative values indicate no
              * limitation i.e. all tranmural connections are added
              */
            PMatrix calc3dAdjacency( const PMatrix &adjSurf,
                                     const PMatrix &neighOrder,
                                     const PMatrix &surfDistance,
                                     double limitFactor,
                                     double maxLengthFact ) const;


            /** @brief Computes the adjacency matrix with transmural
              * anisotropic times slower propagtion velocity. If no anisotropic
              * ratio is supplied the current anisotropic ratio is used.
              * @param adj3D the adjacency matrix
              * @param adjSurf the surface matrix
              * @param anis the anisotropic ratio (default 2)
              * @see mAnisotropyRatio
              */
            PMatrix calcAnisotropicAdjacency( const PMatrix &adj3D, const PMatrix &adjSurf, double anis = DEFAULT_ANISOTROPY ) const;


            /** @brief Computes the adjacency matrix with transmural
              * anisotropic times slower propagtion velocity. If no anisotropic
              * ratio is supplied the current anisotropic ratio is used.
              * @param adj3D the adjacency matrix
              * @param adjSurf the surface matrix
              * @param anis the anisotropic ratio (default 2)
              * @see mAnisotropyRatio
              */
            PMatrix calcAnisotropicAdjacencyFibres( const PMatrix &adj3D, const PMatrix &adjSurf, const P3DVector & longAxis, const PIVector & endoNode, double anis = DEFAULT_ANISOTROPY ) const;


            /** @brief Shortest path algorithm implementation
              * Computes the distances based on the given adjacency matrix.
              * It is assumed here that the adjacency is NON directed, i.e. connections from
              * i to j are the same as from j to i.
              * @param  adj the adjacency matrix
              */
            PMatrix shortestPath( const PMatrix &adj ) const;


            /** @brief Shortest path algorithm implementation
              * Computes the distances based on the given adjacency matrix.
              * It is assumed here that the adjacency is directed, i.e. connections from
              * i to j have a different length than from j to i.
              * @param  adj the adjacency matrix
              */
            PMatrix shortestPathDirectedGraph( const PMatrix &adj ) const;


            /** @brief virtual function that checks the data
              * returns an empty Stringlist if all data is correct
              * otherwise errormessages are created
              * @see PFile
              */
            virtual QStringList checkData( void ) const;


        protected:

            /** @brief Returns the class name
              * @see PFile
              */
            const char *getClassName( void ) const;


            /** @brief Implementation of a pure virtual function (PFile)that
              *  requests the default class data version to write to the case
              * file
              * @see PFile
              */
            int getDefaultVersion( void ) const;


            /** @brief Compute the distance between two vertices.
              * In case of the distance over the surface, the second order
              * neighbor is determined. i1 and i2 belong to the same traingle.
              * j2 shares one side with This triangle but is not part of this
              * triangle
              * @param adj the first order neighbor adjacency matrix
              * @param i1 vertex number 1
              * @param i2 vertex number 2
              * @param j2 vertex number 3
              */
            double getLength( const PMatrix & adj, int i1, int i2, int j2 ) const;


            /** @brief Determines wether a transmural conncetion lies
              * completly within the closed surface
              * @param j1 vertex number 1
              * @param j2 vertex number 2
              */
            bool transmuralEdge( int j1, int j2 ) const;


            /** @brief determine all trans mural edges for the given adjacency
              * matrix and updat this matrix
              * @returns adj AN adjacency matrix with transmural edges
              * @see transmuralEdge(int j1,int j2)
              */
            void determineTransmuralEdges( PMatrix &adj );


        private:

            /** @brief The anisotropy ratio defined as the ratio between
              * lonitudinal and transverse popagtion velocity. The 'anisotropic'
              * distance between one node to all other nodes as computed by
              * the fastest route algorithm.
              * see also van Dam, P., T. Oostendorp and A. van Oosterom. \n
              *  Application of the fastest route algorithm in the interactive
              * simulation of the effect of local ischemia on the ECG.\n
              * Medical & Biological Engineering & Computing 47(1): 11-20, 2009
              * @see mAnisDistance
              */
            double mAnisotropyRatio;


            /** @brief The surface neigh matrix as used by the shortest
              * path algorithm 1 is connected 0 is not connected
              */
            PMatrix mNeighAdjacency;


            /** @brief The surface adjecency matrix as used by the shortest
              * path algorithm
              */
            PMatrix mSurfAdjacency;


            /** @brief The 3d adjecency matrix, with connction through the
              * wall as used by the shortest path algorithm
              */
            PMatrix m3dAdjacency;


            /** @brief The anisotropic adjecency matrix with scaled distances
              * through the wall to mimic anisotropic propagation. All non-zero
              *  elements in the adjacyncy matrix contain the distance between
              * two nodes describing all edges over or through the wall.
              * Elements that are zero describe non-connected nodes.
              */
            PMatrix mAnisAdjacency;


            /** @brief The surface neighbor matrix as created by the shortest
              * path algorithm. the number sin th ematrix determine the order of the
              * neighbor
              */
            PMatrix mNeighDistance;


            /** @brief The surface distance matrix as created by the shortest
              * path algorithm
              */
            PMatrix mSurfDistance;


            /** @brief The anisotropic distance matrix as created by the
              * shortest path algorithm
              */
            PMatrix mAnisDistance;


            /** @brief The 3d distance matrix as created by the shortest
              * path algorithm
              */
            PMatrix m3dDistance;


            /** @brief If needed a progress dialog can be shown
              */
            QProgressDialog *mProgressDialog;


            /** @brief the type of the surface adjacency matrix
              */
            NeigborOrder_type mNeighType;
        };
    }
}

#endif // __PGRAPHGEOMETRY_H
