/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pmatrix.cpp 1179 2014-03-22 20:47:25Z peter $
 *
 */

#include <QTextStream>
#include <QDataStream>
#include <QFile>
#include <QDebug>

#include "math.h"

#include "pmatrix.h"
#include "PIVector.h"

//#ifdef PEACSCLASSES
//#include "nrutils.cc"
//#include "svdcmp.cc"
//#endif

using namespace Peacs::Support;


const int FILE_MATRIX_VERSION_1         = 1;
const int FILE_MATRIX_VERSION_DEFAULT   = 1;
const char* FILE_MATRIX_NAME            = "PMatrix";


PMatrix::PMatrix( void ) :
    PFile(),
    mNr( 0 ),
    mNc( 0 ),
    mMat()
{
}


PMatrix::PMatrix( int nrow, int ncol ) :
    PFile(),
    mNr( nrow ),
    mNc( ncol ),
    mMat( nrow * ncol, 0 )
{

}

PMatrix::PMatrix( int nrow, int ncol, double val ) :
    PFile(),
    mNr( nrow ),
    mNc( ncol ),
    mMat( nrow * ncol, val )
{
}


PMatrix::PMatrix( const PMatrix &m ) :
    PFile( m ),
    mNr( m.mNr ),
    mNc( m.mNc ),
    mMat( m.mMat )
{
}




PMatrix::~PMatrix( void )
{
}


int PMatrix::getDefaultVersion( void ) const
{
    return FILE_MATRIX_VERSION_DEFAULT;
}


const char *PMatrix::getClassName( void ) const
{
    return FILE_MATRIX_NAME;
}


void PMatrix::clear( void )
{
    mMat.clear();

    mNr = 0;
    mNc = 0;
}


void PMatrix::resize( int mnr, int mnc )
{
    if ( mMat.size() != mnr * mnc )
    {
        if ( mnr <= 0 || mnc <= 0 )
        {
            mMat.clear();

            mNr = 0;
            mNc = 0;
        }
        else
        {
            mNr = mnr;
            mNc = mnc;
            mMat.resize( mNr * mNc );
        }
    }
}


PMatrix PMatrix::diag( int n )
{
    PMatrix m( n, n );

    for ( int i = 0; i < n; i++ )
    {
        m.mMat[ i * n + i ] = 1;
    }

    return m;
}


void PMatrix::makeDiag( int n )
{
    resize( n, n );

    mMat.fill( 0 );

    for ( int i = 0; i < n; i++ )
    {
        mMat[ i * n + i ] = 1;
    }
}


PMatrix PMatrix::operator+ ( double f ) const
{
    PMatrix m( *this );
    int     size    = mNr * mNc;
    double* mapData = m.mMat.data();          // Get the matrix memory

    for ( int i = 0; i < size; i++ )
    {
        mapData[i] += f;
    }

    return m;
}


PMatrix PMatrix::operator- ( double f ) const
{
    PMatrix m( *this );
    int     size    = mNr * mNc;
    double* mapData = m.mMat.data();          // Get the matrix memory

    for ( int i = 0; i < size; i++ )
    {
        mapData[i] -= f;
    }

    return m;
}


PMatrix PMatrix::operator* ( double f ) const
{
    PMatrix m( *this );
    int     size    = mNr * mNc;
    double* mapData = m.mMat.data();          // Get the matrix memory

    for ( int i = 0; i < size; i++ )
    {
        mapData[i] *= f;
    }

    return m;
}


PMatrix &PMatrix::operator+= ( double f )
{
    int     size    = mNr * mNc;
    double* mapData = mMat.data();          // Get the matrix memory

    for ( int i = 0; i < size; i++ )
    {
        mapData[i] += f;
    }

    return *this;
}


PMatrix &PMatrix::operator-= ( double f )
{
    int     size    = mNr * mNc;
    double* mapData = mMat.data();          // Get the matrix memory

    for ( int i = 0; i < size; i++ )
    {
        mapData[i] -= f;
    }

    return *this;
}


PMatrix & PMatrix::subtractMatrixFromVal( double f )
{
    int     size    = mNr * mNc;
    double* mapData = mMat.data();          // Get the matrix memory

    for ( int i = 0; i < size; i++ )
    {
        mapData[i] = f - mapData[i];
    }

    return *this;

}



PMatrix & PMatrix::operator*= ( double f )
{
    int     size    = mNr * mNc;
    double* mapData = mMat.data();          // Get the matrix memory

    for ( int i = 0; i < size; i++ )
    {
        mapData[i] *= f;
    }

    return *this;
}


PMatrix PMatrix::operator+ ( const PMatrix & m2 ) const
{
    PMatrix m( *this );

    if ( !m2.isEmpty() )
    {
        if ( !m.isEmpty() )
        {
            if ( m2.mNc != m.mNc &&
                 m2.mNr != m.mNr )
            {
                qDebug("PMatrix operator +: matrices have unequal size");
            }
            else
            {
                int             size        = m.mNr * m.mNc;
                double*         mapData     = m.mMat.data();          // Get the matrix memory
                const double*   map2Data    = m2.mMat.constData();  // Get the m2 matrix memory

                for ( int i = 0; i < size; i++ )
                {
                    mapData[i] += map2Data[i];
                }
            }
        }
        else
        {
            m = m2;
        }
    }

    return m;
}


PMatrix PMatrix::operator- ( const PMatrix & m2 ) const
{
    PMatrix m( *this );

    if ( !m2.isEmpty() )
    {
        int     size = m2.mNr * m2.mNc;
        double* mapData;

        if ( !m.isEmpty() )
        {
            if ( m2.mNc != m.mNc &&
                 m2.mNr != m.mNr )
            {
                qDebug("PMatrix operator -: matrices have unequal size");
            }
            else
            {
                const double* map2Data = m2.mMat.constData();  // Get the m2 matrix memory

                mapData = m.mMat.data();          // Get the matrix memory

                for ( int i = 0; i < size; i++ )
                {
                    mapData[i] -= map2Data[i];
                }
            }
        }
        else
        {
            m = m2;

            mapData = m.mMat.data();          // Get the matrix memory

            for ( int i = 0; i < size; i++ )
            {
                mapData[i] *= -1;
            }
        }
    }

    return m;
}


PMatrix & PMatrix::operator+= ( const PMatrix &m2 )
{
    if ( !m2.isEmpty() )
    {
        if ( isEmpty() )
        {
            *this = m2;
        }
        else
        {
            if ( m2.mNc != mNc &&
                 m2.mNr != mNr )
            {
                qDebug("PMatrix operator +=: matrices have unequal size");
            }
            else
            {
                int             size        = mNr * mNc;
                double*         mapData     = mMat.data();          // Get the matrix memory
                const double*   map2Data    = m2.mMat.constData();  // Get the m2 matrix memory

                for ( int i = 0; i < size; i++ )
                {
                    mapData[i] += map2Data[i];
                }
            }
        }
    }

    return *this;
}


PMatrix & PMatrix::operator-= ( const PMatrix & m2 )
{
    if ( !m2.isEmpty() )
    {
        int     size;
        double* mapData;

        if ( mMat.isEmpty() )
        {
            *this = m2;

            // Now make it negative
            size    = mNr * mNc;
            mapData = mMat.data();        // Get the matrix memory

            for ( int i = 0; i < size; i++ )
            {
                mapData[i] *= -1;
            }
        }
        else
        {
            if ( m2.mNc != mNc &&
                 m2.mNr != mNr )
            {
                qDebug("PMatrix operator -=: matrices have unequal size");
            }
            else
            {
                const double* map2Data = m2.mMat.constData();  // Get the m2 matrix memory

                size        = mNr * mNc;
                mapData     = mMat.data();          // Get the matrix memory

                for ( int i = 0; i < size; i++ )
                {
                    mapData[i] -= map2Data[i];
                }
            }
        }
    }

    return *this;
}


PMatrix PMatrix::operator* ( const PMatrix & m2 ) const
{
    PMatrix m;

    m.mul( *this, m2 );

    return m;
}


PMatrix & PMatrix::operator*= ( const PMatrix & m2 )
{
    PMatrix m1( *this );

    mul( m1, m2 );

    return *this;
}


PMatrix PMatrix::operator/ ( const PMatrix & m2 ) const
{
    PMatrix m( *this );

    if ( m.Nc() == m2.Nc() &&
         m.Nr() == m2.Nr() )
    {
        int             size        = mNr * mNc;
        double*         mapData     = m.mMat.data();        // Get the matrix memory
        const double*   map2Data    = m2.mMat.constData();  // Get the m2 matrix memory

        for ( int i = 0; i < size; i++ )
        {
            mapData[i] /= map2Data[i];
        }
    }

    return m;
}


PMatrix & PMatrix::operator/= ( const PMatrix &m2 )
{
    if ( Nc() == m2.Nc() &&
         Nr() == m2.Nr() )
    {
        double* mapData         = mMat.data();          // Get the matrix memory
        const double* map2Data  = m2.mMat.constData();  // Get the m2 matrix memory

        for ( int i = 0; i < mNr * mNc; i++ )
        {
            mapData[i] /= map2Data[i];
        }
    }

    return *this;
}


PVector PMatrix::operator* ( const PVector & v2 ) const
{
    PVector v( mNr );

    if ( mNc == v2.N() &&
         mNr == v.N() )
    {
        int             k;
        const double *  mapData = mMat.constData();   // Get the matrix memory
        const double *  vecData = v2.data();          // Get the vector memory
        const double *  row;
        double *        vData   = v.data();

        for ( int i = 0; i < mNr; i++ )
        {
            vData[i]    = 0;
            row         = &mapData[ i * mNc ];      // Get the row to work with from the matrix

            for ( k = 0; k < mNc; k++ )
            {
                vData[i] += row[k] * vecData[k];
            }
        }
    }
    else
    {
        qDebug( "PMatrix operator * vextor: mNc==v2.N()" );
    }

    return v;
}


const PMatrix& PMatrix::operator= ( const PMatrix & m )
{
    if ( this != & m )
    {
        mMat    = m.mMat;
        mNc     = m.mNc;
        mNr     = m.mNr;
    }

    return *this;
}


PMatrix & PMatrix::copy( const PMatrix & m, int BeginRow, int EndRowIn, int BeginCol, int EndColIn )
{
    Q_ASSERT( EndRowIn < m.mNr );
    Q_ASSERT( EndColIn < m.mNc );
    int EndRow = qMin(m.mNr - 1, EndRowIn);
    int EndCol = qMin(m.mNc - 1, EndColIn);

    if ( 0          <= BeginRow &&
         EndRow     <= m.mNr    &&
         BeginRow   <  EndRow   &&
         0          <= BeginCol &&
         EndCol     <= m.mNc    &&
         BeginCol   <  EndCol   )
    {
        int ir;
        int newNr = EndRow - BeginRow + 1;  // Add 1 due to zero based indexes
        int newNc = EndCol - BeginCol + 1;

        if ( mMat.size() != newNr * newNc )
        {
            mNr = newNr;
            mNc = newNc;

            mMat.resize( mNr * mNc );
        }

        if ( !mMat.isEmpty() )
        {
            const double*    source  = m.mMat.data();
            double*          dest    = mMat.data();

            for ( ir = 0; ir < mNr; ir++ )
            {
                memcpy( (void *)&dest[ ir * mNc ],
                        (const void *)&source[ (ir + BeginRow) * m.mNc + BeginCol ],
                        size_t( mNc * sizeof(double) ) );
            }
        }
    }
    else
    {
        throw("PMatrix copy: indicated ranges out of matrix range");
    }

    return *this;
}


PMatrix PMatrix::selectRows( const QList<int> & rows ) const
{
    int             n       = rows.size();
    int             ir;
    int             jr;
    PMatrix         destMat( n, mNc );
    const double*   source  = mMat.data();
    double*         dest    = destMat.mMat.data();

    for ( jr = 0; jr < n; jr++ )
    {
        ir = rows[jr];

        if ( (0 <= ir) && (ir < mNr) )
        {
            memcpy( (void *)&dest[ jr * mNc ],
                    (const void *)&source[ ir * mNc ],
                    size_t( mNc * sizeof(double) ) );
        }
        else
        {
            throw("PMatrix selectRows: selected rows out of range");
        }
    }

    return destMat;
}

PMatrix PMatrix::selectColumns( const PIVector &columns ) const
{
    PMatrix m( mNr, columns.N() );
    int     ir;
    int     jc;
    int     ic;

    for ( ir = 0; ir < mNr; ir++ )
    {
        for ( jc = 0; jc < columns.N(); jc++ )
        {
            ic = columns[jc];

            if ( (0 <= ic) && (ic < mNc) )
            {
                m[ ir * m.Nc() + jc ] = mMat[ ir * mNc + ic ];
            }
            else
            {
                qDebug("PMatrix selectcolumns: selected column out of range");
            }
        }
    }

    return m;
}


PMatrix PMatrix::selectRows( int firstrow, int lastrowIn ) const
{
    Q_ASSERT( lastrowIn < mNr );
    int lastrow = qMin(mNr - 1, lastrowIn);


    PMatrix destMat;

    if ( firstrow >= 0 &&
         lastrow  >= firstrow &&
         lastrow  <= mNr )
    {
        destMat.resize( lastrow - firstrow + 1, mNc );

        const double*    source  = mMat.data();
        double*          dest    = destMat.mMat.data();

        for ( int index = 0; index < destMat.Nr(); index++ )
        {
            memcpy( (void *)&dest[ index * mNc ],
                    (const void *)&source[ (index + firstrow) * mNc ],
                    size_t( mNc * sizeof(double) ) );
        }
    }
    else
    {
        qDebug() << "PMatrix::selectRows(): rows to select out of range";
    }

    return destMat;
}


PMatrix PMatrix::selectColumns( int firstcol, int lastcolIn ) const
{

    Q_ASSERT( lastcolIn < mNc );
    int lastcol = qMin(mNc - 1, lastcolIn);

    PMatrix destMat;

    if ( firstcol >= 0 &&
         lastcol  >  firstcol &&
         lastcol  <  mNc )
    {
        destMat.resize( mNr , lastcol-firstcol + 1 );

        const double*   source  = mMat.constData();
        double*         dest    = destMat.mMat.data();

        for ( int ir = 0; ir < mNr; ir++ )
        {

            memcpy( (void *)&dest[ir * destMat.Nc()],
                    (const void *)&source[ir * mNc + firstcol],
                    size_t( destMat.Nc() * sizeof(double) ) );
        }
    }
    else
    {
        qDebug("PMatrix selectcolumns: columns to select out of range");
    }

    return destMat;
}


const PMatrix & PMatrix::addMatrixRows( const PMatrix & rows )
{
    if ( !rows.isEmpty() )
    {
        if ( mMat.isEmpty() )
        {
            *this = rows;
        }
        else if ( mNc == rows.mNc )
        {
            //int     newNr   = mNr + rows.mNr;
            int oldSize = mMat.size();

            mNr += rows.mNr;
            mMat.resize( mNr * mNc );

            if ( !mMat.isEmpty() )
            {
                const double*   source  = rows.mMat.constData();
                double*         dest    = mMat.data();

                // Now copy the new data to the correct memory position
                memcpy( (void *)&dest[oldSize],
                        (const void *)source,
                        size_t( rows.mMat.size() * sizeof(double) ) );
            }
        }
        else
        {
            throw( "PMatrix addMatrixRows: matrices have unequal number of columns" );
        }
    }

    return *this;
}


const PMatrix & PMatrix::addMatrixColumns( const PMatrix & cols )
{
    if ( !cols.isEmpty() )
    {
        if ( mMat.isEmpty() )
        {
            *this = cols;
        }
        else if ( mNr == cols.mNr )
        {
            PMatrix oldMat( *this );
            int     rowIndex        = 0;
            int     rowIndexOld     = 0;
            int     rowIndexCols    = 0;

            mNc += cols.mNc;
            mMat.resize( mNr * mNc );

            double *        matDataPtr  = mMat.data();
            const double *  oldDataPtr  = oldMat.mMat.constData();
            const double *  colsDataPtr = cols.mMat.constData();

            for ( int ir = 0; ir < mNr; ir++ )
            {
                memcpy( (void*)&matDataPtr[rowIndex],
                        (const void *)&oldDataPtr[rowIndexOld],
                        size_t(sizeof(double) * oldMat.Nc()) );

                memcpy( (void*)&matDataPtr[rowIndex + oldMat.Nc()],
                        (const void *)&colsDataPtr[rowIndexCols],
                        size_t(sizeof(double) * cols.Nc()) );

                rowIndex        += mNc;
                rowIndexOld     += oldMat.Nc();
                rowIndexCols    += cols.Nc();
            }
        }
        else
        {
            throw( "PMatrix addMatrixColumns: matrices have unequal number of rows" );
        }
    }

    return *this;
}


void PMatrix::addColumnsToMatrix(const PMatrix &cols, int startColumn)
{
    if ( !cols.isEmpty() && mNr == cols.mNr )
    {
        int lastColumn   = startColumn + cols.Nc();
        if ( mNc < lastColumn )
        {
            resize(mNr, lastColumn );
        }
        double*         mapData     = mMat.data();          // Get the matrix memory
        const double*   map2Data    = cols.mMat.constData();  // Get the m2 matrix memory
        int             map2DataNc  = cols.mNc;
        for ( int i = 0; i < mNr; i++ )
        {
            double*         rowi    = &mapData[i * mNc ];          // Get the matrix memory
            const double*   colsRowi= &map2Data[ i * map2DataNc ];  // Get the m2 matrix memory
            int jc = 0;
            for ( int j = startColumn; j < lastColumn; j++, jc++ )
            {
                rowi[ j ] += colsRowi[ jc ];
            }
        }
    }
    else
    {
        throw( "PMatrix addMatrixToColumns: matrices have unequal number of rows" );
    }
}

PMatrix PMatrix::transposed() const
{
    PMatrix M;
    M.resize(mNc,mNr);
    double*         mapData = M.mMat.data();          // Get the local matrix memory
    const double*   myData  = mMat.constData();  // Get the m1 matrix memory

    for ( int i = 0; i < mNr; i++ )
    {
        const double*   row = &myData[ i * mNc ];

        for ( int j = 0; j < mNc; j++ )
        {
            mapData[ j * mNr + i ] = row[ j ];
        }
    }
    return M;
}


const PMatrix & PMatrix::mul( const PMatrix & m1, const PMatrix & m2 )
{
    if ( m2.Nc() > 100 ) // in the mul function m2 is transposed, which is too much work for small matrices
    {
        return mul(m1,false,m2,false);
    }
    else
    {

        Q_ASSERT( this != &m1 );
        Q_ASSERT( this != &m2 );

        if ( m1.isEmpty() || m2.isEmpty() )
        {
            clear();
        }
        else
        {
            if ( m1.mNc == m2.mNr )
            {
                resize( m1.mNr, m2.mNc );

                if ( !mMat.isEmpty() )
                {
                    double          val;
                    double*         mapData = mMat.data();          // Get the local matrix memory
                    const double*   m1Data  = m1.mMat.constData();  // Get the m1 matrix memory
                    const double*   m2Data  = m2.mMat.constData();  // Get the m2 matrix memory

                    for ( int i = 0; i < mNr; i++ )
                    {
                        const double*   row1 = &m1Data[ i * m1.mNc ];

                        for ( int j = 0; j < mNc; j++ )
                        {
                            val = 0;

                            for ( int k = 0; k < m1.mNc; k++ )
                            {
                                val += row1[ k ] * m2Data[ k * mNc + j ];
                            }

                            mapData[ i * mNc + j ] = val;
                        }
                    }
                }
            }
            else
            {
                qDebug("PMatrix mul: matrices have unequal size  m1.mNc != m2.mNr )");
            }
        }
    }
    return *this;
}

const PMatrix &PMatrix::mul(const PMatrix &m1, bool transposed1, const PMatrix &m2, bool transposed2)
{
    Q_ASSERT( this != &m1 );
    Q_ASSERT( this != &m2 );

    if ( m1.isEmpty() || m2.isEmpty() )
    {
        clear();
    }
    else
    {
        if ( !transposed1 && !transposed2 )
        {
            if ( m1.mNc == m2.mNr )
            {
                resize( m1.mNr, m2.mNc );

                PMatrix m2t( m2.transposed() );

                if ( !mMat.isEmpty() )
                {
                    double          val;
                    double*         mapData = mMat.data();          // Get the local matrix memory

                    const double*   m1Data  = m1.mMat.constData();  // Get the m1 matrix memory
                    const double*   m2Data  = m2t.mMat.constData();  // Get the m2 matrix memory

                    for ( int i = 0; i < mNr; i++ )
                    {
                        const double*   row1    = &m1Data[ i * m1.mNc ];
                        double*         maprow  = &mapData[ i * mNc ];

                        for ( int j = 0; j < mNc; j++ )
                        {
                            const double*   row2 = &m2Data[ j * m2.mNr ];
                            val = 0;

                            for ( int k = 0; k < m1.mNc; k++ )
                            {
                                val += row1[ k ] * row2[ k ];
                            }

                            maprow[ j ] = val;
                        }
                    }
                }
            }
        }
        else if ( !transposed1 && transposed2 )
        {
            if ( m1.mNc == m2.mNc )
            {
                resize( m1.mNr, m2.mNr );

                if ( !mMat.isEmpty() )
                {
                    double          val;
                    double*         mapData = mMat.data();          // Get the local matrix memory

                    const double*   m1Data  = m1.mMat.constData();  // Get the m1 matrix memory
                    const double*   m2Data  = m2.mMat.constData();  // Get the m2 matrix memory

                    for ( int i = 0; i < mNr; i++ )
                    {
                        const double*   row1    = &m1Data[ i * m1.mNc ];
                        double*         maprow  = &mapData[ i * mNc ];

                        for ( int j = 0; j < mNc; j++ )
                        {
                            const double*   row2 = &m2Data[ j * m2.mNc ];
                            val = 0;

                            for ( int k = 0; k < m1.mNc; k++ )
                            {
                                val += row1[ k ] * row2[ k ];
                            }

                            maprow[ j ] = val;
                        }
                    }
                }
            }
        }
        else if ( transposed1 && !transposed2 )
        {
            if ( m1.mNr == m2.mNr )
            {
                resize( m1.mNc, m2.mNc );

                PMatrix m1t( m1.transposed() );
                PMatrix m2t( m2.transposed() );

                if ( !mMat.isEmpty() )
                {
                    double          val;
                    double*         mapData = mMat.data();          // Get the local matrix memory

                    const double*   m1Data  = m1t.mMat.constData();  // Get the m1 matrix memory
                    const double*   m2Data  = m2t.mMat.constData();  // Get the m2 matrix memory

                    for ( int i = 0; i < mNr; i++ )
                    {
                        const double*   row1    = &m1Data[ i * m1.mNr ];
                        double*         maprow  = &mapData[ i * mNc ];

                        for ( int j = 0; j < mNc; j++ )
                        {
                            const double*   row2 = &m2Data[ j * m2.mNr ];
                            val = 0;

                            for ( int k = 0; k < m1.mNr; k++ )
                            {
                                val += row1[ k ] * row2[ k ];
                            }

                            maprow[ j ] = val;
                        }
                    }
                }
            }
        }
        else if ( transposed1 && transposed2 )
        {
            if ( m1.mNr == m2.mNc )
            {
                resize( m1.mNc, m2.mNr );

                PMatrix m1t( m1.transposed() );

                if ( !mMat.isEmpty() )
                {
                    double          val;
                    double*         mapData = mMat.data();          // Get the local matrix memory

                    const double*   m1Data  = m1t.mMat.constData();  // Get the m1 matrix memory
                    const double*   m2Data  = m2.mMat.constData();  // Get the m2 matrix memory

                    for ( int i = 0; i < mNr; i++ )
                    {
                        const double*   row1    = &m1Data[ i * m1.mNr ];
                        double*         maprow  = &mapData[ i * mNc ];

                        for ( int j = 0; j < mNc; j++ )
                        {
                            const double*   row2 = &m2Data[ j * m2.mNc ];
                            val = 0;

                            for ( int k = 0; k < m1.mNr; k++ )
                            {
                                val += row1[ k ] * row2[ k ];
                            }

                            maprow[ j ] = val;
                        }
                    }
                }
            }

        }
        else
        {
            qDebug("PMatrix mul(const PMatrix &m1, bool transposed1, const PMatrix &m2, bool transposed2): matrices have unequal size  )");
        }
    }
    return *this;
}







void PMatrix::mulElement(const PMatrix &m1, const PMatrix &m2)
{
    Q_ASSERT( m1.Nc() == m2.Nc() );
    Q_ASSERT( m1.Nr() == m2.Nr() );


    mMat    = m1.mMat;
    mNc     = m1.mNc;
    mNr     = m1.mNr;

    double*         m1Data  = mMat.data();  // Get the m1 matrix memory
    const double*   m2Data  = m2.mMat.constData();  // Get the m2 matrix memory

    for ( int i=0; i < mNc * mNr; i++)
    {
        m1Data[i] *= m2Data[i];
    }
}


PMatrix & PMatrix::mulElement(const PMatrix &m1)
{
    Q_ASSERT( m1.Nc() == mNc );
    Q_ASSERT( m1.Nr() == mNr );

    double*         mData  = mMat.data();          // Get this matrix memory
    const double*   m1Data = m1.mMat.constData();  // Get the m1 matrix memory

    for ( int i=0; i < mNc * mNr; i++)
    {
        mData[i] *= m1Data[i];
    }
    return *this;
}


PMatrix PMatrix::ATA( void ) const
{
    PMatrix         MATA( mNc, mNc );
    PMatrix         M( (*this).transposed() );

    double          val;

    const double*   mapData = M.mMat.constData();  // Get the matrix memory

    double*         mataData = MATA.mMat.data();      // Get the matrix memory
    for ( int i = 0; i < MATA.mNc; i++ )
    {
        const double*   rowi = &mapData[ i * M.mNc ];
        double*   matarowi = &mataData[ i * mNc ];

        for ( int j = 0; j < MATA.mNc; j++ )
        {
            const double*   rowj = &mapData[ j * M.mNc ];
            val = 0;

            for ( int k = 0; k < M.mNc; k++ )
            {
                val += rowi[k] * rowj[k];
            }
            matarowi[j] = val;
            mataData[ i * mNc + j ] = val;
        }
    }

    return MATA;
}


PMatrix PMatrix::AAT( void ) const
{
    PMatrix         MAAT( mNr, mNr );
    double          val;
    const double*   mapData = mMat.constData();  // Get the matrix memory
    double*         outData = MAAT.mMat.data();  // Get the matrix memory

    for ( int i = 0; i < MAAT.mNr; i++ )
    {
        const double*   rowi = &mapData[ i * mNc ];
        double*   outrowi = &outData[ i * mNr ];


        for ( int j = 0; j < MAAT.mNr; j++ )
        {
            const double*   rowj = &mapData[ j * mNc ];

            val = 0.0;

            for ( int k = 0; k < mNc; k++ )
            {
                val += rowi[k] * rowj[k];
            }
            outrowi[j] = val;
        }
    }
    return MAAT;
}


void PMatrix::ATA(PMatrix &MATA) const
{
    MATA.resize( mNc, mNc );
    PMatrix         M( (*this).transposed() );

    double          val;

    const double*   mapData = M.mMat.constData();  // Get the matrix memory

    double*         mataData = MATA.mMat.data();      // Get the matrix memory
    for ( int i = 0; i < MATA.mNc; i++ )
    {
        const double*   rowi = &mapData[ i * M.mNc ];
        double*   matarowi = &mataData[ i * mNc ];

        for ( int j = 0; j < MATA.mNc; j++ )
        {
            const double*   rowj = &mapData[ j * M.mNc ];
            val = 0;

            for ( int k = 0; k < M.mNc; k++ )
            {
                val += rowi[k] * rowj[k];
            }
            matarowi[j] = val;
        }
    }
}


void PMatrix::AAT(PMatrix &MAAT) const
{
    MAAT.resize( mNr, mNr );
    double          val;
    const double*   mapData = mMat.constData();  // Get the matrix memory
    double*         outData = MAAT.mMat.data();  // Get the matrix memory

    for ( int i = 0; i < MAAT.mNr; i++ )
    {
        const double*   rowi = &mapData[ i * mNc ];
        double*   outrowi = &outData[ i * mNr ];


        for ( int j = 0; j < MAAT.mNr; j++ )
        {
            const double*   rowj = &mapData[ j * mNc ];

            val = 0.0;

            for ( int k = 0; k < mNc; k++ )
            {
                val += rowi[k] * rowj[k];
            }
            outrowi[j] = val;
        }
    }
}

int PMatrix::nonZeros( void ) const
{
    int             nonzeros    = 0;
    const double*   mapData     = mMat.constData();  // Get the matrix memory

    for ( int i = 0; i < mNr * mNc; i++ )
    {
        //if ( mMat[i] != 0.0 )
        if ( mapData[i] != 0.0 )
        {
            nonzeros++;
        }
    }

    return nonzeros;
}


void PMatrix::normalize( void )
{
    double  max     = getMax();
    double  min     = getMin();
    double* mapData = mMat.data();  // Get the matrix memory

    max -= min;

    for ( int i = 0; i < mNc * mNr; i++ )
    {
        mapData[i] -= min;
        mapData[i] /= max;
    }
}


void PMatrix::normalizeRows( void )
{
    double  max;
    double  min;
    int     i;
    int     ir;
    int     ic;
    double* mapData = mMat.data();  // Get the matrix memory

    for ( i = 0, ir = 0; i < mNr; i++, ir += mNc )
    {
        max = mapData[ ir + 0 ];
        min = mapData[ ir + 0 ];

        for ( ic = 1; ic < mNc; ic++ )
        {
            if ( mapData[ ir + ic ] > max )
            {
                max = mapData[ ir + ic ];
            }

            if ( mapData[ ir + ic ] < min )
            {
                min = mapData[ ir + ic ];
            }
        }

        max -= min;

        for ( ic = 0; ic < mNc; ic++ )
        {
            mapData[ ir + ic ] -= min;
            mapData[ ir + ic ] /= max;
        }
    }
}

void PMatrix::normalizeRow( int index )
{
    Q_ASSERT( index >= 0 && index < mNr);

    double  max;
    double  min;
    int     ir = index;
    int     ic;
    double* mapData = mMat.data();  // Get the matrix memory

    max = mapData[ ir + 0 ];
    min = mapData[ ir + 0 ];

    for ( ic = 1; ic < mNc; ic++ )
    {
        if ( mapData[ ir + ic ] > max )
        {
            max = mapData[ ir + ic ];
        }

        if ( mapData[ ir + ic ] < min )
        {
            min = mapData[ ir + ic ];
        }
    }

    max -= min;

    for ( ic = 0; ic < mNc; ic++ )
    {
        mapData[ ir + ic ] -= min;
        mapData[ ir + ic ] /= max;
    }
}


double PMatrix::norm1() const
{
    const double*   mapData = mMat.constData(); // Get the matrix memory
    double          mnorm   = 0;

    //column sum
    for ( int ic = 0; ic < mNc; ic++ )
    {
        double tnorm = 0.0;
        for ( int ir = 0; ir < mNr * mNc; ir += mNc )
        {
            tnorm  += fabs( mapData[ir + ic] );
        }

        mnorm = qMax( mnorm, tnorm);
    }

    return mnorm;
}


double PMatrix::normInf() const
{
    const double*   mapData = mMat.constData(); // Get the matrix memory
    const double*   row;
    double          mnorm   = 0;
    //row sum
    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        double tnorm = 0.0;

        row  = &mapData[ir];      // Get the row to work with from the matrix

        for ( int ic = 0; ic < mNc; ic++ )
        {
            tnorm  += fabs( row[ic] );
        }

        mnorm = qMax( mnorm, tnorm);
    }

    return mnorm;
}

#ifdef PEACSCLASSES

double* dvector(int nl, int nh);
void free_dmatrix(double **m, int nrl, int nrh, int ncl, int nch);
double **dmatrix(int nrl, int nrh, int ncl, int nch);
void free_dvector(double *v, int nl, int nh);

#ifndef svdcmpname
#define svdcmpname	svdcmp
#endif
void svdcmpname(double **a, int m, int n, double *w, double **v);

double PMatrix::norm2() const
{
    int nSigma = qMax(mNr,mNc);
    double **u = dmatrix( 1, mNr, 1, mNc);
    double *sigma = dvector( 1, nSigma );
    double **v = dmatrix( 1, mNc, 1, mNc);
    double maxSigma = 0;

//    const double* mapData   = mMat.data();          // Get the matrix memory


    if ( u != NULL && sigma != NULL && v != NULL)
    {
        for ( int i = 1; i <= mNr; i++ )
        {
            for ( int j = 1; j <= mNc; j++ )
            {
                u[i][j] = value(i-1,j-1);
            }
        }
        //
        svdcmp( u, mNr, mNc, sigma, v );

        // check inverse matrix
        maxSigma = sigma[1];
        for (int i=1; i <= nSigma; i++ )
        {
            maxSigma = qMax(maxSigma,sigma[i]);
        }
    }
    free_dmatrix( v, 1, mNc, 1, mNc );
    free_dvector( sigma, 1, nSigma );
    free_dmatrix( u, 1, mNr, 1, mNc );

    return maxSigma;
}
#endif

double PMatrix::frobeniusNorm() const
{
    int     size            = mNr * mNc;
    const double* mapData   = mMat.data();          // Get the matrix memory
    double frobnorm         = 0.0;
    for ( int i = 0; i < size; i++ )
    {
        double a = mapData[i];
        frobnorm += (a * a);
    }
    frobnorm = sqrt(frobnorm);

    return frobnorm;
}


PMatrix & PMatrix::scaleRows( const PVector & scale)
{
    Q_ASSERT( scale.N() == mNr );

    int             i           = 0;
    const double*   vecPtr      = scale.data();
          double*   mapData     = mMat.data(); // Get the matrix memory
          double*   row;

    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        row         = &mapData[ir];      // Get the row to work with from the matrix
        double value= vecPtr[ir];
        for ( int ic = 0; ic < mNc; ic++ )
        {
            row[ic] *= value;
        }

        i++;
    }
    return *this;
}


PVector PMatrix::minRows( void ) const
{
    PVector         v( mNr );
    int             i           = 0;
    double *        vecPtr      = v.data();
    const double*   mapData     = mMat.data(); // Get the matrix memory
    const double*   row;

    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        row         = &mapData[ir];      // Get the row to work with from the matrix
        //v[i]        = row[0];            // Initialize with the first item
        vecPtr[i]   = row[0];            // Initialize with the first item

        for ( int ic = 1; ic < mNc; ic++ )
        {
            vecPtr[i] = qMin( vecPtr[i], row[ic] );
        }

        i++;
    }

    return v;
}


PVector PMatrix::maxRows( void ) const
{
    PVector         v( mNr );
    int             i           = 0;
    double *        vecPtr      = v.data();
    const double*   mapData     = mMat.constData(); // Get the matrix memory
    const double*   row;

    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        row  = &mapData[ir];    // Get the row to work with from the matrix
        vecPtr[i] = row[0];     // Initialize with the first item

        for ( int ic = 1; ic < mNc; ic++ )
        {
            vecPtr[i] = qMax( vecPtr[i], row[ic]);
        }

        i++;
    }

    return v;
}


PVector PMatrix::meanRows( void ) const
{
    PVector         v( mNr );
    int             i           = 0;
    double *        vecPtr      = v.data();
    const double*   mapData     = mMat.constData(); // Get the matrix memory
    const double*   row;

    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        row  = &mapData[ir];    // Get the row to work with from the matrix
        vecPtr[i] = row[0];     // Initialize with the first item

        for ( int ic = 1; ic < mNc; ic++ )
        {
            vecPtr[i] += row[ic];
        }
        vecPtr[i] /= (double)mNc;

        i++;
    }

    return v;
}

PVector PMatrix::meanCols( void ) const
{
    PVector         v( mNc );
    double *        vecPtr      = v.data();
    const double*   mapData     = mMat.constData(); // Get the matrix memory

    for ( int ic = 0; ic < mNc; ic++ )
    {
        vecPtr[ic] = 0;     // Initialize with the first item
        for ( int ir = 0; ir < mNr * mNc; ir += mNc )
        {
            vecPtr[ic] += mapData[ ir + ic ];
        }
        vecPtr[ic] /= (double)mNr;
    }
    return v;
}



PVector PMatrix::minCols( void ) const
{
    PVector         v( mNc );
    double *        vecPtr  = v.data();
    const double *  dataPtr = mMat.constData();

    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        vecPtr[0] = dataPtr[ir];       // Initialize with the first item

        for ( int ic = 1; ic < mNc; ic++ )
        {
            vecPtr[ic] = qMin( vecPtr[ic], dataPtr[ ir + ic ] );
        }
    }

    return v;
}


PVector PMatrix::maxCols( void ) const
{
    PVector         v( mNc );
    double *        vecPtr  = v.data();
    const double *  dataPtr = mMat.constData();

    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        vecPtr[0] = dataPtr[ir];       // Initialize with the first item

        for ( int ic = 1; ic < mNc; ic++ )
        {
            vecPtr[ic] = qMax( vecPtr[ic], dataPtr[ ir + ic ] );
        }
    }

    return v;
}


double PMatrix::getMin( void ) const
{
    double min = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mMat.constData();

        min = dataPtr[0];

        for ( int i = 1; i < mNr * mNc; i++ )
        {
            if ( dataPtr[i] < min )
            {
                min = dataPtr[i];
            }
        }
    }

    return min;
}


double PMatrix::getMax( void ) const
{
    double max = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mMat.constData();

        max = dataPtr[0];

        for ( int i = 1; i < mNr * mNc; i++ )
        {
            if ( dataPtr[i] > max )
            {
                max = dataPtr[i];
            }
        }
    }
    return max;
}


PVector PMatrix::sumRows() const
{
    PVector         v( mNr );
    int             i           = 0;
    double *        vecPtr      = v.data();
    const double*   mapData     = mMat.constData(); // Get the matrix memory
    const double*   row;

    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        row  = &mapData[ir];    // Get the row to work with from the matrix
        vecPtr[i] = row[0];     // Initialize with the first item

        for ( int ic = 1; ic < mNc; ic++ )
        {
            vecPtr[i] += row[ic];
        }
        i++;
    }
    return v;
}

PVector PMatrix::sumCols() const
{
    PVector         v( mNc );
    double *        vecPtr      = v.data();
    const double*   mapData     = mMat.constData(); // Get the matrix memory

    for ( int ic = 0; ic < mNc; ic++ )
    {
        vecPtr[ic] = 0;     // Initialize with the first item
        for ( int ir = 0; ir < mNr * mNc; ir += mNc )
        {
            vecPtr[ic] += mapData[ ir + ic ];
        }
    }
    return v;
}


void PMatrix::zeromeanRows()
{
    PVector     v( meanRows() );
    double *    vecPtr  = v.data();
    double *    dataPtr = mMat.data();
    int i = 0;
    for ( int ir = 0; ir < mNr * mNc; ir += mNc )
    {
        for ( int ic = 0; ic < mNc; ic++ )
        {
            dataPtr[ ir + ic ] -= vecPtr[i];
        }
        i++;
    }
}



PVector PMatrix::col( int col ) const
{
    Q_ASSERT( col < mNc );
    int i = qMin(mNc - 1, col);

    if ( mMat.isEmpty() || i < 0 || i >= mNc )
    {
        return PVector();
    }
    else
    {
        PVector         v( mNr );
        double *        vecPtr  = v.data();
        const double *  dataPtr = mMat.constData();

        for ( int j = 0; j < mNr; j++ )
        {
            vecPtr[j] = dataPtr[ j * mNc + i ];
        }

        return v;
    }
}


void PMatrix::setCol( int col, const PVector& vec )
{
    Q_ASSERT( col < mNc );
    int i = qMin(mNc - 1, col);

    if ( vec.N() == mNr )
    {
        double *        dataPtr = mMat.data();
        const double *  vecPtr  = vec.data();

        for ( int j = 0; j < mNr; j++ )
        {
            dataPtr[j * mNc + i] = vecPtr[ j ];
        }
    }
    else
    {
        qDebug("PMatrix::setCol vector size not the same as the number of rows");
    }
}


void PMatrix::setRowToZero( int row )
{
    Q_ASSERT( row < mNr );
    int i = qMin(mNr - 1, row);

    if ( i >= 0 && i <= mNr )
    {
        int         ir      = i * mNc;
        double *    dataPtr = mMat.data();

        for ( int j = 0; j < mNc; j++ )
        {
            Q_ASSERT( ir + j < mMat.size() );

            dataPtr[ ir + j ] = 0.0;
        }
    }
}


void PMatrix::setColToZero( int col )
{
    Q_ASSERT( col < mNc );
    int i = qMin(mNc - 1, col);

    if ( i >= 0 && i < mNc )
    {
        double * dataPtr = mMat.data();

        for ( int j = 0; j < mNr; j++ )
        {
            Q_ASSERT( j * mNc + i < mMat.size() );

            dataPtr[ j * mNc + i ] = 0.0;
        }
    }
}


PVector PMatrix::row( int i ) const
{
    PVector v;

    row( i, v );

    return v;
}


bool PMatrix::row( int row, PVector & destination ) const
{
    Q_ASSERT( row < mNr );
    int i = qMin(mNr - 1, row);

    bool done = false;

    if ( mMat.isEmpty() || i < 0 || i >= mNr )
    {
        destination.clear();
        qDebug() << " PMatrix::row( int i ) invalid row " << i;
    }
    else
    {
        const double * row = &( mMat.constData()[ i * mNc ] );

        destination.resize( mNc );

        Q_ASSERT( destination.size() == mNc );

        done = true;

        memcpy( (void*)destination.data(), (const void*)row, size_t( mNc * sizeof(double) ) );
    }

    return done;
}


void PMatrix::setRow( int row, const PVector& vec )
{
    Q_ASSERT( row < mNr );
    int i = qMin(mNr - 1, row);

    if ( vec.N() == mNc )
    {
        double* row = &( mMat.data()[ i * mNc ] );

        memcpy( (void*)row, (const void*)vec.data(), size_t( mNc * sizeof(double) ) );
    }
    else
    {
        qDebug( "PMatrix::setRow vector size not the same as the number of columns" );
    }
}


bool PMatrix::save( const QString &filename, int precision ) const
{
    int             ir;
    int             ic;
    bool            done = false;
    QFile           f( filename );
    const double *  dataPtr = mMat.constData();

    if ( !f.open( QIODevice::WriteOnly ) )
    {
        return false;
    }

    if ( precision <= 0 )
    {
        QDataStream dataStream( &f );

        dataStream.setByteOrder( QDataStream::LittleEndian );
        dataStream.setFloatingPointPrecision( QDataStream::SinglePrecision );

        dataStream << mNr;
        dataStream << mNc;

        for ( ir = 0; ir < mNr && dataStream.status() == QDataStream::Ok; ir++ )
        {
            for ( ic = 0; ic < mNc && dataStream.status() == QDataStream::Ok; ic++ )
            {
                //dataStream << mMat[ ir * mNc + ic ];
                dataStream << dataPtr[ ir * mNc + ic ];
            }
        }

        done = dataStream.status() == QDataStream::Ok;

        f.close();
    }
    else
    {
        QTextStream textStream( &f );

        textStream.setRealNumberPrecision( precision );

        textStream << mNr;
        textStream << " ";
        textStream << mNc;
        textStream << "\n";

        for ( ir = 0; ir < mNr && textStream.status() == QTextStream::Ok; ir++ )
        {
            for ( ic = 0; ic < mNc && textStream.status() == QTextStream::Ok; ic++ )
            {
                //textStream << mMat[ ir * mNc + ic ];
                textStream << dataPtr[ ir * mNc + ic ];
                textStream << " ";
            }

            textStream << "\n";
        }

        done = textStream.status() == QTextStream::Ok;

        f.close();
    }

    return done;
}


bool PMatrix::load( const QString &filename )
{
    bool    ok = false;

    if ( !filename.isEmpty() )
    {
        QFile f( filename);

        if ( f.open( QIODevice::ReadOnly ) )
        {
            qint32      nr;
            qint32      nc;
            QDataStream dataStream( &f );
            qint64      size = f.bytesAvailable();

            dataStream.setByteOrder( QDataStream::LittleEndian ); // default used in matlab

            const char magic[] = ";;mbfmat";
            char buf[10];
            dataStream.readRawData(buf,strlen(magic));
            buf[strlen(magic)]='\0';
                    /* structure of new type binary matrix files:

                    8 bytes		";;mbfmat"	(magic number)
                    1 byte  	bigendian	true if file is bigendian
                    4 bytes		headersize	number of bytes in header
                    <begin of header>
                    1 byte		data sign	0=unsigned, 1=signed
                    1 byte		data type	0=integer, 1=doubleing
                    1 byte		data size	number of bytes per element
                    4 bytes 	number of rows
                    4 bytes 	number of columns
                    */
            if ( strncmp(buf, magic, strlen(magic)) ==0 )
            {
                char c[1];
                qint32 ii;
                dataStream.readRawData(c,1);
                dataStream >> ii;
                dataStream.readRawData(c,1);
                dataStream.readRawData(c,1);
                dataStream.readRawData(c,1);
            }
            else
            {
                dataStream.device()->seek(0);
            }

            dataStream >> nr;

            if ( dataStream.status() == QDataStream::Ok &&  nr >= 0 )
            {
                dataStream >> nc;

                if ( size == nc * nr * sizeof( float ) + 2 * sizeof( qint32 ) )
                {
                    dataStream.setFloatingPointPrecision( QDataStream::SinglePrecision );
                }
                else
                {
                    dataStream.setFloatingPointPrecision( QDataStream::DoublePrecision );
                }

                if ( dataStream.status() == QDataStream::Ok &&  nc >= 0 )
                {
                    resize( nr, nc );

                    if ( !isEmpty() )
                    {
                        double * dataPtr = mMat.data();

                        ok = true;

                        for ( int ir = 0; ir < mNr; ir++ )
                        {
                            for ( int ic = 0; ic < mNc; ic++ )
                            {
                                //dataStream >>  mMat[ ir * mNc + ic ];
                                dataStream >>  dataPtr[ ir * mNc + ic ];

                                if ( dataStream.status() != QDataStream::Ok )
                                {
                                    ok = false;

                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }

        if ( !ok )
        {
            clear();
        }

        f.close();
    }

    return ok;
}


bool PMatrix::read( QDataStream &dataStream )
{
    bool    ok      = false;
    int     version = readHead( dataStream, PMatrix::getClassName() );

    if ( version != READ_HEADER_FAILED )
    {
        qint32 nr;
        qint32 nc;

        switch ( version )
        {
        case FILE_MATRIX_VERSION_1:
            clear();

            dataStream >> nr;
            dataStream >> nc;

            if ( isOk( dataStream ) )
            {
                if ( nr > 0 && nc > 0 )
                {
                    resize( (int)nr, (int)nc );

                    if ( !isEmpty() ) // no memory problems?
                    {
                        qint32      ir;
                        qint32      ic;
                        double *    dataPtr = mMat.data();

                        ok = true;

                        // read the matrix
                        for ( ir = 0; ir < mNr && ok; ir++ )
                        {
                            for ( ic = 0; ic < mNc && ok; ic++ )
                            {
                                //dataStream >> mMat[ ir * mNc + ic ];
                                dataStream >> dataPtr[ ir * mNc + ic ];

                                ok = isOk( dataStream );
                            }
                        }
                    }
                }
                else
                {
                    ok = true;  // Nothing more to read
                }
            }
            break;

        default:
            break;
        }
    }

    return ok;

}


bool PMatrix::write( QDataStream &dataStream )const
{
    bool ok = writeHead( dataStream, PMatrix::getClassName(), PMatrix::getDefaultVersion() );

    if ( ok )
    {
        const double * dataPtr = mMat.constData();

        switch ( PMatrix::getDefaultVersion() )
        {
        case FILE_MATRIX_VERSION_1:
            dataStream << (qint32)mNr;
            dataStream << (qint32)mNc;

            ok = isOk( dataStream );

            for ( int ir = 0; ir < mNr && ok; ir++ )
            {
                for ( int ic = 0; ic < mNc && ok; ic++ )
                {
                    //dataStream << mMat[ ir * mNc + ic ];
                    dataStream << dataPtr[ ir * mNc + ic ];

                    ok = isOk( dataStream );
                }
            }
            break;

        default:
            ok = false;
            break;
        }
    }

    return isOk( ok, dataStream );
}


#ifndef QT_NO_DEBUG

int PMatrix::dataOffset( int indexR, int indexC ) const
{
    Q_ASSERT( !mMat.isEmpty() );                // Test just to be sure
    Q_ASSERT( indexR >= 0 && indexR < mNr );    // Test just to be sure
    Q_ASSERT( indexC >= 0 && indexC < mNc );    // Test just to be sure

    return indexR * mNc + indexC;
}


double & PMatrix::operator[] ( int index )
{
    Q_ASSERT( !mMat.isEmpty() );                    // Test just to be sure
    Q_ASSERT( index >= 0 && index < mMat.size() );  // Test just to be sure

    return mMat[index];
}


const double &PMatrix::operator[] ( int index ) const
{
    Q_ASSERT( !mMat.isEmpty() );                    // Test just to be sure
    Q_ASSERT( index >= 0 && index < mMat.size() );  // Test just to be sure

    return mMat[index];
}


const double &PMatrix::at( int index ) const
{
    Q_ASSERT( !mMat.isEmpty() );                    // Test just to be sure
    Q_ASSERT( index >= 0 && index < mMat.size() );  // Test just to be sure

    return mMat[index];
}


double  &PMatrix::value( int indexR, int indexC )
{
    Q_ASSERT( !mMat.isEmpty() );                // Test just to be sure
    Q_ASSERT( indexR >= 0 && indexR < mNr );    // Test just to be sure
    Q_ASSERT( indexC >= 0 && indexC < mNc );    // Test just to be sure

    return mMat[ indexR * mNc + indexC ];
}


const double  &PMatrix::value( int indexR, int indexC ) const
{
    Q_ASSERT( !mMat.isEmpty() );                // Test just to be sure
    Q_ASSERT( indexR >= 0 && indexR < mNr );    // Test just to be sure
    Q_ASSERT( indexC >= 0 && indexC < mNc );    // Test just to be sure

    return mMat[ indexR * mNc + indexC ];
}


const double  &PMatrix::operator ()( int indexR, int indexC ) const
{
    Q_ASSERT( !mMat.isEmpty() );                // Test just to be sure
    Q_ASSERT( indexR >= 0 && indexR < mNr );    // Test just to be sure
    Q_ASSERT( indexC >= 0 && indexC < mNc );    // Test just to be sure

    return mMat[ indexR * mNc + indexC ];
}

double  &PMatrix::operator ()( int indexR, int indexC )
{
    Q_ASSERT( !mMat.isEmpty() );                // Test just to be sure
    Q_ASSERT( indexR >= 0 && indexR < mNr );    // Test just to be sure
    Q_ASSERT( indexC >= 0 && indexC < mNc );    // Test just to be sure

    return mMat[ indexR * mNc + indexC ];
}


#endif
