/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pmatrix.h 1129 2014-02-28 12:03:49Z peter $
 *
 */

#ifndef __PMATRIX_H
#define __PMATRIX_H


#include <QVector>
#include "pvector.h"
#include "pfile.h"

#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {
        class PVector;
        class PIVector;

        /** @class PMatrix pmatrix.h "support/pmatrix.h"
          * @brief This class hold a 2 dimensional matrix and its functions
          */
        class PEACSSUPPORTSHARED_EXPORT PMatrix : public PFile
        {

        public:

            /** @brief Default constructor.
              * Creates an empty matrix
              */
            PMatrix( void );


            /** @brief Assignment constructor.
              * Copies the given matrix
              * mNr,mNc, mMin and mMax are copied too
              */
            PMatrix( const PMatrix &m );


            /** @brief Constructor.
              * reserve memory (nr*mc)
              * @param  nr is the number of columns
              * @param  nc is the number of rows
              */
            PMatrix( int nr, int nc );

            /** @brief Constructor.
              * reserve memory (nr*mc)
              * @param  nr is the number of columns
              * @param  nc is the number of rows
              * @param  val is used to fill the matrix
              */
            PMatrix( int nr, int nc ,double val);


            /** Destructor.
              * @brief Free memory of the matrix
              */
            virtual ~PMatrix( void );


            /** @brief clear matrix
              */
            void clear( void );


            /** @brief set thesize of the matrix to nr x nc
              */
            void resize( int mnr, int mnc );


            /** @brief Returns the minimum value of all elements in the matrix.
              */
            double getMin( void ) const;


            /** @brief Returns the maximum value of all elements in the matrix.
              */
            double getMax( void ) const;


            /** @returns the sum of the rows
              */
            PVector sumRows() const;


            /** @returns the sum of the rows
              */
            PVector sumCols() const;



            /** @brief Returns true if the Matrix is empty otherwise false.
              */
            inline bool isEmpty( void ) const { return mMat.isEmpty(); }


            /** @brief Returns the number of columns of the matrix.
              */
            inline int Nr( void ) const { return mNr; }


            /** @brief Returns the number of rows of the matrix.
              */
            inline int Nc( void ) const { return mNc; }


            /** @brief Returns the number of elements of the matrix.
              */
            inline int size( void ) const { return mNr * mNc; }


            /** @brief Returns the number of elements of the matrix.
              */
            inline int count( void ) const { return mNr * mNc; }


            /** @brief returns the data pointer
              */
            inline const double * data( void ) const { return mMat.data(); }


            /** @brief returns the data pointer
              */
            inline double * data( void ) { return mMat.data(); }


            /** @brief Copy operator.
              * @see  PMatrix(const PMatrix &m)
              */
            const PMatrix & operator= ( const PMatrix &m );


            /** @brief Copies a part of the matrix, indicated by the range parameters.
              * @param  m        The matrix from which data is copied
              * @param  BeginRow First row (zero based) from which the data is copied
              * @param  EndRow   Last row (zero based) from which the data is copied
              * @param  BeginCol First column (zero based) from which the data is copied
              * @param  EndCol  Last column (zero based) from which the data is copied
              * returns This object with a part of matrix m
              */
            PMatrix & copy( const PMatrix & m, int BeginRow, int EndRow, int BeginCol, int EndCol );


            /** @brief Copies a part of the rows of the matrix, and returns this
              * part of the matrix
              * @param rows list of rows (0-based indexes) that are copied into the return matrix
              * @return part of the matrix (number of columns equals the number
              *  of columns in the original matrix
              */
            PMatrix selectRows( const QList<int> & rows ) const;


// TODO Check this functionality according to PMatrix selectColumns( const PIVector &columns ) const
//            /** @brief Copies a part of the rows of the matrix, and returns this
//              * part of the matrix
//              * @param rows vector indicating (1 or 0) which rows are copied into the return matrix
//              * @return part of the matrix (number of columns equals the number
//              *  of columns in the original matrix
//              */
//            PMatrix selectRows( const PIVector & rows ) const;


            /** @brief Copies a part of the rows of the matrix, and returns this
              * part of the matrix
              * @param firstrow the first row (0-based index) of rows that are to be copied into the return matrix
              * @param lastrow the last row (0-based index) of rows that are to be copied into the return matrix
              * @return part of the matrix (number of columns equals the number
              *  of columns in the original matrix
              */
            PMatrix selectRows( int firstrow, int lastrow ) const;


            /** @brief Copies a part of the columns of the matrix, and returns this
              * part of the matrix
              * @param columns vector indicating which columns (0-based index) are copied into the return matrix
              * @return part of the matrix (number of columns equals the requested columns number)
              *         of rows in the original matrix
              */
            PMatrix selectColumns( const PIVector & columns ) const;


            /** @brief Copies a part of the columns of the matrix, and returns this
              * part of the matrix
              * @param firstrow the first column of columns that are to be copied into the return matrix
              * @param lastrow the last column of columns that are to be copied into the return matrix
              * @return part of the matrix (number of rows equals the number
              *  of rows in the original matrix
              */
            PMatrix selectColumns( int firstcol, int lastcol ) const;


            /** @brief resizes the matrix and adds the rows to the end of the matrix
              * @param rows
              */
            const PMatrix& addMatrixRows( const PMatrix &rows );

            /** @brief resizes the matrix and adds the cols to the end of the matrix
              * @param cols
              */
            const PMatrix& addMatrixColumns( const PMatrix &cols );

            /** @brief resizes the matrix and adds the cols to the end of the matrix
              * @param cols
              */
            void addColumnsToMatrix( const PMatrix &cols, int startColumn);




        #ifndef QT_NO_DEBUG

            /** @brief Returns the index, specified by the column/row indexes, into the data memory matrix array
              * Use this function when accessing the matrix memory directly when using the data() function
              *
              * @param indexR index of the row
              * @param indexC index of the column
              * @return Index into the matrix array
              */
            int dataOffset( int indexR, int indexC ) const;


            /** @brief Returns the item at index position i as a modifiable
              * reference.
              * The matrix is indexed accoring to: index=indexR*mNc+indexC
              * @param index must be a valid index position in the list
              * (i.e., 0 <= index < Nr()*Nc()  ).
              * @see value(int indexR,int indexC)
              */
            double & operator[] ( int index );


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @see double   &operator[] (int index)
              */
            const double & operator[] ( int index ) const;


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @see double   &operator[] (int index)
              */
            const double & at( int index ) const;


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            double & value( int indexR, int indexC );


            /** @brief This is an overloaded member function provided for convenience
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            const double & value( int indexR, int indexC ) const;


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            double & operator() ( int indexR, int indexC );

            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            const double & operator() ( int indexR, int indexC ) const;


        #else

            /** @brief Returns the index, specified by the column/row indexes, into the data memory matrix array
              * Use this function when accessing the matrix memory directly when using the data() function
              *
              * @param indexR index of the row
              * @param indexC index of the column
              * @return Index into the matrix array
              */
            inline int dataOffset( int indexR, int indexC ) const { return indexR * mNc + indexC; }


            /** @brief Returns the item at index position i as a modifiable
              * reference.
              * The matrix is indexed accoring to: index=indexR*mNc+indexC
              * @param index must be a valid index position in the list
              * (i.e., 0 <= index < Nr()*Nc()  ).
              * @see value(int indexR,int indexC)
              */
            inline double & operator[] ( int index ) { return mMat[index]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @see double   &operator[] (int index)
              */
            inline const double & operator[]( int index ) const { return mMat[index]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @see double   &operator[] (int index)
              */
            inline const double & at( int index ) const { return mMat[index]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            inline double & value( int indexR, int indexC ) { return mMat[ indexR * mNc + indexC ]; }


            /** @brief This is an overloaded member function provided for convenience
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            inline const double & value( int indexR, int indexC ) const { return mMat.at( indexR * mNc + indexC ); }

            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            inline double & operator ()( int indexR, int indexC ) { return mMat[ indexR * mNc + indexC ]; }


            /** @brief This is an overloaded member function provided for convenience
              * @param indexR index of the row
              * @param indexC index of the column
              * @see double   &operator[] (int index)
              */
            inline const double & operator ()( int indexR, int indexC ) const { return mMat.at( indexR * mNc + indexC ); }

        #endif

            /** @brief Copies matrix and adds a constant (f) to all elements
              * of the matrix.
              * @returns the matrix
              * @see operator+=(const double &f)
              */
            PMatrix operator+ ( double f ) const;


            /** @brief Copies matrix and subracts a constant (f) from all
              * elements of this new matrix.
              * @returns the matrix
              * @see operator-=(const double &f)
              */
            PMatrix operator- ( double f ) const;


            /** @brief Copies matrix and multiplies all elements in this new
              * matrix with a constant (f).
              * @returns the matrix
              * @see operator*=(const double &f)
              */
            PMatrix operator* ( double f ) const;


            /** @brief Adds a constant (f) to all elements of the matrix
              * and returns modified matrix.
              * @returns this object
              * @see operator+(const double &f)
              */
            PMatrix & operator+= ( double f );


            /** @brief Subracts a constant (f) from all elements of the matrix
              * and returns modified matrix.
              * @returns this object
              * @see operator-(const double &f)
              */
            PMatrix & operator-= ( double f );


            /** @brief Subracts the matrix from a constant (f)
              * elements of this new matrix.
              * @returns the matrix
              * @see operator-=(const double &f)
              */
            PMatrix & subtractMatrixFromVal( double f );


            /** @brief Multiplies all elements in the matrix with a constant
              * (f)and returns modified matrix.
              * @returns this object
              * @see operator*(const double &f)
              */
            PMatrix & operator*= ( double f );


            /** @brief Copies matrix and element wise addition of both matrices.
              * @returns the matrix
              * @see operator+=(const PMatrix &m2)
              */
            PMatrix operator+ ( const PMatrix &m2 ) const;


            /** @brief elementwise addition of both matrices and returns
              * modified matrix.
              * @returns this object
              * @see operator+(const PMatrix &m2)
              */
            PMatrix & operator+= ( const PMatrix &m2 );


            /** @brief Copies matrix and elementwise subtraction of both
              * matrices.
              * @returns the matrix
              * @see operator-=(const PMatrix &m2)
              */
            PMatrix operator- ( const PMatrix &m2 ) const;


            /** @brief elementwise subtraction of both matrices and returns
              * modified matrix.
              * @returns this object
              * @see operator-(const PMatrix &m2)
              */
            PMatrix & operator-= ( const PMatrix &m2 );


            /** @brief Copies matrix and multiplies it with m2.
              * @param m2 matrix (m2.mNr has to be the same as mNc)
              * @returns the modified matrix
              * @see mul(const PMatrix &m1, const PMatrix &m2)
              * @see operator*=(const PMatrix &m2)
              */
            PMatrix operator * ( const PMatrix &m2 ) const;

            /** @brief multiplies this matrix with m2 and returns this object.
              * @returns this object
              * @see mul(const PMatrix &m1, const PMatrix &m2)
              * @see operator*(const PMatrix &m2)
              */
            PMatrix & operator*= ( const PMatrix &m2 );


            /** @brief Copies matrix and divides it by m2.
              * @param m2 matrix (m2.mNr has to be the same as mNc)
              * @returns the modified matrix
              * @see mul(const PMatrix &m1, const PMatrix &m2)
              * @see operator*=(const PMatrix &m2)
              */
            PMatrix operator/ ( const PMatrix &m2 ) const;


            /** @brief divides this matrix by m2 and returns this object.
              * @returns this object
              * @see mul(const PMatrix &m1, const PMatrix &m2)
              * @see operator*(const PMatrix &m2)
              */
            PMatrix & operator/= ( const PMatrix &m2 );



            /** @returns the transposed of the matrix
              *
              */
            PMatrix transposed() const;


            /** @brief multiplies m1 with m2 and stores the result in this
              * object (m2.mNr has to be the same as m1.mNc, otherwise an
              * exeption is raised)
              * @param m1 matrix 1
              * @param m2 matrix 2
              * @see operator*(const PMatrix &m2)
              * @see operator*=(const PMatrix &m2)
              */
            const PMatrix & mul( const PMatrix &m1, const PMatrix &m2 );


            /** @brief multiplies m1 with m2 and stores the result in this
              * object (m2.mNr has to be the same as m1.mNc, otherwise an
              * exeption is raised)
              * @param m1 matrix 1
              * @param m2 matrix 2
              * @see operator*(const PMatrix &m2)
              * @see operator*=(const PMatrix &m2)
              */
            const PMatrix & mul( const PMatrix &m1, bool transposed1, const PMatrix &m2 ,bool transposed2);



            /** @brief element multiplication of m1 with m2 and stores the result in
              * this object (m2 and m1 shoudl have the same size, otherwise an
              * exeption is raised)
              * @param m1 matrix 1
              * @param m2 matrix 2
              */
            void mulElement( const PMatrix &m1, const PMatrix &m2 );

            /** @brief element multiplication of m1 with m2 and stores the result in
              * this object (m2 and m1 shoudl have the same size, otherwise an
              * exeption is raised)
              * @param m1 matrix 1
              * @param m2 matrix 2
              */
            PMatrix & mulElement( const PMatrix &m1 );



            /** @returns the A' * A
              */
            PMatrix ATA( void ) const;

            /** @returns the A * A'
              */
            PMatrix AAT( void ) const;


            /** @returns the A' * A
              */
            void ATA( PMatrix & ata) const;

            /** @returns the A * A'
              */
            void AAT( PMatrix & aat ) const;




            /** @brief multiplies a vector with the matrix and returns a vector.
              * condition: mNc==v2.N()
              * @param v2 the vector that is multiplied with the matrix in this
              * object
              * @returns the resulting vector
               */
            PVector operator* ( const PVector &v2 ) const;


            /** @brief returns a diagonal matrix.
              * @param n size of thediagonal matrix
              */
            static PMatrix diag( int n );

            /** @brief returns a diagonal matrix.
              * @param n size of thediagonal matrix
              */
            void makeDiag( int n );

            /** @brief returns column at index i.
              * i must be a valid index position in the matrix (0 <= i < Nc() ).
              * no range check
              * @param i index of the selected column
              * @returns the vector with the selcted column
              * @see Nc()
              * @see row(int i)
              */
            PVector col( int i ) const;


            /** @brief returns row at index i.
              * Variable i should contain a valid (row) index position in the matrix (0 <= i < Nr() ).
              * If no data is present or i is invalid an empty vector is returns
              * @param i index of the row to return the data for
              * @returns the vector with the selected row data
              * @see Nc()
              * @see col(int i)
              */
            PVector row( int i ) const;


            /** @brief returns row at index i.
              * Variable i should contain a valid (row) index position in the matrix (0 <= i < Nr() ).
              * If no data is present or i is invalid an empty vector is returns
              * @param i index of the row to return the data for
              * @param destination vector to recieve the requested data
              * @returns true if the requested data was copied
              * @see Nc()
              * @see col(int i)
              */
            bool row( int i, PVector & destination ) const;



            /** @brief TODO
              */
            PMatrix &scaleRows( const PVector & scale);



            /** @brief TODO
              */
            PVector minRows( void ) const;


            /** @brief TODO
              */
            PVector maxRows( void ) const;

            /** @brief TODO
              */
            PVector meanRows( void ) const;

            /** @brief TODO
              */
            PVector meanCols( void ) const;


            /** @brief TODO
              */
            PVector minCols( void ) const;


            /** @brief TODO
              */
            PVector maxCols( void ) const;


            void zeromeanRows();


            /** @brief set the row at index i to zero
              * @param i row indes
              */
            void setRowToZero( int i );


            /** @brief sets column at index i to the given vector.
              * i must be a valid index position in the matrix (0 <= i < Nc() ).
              * no range check
              * @param i index of the selected column
              * @param  vec the column vector
              * @see Nc()
              * @see row(int i)
              */
            void setCol( int i, const PVector &vec );


            /** @brief set the col at index i to zero
              * @param i col indes
              */
            void setColToZero( int i );


            /** @brief set all elements to zero
              */
            inline void toZero() { mMat.fill(0); }


            /** @brief set all elements to zero
              */
            inline void fill(double value) { mMat.fill(value); }


            /** @brief sets row to the given vector.
              * i must be a valid index position in the matrix (0 <= i < Nr() ).
              * no range check
              * @param i index of the selected row
              * @param  vec the row vector
              * @see Nc()
              * @see col(int i)
              */
            void setRow( int i, const PVector &vec );


            /** @brief Reads matrix data from file.
              * @param ds stream used for reading
              * @see virtual bool PFile::read(QDatastream &ds);
              */
            bool read( QDataStream &ds );


            /** @brief Write matrix data to file
              * @param ds stream used for writing.
              * @see virtual bool read(QDatastream &ds)=0;
              */
            bool write( QDataStream &ds )const;


            /** @brief Save the matrix to a text file with given precision.
              * If the precision is less than 0 the file is a binary file,
              * otherwise a textfile with precion digets is created
              * @param filename The filname used to save the matrix
              * @param precision number of digets (default set to binary file)
              * @return true if the file is saved
              * @see load(QString filename)
              */
            bool save( const QString &filename, int precision = -1 ) const;


            /** @brief load a the matrix from a binary file.
              * Format file: see code
              * @param filename  Filename of the matrix file
              * @see save(QString filename, unsigned int precision)
              */
            bool load( const QString &filename );


            /** @brief returns the number of non-zero elemenst
              */
            int nonZeros( void ) const;


            /** @brief Normalize the matrix
              * It is assumed that all values are positive, i.e. the minium is zero
              */
            void normalize( void );


            /** @brief Normalize each of the rows in the matrix
              */
            void normalizeRows( void );


            /** @brief Normalize each of the rows in the matrix
              */
            void normalizeRow( int index );



            /** @brief Function to find the 1-norm of a matrix (Forsythe page 46)
              */
            double norm1() const;


#ifdef PEACSCLASSES
            /** @brief Function to find the 2-norm of a matrix (Forsythe page 46)
              */
            double norm2() const;
#endif


            /** @brief Function to find the Inf-norm of a matrix (Forsythe page 46)
              */
            double normInf() const;



            /** @brief frobenius norm of the matrix
              */
            double frobeniusNorm() const;


        protected:

            /** @brief Returns the class name
              * @see PFile
              */
            const char * getClassName( void ) const;


            /** @brief Implementation of a pure virtual function (PFile)that
              * requests the default class data version to write to the case file
              * @see Pfile
              */
            int getDefaultVersion( void ) const;


        private:

            /** @brief The number of columns of the matrix
              */
            int	mNr;


            /** @brief The number of rows of the matrix
              */
            int	mNc;


            /** @brief Pointer to the matrix
              */
            QVector <double> mMat;
        };
    }
}

#endif //PMATRIX_H
