/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pplane.cpp 1118 2014-02-26 16:09:13Z peter $
 *
 */


#include "pplane.h"

#ifdef Q_CC_MSVC
#pragma warning( disable : 4351 )
#endif


using namespace Peacs::Support;



PPlane::PPlane( void ) :
    mPlaneVector( 0, 0, 0 ),
    mPosition( 0 ),
    mData()
{
    updateDataPointers();
}


PPlane::PPlane(const P3DVector& vect, double constSup) :
    mPlaneVector( vect.normalized() ),
    mPosition( constSup / vect.length() ),
    mData()
{
    updateDataPointers();
}


PPlane::PPlane(const P3DVector &vect,const P3DVector &supportPos) :
    mPlaneVector( vect.normalized() ),
    mPosition( 0 ),
    mData()
{
    mPosition = supportPos.dot( mPlaneVector ) * -1.0;

    updateDataPointers();
}


PPlane::PPlane( const PPlane &plane ) :
    mPlaneVector( plane.mPlaneVector ),
    mPosition( plane.mPosition ),
    mData()
{
    updateDataPointers();
}


PPlane::~PPlane( void )
{
}


void PPlane::SetPlane( const P3DVector &vect, double constSup )
{
    mPlaneVector = vect.normalized();
    mPosition    = constSup / vect.length();

    updateDataPointers();
}


void PPlane::SetPlane( const P3DVector &vect, const P3DVector &supportPos )
{
    mPlaneVector    = vect.normalized();
    mPosition       = supportPos.dot( mPlaneVector ) * -1.0;

    updateDataPointers();
}


const PPlane &PPlane::operator= ( const PPlane& plane )
{
    if ( this != &plane )
    {
        mPlaneVector = plane.mPlaneVector;
        mPosition    = plane.mPosition;

        updateDataPointers();
    }

    return *this;
}


const PPlane &PPlane::operator-= ( double deltaPos )
{
    mPosition -= deltaPos;

    updateDataPointers();

    return *this;
}


const PPlane &PPlane::operator+= ( double deltaPos )
{
    mPosition += deltaPos;

    updateDataPointers();

    return *this;
}


bool PPlane::operator== ( const PPlane& plane ) const
{
    return ( mPlaneVector == plane.mPlaneVector &&
             mPosition == plane.mPosition );
}


bool PPlane::operator!= ( const PPlane& plane ) const
{
    return ( mPlaneVector != plane.mPlaneVector &&
            mPosition != plane.mPosition );
}


bool PPlane::isValid( void ) const
{
    bool valid = true;

    if ( mPlaneVector.x() == 0.0 &&
         mPlaneVector.y() == 0.0 &&
         mPlaneVector.z() == 0.0 )
    {
        valid = false;
    }

    return valid;
}


void PPlane::setPos( double pos )
{
    mPosition = pos;

    updateDataPointers();
}


void PPlane::setPos( const P3DVector &pos )
{
    mPosition = mPlaneVector.dot(pos) * -1.0;

    updateDataPointers();
}


void PPlane::setVector( const P3DVector& vect )
{
    mPosition /= vect.length();

    mPlaneVector = vect.normalized();

    updateDataPointers();
}


void PPlane::invertVector( void )
{
    mPosition       *= -1.0;
    mPlaneVector    *= -1.0;

    updateDataPointers();
}


void PPlane::updateDataPointers( void )
{
    mData[0] = mPlaneVector.x();
    mData[1] = mPlaneVector.y();
    mData[2] = mPlaneVector.z();
    mData[3] = mPosition;
}


P3DVector PPlane::projection( const P3DVector & vect ) const
{
     return vect -
            mPlaneVector * ( vect.dot( mPlaneVector ) +
                             mPosition );
}


P3DVertices PPlane::projection( const P3DVertices & ver ) const
{
    P3DVertices projected( ver );

    for ( int i = 0; i < projected.size(); i++ )
    {
        projected[i] = ver.at( i ) - ( mPlaneVector * ( ver.at( i ).dot( mPlaneVector ) + mPosition ) );
    }

    return projected;
}

QVector <PPlane::PlaneSide_type> PPlane::planeSide(const P3DVertices & ver) const
{
    QVector <PlaneSide_type> retVect(ver.N(),ONPLANE);
    for ( int i = 0; i < ver.size(); i++ )
    {
        double val = (mData[0] * ver.at(i).x()) + mData[1] * ver.at(i).y() + mData[2] * ver.at(i).z() + mData[3];
        if ( val > 0 )
        {
            retVect[i] = NORMALSIDE;
        }
        else if ( val > 0 )
        {
            retVect[i] = OTHERSIDE;
        }
    }
    return retVect;
}


bool PPlane::lineCross( const P3DVector & l1, const P3DVector & l2, P3DVector &crossPoint ) const
{

    P3DVector dl = l2 - l1;

    double denom = mPlaneVector.dot(dl);

    if ( fabs(denom) > 1e-10 )
    {
        double lambda = (-mPosition - mPlaneVector.dot(l1) ) / denom;
        crossPoint = l1 + dl * lambda;
        return true;
    }
    else
    {
        return false;
    }

    return false;
}
