/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pplane.h 1118 2014-02-26 16:09:13Z peter $
 *
 */


#ifndef _PPLANE_H
#define _PPLANE_H

#include "P3DVector.h"
#include "P3DVertices.h"

#include "PeacsSupport_global.h"

#include <QVector>

namespace Peacs
{
    namespace Support
    {
        /** @class PPlane pplane.h "support/pplane.h"
          *	brief This class describes a plane by a normal vector and a constant (the position)
          * note: that the plane vector (@see mPlaneVector) is normalized
          * @see PGeometry
          * @see P3DVector
          */
        class PEACSSUPPORTSHARED_EXPORT PPlane
        {
        public:
            enum PlaneSide_type { ONPLANE, NORMALSIDE, OTHERSIDE};


            /** @brief default constructor, creates an invalid plane description
              */
            PPlane( void );


            /** @brief Contructor with the support vector and position
              * @param vect (normal) vector describing the plane
              * @param constSup the constant to put the plane in the right position
              */
            PPlane( const P3DVector &vect, double constSup );


            /** @brief Contructor with the support vector and positionvector
              * @param vect (normal) vector describing the plane
              *	@param supportPos the postion used to put the plane in the right position
              */
            PPlane( const P3DVector &vect, const P3DVector &supportPos );


            /** @brief Copy constructor
              * @param plane is copied into this object
              */
            PPlane( const PPlane &plane );


            /** @brief Destructor
              */
            virtual ~PPlane( void );


            /** @brief Initializes the class members with the support vector and position
              * @param vect (normal) vector describing the plane
              *	@param constSup the constant to put the plane in the right position
              */
            void SetPlane( const P3DVector &vect, double constSup );


            /** @brief Initializes the class members with the support vector and positionvector
              * @param vect (normal) vector describing the plane
              *	@param supportPos the postion used to put the plane in the right position
              */
            void SetPlane( const P3DVector &vect, const P3DVector &supportPos );


            /** @brief Asignment operator.
              * plane is copied into this object and this object is returned
              * @param plane is copied into this object
              * @returns a refrence to this plane object
              */
            const PPlane &operator= ( const PPlane& plane );


            /** @brief adapts the postion of the plane by deltaPos
              * @param deltaPos the adaptation of the postion of this plane
              * @returns a refrence to this plane object
              */
            const PPlane &operator-= ( double deltaPos );


            /** @brief adapts the postion of the plane by deltaPos
              *  @param deltaPos the adaptation of the postion of this plane
              *  @returns a refrence to this plane object
              */
            const PPlane &operator+= ( double deltaPos );


            /** @brief equal operator.
              *	true if the vector and constant are the same
              * @param plane that is compered to
              * @returns true or false
              */
            bool operator== ( const PPlane& plane ) const;


            /** @brief unequal operator.
              * false if one of the values of the vector ore constant are the NOT the same
              * @param plane that is compered to
              * @returns true or false
              */
            bool operator!= ( const PPlane& plane ) const;


            /** @brief sets the postion of the plane
              * @param pos the new postion of the plane
              */
            void setPos( double pos );


            /** @brief sets the postion of the plane
              * @param pos the new postion of the plane
              */
            void setPos( const P3DVector &pos );


            /** @brief sets the the normal vector decsribing the plane
              * @param vect th enew support vector
              */
            void setVector( const P3DVector &vect );


            /** @brief invert the normal of the plane
              */
            void invertVector( void );


            /** @returns  the normal vector decsribing the plane  (for convinience
              */
            inline const P3DVector &vector( void ) const { return mPlaneVector; }


            /** @returns the position of the plane
              */
            inline double pos( void ) const { return mPosition; }


            /** @brief A plane is only valid if the support vector is nonzero
              *	@returns is true if the support is nonzero
              */
            bool isValid( void ) const;


            /** @returns a pointer to the data for Open GL support
              */
            inline const double *data( void ) const { return mData; }


            /** @returns the projection of @param vect on the plane
              */
            P3DVector projection( const P3DVector & vect ) const;


            /** @brief Makes a projection of vect on the plane
              * @param vect used to project with
              * @returns the projection of vect on the plane
              */
            P3DVertices projection( const P3DVertices& vers ) const;


            /** @returns the crossection point of the line defined by l1 and l2
              */
            bool lineCross( const P3DVector & l1, const P3DVector & l2, P3DVector &crossPoint ) const;


            QVector <PlaneSide_type> planeSide(const P3DVertices & ver) const;


        private:

            /** @brief update the data pointers for OPEN GL support
              */
            void updateDataPointers( void );


            /** @brief the normalized plane vector
              */
            P3DVector mPlaneVector;


            /** @brief the postion of the plane
              */
            double mPosition;


            /** @brief a pointer to the data for Open GL support
              */
            double mData[4];
        };
    }
}

#endif // _PPLANE_H
