/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pvector.cpp 1118 2014-02-26 16:09:13Z peter $
 *
 */


#include <QTextStream>
#include <QFile>
#include <QDebug>

#include <float.h>

#include "pvector.h"
#include "pmatrix.h"
#include "PIVector.h"


using namespace Peacs::Support;


const int FILE_VECTOR_VERSION_1         = 1;
const int FILE_VECTOR_VERSION_DEFAULT   = 1;
const char* FILE_VECTOR_NAME            = "PVector";


PVector::PVector( void ) :
    PFile(),
    mVec()
{
}


PVector::PVector( int n, double val ) :
    PFile(),
    mVec( n, val )
{
}


PVector::PVector( int n ) :
    PFile(),
    mVec( n, 0 )
{
}


PVector::PVector( const PVector & v ) :
    PFile( v ),
    mVec( v.mVec )
{
}


PVector::PVector( const QVector<double> & v ) :
    PFile(),
    mVec( v )
{
}


PVector::PVector( const PIVector & vec ) :
    PFile(),
    mVec( vec.N() )
{
    double * dataPtr = mVec.data();

    for ( int i = 0; i < vec.N(); i++ )
    {
        dataPtr[i] = (double)vec[i];
    }
}


int PVector::getDefaultVersion( void ) const
{
    return FILE_VECTOR_VERSION_DEFAULT;
}


const char *PVector::getClassName( void ) const
{
    return FILE_VECTOR_NAME;
}


bool PVector::operator== ( const PVector & vec ) const
{
    bool equal = false;

    if ( mVec.size() == vec.mVec.size() )
    {
        int             i           = 0;
        const double *  data1Ptr    = mVec.data();
        const double *  data2Ptr    = vec.mVec.data();

        equal = true;   // Assume equal

        while ( i < mVec.size() &&
                equal )
        {
            equal = ( data1Ptr[i] == data2Ptr[i] );

            i++;
        }
    }

    return equal;
}


bool PVector::operator!= ( const PVector & vec ) const
{
    bool unequal = false;

    if ( !mVec.size() != vec.mVec.size() )
    {
        int             i           = 0;
        const double *  data1Ptr    = mVec.data();
        const double *  data2Ptr    = vec.mVec.data();

        while ( (i < mVec.size()) && !unequal )
        {
            unequal = ( data1Ptr[i] != data2Ptr[i] );

            i++;
        }
    }

    return unequal;
}


PVector PVector::operator+ ( double f ) const
{
    PVector     v( *this );
    double *    data1Ptr = v.mVec.data();

    for ( int i = 0 ; i < v.mVec.size(); i++ )
    {
        data1Ptr[i] += f;
    }

    return v;
}


PVector PVector::operator- ( double f ) const
{
    PVector     v( *this );
    double *    data1Ptr = v.mVec.data();

    for ( int i = 0; i < v.mVec.size(); i++ )
    {
        data1Ptr[i] -= f;
    }

    return v;
}


PVector PVector::operator* ( double f ) const
{
    PVector     v( *this );
    double *    data1Ptr = v.mVec.data();

    for ( int i = 0 ; i < v.mVec.size(); i++ )
    {
        data1Ptr[i] *= f;
    }

    return v;
}


PVector PVector::operator/ ( double f ) const
{
    PVector     v( *this );
    double *    data1Ptr = v.mVec.data();

    for ( int i = 0 ; i < v.mVec.size(); i++ )
    {
        data1Ptr[i] /= f;
    }

    return v;
}


const PVector & PVector::operator+= ( double f )
{
    double * data1Ptr = mVec.data();

    for ( int i = 0 ; i < mVec.size(); i++ )
    {
        data1Ptr[i] += f;
    }

    return *this;
}


const PVector & PVector::operator-= ( double f )
{
    double * data1Ptr = mVec.data();

    for ( int i = 0 ; i < mVec.size(); i++ )
    {
        data1Ptr[i] -= f;
    }

    return *this;
}


const PVector & PVector::operator*= ( double f )
{
    double * data1Ptr = mVec.data();

    for ( int i = 0 ; i < mVec.size(); i++ )
    {
        data1Ptr[i] *= f;
    }

    return *this;
}


const PVector & PVector::operator/= ( double f )
{
    double * data1Ptr = mVec.data();

    for ( int i = 0 ; i < mVec.size(); i++ )
    {
        data1Ptr[i] /= f;
    }

    return *this;
}


const PVector & PVector::operator= ( double f )
{
    // Don't call base class asignment operator. Version info must always be
    // the last version
    mVec.fill( f );

    return *this;
}


PVector PVector::operator+ ( const PVector & v2 ) const
{
    PVector v( *this );

    if ( !v2.isEmpty() )
    {
        if ( !isEmpty() )
        {
            if ( v2.mVec.size() != v.mVec.size() )
            {
                qDebug("PVector operator +: vectors are not the same size");
            }
            else
            {
                double *        data1Ptr = v.mVec.data();
                const double *  data2Ptr = v2.mVec.data();

                for ( int i = 0 ; i < v.mVec.size(); i++ )
                {
                    data1Ptr[i] += data2Ptr[i];
                }
            }
        }
        else
        {
            v = v2;
        }
    }

    return v;
}


PVector PVector::operator- ( const PVector & v2 ) const
{
    PVector v( *this );

    if ( !v2.isEmpty() )
    {
        if ( !v.isEmpty()  )
        {
            if (v2.mVec.size() != v.mVec.size() )
            {
                qDebug("PVector operator -: vectors are not the same size");
            }
            else
            {
                double *        data1Ptr = v.mVec.data();
                const double *  data2Ptr = v2.mVec.data();

                for ( int i = 0 ; i < v.mVec.size(); i++ )
                {
                    data1Ptr[i] -= data2Ptr[i];
                }
            }
        }
        else
        {
            v -= v2;
        }
    }

    return v;
}


PVector PVector::operator* ( const PVector & v2 ) const
{
    PVector v( *this );

    if ( v2.isEmpty() || isEmpty() )
    {
        v.mVec.fill( 0, qMax( v.mVec.size(), v2.mVec.size() ) );
    }
    else
    {
        if ( v2.mVec.size() != v.mVec.size() )
        {
            qDebug("PVector operator *: vectors are not the same size");
        }
        else
        {
            double *        data1Ptr = v.mVec.data();
            const double *  data2Ptr = v2.mVec.data();

            for ( int i = 0; i < v.mVec.size(); i++ )
            {
                data1Ptr[i] *= data2Ptr[i];
            }
        }
    }

    return v;
}


PVector PVector::operator/ ( const PVector &v2 ) const
{
    PVector v( *this );

    if ( v2.isEmpty() )
    {
        qDebug("PVector operator /: illegal to devide by 0  (v2 is empty)");
    }
    else if ( v.isEmpty() )
    {
        v.mVec.fill( 0, v2.mVec.size() );
    }
    else if ( v2.mVec.size() != v.mVec.size() )
    {
        qDebug("PVector operator /: vectors are not the same size");
    }
    else
    {
        double *        data1Ptr = v.mVec.data();
        const double *  data2Ptr = v2.mVec.data();

        for ( int i = 0; i < mVec.size(); i++ )
        {
            data1Ptr[i] /= data2Ptr[i];
        }
    }

    return v;
}


const PVector & PVector::operator+= ( const PVector & v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            *this = v;
        }
        else
        {
            if ( v.mVec.size() != mVec.size() )
            {
                qDebug("PVector operator +=: vectors are not the same size");
            }
            else
            {
                double *        data1Ptr = mVec.data();
                const double *  data2Ptr = v.mVec.data();

                for ( int i = 0; i < mVec.size(); i++ )
                {
                    data1Ptr[i] += data2Ptr[i];
                }
            }
        }
    }

    return *this;
}


const PVector & PVector::operator-= ( const PVector & v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            resize( v.mVec.size() );

            for ( int i = 0; i < mVec.size(); i++ )
            {
                mVec[i] = -v.mVec[i];
            }

        }
        else if ( v.mVec.size() != mVec.size() )
        {
            qDebug("PVector operator -=: vectors are not the same size");
        }
        else
        {
            double *        data1Ptr = mVec.data();
            const double *  data2Ptr = v.mVec.data();

            for ( int i = 0; i < mVec.size(); i++ )
            {
                data1Ptr[i] -= data2Ptr[i];
            }
        }
    }

    return *this;
}


const PVector & PVector::operator*= ( const PVector & v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            *this = v;
        }
        else
        {
            if ( v.mVec.size() != mVec.size() )
            {
                qDebug("PVector operator *=: vectors are not the same size");
            }
            else
            {
                double *        data1Ptr = mVec.data();
                const double *  data2Ptr = v.mVec.data();

                for ( int i = 0; i < mVec.size(); i++ )
                {
                    data1Ptr[i] *= data2Ptr[i];
                }
            }
        }
    }

    return *this;
}


const PVector & PVector::operator/= ( const PVector & v )
{
    if ( !v.isEmpty() )
    {
        if ( isEmpty() )
        {
            mVec.fill( 0, v.mVec.size() );
        }
        else
        {
            if ( v.mVec.size() != mVec.size() )
            {
                qDebug("PVector operator /=: vectors are not the same size");
            }
            else
            {
                double *        data1Ptr = mVec.data();
                const double *  data2Ptr = v.mVec.data();

                for ( int i = 0; i < mVec.size(); i++ )
                {
                    data1Ptr[i] /= data2Ptr[i];
                }
            }
        }
    }
    else
    {
        qDebug("PVector operator /=: illegal to devide by 0 (v is empty)");
    }

    return *this;
}


const PVector & PVector::operator*= ( const PMatrix & m )
{
    if ( m.Nr() == mVec.size() )
    {
        int     ir;
        int     k;
//        double   val;
//        PVector vec( *this );
        PVector vec( m.Nc(), 0.0 );

//        resize( m.Nc() );

        const double *  dataPtr     = mVec.data();
        double *        dataTmpPtr  = vec.mVec.data();

        for ( int i = 0; i < m.Nc(); i++ )
        {
//            val = 0;

            for ( k = 0, ir = i; k < m.Nr(); k++, ir += m.Nc()  )
            {
//                val += vec[k] * m[ir];
                dataTmpPtr[i] += dataPtr[k] * m[ir];
            }

//            mVec[i] = val;
        }

        mVec = vec.mVec;
    }
    else
    {
        qDebug("PVector operator *=: vectors are not the same size");
    }

    return *this;
}


PVector PVector::operator* ( const PMatrix & m ) const
{
    PVector v( m.Nc(), 0.0 );

    if ( m.Nr() == mVec.size() )
    {
        int             k;
        int             ir;
        const double *  dataPtr     = mVec.data();
        double *        dataTmpPtr  = v.mVec.data();

        for ( int i = 0; i < m.Nc(); i++ )
        {
            for ( k = 0, ir = i; k < m.Nr(); k++, ir += m.Nc()  )
            {
                dataTmpPtr[i] += dataPtr[k] * m[ir];
            }
        }
    }
    else
    {
        qDebug("PVector operator*: vector and rows matrix are not the same size");
    }

    return v;
}


const PVector & PVector::operator= ( const PVector & v )
{
    if ( this != &v )
    {
        mVec = v.mVec;
    }

    return *this;
}


PIVector PVector::operator> ( double val ) const
{
    PIVector        ret( mVec.size(), 0.0 );
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( dataPtr[i] > val )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PVector::operator< ( double val ) const
{
    PIVector        ret( mVec.size(), 0.0 );
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( dataPtr[i] < val )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PVector::operator == ( double val ) const
{
    PIVector        ret( mVec.size(), 0 );
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( dataPtr[i] == val )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PVector::operator>= ( double val ) const
{
    PIVector        ret( mVec.size(), 0 );
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( dataPtr[i] >= val )
        {
            ret[i] = 1;
        }
    }

    return ret;
}


PIVector PVector::operator<= ( double val ) const
{
    PIVector        ret( mVec.size(), 0 );
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( dataPtr[i] <= val )
        {
            ret[i] = 1;
        }
    }

    return ret;
}

PIVector PVector::operator!= ( double val ) const
{
    PIVector        ret( mVec.size(), 0 );
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        if ( dataPtr[i] != val )
        {
            ret[i] = 1;
        }
    }

    return ret;

}

double PVector::sum( void ) const
{
    double          ret = 0;
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        ret += dataPtr[i];
    }

    return ret;
}


PVector PVector::getSqrt( void ) const
{
    PVector         ret( mVec.size(), 0.0 );
    const double *  dataPtr = mVec.data();
    double *        dataOutPtr = ret.mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        dataOutPtr[i] = (double)sqrt( dataPtr[i] );
    }

    return ret;
}


const PVector & PVector::mul( const PMatrix & m, const PVector & v2 )
{
    if ( m.Nc() == v2.mVec.size() )
    {
        double val;

        mVec.fill( 0.0, m.Nr() );

        double *        dataPtr     = mVec.data();
        const double *  data2Ptr    = v2.mVec.data();
        const double *  mPtr        = m.data();

        if ( !mVec.isEmpty() )
        {
            for ( int i = 0; i < m.Nr(); i++ )
            {
                val = 0;

                for ( int k = 0; k < m.Nc(); k++ )
                {
                    dataPtr[i] += mPtr[ m.dataOffset( i, k ) ] * data2Ptr[k];
                }
            }
        }
    }
    else
    {
        qDebug("PVector mul(matrix,vector): not the same size");
    }

    return *this;
}


const PVector & PVector::mul( const PVector & v2, const PMatrix & m )
{
    if ( m.Nr() == v2.mVec.size() )
    {
        double val;

        mVec.fill( 0.0, m.Nr() );

        double *        dataPtr     = mVec.data();
        const double *  data2Ptr    = v2.mVec.data();
        const double *  mPtr        = m.data();

        for ( int i = 0; i < m.Nc(); i++ )
        {
            val = 0;

            for ( int k = 0; k < m.Nr(); k++ )
            {
                dataPtr[i] += mPtr[ m.dataOffset( k, i ) ] * data2Ptr[k];
            }
        }
    }
    else
    {
        qDebug("PVector mul(vector,matrix): not the same size");
    }

    return *this;
}


void PVector::normalize( void )
{
    if ( !mVec.isEmpty() )
    {
        double *    dataPtr = mVec.data();
        double      min     = dataPtr[0];
        double      max     = dataPtr[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( dataPtr[i] < min )
            {
                min = dataPtr[i];
            }

            if ( max < dataPtr[i] )
            {
                max = dataPtr[i];
            }
        }

        if ( max - min != 0 )
        {
            for ( int i = 0; i < mVec.size(); i++ )
            {
                dataPtr[i] = (dataPtr[i] - min) / (max - min);
            }
        }
    }
}


PVector PVector::normalized( void ) const
{
    PVector v( mVec.size(), 0.0 );

    if ( !isEmpty() )
    {
        const double *  dataPtr = mVec.data();
        double          min = dataPtr[0];
        double          max = dataPtr[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( dataPtr[i] < min )
            {
                min = dataPtr[i];
            }

            if ( max < dataPtr[i] )
            {
                max = dataPtr[i];
            }
        }

        if ( max - min != 0 )
        {
            double * dataOutPtr = v.mVec.data();

            for ( int i = 0; i < v.mVec.size(); i++ )
            {
                dataOutPtr[i] = (dataPtr[i] - min) / (max - min);
            }
        }
    }

    return v;
}


double PVector::norm1() const
{
    double          mnorm   = 0;
    const double *  dataPtr = mVec.data();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        mnorm += fabs( dataPtr[i] );
    }

    return mnorm;
}


double PVector::norm2() const
{
    double          mnorm   = 0;
    const double *  dataPtr = mVec.constData();

    for ( int i = 0; i < mVec.size(); i++ )
    {
        mnorm += ( dataPtr[i] * dataPtr[i] );
    }

    return sqrt(mnorm);
}


double PVector::getMin( void ) const
{
    double min = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mVec.constData();

        min = dataPtr[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( dataPtr[i] < min )
            {
                min = dataPtr[i];
            }
        }
    }

    return min;
}


double PVector::getMax( void ) const
{
    double max = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mVec.constData();

        max = dataPtr[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( dataPtr[i] > max )
            {
                max = dataPtr[i];
            }
        }
    }

    return max;
}


int PVector::getMinI( void ) const
{
    int minI = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mVec.constData();

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( dataPtr[i] < dataPtr[minI] )
            {
                minI = i;
            }
        }
    }

    return minI;
}


int PVector::getMaxI( void ) const
{
    int maxI = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mVec.constData();

        for ( int i = 1; i < mVec.size(); i++ )
        {
            if ( dataPtr[i] > dataPtr[maxI] )
            {
                maxI = i;
            }
        }
    }

    return maxI;
}


double PVector::getMean( void ) const
{
    double mean = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mVec.constData();

        for ( int i = 0; i < mVec.size(); i++ )
        {
            mean += dataPtr[i];
        }

        mean /= (double)mVec.size();
    }

    return mean;
}


double PVector::getStd( void ) const
{
    double stdev = 0;

    if ( !isEmpty() )
    {
        int             i;
        double          mean    = 0;
        double          sqSum   = 0;
        double          tmp;
        const double *  dataPtr = mVec.constData();

        for ( i = 0; i < mVec.size(); i++ )
        {
            mean += dataPtr[i];
        }

        mean /= (double)mVec.size();

        for ( i = 0; i < mVec.size(); i++ )
        {
            tmp     =  (dataPtr[i] - mean);
            sqSum   += (tmp * tmp);
        }

        stdev = (double)sqrt( sqSum / ( (double)mVec.size() ) );
    }

    return stdev;
}


double PVector::getRange( void ) const
{
    double min    = 0;
    double max    = 0;

    if ( !isEmpty() )
    {
        const double * dataPtr = mVec.constData();

        min    = dataPtr[0];
        max    = dataPtr[0];

        for ( int i = 1; i < mVec.size(); i++ )
        {
            min = qMin( min, dataPtr[i] );
            max = qMax( max, dataPtr[i] );
        }
    }

    return max - min;
}


void PVector::sort( bool asscending )
{
    if ( asscending )
    {
        qSort( mVec.begin(), mVec.end() );
    }
    else
    {
        qSort( mVec.begin(), mVec.end(), qGreater<double>() );
    }
}



PVector PVector::select(int start, int last)
{
    Q_ASSERT(start < last );
    Q_ASSERT(start >= 0 && start < mVec.size() -1 );
    Q_ASSERT(last  >= 0 && last  < mVec.size() -1 );

    const double * source = mVec.constData();
    PVector vec( last - start + 1 );

    memcpy( (void *)vec.mVec.data(),
            (const void *) & (source[ start ]),
            size_t( vec.mVec.size()  * sizeof(double) ) );

    return vec;
}


PVector PVector::select( const PIVector &selection ) const
{
    const double * source = mVec.constData();
    PVector vec( selection.N() );
    double * sink = vec.mVec.data();
    for ( int i = 0; i < selection.N(); i++)
    {
        sink[i] = source[selection[i]];
    }
    return vec;
}




bool PVector::save( QString filename, unsigned int precision ) const
{
    int     ir;
    bool    ok = false;
    QFile   f( filename );

    if ( f.open( QIODevice::WriteOnly ) )
    {
        QTextStream ts( &f );

        ts.setRealNumberPrecision( precision );

        ts << mVec.size()  << " 1  \n";

        for ( ir = 0; ir < mVec.size() && ts.status() == QTextStream::Ok; ir++ )
        {
            ts << mVec[ir] << "\n";
        }

        ts << "\n";

        f.close();

        ok = ts.status() == QTextStream::Ok;
    }

    return ok;
}


bool PVector::load( QString filename )
{
    bool ok = false;

    if ( !filename.isEmpty() )
    {
        QFile f( filename );

        if ( f.open( QIODevice::ReadOnly ) )
        {
            int         n;
            int         nc;
            QTextStream ts( &f );

            ts >> n >> nc;

            if ( ts.status() == QTextStream::Ok &&  n > 0  && nc == 1 )
            {
                resize( n );

                if ( !isEmpty() )
                {
                    ok = true;

                    for ( int i = 0; i < n && ok; i++ )
                    {
                        ts >> mVec[i];

                        ok = ts.status() == QTextStream::Ok;
                    }
                }
            }
        }

        if ( !ok )
        {
            clear();
        }

        f.close();
    }

    return ok;
}


bool PVector::read( QDataStream &dataStream )
{
    bool    ok      = false;
    int     version = readHead( dataStream, PVector::getClassName() );

    if ( version != READ_HEADER_FAILED )
    {
        qint32 n;

        switch ( version )
        {
        case FILE_VECTOR_VERSION_1:
            clear();

            dataStream >> n;

            if ( isOk( dataStream ) )
            {
                if ( n > 0 )
                {
                    resize( (int)n );

                    if ( !isEmpty() )
                    {
                        ok = true;

                        for ( int i = 0; i < mVec.size() && ok; i++ )
                        {
                            dataStream >> mVec[i];

                            ok = isOk( dataStream );
                        }
                    }
                }
                else
                {
                    ok = true; // Nothing more to read
                }
            }
            break;

        default:
            break;
        }
    }

    return ok;
}


bool PVector::write( QDataStream &dataStream ) const
{
    bool ok = writeHead( dataStream, PVector::getClassName(), PVector::getDefaultVersion() );

    if ( ok )
    {
        switch ( PVector::getDefaultVersion() )
        {
        case FILE_VECTOR_VERSION_1:
            {
                qint32 n = (qint32)mVec.size();

                dataStream << n;

                if ( isOk( ok, dataStream ) && !isEmpty() )
                {
                    for ( int i = 0; i < mVec.size() && isOk( dataStream ); i++ )
                    {
                        dataStream << mVec[i];
                    }
                }
                break;
            }

        default:
            ok = false;
            break;
        }
    }

    return isOk( ok, dataStream );
}


#ifndef QT_NO_DEBUG

double & PVector::operator[]( int i )
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure
    Q_ASSERT( i >= 0 && i < mVec.size() );  // Test just to be sure

    return mVec[i];
}


const double & PVector::operator[] ( int i ) const
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure
    Q_ASSERT( i >= 0 && i < mVec.size() );  // Test just to be sure


    return mVec[i];
}


const double & PVector::at( int i ) const
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure
    Q_ASSERT( i >= 0 && i < mVec.size() );  // Test just to be sure

    return mVec[i];
}


const double & PVector::first() const
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure

    return mVec.first();
}


const double & PVector::last() const
{
    Q_ASSERT( !mVec.isEmpty() );            // Test just to be sure
    return mVec.last();
}

#endif
