/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: pvector.h 1112 2014-02-19 16:23:58Z peter $
 *
 */

#ifndef __PVECTOR_H
#define __PVECTOR_H

#include <QVector>

#include "pfile.h"
#include "math.h"

#include "PeacsSupport_global.h"

namespace Peacs
{
    namespace Support
    {
        class PMatrix;

        class PIVector;

        /** @class PVector pvector.h "support/pvector.h"
          * @brief This class hold a vector (1 demensional array)
          * and its functions
          */
        class PEACSSUPPORTSHARED_EXPORT PVector : public PFile
        {

        public:

            /** @brief Default constructor.
              *	Create a empty vector
              */
            PVector( void );


            /** @brief Constructor.
              * copy constructor
              * @see operator=(const PVector &vector)
              */
            PVector( const PVector & vector );


            /** @brief Constructor
              * Initializes the vector (size n) with val)
              * @param n  size of the vector
              * @param val the value assigned to all elements
              */
            PVector( int n, double val );


            /** @brief Initializer constructor
              * Creates a new vector with size n and initializes all elements to zero
              * @param n  size of the vector
              */
            PVector( int n );


            /** @brief Initializer constructor
              *
              * @param vec The integer vector to initialize with
              */
            PVector( const PIVector & vec );


            /** @brief Initializer constructor
              *
              * @param vec The double vector to initialize with
              */
            PVector( const QVector<double> & v );


            /** @brief destructor.
              */
            virtual ~PVector(void) {}


            /** @brief Returns the minimum value of all elements in the vector
              * This one is slower because it computes the minimum value every time
              */
            double getMin( void ) const;


            /** @brief The index posistion of the minimum in the vector
              */
            int getMinI( void ) const;


            /** @brief Returns the maximum value of all elements in the
              * vector.    This one is slower because it computes the maximum value every time
              * @see max()
              */
            double getMax( void ) const;


            /** @brief The index posistion of the maximum in the vector
              */
            int getMaxI( void ) const;


            /** @brief The mean of the vector
              */
            double getMean( void ) const;


            /** @brief The standard deviation of the vector
              */
            double getStd( void ) const;


            /** @brief The diffrence between max and min of the vector
              */
            double getRange( void ) const;


            /** @brief Sorts the items in the vector
              * @param asscending (default true)
              */
            void sort( bool asscending = true );


            /** @brief Returns the number of elements of the vector.
              */
            inline int N( void ) const { return mVec.size(); }


            /** @brief Returns the number of elements of the vector.
              */
            inline int count( void ) const { return mVec.size(); }


            /** @brief Returns the number of elements of the vector.
              */
            inline int size( void ) const { return mVec.size(); }


            /** @brief Returns true if the vector is empty.
              */
            inline bool isEmpty( void ) const { return mVec.isEmpty(); }


            /** @brief Sets the size of the vector to the size mn
              * If size is greater than the current size, elements are added to the end; the new elements are initialized
              * with a default-constructed value. If size is less than the current size, elements are removed from the end.
              *
              * @param mn The new size of the vector
              */
            inline void resize( int mn ) { mVec.resize( mn ); }


            /** @brief Removes the value at index.
              * @param index the index of the element to remove
              */
            inline void remove( int index ) { mVec.remove( index ); }


            /** @brief Removes count elements from within of the vector, starting at index position pos.
              * @param pos The starting position in the vector from where to remove the elements
              * @param count The number of elements to remove
              */
            inline void remove( int pos, int count ) { mVec.remove( pos, count ); }


            /** @brief Clears the vector (empty)
              */
            inline void clear( void ) { mVec.clear(); }


            /** @brief set all values to zero
              */
            inline void toZero( void ) { mVec.fill( 0 ); }


            /** @brief returns the data pointer
              */
            inline const double * data( void ) const { return mVec.data(); }


            /** @brief returns the data pointer
              */
            inline double * data( void ) { return mVec.data(); }


            /** @brief Returns a QList object with the data contained in this QVector
              */
            inline QList <double> toList() const { return mVec.toList(); }



            /** @brief comparisson operator.
              * compares all elements from vector (vec) with this object
              *
              * @param vec the vetor being compared to
              *
              * @returns true if the data is the same
              */
            bool operator== ( const PVector &vec ) const;


            /** @brief comparisson operator.
              * compares all elements from vector (vec) with this object
              *
              * @param vec the vetor being compared to
              * @returns true if the data is not the same
              */
            bool operator!= ( const PVector &vec ) const;


#ifndef QT_NO_DEBUG

            /** @brief Returns the item at index position i as a modifiable
              * reference. i must be a valid index position in the vecor
              * (i.e. no range check)
              * @param i index postion
              * @see double &operator[](int i) const;
              */
            double & operator[] ( int i );


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see double &operator[](int i);
              */
            const double & operator[] ( int i ) const;


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see double &operator[](int i);
              */
            const double & at( int i ) const;

            /** @brief This is an overloaded member function, provided for
              * convenience. returns the first item in the list
              */
            const double & first() const;


            /** @brief This is an overloaded member function, provided for
              * convenience. returns the last item in the list
              */
            const double & last() const;

#else

            /** @brief Returns the item at index position i as a modifiable
              * reference. i must be a valid index position in the vecor
              * (i.e. no range check)
              * @param i index postion
              * @see double &operator[](int i) const;
              */
            inline double & operator[] ( int i ) { return mVec[i]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see double &operator[](int i);
              */
            inline const double & operator[] ( int i ) const { return mVec[i]; }


            /** @brief This is an overloaded member function, provided for
              * convenience.
              * @param i index postion
              * @see double &operator[](int i);
              */
            inline const double & at( int i ) const { return mVec[i]; }

            /** @brief This is an overloaded member function, provided for
              * convenience. returns the first item in the list
              */
            inline const double & first() const { return mVec.first(); }


            /** @brief This is an overloaded member function, provided for
              * convenience. returns the last item in the list
              */
            inline const double & last() const { return mVec.last(); }

#endif

            /** @brief Copies this vector and adds a constant (f) to all
              * elements of the vector.
              * @param f factor added
              * @see operator+=( double f )
              */
            PVector operator+ ( double f ) const;


            /** @brief Copies this vector and subracts a constant (f) from all
              * elements of this new vector.
              * @param f factor subtracted
              * @see operator-=( double f )
              */
            PVector operator- ( double f ) const;


            /** @brief Copies this vector and multiplies all elements with a
              * constant (f) of this new vector.
              * @param f multiplying factor
              * @see operator*=( double f)
              */
            PVector operator* ( double f ) const;


            /** @brief Copies this vector and divides all elements by a
              * constant (f) of this new vector.
              * @param f deviding factor
              * @see operator/=(double f)
              */
            PVector operator/ ( double f ) const;


            /** @brief Subtracts to all elements of the vector a constant (f)
              * and returns the vector.
              * @param f factor subtracted
              * @see operator+(double f)
              */
            const PVector & operator-= ( double f );


            /** @brief  Adds to all elements of the vector a constant (f) and
              * retruns the vector.
              * @param f factor added
              * @see operator+(double f)
              */
            const PVector & operator+= ( double f );


            /** @brief  Multiplies all elements by a constant (f) of the
              * vector and returns the vector.
              * @param f multiplying factor
              * @see operator*(double f)
              */
            const PVector & operator*= ( double f );


            /** @brief  Divides all elements by a constant (f) of the vector
              * and returns the vector.
              * @param f deviding factor
              * @see operator/(double f)
              */
            const PVector & operator/= ( double f );


            /** @brief Copy operator.
              * All elements of the vector are set to value f
              * @param f value of all elements
              */
            const PVector & operator= ( double f );


            /** @brief Copies vector, adds element wise both vectors and returns the new vector.
              * if both vectors differ in size an exception is raiseds
              * @see operator+=(const PVector &v)
              */
            PVector operator+ ( const PVector & v ) const;


            /** @brief Copies vector, subtracts element wise both vectors and
              * returns the new vector.
              * if both vectors differ in size an exception is raised
              * @param v  input vector
              * @see operator-=(const PVector &v)
              */
            PVector operator- ( const PVector &v ) const;


            /** @brief Copies vector, multiplies element wise both vectors and
              * returns the new vector.
              * if both vectors differ in size an exception is raised
              * @param v  input vector
              * @see operator*=(const PVector &v)
              */
            PVector operator* ( const PVector &v ) const;


            /** @brief Copies vector, divides element wise both vectors and
              * returns the new vector.
              * if both vectors differ in size an exception is raised
              * @param v  input vector
              * @see operator/=(const PVector &v)
              */
            PVector operator/ ( const PVector &v ) const;


            /** @brief Adds element wise both vectors and returns this vector.
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator+(const PVector &v)
              */
            const PVector & operator+= ( const PVector &v );


            /** @brief Subtracts element wise both vectors and returns this vector
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator-(const PVector &v)
              */
            const PVector & operator-= ( const PVector &v );


            /** @brief Multiplies element wise both vectors and returns the vector
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator*(const PVector &v)
              */
            const PVector & operator*= ( const PVector &v );


            /** @brief Divides element wise both vectors and returns the vector.
              * if both vectors differ in size an exception is raised
              * @returns this object
              * @param v  input vector
              * @see operator/(const PVector &v)
              */
            const PVector & operator/= ( const PVector &v );


            /** @brief Copies the vector, multiplies the vector with a vector:
              * vec = vec * M, and returns the vector
              * if the number of rows is unequal to the length of the vector
              * an exception is raised
              * @returns this object
              * @param m input vector
              * @see operator*=(const Pvector &m)
              * @see mul(const PVector &v, const Pvector &m)
              * @see mul(const Pvector &m, const PVector &v)
              */
            PVector operator* ( const PMatrix &m ) const;


            /** @brief Multiplies the vector with a vector: vec=vec*M
              * if the number of rows is unequalto the length of the evector
              * an exception is raised
              * @returns this object
              * @param m  input vector
              * @see operator*(const Pvector &m)
              * @see mul(const PVector &v, const Pvector &m)
              * @see mul(const Pvector &m, const PVector &v)
              */
            const PVector & operator*= ( const PMatrix &m );


            /** @brief copy operator.
              * copy all elements from vector (vec)
              * @returns this object
              * @param vec the vetor being copied
              */
            const PVector & operator= ( const PVector &vec );


            /** @brief comparison operator. If values if >= val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator>= ( double val ) const;


            /** @brief comparison operator. If values == val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator== ( double val ) const;


            /** @brief comparison operator. If values <= val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator<= ( double val ) const;


            /** @brief comparison operator. If values if > val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator> ( double val ) const;


            /** @brief comparison operator. If values , val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator< ( double val ) const;


            /** @brief comparison operator. If values if != val 1 else 0
              * @returns vector with 0 and 1
              */
            PIVector operator!= ( double val ) const;


            /** @returns sum of all items in the vector
              */
            double sum( void ) const;


            /** @returns the root mean square
              */
            PVector getSqrt( void ) const;


            /** @brief Multiplies the matrix with a vector: vec=M*vec and
              * copies the resulting vector in this object.
              * if the number of rows is unequal to the length of the vector
              * an exception is raised
              * @param v input vector v
              * @param m input vector m
              * @returns this object
              * @see operator*(const Pvector &m)
              * @see mul(const PVector &v, const Pvector &m)
              * @see mul(const Pvector &m, const PVector &v)
              */
            const PVector & mul( const PMatrix & m, const PVector & v );


            /** @brief Multiplies the vector (v) with a vector (m): vec=vec*M
              * if the number of rows is unequal to the length of the vector
              * an exception is raised
              * @returns this object
              * @param v input vector v
              * @param m input vector m
              * @see operator*(const Pvector &m)
              * @see mul(const PVector &v, const Pvector &m)
              * @see mul(const Pvector &m, const PVector &v)
              */
            const PVector & mul( const PVector & v, const PMatrix & m );


            /** @brief Normalize the vector
              */
            void normalize( void );


            /** @brief return a normalized copy of this vector
              */
            PVector normalized( void ) const;


            /** @brief Function to find the 1-norm of a matrix (Forsythe page 46)
              */
            double norm1() const;

            /** @brief Function to find the 2-norm of a matrix (Forsythe page 46)
              */
            double norm2() const;


            /** @brief Inserts value at the end of the vector.
              * @param value the value to append
              */
            void append( double value ) { mVec.append( value ); }


            /** @brief Inserts value at the end of the vector.
              * @param vec the vector to append
              */
            void append( const PVector& vec ) { mVec += vec.mVec; }


            /** @brief select a part of the data in the vector
              * @param start first sample
              * @param last last sample
              * @returns the selected part
              */
            PVector select(int start, int last);


            /** @brief select a part of the data in the vector
              * @param selection the indexes of the ones to select
              * @returns the selected part
              */
            PVector select(const PIVector & selection) const;

            /** @brief Reads vector data from file.
              * @see virtual bool PFile::read(QDatastream &ds);
              */
            bool read( QDataStream & ds );


            /** @brief Write vector data to file.
              *  @see virtual bool read(QDatastream &ds)=0;
              */
            bool write( QDataStream & ds ) const;


            /** @brief load the vector from a file.
              *  @see save(QString filename, unsigned int precision)
              */
            bool load( QString filename );


            /** @brief Save the vector to a text file with given precision.
              * @see load(QString filename)
              */
            bool save( QString filename, unsigned int precision = 8 ) const;


        protected:

            /** @brief Returns the class name
              * @see PFile
              */
            const char * getClassName( void ) const;


            /** @brief Implementation of a pure virtual function (PFile)that
              * requests the default class data version to write to the case
              * file
              */
            int getDefaultVersion( void ) const;


        private:

               /** @brief The Vector
                 */
            QVector <double> mVec;
        };
    }
}

#endif //PVECTOR_H
