/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PGrItemBase.h 902 2013-11-01 14:55:29Z eelco $
 *
 */


#ifndef __PGRAPHICSITEMBASE_H__
#define __PGRAPHICSITEMBASE_H__


#include "PeacsSupport_global.h"

#include <QGraphicsItem>

#include "PWidgetDefines.h"



namespace Peacs
{
    namespace Support
    {
        /** @class PGrItemBase PGrItemBase.h "widgets/PGrItemBase.h"
          * @brief The class holds extra information for Peacs graphics item implementations
          * This is the class for maintaining base information for graphics items as
          * implemented by Peacs. It keeps track of the position of this item in a matrix
          */
        class PEACSSUPPORTSHARED_EXPORT PGrItemBase
        {
        public:

            /**
              * The lowest permitted type value for project specific items (subclasses of QGraphicsItem or any of the
              * standard items) implemented in combination with the already customized Peacs Support graphics items
              * This value is used in conjunction with a reimplementation of QGraphicsItem::type() and declaring a
              * Type enum value.
              */
            enum {
                PeacsGrType = QGraphicsItem::UserType + 8700
            };


            /**
              * The type value returned by the virtual type() function in Peacs Support specific graphics item classes.
              * All these graphics item classes in Peacs Support are associated with a unique value for Type
              * This value is used in conjunction with a reimplementation of QGraphicsItem::type() and declaring a
              * Type enum value.
              */
            enum {
                PSignalPlotBaseType = QGraphicsItem::UserType + 5901,
                PSignalPlotType     = QGraphicsItem::UserType + 5902,
                PXYPlotType         = QGraphicsItem::UserType + 5903,
                PSimpleTextItemType = QGraphicsItem::UserType + 5904,
                PSignalDrawType     = QGraphicsItem::UserType + 5905
            };


            /** @brief Constructor
              * Constructs this object with the specified column and row
              * If column and/or row contain negative values, default implementation will
              * hide this item. If other functionality is wanted the user must override the
              * function PGraphicsSignalView::getItemPosition in a derived class to set the
              * correct position.
              *
              * @param column the column at which this item will reside in the matrix (default -1)
              * @param row the row at which this item will reside in the matrix (default -1)
              */
            PGrItemBase( int column = Widget::NOTUSED, int row = Widget::NOTUSED );


            /** @brief destructor
              */
            virtual ~PGrItemBase( void ) {;}


            /** @brief Return mColumn
              * Returns the column this item should be positioned in.
              */
            inline int getColumn( void ) const { return mColumn; }


            /** @brief Sets mColumn
              * Sets the column this item should be positioned in. If set to a negative
              * value, default implementation will hide this item.
              */
            inline void setColumn( int column ) { mColumn = column; }


            /** @brief Return mColumnSpan
              * Returns the column spanning of this item. This value is 1 or bigger
              */
            inline int getColumnSpan( void ) const { return mColumnSpan; }


            /** @brief Sets mColumnSpan
              * Sets the column spanning of this item. This valus must be 1 or bigger
              */
            inline void setColumnSpan( int columnSpan ) { Q_ASSERT( 0 < columnSpan); mColumnSpan = columnSpan; }


            /** @brief Return mRow
              * Returns the row this item should be positioned in.
              */
            inline int getRow( void ) const { return mRow; }


            /** @brief Sets mRow
              * Sets the row this item should be positioned in. If set to a negative
              * value, default implementation will hide this item.
              */
            inline void setRow( int row ) { mRow = row; }


            /** @brief Return mRowSpan
              * Returns the row spanning of this item. This value is 1 or bigger
              */
            inline int getRowSpan( void ) const { return mRowSpan; }


            /** @brief Sets mRowSpan
              * Sets the row spanning of this item. This valus must be 1 or bigger
              */
            inline void setRowSpan( int rowSpan ) { Q_ASSERT( 0 < rowSpan); mRowSpan = rowSpan; }


            /** @brief Returns if this item is configured to be positioned in the matrix
              * If this item is configured at a position within a matrix, mRow and mColumn have a valid
              * value, the positioning can be performed automatically by the base implementation. If not,
              * the application specific implementation is responcible for its positioning
              *
              * @return True if matrix coordinates are specified
              */
            inline bool isInMatrix( void ) const { return (mRow != Widget::NOTUSED) && (mColumn != Widget::NOTUSED); }


            /** Returns mBounds
              * Returns the current bounding rectangle of this items bounds
              */
            inline const QRectF & getBounds( void ) const { return mBounds; }


            /** @brief Sets the bounding (virtual) rectangle used when drawing
              * This internally used virtual drawing board bound will be used to draw within.
              * @note In the derived graphics item classes the function prepareGeometryChange()
              * must be called
              *
              * @param width The new width to use.
              * @param height The new height to use.
              * @return true if the bounding rectangle has changed
              */
            virtual bool setBounds( qreal width, qreal height ) = 0;


            /** @brief
              */
            virtual QSizeF getPreferredSize( void ) { return QSize(); }


        protected:

            /** @brief Will hold the bounding rectangle of the derived items
              */
            QRectF mBounds;


        private:

            /** @brief Holds the column in which this item is position
              * Default all items are positioned in a matrix. This item holds the column
              * for this item. The function PGraphicsSignalView::getItemPosition will
              * use this value to determain and set the position of this item
              *
              * According to the default paramteter in the constructor this value can be NOTUSED.
              * If set to NOTUSED the function getItemPosition must be overridden in a derived
              * class to set the correct position. Default implementation will hide the item.
              */
            int mColumn;


            /** @brief Holds the number of columns this item occupies
              * Default all items are positioned in a matrix. This member holds the number of
              * columns this item will occupy. The function PGraphicsSignalView::getItemPosition will
              * use this value to determain the width of this item
              *
              * The default value is 1 column
              */
            int mColumnSpan;


            /** @brief Holds the row in which this item is position
              * Default all items are positioned in a matrix. This item holds the row
              * for this item. The function PGraphicsSignalView::getItemPosition will
              * use this value to determain and set the position of this item
              *
              * According to the default paramteter in the constructor this value can be -1
              * If set to -1 the function getItemPosition must be overridden in a derived
              * class to set the correct position. Default implementation will hide the item.
              */
            int mRow;


            /** @brief Holds the number of rows this item occupies
              * Default all items are positioned in a matrix. This member holds the number of
              * rows this item will occupy. The function PGraphicsSignalView::getItemPosition will
              * use this value to determain the height of this item
              *
              * The default value is 1 row
              */
            int mRowSpan;
        };
    }
}

#endif // __PGRAPHICSITEMBASE_H__
