/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2013 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id$
 *
 */


#include <QGraphicsScene>
#include <QGraphicsItem>

//#include <QDebug>

#include "PGrSceneBase.h"
#include "PSignalDraw.h"


using namespace Peacs::Support;


// Initialization of static members
#ifdef WIN32
double  PGrSceneBase::mFrequency    = 0;
bool    PGrSceneBase::mFrequencySet = false;
#else
#include <sys/time.h>
#endif


PGrSceneBase::PGrSceneBase( QObject * parentPtr ) :
    QGraphicsScene( parentPtr ),
    mTimerID( 0 ),                  // Zero means no timer ID
    mRefreshInterval( 0 ),
    mRealIntervalTime()
{
}


bool PGrSceneBase::startUpdateTimer( int interval )
{
    bool done = false;

    // Check if a new timer interval is specified
    if ( interval != mRefreshInterval )
    {
        mRefreshInterval = interval;

        // If the timer is already running, kill it to enable (re-)initialization
        if ( 0 < mTimerID )
        {
            killTimer( mTimerID );

            mTimerID = 0;
        }
    }

    if ( 0 == mTimerID )
    {
#ifdef WIN32
        // Retrieve the timer frequency on Windows systems and store for later use
        if( ! mFrequencySet )
        {
            LARGE_INTEGER temp;

            if ( ::QueryPerformanceFrequency( (LARGE_INTEGER*) &temp ) )
            {
                mFrequency      = ((double) temp.QuadPart) / 1000.0;    // Convert to kHz or pulses/msec.

                mFrequencySet   = true;                                 // Frequency found
            }
        }

        if( mFrequencySet )
        {
#endif

        // Now start the times
        mTimerID = startTimer ( mRefreshInterval );

        // and get the current timestamp
#ifdef WIN32
        ::QueryPerformanceCounter( (LARGE_INTEGER *) & mRealIntervalTime );
#else
        gettimeofday( &mRealIntervalTime, 0 );
#endif

        if ( 0 < mTimerID )
        {
            done = true;
        }

#ifdef WIN32
        }
#endif

    }

    return done;
}


void PGrSceneBase::timerEvent( QTimerEvent * event )
{
    if ( event->timerId() == mTimerID )
    {
        // Use of the clock() function here is not appropriate due to the asynchronus handling of the timer events on a
        // multi threaded/application system. Clock() returns the processor time on the system. On the Mac this time is
        // very small compared to the Windows system. Use the code below to get the real interval time
#ifdef WIN32
        LARGE_INTEGER  now;

        QueryPerformanceCounter( (LARGE_INTEGER *) & now );

        int diff = qRound( double( now.QuadPart - mRealIntervalTime.QuadPart ) / mFrequency );
#else
        struct timeval now;

        gettimeofday( &now, 0 );

        // Get the time difference in milli-seconds
        int diff = qRound( (1000.0 * double(now.tv_sec - mRealIntervalTime.tv_sec)) + (0.001 * double(now.tv_usec - mRealIntervalTime.tv_usec)) );
#endif

        // Now we know the elapsed time interval. Inform the classes we know of that they need to be informed
        Q_ASSERT( NULL != dynamic_cast< QGraphicsScene *>( this ) );

        QList< QGraphicsItem * > itemsPtrLst = dynamic_cast< QGraphicsScene *>( this )->items();

        for ( int index = 0; index < itemsPtrLst.size(); index++ )
        {
            PSignalDraw * drawItemPtr = dynamic_cast< PSignalDraw * >( itemsPtrLst[ index ] );

            if ( NULL != drawItemPtr )
            {
                // Update this item because we know here that this items must be updated
                drawItemPtr->updateSignalForTimeElapse( diff );
            }
        }

        // Now, inform the rest
        updateDynItems( diff );

//#ifdef WIN32
//        LARGE_INTEGER  now2;

//        QueryPerformanceCounter( (LARGE_INTEGER *) & now2 );

//        double diff2 = double( now2.QuadPart - mRealIntervalTime.QuadPart ) / mFrequency;
//#else
//        struct timeval now2;

//        gettimeofday( &now2, 0 );

//        // Get the time difference in milli-seconds
//        double diff2 = (1000.0 * double(now2.tv_sec - mRealIntervalTime.tv_sec)) + (0.001 * double(now2.tv_usec - mRealIntervalTime.tv_usec));
//#endif

//        qDebug() << "PGrSceneBase::timerEvent:" << diff << "\t" << diff2 << "\n";

        mRealIntervalTime = now;
    }
}
