/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2013 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id$
 *
 */


#ifndef PGRSCENEBASE_H
#define PGRSCENEBASE_H


#include <QSize>
#include <QTimerEvent>
#include <QGraphicsScene>

#ifdef WIN32
    #include <windows.h>
#endif

#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PGrSceneBase PGrSceneBase.h "widget/PGrSceneBase.h"
          * @brief The interface class must be used for all scenes that can be used in the graph views
          */
        class PEACSSUPPORTSHARED_EXPORT PGrSceneBase : public QGraphicsScene
        {
            Q_OBJECT

        public:

            /** @brief Specified the default interval timeout to use for the timer when used
              */
            static const int DEFAULT_TIMER_INTERVAL_MS   = 20;


            /** @brief Constructor
              *
              * @param parentPtr Pointer to the parent for this scene object
              */
            PGrSceneBase( QObject * parentPtr = 0 );


            /** @brief This function will initialyze/set the bounding rectangle of all items and position them
              * This function should be called after all items haven been added to the scene and
              * when the view is resized
              * Implement this function in derived classes to position all items according to needs of the
              * specified scene
              */
            virtual void updateSceneLayout( void ) = 0;


            /** @brief Starts the timer that will trigger the update of dynamic scene items
              *
              * @param interval Specified the interval in miliseconds (default DEFAULT_TIMER_INTERVAL_MS)
              * @return true if all is initialized and the timer is started
              */
            bool startUpdateTimer( int interval = DEFAULT_TIMER_INTERVAL_MS );


            /** @brief Forces all items to rescale themself
              * This function is called after the autoScaleSlot() slot is triggered by a graph item (plot or draw)
              * and when the update action in a graph item (by calling autoScaleSlot() function of that class) has
              * returned true, indicating a rescal is needed.
              * Default implementation will tell all items to rescale themself according the items signal data
              * min/max values. Override this function to perform a different kind of scaling
              */
            virtual void autoScale( void ) = 0;


        public slots:

            /** @brief Recieves all events from graph items inticating a rescale is needed
              * This signal is passed on to the real (virtual) rescale function of this view
              * @see PGraphMatrixScene::addItem
              */
            void autoScaleSlot( void ) { autoScale(); }


            /** @brief Recieves signals from graph items inticating an amplitude rescale is needed
              * This signal is passed on to the real (virtual) amplitude rescale function of this view
              *
              * @param itemPtr Points to the item initiating the signal
              * @see PGraphMatrixScene::addItem
              */
            void scaleAmplitudeSlot( QObject * itemPtr ) { scaleAmplitude( itemPtr ); }


            /** @brief Recieves signals from graph items inticating a time rescale is needed
              * This signal is passed on to the real (virtual) time rescale function of this view
              *
              * @param itemPtr Points to the item initiating the signal
              * @see PGraphMatrixScene::addItem
              */
            void scaleTimeSlot( QObject * itemPtr ) { scaleTime( itemPtr ); }


        protected:

            /** @brief Handles the signals from graph items inticating an amplitude rescale is perfomed
              * This function is the default implementation of handling the amplitude scaling at a graph item. When
              * an item changes its amplitude scale, it will emit a signal to inform the view. The view is
              * responsible to take the appropriate actions with regards to the other items in the view, i.e.
              * rescale the amplitude of those items the same or just some of them.
              * This default implementation does noting. Implement in a derived class to handle in an
              * application specific way
              *
              * @param itemPtr Points to the item initiating the signal
              * @see PGrSceneBase::scaleAmplitudeSlot
              */
            virtual void scaleAmplitude( QObject * itemPtr ) { Q_UNUSED(itemPtr); }


            /** @brief Handles the signals from graph items inticating a time rescale is perfomed
              * This function is the default implementation of handling the time scaling at a graph item. When
              * an item changes its time scale, it will emit a signal to inform the view. The view is
              * responsible to take the appropriate actions with regards to the other items in the view, i.e.
              * rescale the time of those items the same or just some of them.
              * This default implementation does noting. Implement in a derived class to handle in an
              * application specific way
              *
              * @param itemPtr Points to the item initiating the signal
              * @see PGrSceneBase::scaleTimeSlot
              */
            virtual void scaleTime( QObject * itemPtr ) { Q_UNUSED(itemPtr); }


            /** @brief This timer event handler is used to receive timer events for this object
              *
              * @param event Points to the object containing information about this timer event
              */
            void timerEvent( QTimerEvent * event );


            /** @brief This function is  called while handling the time event
              * Implement this function in a derived class to handle this event and do something application
              * specific action. Default the timer event will update all PSignalDraw (derived) item objects.
              * All custom items should be adjusted in the specific application code
              *
              * @param elapsed Holds the elapsed time in milliseconds since the last occurred time event or
              *                the timer start
              */
            virtual void updateDynItems( int elapsed ) { Q_UNUSED( elapsed ); }


        private:

            /** @brief Will hold the time-out timer ID if correctly created
              */
            int mTimerID;


            /** @brief Defines the interval timeout in miliseconds
              */
            int mRefreshInterval;


            /** @brief Holds the clock timer tick value of the last time the offsets were adjusted
              * This member will be used to retrieve the correct timer elapse time since the last
              * time-out
              */
#ifdef WIN32
            LARGE_INTEGER  mRealIntervalTime;
#else
            struct timeval  mRealIntervalTime;
#endif

#ifdef WIN32
            /** @brief Frequency of the system in kHz
              * Only on Windows systems needed
              */
            static double  mFrequency;


            /** @brief Flag that indicates if the frequency has been determained and set
              * Only on Windows systems needed
              */
            static  bool  mFrequencySet;
#endif
        };
    }
}

#endif // PGRSCENEBASE_H
