/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2013 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id$
 *
 */

//#include <QDebug>
#include <QGraphicsView>

#include "PGraphMatrixScene.h"
#include "PGraphicsSignalView.h"
#include "PGrItemBase.h"
#include "psignalplotBase.h"



/** @brief This value represent the initial space between the items in the view
  */
const int   SPACEBETWEEN = 3;


using namespace Peacs::Support;


PGraphMatrixScene::PGraphMatrixScene( int columns, int rows, QObject * parentPtr ) :
    PGrSceneBase( parentPtr ),
    mColumns( columns ),
    mHideColumns( columns, false ),         // Default all columns are shown
    mRows( rows ),
    mHideRows( rows, false ),               // Default all rows are shown
    mSpaceBetween( SPACEBETWEEN ),
    mMatrixItems(),
    mColumnHeaderItems(),
    mRowHeaderItems(),
    mCustomItems(),
    mMsAmplScaleChanged( NULL ),
    mMsTimeScaleChanged( NULL )
{
    Q_ASSERT( (0 < mColumns) && (0 < mRows) );

    mMsAmplScaleChanged = new QSignalMapper( this );
    mMsTimeScaleChanged = new QSignalMapper( this );

    connect( mMsAmplScaleChanged, SIGNAL( mapped(QObject*) ), this, SLOT( scaleAmplitudeSlot(QObject *) ) );
    connect( mMsTimeScaleChanged, SIGNAL( mapped(QObject*) ), this, SLOT( scaleTimeSlot(QObject *) ) );
}


PGraphMatrixScene::~PGraphMatrixScene( void )
{
}


void PGraphMatrixScene::showColumn( int index, bool show )
{
    Q_ASSERT( 0 <= index && index < mHideColumns.size() );

    if ( mHideColumns.at( index ) == show )
    {
        mHideColumns[ index ] = !show;

        //Hide all items in that column (including the header items)
        hideColumnRowItems( !show, index, true );

        updateSceneLayout();

        invalidate( sceneRect() );      // Redraw all
    }
}


void PGraphMatrixScene::showRow( int index, bool show )
{
    Q_ASSERT( 0 <= index && index < mHideRows.size() );

    if ( mHideRows.at( index ) == show )
    {
        mHideRows[ index ] = !show;

        //Hide all items in that row (including the header items)
        hideColumnRowItems( !show, index, false );

        updateSceneLayout();

        invalidate( sceneRect() );      // Redraw all
    }
}


void PGraphMatrixScene::addItem( PMatrixSceneItemType type, QGraphicsObject * itemPtr )
{
    PGrItemBase *           pGrItemBaseptr  = NULL;
    PGraphicsSignalView *   viewPtr         = NULL;

    // Find the view pointer to connect the signal/slot for
    if ( /*isActive() &&*/  !views().isEmpty() )
    {
        QList <QGraphicsView *> viewList = views();

        // Search for the first view that shows this scene and retrieve its maximum viewport size
        for ( int index = 0; (index < viewList.size()) && !viewPtr; index++ )
        {
            viewPtr = dynamic_cast<PGraphicsSignalView *>( viewList.at( index ) );
        }
    }

    QGraphicsScene::addItem( itemPtr );

    switch( type )
    {
    case matrixItem:
        Q_ASSERT_X( NULL != dynamic_cast<PGrItemBase *>( itemPtr ), "PGraphMatrixScene::addItem", "The item must be derived from PGrItemBase" );
        Q_ASSERT_X( dynamic_cast<PGrItemBase *>( itemPtr )->isInMatrix(), "PGraphMatrixScene::addItem", "The item must be position within the matrix" );
        mMatrixItems.append( itemPtr );

        //Show the new item only when that column or row must be shown
        pGrItemBaseptr = dynamic_cast<PGrItemBase *>( itemPtr );
        itemPtr->setVisible( !mHideColumns.at( pGrItemBaseptr->getColumn() ) && !mHideRows.at( pGrItemBaseptr->getRow() ) );

        // If this is a graph item, connect the auto scale and time/amplitude scale change signal/slot to enable easy rescaling
        if ( viewPtr && dynamic_cast<PSignalPlotBase *>( itemPtr ) )
        {
            PSignalPlotBase * basePtr = dynamic_cast<PSignalPlotBase *>( itemPtr );

            connect( basePtr, SIGNAL(doAutoScale()), this, SLOT(autoScaleSlot()) );

            connect( basePtr, SIGNAL( amplScaleChanged() ), mMsAmplScaleChanged, SLOT( map() ) );
            mMsAmplScaleChanged->setMapping( basePtr, itemPtr );

            connect( basePtr, SIGNAL( timeScaleChanged() ), mMsTimeScaleChanged, SLOT( map() ) );
            mMsTimeScaleChanged->setMapping( basePtr, itemPtr );
        }
        break;

    case columnHeaderItem:
        Q_ASSERT_X( NULL != dynamic_cast<PGrItemBase *>( itemPtr ), "PGraphMatrixScene::addItem", "The item must be derived from PGrItemBase" );
        Q_ASSERT_X( dynamic_cast<PGrItemBase *>( itemPtr )->isInMatrix(), "PGraphMatrixScene::addItem", "The item must be configured for a column" );
        mColumnHeaderItems.append( itemPtr );

        //Show the new header item only when that column must be shown
        pGrItemBaseptr = dynamic_cast<PGrItemBase *>( itemPtr );
        itemPtr->setVisible( !mHideColumns.at( pGrItemBaseptr->getColumn() ) );
        break;

    case rowHeaderItem:
        Q_ASSERT_X( NULL != dynamic_cast<PGrItemBase *>( itemPtr ), "PGraphMatrixScene::addItem", "The item must be derived from PGrItemBase" );
        Q_ASSERT_X( dynamic_cast<PGrItemBase *>( itemPtr )->isInMatrix(), "PGraphMatrixScene::addItem", "The item must be configured for a row" );
        mRowHeaderItems.append( itemPtr );

        //Show the new header item only when that row must be shown
        pGrItemBaseptr = dynamic_cast<PGrItemBase *>( itemPtr );
        itemPtr->setVisible( !mHideRows.at( pGrItemBaseptr->getRow() ) );
        break;

    case customItem:
        mCustomItems.append( itemPtr );
        break;

    default:
        Q_ASSERT( false );
        break;
    }
}


bool PGraphMatrixScene::setSpaceBetween( int spaceBetween )
{
    bool changed = false;

    if ( spaceBetween != mSpaceBetween )
    {
        mSpaceBetween   = spaceBetween;
        changed         = true;

        updateSceneLayout();

        invalidate( sceneRect() );      // Redraw all
    }

    return changed;
}


void PGraphMatrixScene::updateSceneLayout( void )
{
    QSize viewportSize = getViewportSize();

    if ( viewportSize.isValid() )
    {
        int             val = 0;
        int             index;
        int             nrColumns           = mColumns - getHiddenColumnRowCount( true );
        int             nrRows              = mRows - getHiddenColumnRowCount( false );
        qreal           rowHeaderWidth      = 0.0;
        qreal           columnHeaderHeight  = 0.0;
        qreal           initialItemWidth    = viewportSize.width() - mSpaceBetween;
        qreal           initialItemHeight   = viewportSize.height() - mSpaceBetween;
        qreal           offsetX             = mSpaceBetween / 2;
        qreal           offsetY             = mSpaceBetween / 2;
        QSizeF          preferredItemSize;
        QPointF         positionPoint;
        PGrItemBase *   pGrItemBaseptr      = NULL;
        QVector< int >  correctedColumnIndexes( mColumns, -1 );
        QVector< int >  correctedRowIndexes( mRows, -1 );

        setSceneRect( 0, 0 , viewportSize.width(), viewportSize.height() );
    
        // Correct the row indexes that are used to position the items at the right row for
        // rows that are not being shown
        for ( index = 0; index < mHideRows.size(); index++ )
        {
            if ( !mHideRows.at( index ) )
            {
                correctedRowIndexes[ index ] = val;
    
                val++;
            }
        }
    
        // Now correct the column indexes that are used to position the items at the right column for
        // columns that are not being shown
        val = 0;
        for ( index = 0; index < mHideColumns.size(); index++ )
        {
            if ( !mHideColumns.at( index ) )
            {
                correctedColumnIndexes[ index ] = val;
    
                val++;
            }
        }
    
        // Get the initial row header width
        if ( 0 < mRowHeaderItems.size() )
        {
            for( index = 0; index < mRowHeaderItems.size(); index++ )
            {
                pGrItemBaseptr  = dynamic_cast<PGrItemBase *>( mRowHeaderItems[ index ] );
    
                if ( (0 <= pGrItemBaseptr->getRow()) && (pGrItemBaseptr->getRow() < mHideRows.size() ) && !mHideRows.at( pGrItemBaseptr->getRow() ) )
                {
                    // Calling getPreferredSize() will change the bounding rectangle from the item. This rectangle must be corrected later on
                    preferredItemSize = pGrItemBaseptr->getPreferredSize();

                    // We use the minimum width to get the easiest width to handle. The use must change the font size
                    // to get better fitting
                    if ( rowHeaderWidth == 0.0 )
                    {
                        rowHeaderWidth = preferredItemSize.width();
                    }
                    else
                    {
                        rowHeaderWidth = qMin( rowHeaderWidth, preferredItemSize.width() );
                    }
//qDebug() << "Item [" << pGrItemBaseptr->getColumn() << "," << pGrItemBaseptr->getRow() << "has preferred size" << preferredItemSize;
                }
            }
//qDebug() << "Best row header width =" << rowHeaderWidth;
        }

        // Get the initial column header height
        if ( 0 < mColumnHeaderItems.size() )
        {
            for( index = 0; index < mColumnHeaderItems.size(); index++ )
            {
                pGrItemBaseptr  = dynamic_cast<PGrItemBase *>( mColumnHeaderItems[ index ] );

                if ( (0 <= pGrItemBaseptr->getColumn()) && (pGrItemBaseptr->getColumn() < mHideRows.size() ) && !mHideColumns.at( pGrItemBaseptr->getColumn() ) )
                {
                    // Calling getPreferredSize() will change the bounding rectangle from the item. This rectangle must be corrected later on
                    preferredItemSize = pGrItemBaseptr->getPreferredSize();

                    // We use the minimum height to get the easiest height to handle. The use must change the font size
                    // to get better fitting
                    if ( columnHeaderHeight == 0.0 )
                    {
                        columnHeaderHeight = preferredItemSize.height();
                    }
                    else
                    {
                        columnHeaderHeight = qMin( columnHeaderHeight, preferredItemSize.width() );
                    }
//qDebug() << "Item [" << pGrItemBaseptr->getColumn() << "," << pGrItemBaseptr->getRow() << "has preferred size" << preferredItemSize;
                }
            }
//qDebug() << "Best column header height =" << columnHeaderHeight;
        }

        // Now we know the column/row header height/width
        if ( 0.0 < rowHeaderWidth )
        {
            initialItemWidth -= (rowHeaderWidth + mSpaceBetween);
        }
        if ( 0.0 < columnHeaderHeight )
        {
            initialItemHeight -= (columnHeaderHeight + mSpaceBetween);
        }

        // Now find the resulting matrix item width and height by removing the seperating space and
        // then devide the rest over the columns/rows
        initialItemWidth -= mSpaceBetween * (nrColumns - 1);
        initialItemWidth = initialItemWidth / nrColumns;
        initialItemHeight -= mSpaceBetween * (nrRows - 1);
        initialItemHeight = initialItemHeight / nrRows;
//qDebug() << "Item width and height =" << initialItemWidth << initialItemHeight;

        // Now set the bounding rectangle of all items and position them
        // Start with the row headers
        if ( 0 < mRowHeaderItems.size() )
        {
            offsetX = mSpaceBetween / 2;
            offsetY = mSpaceBetween / 2 + columnHeaderHeight + mSpaceBetween;

            for( index = 0; index < mRowHeaderItems.size(); index++ )
            {
                pGrItemBaseptr  = dynamic_cast<PGrItemBase *>( mRowHeaderItems[ index ] );

                if ( (0 <= pGrItemBaseptr->getRow()) && (pGrItemBaseptr->getRow() < mHideRows.size() ) && !mHideRows.at( pGrItemBaseptr->getRow() ) )
                {
                    pGrItemBaseptr->setBounds( rowHeaderWidth,
                                               initialItemHeight * pGrItemBaseptr->getRowSpan() +  (pGrItemBaseptr->getRowSpan() - 1) * mSpaceBetween );

                    positionPoint = adjustForRotation( offsetX,
                                                       offsetY + correctedRowIndexes[ pGrItemBaseptr->getRow() ] * (initialItemHeight + mSpaceBetween),
                                                       mRowHeaderItems[ index ] );

                    mRowHeaderItems[ index ]->setPos( positionPoint );
//qDebug() << "Row header item at [" << pGrItemBaseptr->getRow() << "] is at position" << mRowHeaderItems[ index ]->pos() << "with bounds" << pGrItemBaseptr->getBounds();
                }
            }
        }

        // Now the column headers
        if ( 0 < mColumnHeaderItems.size() )
        {
            offsetX = mSpaceBetween / 2 + rowHeaderWidth + mSpaceBetween;
            offsetY = mSpaceBetween / 2;

            for( index = 0; index < mColumnHeaderItems.size(); index++ )
            {
                pGrItemBaseptr  = dynamic_cast<PGrItemBase *>( mColumnHeaderItems[ index ] );

                if ( (0 <= pGrItemBaseptr->getColumn()) && (pGrItemBaseptr->getColumn() < mHideRows.size() ) && !mHideColumns.at( pGrItemBaseptr->getColumn() ) )
                {
                    pGrItemBaseptr->setBounds( initialItemWidth * pGrItemBaseptr->getColumnSpan() + (pGrItemBaseptr->getColumnSpan() - 1) * mSpaceBetween,
                                               columnHeaderHeight );

                    positionPoint = adjustForRotation( offsetX + correctedColumnIndexes[ pGrItemBaseptr->getColumn() ] * (initialItemWidth + mSpaceBetween),
                                                       offsetY,
                                                       mColumnHeaderItems[ index ] );

                    mColumnHeaderItems[ index ]->setPos( positionPoint );
//qDebug() << "Column header item at [" << pGrItemBaseptr->getColumn() << "] is at position" << mColumnHeaderItems[ index ]->pos() << "with bounds" << pGrItemBaseptr->getBounds();
                }
            }
        }

        // Prepare for the matrix items positioning. Get the starting top left position
        if ( 0.0 == rowHeaderWidth )
        {
            offsetX = mSpaceBetween / 2;
        }
        else
        {
            offsetX = mSpaceBetween / 2 + rowHeaderWidth + mSpaceBetween;
        }

        if ( 0.0 == columnHeaderHeight )
        {
            offsetY = mSpaceBetween / 2;
        }
        else
        {
            offsetY = mSpaceBetween / 2 + columnHeaderHeight + mSpaceBetween;
        }

        // And now the matrix items
        for ( index = 0; index < mMatrixItems.size(); index++ )
        {
            pGrItemBaseptr = dynamic_cast<PGrItemBase *>( mMatrixItems[ index ] );

            if ( !mHideColumns.at( pGrItemBaseptr->getColumn() ) || mHideRows.at( pGrItemBaseptr->getRow() ) )
            {
                qreal itemOffsetX = offsetX +
                                    correctedColumnIndexes[ pGrItemBaseptr->getColumn() ] * (initialItemWidth + mSpaceBetween);
                qreal itemOffsetY = offsetY +
                                    correctedRowIndexes[ pGrItemBaseptr->getRow() ] * (initialItemHeight + mSpaceBetween);

                pGrItemBaseptr->setBounds( initialItemWidth * pGrItemBaseptr->getColumnSpan() + (pGrItemBaseptr->getColumnSpan() - 1) * mSpaceBetween,
                                           initialItemHeight * pGrItemBaseptr->getRowSpan() +  (pGrItemBaseptr->getRowSpan() - 1) * mSpaceBetween );

                mMatrixItems[ index ]->setPos( itemOffsetX, itemOffsetY );
//qDebug() << "item at [" << pGrItemBaseptr->getColumn() << pGrItemBaseptr->getRow() << "] is at position" << mMatrixItems[ index ]->pos() << "with bounds" << pGrItemBaseptr->getBounds();
            }
        }
    }
}


QSize PGraphMatrixScene::getViewportSize( void ) const
{
    QSize size;

    if ( /*isActive() &&*/  !views().isEmpty() )
    {
        QList <QGraphicsView *> viewList = views();

        // Search for the first view that shows this scene and retrieve its maximum viewport size
        for ( int index = 0; (index < viewList.size()) && !size.isValid(); index++ )
        {
            const QGraphicsView * viewPtr = viewList.at( index );

            if ( viewPtr->scene() == this )
            {
                size = viewPtr->maximumViewportSize();
            }
        }
    }

    return size;
}


void PGraphMatrixScene::hideColumnRowItems( bool hide, int columnRowIndex, bool column )
{
    QList<QGraphicsItem*>   sceneItems( items() );      // Get all scene items
    QGraphicsItem *         itemPtr         = NULL;
    PGrItemBase *           pItemBasePtr    = NULL;

    // Show or hide the matrix aligned items
    for ( int index = 0; index < sceneItems.size(); index++ )
    {
        itemPtr         = sceneItems[ index ];
        pItemBasePtr    = dynamic_cast<PGrItemBase *>( itemPtr );

        if ( NULL != pItemBasePtr &&
             pItemBasePtr->isInMatrix() &&
             (column ? (pItemBasePtr->getColumn() == columnRowIndex) : (pItemBasePtr->getRow() == columnRowIndex) ) )
        {
            itemPtr->setVisible( !hide );
        }
    }

    // Now show or hide the column or row aligned item
    if ( column )
    {
        if ( (0 <= columnRowIndex) && (columnRowIndex < mColumnHeaderItems.size()) )
        {
            mColumnHeaderItems[ columnRowIndex ]->setVisible( !hide );
        }
    }
    else
    {
        if ( (0 <= columnRowIndex) && (columnRowIndex < mRowHeaderItems.size()) )
        {
            mRowHeaderItems[ columnRowIndex ]->setVisible( !hide );
        }
    }
}


int PGraphMatrixScene::getHiddenColumnRowCount( bool columns ) const
{
    int                     number      = 0;
    const QVector< bool > * vectorPtr   = NULL;

    if ( columns )
    {
        vectorPtr = &mHideColumns;
    }
    else
    {
        vectorPtr = &mHideRows;
    }

    for ( int index = 0; index < vectorPtr->size(); index++ )
    {
        if ( vectorPtr->at( index ) )
        {
            number++;
        }
    }

    return number;
}


QPointF PGraphMatrixScene::adjustForRotation( qreal xPos, qreal yPos, const QGraphicsItem * itemPtr ) const
{
    QPointF             correctedPoint;
    const PGrItemBase * pItemBasePtr = dynamic_cast<const PGrItemBase *>( itemPtr );

    if ( itemPtr->rotation() == -90 )
    {
        correctedPoint.setX( xPos );
        correctedPoint.setY( yPos + pItemBasePtr->getBounds().width() );
    }
    else if ( itemPtr->rotation() == 90 )
    {
        correctedPoint.setX( xPos + pItemBasePtr->getBounds().height() );
        correctedPoint.setY( yPos );
    }
    else
    {
        correctedPoint.setX( xPos );
        correctedPoint.setY( yPos );
    }

    return correctedPoint;
}


void PGraphMatrixScene::autoScale( void )
{
    // Inform all items that a rescale is to be performed
    for ( int i = 0; i < mMatrixItems.size(); i++ )
    {
        PSignalPlotBase *signal = dynamic_cast<PSignalPlotBase *>( mMatrixItems[i] );

        if ( (signal != NULL) &&
             (signal->getMinYData() < signal->getMaxYData()) &&
             (signal->getMinXData() < signal->getMaxXData()))
        {
            signal->setSignalAmpl( signal->getMinYData(), signal->getMaxYData(), false );
            signal->setTimeRange( signal->getMinXData(), signal->getMaxXData(), false );
        }
    }
}
