/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2013 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id$
 *
 */


#ifndef PGRAPHMATRIXSCENE_H
#define PGRAPHMATRIXSCENE_H


#include <QList>
#include <QSignalMapper>
#include <QGraphicsObject>

#include "PGrSceneBase.h"
#include "PeacsSupport_global.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PGraphMatrixScene PGraphMatrixScene.h "widget/PGraphMatrixScene.h"
          * @brief Is used to place all signal graph items in a matrix shape
          */
        class PEACSSUPPORTSHARED_EXPORT PGraphMatrixScene : public PGrSceneBase
        {
            Q_ENUMS( PMatrixSceneItemType )

        public:

            /** @brief defines the types of items this scene can represent
              */
            enum PMatrixSceneItemType { matrixItem,         // Items aligned/positioned within the matrix
                                        columnHeaderItem,   // Items positioned as header on the top or bottom of the columns
                                        rowHeaderItem,      // Items positioned before or after the rows items
                                        customItem };       // Other items that are added to this scene but for which derived classes or the view is responsible

            /** @brief Constructor
              *
              * @param columns Specifies the number of columns shown in this scene
              * @param rows Specifies the number of rows shown in this scene
              * @param parentPtr Pointer to the parent for this scene object
              */
            PGraphMatrixScene( int columns, int rows, QObject * parentPtr = 0 );


            /** @brief Destructor
              */
            virtual ~PGraphMatrixScene( void );


            /** @brief Returns the number of columns
              *
              * @return mColumns
              */
            int getColumnCount( void ) const { return mColumns; }


            /** @brief Shows or hides the specified column
              *
              * @param index Specifies the column to show or hide
              * @param show If true, the column wil be shown
              */
            void showColumn( int index, bool show );


            /** @brief Returns if the specified column is currently being shown
              *
              * @param index Specifies the column
              * @return True is the specified column is being shown
              */
            bool isColumnShown( int index ) const { Q_ASSERT( 0 <= index && index < mHideColumns.size() ); return !mHideColumns[ index ]; }


            /** @brief Returns the number of rows
              *
              * @return mRows
              */
            int getRowCount( void ) const { return mRows; }


            /** @brief Shows or hides the specified row
              *
              * @param index Specifies the row to show or hide
              * @param show If true, the row wil be shown
              */
            void showRow( int index, bool show );


            /** @brief Returns if the specified row is currently being shown
              *
              * @param index Specifies the row
              * @return True is the specified row is being shown
              */
            bool isRowShown( int index ) const { Q_ASSERT( 0 <= index && index < mHideRows.size() ); return !mHideRows[ index ]; }


            /** @brief This function must be used to add items to this scene
              * Together with the items, the use of that item must be specified with it.
              */
            void addItem( PMatrixSceneItemType type, QGraphicsObject * itemPtr );


            /** @brief Returns the space between the items in the view
              *
              * @return mSpaceBetween
              */
            inline int getSpaceBetween( void ) const { return mSpaceBetween; }


            /** @brief Sets the space between the items in the view
              * This function will scedule an update of the whole scene
              *
              * @param spaceBetween the new value for mSpaceBetween in scene coordinate system
              * @return true if the value has changed
              */
            bool setSpaceBetween( int spaceBetween );


            /** @brief Returns the scene items that are positioned within the matrix
              *
              * @return Returns mMatrixItems as a const object
              */
            inline const QList< QGraphicsItem * > & getMatrixItems( void ) const { return mMatrixItems; }


            /** @brief Returns the scene items that are positioned within the matrix
              *
              * @return Returns mMatrixItems
              */
            inline QList< QGraphicsItem * > & matrixItems( void ) { return mMatrixItems; }


            /** @brief Returns the scene items that are positioned on top of the matrix columns
              *
              * @return Returns mColumnHeaderItems as a const object
              */
            inline const QList< QGraphicsItem * > & getColumnHeaderItems( void ) const { return mColumnHeaderItems; }


            /** @brief Returns the scene items that are positioned on top of the matrix columns
              *
              * @return Returns mColumnHeaderItems
              */
            inline QList< QGraphicsItem * > & columnHeaderItems( void ) { return mColumnHeaderItems; }


            /** @brief Returns the scene items that are positioned before the matrix rows
              *
              * @return Returns mRowHeaderItems as a const object
              */
            inline const QList< QGraphicsItem * > & getRowHeaderItems( void ) const { return mRowHeaderItems; }


            /** @brief Returns the scene items that are positioned before the matrix rows
              *
              * @return Returns mRowHeaderItems
              */
            inline QList< QGraphicsItem * > & rowHeaderItems( void ) { return mRowHeaderItems; }


            /** @brief Returns the scene items that are not positioned within the matrix
              * These items have no real relation to the matrix defined in this scene. Positioning of these
              * items is considered not to be the responsibility of this implementation. Derived classes or
              * the scenes view must position them and determain theire extent
              *
              * @return Returns mCustomItems as a const object
              */
            inline const QList< QGraphicsItem * > & getCustomItems( void ) const { return mCustomItems; }


            /** @brief Returns the scene items that are not positioned within the matrix
              * These items have no real relation to the matrix defined in this scene. Positioning of these
              * items is considered not to be the responsibility of this implementation. Derived classes or
              * the scenes view must position them and determain theire extent
              *
              * @return Returns mCustomItems
              */
            inline QList< QGraphicsItem * > & customItems( void ) { return mCustomItems; }


            /** @brief This function will initialyze/set the bounding rectangle of all items and position them
              * This function should be called after all items haven been added to the scene and
              * when the view is resized
              */
            void updateSceneLayout( void );


            /** @brief Forces all items to rescale themself
              * This function is called after the autoScaleSlot() slot is triggered by a graph item (plot or draw)
              * and when the update action in a graph item (by calling autoScaleSlot() function of that class) has
              * returned true, indicating a rescal is needed.
              * Default implementation will tell all items to rescale themself according the items signal data
              * min/max values. Override this function to perform a different kind of scaling
              */
            virtual void autoScale( void );


        private:

            /** @brief The default function to add items to the scene should not be used
              */
            void addItem ( QGraphicsItem * /*item*/ ) { Q_ASSERT( false ); }


            /** @brief Retrieves the scenes viewport size
              * For this all views that are related to this scene are searched. The first
              * one showing this scene will be used to retrieve the viewport size
              *
              * @return The (maximum) viewport size
              */
            QSize getViewportSize( void ) const;


            /** @brief Hides or shows all items in the specified row or column
              *
              * @param hide If true, the column/row must be hidden
              * @param columnRowIndex Specifies the index of the column/row
              * @param column If true, a column must be hidden/shown
              */
            void hideColumnRowItems( bool hide, int columnRowIndex, bool column );


            /** @brief Returns the number of hidden columns/rows
              *
              * @param columns If true, the hidden items in the columns will be counted
              * @return The number of hidden items
              */
            int getHiddenColumnRowCount( bool columns ) const;


            /** @brief Adjusts the given item position points for the rotation of the item
              * This function handles -90, 0 and 90 degrees
              *
              * @param xPos The starting X position
              * @param yPos The starting Y position
              * @param itemPtr Points to the item the rotation and bounds is determained from
              * @return The adjusted position
              */
            QPointF adjustForRotation( qreal xPos, qreal yPos, const QGraphicsItem * itemPtr ) const;


            /** @brief Defines the number of columns present in the matrix
              */
            int mColumns;


            /** @brief For all columns this list holds the flag indicating if that column must be shown or not
              */
            QVector< bool > mHideColumns;


            /** @brief Defines the number of rows present in the matrix
              */
            int mRows;


            /** @brief For all rows this list holds the flag indicating if that row must be shown or not
              */
            QVector< bool > mHideRows;


            /** @brief This value represent the space between the items in the scene
              * This value is a number represented in scene coordinate system
              */
            int mSpaceBetween;


            /** @brief Used to store the scene items that are positioned within the matrix
              */
            QList< QGraphicsItem * > mMatrixItems;


            /** @brief Used to store the scene items that are positioned on top of the matrix columns
              */
            QList< QGraphicsItem * > mColumnHeaderItems;


            /** @brief Used to store the scene items that are positioned before the matrix rows
              */
            QList< QGraphicsItem * > mRowHeaderItems;


            /** @brief Used to store the scene items that are not positioned within the matrix
              * These items have no real relation to the matrix defined in this scene. Positioning of these
              * items is considered not to be the responsibility of this implementation. Derived classes or
              * the scenes view must position them and determain theire extent
              */
            QList< QGraphicsItem * > mCustomItems;


            /** @brief Is used to connect all PSignalPlotBase::amplScaleChanged signals
              */
            QSignalMapper * mMsAmplScaleChanged;


            /** @brief Is used to connect all PSignalPlotBase::timeScaleChanged signals
              */
            QSignalMapper * mMsTimeScaleChanged;
        };
    }
}

#endif // PGRAPHMATRIXSCENE_H
