/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PGraphicsSignalView.cpp 1172 2014-03-21 13:45:45Z eelco $
 *
 */


#include <QPainter>
#include <QList>
#include <qglobal.h>
#include <QGraphicsScene>
#include <QApplication>
#include <QClipboard>
#include <QToolTip>
#include <QHelpEvent>

#include <math.h>

#include "PGraphicsSignalView.h"
#include "PSignalDraw.h"
#include "PSignalPlot.h"
#include "PSignalPlotBase.h"
#include "PSimpleTextItem.h"
#include "PXYPlot.h"
#include "PGrSceneBase.h"

#include "pfunctions.h"
#include "PSettings.h"

//#include <QDebug>


using namespace Peacs::Support;


PGraphicsSignalView::PGraphicsSignalView( QWidget * parent ) :
    QGraphicsView( parent ),
    mBackGroundColor(),
    mBackGroundColorPrint( Qt::white ),
    mIsPrinting( false )
{
    setAttribute( Qt::WA_AlwaysShowToolTips );

    // Disable scroll bars. They shouldn't be needed
    setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
    setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );

    //setColors();

    setCacheMode( CacheBackground );    // Let Qt handle the cashed background bitmap

    setRenderHints( QPainter::Antialiasing |
                    QPainter::TextAntialiasing |
                    QPainter::SmoothPixmapTransform |
                    QPainter::HighQualityAntialiasing );

//    setViewportUpdateMode(QGraphicsView::FullViewportUpdate);//SmartViewportUpdate);
    setViewportUpdateMode(QGraphicsView::SmartViewportUpdate);
}


PGraphicsSignalView::~PGraphicsSignalView()
{
    removeMyItems();
}


void PGraphicsSignalView::initialize( void )
{
    Q_ASSERT_X( scene(), "PGraphicsSignalView::initialize", "Before initializing the graphics signal view an appropriate scene must be added" );

    Q_ASSERT_X( dynamic_cast<PGrSceneBase *>( scene() ), "PGraphicsSignalView::initialize", "Before initializing the graphics signal view an appropriate scene must be added" );

    removeMyItems();

    initialyzeItems();

    positionItems();
}


void PGraphicsSignalView::initializeView( void )
{
    if ( scene() )
    {
        QList<QGraphicsItem*> scene_items( scene()->items() );

        for ( int i = 0; i < scene_items.size() ; i++ )
        {
            PSignalPlotBase *signal = dynamic_cast<PSignalPlotBase *>( scene_items[i] );

            if ( signal != NULL )
            {
                signal->init();
            }
        }
    }
}


bool PGraphicsSignalView::event( QEvent *event )
{
    if ( event->type() == QEvent::ToolTip )
    {
        QHelpEvent *    helpEvent = static_cast<QHelpEvent *>( event );

        QGraphicsItem * item = itemAt( helpEvent->pos() );
        QString         text;

        if ( item )
        {
            const PSignalPlotBase * signal  = dynamic_cast<PSignalPlotBase *>( item );
            QPointF             scenePoint  = mapToScene( helpEvent->pos() );
            QPointF             itemPoint   = item->mapFromScene( scenePoint );
            QPointF             foundtimeAmpl;

            if ( signal != NULL )
            {
                foundtimeAmpl = const_cast<PSignalPlotBase *>(signal)->toTimeAmplitude( itemPoint );

                text = signal->getToolTipText( foundtimeAmpl.x() );
//qDebug() << "PGraphicsSignalView::event:";
//qDebug() << "eventpoint - " << helpEvent->pos();
//qDebug() << "scene point - " << scenePoint;
//qDebug() << "item point - " << itemPoint;
//qDebug() << "data point - " << foundtimeAmpl;
//qDebug() << "data text - " << text;
            }
        }

        if ( text.isEmpty() )
        {
            QToolTip::hideText();

            event->ignore();
        }
        else
        {
            QToolTip::showText( helpEvent->globalPos(), text );
        }

        return true;
    }
    else
    {
        return QGraphicsView::event( event );
    }
}


void PGraphicsSignalView::resizeEvent( QResizeEvent * event )
{
    if ( event->size().isValid() )
    {
        positionItems();
    }

    QGraphicsView::resizeEvent( event );
}


//void PGraphicsSignalView::showEvent( QShowEvent * event )
//{
//    //if ( !event->spontaneous() )
//    {
//        positionItems();
//    }

//    QGraphicsView::showEvent( event );
//}


void PGraphicsSignalView::setBackGroundColor( const QColor & color )
{
    if ( color.isValid() && mBackGroundColor != color )
    {
        mBackGroundColor = color;

        if ( !mIsPrinting )
        {
            setBackgroundBrush( QBrush( mBackGroundColor ) );

            invalidateScene();
        }
    }
}


bool PGraphicsSignalView::setGrid( Widget::Grid_type grid )
{
    bool changed = false;

    if ( scene() )
    {
        QList<QGraphicsItem*>   scene_items( scene()->items() );

        for ( int i = 0; i < scene_items.size() ; i++ )
        {
            PSignalPlotBase * signal = dynamic_cast<PSignalPlotBase *>( scene_items[i] );

            if ( signal != NULL )
            {
                changed |= signal->setGrid( grid );
            }
        }
    }

    return changed;
}


bool PGraphicsSignalView::setFonts( const QFont & textFont, const QFont & labelFont, const QFont & scaleFont )
{
    bool changed = false;

    if ( scene() )
    {
        QList<QGraphicsItem*>   scene_items( scene()->items() );

        for ( int i = 0; i < scene_items.size() ; i++ )
        {
            PSignalPlotBase * signal = dynamic_cast<PSignalPlotBase *>( scene_items[i] );

            if ( signal != NULL )
            {
                changed |= signal->setLabelFont( labelFont );
                changed |= signal->setScaleFont( scaleFont );
                changed |= signal->setTextFont( textFont );
            }
            else
            {
                PSimpleTextItem * txtItem = qgraphicsitem_cast<PSimpleTextItem *>( scene_items[i] );

                if ( (txtItem != NULL) && (txtItem->font() != textFont) )
                {
                    changed = true;

                    txtItem->setFont( textFont );
                }
            }
        }
    }

    return changed;
}


void PGraphicsSignalView::updateInfo( void )
{
    if ( scene() )
    {
        bool                    rescale = false;
        QList<QGraphicsItem*>   scene_items( scene()->items() );

        for ( int i = 0; i < scene_items.size() ; i++ )
        {
            PSignalPlotBase *signal = dynamic_cast<PSignalPlotBase *>( scene_items[i] );

            if ( signal != NULL )
            {
                PSignalDraw * dynSignalPtr = dynamic_cast<PSignalDraw *>( signal );

                // First update the signal
                if ( signal->updateSignals() )
                {
                    rescale = true;

                    signal->resetSignalsZoomed( true );
                }

                // Then, correct the administration of the dynamic signal, i.e. remove obsolete data, correct for
                // data display when getting behind with drawing of new data, etc.
                if ( dynSignalPtr )
                {
                    dynSignalPtr->correctAdministrationForNewData();
                }
            }
        }

        if ( rescale )
        {
            PGrSceneBase * sceneBasePtr = static_cast<PGrSceneBase *>( scene() );

            if ( NULL != sceneBasePtr )
            {
                sceneBasePtr->autoScale();
            }
            else
            {
                Q_ASSERT( false );
            }
        }
    }
}


void PGraphicsSignalView::updateSceneForeGround( void )
{
    if ( scene() )
    {
        invalidateScene( scene()->sceneRect(), QGraphicsScene::ItemLayer /*ForegroundLayer*/ );
    }
}


void PGraphicsSignalView::updateSceneBackGround( void )
{
    invalidateScene();
}


void PGraphicsSignalView::removeMyItems( void )
{
    if ( scene() )
    {
        if ( !scene()->items().isEmpty() )
        {
            QList<QGraphicsItem *> scene_items = scene()->items();
            QList<QGraphicsItem *> remove_items;

            for ( int i = 0; i < scene_items.size(); ++i )
            {
                // only [re]move top level items. all others are
                // being taken care of by their respective parent
                if (scene_items[i]->parentItem() == 0)
                {
                    scene()->removeItem(scene_items[i]);
                    remove_items << scene_items[i];
                }
            }

            // finally delete them
            for ( int i = 0; i < remove_items.size(); ++i )
            {
                delete remove_items[i];
            }
        }
    }
}


QGraphicsItem * PGraphicsSignalView::ItemAt( int column, int row ) const
{
    QGraphicsItem * item = NULL;

    if ( scene() )
    {
        QList<QGraphicsItem *>  allItems( scene()->items() );

        for ( int index = 0; (index < allItems.count()) && (item == NULL); index++ )
        {
            QGraphicsItem * itemPtr         = allItems.at( index );
            PGrItemBase*    pItemBasePtr    = dynamic_cast<PGrItemBase*>( itemPtr );

            Q_ASSERT( pItemBasePtr != NULL );

            if ( pItemBasePtr && (pItemBasePtr->getColumn() == column) && (pItemBasePtr->getRow() == row) )
            {
                item = itemPtr;
            }
        }
    }

    return item;
}


void PGraphicsSignalView::positionItems( void )
{
    if ( scene() && !scene()->items().isEmpty() )
    {
        PGrSceneBase * sceneBasePtr = static_cast<PGrSceneBase *>( scene() );

        if ( NULL != sceneBasePtr )
        {
            QSize   vpSize;
            QRectF  theSceneRect;

            sceneBasePtr->updateSceneLayout();

            vpSize          = maximumViewportSize();
            theSceneRect    = sceneRect();

            if ( vpSize.isValid() && theSceneRect.isValid() )
            {
                double scaleW   = double( vpSize.width() ) / theSceneRect.width();
                double scaleH   = double( vpSize.height() ) / theSceneRect.height();

                if ( scaleH <= scaleW )
                {
                    setMatrix( QMatrix( scaleH, 0, 0, scaleH, 0, 0 ) );
                }
                else
                {
                    setMatrix( QMatrix( scaleW, 0, 0, scaleW, 0, 0 ) );
                }

                afterSceneRectSet();    // Notify derived classes of the new size of the scene

                redrawBackground();
            }
        }
    }
}


void PGraphicsSignalView::copyToClipBoard( void )
{
    QPixmap         pixmap( width(), height() );
    QClipboard *    cb = QApplication::clipboard();

    setPrinting( true );

    cb->setPixmap( pixmap.grabWidget( this ), QClipboard::Clipboard );

    setPrinting( false );
}


void PGraphicsSignalView::setPrinting( bool doPrint )
{
    QList<QGraphicsItem *> scene_items( scene()->items() );

    mIsPrinting = doPrint;

    if ( doPrint )
    {
        setBackgroundBrush( QBrush( mBackGroundColorPrint ) );
    }
    else if ( mBackGroundColor.isValid() )
    {
        setBackgroundBrush( QBrush( mBackGroundColor ) );
    }

    for ( int i = 0; i < scene_items.size(); i++ )
    {
        PSignalPlotBase * signal = dynamic_cast<PSignalPlotBase *>( scene_items[i] );

        if ( signal != NULL )
        {
            signal->setPrinting( doPrint );
        }
    }
}


void PGraphicsSignalView::paintEvent( QPaintEvent * event )
{
    if ( !scene() && mBackGroundColor.isValid() )
    {
        // If no scene if given yet, paint at least the background in the specified color
        QPainter painter( viewport() );

        painter.fillRect( event->rect(), mBackGroundColor );
    }
    else
    {
        QGraphicsView::paintEvent( event );
    }
}


void PGraphicsSignalView::drawForeground( QPainter * painter, const QRectF & r )
{
    // Call the base implementation first to fill the background image with the specified brush
    QGraphicsView::drawForeground( painter, r );

    if ( scene() )
    {
        // Find all items that need to be updated
        QList<QGraphicsItem *>  itemList    = scene()->items( r, Qt::IntersectsItemBoundingRect );
        PSignalPlotBase *       signalPtr   = NULL;
        QRectF                  itemBoundRect;

//    qDebug() << "*** New background draw **********************************************************";
//    qDebug() << "drawForeground";
//    qDebug() << "given rect:/t" << r;
//    qDebug() << "clipBoundingRect:/t" << painter->clipBoundingRect();
//    qDebug() << "viewport rect:/t" << painter->viewport();
//    qDebug() << "window rect:/t" << painter->window();
//    qDebug() << "Scene rect:/t" << scene()->sceneRect();

        // Paint the background of all items
        for ( int i = 0; i < itemList.size() ; i++ )
        {
            signalPtr = dynamic_cast<PSignalPlotBase *>( itemList[i] );

            if ( signalPtr != NULL )
            {
                itemBoundRect = signalPtr->sceneBoundingRect();

                if ( r.intersects( itemBoundRect ) )
                {
                    itemBoundRect = r.intersected( itemBoundRect );

                    signalPtr->paintForeGround( painter, itemBoundRect );
                }
            }
        }
    }
    else
    {
        // Fill the view with the given background color
        painter->fillRect( r, mBackGroundColor );
    }
}


void PGraphicsSignalView::drawBackground( QPainter * painter, const QRectF & r )
{
    // Call the base implementation first to fill the background image with the specified brush
    QGraphicsView::drawBackground( painter, r );

    if ( scene() )
    {
        // Find all items that need to be updated
        QList<QGraphicsItem *>  itemList    = scene()->items( r, Qt::IntersectsItemBoundingRect );
        PSignalPlotBase *       signalPtr   = NULL;
        QRectF                  itemBoundRect;

//    qDebug() << "*** New background draw **********************************************************";
//    qDebug() << "drawBackground";
//    qDebug() << "given rect:/t" << r;
//    qDebug() << "clipBoundingRect:/t" << painter->clipBoundingRect();
//    qDebug() << "viewport rect:/t" << painter->viewport();
//    qDebug() << "window rect:/t" << painter->window();
//    qDebug() << "Scene rect:/t" << scene()->sceneRect();

        // Paint the background of all items
        for ( int i = 0; i < itemList.size() ; i++ )
        {
            signalPtr = dynamic_cast<PSignalPlotBase *>( itemList[i] );

            if ( signalPtr != NULL )
            {
                itemBoundRect = signalPtr->sceneBoundingRect();

                if ( r.intersects( itemBoundRect ) )
                {
                    itemBoundRect = r.intersected( itemBoundRect );

                    signalPtr->paintGrid( painter, itemBoundRect );
                }
            }
        }
    }
    else
    {
        // Fill the view with the given background color
        painter->fillRect( r, mBackGroundColor );
    }
}


void PGraphicsSignalView::redrawBackground( void )
{
//qDebug() << "redrawBackground called";
    if ( cacheMode() == CacheBackground )
    {
        // Calling this function thiggers a full redraw
        resetCachedContent();
    }
    else
    {
        update();
    }
}
