/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PGraphicsSignalView.h 1147 2014-03-08 16:52:00Z eelco $
 *
 */


#ifndef __PGRAPHICSSIGNALVIEW_H__
#define	__PGRAPHICSSIGNALVIEW_H__


#include <QGraphicsView>
#include <QtGlobal>
#include <QPixmap>

#include "PeacsSupport_global.h"
#include "PWidgetDefines.h"
#include "PGrItemBase.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PGraphicsSignalView PGraphicsSignalView.h "widget/PGraphicsSignalView.h"
          * @brief The view window in which the signal graphs are being displayed
          */
        class PEACSSUPPORTSHARED_EXPORT PGraphicsSignalView : public QGraphicsView
        {
            Q_OBJECT

        public:


            /** @brief The constructor
              *
              * @param parent the parent of this view
              */
            PGraphicsSignalView( QWidget *parent = 0 );


            /** @brief The denstructor
              */
            virtual ~PGraphicsSignalView();


            /**
              */
            //void init();


            /** @brief Set up the view by adding all items
              * This function must be called after construction of this object when all data to
              * create the items in the view is present. This function will call the pure virtual
              * function initialyzeItems() that is responcible for adding the items.
              * After that this function will initialyze the view by setting the bounds of the
              * individual items and position them in the view.
              */
            void initialize( void );


            /** @brief Used to set mBackGroundColor to a (new) value
              * Default mBackGroundColor is initialized with no color specified, i.e. invalid. In
              * that case no background color will be specified for this view
              * By specifying a (new) color that color will be used to fill the background
              * of the this view
              * @note Use this function in stead of QGraphicsView::setBackgroundBrush(). When using
              * that function the background function will be lost after a copy to clipboard action
              *
              * @param color the (new) color to use
              */
            void setBackGroundColor( const QColor & color );


            /** @brief Set mBackGroundColorPrint to the specified color
              * No updating is needed because this coloris only used during printing or copy to clipboard. Default
              * color is white
              *
              * @param color the (new) color to use
              */
            inline void setBackGroundColorForPrinting( const QColor & color ) { mBackGroundColorPrint = color; }


            /** Sets the grid member for all scene items to the specified grid setting
              *
              * Checks if the grid setting has changed and sets redraw background accordingly and signals
              * for an update
              *
              * @param grid the new grid setting to use
              *
              * @return true if the setting has changed
              */
            bool setGrid( Widget::Grid_type grid );


            /** Sets all font members for all scene items to the specified fonts
              *
              * Checks if the fonts have changed and sets redraw background accordingly and signals
              * for an update
              *
              * @param textFont the new font to use as text font
              * @param labelFont the new font to use as label font
              * @param scaleFont the new font to use as scale font
              *
              * @return true if the value has changed
              */
            bool setFonts( const QFont & textFont, const QFont & labelFont, const QFont & scaleFont );


            /** @brief Returns the value of mIsPrinting
              * Returns true if this view is currently in printing mode
              *
              * @return mIsPrinting
              */
            inline bool isPrinting( void ) const { return mIsPrinting; }


            /** @brief returns the item at [column, row]
              * This function searches all items until the one at the specified [column, row]
              * is found and returns this one. NULL is returned if none is found
              *
              * @param column the column to search
              * @param row the tow to search
              *
              * @return the itme at the specified location or NULL if not found
              */
            QGraphicsItem * ItemAt( int column, int row ) const;


            /** @brief Must be called to initialize this view with its items.
              * By calling this function the init() functions of all items contained
              * by this graphics view will be called. Thus initialization can be done
              * with any first data present
              * Implement this function in a derived class to add application specific
              * initialisation. Always call this base implementation
              */
            virtual void initializeView( void );


        public slots:

            /** @brief Must be called to update the information shown in all signal items
              * All items contained by this graphics view will be told to update the data
              */
            void updateInfo( void );


            /** @brief Copies the content of this view to the clipboard
              * Custom colors can be set and will be used for all items. Markers will not be
              * contained by the clipboard image, unless otherwise implemented
              */
            void copyToClipBoard( void );


            /** @brief Function that can be called to refresh this views content except the background
              * This implementation calles the invalidateScene() function forcing a redraw
              * of the view content and foreground. The background image is not redrawn
              * Implement in a derived class to add neccesary functionality to the update
              */
            virtual void updateSceneForeGround( void );


            /** @brief Function that can be called to refresh this views background image
              * This implementation calles the invalidateScene() function forcing a redraw
              * of the whole view content
              * Implement in a derived class to add neccesary functionality to the update
              */
            virtual void updateSceneBackGround( void );


        protected:

            /** @brief Implement this function to add the items to this view
              * You have to implement this function yourself in a derived class. In it
              * you must add all items you desire in the view.
              * This function will be called by the function initialize().
              */
            virtual void initialyzeItems( void ) = 0;


            /** @brief Is used to track events. Here it tracks tooltip events
              * This function will try to retrieve the tooltip text at the give event position
              * if there is a tooltip event. If a text is found, the tooltip is shown with this
              * text. Otherwise the tooltip text is cleared
              *
              * @param event General information about the event occurred
              */
            bool event( QEvent * event );


            /** @brief This event handler receives resize events
              * Makes sure the items are scalled appropriately
              *
              * @param event points to information about the mouse event
              */
            virtual void resizeEvent( QResizeEvent * event );


//            /** @brief This event handler can be reimplemented in a subclass
//              * to receive show events (passed in event), for the viewport() widget.
//              * Makes sure the items are scalle appropriately
//              * @param event the show event information
//              */
//            virtual void showEvent( QShowEvent * event );


            /** @brief This function is called after the rectangle of this scene has been set
              * This function determains the moment that the size of the scene is definitaly set and
              * information related to it can be used. This is for example the pixelsize.
              *
              * This function is triggered by a resize event after all items are positioned and theire
              * bounding rectangles are set by the function setItemPositions().
              *
              * Implement this function in a derived class to perform some scene size depending
              * initialization that is needed after a resize
              */
            virtual void afterSceneRectSet( void ) {;}


            /** @brief Removes all items shown in this view
              */
            virtual void removeMyItems( void );


            /** @brief Is called when the view must paint itself
              *
              * @param event Informamtion needed when painging
              */
            void paintEvent( QPaintEvent * event );


            /** @brief Draws the markers and zoom rectangle on top of the signal items
              * Reimplementation of the QGraphicsView::drawForeground(QPainter *, const QRectF &)
              * All painting is done in scene coordinates. rect is the exposed rectangle.
              *
              * @param painter see Qt documantation
              * @param r see Qt documantation QRectF
              */
            virtual void drawForeground( QPainter * painter, const QRectF & r );


            /** @brief Draw the grid and all related information around the graph like texts and scale info
              * Reimplementation of the QGraphicsView
              * All painting is done in scene coordinates. rect is the exposed rectangle.
              * @param painter see Qt documantation
              * @param r see Qt documantation    QRectF
              */
            virtual void drawBackground ( QPainter * painter, const QRectF & r );


            /** @brief Makes the background redrawn if needed
              * Check's if the background must be redrawn (i.e. VCG always when
              * calling this function), otherwise only when forceRedraw is true
              * @param forceRedraw is true then the background is redrawn always (default true)
              */
            void redrawBackground( void );


            /** @brief Sets the correct plotting/printing colors and the corresponding state of all item
              * In printing mode the printing colors, pens or brushes are used in stead of the normal painting ones.
              * This function is called by the copyToClipBoard() function
              * Override this function in an application specific implementation to perform application specific
              * adjustments, but always call this base implementation
              *
              * @param doPrint determines if the view is printed
              */
            virtual void setPrinting( bool doPrint = false );


        private:

            /** @brief This function will initialyze/set the bounding rectangle of all items and position them
              * This function will be called after all items haven been added to the scene and
              * when the view is resized
              */
            void positionItems( void );


            /** @brief used to specify the default background color within the view
              */
            QColor mBackGroundColor;


            /** @brief used to specify the background color within the view when printing
              */
            QColor mBackGroundColorPrint;


            /** @brief Flag that indicates if the printing colors should be used or the 'normal' screen paint colors
              * If this flag is true, the printing colors, pens, brushes will be used, otherwise the default ones
              */
            bool mIsPrinting;
        };
    }
}
#endif	/* __PGRAPHICSSIGNALVIEW_H__ */

