/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PMarkerEllipse.h 1113 2014-02-19 21:03:28Z eelco $
 *
 */

#ifndef PMARKERECLIPSE_H
#define PMARKERECLIPSE_H


#include "PeacsSupport_global.h"
#include "PSignalMarker.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PMarkerEllipse PMarkerEllipse.h "widgets/PMarkerEllipse.h"
          * @brief Is a specialisation of the class PSignalMarker that creates a line marker
          */
        class PEACSSUPPORTSHARED_EXPORT PMarkerEllipse : public PSignalMarker
        {

        public:

            /** @brief Only present to be used in the type() function
              */
            enum { Type = 2 };


            /** @brief Returns the type of marker.
              * Each derived class, representing a different kind of marker, should specify a unique
              * marker type identifier.
              * Enable the use of marker_cast with this item.
              */
            virtual int type( void ) const { return Type; }


            /** @brief Default constructor
              * This constructor should only be used when claiming object during some sort
              * of initialization. Normally not to be used.
              */
            PMarkerEllipse();


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              */
            PMarkerEllipse( int id, const QString & label = QString() );


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              * @param show flag indicating if the marker must be shown initially
              * @param pen the QPen object to use when drawing the marker
              */
            PMarkerEllipse( int id, const QString & label, bool show, const QPen & pen );


            /** @brief Copy constructor
              *
              * @param source the source to copy from
              */
            PMarkerEllipse( const PMarkerEllipse & source );


            /** @brief Destructor
              */
            ~PMarkerEllipse( void ) {;}


            /** @brief Asignment operator
              *
              * @param source the source to copy from
              */
            virtual const PMarkerEllipse & operator =( const PMarkerEllipse & source );


            /** @brief Sets the radius of the elipse in X and Y direction
              *
              * @param rx The radius in X direction to use
              * @param ry The radius in Y direction to use. If the default value is used (i.e. 0.0) rx is also used for mRy
              */
            inline void setRadii( qreal rx, qreal ry = 0.0 ) { mRx = rx; (ry == 0.0 ? mRy = mRx : mRy = ry); }


            /** @brief Sets the center point of the eclipse in the local maintained member
              * Note that this point should be specified in data coordinate system. During
              * drawing of the eclipse this coordinate is translated into graph coordinate
              * system automatically
              * This function also clears the graph representing eclipse data forcing a new transformation
              *
              * @param center The center point of the new eclipse painterpath
              */
            inline void setMarkerData( const QPointF & center ) { mCenter = center; mMarker = QPainterPath(); }


            /** @brief Added for convenience.
              * @see setMarkerData( const QPointF & first, const QPointF & last )
              * Note that this point should be specified in data coordinate system. During
              * drawing of the eclipse this coordinate is translated into graph coordinate
              * system automatically
              * This function also clears the graph representing eclipse data forcing a new transformation
              *
              * @param center_x The X-position of the center point of the new eclipse painterpath
              * @param center_y The Y-position of the center point of the new eclipse painterpath
              */
            inline void setMarkerData( double center_x, double center_y )  { mCenter = QPointF( center_x, center_y ); mMarker = QPainterPath(); }


            /** @brief This function will be called to translate the internal stored points into a line object in graph coordinates
              * Before the transformation will be performed, there will be checked if the local stored object is still vallid
              *
              * @param translation The transformation object used to transform from data coordinates into graph coordinates
              * @param forceTransformation if true the transformation will be performed without checking the local stored object
              *
              * @return the resulting line object stored in a painterpath
              */
            const QPainterPath & translate( const QTransform & translation, bool forceTransformation = false );


            /** @brief This function will be called to translate the internal stored points into a line object in graph coordinates
              * This version of the function will always perform the transformation without storing the result
              *
              * @param translation The transformation object used to transform from data coordinates into graph coordinates
              *
              * @return the resulting line object stored in a painterpath
              */
            QPainterPath translated( const QTransform & translation );


            /** @brief Checks if there is any data stored to be shown
              * This depends on data sotred in derived classes. The mMarker is only used to store trandsformed data
              * for convenience for speedup the display of the marker
              *
              * @return true if there is data to be shown
              */
            bool isEmpty( void ) const;


            /** @brief Clears all local data and calls the base class implementation
              * @see PSignalMarker::clear()
              */
            void clear( void );


            /** @brief Checks if this marker is at the specified position
              * This class will always return false. Real implementation must be done in derived classes
              *
              * @param pos The position to look at. This position should be in data coordinate system
              * @param range A positive value indicating the allowed distance range of the position to the eclipse marker
              *
              * @return true if this marker is at the position
              */
            bool isMarkerAtPos( const QPointF & pos, double range = 1.0 ) const;


        private:

            /** @brief Represents the radius of the elipse in X-axis direction
              */
            qreal mRx;


            /** @brief Represents the radius of the elipse in Y-axis direction
              */
            qreal mRy;


            /** @brief The center point of the elipse in data coordinate system
              * This point will be translated later into a graph coordinate system point
              */
            QPointF mCenter;
        };
    }
}

#endif // PMARKERECLIPSE_H
