/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PMarkerLine.h 1131 2014-02-28 12:11:29Z eelco $
 *
 */

#ifndef PMARKERLINE_H
#define PMARKERLINE_H


#include "PeacsSupport_global.h"
#include "PSignalMarker.h"

#include <QPointF>


namespace Peacs
{
    namespace Support
    {
        /** @class PMarkerLine PMarkerLine.h "widgets/PMarkerLine.h"
          * @brief Is a specialisation of the class PSignalMarker that creates a line marker
          * Do not derive from this class. Create an own specialisation
          */
        class PEACSSUPPORTSHARED_EXPORT PMarkerLine : public PSignalMarker
        {

        public:

            /** @brief Only present to be used in the type() function
              */
            enum { Type = 1 };


            /** @brief Returns the type of marker.
              * Each derived class, representing a different kind of marker, should specify a unique
              * marker type identifier.
              * Enable the use of marker_cast with this item.
              */
            int type( void ) const { return Type; }


            /** @brief Default constructor
              * This constructor should only be used when claiming object during some sort
              * of initialization. Normally not to be used.
              */
            PMarkerLine();


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              */
            PMarkerLine( int id, const QString & label = QString() );


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              * @param show flag indicating if the marker must be shown initially
              * @param pen the QPen object to use when drawing the marker
              */
            PMarkerLine( int id, const QString & label, bool show, const QPen & pen );


            /** @brief Destructor
              */
            ~PMarkerLine( void ) {;}


            /** @brief Copy constructor
              *
              * @param source the source to copy from
              */
            PMarkerLine( const PMarkerLine & source );


            /** @brief Sets the two line points in the local maintained members
              * Note that these points should be specified in data coordinate system. During
              * drawing of the line these coordinates are translated into graph coordinate
              * system automatically
              * This function also clears the graph representing line data forcing a new transformation
              *
              * @param first The point to start the line at
              * @param last The point to draw the line to
              */
            inline void setMarkerData( const QPointF & first, const QPointF & last ) { mFirst = first; mLast = last; mMarker = QPainterPath(); }


            /** @brief Added for convenience.
              * @see setMarkerData( const QPointF & first, const QPointF & last )
              * Note that these values should be specified in data coordinate system. During
              * drawing of the line these coordinates are translated into graph coordinate
              * system automatically
              * This function also clears the graph representing line data forcing a new transformation
              *
              * @param first_x The X-position of the point to start the line at
              * @param first_y The Y-position of the point to start the line at
              * @param last_x The X-position of the point to draw the line to
              * @param last_y The Y-position of the point to draw the line to
              */
            inline void setMarkerData( double first_x, double first_y, double last_x, double last_y )  { mFirst     = QPointF( first_x, first_y );
                                                                                                         mLast      = QPointF( last_x, last_y );
                                                                                                         mMarker    = QPainterPath(); }


            /** @brief Returns the X-axis value of the first point of this line marker object
              *
              * @return X-value of first point
              */
            inline qreal getMarkerXDataP1( void ) const { return mFirst.x(); }


            /** @brief Returns the X-axis value of the second point of this line marker object
              *
              * @return X-value of second point
              */
            inline qreal getMarkerXDataP2( void ) const { return mLast.x(); }


            /** @brief Returns the Y-axis value of the first point of this line marker object
              *
              * @return Y-value of first point
              */
            inline qreal getMarkerYDataP1( void ) const { return mFirst.y(); }


            /** @brief Returns the Y-axis value of the second point of this line marker object
              *
              * @return Y-value of second point
              */
            inline qreal getMarkerYDataP2( void ) const { return mLast.y(); }


            /** @brief Adjusts the size of this line object according the given minimum/maximum values
              * The given minimum and maximum values define the bordens for this line object
              *
              * @param minX The minimum X-axis extent value
              * @param maxX The maximum X-axis extent value
              * @param minY The minimum Y-axis extent value
              * @param maxY The maximum Y-axis extent value
              * @return True if the line object has been adjusted
              */
            bool adjustForNewExtent( qreal minX, qreal maxX, qreal minY, qreal maxY );


            /** @brief This function will be called to translate the internal stored points into a line object in graph coordinates
              * Before the transformation will be performed, there will be checked if the local stored object is still vallid
              *
              * @param translation The transformation object used to transform from data coordinates into graph coordinates
              * @param forceTransformation if true the transformation will be performed without checking the local stored object
              *
              * @return the resulting line object stored in a painterpath
              */
            const QPainterPath & translate( const QTransform & translation, bool forceTransformation = false );


            /** @brief This function will be called to translate the internal stored points into a line object in graph coordinates
              * This version of the function will always perform the transformation without storing the result
              *
              * @param translation The transformation object used to transform from data coordinates into graph coordinates
              *
              * @return the resulting line object stored in a painterpath
              */
            QPainterPath translated( const QTransform & translation );


            /** @brief Checks if there is any data stored to be shown
              * This depends on data sotred in derived classes. The mMarker is only used to store trandsformed data
              * for convenience for speedup the display of the marker
              *
              * @return true if there is data to be shown
              */
            bool isEmpty( void ) const;


            /** @brief Clears all local data and calls the base class implementation
              * @see PSignalMarker::clear()
              */
            void clear( void );


            /** @brief Checks if this marker is at the specified position
              * This class will always return false. Real implementation must be done in derived classes
              *
              * @param pos The position to look at. This position should be in data coordinate system
              * @param range A positive value indicating the allowed distance range of the position to the line marker
              *
              * @return true if this marker is at the position
              */
            bool isMarkerAtPos( const QPointF & pos, double range = 1.0 ) const;


        private:

            /** @brief The first point of the line in data coordinate system
              * These points will be translated later in graph coordinate system points
              */
            QPointF mFirst;


            /** @brief The second (last) point of the line in data coordinate system
              */
            QPointF mLast;
        };
    }
}

#endif // PMARKERLINE_H
