/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2014 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalData.cpp 1177 2014-03-21 16:10:38Z eelco $
 *
 */


#include "PSignalData.h"

//#include <QDebug>


using namespace Peacs::Support;

int UNINITIALYZED_SIGNAL_DATA = INT_MAX;


PSignalData::PSignalData( void ) :
    mId( UNINITIALYZED_SIGNAL_DATA-- ), // Make sure all signal data object can be uniquly identified
    mLabel(),
    mShow( false ),
    mPen(),
    mPenPrint( Qt::NoPen ),
    mBrush(),
    mBrushPrint(),
    mCorrectionX( 1.0 ),
    mCorrectionY( 1.0 ),
    mXData(),
    mYData(),
    mMinXData( 0.0 ),
    mMaxXData( 0.0 ),
    mMinYData( 0.0 ),
    mMaxYData( 0.0 ),
    mDataPoly()
{
}


PSignalData::PSignalData( int id, const QString & label, double correctionX, double correctionY ) :
    mId( id ),
    mLabel( label ),
    mShow( false ),
    mPen(),
    mPenPrint( Qt::NoPen ),
    mBrush(),
    mBrushPrint(),
    mCorrectionX( correctionX ),
    mCorrectionY( correctionY ),
    mXData(),
    mYData(),
    mMinXData( 0.0 ),
    mMaxXData( 0.0 ),
    mMinYData( 0.0 ),
    mMaxYData( 0.0 ),
    mDataPoly()
{
}


PSignalData::PSignalData( int id, const QString & label, bool show, const QPen & pen, double correctionX, double correctionY ) :
    mId( id ),
    mLabel( label ),
    mShow( show ),
    mPen( pen ),
    mPenPrint( Qt::NoPen ),
    mBrush(),
    mBrushPrint(),
    mCorrectionX( correctionX ),
    mCorrectionY( correctionY ),
    mXData(),
    mYData(),
    mMinXData( 0.0 ),
    mMaxXData( 0.0 ),
    mMinYData( 0.0 ),
    mMaxYData( 0.0 ),
    mDataPoly()
{
}


PSignalData::PSignalData( const PSignalData & source ) :
    mId( source.mId ),
    mLabel( source.mLabel ),
    mShow( source.mShow ),
    mPen( source.mPen ),
    mPenPrint( source.mPenPrint ),
    mBrush( source.mBrush ),
    mBrushPrint( source.mBrushPrint ),
    mCorrectionX( source.mCorrectionX ),
    mCorrectionY( source.mCorrectionY ),
    mXData( source.mXData ),
    mYData( source.mYData ),
    mMinXData( source.mMinXData ),
    mMaxXData( source.mMaxXData ),
    mMinYData( source.mMinYData ),
    mMaxYData( source.mMaxYData ),
    mDataPoly( source.mDataPoly )
{
}


PSignalData & PSignalData::operator =( const PSignalData & source )
{
    if ( this != &source )
    {
        mId                     = source.mId;
        mLabel                  = source.mLabel;
        mShow                   = source.mShow;
        mPen                    = source.mPen;
        mPenPrint               = source.mPenPrint;
        mBrush                  = source.mBrush;
        mBrushPrint             = source.mBrushPrint;
        mCorrectionX            = source.mCorrectionX;
        mCorrectionY            = source.mCorrectionY;
        mXData                  = source.mXData;
        mYData                  = source.mYData;
        mMinXData               = source.mMinXData;
        mMaxXData               = source.mMaxXData;
        mMinYData               = source.mMinYData;
        mMaxYData               = source.mMaxYData;
        mDataPoly               = source.mDataPoly;
    }

    return *this;
}


bool PSignalData::operator ==( const PSignalData & source )
{
    return (mId == source.mId) && (mLabel == source.mLabel);
}


bool PSignalData::operator !=( const PSignalData & source )
{
    return (mId != source.mId) || (mLabel != source.mLabel);
}


QPointF PSignalData::getXYData( int index ) const
{
    Q_ASSERT( (index < mYData.N()) &&
              (mXData.isEmpty() || index < mXData.N()) );

    if ( mXData.isEmpty() )
    {
        return QPointF( qreal(index), mYData[index] );
    }
    else
    {
        return QPointF( mXData[index], mYData[index] );
    }
}


void PSignalData::fillXDataForY( double delta, double start )
{
    if ( mYData.isEmpty() )
    {
        mXData.clear();     // For convenience
        clearPoly();
    }
    else if ( mYData.N() > mXData.N() )
    {
        int index = 0;

        // If mXData is not empty but should become bigger, we start off with the last value
        if ( 0 < mXData.N() )
        {
            index = mXData.N();
            start = mXData.last() + delta;
        }

        mXData.resize( mYData.N() );

        double * xData = mXData.data();

        for ( ; index < mXData.N(); index++ )
        {
            xData[ index ] = start;

            start += delta;
        }
    }
}


bool PSignalData::CreateOrUpdatePolygon( const QTransform & transformation, double pixelSize, uint startIndex, uint stopIndex )
{
    bool changed = false;

    if ( count() > 0 && (startIndex < (uint)count()) &&                 // Only draw when start index is within data range, otherwise no data present
         ((stopIndex == UINT_MAX) || (startIndex < stopIndex + 1)) )    // At least two points are needed to construct a polygon
    {
        if ( mDataPoly.isEmpty() )
        {
            Q_ASSERT( mXData.isEmpty() || mXData.size() == mYData.size() );

            if ( startIndex >= (uint)count() )
            {
                startIndex = 0;
            }

            if ( stopIndex >= (uint)count() )
            {
                stopIndex = count() - 1;    // Point to the last sample
            }

            qreal       diffValY;
            qreal       diffValX;
            qreal       maxDiffValY     = 0;
            qreal       minDiffValY     = 0;
            double *    xData           = mXData.data();
            double *    yData           = mYData.data();
            QPointF     newPolyPoint;
            QPointF     dataPoint( (mXData.isEmpty() ? startIndex : xData[startIndex]) * mCorrectionX,
                                   yData[startIndex] * mCorrectionY );
            QPointF     polyPoint       = transformation.map( dataPoint );
            QPointF     maxVal          = polyPoint;
            QPointF     minVal          = polyPoint;
            QPointF     dataMaxVal      = dataPoint;
            QPointF     dataMinVal      = dataPoint;
            int         indexMaxValY    = startIndex;
            int         indexMinValY    = startIndex;

            changed     = true;

            mMinXData   = dataPoint.x();
            mMaxXData   = mMinXData;
            mMinYData   = dataPoint.y();
            mMaxYData   = mMinYData;

            mDataPoly.append( polyPoint );

            // We need to create the polygon representation of the signal data, but
            // only the data that adds information to the drawing. We test with the
            // pixel size. If the next point is in closer range then the pixel size
            // is, the point is not used. This speeds up drawing
            for ( int index = startIndex + 1; index < (int)stopIndex; index++ )
            {
                dataPoint.setX( (mXData.isEmpty() ? qreal(index) : xData[index]) * mCorrectionX );
                dataPoint.setY( yData[index] * mCorrectionY );

                newPolyPoint    = transformation.map( dataPoint );
                diffValX        = fabs( newPolyPoint.x() - polyPoint.x() );
                diffValY        = ( newPolyPoint.y() - polyPoint.y() );

                if ( diffValY > maxDiffValY )
                {
                    maxDiffValY     = diffValY;
                    maxVal          = newPolyPoint;
                    indexMaxValY    = index;
                    dataMaxVal      = dataPoint;
                }
                else if ( diffValY < minDiffValY )
                {
                    minDiffValY     = diffValY;
                    minVal          = newPolyPoint;
                    indexMinValY    = index;
                    dataMinVal      = dataPoint;
                }

                if ( diffValX >= pixelSize )
                {
                    if ( maxDiffValY > 0 && minDiffValY < 0 )
                    {
                        // first a positive peak negative will be drawn in the next pixel
                        if ( indexMaxValY < indexMinValY )
                        {
                            addPolyPoint( dataMaxVal, maxVal );

                            addPolyPoint( dataMinVal, minVal );
                        }
                        else // first a negative peak the positive will be drawn in the next pixel
                        {
                            addPolyPoint( dataMinVal, minVal );

                            addPolyPoint( dataMaxVal, maxVal );
                        }
                    }
                    else if ( maxDiffValY > 0 ) // draw upstroke value
                    {
                        addPolyPoint( dataMaxVal, maxVal );
                    }
                    else // draw downstroke or same as previous value
                    {
                        addPolyPoint( dataMinVal, minVal );
                    }

                    maxDiffValY = 0;
                    minDiffValY = 0;
                    maxVal      = newPolyPoint;
                    minVal      = maxVal;
                    polyPoint   = maxVal;
                }
            }

            // Always add the last sample to make sure that the whole signal will be present
            int last = stopIndex;

            dataPoint.setX( (mXData.isEmpty() ? qreal(last) : xData[last]) * mCorrectionX );
            dataPoint.setY( yData[last] * mCorrectionY );

            polyPoint = transformation.map( dataPoint );

            addPolyPoint( dataPoint, polyPoint );

//qDebug() << "number of polypoints:" << mLabel << "\t" << mDataPoly.size() << "\t" << mYData.size() << "\t" << pixelSize << "\t" << startIndex << "\t" << stopIndex;
        }
    }

    return changed;
}


void PSignalData::addPolyPoint( const QPointF & dataVal, QPointF polyVal )
{
    if ( mMinXData > dataVal.x() )
    {
        mMinXData = dataVal.x();
    }
    else if ( mMaxXData < dataVal.x() )
    {
        mMaxXData = dataVal.x();
    }

    if ( mMinYData > dataVal.y() )
    {
        mMinYData = dataVal.y();
    }
    else if ( mMaxYData < dataVal.y() )
    {
        mMaxYData = dataVal.y();
    }

    mDataPoly.append( polyVal );
}


bool PSignalData::findMinMaxData( uint startIndex, uint stopIndex )
{
    bool dataPresent = false;

    if ( count() > 0 && (startIndex < (uint)count()) &&                 // Only draw when start index is within data range, otherwise no data present
         ((stopIndex == UINT_MAX) || (startIndex < stopIndex + 1)) )    // At least two points are needed to construct a polygon
    {
        double * xData  = mXData.data();
        double * yData  = mYData.data();

        if ( startIndex >= (uint)count() )
        {
            startIndex = 0;
        }

        if ( stopIndex >= (uint)count() )
        {
            stopIndex = count() - 1;    // Point to the last sample
        }

        QPointF dataPoint( (mXData.isEmpty() ? startIndex : xData[startIndex]) * mCorrectionX,
                           yData[startIndex] * mCorrectionY );

        mMinXData   = dataPoint.x();
        mMaxXData   = mMinXData;
        mMinYData   = dataPoint.y();
        mMaxYData   = mMinYData;

        dataPresent = true;

        // The first data point is used to initialyze. So start with the next data point
        for ( int index = startIndex + 1; index <= (int)stopIndex; index++ )
        {
            dataPoint.setX( (mXData.isEmpty() ? qreal(index) : xData[index]) * mCorrectionX );
            dataPoint.setY( yData[index] * mCorrectionY );

            if ( mMinXData > dataPoint.x() )
            {
                mMinXData = dataPoint.x();
            }
            else if ( mMaxXData < dataPoint.x() )
            {
                mMaxXData = dataPoint.x();
            }

            if ( mMinYData > dataPoint.y() )
            {
                mMinYData = dataPoint.y();
            }
            else if ( mMaxYData < dataPoint.y() )
            {
                mMaxYData = dataPoint.y();
            }
        }
    }

    return dataPresent;
}


bool PSignalData::removeUnusedDataBefore( int beforeIndex )
{
    bool removed = false;

    if ( (0 <= beforeIndex) && (beforeIndex < count()) )
    {
        removed = true;

        mYData.remove( 0, beforeIndex );

        if ( beforeIndex < mXData.count() )
        {
            mXData.remove( 0, beforeIndex );

            Q_ASSERT( mYData.count() == mXData.count() );
        }

        clearPoly();
    }

    return removed;
}
