/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalData.h 1177 2014-03-21 16:10:38Z eelco $
 *
 */


#ifndef PSIGNALDATA_H
#define PSIGNALDATA_H


#include "PeacsSupport_global.h"
#include "pvector.h"

#include <QPen>
#include <QBrush>
#include <QPolygon>
#include <QVector>


namespace Peacs
{
    namespace Support
    {
        /** @class PSignalData PSignalData.h "widgets/PSignalData.h"
          * @brief Holds all data and information about one signal to plot in the signal view
          */
        class PEACSSUPPORTSHARED_EXPORT PSignalData
        {

        public:

            /** @brief Default constructor
              * This constructor should only be used when claiming object during some sort
              * of initialization. Normally not to be used.
              */
            PSignalData( void );


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              * @param correctionX the value to use for mCorrectionX (default 1.0)
              * @param correctionY the value to use for mCorrectionY (default 1.0)
              */
            PSignalData( int id, const QString & label = QString(), double correctionX = 1.0, double correctionY = 1.0 );


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              * @param show flag indicating if the signal must be shown initially
              * @param pen the QPen object to use when drawing the signal.
              * @param correctionX the value to use for mCorrectionX (default 1.0)
              * @param correctionY the value to use for mCorrectionY (default 1.0)
              */
            PSignalData( int id, const QString & label, bool show, const QPen & pen, double correctionX = 1.0, double correctionY = 1.0 );


            /** @brief Copy constructor
              *
              * @param source the source to copy from
              */
            PSignalData( const PSignalData & source );


            /** @brief Asignment operator
              *
              * @param source the source to copy from
              */
            PSignalData & operator =( const PSignalData & source );


            /** Test equal operator
              *
              * @param source the source to test against
              */
            bool operator ==( const PSignalData & source );


            /** Test not equal operator
              *
              * @param source the source to test against
              */
            bool operator !=( const PSignalData & source );


            /** @brief Returns mId
              *
              * @return The identifier of this signal
              */
            inline int getId( void ) const { return mId; }


            /** @brief Sets mId
              * Sets mId to the new value
              *
              * @param id The value for mId
              */
            inline void setId( int id ) { mId = id; }


            /** @brief Returns mLabel
              *
              * @return The label of this signal
              */
            inline const QString & getLabel( void ) const { return mLabel; }


            /** @brief Sets mLabel
              * Sets mLabel to the new value
              *
              * @param label The new value for mLabel
              */
            inline void setLabel( const QString & label ) { mLabel = label; }


            /** @brief Enables of disables this signal data for showing
              *
              * @param show the new show status
              */
            inline void setShow( bool show ) { mShow = show; }


            /** @brief Inverts the show status flag
              */
            inline void toggleShow( void ) { mShow = !mShow; }


            /** @brief returns mShow
              *
              * @return true if this signal data must be shown, otherwise false
              */
            inline bool mustShow( void ) const { return mShow; }


            /** @brief Returns mPen
              *
              * @return The pen that will be used to draw this signal
              */
            inline const QPen & getPen( void ) const { return mPen; }


            /** @brief Returns mPen as a non const object
              * Convenient function that can be used to change the pen object
              *
              * @return The pen as a non const object that will be used to draw this signal
              */
            inline QPen & pen( void ) { return mPen; }


            /** @brief Sets mPen
              * Sets mPen to the new value
              *
              * @param pen The new value for mPen
              */
            inline void setPen( const QPen & pen ) { mPen = pen; }


            /** @brief Initialyzes mPen with a new QPen object with the given color
              *
              * @param color The color to use with the new mPen
              */
            inline void setPen( const QColor & color ) { mPen = QPen( color ); }


            /** @brief Sets the width of mPen
              *
              * @param width The width to use with mPen
              */
            inline void setPenWidth( qreal width ) { mPen.setWidthF( width ); }


            /** @brief Returns mPenPrint
              *
              * @return The pen that will be used to draw this signal when in printing mode
              * @see PSignalPlotBase::mIsPrinting
              */
            inline const QPen & getPrintPen( void ) const { return mPenPrint; }


            /** @brief Returns mPenPrint as a non const object
              * Convenient function that can be used to change the pen object used for printing or screen capturing
              *
              * @return The pen as a non const object that will be used to draw this signal
              * @see PSignalPlotBase::mIsPrinting
              */
            inline QPen & printPen( void ) { return mPenPrint; }


            /** @brief Sets mPenPrint
              * Sets mPen to the new value
              *
              * @param pen The new value for mPenPrint
              * @see PSignalPlotBase::mIsPrinting
              */
            inline void setPrintPen( const QPen & pen ) { mPenPrint = pen; }


            /** @brief Initialyzes mPenPrint with a new QPen object with the given color
              *
              * @param color The color to use with the new mPenPrint
              * @see PSignalPlotBase::mIsPrinting
              */
            inline void setPrintPen( const QColor & color ) { mPenPrint = QPen( color ); }


            /** @brief Sets the width of mPenPrint
              *
              * @param width The width to use with mPenPrint
              * @see PSignalPlotBase::mIsPrinting
              */
            inline void setPrintPenWidth( qreal width ) { mPenPrint.setWidthF( width ); }


            /** @brief Returns mBrush
              *
              * @return The brush that will be used to fill this signals polygon
              */
            inline const QBrush & getBrush( void ) const { return mBrush; }


            /** @brief Returns mBrush as a non const object
              * Convenient function that can be used to change the brush object
              *
              * @return The brush as a non const object that will be used to fill this signals polygon
              */
            inline QBrush & brush( void ) { return mBrush; }


            /** @brief Sets mBrush
              * Sets mBrush to the new value
              *
              * @param brush The new value for mBrush
              */
            inline void setBrush( const QBrush & brush ) { mBrush = brush; }


            /** @brief Initialyzes mBrush with a new QBrush object with the given color
              *
              * @param color The color to use with the new mBrush
              * @param style The brush style to use with the new mBrush
              */
            inline void setBrush( const QColor & color, Qt::BrushStyle style = Qt::SolidPattern )
            { mBrush = QBrush( color, style ); }


            /** @brief Checks if the brush has be set, i.e. its style is unequal to Qt::NoBrush
              *
              * @return true is a brush is defined
              */
            inline bool hasBrush( void ) const { return mBrush.style() != Qt::NoBrush; }


            /** @brief Returns mBrushPrint
              *
              * @return The brush that will be used to fill this signals polygon in printing mode
              * @see PSignalPlotBase::mIsPrinting
              */
            inline const QBrush & getPrintBrush( void ) const { return mBrushPrint; }


            /** @brief Returns mBrushPrint as a non const object
              * Convenient function that can be used to change the brush object used in printing mode
              *
              * @return The brush as a non const object that will be used to fill this signals polygon in printing mode
              * @see PSignalPlotBase::mIsPrinting
              */
            inline QBrush & printBrush( void ) { return mBrushPrint; }


            /** @brief Sets mBrushPrint
              * Sets mBrushPrint to the new value
              *
              * @param brush The new value for mBrushPrint
              * @see PSignalPlotBase::mIsPrinting
              */
            inline void setPrintBrush( const QBrush & brush ) { mBrushPrint = brush; }


            /** @brief Initialyzes mBrushPrint with a new QBrush object with the given color and style
              *
              * @param color The color to use with the new mBrushPrint
              * @param style The brush style to use with the new mBrushPrint
              * @see PSignalPlotBase::mIsPrinting
              */
            inline void setPrintBrush( const QColor & color, Qt::BrushStyle style = Qt::SolidPattern )
            { mBrushPrint = QBrush( color, style ); }


            /** @brief Checks if the brush used in printing mode has be set, i.e. its style is unequal to Qt::NoBrush
              *
              * @return true is the print brush is defined
              * @see PSignalPlotBase::mIsPrinting
              */
            inline bool hasPrintBrush( void ) const { return mBrushPrint.style() != Qt::NoBrush; }


            /** @brief Returns mCorrectionX
              *
              * @return The correction that can be used later to adapt this signal to the wanten unit
              */
            inline double getCorrectionX( void ) const { return mCorrectionX; }


            /** @brief Sets the width of mCorrectionX
              *
              * @param correction The correction to use with mCorrectionX
              */
            inline void setCorrectionX( double correction ) { mCorrectionX = correction; }


            /** @brief Returns mCorrectionY
              *
              * @return The correction that can be used later to adapt this signal to the wanten unit
              */
            inline double getCorrectionY( void ) const { return mCorrectionY; }


            /** @brief Sets the width of mCorrectionY
              *
              * @param correction The correction to use with mCorrectionY
              */
            inline void setCorrectionY( double correction ) { mCorrectionY = correction; }


            /** @brief Returns the data at the specified position on the x-axis if available or index
              * This function returns the data at the specified position in mXData if
              * mXData is not empty. If mXData is empty the index is returned.
              * In the first case, two signals are drawn against each other. In the second
              * case the signal is drawn agains a fixed measurement interval, for example time.
              *
              * @param index the position to return the data from
              *
              * @return the value at the specified position or index
              */
            inline double getXData( int index ) const { Q_ASSERT( mXData.isEmpty() || (0 <= index && index < mXData.N()) );
                                                         return (mXData.isEmpty() ? (double)index : mXData[index]); }


            /** @brief Returns the const reference to the x-axis data object
              *
              * @return the const x-axis data object
              */
            inline const PVector & getXData( void ) const { return mXData; }


            /** @brief Returns a changable reference to the x-axis data object
              * Note that because this object can be used to change the x-axis data, the
              * representing polygon will be cleared in advance of returning this object
              * @return the x-axis data object
              */
            inline PVector & getXData( void ) { clearPoly(); return mXData; }


            /** @brief Sets the data in mXData
              *
              * @param data the data to store in mXData
              */
            inline void setXData( const PVector & data ) { clearPoly(); mXData = data; }


            /** @brief Appends the data in mXData
              *
              * @param data the data to append to mXData
              */
            inline void addToXData( const PVector & data ) { clearPoly(); mXData.append( data ); }


            /** @brief Returns the data at the specified position on the y-axis
              *
              * @param index the position to return the data from
              *
              * @return the value at the specified position
              */
            inline double getYData( int index ) const { Q_ASSERT( 0 <= index && index < mYData.N() );
                                                         return mYData[index]; }


            /** @brief Returns the const reference to the y-axis data object
              *
              * @return the const y-axis data object
              */
            inline const PVector & getYData( void ) const { return mYData; }


            /** @brief Returns a changable reference to the y-axis data object
              * Note that because this object can be used to change the y-axis data, the
              * representing polygon will be cleared in advance of returning this object
              * @return the y-axis data object
              */
            inline PVector & getYData( void ) { clearPoly(); return mYData; }


            /** @brief Sets the data in mYData
              *
              * @param data the data to store in mYData
              */
            inline void setYData( const PVector & data ) { clearPoly(); mYData = data; }


            /** @brief Appends the data in mYData
              *
              * @param data the data to append to mYData
              */
            inline void addToYData( const PVector & data ) { clearPoly(); mYData.append( data ); }


            /** @brief Removes toRemove elements from the beginning of the mYData vector
              *
              * @param toRemove The number of element to remove
              */
            inline void adjustYDataLen( int toRemove ) { Q_ASSERT( 0 <= toRemove && toRemove < mYData.N() );
                                                         clearPoly(); mYData.remove( 0, toRemove ); }


            /** @brief Returns mDataPoly
              *
              * @return The polygon point collection representing the signal data to plot
              */
            inline const QPolygonF & getDataPoly( void ) const { return mDataPoly; }


            /** @brief Returns mDataPoly as a non const object
              * Convenient function that can be used to change the polygon object
              *
              * @return The polygon point collection  as a non const object representing the signal data to plot
              */
            inline QPolygonF & dataPoly( void ) { return mDataPoly; }


            /** @brief Returns the [X,Y] coordinate at the specified position
              * This function returns the y-axis data pared to the corresponding x-axis data
              * The x-axis data can either be a found value in mXData or the asked index
              *
              * @param index the position to return the [X,Y] data from
              *
              * @return The combined [X,Y] value of mXData and mYData
              */
            QPointF getXYData( int index ) const;


            /** @brief Clears all data in mXData, mYData and mDataPoly
              */
            inline void clear( void ) { clearPoly(); mXData.clear(); mYData.clear(); }


            /** @brief Clears all data in mDataPoly
              * This forces recalculation of the drawn polyline. This is usefull when, for example
              * the scale has changed
              */
            inline void clearPoly( void ) { mDataPoly.clear(); }


            /** Returns the number of information points contained by the signal data
              *
              * @return The size of the leading mYData
              */
            inline int count( void ) const { return mYData.N(); }


            /** @brief Returns the minimum data value that must be shown on the x-axis
              *
              * @return mMinXData
              */
            inline double getMinXData( void ) const { return mMinXData; }


            /** @brief Returns the maximum data value that must be shown on the x-axis
              *
              * @return mMaxXData
              */
            inline double getMaxXData( void ) const { return mMaxXData; }


            /** @brief Returns the minimum data value that must be shown on the y-axis
              *
              * @return mMinYData
              */
            inline double getMinYData( void ) const { return mMinYData; }


            /** @brief Returns the maximum data value that must be shown on the y-axis
              *
              * @return mMaxYData
              */
            inline double getMaxYData( void ) const { return mMaxYData; }


            /** @brief Tries to find the minimum and maximum X and Y data values
              * This function wil search the current X and Y data and fills mMinXData,
              * mMaxXData, mMinYData and mMaxYData.
              * This function forces a re-chack. The previous values are neglected
              *
              * @param startIndex The index in the signal data array where to start converting to polygon (default 0)
              * @param stopIndex The index in the signal data array where to stop converting to polygon  (default
              *        UINT_MAX, i.e. up to the end)
              * @return true is these member were updated. False is not, ie. not data present
              */
            bool findMinMaxData( uint startIndex = 0, uint stopIndex = UINT_MAX );


            /** @brief Enables the correction of all datavalues with the factor
              * This function can be used to correct the data maintained in mXData
              * for a specific factor. This can be needed if the unit of the values
              * must be displayed in a different range then they are calculated. For
              * example, calculated in msec. and displayed in sec.
              */
            inline void correctXData( double factor ) { mXData.operator *=( factor ); }


            /** @brief Enables the correction of all datavalues with the factor
              * This function can be used to correct the data maintained in mYData
              * for a specific factor. This can be needed if the unit of the values
              * must be displayed in a different range then they are calculated. For
              * example, calculated in msec. and displayed in sec.
              * mCorrectionY is not being used in this function and not set by it
              */
            inline void correctYData( double factor ) { mYData.operator *=( factor ); }


            /** @brief This function will initialyze or fill mXData when only mYData is set
              * This helper function will fill mXData with values starting with the 'start'
              * value and all subsequent values with sequentially added value 'delta'
              *
              * @param delta the value to add sequentially to each next value
              * @param start the initial value for when mXData is empty
              */
            void fillXDataForY( double delta, double start = 0 );


            /** @brief Creates or updates the polygon given the signal data combined with the specified data
              * The polygon will only be updated if it is empty when calling this function. The polygon will
              * be made empty when a redraw is needed or when new signal data is set.
              *
              * @param transformation The transformation matrix used to transform the signal data into the
              *        wanted polygon coordinates related to the drawing area
              * @param pixelSize Used to filer the signal data; only the data that adds information to the
              *        drawing will be used. Any data in between the pixel size points is neglected
              * @param startIndex The index in the signal data array where to start converting to polygon (default 0)
              * @param stopIndex The index in the signal data array where to stop converting to polygon  (default
              *        UINT_MAX, i.e. up to the end)
              * @return true if the any of the polygons has changed
              */
            bool CreateOrUpdatePolygon( const QTransform & transformation, double pixelSize, uint startIndex = 0, uint stopIndex = UINT_MAX );


            /** @brief Removes all data for this signal before the specified index
              *
              * @param beforeIndex Specifies the index before which all data should be removed from the signal data
              * @return true is for all items the requested data part is removed, false if only for some or none
              */
            bool removeUnusedDataBefore( int beforeIndex );


        private:

            /** @brief Add the given information to the local stored administration and polyline information
              *
              * @param dataVal Represents the signal data point to add. This value is used to calculate the
              *                minimum and maximum values of the signal data
              * @param polyVal Represents the poly information that is related to the data point to add
              */
            void addPolyPoint( const QPointF & dataVal, QPointF polyVal );


            /** @brief Can be used to identify this signal with application specific information
              * This can be an index or a define or something else that can be used to uniquely
              * identify this signal amoungst others
              */
            int mId;


            /** @brief Can be used to hold the label that can be attached to the signal
              */
            QString mLabel;


            /** @brief If true, the signal must be shown
              */
            bool mShow;


            /** @brief The pen that will be used to draw this signal
              */
            QPen mPen;


            /** @brief The pen that will be used to draw this signal when in printing mode
              */
            QPen mPenPrint;


            /** @brief The brush that will be used to, for example, fill this polygon signal
              */
            QBrush mBrush;


            /** @brief The brush that will be used to, for example, fill this polygon signal when in printing mode
              */
            QBrush mBrushPrint;


            /** @brief Can be used to hold the correction of the data stored for the X-axis
              */
            double mCorrectionX;


            /** @brief Can be used to hold the correction of the data stored for the Y-axis
              */
            double mCorrectionY;


            /** @brief The data position on the x-axis if available
              * This member can contain the positions on the x-axis that will be used in
              * combination with the y-axis data to draw this signal. When this vector
              * is empty the index of the position (i.e. [0,1,2,...,n]) of the y-axis
              * positions will be used
              */
            PVector mXData;


            /** @brief The data position on the y-axis
              * This member will contain the positions on the y-axis that will be used in
              * combination with the x-axis data, if present, or the index offset to draw
              * this signal.
              */
            PVector mYData;


            /** @brief Hold the minimum data value that must be shown on the x-axis
              */
            double mMinXData;


            /** @brief Hold the maximum data value that must be shown on the x-axis
              */
            double mMaxXData;


            /** @brief Hold the minimum data value that must be shown on the y-axis
              */
            double mMinYData;


            /** @brief Hold the maximum data value that must be shown on the y-axis
              */
            double mMaxYData;


            /** @brief Convenient member that can be used to store the transformed signal data
              * When plotting the signal the raw data maintained in the local members mXData and
              * mYData must be transformed into a device specific set of coordinates.
              * this member can be used to hold the transformed data
              */
            QPolygonF mDataPoly;
        };
    }
}

#endif // PSIGNALDATA_H
