/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalDraw.cpp 1172 2014-03-21 13:45:45Z eelco $
 *
 */


#include "PSignalDraw.h"
#include "pfunctions.h"
#include "PWidgetDefines.h"

//#include <QDebug>


using namespace Peacs::Support;


// Defines the space that is left blank between the head of the running signal and the
// oldest part that can lye in front of it. The resulting space is the maximum space
// devided by this value
const double UPDATE_SIGNAL_CLEAR_SPACE  = 15.0;


PSignalDraw::PSignalDraw( double zoomT0, double zoomT1, int column, int row ) :
    PSignalPlotBase( false, zoomT0, zoomT1, column, row ),
    mLastSignalSize( 0 ),
    mMemSignalPlot(),
    mLastIntervalTimeStep( 0 ),
    mIndexToNow( 0 ),
    mIndexToStart( 0 ),
    mSignalSeperationSpace( UPDATE_SIGNAL_CLEAR_SPACE )
{
}


PSignalDraw::~PSignalDraw()
{
}


void PSignalDraw::init()
{
    mLastSignalSize         = 0;
    mMemSignalPlot          = QPixmap();    // By giving a empty (isNull() == true) forces a total redraw
    mLastIntervalTimeStep   = 0;
    mIndexToNow             = 0;
    mIndexToStart           = 0;

    PSignalPlotBase::init();
}


void PSignalDraw::updateSignalForTimeElapse( int elapse )
{
    mLastIntervalTimeStep   = elapse;

    if ( 0 < mSignals.count() )
    {
        const PSignalData & sigDataConst    = mSignals.at( 0 );

        // Check if there is data left to shift and show
        if ( mIndexToNow < sigDataConst.count() - 1 )
        {
            Q_ASSERT_X( 0 == sigDataConst.getXData().count(), "PSignalDraw::updateSignalForTimeElapse", "The signal X-data is not equal to the index of the Y-data" );

            int     range           = qRound( getTimeRange() );
            int     newIndexToNow   = mIndexToNow + elapse;
            int     newIndexToStart = mIndexToStart;
            QRectF  updateRect( getGraphRect() );

            if ( newIndexToNow >= sigDataConst.count() )
            {
                // Limit to the maximum of data present
                newIndexToNow = sigDataConst.count() - 1;
            }

            if ( newIndexToNow - newIndexToStart > range )
            {
                newIndexToStart += range;

                // All must now be updated
//if ( 100 > y() ) qDebug() << "1 -> mIndexToStart =" << mIndexToStart << "(" << newIndexToStart << ")\tmIndexToNow =" << mIndexToNow << "(" << newIndexToNow << newIndexToNow - mIndexToNow << ")\telapse =" << elapse << "\tsigCount" << sigDataConst.count();
            }
            else
            {
                // Update only the changed part
                double tmpVal = getGraphTransform().map( QPointF( mIndexToNow - mIndexToStart, 0.0 ) ).x();

                updateRect.setLeft( tmpVal );

                tmpVal = getGraphTransform().map( QPointF( newIndexToNow - newIndexToStart + (getTimeRange() / mSignalSeperationSpace), 0.0 ) ).x();

                updateRect.setRight( tmpVal );
            }

//if ( 100 > y() ) qDebug() << "2 -> mIndexToStart =" << mIndexToStart << "(" << newIndexToStart << ")\tmIndexToNow =" << mIndexToNow << "(" << newIndexToNow << newIndexToNow - mIndexToNow << ")\telapse =" << elapse << "\tsigCount" << sigDataConst.count() << "\tupdateRect" << updateRect;

            if ( mIndexToNow != newIndexToNow )
            {
                mIndexToNow    = newIndexToNow;
                mIndexToStart  = newIndexToStart;

                scene()->invalidate( mapRectToScene( updateRect ), QGraphicsScene::ItemLayer );
            }
        }
    }
}


void PSignalDraw::paintForeGround( QPainter * painter, const QRectF & updateRect )
{
    QRectF  paintRect( getGraphRect() );
    QPointF itemPos( pos() );

    paintRect.translate( itemPos );
    paintRect = paintRect.intersected( updateRect );

    painter->save();

    if ( isSignalClipping() )
    {
        painter->setClipping( true );
        painter->setClipRect( paintRect ); // do not draw outside the draw area
    }

    // Only draw the markers and selection rectangle when painting to the screen
    if ( !isPrinting() )
    {
        double          pixelSize       = getPixelSize();
        QRectF          objectRect;
        PSignalMarker * markerPtr       = NULL;
        QPainterPath    markerPath;

        for ( int markerNr = mMarkers.count() - 1; markerNr >= 0; markerNr-- )
        {
            markerPtr = mMarkers[ markerNr ];

            if ( markerPtr->mustShow() && !markerPtr->isEmpty() )
            {
                painter->setPen( markerPtr->getPen() );

                if ( markerPtr->getMarker().isEmpty() )
                {
                    markerPtr->translate( getGraphTransform() );
                }

                markerPath = markerPtr->getMarker();

                markerPath.translate( itemPos );

                objectRect = markerPath.boundingRect();

                if ( 0 == objectRect.width() )
                {
                    // Most time the width of a line is zero. In that case the intersection will fail. Just add a width
                    objectRect.setWidth( 1.0 );
                }

                if ( paintRect.intersects( objectRect ) )
                {
                    painter->drawPath( markerPath );
                }
            }
        }

        if ( getSelectRect().isValid() )
        {
            painter->setPen( QPen( QBrush( getSelectRectColor() ), pixelSize ) );
            painter->drawRect( getSelectRect() );
        }
    }

    painter->restore();

    PSignalPlotBase::paintForeGround( painter, updateRect );
}


void PSignalDraw::paint( QPainter *painter,
                         const QStyleOptionGraphicsItem * option,
                         QWidget * widget )
{
    Q_UNUSED( widget );

    double pixelSize = getPixelSize( painter );

    // Create or update the polygon information according the given signal data
    if ( mSignals.CreateOrUpdatePolygons( getGraphTransform(), pixelSize ) )
    {
        // Check if the size of the image in memory that is used to draw the signal information
        // is big enough. Note that we have to add the graph rectangle offset because the polygons
        // that represent the signals are mapped using the default mapper (getGraphTransform()) that
        // has the offset taken into account as well. The memory image will tehn have that offset
        // also. Keep this in mind when drawing the image onto the screen.
        const QPolygonF &   constDataPoly   = mSignals.at( 0 ).getDataPoly();
        const QPointF &     minValPoint     = constDataPoly.first();
        const QPointF &     maxValPoint     = constDataPoly.last();
        QSize               neededMemImageSize( int( maxValPoint.x() - minValPoint.x() + getGraphRect().x() ),
                                                int( getGraphRect().height() + getGraphRect().y() ) );

        // Make sure that the memory pixmap used to draw is big enough. Making it smaller isn't realy needed
        if ( mMemSignalPlot.isNull() ||
             (mMemSignalPlot.height() < neededMemImageSize.height()) ||
             (mMemSignalPlot.width() < neededMemImageSize.width()) )
        {
            mMemSignalPlot  = QPixmap( neededMemImageSize/*(maxValPoint.x() - minValPoint.x()), getGraphRect().height()*/ );
//if ( 100 > y() ) qDebug() <<  "minValPointX =" << minValPoint.x() << "maxValPointX =" << maxValPoint.x() << "MemSignalPlot rectangle =" << mMemSignalPlot.rect();
        }

        // New signal poly-data to draw is created. The memory bitmap must be updated
        // Clear the last drawn information
        mMemSignalPlot.fill( Qt::transparent );

        QPainter drawSignal( &mMemSignalPlot );

        drawSignal.setRenderHint( QPainter::Antialiasing );     // Use improved drawing

        for ( int sigNr = 0; sigNr < mSignals.count(); sigNr++ )
        {
            const PSignalData & constSigData = mSignals.at( sigNr );

            if ( constSigData.mustShow() && (constSigData.getDataPoly().count() > 0) )
            {
                QPen pen( constSigData.getPen() );

                if ( isPrinting() && (constSigData.getPrintPen().style() != Qt::NoPen) )
                {
                    pen = constSigData.getPrintPen();
                }

                pen.setWidthF( 1.3 * pixelSize );
                drawSignal.setPen( pen );

                drawSignal.drawPolyline( constSigData.getDataPoly() );
            }
        }
    }

    // All is ready now for drawing. Create or update the image containging the plotted signals
    if ( !mMemSignalPlot.isNull() )
    {
        // All is ready for displaying on the screen

        painter->save();

        if ( isSignalClipping() )
        {
            painter->setClipping( true );
            painter->setClipRect( getGraphRect().intersected( option->exposedRect ) ); // do not draw outside the draw area
        }

//if ( 100 > y() ) qDebug() << "2 -> clip rect =" << getGraphRect().intersected( option->exposedRect ) << "\t" << option->exposedRect;

//        // Just some extra to make it more fun. Fade-out the current content so the new
//        // signal will come more clear to the rest and the older the added content the
//        // more faded-out it will become
//        painter->setCompositionMode(QPainter::CompositionMode_DestinationIn);
//        painter->fillRect( getGraphRect(), QColor(0, 0, 0, 180) );
//        // Restore the way of painting so the rest can be added
//        painter->setCompositionMode( QPainter::CompositionMode_SourceOver );


        double left     = getGraphTransform().map( QPointF( mIndexToStart, 0.0 ) ).x();
        double right    = getGraphTransform().map( QPointF( mIndexToNow, 0.0 ) ).x();
        double width    = right - left;

        QRectF firstPartRect( left, mMemSignalPlot.rect().y(), width, mMemSignalPlot.rect().height() );
        QRectF secondPartRect;

        painter->drawPixmap( QPointF( getGraphRect().x(), 0.0 )/*getGraphRect().topLeft()*/, mMemSignalPlot, firstPartRect );

        // We add the half pixelsize to prevent very small parts behind the deciaml point to be of anny influence.
        if ( (0 < mIndexToStart) && ((width + (0.5 * pixelSize)) < getGraphRect().width()) )
        {
            secondPartRect = firstPartRect;


//double left1 = left;
//double right1 = right;
//double width1 = width;

            right   = left;
            width   = getGraphTransform().map( QPointF( getTimeRange() -
                                                                    (mIndexToNow - mIndexToStart) -
                                                                    (getTimeRange() / mSignalSeperationSpace), 0.0 ) ).x() - getGraphRect().x();
            left    = right - width;

            if ( width > 0 )
            {
                secondPartRect.setWidth( width );

                secondPartRect.moveRight( right );

                painter->drawPixmap( QPointF( getGraphRect().right() - width, 0.0 ), mMemSignalPlot, secondPartRect );

//if ( 100> y() ) qDebug() << "mIndexToStart" << mIndexToStart << "\tmIndexToNow" << mIndexToNow << "\tleft" << left1 << "\tright" << right1 << "\twidth" << width1 << "\tleft2" << left << "\tright2" << right << "\twidth2" << width;
            }
        }
//        else
//        {
//if ( 100> y() ) qDebug() << "mIndexToStart" << mIndexToStart << "\tmIndexToNow" << mIndexToNow << "\tleft" << left << "\tright" << right << "\twidth" << width << "\tgraph left" << getGraphRect().left() << "\tgraph right" << getGraphRect().right();
//        }

        painter->restore();
    }

    // Call the base implementation to reset the stop mouse move handling flag
    PSignalPlotBase::paint( painter, option, widget );
}


void PSignalDraw::correctAdministrationForNewData( void )
{
    if ( 0 < mSignals.count() )
    {
        bool                changed         = false;
        int                 removed         = 0;
        int                 range           = qRound( getTimeRange() );
        const PSignalData & sigDataConst    = mSignals.at( 0 );


        // Check if the drawing of the dynamic signal is getting behind. If so, we have to
        // speed things a little it up. This is done by just showing all of the previous data
        // mLastIntervalTimeStep represents the last elapsed interval, normally about 20 msec.
        // When more then, just take smoething, 2 times this amount is not being shown we
        // assume that we are getting behind.
        if ( (0 < mLastIntervalTimeStep) &&
             (mLastSignalSize < (unsigned int)sigDataConst.count()) &&              // New data has been added
             (2 * mLastIntervalTimeStep < int(mLastSignalSize) - mIndexToNow) )     // And the previous data isn't completely shown
        {
//qDebug() << "Setting mIndexToNow" << mIndexToNow << "to" << mLastSignalSize << "because of" << mLastIntervalTimeStep << "and" << sigDataConst.count();
            // Go to the end of the previous available data
            mIndexToNow = mLastSignalSize;
            changed      = true;

            if ( mIndexToNow >= sigDataConst.count() )
            {
//qDebug() << "Setting mIndexToNow" << mIndexToNow << " to" << sigDataConst.count() - 1;
                // Limit to the maximum of data present
                mIndexToNow = sigDataConst.count() - 1;
            }

            // Correct the start of the signal accordingly
            if ( mIndexToNow - mIndexToStart > range )
            {
//                int times = int((mIndexToNow - mIndexToStart) / range);

                mIndexToStart += int((mIndexToNow - mIndexToStart) / range) * range;
            }
        }

        // Check if there is something to remove
        if ( (mIndexToNow < sigDataConst.count()) && (0 <= mIndexToNow) && (mIndexToNow > range) )
        {
            // We now assume thet when an dynamic signal is being shown, there isn't real X-axis data. The X-axis data
            // represents the indexes to the Y-axis data to show. A fixed interval is requiered. The timer interval used
            // to update the dynamic signal elapses in mili-seconds, thus an interval of mili-seconds shoudl be fine
            Q_ASSERT_X( 0 == sigDataConst.getXData().count(), "PSignalDraw::findAndRemoveOldData", "The signal X-data is not equal to the index of the Y-data" );

            removed = mIndexToNow - range;

            if ( (0 > removed) && (removed >= sigDataConst.count()) )
            {
                removed = 0;
            }
        }

        // Check if there is more data up to the 'now' moment then there can be displayed
        if ( 0 < removed )
        {
            mIndexToNow     = qMax( mIndexToNow - removed, 0 );
            mIndexToStart   = qMax( mIndexToStart - removed, 0 );
            changed          = true;

            if ( !mSignals.removeUnusedDataBefore( removed ) )
            {
                Q_ASSERT( false );  // Should not be possible
            }

//qDebug() << "removed =" << removed << "mIndexToStart =" << mIndexToStart << "\tmIndexToNow =" << mIndexToNow << "\tsigDataCount =" << mLastSignalSize << sigDataConst.count();
        }

        if ( changed )
        {
            scene()->invalidate( mapRectToScene( getGraphRect() ), QGraphicsScene::ItemLayer );
        }

        mLastSignalSize = sigDataConst.count();
    }
}
