/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalDraw.h 1172 2014-03-21 13:45:45Z eelco $
 *
 */

#ifndef __PSIGNALDRAW_H__
#define	__PSIGNALDRAW_H__


#include <QStyleOptionGraphicsItem>
#include <QTimer>

#ifdef WIN32
    #include <windows.h>
#endif

#include "psignalplotBase.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PSignalDraw PSignalDraw.h "widgets/PSignalDraw.h"
          * @brief The graphics scene item that can show a graph with an
          * in time plotting signal
          */
        class PEACSSUPPORTSHARED_EXPORT PSignalDraw : public PSignalPlotBase
        {
//            Q_OBJECT

        public:

            /** @brief Only present to be used in the type() function
              * @see PGrItemBase
              */
            enum { Type = PGrItemBase::PSignalDrawType };


            /** @brief Returns the type of graphics item.
              * Enable the use of qgraphicsitem_cast with this item.
              * @see PGrItemBase
              */
            inline int type( void ) const { return Type; }


            /** @brief constructor
              *
              * @param zoomT0 Specified the initial starting point, i.e. left extent of the X-axis
              * @param zoomT1 Specified the initial end point, i.e. right extent of the X-axis
              * @param column the column at which this item will reside in the matrix (default -1)
              * @param row the row at which this item will reside in the matrix (default -1)
              *
              * @see PGrItemBase for information about column and row
              */
            PSignalDraw( double zoomT0, double zoomT1, int column = -1, int row = -1 );


            /** @brief destructor: Nothing to do
              */
            virtual ~PSignalDraw();


            /** @brief initializes the class with the data provided before
              * This function must be called before displaying the desired data. It
              * will initialize local members and in a derived class implementation
              * specific data can be set, like labels.
              */
            virtual void init();


            /** @brief In this class no real function is implemented, override to add needed function
              * Override this function in derived classes to enable drawing data on top of the background
              * and signal information
              *
              * @param painter the painter see Qt documentation
              * @param updateRect the rectangle, in scene coordinates, of the rectangle to (re-)draw
              */
            virtual void paintForeGround( QPainter * painter, const QRectF & updateRect );


            /** @brief Paint the signals within the graph
              *
              * @param painter see PGraphicsSignalView
              * @param option used to determine the scale
              * @param widget not used
              */
            virtual void paint( QPainter * painter, const QStyleOptionGraphicsItem * option, QWidget * widget = 0 );


            /** @brief Returns the space that is left blank between the head of the running signal and the oldest part
              * This space is defined by the maximum signal range shown on the X-axis devided by this value
              *
              * @return mSignalSeperationSpace
              */
            inline double getSignalSeperationSpace( void ) const { return mSignalSeperationSpace; }


            /** @brief Sets the space that is left blank between the head of the running signal and the oldest part
              * This space is defined by the maximum signal range shown on the X-axis devided by this value
              *
              * @param newSpace The new value to use for mSignalSeperationSpace
              */
            inline void setSignalSeperationSpace( double newSpace ) { mSignalSeperationSpace = newSpace; }


            /** @brief Must be called to update the shown signal for the next dynamic part
              *
              * @param elapse Holds the delta time in miliseconds the shown signal must be updated (shifted) for
              */
            void updateSignalForTimeElapse( int elapse );


            /** @brief Corrects the administration of the dynamic signal
              * This function will be called by PGraphicsSignalView::updateInfo() after adding newly recieved
              * data. This function will check if there is data that has already been shown and thus can be
              * removed. It also corrects the aministration indexes for when drawing gets behind with new
              * signal data.
              */
            void correctAdministrationForNewData( void );


        protected:

//            /** @brief Implement this function to adjust the position of the marker for the given position
//              * This function will be called by the mouse click handling function to indicate a user action
//              * at the specified position in the graph. The given positions will be in data coordinates
//              * system and graph coordinate system to enable maximum control.
//              * Derived, application specific classes should implement the correct marker behaviour for this.
//              *
//              * This function should not be called here but in a derived class
//              *
//              * @param dataPos The position to adjust with. This position is in data coordinate system
//              * @param graphPos The position to adjust with. This position is in graph coordinate system
//              * @param mouseReleased If true, this function is being called from the mouse release event, otherwise the mouse move event
//              *
//              * @return true if the position has changed
//              */
//            virtual bool adjustMarkerPosition( const QPointF & dataPos, const QPointF & graphPos, bool mouseReleased )
//                { Q_UNUSED(dataPos) Q_UNUSED(graphPos) Q_UNUSED(mouseReleased) Q_ASSERT( false ); return false; }


        private:

            /** @brief Holds the last size of the recieved signaldata
              * This member is used to track changes due to new added data
              */
            unsigned int mLastSignalSize;


            /** @brief Member used to draw the signal in memory
              */
            QPixmap mMemSignalPlot;


            /** @brief Hold the last time refresh interval step
              * This member contains the time step that was used to shift the signal window to make
              * this graph display the next part
              */
            int mLastIntervalTimeStep;


            /** @brief Hold the index to the sinal data up to where the signal is now being displayed
              */
            int mIndexToNow;


            /** @brief Hold the index to the sinal data up to where the signal is now being displayed at the left side
              */
            int mIndexToStart;


            /** @brief Defines the space that is left blank between the head of the running signal and the oldest part
              * Due to the 'drawing' activity of the signal, the older part of the signal can lies in front of it. This
              * member defines the space left between the the two parts of the drawn signal
              * This space is defined by the maximum signal range shown on the X-axis devided by this value
              */
            double mSignalSeperationSpace;
        };
    }
}

#endif	/* __PSIGNALDRAW_H__ */

