/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalMarker.h 1105 2014-02-19 08:38:33Z eelco $
 *
 */

#ifndef PSIGNALMARKER_H
#define PSIGNALMARKER_H


#include "PeacsSupport_global.h"

#include <QPen>
#include <QPainterPath>


namespace Peacs
{
    namespace Support
    {
        /** @class PSignalMarker PSignalMarker.h "widgets/PSignalMarker.h"
          * @brief Holds all data and information about one marker to plot within the signal view
          */
        class PEACSSUPPORTSHARED_EXPORT PSignalMarker
        {

        public:

            /** @brief Only present to be used in the type() function
              */
            enum { Type = 0 };


            /** @brief Returns the type of marker.
              * Each derived class, representing a different kind of marker, should specify a unique
              * marker type identifier.
              * Enable the use of marker_cast with this item.
              */
            virtual int type( void ) const { return Type; }

            /** @brief Default constructor
              * This constructor should only be used when claiming object during some sort
              * of initialization. Normally not to be used.
              */
            PSignalMarker( void );


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              */
            PSignalMarker( int id, const QString & label = QString() );


            /** @brief Initialyzer constructor
              *
              * @param id the value to use for mId
              * @param label the value to use for mLabel (default an empty string)
              * @param show flag indicating if the marker must be shown initially
              * @param pen the QPen object to use when drawing the marker.
              */
            PSignalMarker( int id, const QString & label, bool show, const QPen & pen );


            /** @brief Destructor
              */
            virtual ~PSignalMarker( void ) {;}


            /** @brief Copy constructor
              *
              * @param source the source to copy from
              */
            PSignalMarker( const PSignalMarker & source );


            /** @brief Asignment operator
              *
              * @param source the source to copy from
              */
            virtual const PSignalMarker & operator =( const PSignalMarker & source );


            /** Test equal operator
              *
              * @param source the source to test against
              */
            virtual bool operator ==( const PSignalMarker & source ) const;


            /** Test not equal operator
              *
              * @param source the source to test against
              */
            virtual bool operator !=( const PSignalMarker & source ) const;


            /** @brief Returns mId
              *
              * @return The identifier of this marker
              */
            inline int getId( void ) const { return mId; }


            /** @brief Sets mId
              * Sets mId to the new value
              *
              * @param id The value for mId
              */
            inline void setId( int id ) { mId = id; }


            /** @brief Returns mLabel
              *
              * @return The label of this marker
              */
            inline const QString & getLabel( void ) const { return mLabel; }


            /** @brief Sets mLabel
              * Sets mLabel to the new value
              *
              * @param label The new value for mLabel
              */
            inline void setLabel( const QString & label ) { mLabel = label; }


            /** @brief Enables of disables this marker data for showing
              *
              * @param show the new show status
              * @return true if the flag was changed, otherwise false
              */
            inline bool setShow( bool show ) { bool changed = mShow != show; mShow = show; return changed; }


            /** @brief Inverts the show status flag
              */
            inline void toggleShow( void ) { mShow = !mShow; }


            /** @brief returns mShow
              *
              * @return true if this marker data must be shown, otherwise false
              */
            inline bool mustShow( void ) const { return mShow; }


            /** @brief Returns mPen
              *
              * @return The pen that will be used to draw this marker
              */
            inline const QPen & getPen( void ) const { return mPen; }


            /** @brief Returns mPen as a non const object
              * Convenient function that can be used to change the pen object
              *
              * @return The pen as a non const object that will be used to draw this marker
              */
            inline QPen & pen( void ) { return mPen; }


            /** @brief Sets mPen
              * Sets mPen to the new value
              *
              * @param pen The new value for mPen
              */
            inline void setPen( const QPen & pen ) { mPen = pen; }


            /** @brief Initialyzes mPen with a new QPen object with the given color
              *
              * @param color The color to use with the new mPen
              */
            inline void setPen( const QColor & color ) { mPen = QPen( color ); }


            /** @brief Sets the width of mPen
              *
              * @param width The width to use with mPen
              */
            inline void setPenWidth( qreal width ) { mPen.setWidthF( width ); }


            /** @brief Returns mMarker
              *
              * @return The marker object to draw
              */
            inline const QPainterPath & getMarker( void ) const { return mMarker; }


            /** @brief Sets mMarker
              * Note that the object specified in the painter path should be in graph coordinate system and
              * not in the data coordinate system, or the user should tranform it seperately
              *
              * @param marker The object to use as marker
              */
            inline void setMarker( const QPainterPath & marker ) { mMarker = marker; }


            /** @brief Clears all data in mMarker
              */
            virtual void clear( void ) { mMarker = QPainterPath(); }


            /** @brief Clears all data in mMarker
              */
            inline void clearPath( void ) { mMarker = QPainterPath(); }


            /** @brief This function must be implementedin a derived class to transform local stored values into the marker object
              * This function must translate any, in the derived class internal stored data, from data coordinate system into
              * an object in graph coordinates, by using the QTransform object, ready to display. This object can then be stored in
              * mMarker for later re-use
              * Before the transformation will be performed, there will be checked if the local stored object, mMarker, is still valid
              *
              * @param translation The transformation object used to transform from data coordinates into graph coordinates
              * @param forceTransformation if true the transformation will be performed without checking the local stored object
              *
              * @return the resulting object stored in a painterpath mMarker
              */
            virtual const QPainterPath & translate( const QTransform & translation, bool forceTransformation = false );


            /** @brief This function must be implementedin a derived class to transform local stored values into the marker object
              * This function must translate any, in the derived class internal stored data, from data coordinate system into
              * an object in graph coordinates, by using the QTransform object, ready to display.
              * This version of the function must always perform the transformation without storing the result in mMarker
              *
              * @param translation The transformation object used to transform from data coordinates into graph coordinates
              *
              * @return the resulting marker object stored in a painterpath
              */
            virtual QPainterPath translated( const QTransform & translation );


            /** @brief Checks if there is any data stored to be shown
              * This depends on data sotred in derived classes. The mMarker is only used to store trandsformed data
              * for convenience for speedup the display of the marker
              * Should be filled in in a derived class
              *
              * @return true if there is data to be shown
              */
            virtual bool isEmpty( void ) const;


            /** @brief Checks if this marker is at the specified position
              * This class will always return false. Real implementation must be done in derived classes
              *
              * @param pos The position to look at. This position should be in data coordinate system
              * @param range A positive value indicating the allowed distance range of the position to the marker
              *
              * @return true if this marker is at the position
              */
            virtual bool isMarkerAtPos( const QPointF & pos, double range = 1.0 ) const;


        protected:

            /** @brief The painterpath object that contains the marker figure
              */
            QPainterPath mMarker;


        private:

            /** @brief Can be used to identify this marker with application specific information
              * This can be an index or a define or something else that can be used to uniquely
              * identify this marker amoungst others
              */
            int mId;


            /** @brief Can be used to hold the label that can be attached to the marker
              * Can also be used as an identifier
              */
            QString mLabel;


            /** @brief If true, the marker must be shown
              */
            bool mShow;


            /** @brief The pen that will be used to draw this marker
              */
            QPen mPen;
        };


        /** @brief convenient cast function that also tests of the marker is of the requested type
          *
          * @return the specified object casted to the requested type if possible, otherwise 0
          */
        template <class T> inline T marker_cast( PSignalMarker * markerPtr )
        {
            return ( int( static_cast<T>(0)->Type ) == int( PSignalMarker::Type ) ) ||
                   ( markerPtr && int(static_cast<T>(0)->Type) == markerPtr->type()) ? static_cast<T>( markerPtr ) : 0;
        }


        /** @brief convenient cast function that also tests of the marker is of the requested type
          *
          * @return the specified const object casted to the requested type if possible, otherwise 0
          */
        template <class T> inline T marker_cast( const PSignalMarker * markerPtr )
        {
            return ( int( static_cast<T>(0)->Type ) == int( PSignalMarker::Type ) ) ||
                   ( markerPtr && int(static_cast<T>(0)->Type) == markerPtr->type()) ? static_cast<T>( markerPtr ) : 0;
        }
    }
}

#endif // PSIGNALMARKER_H
