/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalMarkers.h 1113 2014-02-19 21:03:28Z eelco $
 *
 */


#ifndef PSIGNALMARKERS_H
#define PSIGNALMARKERS_H


#include <QVector>

#include "PeacsSupport_global.h"
#include "PSignalMarker.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PSignalMarkers PSignalMarkers.h "widgets/PSignalMarkers.h"
          * @brief Holds a list with data and information about a collection of signals to plot in a signal view
          */
        class PEACSSUPPORTSHARED_EXPORT PSignalMarkers
        {

        public:

            enum
            {
                NO_MARKER_FOUND = -1
            };


            /** @brief Constructor
              */
            PSignalMarkers( void );


            /** @brief Initialyzer constructor
              * Creates a vector with 'size' position reserved for signal data objects
              *
              * @param size the number of objects to instantiate during construction
              */
            PSignalMarkers( int size );


            /** @brief Destructor
              */
            ~PSignalMarkers( void );


            /** @brief Removes all the elements from the vector and releases the memory used by the vector.
              */
            void clear( void );


            /** @brief Returns the number of PSignalMarkers items in the vector.
              */
            inline int size( void ) const { return mData.size(); }


            /** @brief Returns the number of PSignalMarkers items in the vector.
              * This is an overloaded function.
              * @see size()
              */
            inline int count( void ) const { return mData.count(); }


            /** @brief Sets the number of marker pointers reserved to size
              * This function reserves the amount of pointer space to store markers. If the current number of
              * contained markers is bigger, the exceeding number of marker pointers are freed. The marker
              * objects themself are deleted. Be sure that no references to these marker objects are stored
              * elsewher. If the current number of marker pointers is smaller, more space is reserved. The
              * extra positions are set to 0 (zero)
              *
              * @param size The new number of marker object pointers to maintain
              * @return If the new size is set, true is returend
              */
            bool setSize( int size );


            /** @brief Returns a reference to the last PSignalMarker item in the vector.
              * This function assumes that the vector isn't empty.
              */
            inline PSignalMarker * last( void ) { Q_ASSERT( !isEmpty() ); return mData.last(); }


            /** @brief This is an overloaded function.
              * @see last()
              */
            inline const PSignalMarker * last( void ) const { Q_ASSERT( !isEmpty() ); return mData.last(); }


            /** @brief Returns true if the vector has size 0; otherwise returns false.
              */
            inline bool isEmpty( void ) const { return mData.isEmpty(); }


            /** @brief Returns the PSignalMarker item at index position sdIndex in the vector.
              *
              * @param sdIndex must be a valid index position in the vector (i.e., 0 <= sdIndex < size())
              */
            inline const PSignalMarker * at( int sdIndex ) const
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < mData.size(), "PSignalMarkers::at", "index out of range");
                return mData.at( sdIndex );
            }


            /** @brief Returns the PSignalMarker item at index position sdIndex in the vector.
              *
              * @param sdIndex must be a valid index position in the vector (i.e., 0 <= sdIndex < size())
              */
            inline const PSignalMarker * operator[]( int sdIndex ) const
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < mData.size(), "const PSignalMarkers::operator[]", "index out of range");
                return mData[ sdIndex ];
            }


            /** @brief Returns the PSignalMarker item at index position sdIndex as a modifiable reference.
              *
              * @param sdIndex must be a valid index position in the vector (i.e., 0 <= sdIndex < size())
              */
            inline PSignalMarker * & operator[]( int sdIndex )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < mData.size(), "PSignalMarkers::operator[]", "index out of range");
                return mData[ sdIndex ];
            }


            /** @brief Adds a new marker object to the vector
              * This function takes a pointer to the object. When added, this class will take ownership
              * and will be responsible for deleting the pointer object
              *
              * @param marker Pointer to the marker object to add
              */
            inline void addMarker( PSignalMarker * marker ) { mData.append( marker ); }


            /** @brief Searches all markers for the marker that is positioned at, or near, the specifed position
              *
              * @param pos The position to look at
              * @param range A positive value indicating the allowed distance range of the position to the marker
              * @param onlyVisible If true (default) only visible markers are searched
              *
              * @return The index of the found marker or NO_MARKER_FOUND if none was found
              */
            int findMarker( const QPointF & pos, double range = 1.0, bool onlyVisible = true ) const;


            /** @brief Returns the index of the marker that is selected
              *
              * @return mSelectedIndex
              */
            inline int getSelectedMarkerIndex( void ) const { return mSelectedIndex; }


            /** @brief Returns the index of the marker that is selected
              *
              * @return mSelectedIndex
              */
            inline PSignalMarker * getSelectedMarker( void ) const
            {
                Q_ASSERT_X(mSelectedIndex >= 0 && mSelectedIndex < mData.size(), "const PSignalMarkers::getSelectedMarker", "mSelectedIndex not valid");
                return (mSelectedIndex > NO_MARKER_FOUND ? mData[mSelectedIndex] : 0);
            }


            /** @brief Sets mSelectedIndex
              *
              * @param index The index of the marker that must be set as being selected
              */
            inline void setSelectedMarker( int index )
            {
                Q_ASSERT_X(index >= NO_MARKER_FOUND && index < mData.size(), "const PSignalMarkers::setSelectedMarker", "index out of range");
                mSelectedIndex = index;
            }


            /** @brief Resets mSelectedIndex to none
              */
            inline void resetSelectedMarker( void ) { mSelectedIndex = NO_MARKER_FOUND; }


            /** @brief Returns if there is a marker set selected
              *
              * @return true if mSelectedIndex != NO_MARKER_FOUND
              */
            inline bool hasSelectedMarker( void ) { return mSelectedIndex != NO_MARKER_FOUND; }


            /** @brief Enables of disables the specified marker data for showing
              *
              * @param sdIndex must be a valid index position in the vector (i.e., 0 <= sdIndex < size())
              * @param show the new show status of the specified marker
              * @return true if the show flag for the marker was changed, otherwise false
              */
            inline bool setShow( int sdIndex, bool show )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < mData.size(), "PSignalMarkers::setShow", "index out of range");
                return (sdIndex >= 0 && sdIndex < mData.size() ? mData[ sdIndex ]->setShow( show ) : false);
            }


        private:

            Q_DISABLE_COPY( PSignalMarkers )


            /** @brief Hold the data this class is responsible for
              */
            QVector< PSignalMarker * > mData;


            /** @brief Holds the index of the marker that is currently being selected
              */
            int mSelectedIndex;
        };
    }
}

#endif // PSIGNALMARKERS_H
