/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalPlotBase.cpp 1161 2014-03-12 12:21:14Z eelco $
 *
 */


#include <QGraphicsScene>

#include <QDebug>
#include <QGraphicsView>

#include "pfunctions.h"
#include "PStatistics.h"
#include "psignalPlotBase.h"
#include "PMarkerLine.h"
#include "PMarkerEllipse.h"
#include "PMarkerVector.h"

#include "math.h"

using namespace Peacs::Support;


const double    DEF_AMPLITUDE               = 100.0;
// When a user presses the mouse double but also moves the mouse, sometimes the second mouse pres and move
// will result in a selection tracking action what will result in a very small selected area. This is not
// desired behaviour. Set the minimum mouse move distance to 10 units to prevent most of these errors
const qreal     SELECT_RECT_MINIMUM_SIZE    = 10;


// Constant offsets used to get the correspnding items from mItemSizes
// All the tectangles indexed by these offsets are in scene coordinates!
const int SIZES_INDEX_TEXT          = 0;
const int SIZES_INDEX_LABEL_X       = 1;
const int SIZES_INDEX_LABEL_Y       = 2;
const int SIZES_INDEX_UNIT_X        = 3;
const int SIZES_INDEX_UNIT_Y        = 4;
const int SIZES_INDEX_GRAPH         = 5;
const int SIZES_INDEX_TOTAL         = 6;



PSignalPlotBase::PSignalPlotBase( bool xAxisShowsIndex, double zoomT0, double zoomT1, int column, int row ) :
    QGraphicsObject(),
    PGrItemBase( column, row ),
    mSignals(),
    mMarkers(),
    mGrid( Widget::SHOW_GRID ),
    mXAxisShowsIndex( xAxisShowsIndex ),
    mTextFont( Widget::FONTFAMILY, -1, QFont::DemiBold ),
    mLabelFont( Widget::FONTFAMILY, -1, QFont::DemiBold ),
    mScaleFont( Widget::FONTFAMILY, -1, QFont::DemiBold ),
    mText(),
    mInGraphText(),
    mXLabel(),
    mYLabel(),
    mXUnit(),
    mYUnit(),
    mMaxSigAmpl( DEF_AMPLITUDE ),
    mMinSigAmpl( -DEF_AMPLITUDE ),
    mItemSizes(),
    mShowXScaleTexts( true ),
    mXScaleTexts(),
    mXScaleTextRect(),
    mShowYScaleTexts( true ),
    mYScaleTexts(),
    mYScaleTextRect(),
    mShowYZeroLine( true ),
    mGraphRect(),
    mGraphTransform(),
    mIsSymetric( false ),
    mDoAdaptXAxis( true ),
    mLimitAdaptXAxis( false ),
    mDoAdaptYAxis( true ),
    mLimitAdaptYAxis( false ),
    mMaximizeGraph( false ),
    mAllowZoomSelect( false ),
    mAmplitudeRange( mMaxSigAmpl - mMinSigAmpl ),
    mAmplitudeStep( mAmplitudeRange / 10.0 ),
    mInvertXDirection( false ),
    mInvertYDirection( false ),
    mXfactor( 10 ),
    mYfactor( 10 ),
    mXLabelFactor( 1 ),
    mYLabelFactor( 1 ),
    mZoomT0( zoomT0 ),
    mZoomT1( zoomT1 ),
    mTimeStep( 0 ),
    mTimeRange( zoomT1 - zoomT0 ),
    mSelectRect(),
    mZeroLine( Widget::ZEROLINECOLOR ),
    mZeroLinePrint( Widget::ZEROLINECOLOR_PRINT ),
    mGridLine( Widget::LINECOLOR ),
    mGridLinePrint( Widget::GRIDCOLOR_PRINT ),
    mGridPoint( Widget::GRIDCOLOR_DEFAULT ),
    mGridPointPrint( Widget::GRIDCOLOR_PRINT ),
    mBackGroundColor( /*Widget::DEFAULTBACKGROUNDCOLOR*/ ),
    mBackGroundColorPrint(),
    mTextColor( Widget::DEFAULTGRAPHTEXTCOLOR ),
    mTextColorPrint( Widget::DEFAULTGRAPHTEXTCOLOR ),
    mSelectRectColor( Qt::black ),
    mIsPrinting( false ),
    mClipSignal( true ),
    mWaitForDraw( false )
{
    setAcceptedMouseButtons( Qt::LeftButton );

    calcTimeScale();        // Initialize the default axis scales
    calcAmplitudeScale();
}


PSignalPlotBase::~PSignalPlotBase()
{
}


void PSignalPlotBase::init()
{
    // Make sure that all previous information is cleared
    updateWithBackground();
}


void PSignalPlotBase::updateWithBackground( QGraphicsScene::SceneLayers layers )
{
    if ( scene() )
    {
        QRectF itemRectInScene( mBounds );

        itemRectInScene.translate( pos() );
//qDebug() << "updateWithBackground called";
        scene()->invalidate( itemRectInScene, layers );
    }
}


void PSignalPlotBase::forceRedrawSignalsAndMarkers( void )
{
    for ( int sigNr = 0; sigNr < mSignals.count(); sigNr++ )
    {
        // Force a recalculation of the polygons to draw
        mSignals[ sigNr ].clearPoly();
    }

    for ( int markerNr = 0; markerNr < mMarkers.count(); markerNr++ )
    {
        // Force a recalculation of the marker position to draw
        mMarkers[ markerNr ]->clearPath();
    }
}


bool PSignalPlotBase::setBounds( qreal width, qreal height )
{
    bool changed = false;

    if ( (mBounds.width() != width) || (mBounds.height() != height) )
    {
        mBounds.setWidth( width );
        mBounds.setHeight( height );

        mItemSizes.clear();         // Force updating the calculation of the text sizes

        forceRedrawSignalsAndMarkers();

        prepareGeometryChange();    // Notify the view of the new bounds (update will be called if needed)

        changed = true;
    }

    return changed;
}


void PSignalPlotBase::setXfactor( int factor )
{
    if ( mXfactor != factor )
    {
        mXfactor = factor;

        calcTimeScale();
    }
}


void PSignalPlotBase::setYfactor( int factor )
{
    if ( mYfactor != factor )
    {
        mYfactor = factor;

        calcAmplitudeScale();

    }
}


void PSignalPlotBase::setXLabelFactor( double factor )
{
    if ( mXLabelFactor != factor )
    {
        mXLabelFactor       = factor;

        updateWithBackground();     // Update the background of this item
    }
}


void PSignalPlotBase::setYLabelFactor( double factor )
{
    if ( mYLabelFactor != factor )
    {
        mYLabelFactor       = factor;

        updateWithBackground();     // Update the background of this item
    }
}


double PSignalPlotBase::toItemAmplitude( double ampl ) const
{
    double tmpValue;

    if ( mInvertYDirection )
    {
        tmpValue = ampl - mMinSigAmpl;
    }
    else
    {
        tmpValue = mMaxSigAmpl - ampl;
    }

    Q_ASSERT( !mGraphRect.isNull() );

    return  mGraphRect.height() * tmpValue / mAmplitudeRange + mGraphRect.y();
//    return ( mBounds.height() - LEDGENDX ) * tmpValue / mAmplitudeRange;
}


double PSignalPlotBase::toAmplitude( double pos ) const
{
    Q_ASSERT( !mGraphRect.isNull() );

//    double height   = mBounds.height() - LEDGENDX;
//    double tmpValue = pos * mAmplitudeRange / height;
    double posCorrected = pos - mGraphRect.y();
    double tmpValue = posCorrected * mAmplitudeRange / mGraphRect.height();
    //double tmpValue = pos * mAmplitudeRange / mGraphRect.height();

    if ( mInvertYDirection )
    {
        tmpValue = tmpValue + mMinSigAmpl;
    }
    else
    {
        tmpValue = mMaxSigAmpl - tmpValue;
    }

    return qBound( mMinSigAmpl, tmpValue, mMaxSigAmpl );
}


double PSignalPlotBase::toItemTime( double t ) const
{
    double tmpValue;

    if ( mInvertXDirection )
    {
        tmpValue = mZoomT1 - t;
    }
    else
    {
        tmpValue = t - mZoomT0;
    }

    Q_ASSERT( !mGraphRect.isNull() );

    return mGraphRect.width() * tmpValue / mTimeRange + mGraphRect.x();
//    return ( mBounds.width() - LEDGENDY ) * tmpValue / mTimeRange + LEDGENDY;
}


double PSignalPlotBase::toTime( double pos ) const
{
    Q_ASSERT( !mGraphRect.isNull() );

//    double width        = (mBounds.width() - LEDGENDY);
    double posCorrected = pos - mGraphRect.x();
//    double tmpValue     = mTimeRange * posCorrected / width;
    double tmpValue     = mTimeRange * posCorrected / mGraphRect.width();
    //double tmpValue     = mTimeRange * pos / mGraphRect.width();

    if ( mInvertXDirection )
    {
        tmpValue = mZoomT1 - tmpValue;
    }
    else
    {
        tmpValue = tmpValue + mZoomT0;
    }

    return qBound( mZoomT0, tmpValue, mZoomT1 );
}


bool PSignalPlotBase::setGrid( Widget::Grid_type grid )
{
    if ( mGrid != grid )
    {
        mGrid               = grid;

        updateWithBackground();     // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setTextFont( const QFont& font )
{
    if ( mTextFont != font )
    {
        mTextFont = font;

        if ( !mText.isEmpty() )
        {
            mItemSizes.clear();         // Force updating the calculation of the text sizes

            forceRedrawSignalsAndMarkers();

            updateWithBackground();     // Update the background of this item
        }

        return true;
    }

    return false;
}


bool PSignalPlotBase::setLabelFont( const QFont& font )
{
    if ( mLabelFont != font )
    {
        mLabelFont = font;

        if ( !mXLabel.isEmpty() || !mYLabel.isEmpty() )
        {
            mItemSizes.clear();         // Force updating the calculation of the text sizes

            forceRedrawSignalsAndMarkers();

            updateWithBackground();     // Update the background of this item
        }

        return true;
    }

    return false;
}


bool PSignalPlotBase::setScaleFont( const QFont& font )
{
    if ( mScaleFont != font )
    {
        mScaleFont = font;

        mXScaleTexts.clear();           // Force updating the calculation of the X-scale text sizes
        mYScaleTexts.clear();           // Force updating the calculation of the Y-scale text sizes

        forceRedrawSignalsAndMarkers();

        if ( !mXUnit.isEmpty() || !mYUnit.isEmpty() )
        {
            mItemSizes.clear();         // Force updating the calculation of the text sizes
        }
        updateWithBackground();         // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setXUnit( const QString & unit )
{
    if ( mXUnit != unit )
    {
        mXUnit              = unit;

        mItemSizes.clear();         // Force updating the calculation of the text sizes
        updateWithBackground();     // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setYUnit( const QString & unit )
{
    if ( mYUnit != unit )
    {
        mYUnit              = unit;

        mItemSizes.clear();         // Force updating the calculation of the text sizes
        updateWithBackground();     // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setXLabel( const QString & label )
{
    if ( mXLabel != label )
    {
        mXLabel             = label;

        mItemSizes.clear();         // Force updating the calculation of the text sizes
        updateWithBackground();     // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setYLabel( const QString & label )
{
    if ( mYLabel != label )
    {
        mYLabel             = label;

        mItemSizes.clear();         // Force updating the calculation of the text sizes
        updateWithBackground();     // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setText( const QString & text )
{
    if ( mText != text )
    {
        mText = text;

        mItemSizes.clear();         // Force updating the calculation of the text sizes
        updateWithBackground();     // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setInGraphText( const QString & text )
{
    if ( mInGraphText != text )
    {
        mInGraphText = text;

        updateWithBackground();     // Update the background of this item

        return true;
    }

    return false;
}


bool PSignalPlotBase::setSignalAmpl( double min, double max, bool doCheck )
{
    bool changed = false;

    if ( ( min < max ) &&
         ( (mMinSigAmpl != min) || (mMaxSigAmpl != max) ) )
    {
        if ( !doCheck ||
             (fabs(max - min) >= fabs(mAmplitudeStep)) )        // Allow the amplitude to shrink to minimum of one step)
        {
            bool up = fabs( mMaxSigAmpl - mMinSigAmpl ) <= fabs( max - min ) || !doCheck;

//qDebug() << "Init: max =" << max << "- min =" << min << "- doCheck =" << doCheck << "- mMaxSigAmpl =" << mMaxSigAmpl << "- mMinSigAmpl =" << mMinSigAmpl;

            mMaxSigAmpl = max;
            mMinSigAmpl = min;
//qDebug() << "Pre: mMaxSigAmpl =" << mMaxSigAmpl << "- mMinSigAmpl =" << mMinSigAmpl << "- mAmplitudeStep =" << mAmplitudeStep << "- mAmplitudeRange =" << mAmplitudeRange << "- up =" << up;
            calcAmplitudeScale( up || !doCheck );
//qDebug() << "Post: mMaxSigAmpl =" << mMaxSigAmpl << "- mMinSigAmpl =" << mMinSigAmpl << "- mAmplitudeStep =" << mAmplitudeStep << "- mAmplitudeRange =" << mAmplitudeRange;

//            mYScaleTexts.clear();       // Force updating the calculation of the Y-scale text sizes

//            updateWithBackground();     // Update the background of this item

            adjustMarkerExtent();

            changed = true;
        }
    }

    return changed;
}


bool PSignalPlotBase::setTimeRange( double min, double max, bool doCheck )
{
    bool changed = false;

    if ( ( min < max ) &&
         ( (mZoomT0 != min) || (mZoomT1 != max) ) )
    {
        if ( !doCheck ||
             (fabs(max - min) >= fabs(mTimeStep)) )     // Allow the time range to shrink to minimum of one step
        {
            bool up = fabs( mZoomT1 - mZoomT0 ) <= fabs( max - min ) || !doCheck;

//qDebug() << "Init: max =" << max << "- min =" << min << "- doCheck =" << doCheck << "- mZoomT1 =" << mZoomT1 << "- mZoomT0 =" << mZoomT0;

            mZoomT0 = min;
            mZoomT1 = max;

//qDebug() << "Pre: mZoomT1 =" << mZoomT1 << "- mZoomT0 =" << mZoomT0 << "- mTimeStep =" << mTimeStep << "- mTimeRange =" << mTimeRange << "- up =" << up;
            calcTimeScale( up );
//qDebug() << "Post: mZoomT1 =" << mZoomT1 << "- mZoomT0 =" << mZoomT0 << "- mTimeStep =" << mTimeStep << "- mTimeRange =" << mTimeRange;

//            mXScaleTexts.clear();       // Force updating the calculation of the X-scale text sizes

//            updateWithBackground();     // Update the background of this item

            adjustMarkerExtent();

            changed = true;
        }
    }

    return changed;
}


void PSignalPlotBase::setMaximizeGraph( bool maximize )
{
    if ( mMaximizeGraph != maximize )
    {
        mMaximizeGraph = maximize;

        forceRedrawSignalsAndMarkers();

        mItemSizes.clear();         // Force updating the calculation of the text sizes

        updateWithBackground();         // Update the background of this item
    }
}


void PSignalPlotBase::calculateTextSizes( QPainter * painter )
{
    QFontMetrics    metrics = painter->fontMetrics();
    QRectF          textBound;
    QRectF          graphRect( mBounds );   // Default draw the graph within the whole item

    if ( !mText.isEmpty() && !mMaximizeGraph )
    {
        painter->setFont( mTextFont );

        metrics = painter->fontMetrics();

        // The text item will be position at the top of the graph with text centered
        textBound   = metrics.boundingRect( mText );
        textBound.moveTopLeft( QPointF( 0, 0 ) );
        textBound = textBound.intersected( mBounds );   // Make sure only to draw inside the items rect
        if ( textBound.width() < getBounds().width() )
        {
            textBound.setWidth( getBounds().width() );      // Make the text being drawn centered in the item
        }
        textBound.moveTopLeft( pos() );             // Move to desired location

        graphRect.setTop( textBound.height() );     // Remove the top text from the graph drawing space

        mItemSizes.append( textBound );
    }
    else
    {
        mItemSizes.append( QRect() );
    }

    if ( !mXLabel.isEmpty() && !mMaximizeGraph )
    {
        painter->setFont( mLabelFont );

        metrics = painter->fontMetrics();

        // The text item will be position at the bottom of the graph with text right aligned
        textBound   = metrics.boundingRect( mXLabel );
        textBound.moveTopLeft( QPointF( 0, 0 ) );
        textBound.setWidth( mBounds.width() );         // Make sure only to draw inside the items rect
        //textBound = textBound.intersected( mBounds );   // Make sure only to draw inside the items rect
        textBound.moveTopLeft( pos() +
                               QPointF( getBounds().width() - textBound.width(),
                                        getBounds().height() - textBound.height() ) );    // Move to desired location

        graphRect.setHeight( graphRect.height() - textBound.height() );     // Adjust the graph drawing space for the X-label

        mItemSizes.append( textBound );
    }
    else
    {
        mItemSizes.append( QRect() );
    }

    if ( !mYLabel.isEmpty() && !mMaximizeGraph )
    {
        painter->setFont( mLabelFont );

        metrics = painter->fontMetrics();

        // The text item will be position at the right side of the graph with text right aligned and 90 degree rotated
        textBound   = metrics.boundingRect( mYLabel );
        textBound.moveTopLeft( QPointF( 0, 0 ) );

        textBound.setWidth( mBounds.height() );         // Make sure only to draw inside the items rect
        if ( textBound.height() > mBounds.width() )     // We have to correct this manually because the
        {                                               // label will be drawn with 90 degrees shift
            textBound.setHeight( mBounds.width() );
        }

        textBound = QRectF( 0, textBound.width(), textBound.height(), -textBound.width() );  // Change to a vertical version
        if ( !mItemSizes[SIZES_INDEX_TEXT].isNull() )
        {
            // Adjust for the top position text item. This label should stay underneath it
            textBound.setHeight( textBound.height() + mItemSizes[SIZES_INDEX_TEXT].height() );
        }
        textBound = QRectF( pos().x(), pos().y() + textBound.y(), textBound.width(), textBound.height() );

        graphRect.setTopLeft( QPointF( graphRect.left() + textBound.width(), graphRect.top() ) );     // Adjust the graph drawing space for the Y-label

        mItemSizes.append( textBound );
    }
    else
    {
        mItemSizes.append( QRect() );
    }

    if ( !mXUnit.isEmpty() && !mMaximizeGraph )
    {
        painter->setFont( mScaleFont );

        metrics = painter->fontMetrics();

        // The text item will be position over the X-label in the graph with text right aligned
        textBound   = metrics.boundingRect( mXUnit );
        textBound.moveTopLeft( QPointF( 0, 0 ) );

        textBound.setWidth( mBounds.width() );         // Make sure only to draw inside the items rect
        if ( !mItemSizes[SIZES_INDEX_LABEL_Y].isNull() )
        {
            // Adjust for the Y-label position text item. This unit label should stay on the right side of it
            textBound.moveTopLeft( QPointF( mItemSizes[SIZES_INDEX_LABEL_Y].width(), 0 ) );
        }
        //textBound = textBound.intersected( mBounds );   // Make sure only to draw inside the items rect
        if ( !mItemSizes[SIZES_INDEX_LABEL_X].isNull() )
        {
            // Adjust for the Y-label position text item. This unit label should stay on the right side of it
            // Also the unit should stay above the X-label
            textBound.moveTopLeft( pos() +
                                   QPointF( getBounds().width() - textBound.width(),
                                            getBounds().height() - textBound.height() - mItemSizes[SIZES_INDEX_LABEL_X].height() ) );    // Move to desired location
        }
        else
        {
            // Adjust for the Y-label position text item. This unit label should stay on the right side of it
            textBound.moveTopLeft( pos() +
                                   QPointF( getBounds().width() - textBound.width(),
                                            getBounds().height() - textBound.height() ) );    // Move to desired location
        }

        graphRect.setHeight( graphRect.height() - textBound.height() );     // Adjust the graph drawing space for the X-unit

        mItemSizes.append( textBound );
    }
    else
    {
        mItemSizes.append( QRect() );
    }

    if ( !mYUnit.isEmpty() && !mMaximizeGraph )
    {
        painter->setFont( mScaleFont );

        metrics = painter->fontMetrics();

        // The text item will be position to the right of the Y-label in the graph with text right aligned and 90 degree rotated
        textBound   = metrics.boundingRect( mYUnit );
        textBound.moveTopLeft( QPointF( 0, 0 ) );

        textBound.setWidth( mBounds.height() );         // Make sure only to draw inside the items rect
        if ( textBound.height() > mBounds.width() )     // We have to correct this manually because the
        {                                               // unit text will be drawn with 90 degrees shift
            textBound.setHeight( mBounds.width() );
        }

        textBound = QRectF( 0, textBound.width(), textBound.height(), -textBound.width() );  // Change to a vertical version
        if ( !mItemSizes[SIZES_INDEX_TEXT].isNull() )
        {
            // Adjust for the top position text item. This unit text should stay underneath it
            textBound.setHeight( textBound.height() + mItemSizes[SIZES_INDEX_TEXT].height() );
        }
        if ( !mItemSizes[SIZES_INDEX_LABEL_Y].isNull() )
        {
            textBound = QRectF( pos().x() + mItemSizes[SIZES_INDEX_LABEL_Y].width(), pos().y() + textBound.y(), textBound.width(), textBound.height() );
        }
        else
        {
            textBound = QRectF( pos().x(), pos().y() + textBound.y(), textBound.width(), textBound.height() );
        }

        graphRect.setTopLeft( QPointF( graphRect.left() + textBound.width(), graphRect.top() ) );     // Adjust the graph drawing space for the Y-label

        mItemSizes.append( textBound );
    }
    else
    {
        mItemSizes.append( QRect() );
    }

    // Now store the area where the graph with axis unit values must be drawn
    graphRect = QRectF( pos().x() + graphRect.left(),
                        pos().y() + graphRect.top(),
                        graphRect.width(),
                        graphRect.height() );
    mItemSizes.append( graphRect );         // Move to desired location
}


void PSignalPlotBase::paint( QPainter * /*painter*/,
                             const QStyleOptionGraphicsItem * /*option*/,
                             QWidget * /*widget = 0*/ )
{
    // Reset this flag to enable handling of mouse move events.
    mWaitForDraw = false;
}


void PSignalPlotBase::paintGrid( QPainter * painter, const QRectF & updateRect )
{
    QPen            pen( painter->pen() );
    int             nLines      = 5;
    int             index       = 0;
    int             indexVal    = 0;
    double          timeStep    = mTimeStep / (double)nLines;
    int             timeSteps   = ceil((mZoomT1 - mZoomT0) / timeStep);
    double          amplStep    = mAmplitudeStep / (double) nLines;
    int             amplSteps   = ceil((mMaxSigAmpl - mMinSigAmpl) / amplStep);
    QList <double>  x;
    QList <double>  y;
    QFontMetrics    metrics     = painter->fontMetrics();
    double          pixelSize   = getPixelSize( painter );
    QRectF          prevItemRect;
    QRectF          textRect;

    painter->save();

    painter->setClipping( true );
    painter->setClipRect( updateRect.intersected( sceneBoundingRect() ) );

    pen.setWidthF( pixelSize );
    if ( mIsPrinting && mTextColorPrint.isValid() )
    {
        pen.setColor( mTextColorPrint );
    }
    else
    {
        pen.setColor( mTextColor );
    }
    painter->setPen( pen );

//qDebug() << "mZoomT0" << mZoomT0 << "- mZoomT1" << mZoomT1 << "- mTimeStep" << mTimeStep;
    Q_ASSERT_X( (mZoomT1 > mZoomT0 && mTimeStep > 0) ||
                (mTimeStep != 0) ||
                (mZoomT1 < mZoomT0 && mTimeStep < 0), "paintGrid: X-as", "schaal" );

//qDebug() << "mMinSigAmpl" << mMinSigAmpl << "- mMaxSigAmpl" << mMaxSigAmpl << "- mAmplitudeStep" << mAmplitudeStep;
    Q_ASSERT_X( (mMaxSigAmpl > mMinSigAmpl && mAmplitudeStep > 0) ||
                (mAmplitudeStep != 0) ||
                (mMaxSigAmpl < mMinSigAmpl && mAmplitudeStep < 0), "paintGrid: Y-as", "schaal" );

    // First get the maximum size of the unit values along the X-axis
    if ( mXScaleTexts.isEmpty() )
    {
        painter->setFont( mScaleFont );

        metrics = painter->fontMetrics();

        mXScaleTextRect = QRectF();

        for ( index = 0; index <= timeSteps; index++ )
        {
            double timeVal = (mZoomT0 + index * timeStep);

            // when calculating the steps with doubles there can remain a very small number
            // in the value that should 0.0. To prevent showing a number of 165e-19 we correct
            // this value for this.
            if ( fabs( timeVal ) < fabs( timeStep ) / 10000.0 )
            {
                timeVal = 0.0;
            }

            if ( (index % (2 * nLines) == 0) /*&& (timeVal != 0)*/ )
            {
                mXScaleTexts.append( QString( "%1" ).arg( timeVal / mXLabelFactor ) );
                textRect = metrics.boundingRect( mXScaleTexts.last() );

                if ( mXScaleTextRect.height() < textRect.height() )
                {
                    mXScaleTextRect.setHeight( textRect.height() );
                }
                if ( mXScaleTextRect.width() < textRect.width() )
                {
                    mXScaleTextRect.setWidth( textRect.width() );
                }
            }
        }
    }

    // Second get the maximum size of the unit values along the Y-axis
    if ( mYScaleTexts.isEmpty() )
    {
        painter->setFont( mScaleFont );

        metrics = painter->fontMetrics();

        mYScaleTextRect = QRectF();   // Just to be sure

        mYScaleTextRect = metrics.boundingRect( "0000" );

        for ( index = 0; index <= amplSteps; index++ )
        {
            if ( index % (2 * nLines) == 0 )
            {
                double amplitudeVal = (mMinSigAmpl + index * amplStep);

                // when calculating the steps with doubles there can remain a very small number
                // in the value that should 0.0. To prevent showing a number of 165e-19 we correct
                // this value for this.
                if ( fabs( amplitudeVal ) < fabs( mAmplitudeStep ) / 10000.0 )
                {
                    amplitudeVal = 0.0;
                }

                mYScaleTexts.append( QString( "%1" ).arg( amplitudeVal / mYLabelFactor ) );
                textRect = metrics.boundingRect( mYScaleTexts.last() );

                if ( mYScaleTextRect.height() < textRect.height() )
                {
                    mYScaleTextRect.setHeight( textRect.height() );
                }
                if ( mYScaleTextRect.width() < textRect.width() )
                {
                    mYScaleTextRect.setWidth( textRect.width() );
                }
            }
        }
    }

    // Third get the rectangles of all text items that needs to be drawn inside the graph
    if ( mItemSizes.isEmpty() )
    {
        // Re-calculate the shown text sizes
        calculateTextSizes( painter );
    }

    Q_ASSERT( mItemSizes.count() == SIZES_INDEX_TOTAL );
    Q_ASSERT( !mItemSizes[SIZES_INDEX_GRAPH].isNull() );

    // Finally, get the rectangle where the graph must be drawn.
    // First in scene coordinates to be used within this function
    QRectF sceneGraphRect = mItemSizes[SIZES_INDEX_GRAPH];  // Initialyze with the rectangle that will hold the graph and the axis unit values
    if ( !mMaximizeGraph )
    {
        Q_ASSERT( !mXScaleTextRect.isNull() );
        Q_ASSERT( !mYScaleTextRect.isNull() );

        sceneGraphRect.setLeft( sceneGraphRect.left() + mYScaleTextRect.width() + pixelSize );
        sceneGraphRect.setBottom( sceneGraphRect.bottom() - mXScaleTextRect.height() );
        // Adjust for the last scale value in the graph so they will show up correctly
        sceneGraphRect.setTop( sceneGraphRect.top() + (mYScaleTextRect.height() / 2) );
        sceneGraphRect.setWidth( sceneGraphRect.width() - (mXScaleTextRect.width() / 2) );
    }
    // Now in item rectangle coordinates
    mGraphRect = sceneGraphRect;
    mGraphRect.moveTop( mGraphRect.top() - pos().y() );
    mGraphRect.moveLeft( mGraphRect.left() - pos().x() );

    mGraphTransform = QTransform( toItemTime( 2.0 ) - toItemTime( 1.0 ),
                                  0,
                                  0,
                                  toItemAmplitude( 1.0 ) - toItemAmplitude( 0.0 ),
                                  toItemTime( 0.0 ),
                                  toItemAmplitude( 0.0 ) );

////Testcode
//    QBrush bgBrush( Qt::cyan );
//    painter->fillRect( updateRect, bgBrush );

////Testcode
//    if ( !mItemSizes[SIZES_INDEX_GRAPH].isNull() )
//    {
//        QBrush grBrush( Qt::red );
//        painter->fillRect( mItemSizes[SIZES_INDEX_GRAPH], grBrush );
//    }

    if ( mIsPrinting && mBackGroundColorPrint.isValid() )
    {
// No test code
        QBrush grbgBrush( mBackGroundColorPrint );
        painter->fillRect( sceneGraphRect, grbgBrush );
    }
    else if ( !mIsPrinting && mBackGroundColor.isValid() )
    {
// No test code
        QBrush grbgBrush( mBackGroundColor );
        painter->fillRect( sceneGraphRect, grbgBrush );
    }
//    else
//    {
////Testcode
//        QBrush grhBrush( Qt::yellow );
//        painter->fillRect( sceneGraphRect, grhBrush );
//    }

    if ( !mText.isEmpty() && !mItemSizes[SIZES_INDEX_TEXT].isNull() && !mMaximizeGraph )
    {
        painter->setFont( mTextFont );
        painter->drawText( mItemSizes[SIZES_INDEX_TEXT], Qt::AlignCenter, mText );
    }

    if ( !mXLabel.isEmpty() && !mItemSizes[SIZES_INDEX_LABEL_X].isNull() && !mMaximizeGraph )
    {
        painter->setFont( mLabelFont );
        painter->drawText( mItemSizes[SIZES_INDEX_LABEL_X], Qt::AlignVCenter | Qt::AlignRight, mXLabel );
    }

    if ( !mYLabel.isEmpty() && !mItemSizes[SIZES_INDEX_LABEL_Y].isNull() && !mMaximizeGraph )
    {
        painter->setFont( mLabelFont );
        painter->rotate( -90 );
        QRectF rotated( -mItemSizes[SIZES_INDEX_LABEL_Y].y(),
                        mItemSizes[SIZES_INDEX_LABEL_Y].x(),
                        -mItemSizes[SIZES_INDEX_LABEL_Y].height(),
                        mItemSizes[SIZES_INDEX_LABEL_Y].width() );
        painter->drawText( rotated, Qt::AlignVCenter | Qt::AlignRight, mYLabel );
        painter->rotate( 90 );
    }

    if ( !mXUnit.isEmpty() && !mItemSizes[SIZES_INDEX_UNIT_X].isNull() && !mMaximizeGraph )
    {
        painter->setFont( mScaleFont );
        painter->drawText( mItemSizes[SIZES_INDEX_UNIT_X], Qt::AlignVCenter | Qt::AlignRight, mXUnit );
    }

    if ( !mYUnit.isEmpty() && !mItemSizes[SIZES_INDEX_UNIT_Y].isNull() && !mMaximizeGraph )
    {
        painter->setFont( mScaleFont );
        painter->rotate( -90 );
        QRectF rotated( -mItemSizes[SIZES_INDEX_UNIT_Y].y(),
                        mItemSizes[SIZES_INDEX_UNIT_Y].x(),
                        -mItemSizes[SIZES_INDEX_UNIT_Y].height(),
                        mItemSizes[SIZES_INDEX_UNIT_Y].width() );
        painter->drawText( rotated, Qt::AlignVCenter | Qt::AlignRight, mYUnit );
        painter->rotate( 90 );
    }

    painter->setFont( mScaleFont );

    // Draw the Y-axis unit values
    for ( index = 0; index <= amplSteps; index++ )
    {
        double amplitudeVal = (mMinSigAmpl + index * amplStep);

        // when calculating the steps with doubles there can remain a very small number
        // in the value that should 0.0. To prevent showing a number of 165e-19 we correct
        // this value for this.
        if ( fabs( amplitudeVal ) < fabs( mAmplitudeStep ) / 10000.0 )
        {
            amplitudeVal = 0.0;
        }

        // Store the position of the Y-axis scale grid lines
        y.append( toItemAmplitude( amplitudeVal ) + pos().y() );

        if ( !mYScaleTextRect.isNull() && mShowYScaleTexts && (index % (2 * nLines) == 0) && !mMaximizeGraph )
        {
            Q_ASSERT( indexVal < mYScaleTexts.size() );

            QRectF itemRect( mItemSizes[SIZES_INDEX_GRAPH].x(),
                             y.last() - (mYScaleTextRect.height() / 2),
                             mYScaleTextRect.width(),
                             mYScaleTextRect.height() );

            // Now draw, only when there is text rectangle created and when this rectangle doesn't intersect with a previous drawn rectangle
            //if ( mItemSizes[SIZES_INDEX_GRAPH].contains( itemRect ) )
            if ( indexVal < mYScaleTexts.size() && (!prevItemRect.isValid() || !prevItemRect.intersects( itemRect )) )
            {
                painter->drawText( itemRect,
                                   Qt::AlignVCenter | Qt::AlignRight,
                                   mYScaleTexts[indexVal] );

                prevItemRect = itemRect;
            }

            indexVal++;
        }
    }

    indexVal = 0;

    for ( index = 0; index <= timeSteps; index++ )
    {
        double timeVal = (mZoomT0 + index * timeStep);

        // when calculating the steps with doubles there can remain a very small number
        // in the value that should 0.0. To prevent showing a number of 165e-19 we correct
        // this value for this.
        if ( fabs( timeVal ) < fabs( timeStep ) / 10000.0 )
        {
            timeVal = 0.0;
        }

        // Store the position of the X-axis scale grid lines
        x.append( toItemTime( timeVal ) + pos().x() );

        if ( !mXScaleTextRect.isNull() && mShowXScaleTexts && !mMaximizeGraph && (index % (2 * nLines) == 0) /*&& (timeVal != 0)*/ )
        {
            // Don't show the first value on the X-axis because it will be drawn a bit through the Y-axis value
            if ( 0 < index )
            {
                Q_ASSERT( indexVal < mXScaleTexts.size() );

                QRectF itemRect( x.last() - mXScaleTextRect.width() / 2,
                                 sceneGraphRect.bottom(),
                                 mXScaleTextRect.width(),
                                 mXScaleTextRect.height() );

                // Now draw, only when there is text rectangle created and when this rectangle doesn't intersect with a previous drawn rectangle
                //if ( mItemSizes[SIZES_INDEX_GRAPH].contains( itemRect ) )
                if ( indexVal < mXScaleTexts.size() && (!prevItemRect.isValid() || !prevItemRect.intersects( itemRect )) )
                {
                    painter->drawText( itemRect,
                                       Qt::AlignHCenter | Qt::AlignTop,
                                       mXScaleTexts[indexVal] );

                    prevItemRect = itemRect;
                }
            }

            indexVal++;
        }
    }

    if ( ( mGrid != Widget::NO_GRID ) /*&&
         ( pixelSize < 2.2 )*/ )  // Don't draw the grid when the user sized the window to small
    {
        if ( ( mGrid == Widget::SHOW_FULL_GRID ) /*&& ( pixelSize < 0.5 )*/ )
        {
            if ( mIsPrinting && mGridPointPrint.isValid() )
            {
                pen.setColor( mGridPointPrint );
            }
            else
            {
                pen.setColor( mGridPoint );
            }
            pen.setWidthF( pixelSize );

            painter->setPen( pen );

            for ( int i = 0; i < x.size(); i++ )
            {
                if ( i % nLines != 0 )
                {
                    for ( int j = 0; j < y.size(); j++ )
                    {
                        if ( j % nLines != 0 )
                        {
                            painter->drawPoint( QPointF(x[i], y[j]) );
                        }
                    }
                }
            }
        }

        if ( mIsPrinting && mGridLinePrint.isValid() )
        {
            pen.setColor( mGridLinePrint );
        }
        else
        {
            pen.setColor( mGridLine );
        }
        pen.setWidthF( pixelSize );

        painter->setPen( pen );

        if ( 0 < x.size() )
        {
            for ( int i = 0; i < y.size(); i++ )
            {
                if ( i % nLines == 0 )
                {
                    // We must use the extra QLineF here to prevent rounding off errors; the
                    // drawline function with only the single x and y values takes only integers
                    QLineF upperline( x.first(), y[i], x.last() , y[i] );

                    painter->drawLine( upperline );
                }
            }
        }

        if ( 0 < y.size() )
        {
            for ( int j = 0; j < x.size(); j++ )
            {
                if ( j % nLines == 0 )
                {
                    QLineF line( x[j], y.first(), x[j] , y.last());

                    painter->drawLine( line );
                }
            }
        }
    }

    // Draw the zero amplitude line
    if ( mMinSigAmpl <= 0 && mMaxSigAmpl >= 0 )
    {
        if ( mIsPrinting && mZeroLinePrint.isValid() )
        {
            pen.setColor( mZeroLinePrint );
        }
        else
        {
            pen.setColor( mZeroLine );
        }
        pen.setWidthF( pixelSize * 2.0 );
        painter->setPen( pen );

        QPointF xPoint = toItemCoor( mZoomT0, 0.0 ) + pos();
        QPointF yPoint = toItemCoor( mZoomT1, 0.0 ) + pos();

        painter->drawLine( xPoint, yPoint );
    }

    // Draw the zero time line
    if ( mShowYZeroLine && mZoomT0 <= 0 && mZoomT1 >= 0 )
    {
        if ( mIsPrinting && mZeroLinePrint.isValid() )
        {
            pen.setColor( mZeroLinePrint );
        }
        else
        {
            pen.setColor( mZeroLine );
        }
        pen.setWidthF( pixelSize * 2.0 );
        painter->setPen( pen );

        QPointF xPoint = toItemCoor( 0.0, mMinSigAmpl ) + pos();
        QPointF yPoint = toItemCoor( 0.0, mMaxSigAmpl ) + pos();

        painter->drawLine( xPoint, yPoint );
    }

    if ( !mInGraphText.isEmpty() )
    {
        if ( mIsPrinting && mTextColorPrint.isValid() )
        {
            pen.setColor( mTextColorPrint );
        }
        else
        {
            pen.setColor( mTextColor );
        }
        painter->setPen( pen );
        painter->setFont( mTextFont );

        metrics = painter->fontMetrics();

        textRect = metrics.boundingRect( mInGraphText );

        textRect.moveTop( sceneGraphRect.top() + pixelSize );
        textRect.setLeft( sceneGraphRect.left() + pixelSize );
        textRect.setWidth( sceneGraphRect.width() - pixelSize );

        painter->drawText( textRect, Qt::AlignCenter, mInGraphText );
    }

    painter->restore();
}


void PSignalPlotBase::mousePressEvent( QGraphicsSceneMouseEvent * event )
{
    if ( event->button() == Qt::LeftButton )
    {
        event->accept();

        if ( (mMarkers.count() > 0) &&
             mGraphRect.contains( event->pos() ) )
        {
            QPointF posInDataCoord  = getGraphTransform().inverted().map( event->pos() );
            int     marker          = mMarkers.findMarker( posInDataCoord, getPixelSize() * 10.0 );

            mMarkers.setSelectedMarker( marker );
        }
    }
    else
    {
        QGraphicsObject::mousePressEvent( event );
    }
}


void PSignalPlotBase::mouseMoveEvent( QGraphicsSceneMouseEvent * event )
{
    if ( mMarkers.hasSelectedMarker() )
    {
        QPointF posInDataCoord = getGraphTransform().inverted().map( event->pos() );

        if ( !mWaitForDraw )
        {
            // Set his flag to wait for the update paint event of this handler before handling it again
            mWaitForDraw = true;

            if ( !adjustMarkerPosition( posInDataCoord, event->pos(), false ) )
            {
                mWaitForDraw = false;   // Reset again. The new value isn't used
            }
        }

        event->accept();
    }
    else if ( event->buttons().testFlag( Qt::LeftButton ) && mAllowZoomSelect )
    {
        mSelectRect = QRectF( event->pos(), event->buttonDownPos( Qt::LeftButton ) ).intersected( mGraphRect );

        // Check if the mouse move has reached at least the minimum size we want for the tracking rectangle
        if ( SELECT_RECT_MINIMUM_SIZE >= mSelectRect.width() && SELECT_RECT_MINIMUM_SIZE >= mSelectRect.height() )
        {
            // Make invalid again
            mSelectRect.setWidth( 0 );
            mSelectRect.setHeight( 0 );
        }

        scene()->invalidate( mapRectToScene( mGraphRect ), QGraphicsScene::ForegroundLayer );

        event->accept();
    }
    else
    {
        QGraphicsObject::mouseMoveEvent( event );
    }
}


void PSignalPlotBase::mouseReleaseEvent( QGraphicsSceneMouseEvent * event )
{
    if ( event->button() == Qt::LeftButton )
    {
        if ( (mMarkers.count() > 0) &&
             mGraphRect.contains( event->pos() ) &&
             ( event->pos() == event->buttonDownPos( Qt::LeftButton ) || mMarkers.hasSelectedMarker() ) )
        {
            QPointF posInDataCoord = getGraphTransform().inverted().map( event->pos() );

            adjustMarkerPosition( posInDataCoord, event->pos(), true );

            event->accept();
        }
        else if ( mAllowZoomSelect && mSelectRect.isValid() )
        {
            // First get the selected rectangle in item coordinates with borders set to the items bounding rectangle
            mSelectRect = QRectF( event->pos(), event->buttonDownPos( Qt::LeftButton ) ).intersected( mGraphRect );

            // Check if the mouse move has reached at least the minimum size we want for the tracking rectangle
            if ( SELECT_RECT_MINIMUM_SIZE < mSelectRect.width() && SELECT_RECT_MINIMUM_SIZE < mSelectRect.height() )
            {
                // Now transform into data system coordinates
                mSelectRect = getGraphTransform().inverted().mapRect( mSelectRect );

                if ( mIsSymetric )
                {
                    qreal   maxVal;
                    double  minX = mSignals.getMinXData();
                    double  maxX = mSignals.getMaxXData();
                    double  minY = mSignals.getMinYData();
                    double  maxY = mSignals.getMaxYData();

                    if ( (0 > minX) && (0 < maxX) )
                    {
                        // The X-axis can be symmetric
                        maxVal = qMax( fabs( mSelectRect.left() ), fabs( mSelectRect.right() ) );

                        mSelectRect.setLeft( -maxVal );
                        mSelectRect.setWidth( 2 * maxVal );
                    }

                    if ( (0 > minY) && (0 < maxY) )
                    {
                        // The Y-axis can be symmetric
                        maxVal = qMax( fabs( mSelectRect.top() ), fabs( mSelectRect.bottom() ) );

                        mSelectRect.setTop( -maxVal );
                        mSelectRect.setHeight( 2 * maxVal );
                    }
                }

                if ( mSelectRect.left() < mSelectRect.right() )
                {
                    setTimeRange( ( mZoomT0 > mSelectRect.left() ? mZoomT0 : mSelectRect.left() ),
                                  ( mZoomT1 < mSelectRect.right() ? mZoomT1 : mSelectRect.right() ),
                                  false );
                }
                else
                {
                    setTimeRange( ( mZoomT0 > mSelectRect.right() ? mZoomT0 : mSelectRect.right() ),
                                  ( mZoomT1 < mSelectRect.left() ? mZoomT1 : mSelectRect.left() ),
                                  false );
                }

                if ( mSelectRect.top() < mSelectRect.bottom() )
                {
                    setSignalAmpl( mSelectRect.top(), mSelectRect.bottom(), false );
                }
                else
                {
                    setSignalAmpl( mSelectRect.bottom(), mSelectRect.top(), false );
                }

                event->accept();
            }
            else
            {
                // Update the view to make sure that the eventually drawn selection rectangle is removed
                scene()->invalidate( mapRectToScene( mGraphRect ), QGraphicsScene::ForegroundLayer );
            }
        }
    }
    else
    {
        QGraphicsObject::mouseReleaseEvent( event );
    }

    mMarkers.resetSelectedMarker();

    mSelectRect = QRectF();
}


void PSignalPlotBase::wheelEvent( QGraphicsSceneWheelEvent * event )
{
    // Only use the information when scrolled inside the graph itself
    if ( mGraphRect.contains( event->pos() ) )
    {
        //double fact = pow( 1.25, (double)event->delta() / 120.0 );    // Can be used when multiply the extent values
        double fact = 0.0;

        if ( event->modifiers() & Qt::ControlModifier )
        {
            if ( mDoAdaptXAxis )
            {
                fact = ( event->delta() > 0 ? 1 : -1 ) * mTimeStep;

                if ( mIsSymetric )
                {
                    if ( !mLimitAdaptXAxis ||
                         ( (mSignals.getMinXData() <= mZoomT0 + fact) && (mZoomT1 + fact <= mSignals.getMaxXData()) ) )
                    {
                        setTimeRange( mZoomT0 - fact, mZoomT1 + fact );
                    }
                }
                else if ( event->modifiers() & Qt::ShiftModifier )
                {
                    if ( !mLimitAdaptXAxis || (mSignals.getMinXData() <= mZoomT0 + fact) )
                    {
                        setTimeRange( mZoomT0 + fact, mZoomT1 );
                    }
                }
                else
                {
                    if ( !mLimitAdaptXAxis || (mZoomT1 + fact <= mSignals.getMaxXData()) )
                    {
                        setTimeRange( mZoomT0, mZoomT1 + fact );
                    }
                }
            }
        }
        else
        {
            if ( mDoAdaptYAxis )
            {
                fact = ( event->delta() > 0 ? -1 : 1 ) * mAmplitudeStep;
//qDebug() << "Before: mMaxSigAmpl =" << mMaxSigAmpl << "- mMinSigAmpl =" << mMinSigAmpl << "- mAmplitudeStep =" << mAmplitudeStep << "- mAmplitudeRange =" << mAmplitudeRange;
                if ( mIsSymetric )
                {
                    if ( !mLimitAdaptYAxis ||
                         ( (mSignals.getMinYData() <= mMinSigAmpl + fact) && (mMaxSigAmpl + fact <= mSignals.getMaxYData()) ) )
                    {
                        setSignalAmpl( mMinSigAmpl - fact, mMaxSigAmpl + fact );
                    }
                }
                else if ( event->modifiers() & Qt::ShiftModifier )
                {
                    if ( !mLimitAdaptYAxis || (mSignals.getMinYData() <= mMinSigAmpl + fact) )
                    {
                        setSignalAmpl( mMinSigAmpl + fact, mMaxSigAmpl );
                    }
                }
                else
                {
                    if ( !mLimitAdaptYAxis || (mMaxSigAmpl + fact <= mSignals.getMaxYData()) )
                    {
                        setSignalAmpl( mMinSigAmpl, mMaxSigAmpl + fact );
                    }
                }
            }
        }

        event->accept();
    }
    else
    {
        QGraphicsObject::wheelEvent( event );
    }
}


void PSignalPlotBase::hoverMoveEvent( QGraphicsSceneHoverEvent * event )
{
    QPointF posInDataCoord  = getGraphTransform().inverted().map( event->pos() );
    int     marker          = mMarkers.findMarker( posInDataCoord, getPixelSize() * 10.0 );
//qDebug() << "Marker hover event:" << marker << "- position:" << event->pos() << "- data:" << posInDataCoord;

    // The mouse is on top of a marker or perhaps left one. Lets inform derived classes
    onMarkerHoverEvent( marker );

    QGraphicsObject::hoverMoveEvent( event );
}


void PSignalPlotBase::hoverLeaveEvent( QGraphicsSceneHoverEvent * event )
{
    Q_UNUSED( event )

    onMarkerHoverEvent( PSignalMarkers::NO_MARKER_FOUND );

    QGraphicsObject::hoverLeaveEvent( event );
}


void PSignalPlotBase::setObjectColor( const QColor & color, bool print, QColor & paintClr, QColor  & printClr )
{
    if ( !print )
    {
        if ( paintClr != color )
        {
            paintClr = color;

            updateWithBackground();
        }
    }
    else
    {
        printClr = color;
    }
}


void PSignalPlotBase::changeAmplScale( double maxSigAmpl, double minSigAmpl, double amplitudeStep, double amplitudeRange )
{
    if ( maxSigAmpl     != mMaxSigAmpl      ||
         minSigAmpl     != mMinSigAmpl      ||
         amplitudeStep  != mAmplitudeStep   ||
         amplitudeRange != mAmplitudeRange  )
    {
        mMaxSigAmpl     = maxSigAmpl;
        mMinSigAmpl     = minSigAmpl;
        mAmplitudeStep  = amplitudeStep;
        mAmplitudeRange = amplitudeRange;

        //mItemSizes.clear();         // Force updating the calculation of the text sizes
        mYScaleTexts.clear();       // Force updating the calculation of the Y-scale text sizes

        for ( int sigNr = 0; sigNr < mSignals.count(); sigNr++ )
        {
            // Force a recalculation of the polygons to draw
            mSignals[ sigNr ].clearPoly();
        }

        for ( int markerNr = 0; markerNr < mMarkers.count(); markerNr++ )
        {
            // Force a recalculation of the marker position to draw
            mMarkers[ markerNr ]->clearPath();
        }

        adjustMarkerExtent();

        updateWithBackground();     // Update the background of this item
    }
}


void PSignalPlotBase::changeTimeScale( double maxTime, double minTime, double timeStep, double timeRange )
{
    if ( maxTime    != mZoomT1      ||
         minTime    != mZoomT0      ||
         timeStep   != mTimeStep    ||
         timeRange  != mTimeRange   )
    {
        mZoomT1     = maxTime;
        mZoomT0     = minTime;
        mTimeStep   = timeStep;
        mTimeRange  = timeRange;

        //mItemSizes.clear();         // Force updating the calculation of the text sizes
        mXScaleTexts.clear();       // Force updating the calculation of the X-scale text sizes

        for ( int sigNr = 0; sigNr < mSignals.count(); sigNr++ )
        {
            // Force a recalculation of the polygons to draw
            mSignals[ sigNr ].clearPoly();
        }

        for ( int markerNr = 0; markerNr < mMarkers.count(); markerNr++ )
        {
            // Force a recalculation of the marker position to draw
            mMarkers[ markerNr ]->clearPath();
        }

        if ( mXAxisShowsIndex )
        {
            // We can limit the data processed when drawing the signal
            mSignals.setZoomed( mZoomT0, mZoomT1 );
        }

        updateWithBackground();     // Update the background of this item
    }
}


void PSignalPlotBase::calcAmplitudeScale( bool up )
{
    mAmplitudeStep  = calcScaleStep( mMinSigAmpl,
                                     mMaxSigAmpl,
                                     qMax( fabs( mMaxSigAmpl ), fabs( mMinSigAmpl ) ) / mYfactor,
                                     up );

    mAmplitudeRange = mMaxSigAmpl - mMinSigAmpl;

//    mItemSizes.clear();         // Force updating the calculation of the text sizes
    mYScaleTexts.clear();       // Force updating the calculation of the Y-scale text sizes

    for ( int sigNr = 0; sigNr < mSignals.count(); sigNr++ )
    {
        // Force a recalculation of the polygons to draw
        mSignals[ sigNr ].clearPoly();
    }

    for ( int markerNr = 0; markerNr < mMarkers.count(); markerNr++ )
    {
        // Force a recalculation of the marker position to draw
        mMarkers[ markerNr ]->clearPath();
    }

    updateWithBackground();     // Update the background of this item

    emit amplScaleChanged();
}


void PSignalPlotBase::calcTimeScale( bool up )
{
    mTimeStep   = calcScaleStep( mZoomT0, mZoomT1, (mZoomT1 - mZoomT0) / mXfactor, up );

    mTimeRange  = mZoomT1 - mZoomT0;

    mXScaleTexts.clear();       // Force updating the calculation of the X-scale text sizes

    for ( int sigNr = 0; sigNr < mSignals.count(); sigNr++ )
    {
        // Force a recalculation of the polygons to draw
        mSignals[ sigNr ].clearPoly();
    }

    for ( int markerNr = 0; markerNr < mMarkers.count(); markerNr++ )
    {
        // Force a recalculation of the marker position to draw
        mMarkers[ markerNr ]->clearPath();
    }

    if ( mXAxisShowsIndex )
    {
        // We can limit the data processed when drawing the signal
        mSignals.setZoomed( mZoomT0, mZoomT1 );
    }

    updateWithBackground();     // Update the background of this item

    emit timeScaleChanged();
}


double PSignalPlotBase::getPixelSize( QPainter * painter ) const
{
    double pixelSize = 0.0;  // Just init with something

    QList <QGraphicsView *> viewList( scene()->views() );

    // Find the real pixel size transalted to this items
    if ( viewList.size() > 0 )
    {
        for ( int in = 0; in < viewList.size(); in++ )
        {
            QSize viewSize = viewList.at( in )->maximumViewportSize();

            pixelSize = qMax( pixelSize,
                              qMax( fabs( scene()->sceneRect().height() / viewSize.height() ),
                                    fabs( scene()->sceneRect().width() / viewSize.width() ) ) );
        }
    }
    else if ( painter != 0 )
    {
        // Use the less correct method (only valid when the whole scene is being updated) as fallback
        pixelSize   = qMax( fabs( scene()->sceneRect().height() / painter->device()->height() ),
                            fabs( scene()->sceneRect().width() / painter->device()->width() ) );
    }

    return pixelSize;
}


void PSignalPlotBase::adjustMarkerExtent( void )
{
    for ( int index = 0; index < mMarkers.size(); index++ )
    {
        PMarkerLine * markerPtr = marker_cast< PMarkerLine * >( mMarkers[index] );

        if ( markerPtr )
        {
            if ( !markerPtr->adjustForNewExtent( mZoomT0, mZoomT1, mMinSigAmpl, mMaxSigAmpl ) )
            {
                qWarning() << "PLineMarker object could not be adjusted for changed graph extent.";
            }
        }
        else if ( !( (marker_cast< PMarkerEllipse * >( mMarkers[index] ) != NULL) ||
                     (marker_cast< PMarkerVector * >( mMarkers[index] ) != NULL) ) )  // Unknown marker found
        {
            qWarning() << "Unknown marker object found while adjusting for changed graph extent.";
        }
    }
}
