/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalPlotBase.h 1172 2014-03-21 13:45:45Z eelco $
 *
 */


#ifndef __PSIGNALPLOTBASE_H__
#define	__PSIGNALPLOTBASE_H__


#include <QGraphicsObject>

#include <QPainter>
#include <QColor>
#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>
#include <QRectF>
#include <QPointF>
#include <QString>
#include <QFont>

#include "PGrItemBase.h"
#include "PSignalsData.h"
#include "PSignalMarkers.h"
#include "PeacsSupport_global.h"
#include "PWidgetDefines.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PSignalPlotBase PSignalPlotBase.h "widgets/PSignalPlotBase.h"
          * @brief The graphics scene item that holds the base functionality to show a graph
          */
        class PEACSSUPPORTSHARED_EXPORT PSignalPlotBase : public QGraphicsObject, public PGrItemBase
        {
            Q_OBJECT

        public:

            /** @brief Only present to be used in the type() function
              * @see PGrItemBase
              */
            enum { Type = PGrItemBase::PSignalPlotBaseType };


            /** @brief Returns the type of graphics item.
              * Enable the use of qgraphicsitem_cast with this item.
              * @see PGrItemBase
              */
            inline int type( void ) const { return Type; }


            /** @brief constructor
              *
              * @param xAxisShowsIndex If true, the X-axis shows index data, i.e. starting at 0 with intervals of 1. @see mXAxisShowsIndex
              * @param zoomT0 Specified the initial starting point, i.e. left extent of the X-axis
              * @param zoomT1 Specified the initial end point, i.e. right extent of the X-axis
              * @param column the column at which this item will reside in the matrix (default -1)
              * @param row the row at which this item will reside in the matrix (default -1)
              *
              * @see PGrItemBase for information about column and row
              */
            PSignalPlotBase( bool xAxisShowsIndex, double zoomT0, double zoomT1, int column = -1, int row = -1 );


            /** @brief destructor
              */
            virtual ~PSignalPlotBase();


            /** @brief initializes the class with the data provided before
              * This function must be called before displaying the desired data. It
              * will initialize local members and in a derived class implementation
              * specific data can be set, like labels.
              */
            virtual void init();


            /** @brief Sets the bounding (virtual) rectangle used when drawing
              * This internally used virtual drawing board bound will be used to draw within.
              *
              * @param width The new width to use.
              * @param height The new height to use.
              * @return true if the bounding rectangle has changed
              */
            virtual bool setBounds( qreal width, qreal height );


            /** @brief returns the outer bounds of the item as a rectangle.
              * All painting must be restricted to inside an item's bounding rect.
              */
            virtual QRectF boundingRect( void ) const
            {
                // Don't remove this comment. When removed the app crashes at an untraceable location (?????)
                //Q_ASSERT_X( !mBounds.isEmpty(), "boundingRect", "Use setBounds before calling this function" );
                return mBounds;
            }


            /** @brief Sets mXfactor
              *
              * @param factor the new value for mXfactor
              */
            void setXfactor( int factor );


            /** @brief returns mXfactor
              */
            inline int getXfactor( void ) const { return mXfactor; }


            /** @brief Sets mYfactor
              *
              * @param factor the new value for mYfactor
              */
            void setYfactor( int factor );


            /** @brief returns mYfactor
              */
            inline int getYfactor( void ) const { return mYfactor; }


            /** @brief Sets mXLabelFactor
              *
              * @param factor the new value for mXLabelFactor
              */
            void setXLabelFactor( double factor );


            /** @brief returns mXLabelFactor
              */
            inline double getXLabelFactor( void ) const { return mXLabelFactor; }


            /** @brief Sets mYLabelFactor
              *
              * @param factor the new value for mYLabelFactor
              */
            void setYLabelFactor( double factor );


            /** @brief returns mYLabelFactor
              */
            inline double getYLabelFactor( void ) const { return mYLabelFactor; }


           /** Returns member mGrid
             *
             * @return mGrid
             */
           inline Widget::Grid_type getGrid( void ) const { return mGrid; }


           /** Sets mGrid to the new grid setting to use
             *
             * Checks if the value is changed and scedules for an update accordingly
             *
             * @param grid the new grid setting to use
             *
             * @return true if the value has changed
             */
           bool setGrid( Widget::Grid_type grid );


           /** Returns member mTextFont
             *
             * @return constant reference to mTextFont
             */
           inline const QFont & getTextFont( void ) const { return mTextFont; }


           /** Sets mTextFont to the new font
             *
             * Checks if the value is changed and scedules for an update accordingly
             *
             * @param font the new font to use
             *
             * @return true if the value has changed
             */
           bool setTextFont( const QFont& font );


           /** Returns member mLabelFont
             *
             * @return constant reference to mLabelFont
             */
           inline const QFont & getLabelFont( void ) const { return mLabelFont; }


           /** Sets mLabelFont to the new font
             *
             * Checks if the value is changed and scedules for an update accordingly
             *
             * @param font the new font to use
             *
             * @return true if the value has changed
             */
           bool setLabelFont( const QFont& font );


           /** Returns member mScaleFont
             *
             * @return constant reference to mScaleFont
             */
           inline const QFont & getScaleFont( void ) const { return mScaleFont; }


           /** Sets mScaleFont to the new font
             *
             * Checks if the value is changed and scedules for an update accordingly
             *
             * @param font the new font to use
             *
             * @return true if the value has changed
             */
           bool setScaleFont( const QFont& font );


           /** @brief sets the unit shown near the X-axis
             * @param unit the new unit for the X-axis
             * @return true if the value has changed
             */
           bool setXUnit( const QString & unit );


           /** @brief sets the unit shown near the Y-axis
             * @param unit the new unit for the Y-axis
             * @return true if the value has changed
             */
           bool setYUnit( const QString & unit );


            /** @brief sets the label shown near the X-axis
              * @param label the new name X-axis label name
              * @return true if the value has changed
              */
           bool setXLabel( const QString & label );


            /** @brief sets the label shown near the Y-axis
              * @param label the new name Y-axis label name
              * @return true if the value has changed
              */
            bool setYLabel( const QString & label );


            /** @brief Sets an optional text shown above the graph
              * @param text the new text to show
              * @return true if the value has changed
              */
            bool setText( const QString & text );


            /** @brief Sets an optional text shown inside the graph area
              * @param text the new text to show
              * @return true if the value has changed
              */
            bool setInGraphText( const QString & text );


            /** @brief Returns mShowXScaleTexts
              * @return true if the scale texts along the X-axis should be shown
              */
            inline bool showXScaleTexts( void ) const { return mShowXScaleTexts; }


            /** @brief Sets mShowXScaleTexts
              * @param show the new value for mShowXScaleTexts
              */
            inline void setShowXScaleTexts( bool show ) { mShowXScaleTexts = show; }


            /** @brief Returns mShowYScaleTexts
              * @return true if the scale texts along the Y-axis should be shown
              */
            inline bool showYScaleTexts( void ) const { return mShowYScaleTexts; }


            /** @brief Sets mShowYScaleTexts
              * @param show the new value for mShowYScaleTexts
              */
            inline void setShowYScaleTexts( bool show ) { mShowYScaleTexts = show; }


            /** @brief Returns mShowYZeroLine
              * @return true if the zero line for the Y-axis should be shown
              */
            inline bool showYZeroLine( void ) const { return mShowYZeroLine; }


            /** @brief Sets mShowYZeroLine
              * @param show the new value for mShowYZeroLine
              */
            inline void setShowYZeroLine( bool show ) { mShowYZeroLine = show; }


            /** @brief Sets mZeroLine or mZeroLinePrint to the new color
              *
              * @param color the new color to use
              * @param print If true, the color used for printing (screen capture) is set, otherwise the
              *              screen painting color (default)
              */
            inline void setZeroLineColor( const QColor & color, bool print = false )
            { setObjectColor( color, print, mZeroLine, mZeroLinePrint ); }


            /** @brief Sets mGridLine or mGridLinePrint to the new color
              *
              * @param color the new color to use
              * @param print If true, the color used for printing (screen capture) is set, otherwise the
              *              screen painting color (default)
              */
            inline void setGridLineColor( const QColor & color, bool print = false )
            { setObjectColor( color, print, mGridLine, mGridLinePrint ); }


            /** @brief Sets mGridPoint or mGridPointPrint to the new color
              *
              * @param color the new color to use
              * @param print If true, the color used for printing (screen capture) is set, otherwise the
              *              screen painting color (default)
              */
            inline void setGridPointColor( const QColor & color, bool print = false )
            { setObjectColor( color, print, mGridPoint, mGridPointPrint ); }


            /** @brief Used to set mBackGroundColor or mBackGroundColorPrint to a (new) value
              * Default mBackGroundColor is initialized with no color specified, i.e. invalid. In
              * that case the scene background color is used
              * By specifying a (new) color that color will be used to fill the background
              * of the this graph item only and it will override the view background color
              *
              * @param color the (new) color to use
              * @param print If true, the color used for printing (screen capture) is set, otherwise the
              *              screen painting color (default)
              */
            inline void setBackGroundColor( const QColor & color, bool print = false )
            { setObjectColor( color, print, mBackGroundColor, mBackGroundColorPrint ); }


            /** @brief Sets mTextColor or mTextColorPrint to the new color
              *
              * @param color the new color to use
              * @param print If true, the color used for printing (screen capture) is set, otherwise the
              *              screen painting color (default)
              */
            inline void setTextColor( const QColor & color, bool print = false )
            { setObjectColor( color, print, mTextColor, mTextColorPrint ); }


            /** @brief Sets mSelectRectColor to the new color
              *
              * @param color the new color to use
              */
            inline void setSelectRectColor( const QColor & color ) { mSelectRectColor = color; }


            /** @brief Returns mSelectRectColor
              *
              * @return The color used to draw the zoom selection tracking rectangle
              */
            inline const QColor & getSelectRectColor( void ) { return mSelectRectColor; }


            /** @brief Sets mMinSigAmpl and mMaxSigAmpl and forces a total update
              * When the amplitude value have changed, calcAmplitudeScale() will be called
              * to update the scale values
              * This function also forces updating the item within the view
              *
              * @param min the new value for mMinSigAmpl
              * @param max the new value for mMaxSigAmpl
              * @param doCheck if true (default) checking of the new values is done before using them
              *
              * @return true if the amplitude has changed, otherwise false
              */
            bool setSignalAmpl( double min, double max, bool doCheck = true );


            /** @brief Returns mMaxSigAmpl
              */
            inline double getMaxSigAmpl( void ) const { return mMaxSigAmpl; }


            /** @brief Returns mMinSigAmpl
              */
            inline double getMinSigAmpl( void ) const { return mMinSigAmpl; }


            /** @brief Returns mAmplitudeRange
              */
            inline double getAmplitudeRange( void ) const { return mAmplitudeRange; }


            /** @brief Returns mAmplitudeStep
              */
            inline double getAmplitudeStep( void ) const { return mAmplitudeStep; }


            /** @brief Sets mZoomT0 and mZoomT1 and forces a total update
              * When the time scale value have changed, calcTimeScale() will be called
              * to update the scale values
              * This function also forces updating the item within the view
              *
              * @param min the new value for mZoomT0
              * @param max the new value for mZoomT1
              * @param doCheck if true (default) checking of the new values is done before using them
              *
              * @return true if the timescale has changed, otherwise false
              */
            bool setTimeRange( double min, double max, bool doCheck = true );


            /** Returns mZoomT0
              */
            inline double getZoomT0( void ) const { return mZoomT0; }


            /** Returns mZoomT1
              */
            inline double getZoomT1( void ) const { return mZoomT1; }


            /** @brief Returns mTimeStep
              * @return mTimeStep
              */
            inline double getTimeStep( void ) const { return mTimeStep; }


            /** @brief Returns mTimeRange
              * @return mTimeRange
              */
            inline double getTimeRange( void ) const { return mTimeRange; }


            /** @brief Sets mIsSymetric
              * The value of mIsSymetric is used when the used scrolls the mouse-wheel. When set,
              * the scrolling changes both the range of the X-axis and Y-axis at the same time.
              * Using the control modifier key will have no extra function in that case
              *
              * @param sym The new value for mIsSymetric
              */
            inline void setSysmetric( bool sym ) { mIsSymetric = sym; }


            /** @brief Sets mDoAdaptXAxis
              * If set to true, the scrollwheel can be used to adapt the range of the X-axis
              * The use of mDoAdaptXAxis is also influenced by mIsSymetric
              *
              * @param use the new value for mDoAdaptXAxis
              */
            inline void setDoAdaptXAxisOnScroll( bool use ) { mDoAdaptXAxis = use; }


            /** @brief Sets mLimitAdaptXAxis to the new value
              * If true, adapting the range of the X-axis will be limited to the data range
              * Default value is false, i.e. diabled and X-axis range can be changed unlimited
              *
              * @param limit The new value to use
              */
            inline void setLimmitAdaptXAxis( bool limit ) { mLimitAdaptXAxis = limit; }


            /** @brief Sets mDoAdaptYAxis
              * If set to true, the scrollwheel can be used to adapt the range of the Y-axis
              * The use of mDoAdaptYAxis is also influenced by mIsSymetric
              *
              * @param use the new value for mDoAdaptYAxis
              */
            inline void setDoAdaptYAxisOnScroll( bool use ) { mDoAdaptYAxis = use; }


            /** @brief Sets mLimitAdaptYAxis to the new value
              * If true, adapting the range of the Y-axis will be limited to the data range
              * Default value is false, i.e. diabled and Y-axis range can be changed unlimited
              *
              * @param limit The new value to use
              */
            inline void setLimmitAdaptYAxis( bool limit ) { mLimitAdaptYAxis = limit; }


            /** @brief Sets mMaximizeGraph to the new value
              * When true, all labels, texts and scale information are not shown and the graph
              * will occupy all space in the scene given to this item
              *
              * @param maximize The new value to use
              */
            void setMaximizeGraph( bool maximize );


            /** @brief Sets mAllowZoomSelect to the new value
              * If true, the mouse can be used to track a zoom ractangle. Default value is false
              *
              * @param allowZoom The new value to use
              */
            inline void setAllowZoomSelect( bool allowZoom ) { mAllowZoomSelect = allowZoom; }


            /** @brief Returns mInvertYDirection
              */
            inline bool isInvertY( void ) const { return mInvertYDirection; }


            /** @brief Sets mInvertYDirection
              *
              * @param invert the new value for mInvertYDirection
              */
            inline void setInvertY( bool invert ) { mInvertYDirection = invert; }


            /** @brief Returns mInvertYDirection
              */
            inline bool isInvertX( void ) const { return mInvertXDirection; }


            /** @brief Sets mInvertXDirection
              *
              * @param invert the new value for mInvertXDirection
              */
            inline void setInvertX( bool invert ) { mInvertXDirection = invert; }


            /** @brief compute the time and amplitude values from the given graphicsItem coordinates
             * @param pos the graphicsItem coordinates
             */
            inline QPointF toTimeAmplitude( QPointF pos ) const { return QPointF( toTime( pos.x() ), toAmplitude( pos.y() ) ); }


            /** @brief Recalcuates the mAmplitudeStep and mAmplitudeRange according mMinSigAmpl and mMaxSigAmpl
              * This function will be called after changing settings that influence the amplitude scale
              *
              * @param up flag that indicates how the functions floor() and ceil() are to be used. Default this value is true.
              *           This parameter is usefull when handling user input to change the scale. When the scale is to be made smaller
              *           this value should be set to false
              */
            virtual void calcAmplitudeScale( bool up = true );


            /** @brief Recalcuates the mTimeStep and mTimeRange according mZoomT0 and mZoomT1
              * This function will be called after changing settings that influence the time scale
              *
              * @param up flag that indicates how the functions floor() and ceil() are to be used. Default this value is true.
              *           This parameter is usefull when handling user input to change the scale. When the scale is to be made smaller
              *           this value should be set to false
              */
            virtual void calcTimeScale( bool up = true );


            /** @brief Returns the minimum data value that must be shown on the x-axis
              * For all signal that must be shown, the minimum value will be determained
              * @see PSignalsData::getMinXData()
              *
              * @return The minimum X-axis value to show
              */
            inline double getMinXData( void ) const { return mSignals.getMinXData(); }


            /** @brief Returns the maximum data value that must be shown on the x-axis
              * For all signal that must be shown, the maximum value will be determained
              * @see PSignalsData::getMaxXData()
              *
              * @return The maximum X-axis value to show
              */
            inline double getMaxXData( void ) const { return mSignals.getMaxXData(); }


            /** @brief Returns the minimum data value that must be shown on the y-axis
              * For all signal that must be shown, the minimum value will be determained
              * @see PSignalsData::getMinYData()
              *
              * @return The minimum Y-axis value to show
              */
            inline double getMinYData( void ) const { return mSignals.getMinYData(); }


            /** @brief Returns the maximum data value that must be shown on the y-axis
              * For all signal that must be shown, the maximum value will be determained
              * @see PSignalsData::getMaxYData()
              *
              * @return The maximum Y-axis value to show
              */
            inline double getMaxYData( void ) const { return mSignals.getMaxYData(); }


            /** @brief Returns the select rectangle
              * The returned rectangle will have a valid rectangle is the user is currently
              * selecting a rectangle. Otherwise and umpty (null) rectangle is returned
              *
              * @return mSelectRect
              */
            inline const QRectF & getSelectRect( void ) const { return mSelectRect; }


            /** @brief schedules this item for an update including the background
              * If the @link cacheMode is set to @link CacheBackground, the background
              * will not be updated with a simple call to @link update(). We have to add
              * this items rectangle to the appropriate member of the GraphicsView to force
              * updating the background
              *
              * @param layers defines the layers that must be invaluidated and thus be redrawn
              */
            void  updateWithBackground( QGraphicsScene::SceneLayers layers = QGraphicsScene::AllLayers );


            /** @brief Clears all signal and marker poly information
              * By doing so all polyline information used to draw the signals and markers is forced to be
              * recalculated
              */
            void forceRedrawSignalsAndMarkers( void );


            /** @brief Returns the final rectangle with which the graph itself will be drawn.
              * The returned rectangle is the items rectangle minus all text-items drawn around it
              *
              * This rectangle holds coordinates that are in graphics item coordinates, thus can be used for
              * drawing within this item
              *
              * @return mGraphRect
              */
            inline const QRectF & getGraphRect( void ) const { return mGraphRect; }


            /** @brief Returns the transformation matrix that can be used to transform data coordinates into graph coordinates
              *
              * @return mGraphTransform
              */
            inline const QTransform & getGraphTransform( void ) const { return mGraphTransform; }


            /** @brief Helper function that can be used to return the size of a pixel when drawing
              *
              * @param painter A painter object that can be used as a fallback within the function
              *
              * @return The pixel size for the current situation
              */
            double getPixelSize( QPainter * painter = 0 ) const;


            /** @brief Sets mSignals members mStartIndex and mStopIndex to the given values, i.e. zooming to a specified range
              * This function will check the given range with the maintained signals data, sets the range and clears
              * the polygons if valid
              *
              * @param startIndex The index in the signal data array where to start converting to polygon (default 0)
              * @param stopIndex The index in the signal data array where to stop converting to polygon  (default
              *        UINT_MAX, i.e. up to the end)
              * @param updateMinMax If true the minimum and maximum signal data values of the given range will be
              *        determained (default false)
              * @return true if the specified indexes are set, false if they are invalid or the same
              */
            inline bool setSignalsZoomed( uint startIndex = 0, uint stopIndex = UINT_MAX, bool updateMinMax = false )
            { return mSignals.setZoomed( startIndex, stopIndex, updateMinMax ); }


            /** @brief Resets mSignals to indicate to use the full range
              * Sets mSignals members mStartIndex to 0 and mStopIndex to UINT_MAX. This function will reset the
              * range and clears the polygons
              *
              * @param updateMinMax If true the minimum and maximum signal data values of the given range will be
              *        determained (default false)
              * @return true if the 'zoomed' state was reset
              */
            inline bool resetSignalsZoomed( bool updateMinMax = false ) { return mSignals.resetZoomed( updateMinMax ); }


            /** @brief Returns if the whole signals data is being used or a subset, i.e. zoomed
              *
              * @return true if mSignals is set to use the full range
              */
            inline bool isSignalsZoomed( void ) const { return mSignals.isZoomed(); }


            /** @brief Tries to find the minimum and maximum X and Y data values of all signals
              * This function wil search the current X and Y signals data and fills mMinXData,
              * mMaxXData, mMinYData and mMaxYData for all signal data object.
              * This function forces a re-chack. The previous values are neglected
              *
              * @return true is these member were updated. False is not, ie. not data present
              */
            inline bool findMinMaxData( void ) { return mSignals.findMinMaxData(); }


            /** @brief Called to update the data
              * This is a convenient interface function that can be used to signal all diffent kind of
              * graph classes in a uniformed way to update the signal data they contain due to, for
              * example, data changes
              * Derived, application specific classes should implement the correct update behaviour for this.
              *
              * @return true if the added data should initialte a rescale
              */
            virtual bool updateSignals( void ) = 0;


            /** @brief Is called to get the tolltip text for the specified time position
              * This function is called from the event handling function that handles tooltip request
              * events, @see PGraphicsSignalView::events().
              * Implement this function is a derived class to return the corresponding tooltip text or
              * to return an empty text to show no tooltip
              *
              * @param timePosition The requested time position in data coordinate system
              * @return The text to use as the tooltip or an empty string if no tooltip should be shown
              */
            virtual QString getToolTipText( double timePosition ) const = 0;


            /** @brief This function will be called when the extent (amplitued or time-range) of the graph is changed
              * This function wil handle the known types of markers. For exaple the line marker will be updated according
              * the changed extent but the eclipse marker won't be changed.
              * When a custom marker was created, derive an own implementation of this function to handle that
              * kind of markers
              */
            virtual void adjustMarkerExtent( void );


            /** @brief Helper function that will signal the view to perform a rescale of all graphs
              * @see PGraphicsSignalView::autoScaleSlot()
              */
            inline void triggerRescale( void ) { emit doAutoScale(); }


            /** @brief In this class no real function is implemented, override to add needed function
              * Override this function in derived classes to enable drawing data on top of the background
              * and signal information
              *
              * @param painter the painter see Qt documentation
              * @param updateRect the rectangle, in scene coordinates, of the rectangle to (re-)draw
              */
            virtual void paintForeGround( QPainter * /*painter*/, const QRectF & /*updateRect*/ ) { mWaitForDraw = false; }


            /** @brief Draw the grid and any configured texts for the graph
              * @param painter the painter see Qt documentation
              * @param updateRect the rectangle, in scene coordinates, of the rectangle to (re-)draw
              */
            virtual void paintGrid( QPainter * painter, const QRectF & updateRect );


            /** @brief Sets this item in the printing mode
              * In this mode the printing colors, pens or brushes are used in stead of the normal painting ones.
              *
              * @param doPrint determines if the view is printed
              */
            inline void setPrinting( bool doPrint = false ) { mIsPrinting = doPrint; }


            /** @brief Returns the value of mIsPrinting
              * Returns true if this items is currently in printing mode
              *
              * @return mIsPrinting
              */
            inline bool isPrinting( void ) const { return mIsPrinting; }

            /** @brief sets the clipping of the signals drawn
              * @param clip the new clipping value
              */
            inline void setSignalClipping(bool clip) { mClipSignal = clip; }


            /** @brief returns the clipping of the signals drawn
              *
              */
            inline bool isSignalClipping() const { return mClipSignal; }


            /** @brief Is called to paint this graph item
              * Overloaded member function of the QT base class. This handler base implementation should
              * be called to reset the mWaitForDraw flag.
              * See for more information the Qt documentation
              *
              * No params are used in this base implementation
              */
            virtual void paint( QPainter * /*painter*/,
                                const QStyleOptionGraphicsItem * /*option*/,
                                QWidget * /*widget = 0*/ );


        signals:

            /** @brief Signal indicating that the scale of the Y-axis for this graph has changed
              * This signal is automatically connected by PGraphMatrixScene::addItem
              */
            void amplScaleChanged( void );


            /** @brief Signal indicating that the scale of the X-axis for this graph has changed
              * This signal is automatically connected by PGraphMatrixScene::addItem
              */
            void timeScaleChanged( void );


            /** @brief Informs the reciever that the min and max axis values must be checked and adjusted if needed
              */
            void doAutoScale( void );


        public slots:

            /** @brief Recieves a signal that the scale of the Y-axis must change accordingly
              */
            void changeAmplScale( double maxSigAmpl, double minSigAmpl, double amplitudeStep, double amplitudeRange );


            /** @brief Recieves a signal that the scale of the X-axis must change accordingly
              */
            void changeTimeScale( double maxTime, double minTime, double timeStep, double timeRange );


        protected:

            /** @brief Implement this function to adjust the position of the marker for the given position
              * This function will be called by the mouse click handling function to indicate a user action
              * at the specified position in the graph. The given positions will be in data coordinates
              * system and graph coordinate system to enable maximum control.
              * Derived, application specific classes should implement the correct marker behaviour for this.
              *
              * @param dataPos The position to adjust with. This position is in data coordinate system
              * @param graphPos The position to adjust with. This position is in graph coordinate system
              * @param mouseReleased If true, this function is being called from the mouse release event, otherwise the mouse move event
              *
              * @return true if the position has changed
              */
            virtual bool adjustMarkerPosition( const QPointF & dataPos, const QPointF & graphPos, bool mouseReleased ) = 0;


            /** @brief Reacts to a mouse press event
              *   Used for zoom in functionlity
              * @param event mouse event see Qt documentation
              */
            virtual void mousePressEvent( QGraphicsSceneMouseEvent * event );


            /** @brief Reacts to a mouse move event
              *   Used for zoom in functionlity
              * @param event mouse event see Qt documentation
              */
            virtual void mouseMoveEvent( QGraphicsSceneMouseEvent * event );


            /** @brief Reacts to a mouse release event
              *  Used for zoom in functionlity
              *  @param event mouse event see Qt documentation
              */
            virtual void mouseReleaseEvent( QGraphicsSceneMouseEvent * event );


            /** @brief Reimplementation of the wheelEvent
              * This function is used to modify the maximum amplitude
              */
            virtual void wheelEvent( QGraphicsSceneWheelEvent * event );


//            /** @brief This function is called when the mouse is newly over this graph item
//              * This function is used to check if the mouse cursor should change
//              * @note Default hover events are not being passed on by Qt. Enable this by calling
//              *
//              * setAcceptHoverEvents( true ) in the derived class
//              *
//              * @param event Holds the hover event information
//              */
//            virtual void hoverEnterEvent( QGraphicsSceneHoverEvent * event );


            /** @brief This function is called when the mouse is moving over this graph item
              * This function is used to check if the mouse cursor should change
              *
              * @note Default hover events are not being passed on by Qt. Enable this by calling
              * setAcceptHoverEvents( true ) in the derived class
              *
              * @param event Holds the hover event information
              */
            virtual void hoverMoveEvent( QGraphicsSceneHoverEvent * event );


            /** @brief This function is called when the mouse is leaving this graph item
              * This function is used to check if the mouse cursor should be reset
              *
              * @note Default hover events are not being passed on by Qt. Enable this by calling
              * setAcceptHoverEvents( true ) in the derived class
              *
              * @param event Holds the hover event information
              */
            virtual void hoverLeaveEvent( QGraphicsSceneHoverEvent * event );


            /** @brief TODO
              */
            double toTime( double pos ) const;


            /** @brief TODO
              */
            double toAmplitude( double pos ) const;


            /** @brief compute the time and amplitude values to the graphicsItem coordinates
              * @param t the time
              * @param ampl the amplitude
              */
            inline QPointF toItemCoor( double t, double ampl ) const { return QPointF( toItemTime( t ), toItemAmplitude( ampl ) ); }


            /** @brief the coordinate system works from top to bottem, therefore a "-" sign is
              * used
              */
            double toItemAmplitude( double ampl ) const;


            /** @brief the coordinate system works from top to bottem, therefore a "-" sign is
              * used
              */
            double toItemTime( double t ) const;


            /** @brief This function is called when the mouse cursor has been moved over this item
              * This function will indicate if a marker has been hovered or not. In derived classes
              * appropriate action can be taken, like changing the mouse cursor
              *
              * @param markerIndex The index of the marker hovered above or NO_MARKER_FOUND if no marker was hovered
              */
            virtual void onMarkerHoverEvent( int markerIndex ) { Q_UNUSED( markerIndex ) }


            /** @brief Resets the flag mWaitForDraw to false
              * By resetting this flag, the mouse move events will be processed again
              */
            inline void resetWaitForDraw( void ) { mWaitForDraw = false; }


            /** @brief Holds the signals as raw data with signal related information
              * This is the central store container for all signal related data
              */
            PSignalsData mSignals;


            /** @brief Holds the markers as raw data with all related information
              * This is the central store container for all marker related data
              */
            PSignalMarkers mMarkers;


        private:

            /** @brief this function will be called before drawing the grid to calculate the space occupyed by all texts to show
              *
              * @param painter The painter object to use during the calculation
              */
            void calculateTextSizes( QPainter * painter );


            /** @brief Sets the specified color objects to the new color
              * This helper function is used to set the member color objects according the specified 'print'
              * flag to the given color
              *
              * @param color the new font to use
              * @param print If true, the color used for printing (screen capture) is set, otherwise the
              *              screen painting color (default)
              * @param paintClr The member object that is used for screen painting. This member can be set
              * @param printClr The member object that is used for print painting. This member can be set
              */
            void setObjectColor( const QColor & color, bool print, QColor & paintClr, QColor  & printClr );


            /** @brief reference member
              */
            Widget::Grid_type mGrid;


            /** @brief If true, the X-axis shows index data, i.e. starting at 0 with intervals of 1
              * This should be set to true if a positive evolving (time) graph with fixed interval of 1 is
              * being shown. An example would be an ECG signal. If two signals are plotted against each
              * other, where interval varies and where the signal evolves in multiple directions, this flag
              * should be set to false. An example would be a VCG or PV-loop signal
              */
            bool mXAxisShowsIndex;


            /** @brief the font used to draw extra text on the graph
              */
            QFont mTextFont;


            /** @brief the font used to draw the label along the graphs axis
              */
            QFont mLabelFont;


            /** @brief the font used to draw the scale along the graphs axis
              */
            QFont mScaleFont;


            /** @brief Extra text that can be added to the graph
              */
            QString mText;


            /** @brief Extra text that can be added to the graph and shown within the graph area
              */
            QString mInGraphText;


            /** @brief The name that must be shown allong the X-axis
              */
            QString mXLabel;


            /** @brief The name that must be shown allong the Y-axis
              */
            QString mYLabel;

            /** @brief The unit that must be shown allong the X-axis
              */
            QString mXUnit;

            /** @brief The unit that must be shown allong the Y-axis
              */
            QString mYUnit;


            /** @brief The maximum value shown on the Y-axis
              */
            double mMaxSigAmpl;


            /** @brief The minimum value shown on the Y-axis
              */
            double mMinSigAmpl;


            /** @brief Holds the size of all labels, unit strings and other texts
              *
              * This member is used to hold the bounding rectangles of all text items
              * that can be displayed in the graph. If a text should not be displayed
              * (i.e. is empty) an empty rectangle (isNull is true) will be found in the
              * list.
              * The items in this array hold coordinates in the scene coordinate system. Only
              * the texts that will be drawn vertical must be translated.
              */
            QVector<QRectF> mItemSizes;


            /** @brief Flag that indicates if the scale texts along the X-axis will be shown
              * Default true
              */
            bool mShowXScaleTexts;


            /** @brief Holds the texts that are being displayed as scale units along the X-axis
              */
            QStringList mXScaleTexts;


            /** @brief Holds the maximum bounding rectangle for all texts that are being displayed as scale units along the X-axis
              */
            QRectF mXScaleTextRect;


            /** @brief Flag that indicates if the scale texts along the Y-axis will be shown
              * Default true
              */
            bool mShowYScaleTexts;


            /** @brief Holds the texts that are being displayed as scale units along the X-axis
              */
            QStringList mYScaleTexts;


            /** @brief Holds the maximum bounding rectangle for all texts that are being displayed as scale units along the Y-axis
              */
            QRectF mYScaleTextRect;


            /** @brief Flag that indicates if the zero line for the Y-axis will be shown
              * Default true
              */
            bool mShowYZeroLine;


            /** @brief Holds the final rectangle with which the graph itself will be drawn.
              * This rectangle is the items rectangle minus all text-items drawn around it
              *
              * This item holds coordinates that are in graphics item coordinates, thus can be used for
              * drawing within this item
              */
            QRectF mGraphRect;


            /** @brief Holds the transformation matrix that can be used to transform data coordinates into graph coordinates
              */
            QTransform mGraphTransform;


            /** @brief if true the amplitude/time scale is symetric, i.e. mMaxSigAmpl == -mMinSigAmpl
              * This value is used when the user scrolls the mouse-wheel or selects and area in the graph. When
              * set and scrolling, it changes both the range of the X-axis and Y-axis at the same time.
              * Using the <CTRL> or <ALT> modifier key will have no extra function in that case
              */
            bool mIsSymetric;


            /** @brief If true, the scrollwheel of the mouse can be used to adapt the range of the X-axis
              * Default value is true, i.e. enabled
              */
            bool mDoAdaptXAxis;


            /** @brief If true, adapting the range of the X-axis will be limited to the data range
              * Default value is false, i.e. diabled and X-axis range can be changed unlimited
              */
            bool mLimitAdaptXAxis;


            /** @brief If true, the scrollwheel of the mouse can be used to adapt the range of the Y-axis
              * Default value is true, i.e. enabled
              */
            bool mDoAdaptYAxis;


            /** @brief If true, adapting the range of the Y-axis will be limited to the data range
              * Default value is false, i.e. diabled and Y-axis range can be changed unlimited
              */
            bool mLimitAdaptYAxis;


            /** @brief Flag that indicates if the graph should be shown maximized
              * When true, all labels, texts and scale information are not shown and the graph
              * will occupy all space in the scene given to this item
              */
            bool mMaximizeGraph;


            /** @brief
              */
            bool mAllowZoomSelect;


            /** @brief
              */
            double mAmplitudeRange;


            /** @brief
              */
            double mAmplitudeStep;


            /** @brief
              */
            bool mInvertXDirection;


            /** @brief
              */
            bool mInvertYDirection;


            /** @brief the scaling factor of the x axis
              */
            int mXfactor;

            /** @brief the scaling factor of the y axis
              */
            int mYfactor;

            /** @brief the scaling factor of the x axis label
              */
            double mXLabelFactor;

            /** @brief the scaling factor of the y axis label
              */
            double mYLabelFactor;


            /** @brief The minimum time value to include in the graph when drawing
              */
            double mZoomT0;


            /** @brief The maximum time value to include in the graph when drawing
              */
            double mZoomT1;


            /**
              */
            double mTimeStep;


            /** @brief time range in ms
              */
            double mTimeRange;


            /** @brief Used in the Mouse event handling
              */
            QRectF mSelectRect;


            /** @brief Pen used to draw the zero line
              */
            QColor mZeroLine;


            /** @brief Pen used to draw the zero line when in printing mode
              */
            QColor mZeroLinePrint;


            /** @brief Pen used to draw the major grid lines
              */
            QColor mGridLine;


            /** @brief Pen used to draw the major grid lines when in printing mode
              */
            QColor mGridLinePrint;


            /** @brief Color used to draw the grid points
              */
            QColor mGridPoint;


            /** @brief Color used to draw the grid points when in printing mode
              */
            QColor mGridPointPrint;


            /** @brief used to specify a different background color within the graph
              */
            QColor mBackGroundColor;


            /** @brief used to specify a different background color within the graph only when in printing mode
              * Default the scene determains the background color
              */
            QColor mBackGroundColorPrint;


            /** @brief the color used for drawing text next to the graph
              */
            QColor mTextColor;


            /** @brief the color used for drawing text next to the graph when in printing mode
              */
            QColor mTextColorPrint;


            /** @brief the color used for drawing a zoom selection tracking rectangle onto the graph
              */
            QColor mSelectRectColor;


            /** @brief Flag that indicates if the printing colors should be used or the 'normal' screen paint colors
              * If this flag is true, the printing colors, pens, brushes will be used, otherwise the default ones
              */
            bool mIsPrinting;


            /** @brief sets clipping of the signal
              *
              */
            bool mClipSignal;


            /** @brief Flag that is used to filter the mouse move events so drawing isn't prevented by Qt
              * When the mouse is being moved while a handler is selected, that mouse move will be used by the
              * implementation to change the (position) of the handler according its use, see
              * handleMouseAction(const QPointF &, bool). When each mouse move is handled this way, Qt will
              * prevent quick updating of the changes. This flag will prevent handling of mouse move events
              * untill the handler is re-painted.
              * @note This means that in the paint event handling function this flag must be reset. This can be
              * done by calling the base implementation of this class.
              */
            bool mWaitForDraw;
        };
    }
}

#endif	/* __PSIGNALPLOTBASE_H__ */

