/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalsData.cpp 1172 2014-03-21 13:45:45Z eelco $
 *
 */


#include "PSignalsData.h"

#include <QDebug>


using namespace Peacs::Support;


PSignalsData::PSignalsData( void ) :
    mStartIndex( 0 ),
    mStopIndex( UINT_MAX ),
    mData()
{
}


PSignalsData::PSignalsData( int size ) :
    mStartIndex( 0 ),
    mStopIndex( UINT_MAX ),
    mData( size )
{
}


PSignalsData::PSignalsData( const PSignalsData & source ) :
    mStartIndex( source.mStartIndex ),
    mStopIndex( source.mStopIndex ),
    mData( source.mData )
{
}


void PSignalsData::clearSignals( void )
{
    for ( int index = 0; index < mData.size(); index++ )
    {
        mData[ index ].clear();
    }
}


void PSignalsData::clearPoly( void )
{
    for ( int index = 0; index < mData.size(); index++ )
    {
        mData[ index ].clearPoly();
    }
}


bool PSignalsData::setZoomed( uint startIndex, uint stopIndex, bool updateMinMax )
{
    bool done = false;

    if ( mData.size() )
    {
        const PSignalData & sigData = mData.at( 0 );    // Use the first signal to check

        // Check the specified range. If not valid, adapt for its range
        if ( startIndex >= (uint)sigData.count() )
        {
            startIndex = 0;
        }

        if ( (stopIndex != UINT_MAX) && (stopIndex >= (uint)sigData.count()) )
        {
            stopIndex = UINT_MAX;
        }
    }

    if ( (startIndex < stopIndex) &&
         ( (mStartIndex != startIndex) || (mStopIndex != stopIndex) ) )
    {
        mStartIndex = startIndex;
        mStopIndex  = stopIndex;

        clearPoly();

        if ( updateMinMax )
        {
            findMinMaxData();
        }
    }

    return done;
}


bool PSignalsData::resetZoomed( bool updateMinMax )
{
    bool done = false;

    if ( updateMinMax || isZoomed() )
    {
        mStartIndex = 0;
        mStopIndex  = UINT_MAX;
        done        = true;

        clearPoly();

        if ( updateMinMax )
        {
            findMinMaxData();
        }
    }

    return done;
}


bool PSignalsData::CreateOrUpdatePolygons( const QTransform & transformation, double pixelSize/*, uint startIndex, uint stopIndex*/ )
{
    bool changed = false;

    for ( int index = 0; index < mData.size(); index++ )
    {
        if ( mData[ index ].CreateOrUpdatePolygon( transformation, pixelSize, mStartIndex, mStopIndex ) )
        {
            changed = true;
        }
    }

    return changed;
}


bool PSignalsData::removeUnusedDataBefore( int beforeIndex )
{
    bool allRemoved = false;

    if ( (0 < mData.count()) && (0 <= beforeIndex) && (beforeIndex < mData.at( 0 ).count()) )
    {
        // The request is valid
        allRemoved = true;

        for ( int index = 0; index < mData.size(); index++ )
        {
            if ( !mData[ index ].removeUnusedDataBefore( beforeIndex ) )
            {
                allRemoved = false;
            }
        }
    }

    return allRemoved;
}


double PSignalsData::getMinXData( void ) const
{
    double min = 0.0;

    if ( 0 < mData.size() )
    {
        bool first = true;

        for ( int index = 0; index < mData.size(); index++ )
        {
            if ( mData.at( index ).mustShow() &&
                 (first || (min > mData.at( index ).getMinXData())) )
            {
                first   = false;
                min     = mData.at( index ).getMinXData();
            }
        }
    }

    return min;
}


double PSignalsData::getMaxXData( void ) const
{
    double max = 0.0;

    if ( 0 < mData.size() )
    {
        bool first = true;

        for ( int index = 0; index < mData.size(); index++ )
        {
            if ( mData.at( index ).mustShow() &&
                 (first || (max < mData.at( index ).getMaxXData())) )
            {
                first   = false;
                max     = mData.at( index ).getMaxXData();
            }
        }
    }

    return max;
}


double PSignalsData::getMinYData( void ) const
{
    double min = 0.0;

    if ( 0 < mData.size() )
    {
        bool first = true;

        for ( int index = 0; index < mData.size(); index++ )
        {
            if ( mData.at( index ).mustShow() &&
                 (first || (min > mData.at( index ).getMinYData())) )
            {
                first   = false;
                min     = mData.at( index ).getMinYData();
            }
        }
    }

    return min;
}


double PSignalsData::getMaxYData( void ) const
{
    double max = 0.0;

    if ( 0 < mData.size() )
    {
        bool first = true;

        for ( int index = 0; index < mData.size(); index++ )
        {
            if ( mData.at( index ).mustShow() &&
                 (first || (max < mData.at( index ).getMaxYData())) )
            {
                first   = false;
                max     = mData.at( index ).getMaxYData();
            }
        }
    }

    return max;
}


bool PSignalsData::findMinMaxData( void )
{
    bool dataPresent = false;

    if ( 0 < mData.size() )
    {
        PSignalData * dataPtr = mData.data();

        dataPresent = true;

        for ( int index = 0; (index < mData.size()) && dataPresent; index++ )
        {
            dataPresent = dataPtr[ index ].findMinMaxData( mStartIndex, mStopIndex );
        }
    }

    return dataPresent;
}
