/*
 *  This file is part of the Peacs support package
 *
 *  Copyright (C) 2010 - 2012 Peacs (www.peacs.nl)
 *
 *  The Peacs Support package is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  The Peacs support classes are distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Peacs support package.  If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 *    Project:              Peacs Support
 *    Qt version:           4.7.x
 *    Compiler:             QMake, MinGW, gcc 4.4.0
 *    Target platforms:     Windows, Mac OS-X
 *
 *    Subversion info:
 *
 *    $Id: PSignalsData.h 1172 2014-03-21 13:45:45Z eelco $
 *
 */


#ifndef PSIGNALSDATA_H
#define PSIGNALSDATA_H


#include <QVector>

#include "PeacsSupport_global.h"
#include "PSignalData.h"


namespace Peacs
{
    namespace Support
    {
        /** @class PSignalsData PSignalsData.h "widgets/PSignalsData.h"
          * @brief Holds a list with data and information about a collection of signals to plot in a signal view
          */
        class PEACSSUPPORTSHARED_EXPORT PSignalsData
        {

        public:

            /** @brief Constructor
              */
            PSignalsData( void );


            /** @brief Initialyzer constructor
              * Creates a vector with 'size' position reserved for signal data objects
              *
              * @param size the number of objects to instantiate during construction
              */
            PSignalsData( int size );


            /** @brief Initialyzer constructor
              *
              * @param source the source to copy from
              */
            PSignalsData( const PSignalsData & source );


            /** @brief Destructor
              */
            ~PSignalsData( void ) {;}


            /** @brief Removes all the elements from the vector and releases the memory used by the vector.
              */
            inline void clear( void ) { mData.clear(); }


            /** @brief Removes from all signals the data.
              */
            void clearSignals( void );


            /** @brief Removes from all signals the polyline and related data.
              * This forces recalculation of all drawn polyline. This is usefull when, for example
              * the scale has changed
              */
            void clearPoly( void );


            /** @brief Returns the number of PSignalsData items in the vector.
              */
            inline int size( void ) const { return mData.size(); }


            /** @brief Returns the number of PSignalsData items in the vector.
              * This is an overloaded function.
              * @see size()
              */
            inline int count( void ) const { return mData.count(); }


            /** @brief Returns a reference to the last PSignalData item in the vector.
              * This function assumes that the vector isn't empty.
              */
            inline PSignalData & last( void ) { Q_ASSERT( !isEmpty() ); return mData.last(); }


            /** @brief This is an overloaded function.
              * @see last()
              */
            inline const PSignalData & last( void ) const { Q_ASSERT( !isEmpty() ); return mData.last(); }


            /** @brief Returns true if the vector has size 0; otherwise returns false.
              */
            inline bool isEmpty( void ) const { return mData.isEmpty(); }


            /** @brief Returns the PSignalData item at index position sdIndex in the vector.
              *
              * @param sdIndex must be a valid index position in the vector (i.e., 0 <= sdIndex < size())
              */
            inline const PSignalData & at( int sdIndex ) const
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < mData.size(), "PSignalsData::at", "index out of range");
                return mData.at( sdIndex );
            }


            /** @brief Returns the PSignalData item at index position sdIndex in the vector.
              *
              * @param sdIndex must be a valid index position in the vector (i.e., 0 <= sdIndex < size())
              */
            inline const PSignalData & operator[]( int sdIndex ) const
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < mData.size(), "const PSignalsData::operator[]", "index out of range");
                return mData[ sdIndex ];
            }


            /** @brief Returns the PSignalData item at index position sdIndex as a modifiable reference.
              *
              * @param sdIndex must be a valid index position in the vector (i.e., 0 <= sdIndex < size())
              */
            inline PSignalData & operator[]( int sdIndex )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < mData.size(), "PSignalsData::operator[]", "index out of range");
                return mData[ sdIndex ];
            }


            /** @brief Adds a new signal data object to the vector
              *
              * @param id the value to use for mId of the new object (default -1)
              * @param label the value to use for mLabel of the new object (default an empty string)
              * @param correctionX the value to use for mCorrectionX of the new object (default 1.0)
              * @param correctionY the value to use for mCorrectionY of the new object (default 1.0)
              */
            inline void append( int id = -1, const QString & label = QString(), double correctionX = 1.0, double correctionY = 1.0 )
            { mData.append( PSignalData( id, label, correctionX, correctionY ) ); }


            /** @brief Adds a new signal data object to the vector
              *
              * @param id the value to use for mId of the new object (default -1)
              * @param label the value to use for mLabel of the new object (default an empty string)
              * @param correctionX the value to use for mCorrectionX of the new object (default 1.0)
              * @param correctionY the value to use for mCorrectionY of the new object (default 1.0)
              */
            inline void append( int id, const QString & label, bool show, const QPen & pen, double correctionX = 1.0, double correctionY = 1.0 )
            { mData.append( PSignalData( id, label, show, pen, correctionX, correctionY ) ); }


            /** @brief Returns the data at the specified position on the x-axis if available or index
              * This function returns the data at the specified position in mXData if
              * mXData is not empty. If mXData is empty the index is returned.
              * In the first case, two signals are drawn against each other. In the second
              * case the signal is drawn agains a fixed measurement interval, for example time.
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param index the position to return the data from
              *
              * @return the value at the specified position or index
              */
            inline double getXData( int sdIndex, int index ) const
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::getXData", "index out of range");
                return mData.at( sdIndex ).getXData( index );
            }


            /** @brief Sets the data in mXData
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param data the data to store in mXData
              */
            inline void setXData( int sdIndex, const PVector & data )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::setXData", "index out of range");
                mData[ sdIndex ].setXData( data );
            }


            /** @brief Appends the data in mXData
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param data the data to append to mXData
              */
            inline void addToXData( int sdIndex, const PVector & data )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::addToXData", "index out of range");
                mData[ sdIndex ].addToXData( data );
            }


            /** @brief Returns the data at the specified position on the y-axis
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param index the position to return the data from
              *
              * @return the value at the specified position
              */
            inline double getYData( int sdIndex, int index ) const
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::getYData", "index out of range");
                return mData[ sdIndex ].getYData( index );
            }


            /** @brief Sets the data in mYData
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param data the data to store in mYData
              */
            inline void setYData( int sdIndex, const PVector & data )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::setYData", "index out of range");
                mData[ sdIndex ].setYData( data );
            }


            /** @brief Appends the data in mYData
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param data the data to append to mYData
              */
            inline void addToYData( int sdIndex, const PVector & data )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::addToYData", "index out of range");
                mData[ sdIndex ].addToYData( data );
            }


            /** @brief Removes toRemove elements from the beginning of the specified mYData vector
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param toRemove The number of element to remove
              */
            inline void adjustYDataLen( int sdIndex, int toRemove )
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::adjustYDataLen", "index out of range");
                mData[ sdIndex ].adjustYDataLen( toRemove );
            }


            /** @brief Removes toRemove elements from the beginning of all mYData vectors
              *
              * @param toRemove The number of element to remove
              */
            inline void adjustYDataLen( int toRemove )
            {
                for ( int index = 0; index < mData.size(); index++ )
                {
                    mData[ index ].adjustYDataLen( toRemove );
                }
            }


            /** @brief Returns the [X,Y] coordinate at the specified position
              * This function returns the y-axis data pared to the corresponding x-axis data
              * The x-axis data can either be a found value in mXData or the asked index
              *
              * @param sdIndex the index of the signal data in the signals data vector
              * @param index the position to return the [X,Y] data from
              *
              * @return The combined [X,Y] value of mXData and mYData
              */
            QPointF getXYData( int sdIndex, int index ) const
            {
                Q_ASSERT_X(sdIndex >= 0 && sdIndex < size(), "PSignalsData::getXYData", "index out of range");
                return mData.at( sdIndex ).getXYData( index );
            }


            /** @brief Sets mStartIndex and mStopIndex to the given values, i.e. zooming to a specified range
              * This function will check the given range with the maintained signals data, sets the range and clears
              * the polygons if valid
              *
              * @param startIndex The index in the signal data array where to start converting to polygon (default 0)
              * @param stopIndex The index in the signal data array where to stop converting to polygon  (default
              *        UINT_MAX, i.e. up to the end)
              * @param updateMinMax If true the minimum and maximum signal data values of the given range will be
              *        determained (default false)
              * @return true if the specified indexes are set, false if they are invalid or the same
              */
            bool setZoomed( uint startIndex = 0, uint stopIndex = UINT_MAX, bool updateMinMax = false );


            /** @brief Resets mStartIndex and mStopIndex to indicate to use the full range
              * Sets mStartIndex to 0 and mStopIndex to UINT_MAX. This function will reset the range and clears
              * the polygons
              *
              * @param updateMinMax If true the minimum and maximum signal data values of the given range will be
              *        determained (default false)
              * @return true if the 'zoomed' state was reset
              */
            bool resetZoomed( bool updateMinMax = false );


            /** @brief Returns if the whole signal data is being used or a subset, i.e. zoomed
              *
              * @return true if mStartIndex and mStopIndex are set to use the full range, i.e. [0 .. UINT_MAX]
              */
            inline bool isZoomed( void ) const { return (0 < mStartIndex) || (mStopIndex < UINT_MAX); }


            /** @brief Creates or updates the signal polygons for all the signals data combined with the specified data
              * The polygon will only be updated if it is empty when calling this function. The polygon will
              * be made empty when a redraw is needed or when new signal data is set.
              *
              * @param transformation The transformation matrix used to transform the signals data into the
              *        wanted polygons coordinates related to the drawing area
              * @param pixelSize Used to filer the signals data; only the data that adds information to the
              *        drawing will be used. Any data in between the pixel size points is neglected
              * @return true if the any of the polygons has changed
              */
            bool CreateOrUpdatePolygons( const QTransform & transformation, double pixelSize );


            /** @brief Removes all data for all signals before the specified index
              *
              * @param beforeIndex Specifies the index before which all data should be removed from the signal data
              * @return true is for all items the requested data part is removed, false if only for some or none
              */
            bool removeUnusedDataBefore( int beforeIndex );


            /** @brief Returns the minimum data value that must be shown on the x-axis
              * For all signal that must be shown, the minimum value will be determained
              *
              * @return The minimum X-axis value to show
              */
            double getMinXData( void ) const;


            /** @brief Returns the maximum data value that must be shown on the x-axis
              * For all signal that must be shown, the maximum value will be determained
              *
              * @return The maximum X-axis value to show
              */
            double getMaxXData( void ) const;


            /** @brief Returns the minimum data value that must be shown on the y-axis
              * For all signal that must be shown, the minimum value will be determained
              *
              * @return The minimum Y-axis value to show
              */
            double getMinYData( void ) const;


            /** @brief Returns the maximum data value that must be shown on the y-axis
              * For all signal that must be shown, the maximum value will be determained
              *
              * @return The maximum Y-axis value to show
              */
            double getMaxYData( void ) const;


            /** @brief Tries to find the minimum and maximum X and Y data values of all signals
              * This function wil search the current X and Y signals data and fills mMinXData,
              * mMaxXData, mMinYData and mMaxYData for all signal data object.
              * This function forces a re-chack. The previous values are neglected.
              * This function will use the mStartIndex and mStopIndex as the range to determain
              * the min and max values
              *
              * @return true is these member were updated. False is not, ie. not data present
              */
            bool findMinMaxData( void );


        private:

            /** @brief Hold the index within the signals data from where the data should be used
              * This member will contain the value 0 if all data from the beginning up untill mStopIndex
              * should be used
              */
            uint mStartIndex;


            /** @brief Hold the index within the signals data upto where the data should be used
              * This member will contain the value UINT_MAX if all data from mStartIndex untill the
              * end should be used
              */
            uint mStopIndex;


            /** @brief Hold the data this class is responsible for
              */
            QVector< PSignalData > mData;
        };
    }
}

#endif // PSIGNALSDATA_H
