/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: PCircSignalDraw.h 931 2013-11-27 19:03:51Z eelco $
 *
 */


#ifndef __PCIRCSIGNALDRAW_H__
#define	__PCIRCSIGNALDRAW_H__

#include <QMenu>
#include <QList>

#include "PSignalDraw.h"
#include "PCircAdaptDefines.h"
#include "ucirculation.h"


namespace Peacs
{
    using namespace Support;

    namespace CircAdapt
    {
        /** @class PGraphicsECGItem PGraphicsECGItem.h "ECGView/PGraphicsECGItem.h"
         * @brief The view window in which the ECGs are displayed
         *
         *
         * */
        class PCircSignalDraw : public PSignalDraw
        {
            Q_OBJECT

        public:

            /** @brief constructor: ALl paramters are used to scale the signals the same in all these
              * graphics Items
              *
              * @param zoomT0
              * @param zoomT1
              * @param column the column at which this item will reside in the matrix (default -1)
              * @param row the row at which this item will reside in the matrix (default -1)
              *
              * @see PPlotBase for information about column and row
              */
            PCircSignalDraw( signal_type      type,
                             subSignal_type   subtype,
                             UCirculation   & data,
                             double           zoomT0,
                             double           zoomT1,
                             int              column = -1,
                             int              row = -1 );


            /** @brief TODO
              */
            void init( void );


            /** @brief Returns the specified signal type stored in mType
              */
            inline signal_type getType( void ) const { return mType; }


            /** @brief Returns the specified signal sub type stored in mSubType
              */
            inline subSignal_type getSubType( void ) const { return mSubType; }


            /** @brief Called to update the data
              * This is a convenient interface function that is used to signal this graph instance
              * classes in a uniformed way to update the signal data it contains due to, for
              * example, data changes
              *
              * @return true if the added data should initialte a rescale
              */
            bool updateSignals( void );


            /** @brief Is called to get the tolltip text for the specified time position
              * This function is called from the event handling function that handles tooltip request
              * events, @see PGraphicsSignalView::events().
              *
              * @param timePosition The requested time position in data coordinate system
              * @return The text to use as the tooltip or an empty string if no tooltip should be shown
              */
            QString getToolTipText( double timePosition ) const;


            /** @brief Sets the time marker to the specified value and scedules for a redraw
              *
              * @param timeMarker The new time marker value
              */
            void setTimeMarker( double timeMarker );


        public slots:

            /** @brief Is called when the user has selected and item in the popup menu for this item
              *
              * @param action Information about the items selected
              */
            void on_contextMenu( QAction * action );


            /** @brief TODO
              */
            void updateShowSignals(int);


        signals:

            /** @brief TODO
              */
            void showSignals(int );


        protected:

            /** @brief This function is called to adjust the position of the marker for the given position
              * This function will be called from the base class by the mouse click handling function to
              * indicate a user action at the specified position in the graph. The given positions will be
              * in data coordinates system and graph coordinate system to enable maximum control.
              *
              * @param dataPos The position to adjust with. This position is in data coordinate system
              * @param graphPos The position to adjust with. This position is in graph coordinate system
              * @param mouseReleased If true, this function is being called from the mouse release event, otherwise the mouse move event
              *
              * @return true if the position has changed
              */
            bool adjustMarkerPosition( const QPointF & dataPos, const QPointF & graphPos, bool mouseReleased );


            /** @brief Is called when the user presses the right mouse button over this item to show the popup menu
              *
              * @param event Points to information about the event
              */
            void contextMenuEvent(QGraphicsSceneContextMenuEvent *event);


            /** @brief Helper function that returns the correction to use on the data to correct for display needs
              * This function returns a correction that is multiplied with the data values to get the values as wanted
              * for showing.
              *
              * @return The correction factor
              */
            double getCorrection( void ) const;


        private:

            /**
              */
            UCirculation & mData;


            /**
              */
            signal_type mType;


            /** @brief Specified the sub type defined by subSignal_type
              */
            subSignal_type mSubType;


            /** Pointer to the context menu instance created for this graphics view item
              */
            QMenu * mContextMenu;


            /** used to keep the menu-items from the context menu grouped
              */
            QActionGroup *  mMenuGroupPtr;
        };
    }
}
#endif	/* __PCIRCSIGNALDRAW_H__ */

