/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: pcircpvloop.cpp 1173 2014-03-21 14:12:14Z eelco $
 *
 */

//#include <QDebug>
#include <QGraphicsScene>

#include "pcircpvloop.h"
#include "PCircAdaptDefines.h"
#include "PMarkerEllipse.h"



using namespace Peacs::Support;
using namespace Peacs::CircAdapt;



// These offsets are used to identify the the signals within the Signal Data vector
const int OFFSET_SIGNAL_DATA_CURRENT    = 0x0000;
const int OFFSET_SIGNAL_DATA_INITIAL    = 0x1000;
const int OFFSET_SIGNAL_DATA_SAVED      = 0x2000;
const int SIGNAL_TYPE_COUNT             = 3;        // Current, initial and saved


PPVloop::PPVloop( UCirculation & data, int column, int row ) :
    PXYPlot( column, row ),
    mData( data ),
    mDrawInitial( false ),
    mDrawSaved( false ),
    mContextMenu( NULL ),
    mMenuGroupPtr( NULL )//,
//    mPreviousHeartbeatData(),
//    mPreviousHeartbeatData2()
{
    setXLabel( signal_type_str[VOLUME] );
    setYLabel( signal_type_str[PRESSURE] );

    setTimeRange( 0, 100 );
    setSignalAmpl( 0, 100 );

    setDrawInitial( true );
    setDrawSaved( true );

    setAcceptedMouseButtons( acceptedMouseButtons() | Qt::RightButton );
}


void PPVloop::on_contextMenu( QAction * action )
{
    QList < pvloop_type>    pvs( mData.getPVTypes() );
    bool                    done        = false;
    QFont                   actionFont  = action->font();

    actionFont.setBold( action->isChecked() );
    action->setFont( actionFont );

    for ( int k = 0; k < pvs.size() && !done; k++ )
    {
        int sigIndex = k * SIGNAL_TYPE_COUNT;

        if( action->data() == mData.XPVSignal( pvs[k] ).getName() &&
            action->data() == mData.YPVSignal( pvs[k] ).getName() )
        {
            Q_ASSERT( mSignals.count() > sigIndex );

            if ( (mSignals.count() > sigIndex) &&
                 (mSignals.at( sigIndex ).mustShow() != action->isChecked()) )
            {
                Q_ASSERT( mMarkers.count() > sigIndex + 2 );    // All signal type have a corresponding marker

                // Make sure that the current, initial and saved signal are all shown/hiden
                // 'sigIndex' is used and then incremented
                mSignals[ sigIndex++ ].setShow( action->isChecked() );
                //mSignals[ sigIndex ].setShow( action->isChecked() );      // The current time marker is not being used at the moment
                //mMarkers[ sigIndex++ ]->setShow( action->isChecked() );

                if ( mDrawInitial )
                {
                    mSignals[ sigIndex ].setShow( action->isChecked() );
                    mMarkers[ sigIndex++ ]->setShow( action->isChecked() );
                }
                else
                {
                    sigIndex++;
                }

                if ( mDrawSaved )
                {
                    mSignals[ sigIndex ].setShow( action->isChecked() );
                    mMarkers[ sigIndex ]->setShow( action->isChecked() );
                }

                // New signal added or removed. The needed scale may be changed
                triggerRescale();

                // Redraw the foreground
                scene()->invalidate( sceneBoundingRect(), QGraphicsScene::ItemLayer | QGraphicsScene::ForegroundLayer );
            }

            done = true;
        }
    }
}


void PPVloop::setDrawInitial( bool draw )
{
    if ( mDrawInitial != draw )
    {
        QList < pvloop_type>    pvs( mData.getPVTypes() );

        mDrawInitial = draw;

        if ( mSignals.count() == pvs.size() * SIGNAL_TYPE_COUNT )
        {
            for ( int k = 0; k < pvs.size(); k++ )
            {
                int sigIndexCurr = k * SIGNAL_TYPE_COUNT;
                int sigIndexInit = sigIndexCurr + 1;

                Q_ASSERT( mMarkers.count() > sigIndexInit );

                if ( mDrawInitial )
                {
                    // Only shown this initial signal when the current signal is shown as well
                    mSignals[ sigIndexInit ].setShow( mSignals[ sigIndexCurr ].mustShow() );
                    mMarkers[ sigIndexInit ]->setShow( mSignals[ sigIndexCurr ].mustShow() );
                }
                else
                {
                    mSignals[ sigIndexInit ].setShow( false );
                    mMarkers[ sigIndexInit ]->setShow( false );
                }

                // Redraw the foreground
                scene()->invalidate( sceneTransform().mapRect( getBounds() ),
                                     QGraphicsScene::ItemLayer | QGraphicsScene::ForegroundLayer );
            }
        }
    }
}


void PPVloop::setDrawSaved( bool draw )
{
    if ( mDrawSaved != draw )
    {
        QList < pvloop_type>    pvs( mData.getPVTypes() );

        mDrawSaved = draw;

        if ( mSignals.count() == pvs.size() * SIGNAL_TYPE_COUNT )
        {
            for ( int k = 0; k < pvs.size(); k++ )
            {
                int sigIndexCurr = k * SIGNAL_TYPE_COUNT;
                int sigIndexSav  = sigIndexCurr + 2;

                Q_ASSERT( mMarkers.count() > sigIndexSav );

                if ( mDrawSaved )
                {
                    // Only shown this saved signal when the current signal is shown as well
                    mSignals[ sigIndexSav ].setShow( mSignals[ sigIndexCurr ].mustShow() );
                    mMarkers[ sigIndexSav ]->setShow( mSignals[ sigIndexCurr ].mustShow() );
                }
                else
                {
                    mSignals[ sigIndexSav ].setShow( false );
                    mMarkers[ sigIndexSav ]->setShow( false );
                }

                // Redraw the foreground
                scene()->invalidate( sceneTransform().mapRect( getBounds() ),
                                     QGraphicsScene::ItemLayer | QGraphicsScene::ForegroundLayer );
            }
        }
    }
}


void PPVloop::init( void )
{
    QList < pvloop_type>    pvs( mData.getPVTypes() );

    mSignals.clear();
    mMarkers.clear();
//    mPreviousHeartbeatData.clear();
//    mPreviousHeartbeatData2.clear();

    if ( !pvs.isEmpty() )
    {
        int                         signalIndex         = 0;
        bool                        done                = false;
        const UCircAdaptSignal *    xSigPtr             = &( mData.getXPVSignal( pvs[ signalIndex ] ) );
        const UCircAdaptSignal *    ySigPtr             = &( mData.getYPVSignal( pvs[ signalIndex ] ) );
        double                      volumeCorrection    = mData.toVolume( 1.0 );
        double                      pressureCorrection  = mData.toPressure( 1.0 );

        //setXUnit( QString( "[%1]" ).arg( xSigPtr->getUnits() ) );
        //setYUnit( QString( "[%1]" ).arg( ySigPtr->getUnits() ) );
        setXLabel( signal_type_str[VOLUME] + QString( " [%1]" ).arg( xSigPtr->getUnits() ) );
        setYLabel( signal_type_str[PRESSURE] + QString( " [%1]" ).arg( ySigPtr->getUnits() ) );

        // Initialyze all signal data objects
        while ( !done )
        {
            QColor              penColor( xSigPtr->getColor() );
            QPen                signalPen( penColor );
            PMarkerEllipse *    markerPtr = NULL;

            // Initialyze the current signal data object
            mSignals.append( signalIndex + OFFSET_SIGNAL_DATA_CURRENT,
                             xSigPtr->getName(),
                             false,
                             signalPen,
                             volumeCorrection,
                             pressureCorrection );

            markerPtr = new PMarkerEllipse( signalIndex + OFFSET_SIGNAL_DATA_CURRENT,
                                            xSigPtr->getName(),
                                            false,
                                            QPen( TIMELINECOLOR ) );
            markerPtr->setRadii( 6.0 );
            markerPtr->pen().setWidthF( 5.0 );
            mMarkers.addMarker( markerPtr );

            // Initialyze the initial signal data object with alpha color and dotted line
            penColor.setAlphaF( 0.5 );
            signalPen.setColor( penColor );
            signalPen.setStyle( Qt::DotLine );

            mSignals.append( signalIndex + OFFSET_SIGNAL_DATA_INITIAL,
                             xSigPtr->getName(),
                             false,
                             signalPen,
                             volumeCorrection,
                             pressureCorrection );

            markerPtr = new PMarkerEllipse( signalIndex + OFFSET_SIGNAL_DATA_INITIAL,
                                            xSigPtr->getName(),
                                            false,
                                            QPen( TIMELINECOLORINIT ) );
            markerPtr->setRadii( 6.0 );
            markerPtr->pen().setWidthF( 5.0 );
            mMarkers.addMarker( markerPtr );

            // Initialyze the initial signal data object with alpha color and dashed line
            signalPen.setStyle( Qt::DashLine );

            mSignals.append( signalIndex + OFFSET_SIGNAL_DATA_SAVED,
                             xSigPtr->getName(),
                             false,
                             signalPen,
                             volumeCorrection,
                             pressureCorrection );

            markerPtr = new PMarkerEllipse( signalIndex + OFFSET_SIGNAL_DATA_SAVED,
                                            xSigPtr->getName(),
                                            false,
                                            QPen( TIMELINECOLORSAVED ) );
            markerPtr->setRadii( 6.0 );
            markerPtr->pen().setWidthF( 5.0 );
            mMarkers.addMarker( markerPtr );

            // Go to the next signal
            signalIndex++;

            if ( signalIndex < pvs.size() )
            {
                xSigPtr = &( mData.getXPVSignal( pvs[ signalIndex ] ) );
                //ySigPtr = &( mData.getYPVSignal( pvs[ signalIndex ] ) );
            }
            else
            {
                done = true;
            }
        }
    }

    if ( mContextMenu != NULL )
    {
        if ( mMenuGroupPtr != NULL )
        {
            disconnect( mMenuGroupPtr, SIGNAL(triggered(QAction*)), this, SLOT(on_contextMenu(QAction*)) );

            delete mMenuGroupPtr;
            mMenuGroupPtr = NULL;
        }

        delete mContextMenu;
        mContextMenu = NULL;
    }

    PXYPlot::init();
}


QString PPVloop::getToolTipText( double position ) const
{
    Q_UNUSED( position );

    QString valStr;

// NOT USED FOR NOW
//    if ( getCurrentTimeMarker() >= 0 )
//    {
//        QList < pvloop_type>    pvs( mData.getPVTypes() );

//        for ( int k = 0; k < pvs.size(); k++ )
//        {
//            int                         sigIndex    = k * SIGNAL_TYPE_COUNT;
//            const UCircAdaptSignal &    xSig        = mData.getXPVSignal( pvs[k] );
//            const UCircAdaptSignal &    ySig        = mData.getYPVSignal( pvs[k] );

//            // TODO: We should only shown tooltips when over the current timemarker
//            // or we should determain the mouse location relative to the shown signals
//            // and show when over a signal at that position
//            // Problem is that the found x_y coordinate will almost always be different
//            // then any position in the signal. We have to take the pixel size into
//            // account to find a point that will fit
//            if ( (mSignals.count() > sigIndex) &&
//                 mSignals.at( sigIndex ).mustShow() &&
//                 getInitialTimeMarker() > 0 &&
//                 getInitialTimeMarker() < xSig.getReferenceSignal().N() &&
//                 getInitialTimeMarker() < ySig.getReferenceSignal().N() /*&&
//                 x_y.x() == xSig.getSignalSample( getCurrentTimeMarker() ) &&
//                 x_y.y() == ySig.getSignalSample( getCurrentTimeMarker() )*/ )
//            {
//                if ( valStr.length() > 0 )
//                {
//                    valStr += "\n";
//                }

//                valStr += QString("%1:\n\t%2 %3\n\t%4 %5").
//                        arg( xSig.getName() ).
//                        arg( mData.toVolume( xSig.getSignalSample( getInitialTimeMarker() ) ), 5, 'g', 3 ).
//                        arg( xSig.getUnits() ).
//                        arg( mData.toPressure( ySig.getSignalSample( getInitialTimeMarker() ) ), 5, 'g', 3 ).
//                        arg( ySig.getUnits() );
//            }
//        }
//    }

    return valStr;
}


bool PPVloop::updateSignals( void )
{
    bool                    rescale = false;
    QList < pvloop_type>    pvs( mData.getPVTypes() );

    if ( !pvs.isEmpty() )
    {
        bool updateInitalMarker = false;
        bool updateSavedMarker = false;

// THE 'mPreviousHeartbeatData' AND 'mPreviousHeartbeatData2' CAN BE USED TO ADD PREVIOUS HEEARTBEAT DATA
// TO THE CURRENT SIGNAL. THIS COULD BE NICE TO SEE THE TRANSITION OF SIGNAL STATE AFTER THE USER HAS
// CHANGED SOME PARAMETERS.
// FOR NOW NOT USED
//        if ( mPreviousHeartbeatData.size() != pvs.size() )
//        {
//            mPreviousHeartbeatData.resize( pvs.size() );
//        }

//        if ( mPreviousHeartbeatData2.size() != mPreviousHeartbeatData.size() )
//        {
//            mPreviousHeartbeatData2.resize( mPreviousHeartbeatData.size() );
//        }

        for ( int k = 0; k < pvs.size(); k++ )
        {
            int sigIndex = k * SIGNAL_TYPE_COUNT;

            Q_ASSERT( mSignals.count() > sigIndex + 2 );
//            Q_ASSERT( mPreviousHeartbeatData.size() == pvs.size() );
//            Q_ASSERT( mPreviousHeartbeatData2.size() == mPreviousHeartbeatData.size() );

            if ( (mSignals.count() > sigIndex) )
            {
                const UCircAdaptSignal &    xSig        = mData.getXPVSignal( pvs[k] );
                const UCircAdaptSignal &    ySig        = mData.getYPVSignal( pvs[k] );
                PSignalData *               sigData     = &(mSignals[ sigIndex++ ]);
//                signalHeartBeatType &       prevData    = mPreviousHeartbeatData[ k ];
//                signalHeartBeatType &       prePrevData = mPreviousHeartbeatData2[ k ];

                // Set the new current data
//                if ( 0 == prePrevData.xData.count() )
//                {
//                    if ( 0 == prevData.xData.count() )
//                    {
                        sigData->setXData( *(xSig.getLastBeat()) );
                        sigData->setYData( *(ySig.getLastBeat()) );
//                    }
//                    else
//                    {
//                        // We want to have a little lagging in the displayed signal when it is changing due to
//                        // user changes. To obtain this, we store this signal (after this 'else') and initialyze
//                        // each new updated signal with the previous heart beat. In this way, two heart beats
//                        // will be shown on each update; the previous one and the new one
//                        sigData->setXData( prevData.xData );
//                        sigData->setYData( prevData.yData );

//                        sigData->addToXData( *(xSig.getLastBeat()) );
//                        sigData->addToYData( *(ySig.getLastBeat()) );
//                    }
//                }
//                else
//                {
//                    // We want to have a little lagging in the displayed signal when it is changing due to
//                    // user changes. To obtain this, we store this signal (after this 'else') and initialyze
//                    // each new updated signal with the previous heart beat. In this way, two heart beats
//                    // will be shown on each update; the previous one and the new one
//                    sigData->setXData( prePrevData.xData );
//                    sigData->setYData( prePrevData.yData );

//                    sigData->addToXData( prevData.xData );
//                    sigData->addToYData( prevData.yData );

//                    sigData->addToXData( *(xSig.getLastBeat()) );
//                    sigData->addToYData( *(ySig.getLastBeat()) );
//                }

//                // Store this heart beat for later (re-)use
//                if ( 0 < prevData.xData.count() )
//                {
//                    prePrevData.xData = prevData.xData;
//                    prePrevData.yData = prevData.yData;
//                }
//                prevData.xData = *(xSig.getLastBeat());
//                prevData.yData = *(ySig.getLastBeat());

                // Set the initial data
                if ( xSig.isReferenceUpdated() || ySig.isReferenceUpdated() )
                {
                    sigData = &(mSignals[ sigIndex++ ]);
                    sigData->setXData( xSig.getReferenceSignal() );
                    sigData->setYData( ySig.getReferenceSignal() );

                    updateInitalMarker  = true;
                    rescale             = true; // By setting this value to true, we force correct scaling the first time
                }
                else
                {
                    sigIndex++;
                }

                // Get the saved (snapshot) data
                if ( xSig.isSavedUpdated() || ySig.isSavedUpdated() )
                {
                    sigData = &(mSignals[ sigIndex ]);
                    sigData->setXData( xSig.getSavedSignal() );
                    sigData->setYData( ySig.getSavedSignal() );

                    updateSavedMarker = true;
                }
            }
        }

        if ( updateInitalMarker )
        {
            setInitialTimeMarker( mData.getInitialTimeMarker() );
        }

        if ( updateSavedMarker )
        {
            setSavedTimeMarker( mData.getSavedTimeMarker() );
        }

        scene()->invalidate( sceneTransform().mapRect( getBounds() ),
                             QGraphicsScene::ItemLayer | QGraphicsScene::ForegroundLayer );
    }

    return rescale;
}


bool PPVloop::adjustMarkerPosition( const QPointF & dataPos, const QPointF & graphPos, bool mouseReleased )
{
    Q_UNUSED( dataPos )
    Q_UNUSED( graphPos )
    Q_UNUSED( mouseReleased )

    bool markerAdjusted = false;

//    if ( mMarkers.count() > 0 )
//    {
//        PMarkerEllipse * markerPtr;

//        if ( mMarkers.hasSelectedMarker() )
//        {
//            // TODO
//            // We have to check if the selected position is near a line of the same type
//            markerPtr = marker_cast< PMarkerEllipse * >( mMarkers.getSelectedMarker() );

//            Q_ASSERT( markerPtr );

//            markerPtr->setMarkerData( dataPos );
//        }
//        else
//        {
//            // Not realy possible
//            markerPtr = marker_cast< PMarkerEllipse * >( mMarkers[0] );

//            Q_ASSERT( markerPtr );

//            markerPtr->setMarkerData( dataPos );
//        }

//        markerAdjusted = mData.setTimeMarkers( dataPos.x(), mSubType );
//    }

    return markerAdjusted;
}


bool PPVloop::setCurrentTimeMarker( double timeMarker )
{
    Q_UNUSED( timeMarker )

    bool changed = false;

//    if ( timeMarker >= 0 )
//    {
//        QList< pvloop_type > pvs( mData.getPVTypes() );

//        if ( !pvs.isEmpty() )
//        {
//            double volumeCorrection     = mData.toVolume( 1.0 );
//            double pressureCorrection   = mData.toPressure( 1.0 );

//            for ( int k = 0; k < pvs.size(); k++ )
//            {
//                int markerIndex = k * SIGNAL_TYPE_COUNT;

//                Q_ASSERT( mMarkers.count() > markerIndex );

//                if ( (mMarkers.count() > markerIndex) )
//                {
//                    const UCircAdaptSignal &    xSig    = mData.getXPVSignal( pvs[k] );
//                    const UCircAdaptSignal &    ySig    = mData.getYPVSignal( pvs[k] );

//                    if ( xSig.getLastBeat()->N() > 0 && ySig.getLastBeat()->N() > 0 )   // There is at least one beat
//                    {
//                        PMarkerEllipse * markerPtr = marker_cast< PMarkerEllipse * >( mMarkers[ markerIndex ] );

//                        Q_ASSERT( markerPtr );

//                        markerPtr->setMarkerData( xSig.getSignalSample( timeMarker ) * volumeCorrection,
//                                                  ySig.getSignalSample( timeMarker ) * pressureCorrection );

//                        changed = true;
//                    }
//                }
//            }
//        }
//    }

    return changed;
}


bool PPVloop::setInitialTimeMarker( double timeMarker )
{
    bool changed = false;

    if ( timeMarker >= 0 )
    {
        QList< pvloop_type > pvs( mData.getPVTypes() );

        if ( !pvs.isEmpty() )
        {
            double volumeCorrection     = mData.toVolume( 1.0 );
            double pressureCorrection   = mData.toPressure( 1.0 );

            for ( int k = 0; k < pvs.size(); k++ )
            {
                int markerIndex = k * SIGNAL_TYPE_COUNT + 1;

                Q_ASSERT( mMarkers.count() > markerIndex );

                if ( (mMarkers.count() > markerIndex) )
                {
                    const PVector &     xSigData        = mData.getXPVSignal( pvs[k] ).getReferenceSignal();
                    const PVector &     ySigData        = mData.getYPVSignal( pvs[k] ).getReferenceSignal();

                    if ( xSigData.N() > 0 && ySigData.N() > 0 )
                    {
                        PMarkerEllipse *    markerPtr       = marker_cast< PMarkerEllipse * >( mMarkers[ markerIndex ] );
                        int                 locTimeMarker   = (int)timeMarker;

                        Q_ASSERT( markerPtr );

                        // Correct the time for the maximum signal data length
                        while( locTimeMarker >= xSigData.N() )
                        {
                            locTimeMarker -= xSigData.N();
                        }

                        Q_ASSERT( ySigData.N() > locTimeMarker );

                        markerPtr->setMarkerData( xSigData[ locTimeMarker ] * volumeCorrection,
                                                  ySigData[ locTimeMarker ] * pressureCorrection );

                        changed = true;
                    }
                }
            }
        }
    }

    return changed;
}


bool PPVloop::setSavedTimeMarker( double timeMarker )
{
    bool changed = false;

    if ( timeMarker >= 0 )
    {
        QList< pvloop_type > pvs( mData.getPVTypes() );

        if ( !pvs.isEmpty() )
        {
            double volumeCorrection     = mData.toVolume( 1.0 );
            double pressureCorrection   = mData.toPressure( 1.0 );

            for ( int k = 0; k < pvs.size(); k++ )
            {
                int markerIndex = k * SIGNAL_TYPE_COUNT + 2;

                Q_ASSERT( mMarkers.count() > markerIndex );

                if ( (mMarkers.count() > markerIndex) )
                {
                    const PVector &     xSigData        = mData.getXPVSignal( pvs[k] ).getSavedSignal();
                    const PVector &     ySigData        = mData.getYPVSignal( pvs[k] ).getSavedSignal();

                    if ( xSigData.N() > 0 && ySigData.N() > 0 )
                    {
                        PMarkerEllipse *    markerPtr       = marker_cast< PMarkerEllipse * >( mMarkers[ markerIndex ] );
                        int                 locTimeMarker   = (int)timeMarker;

                        Q_ASSERT( markerPtr );

                        // Correct the time for the maximum signal data length
                        while( locTimeMarker >= xSigData.N() )
                        {
                            locTimeMarker -= xSigData.N();
                        }

                        Q_ASSERT( ySigData.N() > locTimeMarker );

                        markerPtr->setMarkerData( xSigData[ locTimeMarker ] * volumeCorrection,
                                                  ySigData[ locTimeMarker ] * pressureCorrection );

                        changed = true;
                    }
                }
            }
        }
    }

    return changed;
}


void PPVloop::contextMenuEvent(QGraphicsSceneContextMenuEvent *event)
{
    QList < pvloop_type> pvs( mData.getPVTypes() );

    if ( mContextMenu == NULL )
    {
        mMenuGroupPtr = new QActionGroup( this );
        if ( mMenuGroupPtr )
        {
            QPixmap pixmap( 16, 16 );

            mMenuGroupPtr->setExclusive( false );

            for ( int k = 0; k < pvs.size(); k++ )
            {
                int                         sigIndex    = k * SIGNAL_TYPE_COUNT;
                const UCircAdaptSignal &    xSig        = mData.getXPVSignal( pvs[k] );
                const UCircAdaptSignal &    ySig        = mData.getYPVSignal( pvs[k] );

                if ( xSig.getName() == ySig.getName() )
                {
                    QAction * action = new QAction( xSig.getName(), this );

                    pixmap.fill( xSig.getColor() );

                    action->setData( xSig.getName() );
                    action->setCheckable( true );
                    if ( mSignals.count() > sigIndex )
                    {
                        QFont actionFont = action->font();

                        action->setChecked( mSignals.at( sigIndex ).mustShow() );

                        actionFont.setBold( action->isChecked() );
                        action->setFont( actionFont );
                    }
                    action->setIcon( QIcon( pixmap ) );

                    mMenuGroupPtr->addAction( action );
                }
                else
                {
                    Q_ASSERT_X( false, "Error", "Non corresponding signales found for PV-loop drawing" );
                }
            }

            connect( mMenuGroupPtr, SIGNAL(triggered(QAction*)), this, SLOT(on_contextMenu(QAction*)) );

            mContextMenu = new QMenu( tr("signals") );
            mContextMenu->addActions( mMenuGroupPtr->actions() );
        }
    }

    if ( mContextMenu != NULL )
    {
        mContextMenu->exec( event->screenPos() );
    }
}
