/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: pcircpvloop.h 931 2013-11-27 19:03:51Z eelco $
 *
 */

#ifndef __PPVLOOP_H
#define	__PPVLOOP_H


#include <QMenu>

#include "PXYPlot.h"
#include "ucirculation.h"


namespace Peacs
{
    namespace CircAdapt
    {
        /** @class PGraphicsECGItem PGraphicsECGItem.h "ECGView/PGraphicsECGItem.h"
          * @brief The view window in which the ECGs are displayed
          */
        class PPVloop: public Support::PXYPlot
        {
            Q_OBJECT

        public:

            /** @brief constructor: ALl paramters are used to scale the signals the same in all these
              * graphics Items
              *
              * @param column the column at which this item will reside in the matrix (default -1)
              * @param row the row at which this item will reside in the matrix (default -1)
              * @param name name of the signal
              * @see PECGView
              */
            PPVloop( UCirculation & data, int column, int row );


            /** @brief
              */
            void init( void );


            /** @brief Called to update the data
              * This is a convenient interface function that is used to signal this graph instance
              * classes in a uniformed way to update the signal data it contains due to, for
              * example, data changes
              *
              * @return true if the added data should initialte a rescale
              */
            bool updateSignals( void );


            /** @brief Is called to get the tolltip text for the specified time position
              * This function is called from the event handling function that handles tooltip request
              * events, @see PGraphicsSignalView::events().
              *
              * @param timePosition The requested time position in data coordinate system
              * @return The text to use as the tooltip or an empty string if no tooltip should be shown
              */
            QString getToolTipText( double timePosition ) const;


            /** @brief Sets mDrawInitial
              * This will force a redraw of this item to make the initial signal show/hide
              *
              * @param draw the new value for mDrawInitial
              */
            void setDrawInitial( bool draw );


            /** @brief Returns mDrawInitial
              */
            inline double isDrawInitial( void ) const { return mDrawInitial; }


            /** @brief Sets mDrawSaved
              * This will force a redraw of this item to make the saved signal show/hide
              *
              * @param draw the new value for mDrawSaved
              */
            void setDrawSaved( bool draw );


            /** @brief Returns mDrawSaved
              */
            inline double isDrawSaved( void ) const { return mDrawSaved; }


            /** Sets mCurrentTimeMarker to the new value
              * Checks if the timemarker shall change and sets mCurrentTimeMarker
              * accordingly
              *
              * @param timeMarker the new current timemarker value
              * @return true if the timemarker value was changed
              */
            bool setCurrentTimeMarker( double timeMarker );


            /** Sets mInitialTimeMarker to the new value
              * Checks if the timemarker shall change and sets mInitialTimeMarker
              * accordingly
              *
              * @param timeMarker the new initial timemarker value
              * @return true if the initial timemarker value was changed
              */
            bool setInitialTimeMarker( double timeMarker );


            /** Sets mSavedTimeMarker to the new value
              * Checks if the timemarker shall change and sets mSavedTimeMarker
              * accordingly
              *
              * @param timeMarker the new snapshot timemarker value
              * @return true if the snapshot timemarker value was changed
              */
            bool setSavedTimeMarker( double timeMarker );


        public slots:

            /** @brief Is called when the user has selected and item in the popup menu for this item
              *
              * @param action Information about the items selected
              */
            void on_contextMenu( QAction * action );


        protected:

            /** @brief This function is called to adjust the position of the marker for the given position
              * This function will be called from the base class by the mouse click handling function to
              * indicate a user action at the specified position in the graph. The given positions will be
              * in data coordinates system and graph coordinate system to enable maximum control.
              *
              * @param dataPos The position to adjust with. This position is in data coordinate system
              * @param graphPos The position to adjust with. This position is in graph coordinate system
              * @param mouseReleased If true, this function is being called from the mouse release event, otherwise the mouse move event
              *
              * @return true if the position has changed
              */
            bool adjustMarkerPosition( const QPointF & dataPos, const QPointF & graphPos, bool mouseReleased );


            /** @brief Is called when the user presses the right mouse button over this item to show the popup menu
              *
              * @param event Points to information about the event
              */
            void contextMenuEvent( QGraphicsSceneContextMenuEvent * event );


        private:

            /** @brief This structure is used to store the previous signal data that will be added to the next signal update
              */
            typedef struct
            {
                PVector xData;
                PVector yData;
            } signalHeartBeatType;


            /** @brief
              */
            UCirculation & mData;


            /** @brief determins whether the initial signals are drawn.
              */
            bool mDrawInitial;


            /** @brief determins whether the saved signals are drawn.
              */
            bool mDrawSaved;


            /** Pointer to the context menu instance created for this graphics view item
              */
            QMenu * mContextMenu;


            /** used to keep the menu-items from the context menu grouped
              */
            QActionGroup *  mMenuGroupPtr;


// FOR NOW NOT USED
//            /** @brief This member will be used to store the previous heartbeat data
//              * This heartbeat data will then be used to add it again when a new heartbeat is
//              * recieved. In this way a loop of moe then one heartbeat is shown. Gives a nice
//              * effect when the signal is changing
//              */
//            QVector<signalHeartBeatType> mPreviousHeartbeatData;


//            /** @brief This member will be used to store the previous-previous heartbeat data
//              * This heartbeat data will then be used to add it again when a new heartbeat is
//              * recieved. In this way a loop of moe then one heartbeat is shown. Gives a nice
//              * effect when the signal is changing
//              * By using this member there will be shown three heartbeats
//              */
//            QVector<signalHeartBeatType> mPreviousHeartbeatData2;
        };
    }
}
#endif	/* __PCIRCSIGNAL_H */

