/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: pcircsignal.cpp 1173 2014-03-21 14:12:14Z eelco $
 *
 */

#include <QDebug>

#include "pcircsignal.h"
#include "PStatistics.h"
#include "PCircAdaptDefines.h"
#include "PMarkerLine.h"


#include "math.h"


using namespace Peacs::Support;
using namespace Peacs::CircAdapt;



PGraphicsCircSignal::PGraphicsCircSignal( signal_type     type,
                                          subSignal_type  subtype,
                                          UCirculation  & data,
                                          double          zoomT0,
                                          double          zoomT1,
                                          int             column,
                                          int             row ) :
    PSignalPlot( false, zoomT0, zoomT1, column, row ),
    mData( data ),
    mType( type ),
    mSubType( subtype ),
    mMarkerHoverCursorShown( false ),
    mContextMenu( NULL ),
    mMenuGroupPtr( NULL )
{
    setSignalAmpl( 0, 100 );

    setAcceptedMouseButtons( acceptedMouseButtons() | Qt::RightButton );
    setAcceptHoverEvents( true );   // Enable tracking of marker hover events
}


QString PGraphicsCircSignal::getToolTipText( double timePosition ) const
{
    QString valStr;

    if ( timePosition >= getZoomT0() &&
         timePosition < getZoomT1() )
    {
        const QList < UCircAdaptSignal *> & mySignals   = mData.getSignals( mType );

        valStr += QString("time:  %1 s").arg( timePosition / getXLabelFactor(), 5, 'f', 2 );

        for ( int k = 0; k < mSignals.size(); k++ )
        {
            const PSignalData & sigData = mSignals.at( k );

            if ( sigData.mustShow() &&
                 sigData.count() > timePosition )
            {
                valStr += QString("\n%1:  %2 %3").
                        arg( sigData.getLabel() ).
                        arg( sigData.getYData( timePosition ) * sigData.getCorrectionY(), 5, 'g', 3 ).
                        arg( mySignals[k]->getUnits() );
            }
        }
    }

    return valStr;
}


void PGraphicsCircSignal::on_contextMenu(QAction *action)
{
    const QList < UCircAdaptSignal *> & mySignals   = mData.getSignals( mType );
    bool                                done        = false;
    QFont                               actionFont  = action->font();

    actionFont.setBold( action->isChecked() );
    action->setFont( actionFont );

    for ( int k = 0; k < mySignals.size() && !done; k++ )
    {
        if ( action->data() == mySignals[k]->getName() )
        {
            Q_ASSERT( mSignals.count() > k );

            if ( (mSignals.count() > k) &&
                 (mSignals.at( k ).mustShow() != action->isChecked()) )
            {
                mSignals[ k ].setShow( action->isChecked() );

                emit( showSignals( k ) );

                triggerRescale();

                // Redraw the foreground
                scene()->invalidate( sceneBoundingRect(), QGraphicsScene::ItemLayer /*ForegroundLayer*/ );
            }

            done = true;
        }
    }
}


void PGraphicsCircSignal::updateShowSignals( int show )
{
    Q_ASSERT( show >= 0 && show < mSignals.count() );

    if ( show >= 0 && show < mSignals.count() )
    {
        // Toggle the show status for this signal
        mSignals[ show ].toggleShow();

        Q_ASSERT( ( show < mMenuGroupPtr->actions().size() ) );

        if ( show < mMenuGroupPtr->actions().size() )
        {
            QAction *   actionPtr = mMenuGroupPtr->actions().at( show );
            QFont       actionFont = actionPtr->font();

            actionPtr->toggle();

            actionFont.setBold( actionPtr->isChecked() );
            actionPtr->setFont( actionFont );
        }

        // New signal added or removed. The needed scale may be changed
        triggerRescale();

        // Redraw the foreground
        scene()->invalidate( sceneBoundingRect(), QGraphicsScene::ItemLayer /*ForegroundLayer*/ );
    }
}


void PGraphicsCircSignal::init( void )
{
    const QList < UCircAdaptSignal *> & mySignals   = mData.getSignals( mType );
    double                              correction  = getCorrection();

    mSignals.clear();
    mMarkers.clear();

    // First, initialyze all signals that can be shown
    if ( !mySignals.isEmpty() )
    {
        for ( int k = 0; k < mySignals.size(); k++ )
        {
            const UCircAdaptSignal *    sigPtr( mySignals[k] );
            QPen                        signalPen( sigPtr->getColor() );
            bool                        show( false );

            // Hard coded signals that must be shown initially
            if ( mType == PRESSURE )
            {
                show = sigPtr->getName() == mData.aorta().getName() /*||
                       sigPtr->getName() == mData.heart().ventricles().leftVentricle().getName()*/;
            }
            /*else if ( mType == VOLUME )
            {
                show = sigPtr->getName() == mData.heart().ventricles().leftVentricle().getName();
            }*/

            // Initialyze the current signal data object
            mSignals.append( k, sigPtr->getName(), show, signalPen, 1.0, correction );
        }

        if ( mSubType == INITIAL_PLOT )
        {
            //setYLabel( QString("[%1]").arg( mySignals[0]->getUnits() ) );
            //setYUnit( QString("[%1]").arg( mySignals[0]->getUnits() ) );
        }
        else
        {
            // Only shown the Y-axis labels at the left side in the graphics view graphs
            //setShowYScaleTexts( false );
        }
    }
    else
    {
//        if ( mSubType == INITIAL_PLOT )
//        {
//            setYLabel( QString("") );
//            //setYUnit( QString("") );
//        }
    }

    if ( mType == FLOWVELOCITY )
    {
        setXLabel( "Time [s]" );
        //setXUnit( "Time [s]" );
    }
    else
    {
        // Only shown the X-axis labels at the bottom of the graphics view graphs
        setShowXScaleTexts( false );
    }

    // Now initialyze, if needed, the marker(s) that will be used
    if ( mMarkers.isEmpty() )
    {
        if ( mSubType == INITIAL_PLOT )
        {
            PMarkerLine * markerPtr = new PMarkerLine( 0, subSignal_type_str[ mSubType ], true, QPen( TIMELINECOLORINIT ) );
            markerPtr->pen().setWidthF( 2.0 );
            markerPtr->setMarkerData( 0.0, getMaxSigAmpl(), 0.0, getMinSigAmpl() );

            mMarkers.addMarker( markerPtr );
        }
        else if ( mSubType == SAVED_PLOT )
        {
            PMarkerLine * markerPtr = new PMarkerLine( 0, subSignal_type_str[ mSubType ], true, QPen( TIMELINECOLORSAVED ) );
            markerPtr->pen().setWidthF( 2.0 );
            markerPtr->setMarkerData( 0.0, getMaxSigAmpl(), 0.0, getMinSigAmpl() );

            mMarkers.addMarker( markerPtr );
        }
    }

    if ( mContextMenu != NULL )
    {
        if ( mMenuGroupPtr != NULL )
        {
            disconnect( mMenuGroupPtr, SIGNAL(triggered(QAction*)), this, SLOT(on_contextMenu(QAction*)) );

            delete mMenuGroupPtr;
            mMenuGroupPtr = NULL;
        }

        delete mContextMenu;
        mContextMenu = NULL;
    }

    mMenuGroupPtr = new QActionGroup( this );
    if ( mMenuGroupPtr )
    {
        QPixmap pixmap( 16, 16 );

        mMenuGroupPtr->setExclusive( false );

        for ( int k = 0; k < mySignals.size(); k++ )
        {
            QAction * action = new QAction( mySignals[k]->getName(), this );

            pixmap.fill( mySignals[k]->getColor() );

            action->setData( mySignals[k]->getName() );
            action->setCheckable( true );
            if ( mSignals.count() > k )
            {
                QFont actionFont = action->font();

                action->setChecked( mSignals.at( k ).mustShow() );

                actionFont.setBold( action->isChecked() );
                action->setFont( actionFont );
            }
            action->setIcon( QIcon( pixmap ) );

            mMenuGroupPtr->addAction( action );
        }

        connect( mMenuGroupPtr, SIGNAL(triggered(QAction*)), this, SLOT(on_contextMenu(QAction*)) );

        mContextMenu = new QMenu( tr("signals") );
        mContextMenu->addActions( mMenuGroupPtr->actions() );
    }

    PSignalPlot::init();    // Init base class
}


void PGraphicsCircSignal::setTimeMarker( double timeMarker )
{
    if ( mMarkers.count() > 0 )
    {
        PMarkerLine * markerPtr = marker_cast< PMarkerLine * >( mMarkers[0] );

        Q_ASSERT( markerPtr );

        if ( markerPtr )
        {
            QPainterPath    pPath = markerPtr->translate( getGraphTransform() );    // Get the current marker polygon
            QRectF          invalidateRect;

            // Set the new position
            markerPtr->setMarkerData( timeMarker, getMaxSigAmpl(), timeMarker, getMinSigAmpl() );

            pPath.addPath( markerPtr->translate( getGraphTransform(), true ) );     // Add the second marker polygon to the first one
            invalidateRect = pPath.boundingRect();                                  // Get the extent of both marker positions

            if ( scene() )
            {
                scene()->invalidate( mapRectToScene( invalidateRect ), QGraphicsScene::ForegroundLayer );
            }
            else
            {
                resetWaitForDraw();
            }
        }
        else
        {
            resetWaitForDraw();
        }
    }
}


double PGraphicsCircSignal::getCorrection( void ) const
{
    double correction = 1.0;

    if ( mType == PRESSURE )
    {
        correction = mData.toPressure( 1.0 );
    }
    else if ( mType == VOLUME )
    {
        correction = mData.toVolume( 1.0 );
    }
    else if ( mType == FLOW )
    {
        correction = 1e4;
    }

    return correction;
}


bool PGraphicsCircSignal::updateSignals( void )
{
    bool                                rescale     = false;
    const QList < UCircAdaptSignal *> & mySignals   = mData.getSignals( mType );

    if ( !mySignals.isEmpty() && (mSignals.size() == mySignals.size()) )
    {
        bool needUpdate     = false;

        for ( int k = 0; k < mySignals.size(); k++ )
        {
            bool                        signalChanged   = false;
            const UCircAdaptSignal *    sourceSignalPtr = mySignals.at( k );
            PSignalData &               sigData         = mSignals[ k ];
            double                      dt              = sourceSignalPtr->getSampleTime();

            switch ( mSubType )
            {
            case INITIAL_PLOT:
                if ( sourceSignalPtr->isReferenceUpdated() )
                {
                    signalChanged   = true;
                    needUpdate      = true;
                    rescale         = true; // By setting this value to true, we force correct scaling the first time

                    sigData.clear();
                    sigData.setYData( sourceSignalPtr->getReferenceSignal( getZoomT0() / dt, getZoomT1() / dt ) );
                }
                break;
            case SAVED_PLOT:
                if ( sourceSignalPtr->isSavedUpdated() )
                {
                    signalChanged   = true;
                    needUpdate      = true;

                    sigData.clear();
                    sigData.setYData( sourceSignalPtr->getSavedSignal( getZoomT0() / dt, getZoomT1() / dt ) );
                }
                break;
            case CURRENT_PLOT:
                // Nothing to do here. This drawing is done with class PCircSignalDraw
                break;
            default:
                Q_ASSERT( false );
                break;
            }

            if ( signalChanged )
            {
                if ( dt != 1 && dt != 0 )   // 0 is illegal and 1 is not needed; the PSignalData will handle it correctly
                {
                    sigData.fillXDataForY( dt, getZoomT0() / dt );
                }
            }
        }

        if ( needUpdate )
        {
            scene()->invalidate( sceneBoundingRect(), QGraphicsScene::ItemLayer /*ForegroundLayer*/ );
        }
    }

    return rescale;
}


bool PGraphicsCircSignal::adjustMarkerPosition( const QPointF & dataPos, const QPointF & graphPos, bool mouseReleased )
{
    Q_UNUSED( graphPos )
    Q_UNUSED( mouseReleased )

    bool markerAdjusted = false;

    if ( (dataPos.x() >= getZoomT0()) && (dataPos.x() <=getZoomT1()) && (mMarkers.count() > 0) )
    {
        markerAdjusted = mData.setTimeMarkers( dataPos.x(), mSubType );
    }

    return markerAdjusted;
}


void PGraphicsCircSignal::onMarkerHoverEvent( int markerIndex )
{
//qDebug() << "Marker hover event:" << markerIndex;
    if ( !mMarkerHoverCursorShown && (PSignalMarkers::NO_MARKER_FOUND != markerIndex) )
    {
        setCursor( Qt::SizeHorCursor );

        mMarkerHoverCursorShown = true;
    }
    else if ( mMarkerHoverCursorShown && (PSignalMarkers::NO_MARKER_FOUND == markerIndex) )
    {
        unsetCursor();

        mMarkerHoverCursorShown = false;
    }
}


void PGraphicsCircSignal::contextMenuEvent(QGraphicsSceneContextMenuEvent *event)
{
    if ( mContextMenu != NULL )
    {
        mContextMenu->exec(event->screenPos());
    }
}
