/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: tchamber.cpp 776 2013-07-11 20:33:36Z eelco $
 *
 */

#include <QDebug>

#include "math.h"
#include "defs.h"
#include "chamber.h"
#include "tchamber.h"
#include "uchamber.h"

using namespace Peacs::CircAdapt;



// Constructor

TChamber::TChamber():
    TCompliance(),
    mAmRef(0.0),
    mAmDead(0.0),
    mVWall(0.0),
    mVm(0.0),
    mXm(0.0),
    mYm(0.0),
    mSheetAreaMid(0.0),
    mCm(0.0),
    mT(0.0),
    mPtrans(0.0),
    mDTDAm(0.0),
    mY(0.0),
    mYDot(0.0),
//    mLRef(0.0),
//    mEAmRef(0.0),
    mPatches(),
    mChamberType(UNDEFINED_TYPE),
    mAdaptToExercise(false),
    mUseTriSeg(false)
{

}

TChamber::TChamber(const TChamber & chamber):
    TCompliance(chamber),
    mAmRef(chamber.mAmRef),
    mAmDead(chamber.mAmDead),
    mVWall(chamber.mVWall),
    mVm(chamber.mVm),
    mXm(chamber.mXm),
    mYm(chamber.mYm),
    mSheetAreaMid(chamber.mSheetAreaMid),
    mCm(chamber.mCm),
    mT(chamber.mT),
    mPtrans(chamber.mPtrans),
    mDTDAm(chamber.mDTDAm),
    mY(chamber.mY),
    mYDot(chamber.mYDot),
//    mLRef(chamber.getLRef()),
//    mEAmRef(chamber.getEAmRef()),
    mPatches(/*chamber.mPatches*/),
    mChamberType(chamber.mChamberType),
    mAdaptToExercise(chamber.mAdaptToExercise),
    mUseTriSeg(chamber.mUseTriSeg)
{
    for (int i=0; i < chamber.getPatches().size(); i++)
    {
        mPatches.append(new TPatch( *chamber.getPatches().at(i) ));
    }
}



TChamber::TChamber(const SChamber & chamber):
    TCompliance(chamber),
    mAmRef(chamber.getAmRef()),
    mAmDead(chamber.getAmDead()),
    mVWall(chamber.getVwall()),
    mVm(),
    mXm(),
    mYm(),
    mSheetAreaMid(),
    mCm(),
    mT(),
    mPtrans(),
    mDTDAm(),
    mY(),
    mYDot(),
//    mLRef(chamber.getLRef()),
//    mEAmRef(chamber.getEAmRef()),
    mPatches(),
    mChamberType(chamber.getChamberType()),
    mAdaptToExercise(chamber.isAdapationOn()),
    mUseTriSeg(chamber.isTriSeg())
{
    for (int i=0; i < chamber.getPatches().size(); i++)
    {
        mPatches.append(new TPatch( *chamber.getPatches().at(i) ));
    }
}

TChamber::~TChamber()
{
    for ( int i=0; i < mPatches.size(); i++)
    {
        delete mPatches[i];
    }
}


const TChamber & TChamber::operator =(const TChamber &chamber)
{
    if ( this != &chamber)
    {
        TCompliance::operator =(chamber);

        mAmRef              = chamber.mAmRef;
        mAmDead             = chamber.mAmDead;
        mVWall              = chamber.mVWall;
        mVm                 = chamber.mVm;
        mXm                 = chamber.mXm;
        mYm                 = chamber.mYm;
        mSheetAreaMid       = chamber.mSheetAreaMid;
        mCm                 = chamber.mCm;
        mT                  = chamber.mT;
        mPtrans             = chamber.mPtrans;
        mDTDAm              = chamber.mDTDAm;
        mY                  = chamber.mY;
        mYDot               = chamber.mYDot;
//        mLRef               = chamber.mLRef;
//        mEAmRef             = chamber.mEAmRef;
        mChamberType        = chamber.mChamberType;
        mAdaptToExercise    = chamber.mAdaptToExercise;
        mUseTriSeg          = chamber.mUseTriSeg;

        for ( int i = 0; i < chamber.mPatches.size(); i++ )
        {
            if ( mPatches.size() <= i )
            {
                mPatches.append(new TPatch( ));
            }

            *mPatches[i] = *chamber.mPatches[i];
        }
    }

    return *this;
}

void TChamber::copyUserSpecifications(const UChamber &chamber)
{
    TCompliance::copyUserSpecifications( chamber );
    mAmRef          = chamber.getAmRef();
    mAmDead         = chamber.getAmDead();
    mVWall          = chamber.getVwall();
    mAdaptToExercise= chamber.isAdapationOn();
    mUseTriSeg      = chamber.isTriSeg();
//    mEAmRef         = chamber.getEAmRef();


    if ( mPatches.size() != chamber.getPatches().size() )
    {
        for ( int i=0; i < mPatches.size(); i++)
        {
            delete mPatches[i];
        }

        mPatches.clear();

        for (int i=0; i < chamber.getPatches().size(); i++)
        {
            mPatches.append(new TPatch( ));
        }
    }

    for (int i=0; i < chamber.getPatches().size(); i++)
    {
        mPatches[i]->copyUserSpecifications( *chamber.getPatches().at(i) );
    }

}


void TChamber::copySpecifications(const SChamber &chamber)
{
    TCompliance::copySpecifications( chamber );
    mAmRef          = chamber.getAmRef();
    mAmDead         = chamber.getAmDead();
    mVWall          = chamber.getVwall();
    mAdaptToExercise= chamber.isAdapationOn();
    mUseTriSeg      = chamber.isTriSeg();
//    mEAmRef         = chamber.getEAmRef();


    if ( mPatches.size() != chamber.getPatches().size() )
    {
        for ( int i=0; i < mPatches.size(); i++)
        {
            delete mPatches[i];
        }

        mPatches.clear();

        for (int i=0; i < chamber.getPatches().size(); i++)
        {
            mPatches.append(new TPatch( *chamber.getPatches().at(i) ) );
        }
    }
    else
    {
        for (int i=0; i < chamber.getPatches().size(); i++)
        {
            mPatches[i]->copySpecifications( *chamber.getPatches().at(i) );
        }
    }

}

void TChamber::resetData( int size )
{
    TCompliance::resetData( size );
    mVm.resize(size);             // Volume enclosed by midwall
    mXm.resize(size);             // Central bulge distance of midwall surface
    mYm.resize(size);             // Radius of midwall at the junctional edge
    mSheetAreaMid.resize(size);   // Area of wall laid as sheet
    mCm.resize(size);             // Curvature = 1/radius of midwall
    mT.resize(size);              // Wall tension [N/m]
    mPtrans.resize(size);         // Transmural pressure
    mDTDAm.resize(size);          // dTension/dArea compliance (wall stiffness)
    mY.resize(size);              // State variable Y for TriSeg computation
    mYDot.resize(size);           // d/dt Y for TriSeg computation
//    mLRef.resize(size);

    for (int i=0; i < mPatches.size(); i++)
    {
        mPatches[i]->resetData( size );
    }
}

void TChamber::addSample( const SChamber & sim, int pos)
{
    TCompliance::addSample(sim,pos);
    mVm[pos] = sim.getVm();             // Volume enclosed by midwall
    mXm[pos] = sim.getXm();             // Central bulge distance of midwall surface
    mYm[pos] = sim.getYm();             // Radius of midwall at the junctional edge
    mSheetAreaMid[pos] = sim.getAm();   // Area of wall laid as sheet
    mCm[pos] = sim.getCm();             // Curvature = 1/radius of midwall
    mT[pos] = sim.getT();              // Wall tension [N/m]
    mPtrans[pos] = sim.getPtrans();         // Transmural pressure
    mDTDAm[pos] = sim.getDTDAm();          // dTension/dArea compliance (wall stiffness)
    mY[pos] = sim.getYm();              // State variable Y for TriSeg computation
    mYDot[pos] = sim.getYdot();           // d/dt Y for TriSeg computation
//    mLRef[pos] = sim.getLRef();

    for (int i=0; i < mPatches.size(); i++)
    {
        mPatches[i]->addSample( *sim.getPatches().at(i), pos );
    }
}
