/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: tchamber.h 789 2013-07-18 13:22:28Z peter $
 *
 */

#ifndef __TCHAMBER_H_
#define __TCHAMBER_H_

#include <QMetaType>

#include "tpatch.h"
#include "tcompliance.h"

#include "pvector.h"
#include "PCircAdaptDefines.h"


using namespace Peacs::Support;

namespace Peacs
{
    using namespace  Support;
    namespace CircAdapt
    {

        class SChamber;
        class UChamber;

        /**
        Class for muscular chamber. Inherits from UCompliance.
        */
        class TChamber: public TCompliance
        {
            Q_OBJECT

        public:
            TChamber();


            TChamber(const TChamber & chamber);

            /* @brief Constructor */
            TChamber(const SChamber & chamber);

            virtual ~TChamber();

            virtual const TChamber& operator = (const TChamber &chamber);


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const UChamber &chamber);


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const SChamber &);


            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );


            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            void addSample( const SChamber & sim, int pos);





            void init();

            bool addPatch();


            double getMinActivationTime() const;

            inline chamber_type getChamberType() const { return mChamberType; }

            inline bool isAdapationOn() const { return mAdaptToExercise; }

            inline bool isVentricle() const { return mChamberType == RIGHT_VENTRICLE || mChamberType == SEPTUM || mChamberType == LEFT_VENTRICLE ; }

            inline bool isTriSeg() const { return mUseTriSeg; }

            /// Method to get wall volume
            inline double getWallVolume() const { return mVWall; }

//            inline const PVector &  getLRef() const { return mLRef; }

            inline const PVector & getY() const { return mY; }

            inline const PVector & getAm() const { return mSheetAreaMid; }

            inline double getAmRef() const { return mAmRef; }


            inline double getAmDead() const { return mAmDead; }

            inline const PVector &getCm() const { return mCm; }

            inline const PVector &getDTDAm() const { return mDTDAm; }

//            inline double getEAmRef() const { return mEAmRef; }

            inline const PVector &getPtrans() const { return mPtrans; }

            inline const PVector &getT() const { return mT; }

            inline const PVector &getVm() const { return mVm; }

            inline double getVwall() const { return mVWall; }

            inline const PVector &getYdot() const { return mYDot; }

            inline const PVector &getXm() const { return mXm; }

            inline const PVector &getYm() const { return mYm; }

            inline const QList < TPatch * > &getPatches() const { return mPatches; }

        protected:
            // Specification parameters
            double mAmRef;     // Reference area of wall laid as sheet
            double mAmDead;    // Non-contractile area (Valve openings)
            double mVWall;     // Wall volume

            // Simulation state variables
            PVector mVm;             // Volume enclosed by midwall
            PVector mXm;             // Central bulge distance of midwall surface
            PVector mYm;             // Radius of midwall at the junctional edge
            PVector mSheetAreaMid;   // Area of wall laid as sheet
            PVector mCm;             // Curvature = 1/radius of midwall
            PVector mT;              // Wall tension [N/m]
            PVector mPtrans;         // Transmural pressure
            PVector mDTDAm;          // dTension/dArea compliance (wall stiffness)
            PVector mY;              // State variable Y for TriSeg computation
            PVector mYDot;           // d/dt Y for TriSeg computation

            // List of pointers to patches
            // Patches form the sheet encapsulating the cavity
            QList < TPatch * > mPatches;

            chamber_type mChamberType;

            bool mAdaptToExercise;

            bool mUseTriSeg;

        };
    }
}

// This makes it possible for TTubeState values to be stored in QVariant
// objects and retrieved later
Q_DECLARE_METATYPE( Peacs::CircAdapt::TChamber )


#endif
