/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: tcirculation.cpp 781 2013-07-14 17:01:20Z peter $
 *
 */

#include <QDebug>


#include "tcirculation.h"
#include "circulation.h"
#include "ucirculation.h"


using namespace Peacs::CircAdapt;

TCirculation::TCirculation() :
    TCircElement(),
    mPressureControl( false ),
    mAdaptRest( false ),
    mAdaptExercise( false ),
    mAorta(),
    mPulmonaryVeins(),
    mPulmonaryArtery(),
    mSystemicVein(),
    mHeart(),
    mTubes(),
    mResistances(),
    mQ0(8.5000e-005),
    mP0(12200),
    mHeartbeatTime( HEARTBEAT_NOT_SET ),
    mDropPulm(2500.0),
    mAdtapStatus(TransitionAdaptation)
{
}


// Constructor: initialization only
TCirculation::TCirculation(const TCirculation &circ) :
    TCircElement(    circ ),
    mPressureControl(circ.mPressureControl ),
    mAdaptRest(      circ.mAdaptRest ),
    mAdaptExercise(  circ.mAdaptExercise ),
    mAorta(          circ.mAorta ),
    mPulmonaryVeins( circ.mPulmonaryVeins ),
    mPulmonaryArtery(circ.mPulmonaryArtery ),
    mSystemicVein(   circ.mSystemicVein),
    mHeart(          circ.mHeart),
    mTubes(          circ.mTubes   ),
    mResistances(    circ.mResistances ),
    mQ0(             circ.venousReturn()),
    mP0(             circ.meanArterialPressure()),
    mHeartbeatTime(  circ.mHeartbeatTime ),
    mDropPulm(       circ.mDropPulm ),
    mAdtapStatus(    circ.mAdtapStatus)
{
//    for ( int i=0; i < mResistances.size(); i++)
//    {
//        mResistances.append(TResistance(circ.getResistances().at(i)));
//    }

//    for ( int i=0; i < mTubes.size(); i++)
//    {
//        mTubes.append(TTube(circ.getTubes().at(i)));
//    }
}


TCirculation::TCirculation( const SCirculation &circ ) :
    TCircElement(    circ.getName() ),
    mPressureControl(circ.pressureControl() ),
    mAdaptRest(      circ.adaptForRest() ),
    mAdaptExercise(  circ.adaptForExercise() ),
    mAorta(          circ.getAorta() ),
    mPulmonaryVeins( circ.getPulmonaryVeins() ),
    mPulmonaryArtery(circ.getPulmonaryArtery() ),
    mSystemicVein(   circ.getSystemicVein() ),
    mHeart(          circ.getHeart() ),
    mQ0(             circ.venousReturn()),
    mP0(             circ.meanArterialPressure()),
    mHeartbeatTime( HEARTBEAT_NOT_SET ),
    mDropPulm(       circ.getPulmDrop()),
    mAdtapStatus(    circ.getAdaptStatus())
{
    for ( int i = 0; i < circ.getResistances().size(); i++ )
    {
        mResistances.append(TResistance(circ.getResistances().at(i)));
    }

    for ( int i = 0; i < circ.getTubes().size(); i++ )
    {
        mTubes.append( TTube( circ.getTubes().at( i ) ) );
    }
}


TCirculation::~TCirculation()
{
}


const TCirculation& TCirculation::operator = (const TCirculation &circ)
{
    if ( this != &circ )
    {
        TCircElement::operator  =( circ );

        mPressureControl    = circ.mPressureControl;
        mAdaptRest          = circ.mAdaptRest;
        mAdaptExercise      = circ.mAdaptExercise;
        mAorta              = circ.mAorta;
        mPulmonaryVeins     = circ.mPulmonaryVeins;
        mPulmonaryArtery    = circ.mPulmonaryArtery;
        mSystemicVein       = circ.mSystemicVein;
        mHeart              = circ.mHeart;
        mTubes              = circ.mTubes;
        mResistances        = circ.mResistances;
        mQ0                 = circ.venousReturn();
        mP0                 = circ.meanArterialPressure();
        mHeartbeatTime      = circ.mHeartbeatTime;
        mDropPulm           = circ.mDropPulm;
        mAdtapStatus        = circ.mAdtapStatus;
    }

    return *this;
}


void TCirculation::resetData( int size)
{
    mHeart.resetData( size );
    mAorta.resetData( size );
    mPulmonaryVeins.resetData( size );
    mPulmonaryArtery.resetData( size );
    mSystemicVein.resetData( size );

    for ( int i = 0; i < mTubes.size(); i++ )
    {
        mTubes[i].resetData( size );
    }

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        mResistances[i].resetData( size );
    }
}


void TCirculation::copyUserSpecifications( const UCirculation & circ )
{
    mPressureControl = circ.pressureControl();
    mAdaptRest       = circ.adaptForRest();
    mAdaptExercise   = circ.adaptForExercise();
    mQ0              = circ.venousReturn();
    mP0              = circ.meanArterialPressure();

    mAorta.copyUserSpecifications( circ.getAorta() );
    mPulmonaryVeins.copyUserSpecifications( circ.getPulmonaryVeins() );
    mPulmonaryArtery.copyUserSpecifications( circ.getPulmonaryArtery() );
    mSystemicVein.copyUserSpecifications( circ.getSystemicVein() );
    mHeart.copyUserSpecifications( circ.getHeart() );

    mResistances.resize( circ.getResistances().size() );
    mTubes.resize( circ.getTubes().size() );

    for ( int i=0; i < mTubes.size(); i++)
    {
        mTubes[i].copyUserSpecifications(  circ.getTubes().at(i) );
    }
    for ( int i=0; i < mResistances.size(); i++)
    {
        mResistances[i].copyUserSpecifications( circ.getResistances().at(i) );
    }
    mDropPulm = circ.getPulmDrop();
}


void TCirculation::copySpecifications( const SCirculation & circ )
{
    mPressureControl = circ.pressureControl();
    mAdaptRest       = circ.adaptForRest();
    mAdaptExercise   = circ.adaptForExercise();
    mQ0              = circ.venousReturn();
    mP0              = circ.meanArterialPressure();
    mAdtapStatus     = circ.getAdaptStatus();


    mAorta.copySpecifications( circ.getAorta() );
    mPulmonaryVeins.copySpecifications( circ.getPulmonaryVeins() );
    mPulmonaryArtery.copySpecifications( circ.getPulmonaryArtery() );
    mSystemicVein.copySpecifications( circ.getSystemicVein() );
    mHeart.copySpecifications( circ.getHeart() );

    mResistances.resize( circ.getResistances().size() );
    mTubes.resize( circ.getTubes().size() );

    for ( int i = 0; i < mTubes.size(); i++ )
    {
        mTubes[i].copySpecifications(  circ.getTubes().at(i) );
    }

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        mResistances[i].copySpecifications( circ.getResistances().at(i) );
    }
    mDropPulm = circ.getPulmDrop();

}


void TCirculation::addSample( const SCirculation & circ ,int pos)
{
    mHeart.addSample(            circ.getHeart(), pos );

    for ( int i = 0; i < mTubes.size(); i++ )
    {
        mTubes[i].addSample( circ.getTubes().at( i ), pos );
    }

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        mResistances[i].addSample( circ.getResistances().at( i ), pos );
    }

    mAorta.addSample(            circ.getAorta() , pos);
    mPulmonaryVeins.addSample( circ.getPulmonaryVeins(), pos );
    mPulmonaryArtery.addSample(  circ.getPulmonaryArtery(), pos );
    mSystemicVein.addSample(     circ.getSystemicVein(), pos );
}


TTube * TCirculation::getTube( const QString & name )
{
    if ( mAorta.getName() == name )
    {
        return & mAorta;
    }
    else if ( mPulmonaryVeins.getName() == name )
    {
        return & mPulmonaryVeins;
    }
    else if ( mPulmonaryArtery.getName() == name )
    {
        return & mPulmonaryArtery;
    }
    else if ( mSystemicVein.getName() == name )
    {
        return & mSystemicVein;
    }
    else
    {
        for ( int i = 0; i < mTubes.size(); i++ )
        {
            if ( mTubes[i].getName() == name )
            {
                return & mTubes[i];
            }
        }
    }

    return NULL;
}


TResistance * TCirculation::getResistance( const QString & name )
{

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        if ( mResistances[i].getName() == name )
        {
            return & mResistances[i];
        }
    }

    return NULL;
}
