/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: tcirculation.h 781 2013-07-14 17:01:20Z peter $
 *
 */

#ifndef _TCIRCULATION_H_
#define _TCIRCULATION_H_

#include <QList>
#include <QMetaType>

#include "theart.h"
#include "ttube.h"
#include "tresistance.h"

#include "PCircAdaptDefines.h"

namespace Peacs
{

    namespace CircAdapt
    {
        const int HEARTBEAT_NOT_SET = -1;

        class SCirculation;
        class UCirculation;

        /** @brief Class to simulate heart and circulation
          */
        class TCirculation : public TCircElement
        {
            Q_OBJECT

        public:

            /** @brief Default Constructor
              */
            TCirculation( );


            TCirculation( const TCirculation & circ);


            TCirculation( const SCirculation & circ);

            /// Destructor
            virtual ~TCirculation();

            const TCirculation& operator = (const TCirculation &circ);



            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const UCirculation &circ);



            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const SCirculation &circ);


            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );


            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            void addSample( const SCirculation & sim, int pos);



            inline const THeart & getHeart() const { return mHeart; }

            inline const TTube & getAorta() const { return mAorta; }

            inline const TTube & getPulmonaryArtery() const { return mPulmonaryArtery; }

            inline const TTube & getPulmonaryVeins() const { return mPulmonaryVeins; }


            inline const TTube & getSystemicVein() const { return mSystemicVein; }


            /** @brief TODO
              */
            inline TTube & systemicVein( void ) { return mSystemicVein; }


            inline const QVector <TTube> & getTubes() const { return mTubes; }

            inline const QVector <TResistance> & getResistances() const { return mResistances; }


            /** @return the status of the pressure contral flag on/off
              */
            inline bool   pressureControl() const  { return  mPressureControl; }

            /** @brief set the status of the pressure contral flag on/off
              * @param control the new value
              */
            inline void setPressureControl( bool control ) { mPressureControl  = control; }


            /** @brief the status of the adapt for rest flag
              */
            inline bool   adaptForRest() const { return mAdaptRest; }


            /** @brief adapt with rest setting when true
              * @param control the new value
              */
            inline void setAdaptForRest(bool adapt) { mAdaptRest = adapt; }


            /** @brief the status of the adapt for rest flag
              */
            inline bool   adaptForExercise() const { return mAdaptExercise; }

            /** @brief adapt with exercise setting when true
              * @param control the new value
              */
            inline void setAdaptForExercise(bool adapt) { mAdaptExercise = adapt; }


            /** @brief venous return  */
            double venousReturn() const { return mQ0; }


            /** @brief set the venous return */
            void setVenousReturn(double CO) { mQ0 = CO; }

            /** @brief venous return */
            double meanArterialPressure() const { return mP0; }

            /** @brief set the venous return */
            void setMeanArterialPressure(double mAP) { mP0 = mAP; }


            /** @brief Returns mHeartbeatTime
              */
            inline int getHeartbeatTime( void ) const { return mHeartbeatTime; }


            /** @brief Sets mHeartbeatTime to the new value
              *
              * @param beattime the new value for mHeartbeatTime
              */
            inline void setHeartbeatTime( int beattime ) { mHeartbeatTime = beattime; }

            /** @brief pulm drop*/
            inline void setPulmDrop(double drop) { mDropPulm = drop; }

            inline double getPulmDrop() const { return mDropPulm; }

            void setAdaptStatus(AdaptStatusType adaptStatus) { mAdtapStatus = adaptStatus; }

            AdaptStatusType getAdaptStatus() const { return mAdtapStatus; }


        private:

            TTube* getTube(const QString & name);


            TResistance* getResistance(const QString & name);


        private:

            bool   mPressureControl;     // Control blood pressure when TRUE

            bool   mAdaptRest;           // Adaptation for rest when TRUE

            bool   mAdaptExercise;       // Adaptation for exercise when TRUE

            TTube  mAorta;                // Lart

            TTube  mPulmonaryVeins;       // LVen

            TTube  mPulmonaryArtery;       // Rart

            TTube  mSystemicVein;         // RVen

            THeart mHeart;

            // List of pointers to all tubes
            QVector <TTube> mTubes;

            // List of pointers to all resistances
            QVector <TResistance> mResistances;

            double mQ0;                 // Actual systemic flow

            double mP0;                 // Actual systemic pressure

            int mHeartbeatTime;         // The time in mSec. for one heartbeat

            double mDropPulm;           // Pulmonary pressure drop

            AdaptStatusType mAdtapStatus;


        };
    }
}

// This makes it possible for TCirculation values to be stored in QVariant
// objects and retrieved later
Q_DECLARE_METATYPE( Peacs::CircAdapt::TCirculation )

#endif
