/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: tcompliance.h 776 2013-07-11 20:33:36Z eelco $
 *
 */


#ifndef __TCOMPLIANCE_H_
#define __TCOMPLIANCE_H_

#include "tcircelement.h"
#include "pvector.h"

namespace Peacs
{
    using namespace Support;
    namespace CircAdapt
    {

        class SCompliance;
        class UCompliance;
        /**
        Abstract class for compliant chamber or blood vessel.
        Inherits from CircElement.
        */
        class TCompliance: public TCircElement
        {
            Q_OBJECT


        public:
            TCompliance();

            TCompliance(const TCompliance &comp);


            TCompliance(const SCompliance &comp);

            virtual const TCompliance& operator = (const TCompliance &comp);


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const UCompliance &);

            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            virtual void copySpecifications(const SCompliance &);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            virtual void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            virtual void addSample( const SCompliance & sim, int pos);


            /// Method to get cross-section
            inline const PVector & getArea() const { return mA; }

            /// Method to get volume of compliance
            inline const PVector & getVolume() const { return mVolume; }

            inline const PVector & getVolumeDot() const { return mVDot; }

            /// Method to get pressure of compliance
            inline const PVector & getPressure() const { return mP; }

            /// Method to get pressure at entrance
            inline const PVector & getPressureAtEntrance() const { return mPin; }

            /// Method to get pressure at entrance
            inline const PVector & getImpedance() const { return mZ; }

            inline double getInitialVolume ( ) const  { return mInitialVolume; }

            inline void setInitialVolume ( double newVol ) { mInitialVolume = newVol; }

        protected:
            // Simulation state variables
            PVector mVolume;       // Volume of compliance

            PVector mVDot;    // d/dt Volume

            PVector mA;       // Cross-section of compliance

            PVector mP;       // Pressure of compliance

            PVector mZ;       // Impedance

            PVector mPin;     // Pressure at entrance

            PVector mQRemod;  // Through flow used for diameter adaptation

            double  mInitialVolume;

        };
    }
}

// This makes it possible for TTubeState values to be stored in QVariant
// objects and retrieved later
Q_DECLARE_METATYPE( Peacs::CircAdapt::TCompliance )

#endif
