/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: tpatch.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef __TPATCH_H_
#define __TPATCH_H_

#include "tcircelement.h"
#include "tsarcomere.h"
#include "PCircAdaptDefines.h"


namespace Peacs
{
    using namespace  Support;

    namespace CircAdapt
    {
        /**
        Class for a patch of the sheet encapsulating a cavity.
        Inherits from CircElement.
        */

        class TChamber;

        class SPatch;
        class UPatch;


        class TPatch: public TCircElement {
            Q_OBJECT


        public:
            /** @brief default constructor */
            TPatch ();

            TPatch (const TPatch &patch);

            TPatch (const SPatch &patch);

            virtual ~TPatch() { ; }


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const UPatch &patch);

            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const SPatch &patch);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            void addSample( const SPatch & sim, int pos);


            chamber_type getChamberType() const;

            bool isAdapationOn() const;

            bool isVentricle() const;

            const TPatch& operator = (const TPatch &element);

            void init();

            inline const TSarcomere & getSarcomere() const{ return mSarcomere; }

            /// Method to obtain wall tension (N/m)
            inline const PVector & getWallTension() const { return mT; }

            /// Method to obtain transmural pressure
            inline const PVector &  getTransmuralPressure() const { return mPTrans; }

            /// Method to obtain wall stiffness
            inline const PVector &  getWallStiffness() const { return mDTDAm; }

            /// Method to obtain wall volume
            inline double getWallVolume() const { return mVWall; }

            /// Method to obtain reference patch area
            inline const PVector &  getArea() const { return mPatchAreaMid; }

            /// Method to obtain reference patch area
            double getRefArea() const { return mAmRef; }

            /// Method to obtain reference patch area
            const PVector & getCurvature() const { return mCm; }

            /// Method to obtain reference patch area
            const PVector & getFiberStrain() const { return mEf; }



        private:
            // Specification parameters
            double mAmRef;  // Area of patch for reference sarcomere length
            double mVWall;  // Wall volume of patch

            // Simulation state variables
            PVector mPatchAreaMid;     // Current area of patch
            PVector mCm;               // Curvature = 1/radius
            PVector mT;                // Wall tension (N/m)
            PVector mPTrans;           // Transmural pressure
            PVector mDTDAm;            // Wall stiffness

            // Simulation state variables of fiber
            PVector mEf;     // Fiber strain (natural strain)


            // Sarcomere describes mechanics of myocardium
            TSarcomere mSarcomere;
        };
    }
}
// This makes it possible for TTubeState values to be stored in QVariant
// objects and retrieved later
Q_DECLARE_METATYPE( Peacs::CircAdapt::TPatch )

#endif

