/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: tsarcomere.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef __TSARCOMERE_H_
#define __TSARCOMERE_H_



#include "tcircelement.h"
#include "pvector.h"

namespace Peacs
{

using namespace  Support;

    namespace CircAdapt
    {
        class SSarcomere;
        class USarcomere;


        /**
        Class for mechanics of myocardium.
        Inherits from CircElement.
        */
        class TSarcomere: public TCircElement {
            Q_OBJECT


        public:
            /** @brief default constructor added for debug purposes*/
            TSarcomere();


            TSarcomere (const SSarcomere &sarco);

            virtual ~TSarcomere() { ;}

            const TSarcomere& operator = (const TSarcomere &sarco);


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const USarcomere &sarco);


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              *
              */
            void copySpecifications(const SSarcomere &sarco);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            void addSample( const SSarcomere & sim, int pos);


            //------------------------ memeber access function ------------

            inline void setActiovationDelay(double delay) { mActivationDelay = delay; }
            inline double getActivationdelay() const   { return mActivationDelay; }

            inline double getLsRef() const {return mLsRef; }
            inline double getLs0Pas() const {return mLs0Pas; }
            inline double getdLsPas() const {return mdLsPas; }
            inline double getSfPas() const {return mSfPas; }
            inline double getLsi0Act() const {return mLsi0Act;}          // Sarcomere length when active stress is zero
            inline double getLenSeriesElement() const {return mLenSeriesElement;} // Reference length of series elastic element
            inline double getSfAct() const {return mSfAct;}
            inline double getvMax() const {return mvMax; }
            inline double getTimeAct() const {return mTimeAct;}          // Time scale of contraction pulse
            inline double getTR() const {return mTR; }
            inline double getTD() const {return mTD;}
            inline double getCRest() const {return mCRest;}
            inline double getAdapt_LsBe() const {return mAdapt_LsBe;}       // For adaptation
            inline double getAdapt_LsEe() const {return mAdapt_LsEe; }      // For adaptation
            inline double getAdapt_SfPasMax() const {return mAdapt_SfPasMax;}   // For adaptation

            // Simulation state variables
            inline const PVector &  getFiberStrain() const {return mFiberStrain; }// Fiber strain
            inline const PVector &  getC() const {return mC;}           // Contraction
            inline const PVector &  getCDot() const {return mCDot;}        // d/dt Contraction
            inline const PVector &  getLsi() const {return mLsi;  }       // Internal sarcomere length
            inline const PVector &  getLsiDot() const {return mLsiDot;}      // d/dt Internal sarcomere length
            inline const PVector &  getLs() const {return mLs;  }        // Sarcomere length
            inline const PVector &  getSfPasT() const {return mSfPasT; }     // Fiber passive stress
            inline const PVector &  getSf() const {return mSf; }         // Fiber stress
            inline const PVector &  getDSfDEf() const {return mDSfDEf; }     // Fiber stiffness

        private:
            // Specification parameters
            double mActivationDelay;
            double mLsRef;
            double mLs0Pas;
            double mdLsPas;
            double mSfPas;
            double mLsi0Act;          // Sarcomere length when active stress is zero
            double mLenSeriesElement; // Reference length of series elastic element
            double mSfAct;
            double mvMax;
            double mTimeAct;          // Time scale of contraction pulse
            double mTR;
            double mTD;
            double mCRest;

            double mAdapt_LsBe;       // For adaptation
            double mAdapt_LsEe;       // For adaptation
            double mAdapt_SfPasMax;   // For adaptation

            // Simulation state variables
            PVector mFiberStrain; // Fiber strain

            PVector mC;           // Contraction

            PVector mCDot;        // d/dt Contraction

            PVector mLsi;         // Internal sarcomere length

            PVector mLsiDot;      // d/dt Internal sarcomere length

            PVector mLs;          // Sarcomere length

            PVector mSfPasT;      // Fiber passive stress

            PVector mSf;          // Fiber stress

            PVector mDSfDEf;      // Fiber stiffness

        };
    }
}
// This makes it possible for TTubeState values to be stored in QVariant
// objects and retrieved later
Q_DECLARE_METATYPE( Peacs::CircAdapt::TSarcomere )


#endif
