/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: uchamber.cpp 784 2013-07-16 17:35:19Z peter $
 *
 */

#include <QDebug>

#include "math.h"
#include "defs.h"
#include "tchamber.h"
#include "uchamber.h"

using namespace Peacs::CircAdapt;



// Constructor

UChamber::UChamber(int size, const QColor &color):
    UCompliance(size,color),
    mAmRef(0.0),
    mAmDead(0.0),
    mVWall(0.0),
    mVWallRef(mVWall),
    mVm( *this, VOLUME_UNITS,size),
    mXm( *this, LENGTH_UNITS,size),
    mYm( *this, LENGTH_UNITS,size),
    mSheetAreaMid( *this, AREA_UNITS,size),
    mCm( *this, CURVATIRE_UNITS,size),
    mT( *this, TENSION_UNITS,size),
    mPtrans( *this, PRESSURE_UNITS,size),
    mDTDAm( *this, DTENSION_UNITS,size),
    mY( *this, LENGTH_UNITS,size),
    mYDot( *this, DLENGTH_UNITS,size),
    mPatches(),
    mChamberType(UNDEFINED_TYPE),
    mAdaptToExercise(false),
    mUseTriSeg(false)
{

}



UChamber::UChamber(const UChamber & chamber):
    UCompliance(chamber),
    mAmRef(chamber.getAmRef()),
    mAmDead(chamber.getAmDead()),
    mVWall(chamber.getVwall()),
    mVWallRef(mVWall),
    mVm(chamber.getVm()),
    mXm(chamber.getXm()),
    mYm(chamber.getYm()),
    mSheetAreaMid(chamber.getAm()),
    mCm(chamber.getCm()),
    mT(chamber.getT()),
    mPtrans(chamber.getPtrans()),
    mDTDAm(chamber.getDTDAm()),
    mY(chamber.getY()),
    mYDot(chamber.getYdot()),
    mPatches(),
    mChamberType(chamber.getChamberType()),
    mAdaptToExercise(chamber.isAdapationOn()),
    mUseTriSeg(chamber.isTriSeg())
{
    for (int i=0; i < chamber.mPatches.size(); i++)
    {
        mPatches.append(new UPatch( *chamber.mPatches.at(i) ));
    }
}

UChamber::~UChamber()
{
    for ( int i=0; i < mPatches.size(); i++)
    {
        delete mPatches[i];
    }
}


const UChamber & UChamber::operator =(const UChamber &chamber)
{
    if ( this != &chamber)
    {
        UCompliance::operator =(chamber);

        mAmRef              = chamber.mAmRef;
        mAmDead             = chamber.mAmDead;
        mVWall              = chamber.mVWall;
        mVWallRef           = mVWall;
        mVm                 = chamber.mVm;
        mXm                 = chamber.mXm;
        mYm                 = chamber.mYm;
        mSheetAreaMid       = chamber.mSheetAreaMid;
        mCm                 = chamber.mCm;
        mT                  = chamber.mT;
        mPtrans             = chamber.mPtrans;
        mDTDAm              = chamber.mDTDAm;
        mY                  = chamber.mY;
        mYDot               = chamber.mYDot;
//        mLRef               = chamber.mLRef;
//        mEAmRef             = chamber.mEAmRef;
        mPatches            = chamber.mPatches;
        mChamberType        = chamber.mChamberType;
        mAdaptToExercise    = chamber.mAdaptToExercise;
        mUseTriSeg          = chamber.mUseTriSeg;

        for ( int i = 0; i < mPatches.size(); i++ )
        {
            delete mPatches[i];
        }

        mPatches.clear();

        for ( int i = 0; i < chamber.mPatches.size(); i++ )
        {
            mPatches.append(new UPatch( *chamber.mPatches.at(i) ));
        }
    }
    return *this;
}

void UChamber::copySpecifications(const TChamber &chamber)
{
    UCompliance::copySpecifications(chamber);
    mAmRef          = chamber.getAmRef();
    mAmDead         = chamber.getAmDead();
    mVWall          = chamber.getVwall();
    mAdaptToExercise= chamber.isAdapationOn();
    mUseTriSeg      = chamber.isTriSeg();
//    mEAmRef         = chamber.getEAmRef();

    if ( mPatches.size() != chamber.getPatches().size() )
    {
        for ( int i = 0; i < mPatches.size(); i++ )
        {
            delete mPatches[i];
        }

        int n = mVm.getSize();

        mPatches.clear();

        for ( int i = 0; i < chamber.getPatches().size(); i++ )
        {
            mPatches.append(new UPatch( n, getColor() ));
        }
    }

    for (int i=0; i < chamber.getPatches().size(); i++)
    {
        mPatches[i]->copySpecifications( *chamber.getPatches().at(i) );
    }
}

void UChamber::resetData( int size )
{
    UCompliance::resetData(size);
    mVm.reset(size);             // Volume enclosed by midwall
    mXm.reset(size);             // Central bulge distance of midwall surface
    mYm.reset(size);             // Radius of midwall at the junctional edge
    mSheetAreaMid.reset(size);   // Area of wall laid as sheet
    mCm.reset(size);             // Curvature = 1/radius of midwall
    mT.reset(size);              // Wall tension [N/m]
    mPtrans.reset(size);         // Transmural pressure
    mDTDAm.reset(size);          // dTension/dArea compliance (wall stiffness)
    mY.reset(size);              // State variable Y for TriSeg computation
    mYDot.reset(size);           // d/dt Y for TriSeg computation
//    mLRef.resize(size);


    for (int i=0; i < mPatches.size(); i++)
    {
        mPatches[i]->resetData( size );
    }
}

void UChamber::addData(const TChamber &sim)
{
    UCompliance::addData(sim);
    mVm.addBeat(sim.getVm());             // Volume enclosed by midwall
    mXm.addBeat(sim.getXm());             // Central bulge distance of midwall surface
    mYm.addBeat(sim.getYm());             // Radius of midwall at the junctional edge
    mSheetAreaMid.addBeat(sim.getAm());   // Area of wall laid as sheet
    mCm.addBeat(sim.getCm());             // Curvature = 1/radius of midwall
    mT.addBeat(sim.getT());              // Wall tension [N/m]
    mPtrans.addBeat(sim.getPtrans());         // Transmural pressure
    mDTDAm.addBeat(sim.getDTDAm());          // dTension/dArea compliance (wall stiffness)
    mY.addBeat(sim.getYm());              // State variable Y for TriSeg computation
    mYDot.addBeat(sim.getYdot());           // d/dt Y for TriSeg computation
//    mLRef.addBeat(sim.getLRef());

    if ( mPatches.size() != sim.getPatches().size() )
    {
        mPatches.append( new UPatch( mVm.getSize() ) );
    }

    for ( int i = 0; i < mPatches.size(); i++ )
    {
        mPatches[i]->addData( *sim.getPatches().at(i) );
    }
}

void UChamber::setRefValues()
{
//    UCompliance::setRefValues();
    mVWallRef       = mVWall;
    for ( int i = 0; i < mPatches.size(); i++ )
    {
        mPatches[i]->setRefValues();
    }
}

void UChamber::saveSignals()
{
    UCompliance::saveSignals();
    mVm.startSave();
    mXm.startSave();
    mYm.startSave();
    mSheetAreaMid.startSave();
    mCm.startSave();
    mT.startSave();
    mPtrans.startSave();
    mDTDAm.startSave();
    mY.startSave();
    mYDot.startSave();
//    mLRef.addBeat(sim.getLRef());

    for (int i=0; i < mPatches.size(); i++)
    {
        mPatches[i]->saveSignals();
    }

}


void UChamber::setRelSfAct(double percentage)
{
    for ( int i=0 ; i < mPatches.size(); i++ )
    {
        mPatches[i]->sarcomere().setRelSfAct( percentage );
    }
}

void UChamber::setRelSfPas(double percentage)
{
    for ( int i=0 ; i < mPatches.size(); i++ )
    {
       mPatches[i]->sarcomere().setRelSfPas( percentage );
    }
}


void UChamber::setActivationDelay(double delay)
{
    for ( int i=0 ; i < mPatches.size(); i++ )
    {
       mPatches[i]->sarcomere().setActivationDelay( delay );
    }
}

double UChamber::getActivationDelay() const
{
    return mPatches.first()->sarcomere().getActivationdelay();
}






double UChamber::getRelSfAct() const
{
    return mPatches.first()->sarcomere().getRelSfAct();
}


double UChamber::getRelSfPas() const
{
    return mPatches.first()->sarcomere().getRelSfPas();
}

