/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: uchamber.h 784 2013-07-16 17:35:19Z peter $
 *
 */

#ifndef __UCHAMBER_H_
#define __UCHAMBER_H_

#include "upatch.h"
#include "ucompliance.h"
#include "ucircadaptsignal.h"
#include "PCircAdaptDefines.h"


using namespace Peacs::Support;

namespace Peacs
{
    using namespace  Support;
    namespace CircAdapt
    {

        class TChamber;
        /**
        Class for muscular chamber. Inherits from UCompliance.
        */
        class UChamber: public UCompliance
        {

        public:
            UChamber(int size= DEFAULT_SIGNAL_LENGTH,const QColor & color = QColor(255,255,255));

            /* @brief Constructor */
            UChamber(const UChamber & chamber);

            virtual ~UChamber();

            virtual const UChamber& operator = (const UChamber &chamber);


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const TChamber &);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            virtual void addData( const TChamber & sim);

            virtual void setRefValues();

            void saveSignals();

            void init();

            bool addPatch();


            inline chamber_type getChamberType() const { return mChamberType; }

            inline bool isAdapationOn() const { return mAdaptToExercise; }

            inline bool isVentricle() const { return mChamberType == RIGHT_VENTRICLE || mChamberType == SEPTUM || mChamberType == LEFT_VENTRICLE ; }

            inline bool isTriSeg() const { return mUseTriSeg; }

            /// Method to get wall volume
            inline double getWallVolume() const { return mVWall; }

            inline double getRelWallVolume() const { return mVWall / mVWallRef;}

            inline void setRelWallVolume(double fact) { mVWall = mVWallRef * fact;}

            void setRelSfAct(double percentage);

            void setRelSfPas(double percentage);

            double getRelSfAct() const;

            double getRelSfPas() const;

            void setActivationDelay(double delay);

            double getActivationDelay() const;


//            inline const UCircAdaptSignal &  getLRef() const { return mLRef; }

            inline const UCircAdaptSignal & getY() const { return mY; }

            inline const UCircAdaptSignal & getAm() const { return mSheetAreaMid; }

            inline double getAmDead() const { return mAmDead; }

            inline const UCircAdaptSignal &getCm() const { return mCm; }

            inline const UCircAdaptSignal &getDTDAm() const { return mDTDAm; }

            inline double getAmRef() const { return mAmRef; }

            inline const UCircAdaptSignal &getPtrans() const { return mPtrans; }

            inline const UCircAdaptSignal &getT() const { return mT; }

            inline const UCircAdaptSignal &getVm() const { return mVm; }

            inline double getVwall() const { return mVWall; }

            inline const UCircAdaptSignal &getYdot() const { return mYDot; }

            inline const UCircAdaptSignal &getXm() const { return mXm; }

            inline const UCircAdaptSignal &getYm() const { return mYm; }

            inline const QList < UPatch * > &getPatches() const { return mPatches; }

        protected:
            // Specification parameters
            double mAmRef;     // Reference area of wall laid as sheet
            double mAmDead;    // Non-contractile area (Valve openings)
            double mVWall;     // Wall volume
            double mVWallRef;     // Wall volume



            // Simulation state variables
            UCircAdaptSignal mVm;             // Volume enclosed by midwall
            UCircAdaptSignal mXm;             // Central bulge distance of midwall surface
            UCircAdaptSignal mYm;             // Radius of midwall at the junctional edge
            UCircAdaptSignal mSheetAreaMid;   // Area of wall laid as sheet
            UCircAdaptSignal mCm;             // Curvature = 1/radius of midwall
            UCircAdaptSignal mT;              // Wall tension [N/m]
            UCircAdaptSignal mPtrans;         // Transmural pressure
            UCircAdaptSignal mDTDAm;          // dTension/dArea compliance (wall stiffness)
            UCircAdaptSignal mY;              // State variable Y for TriSeg computation
            UCircAdaptSignal mYDot;           // d/dt Y for TriSeg computation


            // List of pointers to patches
            // Patches form the sheet encapsulating the cavity
            QList < UPatch * > mPatches;

            chamber_type mChamberType;

            bool mAdaptToExercise;

            bool mUseTriSeg;

        };
    }
}
#endif
