/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: ucirculation.cpp 1174 2014-03-21 14:15:01Z eelco $
 *
 */

#include <QDebug>

#include "ucirculation.h"
#include "tcirculation.h"

using namespace Peacs::CircAdapt;

QDebug & operator<< ( QDebug & dbg, const UCirculation & state )
{
    dbg.nospace() << "UCirculation:";
    dbg.nospace() << "UCirculation::mHeart";
//    dbg.nospace() << state.heart();
    dbg.nospace() << "UCirculation::mAorta";
//    dbg.nospace() << state.aorta();
    dbg.nospace() << "UCirculation::mPulmonaryVeins";
//    dbg.nospace() << state.getLungVeins();
    dbg.nospace() << "UCirculation::mPulmonaryArtery";
//    dbg.nospace() << state.getRVOT();
    dbg.nospace() << "UCirculation::mVenaCava";
//    dbg.nospace() << state.getVenaCava();

    dbg.nospace() << "UCirculation::mTubesState TODO";

    dbg.nospace() << "UCirculation::mResistancesState TODO";

    return dbg.maybeSpace();
}


// Constructor: initialization only
UCirculation::UCirculation(int size):
    QObject(),
    mPressureControl( false ),
    mAdaptRest( false ),
    mAdaptExercise( false ),
    mHeart( size, HEART_COLOR ),
    mAorta( size, AORTA_COLOR ),
    mPulmonaryVeins( size, PULMONARYVEINS_COLOR ),
    mPulmonaryArtery( size, PULMONARYARTERY_COLOR ),
    mVenaCava( size, SYSTEMICVEINS_COLOR ),
    mTubes(),
    mResistances(),
    mPressureSignals(),
    mVolumeSignals(),
    mAreaSignals(),
    mFlowSignals(),
    mCurrentTimeMarker( -1.0 ),     // Negative means not used
    mInitialTimeMarker( -1.0 ),     // Negative means not used
    mSavedTimeMarker( -1.0 ),       // Negative means not used
    mQ0(8.5000e-005),
    mP0(12200),
    mHeartbeatTime( HEARTBEAT_NOT_SET ), // mSec.
    mDropPulm(2500.0),
    mAdtapStatus(TransitionAdaptation)
{
//    initSignals();
}


UCirculation::UCirculation(const UCirculation &state) :
    QObject(),
    mPressureControl(   state.mPressureControl ),
    mAdaptRest(         state.mAdaptRest ),
    mAdaptExercise(     state.mAdaptExercise ),
    mHeart(             state.mHeart ),
    mAorta(             state.mAorta ),
    mPulmonaryVeins(    state.mPulmonaryVeins ),
    mPulmonaryArtery(   state.mPulmonaryArtery ),
    mVenaCava(          state.mVenaCava ),
    mTubes(             state.mTubes ),
    mResistances(       state.mResistances ),
    mPressureSignals(   state.mPressureSignals ),
    mVolumeSignals(     state.mVolumeSignals ),
    mAreaSignals(       state.mAreaSignals ),
    mFlowSignals(       state.mFlowSignals ),
    mCurrentTimeMarker( state.mCurrentTimeMarker ),
    mInitialTimeMarker( state.mInitialTimeMarker ),
    mSavedTimeMarker(   state.mInitialTimeMarker ),
    mQ0(                state.mQ0),
    mP0(                state.mP0),
    mHeartbeatTime(     state.mHeartbeatTime ),
    mDropPulm(          state.mDropPulm),
    mAdtapStatus(       state.mAdtapStatus)
{
}


UCirculation::UCirculation(const TCirculation &circ, int size):
    QObject(),
    mPressureControl(circ.pressureControl()),
    mAdaptRest(      circ.adaptForRest()),
    mAdaptExercise(  circ.adaptForExercise()),
    mHeart( size ),
    mAorta( size ),
    mPulmonaryVeins( size ),
    mPulmonaryArtery( size ),
    mVenaCava( size ),
    mTubes( ),
    mResistances( ),
    mPressureSignals(),
    mVolumeSignals(),
    mAreaSignals(),
    mFlowSignals(),
    mQ0(circ.venousReturn()),
    mP0(circ.meanArterialPressure()),
    mHeartbeatTime( circ.getHeartbeatTime() ),
    mDropPulm(      circ.getPulmDrop() ),
    mAdtapStatus(    circ.getAdaptStatus())
{
    for ( int i = 0; i < circ.getTubes().size(); i++ )
    {
        mTubes.append( UTube( size) );
    }

    for ( int i = 0; i < circ.getResistances().size(); i++ )
    {
        mResistances.append( UResistance( size ) );
    }

    copySpecifications( circ );

    initSignals();
}


UCirculation::~UCirculation( void )
{
}


void UCirculation::initializeUser(const TCirculation &circ)
{
    initSignals();

    resetData( circ );
//    addData( circ, false );   // Not here any more
    copySpecifications( circ );

    setRefValues();

    emit( initialized() );
}


void UCirculation::setRefValues()
{
    mHeart.setRefValues();

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        mResistances[i].setRefValues();
    }
}

void UCirculation::initSignals()
{
    if ( mPressureSignals.isEmpty() )
    {
        mPressureSignals.append( &mAorta.pressureAtEntrance() );
        mPressureSignals.append( &mPulmonaryArtery.pressureAtEntrance() );
        mPressureSignals.append( &mPulmonaryVeins.pressureAtEntrance() );
        mPressureSignals.append( &mVenaCava.pressureAtEntrance() );

        mPressureSignals.append( &mHeart.atria().rightAtrium().pressure() );
        mPressureSignals.append( &mHeart.atria().leftAtrium().pressure() );
        mPressureSignals.append( &mHeart.ventricles().leftVentricle().pressure() );
        mPressureSignals.append( &mHeart.ventricles().rightVentricle().pressure() );

        mVolumeSignals.append( &mHeart.atria().rightAtrium().volume() );
        mVolumeSignals.append( &mHeart.atria().leftAtrium().volume() );
        mVolumeSignals.append( &mHeart.ventricles().leftVentricle().volume());
        mVolumeSignals.append( &mHeart.ventricles().rightVentricle().volume() );

        mFlowSignals.append( &mHeart.valveLArt().flow() );
        mFlowSignals.append( &mHeart.valveLAv().flow() );
        mFlowSignals.append( &mHeart.valveLVen().flow() );
        mFlowSignals.append( &mHeart.valveRArt().flow() );
        mFlowSignals.append( &mHeart.valveRAv().flow() );
        mFlowSignals.append( &mHeart.valveRVen().flow() );
        mFlowSignals.append( &mHeart.valveDUCT().flow() );
        mFlowSignals.append( &mHeart.valveASD().flow() );
        mFlowSignals.append( &mHeart.valveVSD().flow() );

        mFlowVelocitySignals.append( &mHeart.valveLArt().flowVelocity() );
        mFlowVelocitySignals.append( &mHeart.valveLAv().flowVelocity() );
        mFlowVelocitySignals.append( &mHeart.valveRArt().flowVelocity() );
        mFlowVelocitySignals.append( &mHeart.valveRAv().flowVelocity() );
        mFlowVelocitySignals.append( &mHeart.valveASD().flowVelocity() );
        mFlowVelocitySignals.append( &mHeart.valveVSD().flowVelocity() );
        mFlowVelocitySignals.append( &mHeart.valveDUCT().flowVelocity() );
    }

    // atria
    //mHeart.atria().leftAtrium().pressure().setUse(UCircAdaptSignal::STORE_SHOW);
//    mHeart.atria().leftAtrium().pressure().setUse(UCircAdaptSignal::STORE);
//    mHeart.atria().leftAtrium().volume().setUse(UCircAdaptSignal::STORE);
//    mHeart.atria().rightAtrium().volume().setUse(UCircAdaptSignal::STORE);

//    // ventricles
//    //mHeart.ventricles().leftVentricle().pressure().setUse(UCircAdaptSignal::STORE_SHOW);
//    mHeart.ventricles().leftVentricle().pressure().setUse(UCircAdaptSignal::STORE);
//    mHeart.ventricles().leftVentricle().volume().setUse(UCircAdaptSignal::STORE_SHOW);
//    mHeart.ventricles().rightVentricle().volume().setUse(UCircAdaptSignal::STORE);


//    // Aortic valve
//    mHeart.valveLArt().flow().setUse(UCircAdaptSignal::STORE);
//    //mHeart.valveLArt().flowVelocity().setUse(UCircAdaptSignal::STORE_SHOW);
//    mHeart.valveLArt().flowVelocity().setUse(UCircAdaptSignal::STORE);

//    // mitral valve
//    //mHeart.valveLAv().flowVelocity().setUse(UCircAdaptSignal::STORE_SHOW);
//    mHeart.valveLAv().flowVelocity().setUse(UCircAdaptSignal::STORE);

//    // tubes
//    mAorta.pressureAtEntrance().setUse(UCircAdaptSignal::STORE_SHOW);
//    mAorta.volume().setUse(UCircAdaptSignal::STORE);

//    mVenaCava.volume().setUse(UCircAdaptSignal::STORE);
//    mVenaCava.pressureAtEntrance().setUse(UCircAdaptSignal::STORE);

//    mPulmonaryArtery.volume().setUse(UCircAdaptSignal::STORE);
//    mPulmonaryVeins.volume().setUse(UCircAdaptSignal::STORE);

//    for ( int i = 0; i < mResistances.size(); i++ )
//    {
//        mResistances[i].flow().setUse(UCircAdaptSignal::STORE);
//    }

//    for ( int i = 0; i < mTubes.size(); i++ )
//    {
//        mTubes[i].volume().setUse(UCircAdaptSignal::STORE);
//    }
}


const UCirculation & UCirculation::operator= ( const UCirculation & state )
{
    if ( this != &state )
    {
        mPressureControl    = state.mPressureControl;
        mAdaptRest          = state.mAdaptRest;
        mAdaptExercise      = state.mAdaptExercise;
        mHeart              = state.mHeart;
        mAorta              = state.mAorta;
        mPulmonaryVeins     = state.mPulmonaryVeins;
        mPulmonaryArtery    = state.mPulmonaryArtery;
        mVenaCava           = state.mVenaCava;
        mTubes              = state.mTubes;
        mResistances        = state.mResistances;
        mCurrentTimeMarker  = state.mCurrentTimeMarker;
        mInitialTimeMarker  = state.mInitialTimeMarker;
        mSavedTimeMarker    = state.mSavedTimeMarker;
        mQ0                 = state.mQ0;
        mP0                 = state.mP0;
        mHeartbeatTime      = state.mHeartbeatTime;
        mDropPulm           = state.mDropPulm;
        mAdtapStatus        = state.getAdaptStatus();
    }

    return *this;
}

void UCirculation::copySpecifications(const TCirculation &circ)
{
    mPressureControl = circ.pressureControl();
    mAdaptRest       = circ.adaptForRest();
    mAdaptExercise   = circ.adaptForExercise();
    mQ0              = circ.venousReturn();
    mP0              = circ.meanArterialPressure();
    mAdtapStatus     = circ.getAdaptStatus();

    mHeart.copySpecifications(circ.getHeart());
    mAorta.copySpecifications( circ.getAorta() );
    mPulmonaryVeins.copySpecifications( circ.getPulmonaryVeins() );
    mPulmonaryArtery.copySpecifications( circ.getPulmonaryArtery() );
    mVenaCava.copySpecifications( circ.getSystemicVein() );

    for ( int i = 0; i < circ.getTubes().size(); i++ )
    {
        mTubes[i].copySpecifications(circ.getTubes().at(i) );
    }

    for ( int i = 0; i < circ.getResistances().size(); i++ )
    {
        mResistances[i].copySpecifications(circ.getResistances().at( i ));
    }

    mDropPulm = circ.getPulmDrop();
}


void UCirculation::resetData(const TCirculation & circ, int size)
{
    mPressureControl = circ.pressureControl();
    mAdaptRest       = circ.adaptForRest();
    mAdaptExercise   = circ.adaptForExercise();
    mQ0              = circ.venousReturn();
    mP0              = circ.meanArterialPressure();

    mHeart.resetData(size);
    mAorta.resetData(size);
    mPulmonaryVeins.resetData(size);
    mPulmonaryArtery.resetData(size);
    mVenaCava.resetData(size);

    if ( mTubes.size()!= circ.getTubes().size() )
    {
        mTubes.resize(circ.getTubes().size());
    }
    if ( mResistances.size()!= circ.getResistances().size() )
    {
        mResistances.resize(circ.getResistances().size());
    }

    for ( int i = 0; i < mTubes.size(); i++ )
    {
        mTubes[i].resetData(size);
    }

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        mResistances[i].resetData(size);
    }

    mHeartbeatTime = HEARTBEAT_NOT_SET;

    mDropPulm = circ.getPulmDrop();
}


void UCirculation::addData( const TCirculation & circ, bool initDone )
{
    mHeart.addData(             circ.getHeart()             );
    mAorta.addData(             circ.getAorta()             );
    mPulmonaryVeins.addData(    circ.getPulmonaryVeins()    );
    mPulmonaryArtery.addData(   circ.getPulmonaryArtery()   );
    mVenaCava.addData(          circ.getSystemicVein()      );

    for ( int i = 0; i < mTubes.size(); i++ )
    {
        mTubes[i].addData( circ.getTubes().at( i ) );
    }

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        mResistances[i].addData( circ.getResistances().at( i ) );
    }

    mPressureControl = circ.pressureControl();

    mHeartbeatTime = circ.getHeartbeatTime();

//    if ( mAdaptRest || mAdaptExercise )
    {
        copySpecifications( circ );
    }

    if ( initDone )
    {
        // Only signal if the initialization is done. This function is also called from
        // within the initializeUser function. In that case the update signal should not
        // be emitted
        emit( updatedData() );
    }
}


void UCirculation::saveSignals()
{
    mHeart.saveSignals();
    mAorta.saveSignals();
    mPulmonaryVeins.saveSignals();
    mPulmonaryArtery.saveSignals();
    mVenaCava.saveSignals();

    for ( int i = 0; i < mTubes.size(); i++ )
    {
        mTubes[i].saveSignals();
    }

    for ( int i = 0; i < mResistances.size(); i++ )
    {
        mResistances[i].saveSignals();
    }
}


const QList < UCircAdaptSignal *> &UCirculation::getSignals(signal_type type) const
{
    switch ( type )
    {
    case PRESSURE:
        return mPressureSignals;
    case FLOW:
        return mFlowSignals;
    case VOLUME:
        return mVolumeSignals;
    case FLOWVELOCITY:
        return mFlowVelocitySignals;
    default:
        return mPressureSignals;
    }
    return mPressureSignals;
}


QList < pvloop_type > UCirculation::getPVTypes() const
{
    QList< pvloop_type > list;
    list.append( PV_RIGHT_ATRIA );
    list.append( PV_LEFT_ATRIA );
    list.append( PV_RIGHT_VENTRICLES );
    list.append( PV_LEFT_VENTRICLES );
    return list;
}


const UCircAdaptSignal &UCirculation::getYPVSignal(pvloop_type type) const
{
    switch ( type )
    {
    case PV_RIGHT_ATRIA:
        return mHeart.getAtria().getRightAtrium().getPressure();
    case PV_LEFT_ATRIA:
        return mHeart.getAtria().getLeftAtrium().getPressure();
    case PV_RIGHT_VENTRICLES:
        return mHeart.getVentricles().getRightVentricle().getPressure();
    case PV_LEFT_VENTRICLES:
        return mHeart.getVentricles().getLeftVentricle().getPressure();
    default:
        qDebug().nospace() << "error in selection PV x signal";
    }
    return mHeart.getVentricles().getLeftVentricle().getPressure();
}


const UCircAdaptSignal &UCirculation::getXPVSignal(pvloop_type type) const
{
    switch ( type )
    {
    case PV_RIGHT_ATRIA:
        return mHeart.getAtria().getRightAtrium().getVolume();
    case PV_LEFT_ATRIA:
        return mHeart.getAtria().getLeftAtrium().getVolume();
    case PV_RIGHT_VENTRICLES:
        return mHeart.getVentricles().getRightVentricle().getVolume();
    case PV_LEFT_VENTRICLES:
        return mHeart.getVentricles().getLeftVentricle().getVolume();
    default:
        qDebug().nospace() << "error in selection PV y signal";
    }
    return mHeart.getVentricles().getLeftVentricle().getVolume();
}


UCircAdaptSignal &UCirculation::YPVSignal(pvloop_type type)
{
    switch ( type )
    {
    case PV_RIGHT_ATRIA:
        return mHeart.atria().rightAtrium().pressure();
    case PV_LEFT_ATRIA:
        return mHeart.atria().leftAtrium().pressure();
    case PV_RIGHT_VENTRICLES:
        return mHeart.ventricles().rightVentricle().pressure();
    case PV_LEFT_VENTRICLES:
        return mHeart.ventricles().leftVentricle().pressure();
    default:
        qDebug().nospace() << "error in selection PV x signal";
    }
    return mHeart.ventricles().leftVentricle().pressure();
}


UCircAdaptSignal &UCirculation::XPVSignal(pvloop_type type)
{
    switch ( type )
    {
    case PV_RIGHT_ATRIA:
        return mHeart.atria().rightAtrium().volume();
    case PV_LEFT_ATRIA:
        return mHeart.atria().leftAtrium().volume();
    case PV_RIGHT_VENTRICLES:
        return mHeart.ventricles().rightVentricle().volume();
    case PV_LEFT_VENTRICLES:
        return mHeart.ventricles().leftVentricle().volume();
    default:
        qDebug().nospace() << "error in selection PV y signal";
    }
    return mHeart.ventricles().leftVentricle().volume();
}


bool UCirculation::setTimeMarkers( double marker, subSignal_type subtype )
{
    bool changed = false;

    switch ( subtype )
    {
    case INITIAL_PLOT:
        if ( mInitialTimeMarker != marker )
        {
            mInitialTimeMarker  = marker;
            changed             = true;
        }
        break;
    case SAVED_PLOT:
        if ( mSavedTimeMarker != marker )
        {
            mSavedTimeMarker    = marker;
            changed             = true;
        }
        break;
    case CURRENT_PLOT:
    default:
        if ( mCurrentTimeMarker != marker )
        {
            mCurrentTimeMarker  = marker;
            changed             = true;
        }
        break;
    }

    if ( changed )
    {
        emit ( updateTimeMarker() );
    }

    return changed;
}


double UCirculation::getVolumeLastBeat() const
{
    double volume = mHeart.getVolume();

    if ( mAorta.getVolume().getLastBeat() != NULL )
    {
        volume += mAorta.getVolume().getLastBeat()->last();
    }

    if ( mPulmonaryArtery.getVolume().getLastBeat() != NULL )
    {
        volume += mPulmonaryArtery.getVolume().getLastBeat()->last();
    }

    if ( mVenaCava.getVolume().getLastBeat() != NULL )
    {
        volume += mVenaCava.getVolume().getLastBeat()->last();
    }

    if ( mPulmonaryVeins.getVolume().getLastBeat() != NULL )
    {
        volume += mPulmonaryVeins.getVolume().getLastBeat()->last();
    }

    for ( int i=0; i < mTubes.size(); i++ )
    {
        if ( mTubes[i].getVolume().getLastBeat() != NULL )
        {
            volume += mTubes[i].getVolume().getLastBeat()->last();
        }
    }
    return volume;
}
