/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: ucirculation.h 1174 2014-03-21 14:15:01Z eelco $
 *
 */

#ifndef __UCIRCULATION_H
#define __UCIRCULATION_H

#include <QMetaType>
#include <QObject>

#include "uheart.h"
#include "utube.h"
#include "uresistance.h"
#include "ucircadaptsignal.h"
#include "PCircAdaptDefines.h"

const double M3TOML = 1e6;

namespace Peacs
{
    namespace CircAdapt
    {
        class TCirculation;

        /**
        Class to simulate heart and circulation
        */
        class UCirculation : public QObject
        {
            Q_OBJECT

        public:

            /** @brief Copy Constructor
              * Creates a new instance of UCirculation and initialyzes its members
              * with the specified source
              * @param state the source to copy the data from
              */
            UCirculation( int size = DEFAULT_SIGNAL_LENGTH );


            UCirculation( const UCirculation & state );


            UCirculation( const TCirculation & state, int size );


            /** @brief Destructor
              */
            ~UCirculation( void );


            /**
              */
            void initSignals();


            /** @brief Asignment operator
              * @param state the source to copy the data from
              */
            const UCirculation & operator= ( const UCirculation & state );


        public slots:

            /** @brief function called when this object must be initialyzed
              */
            void initializeUser( const TCirculation & circ );


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const TCirculation &circ);


            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData(const TCirculation & circ, int size = DEFAULT_SIGNAL_LENGTH );


            /** @brief add a sample to the state variables
              * This function is also called from within the initializeUser function. In that case
              * the update signal should not be emitted (initDone = false)
              *
              * @param circ the recieved simulation data
              * @param initDone flag indicating if the initialization is done or not.
              */
            void addData( const TCirculation & circ, bool initDone = true );


            /** @brief save the signals in the variables
              * @see UCircAdaptSignal
              */
            void saveSignals();


            void setRefValues();


        signals:

            void updatedData();


            void initialized();


            void updateTimeMarker();


        public:

            inline double toPressure(double val) const { return val * PASCAL2MMHG; }


            inline double toVolume(double val) const { return val * M3TOML; }


            inline double toFlow(double val) const { return val * 60000.0; }


            inline double fromPressure(double val) const { return val / PASCAL2MMHG; }


            inline double fromVolume(double val) const { return val / M3TOML; }


            inline double fromFlow(double val) const { return val / 60000.0; }


            const QList < UCircAdaptSignal *> &getSignals(signal_type type) const;


            QList < pvloop_type > getPVTypes() const;


            const UCircAdaptSignal &getXPVSignal( pvloop_type ) const;


            const UCircAdaptSignal &getYPVSignal( pvloop_type ) const;


            UCircAdaptSignal &XPVSignal( pvloop_type );


            UCircAdaptSignal &YPVSignal( pvloop_type );


            /** @brief TODO
              */
            inline UHeart & heart( void ) { return mHeart; }


            /** @brief TODO
              */
            inline UTube & aorta( void ) { return mAorta; }


            /** @brief TODO
              */
            inline UTube & pulmonaryVeins( void ) { return mPulmonaryVeins; }


            /** @brief TODO
              */
            inline UTube & pulmonaryArtery( void ) { return mPulmonaryArtery; }


            /** @brief TODO
              */
            inline UTube & systemicVein( void ) { return mVenaCava; }


            /** @brief TODO
              */
            inline QVector<UTube> & tubes( void ) { return mTubes; }


            /** @brief TODO
              */
            inline QVector<UResistance> & resistances( void )  { return mResistances; }


            /** @brief TODO
              */
            inline const UHeart & getHeart( void ) const { return mHeart; }


            /** @brief TODO
              */
            inline const UTube & getAorta( void ) const { return mAorta; }


            /** @brief TODO
              */
            inline const UTube & getPulmonaryVeins( void )const  { return mPulmonaryVeins; }


            /** @brief TODO
              */
            inline const UTube & getPulmonaryArtery( void ) const { return mPulmonaryArtery; }


            /** @brief TODO
              */
            inline const UTube & getSystemicVein( void ) const { return mVenaCava; }


            /** @brief TODO
              */
            inline const QVector<UTube> & getTubes( void ) const { return mTubes; }


            /** @brief TODO
              */
            inline const QVector<UResistance> & getResistances( void ) const  { return mResistances; }

            /** @return the status of the pressure contral flag on/off
              */
            bool   pressureControl() const  { return  mPressureControl; }

            /** @brief set the status of the pressure contral flag on/off
              * @param control the new value
              */
            void setPressureControl( bool control ) { mPressureControl  = control; }


            /** @brief the status of the adapt for rest flag
              */
            bool   adaptForRest() const { return mAdaptRest; }


            /** @brief adapt with rest setting when true
              * @param control the new value
              */
            void setAdaptForRest(bool adapt) { mAdaptRest = adapt; }


            /** @brief the status of the adapt for rest flag
              */
            bool   adaptForExercise() const { return mAdaptExercise; }

            /** @brief adapt with exercise setting when true
              * @param control the new value
              */
            void setAdaptForExercise(bool adapt) { mAdaptExercise = adapt; }


            int getCurrentTimeMarker() const { return mCurrentTimeMarker; }

            int getInitialTimeMarker() const { return mInitialTimeMarker; }

            int getSavedTimeMarker() const { return mSavedTimeMarker; }


            /** @brief Sets the time marker position for the specified graph sub-type
              * Markers are shown for the initial, snapshot and, perhaps in the future, the current graphs.
              * these markers can be moved by the user. This function will handle the changes in marker
              * position.
              *
              * @param marker The new marker time value
              * @param subtype The bub-type of the marker, i.e. initial, snapshot or current
              * @return true if the marker value has changed
              */
            bool setTimeMarkers( double marker, subSignal_type subtype );


            inline void setCurrentTimeMarker(double t)  { mCurrentTimeMarker = t; emit(updateTimeMarker()); }

            inline void setInitialTimeMarker(double t)  { mInitialTimeMarker = t; emit(updateTimeMarker());}

            inline void setSavedTimeMarker(double t)    { mSavedTimeMarker = t; emit(updateTimeMarker());}

            double getVolumeLastBeat() const;

//            double getRefVolume() const { return mRefCircVolume; }


            /** @brief cardiacOutput */
            double venousReturn() const { return mQ0; }


            /** @brief set the cardiacOutput */
            void setVenousReturn( double CO ) { mQ0 = CO; }

            /** @brief venous return */
            double meanArterialPressure() const { return mP0; }

            /** @brief set the venous return */
            void setMeanArterialPressure(double mAP) { mP0 = mAP; }


            /** @brief Returns mHeartbeatTime
              */
            int getHeartbeatTime( void ) const { return mHeartbeatTime; }


            /** @brief Sets mHeartbeatTime to the new value
              *
              * @param beattime the new value for mHeartbeatTime
              */
            void setHeartbeatTime( int beattime ) { mHeartbeatTime = beattime; }

            /** @brief pulm drop*/
            void setPulmDrop(double drop) { mDropPulm = drop; }

            double getPulmDrop() const { return mDropPulm; }

            void setAdaptStatus(AdaptStatusType adaptStatus) { mAdtapStatus = adaptStatus; }

            AdaptStatusType getAdaptStatus() const { return mAdtapStatus; }

        private:
            bool   mPressureControl;     // Control blood pressure when TRUE

            bool   mAdaptRest;           // Adaptation for rest when TRUE

            bool   mAdaptExercise;       // Adaptation for exercise when TRUE

            /** @brief TODO
              */
            UHeart          mHeart;


            /** @brief TODO
              */
            UTube           mAorta;


            /** @brief TODO
              */
            UTube           mPulmonaryVeins;


            /** @brief TODO
              */
            UTube           mPulmonaryArtery;


            /** @brief TODO
              */
            UTube           mVenaCava;


            /** List of pointers to all tubes
              */
            QVector <UTube> mTubes;


            /** List of pointers to all resistances
              */
            QVector <UResistance> mResistances;


            QList < UCircAdaptSignal *> mPressureSignals;

            QList < UCircAdaptSignal *> mVolumeSignals;

            QList < UCircAdaptSignal *> mAreaSignals;

            QList < UCircAdaptSignal *> mFlowSignals;

            QList < UCircAdaptSignal *> mFlowVelocitySignals;


            double mCurrentTimeMarker;


            double mInitialTimeMarker;


            double mSavedTimeMarker;


            double mQ0;                 // Actual systemic flow


            double mP0;                 // Actual systemic pressure


            int mHeartbeatTime;     // The time in mSec. for one heartbeat


            double mDropPulm;           // Pulmonary pressure drop

            AdaptStatusType mAdtapStatus;

        };
    }
}
/** @brief TODO
  */
QDebug & operator<< ( QDebug & dbg, const Peacs::CircAdapt::UCirculation & state );


// This makes it possible for UCirculation values to be stored in QVariant
// objects and retrieved later
//Q_DECLARE_METATYPE( Peacs::CircAdapt::UCirculation );


#endif
