/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: ucompliance.h 776 2013-07-11 20:33:36Z eelco $
 *
 */


#ifndef __UCOMPLIANCE_H_
#define __UCOMPLIANCE_H_


#include "tcompliance.h"
#include "ucircelement.h"
#include "ucircadaptsignal.h"
#include "PCircAdaptDefines.h"


namespace Peacs
{
    using namespace Support;

    namespace CircAdapt
    {
        /**
        Abstract class for compliant chamber or blood vessel.
        Inherits from CircElement.
        */
        class UCompliance: public UCircElement {

        public:
            UCompliance(int size = DEFAULT_SIGNAL_LENGTH,const QColor & color = QColor(255,255,255));


            UCompliance(const UCompliance &comp);

            virtual const UCompliance& operator = (const UCompliance &comp);


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            virtual void copySpecifications(const TCompliance &);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            virtual void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            virtual void addData( const TCompliance & sim);

            /** @brief save the signals in the variables
              * @see UCircAdaptSignal*/
            virtual void saveSignals();

            /// Method to get cross-section
            inline UCircAdaptSignal & area() { return mA; }

            /// Method to get volume of compliance
            inline UCircAdaptSignal & volume() { return mVolume; }

            inline UCircAdaptSignal & volumeDot() { return mVDot; }

            /// Method to get pressure of compliance
            inline UCircAdaptSignal & pressure() { return mP; }

            /// Method to get pressure at entrance
            inline UCircAdaptSignal & pressureAtEntrance() { return mPin; }

            /// Method to get pressure at entrance
            inline UCircAdaptSignal & impedance() { return mZ; }


            /// Method to get cross-section
            inline const UCircAdaptSignal & getArea() const { return mA; }

            /// Method to get volume of compliance
            inline const UCircAdaptSignal & getVolume()const  { return mVolume; }

            inline const UCircAdaptSignal & getVolumeDot()const { return mVDot; }

            /// Method to get pressure of compliance
            inline const UCircAdaptSignal & getPressure() const { return mP; }

            /// Method to get pressure at entrance
            inline const UCircAdaptSignal & getPressureAtEntrance() const { return mPin; }

            /// Method to get pressure at entrance
            inline const UCircAdaptSignal & getImpedance() const { return mZ; }

        protected:
            // Simulation state variables
            UCircAdaptSignal mVolume;       // Volume of compliance

            UCircAdaptSignal mVDot;    // d/dt Volume

            UCircAdaptSignal mA;       // Cross-section of compliance

            UCircAdaptSignal mP;       // Pressure of compliance

            UCircAdaptSignal mZ;       // Impedance

            UCircAdaptSignal mPin;     // Pressure at entrance

            UCircAdaptSignal mQRemod;  // Through flow used for diameter adaptation
        };
    }
}
#endif
