/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: uheart.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef __UHEART_H_
#define __UHEART_H_


#include <QMetaType>
#include <QList>
#include "ucircadaptsignal.h"
#include "theart.h"
#include "uatria.h"
#include "uventricle.h"
#include "usack.h"
#include "uvalve.h"

using namespace Peacs::Support;

namespace Peacs
{

    namespace CircAdapt
    {
        class UHeart : public UCircElement
        {

        public:

            enum pressure_type { RAP = 0, RVP=1, LAP=2, LVP=3 };

            enum volume_type   { RAV = 0, RVV=1, LAV=2, LVV=3 };

            enum flow_type     { RAF = 0, RVF=1, LAF=2, LVF=3 };


            UHeart(int size = DEFAULT_SIGNAL_LENGTH,const QColor & color = QColor(255,255,255));

            /** @brief Copy Constructor
              * Creates a new instance of THeartState and initialyzes its members
              * with the specified source
              * @param state the source to copy the data from
              */
            UHeart(const UHeart &heart);

            /** @brief Destructor
              */
            ~UHeart( void ) {;}


            /** @brief Asignment operator
              * @param state the source to copy the data from
              */
            virtual const UHeart & operator= ( const UHeart & state );


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const THeart &heart);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            void addData( const THeart& sim);

            /** @brief save the signals in the variables
              * @see UCircAdaptSignal*/
            void saveSignals();


            void setRefValues();


            inline UAtria &atria() { return mAtria; }

            inline UVentricle &ventricles() { return mVentricle; }

            inline USack &vack() { return mPericardialSack; }

            inline UValve &valveLVen() { return mValveLVen; }           // Inlet left atrium
            inline UValve &valveLAv()  { return mValveLAv;   }        // Mitral valve
            inline UValve &valveLArt() { return mValveLArt;  }          // Aortic valve
            inline UValve &valveRVen() { return mValveRVen; }            // Inlet right atrium
            inline UValve &valveRAv()  { return mValveRAv; }            // Tricuspid valve
            inline UValve &valveRArt() { return mValveRArt; }           // Pulmonary valve
            inline UValve &valveASD()  { return mValveASD;  }           // Valve for ASD
            inline UValve &valveVSD()  { return mValveVSD;  }           // Valve for VSD
            inline UValve &valveDUCT() { return mValveDUCT; }            // Valve for DUCT

            inline const UAtria &getAtria() const { return mAtria; }

            inline const UVentricle &getVentricles() const { return mVentricle; }

            inline const USack &getSack() const { return mPericardialSack; }

            inline const UValve &getValveLVen() const { return mValveLVen; }           // Inlet left atrium
            inline const UValve &getValveLAv()  const { return mValveLAv;   }        // Mitral valve
            inline const UValve &getValveLArt() const { return mValveLArt;  }          // Aortic valve
            inline const UValve &getValveRVen() const { return mValveRVen; }            // Inlet right atrium
            inline const UValve &getValveRAv()  const { return mValveRAv; }            // Tricuspid valve
            inline const UValve &getValveRArt() const { return mValveRArt; }           // Pulmonary valve
            inline const UValve &getValveASD()  const { return mValveASD;  }           // Valve for ASD
            inline const UValve &getValveVSD()  const { return mValveVSD;  }           // Valve for VSD
            inline const UValve &getValveDUCT() const { return mValveDUCT; }            // Valve for DUCT

//            inline const UCircAdaptSignal & getPressure( pressure_type type ) const { return mPressures[type]; }

            double getVolume() const;


        private:
            UAtria      mAtria;

            UVentricle  mVentricle;

            USack        mPericardialSack;

            UValve       mValveLVen;            // Inlet left atrium
            UValve       mValveLAv;             // Mitral valve
            UValve       mValveLArt;            // Aortic valve
            UValve       mValveRVen;            // Inlet right atrium
            UValve       mValveRAv;             // Tricuspid valve
            UValve       mValveRArt;            // Pulmonary valve
            UValve       mValveASD;             // Valve for ASD
            UValve       mValveVSD;             // Valve for VSD
            UValve       mValveDUCT;            // Valve for DUCT


        };

    }
}

QDebug & operator<< ( QDebug & dbg, const Peacs::CircAdapt::UHeart & state );


#endif
