/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: upatch.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef __UPATCH_H_
#define __UPATCH_H_

#include "ucircelement.h"
#include "usarcomere.h"
#include "PCircAdaptDefines.h"
#include "ucircadaptsignal.h"

namespace Peacs
{
    using namespace  Support;

    namespace CircAdapt
    {
        /**
        Class for a patch of the sheet encapsulating a cavity.
        Inherits from CircElement.
        */

        class TChamber;

        class TPatch;


        class UPatch: public UCircElement {

        public:
            /** @brief default constructor */
            UPatch (int size = DEFAULT_SIGNAL_LENGTH,const QColor & color = QColor(255,255,255));

            UPatch (const UPatch &patch);

            virtual ~UPatch() { ; }


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const TPatch &patch);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            void addData( const TPatch & sim);

            /** @brief save the signals in the variables
              * @see UCircAdaptSignal*/
            void saveSignals();

            virtual void setRefValues();

            chamber_type getChamberType() const;

            bool isAdapationOn() const;

            bool isVentricle() const;

            const UPatch& operator = (const UPatch &element);

            void init();

            const USarcomere & getSarcomere() const { return mSarcomere; }

            USarcomere & sarcomere() { return mSarcomere; }

            /// Method to obtain wall tension (N/m)
            inline const UCircAdaptSignal & getWallTension() const { return mT; }

            /// Method to obtain transmural pressure
            inline const UCircAdaptSignal &  getTransmuralPressure() const { return mPTrans; }

            /// Method to obtain wall stiffness
            inline const UCircAdaptSignal &  getWallStiffness() const { return mDTDAm; }

            /// Method to obtain wall volume
            inline double getWallVolume() const { return mVWall; }

            /// Method to obtain reference patch area
            inline const UCircAdaptSignal &  getArea() const { return mPatchAreaMid; }

            /// Method to obtain reference patch area
            inline double getRefArea() const { return mAmRef; }

            /// Method to obtain reference patch area
            inline const UCircAdaptSignal &getCurvature() const { return mCm; }

            /// Method to obtain reference patch area
            inline const UCircAdaptSignal &getFiberStrain() const { return mEf; }



        private:
            // Specification parameters
            double mAmRef;  // Area of patch for reference sarcomere length
            double mVWall;  // Wall volume of patch

            // Simulation state variables
            UCircAdaptSignal mPatchAreaMid;     // Current area of patch
            UCircAdaptSignal mCm;               // Curvature = 1/radius
            UCircAdaptSignal mT;                // Wall tension (N/m)
            UCircAdaptSignal mPTrans;           // Transmural pressure
            UCircAdaptSignal mDTDAm;            // Wall stiffness

            // Simulation state variables of fiber
            UCircAdaptSignal mEf;     // Fiber strain (natural strain)


            // Sarcomere describes mechanics of myocardium
            USarcomere mSarcomere;
        };
    }
}

#endif

