/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: usarcomere.h 784 2013-07-16 17:35:19Z peter $
 *
 */

#ifndef __USARCOMERE_H_
#define __USARCOMERE_H_



#include "ucircelement.h"
#include "ucircadaptsignal.h"
#include "PCircAdaptDefines.h"



namespace Peacs
{

using namespace  Support;

    namespace CircAdapt
    {
        class TSarcomere;

        /**
        Class for mechanics of myocardium.
        Inherits from CircElement.
        */
        class USarcomere: public UCircElement {

        public:
            /** @brief default constructor added for debug purposes*/
            USarcomere(int size = DEFAULT_SIGNAL_LENGTH,const QColor & color = QColor(255,255,255));


            USarcomere (const USarcomere &sarco);

            virtual ~USarcomere() { ;}

            const USarcomere& operator = (const USarcomere &sarco);


            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const TSarcomere &sarc);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            void addData( const TSarcomere & sim);


            /** @brief save the signals in the variables
              * @see UCircAdaptSignal*/
            void saveSignals();

            void setRefValues();


            //------------------------ memeber access function ------------

            inline void setActivationDelay(double delay) { mActivationDelay = delay; }
            inline double getActivationdelay() const   { return mActivationDelay; }

            inline double getLsRef() const {return mLsRef; }
            inline double getLs0Pas() const {return mLs0Pas; }
            inline double getdLsPas() const {return mdLsPas; }
            inline double getSfPas() const {return mSfPas; }

            inline double getRelSfPas() const {return mSfPas / mSfPasRef;}

            inline void setRelSfPas(double sfact) { mSfPas = mSfPasRef * sfact;}

            inline double getLsi0Act() const {return mLsi0Act;}          // Sarcomere length when active stress is zero
            inline double getLenSeriesElement() const {return mLenSeriesElement;} // Reference length of series elastic element

            inline double getSfAct() const {return mSfAct;}

            inline double getRelSfAct() const {return mSfAct / mSfActRef;}

//            inline void setSfAct(double sfact) { mSfAct = sfact;}

            inline void setRelSfAct(double sfact) { mSfAct = mSfActRef * sfact;}



            inline double getvMax() const {return mvMax; }
            inline double getTimeAct() const {return mTimeAct;}          // Time scale of contraction pulse
            inline double getTR() const {return mTR; }
            inline double getTD() const {return mTD;}
            inline double getCRest() const {return mCRest;}
            inline double getAdapt_LsBe() const {return mAdapt_LsBe;}       // For adaptation
            inline double getAdapt_LsEe() const {return mAdapt_LsEe; }      // For adaptation
            inline double getAdapt_SfPasMax() const {return mAdapt_SfPasMax;}   // For adaptation

            // Simulation state variables
            inline const UCircAdaptSignal &  getFiberStrain() const {return mFiberStrain; }// Fiber strain
            inline const UCircAdaptSignal &  getC() const {return mC;}           // Contraction
            inline const UCircAdaptSignal &  getCDot() const {return mCDot;}        // d/dt Contraction
            inline const UCircAdaptSignal &  getLsi() const {return mLsi;  }       // Internal sarcomere length
            inline const UCircAdaptSignal &  getLsiDot() const {return mLsiDot;}      // d/dt Internal sarcomere length
            inline const UCircAdaptSignal &  getLs() const {return mLs;  }        // Sarcomere length
            inline const UCircAdaptSignal &  getSfPasT() const {return mSfPasT; }     // Fiber passive stress
            inline const UCircAdaptSignal &  getSf() const {return mSf; }         // Fiber stress
            inline const UCircAdaptSignal &  getDSfDEf() const {return mDSfDEf; }     // Fiber stiffness

        private:
            // Specification parameters
            double mActivationDelay;

            double mLsRef;

            double mLs0Pas;

            double mdLsPas;

            double mSfPas;

            double mSfPasRef;

            double mLsi0Act;          // Sarcomere length when active stress is zero

            double mLenSeriesElement; // Reference length of series elastic element

            double mSfAct;

            double mSfActRef;

            double mvMax;

            double mTimeAct;          // Time scale of contraction pulse

            double mTR;

            double mTD;

            double mCRest;

            double mAdapt_LsBe;       // For adaptation

            double mAdapt_LsEe;       // For adaptation

            double mAdapt_SfPasMax;   // For adaptation

            // Simulation state variables
            UCircAdaptSignal mFiberStrain; // Fiber strain
            UCircAdaptSignal mC;           // Contraction
            UCircAdaptSignal mCDot;        // d/dt Contraction
            UCircAdaptSignal mLsi;         // Internal sarcomere length
            UCircAdaptSignal mLsiDot;      // d/dt Internal sarcomere length
            UCircAdaptSignal mLs;          // Sarcomere length
            UCircAdaptSignal mSfPasT;      // Fiber passive stress
            UCircAdaptSignal mSf;          // Fiber stress
            UCircAdaptSignal mDSfDEf;      // Fiber stiffness

        };
    }
}

#endif
