/*
 *  This file is part of the CircAdapt package (see http:/www.circadapt.org)
 *
 *  Copyright (C) 2012 Peacs (www.peacs.nl)
 *
 *  CircAdapt is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  CircAdapt is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CircAdapt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    Project:              CircAdapt
 *    Compiler:             MinGW, gcc
 *    Tatrget platforms:    WIndows, Mac OS-X
 *
 *
 *    Subversion info:
 *
 *    $Id: utube.h 776 2013-07-11 20:33:36Z eelco $
 *
 */

#ifndef __UTUBE_H_
#define __UTUBE_H_

#include <QMetaType>
#include <QString>
#include "ucompliance.h"
#include "ucircadaptsignal.h"
#include "PCircAdaptDefines.h"

using namespace Peacs::Support;

namespace Peacs
{

    namespace CircAdapt
    {

        class TTube;

        /** @brief Class for compliant blood vessel. Inherits from Compliance.
          */
        class UTube :public UCompliance
        {

        public:

            /** @brief Default Constructor
              */
            UTube(  int size = DEFAULT_SIGNAL_LENGTH, const QColor & color = QColor(255,255,255));


            /** @brief Copy Constructor
              * Creates a new instance of TTubeState and initialyzes its members
              * with the specified source
              * @param state the source to copy the data from
              */
            UTube( const UTube & tube );

            /** @brief Destructor
              */
            ~UTube( void ) {;}

            /** @brief Asignment operator
              * @param state the source to copy the data from
              */
            virtual const UTube & operator= ( const UTube & state );


            /** @brief copy the sepcification parameters from the user class to the transfer class
              */
            void copyUserSpecifications(const UTube &);

            /** @brief copy the sepcification parameters from the simulation class to the transfer class
              * no specification parameters in this class
              */
            void copySpecifications(const TTube &tube);

            /** @brief At the start of a simulation run the state variables are resized
              * @param size the number of samples in this heart beat
              */
            void resetData( int size );

            /** @brief add a sample to the state variables
              * @param sim the simulate class
              * @param pos the sample number in the state variables
              */
            virtual void addData( const TTube & sim);

            /** @brief save the signals in the variables
              * @see UCircAdaptSignal*/
            virtual void saveSignals();


            // =================== access functions =================
            /** @brief set stiffness exponent of wall material
              * @param k new stiffness
              */
            inline void setStiffnessWall(double k)  { mK =k; }

            /** @returns stiffness exponent of wall material
              */
            inline double stiffnessWall()  const { return mK; }


            /** @brief set Mean flow at rest
              * @param k new stiffness
              */
            inline void setMeanRestFlow(double q0)  { mQ0 = q0; }

            /** @returns Mean flow at rest
              */
            inline double meanRestFlow()  const { return mQ0; }


            /** @brief set Mean flow at rest
              * @param p0 new p0
              */
            inline void setMeanRestPressure(double p0)  { mP0 = p0; }

            /** @returns Mean flow at rest
              */
            inline double meanRestPressure()  const { return mP0; }


            /** @brief set Effective length of vessel (compliance)
              * @param vl new vessel length
              */
            inline void setVesselLength(double vl)  { mVesselLength = vl; }

            /** @returns Effective length of vessel (compliance)
              */
            inline double meanVesselLength()  const { return mVesselLength; }


            /** @brief set Maximum pressure
              * @param pmax new maximum pressure
              */
            inline void setPmax(double pmax)  { mPmax = pmax; }

            /** @returns Maximum pressure
              */
            inline double Pmax()  const { return mPmax; }


            /** @brief set Cross-sectional area of wall
              * @param Awall new Cross-sectional area of wall
              */
            inline void setAwall(double awall)  { mAWall = awall; }

            /** @returns Cross-sectional area of wall
              */
            inline double Awall()  const { return mAWall; }


            /** @brief set Cross-sectional area of wall
              * @param a0 new Mean cross-section at rest
              */
            inline void setA0(double a0)  { mA0 = a0; }

            /** @returns Cross-sectional area of wall
              */
            inline double A0()  const { return mA0; }


            // Specification parameters for adaptation
            /** @brief set Maximum wall stress
              * @param wallStressAdapt new Maximum wall stress
              */
            inline void setWallStressAdapt(double wallStressAdapt)  { mWallStressAdapt = wallStressAdapt; }

            /** @returns Maximum wall stress
              */
            inline double wallStressAdapt()  const { return mWallStressAdapt; }


            /** @brief set Mean flow velocity
              * @param vFlowMeanAdapt new Mean flow velocity
              */
            inline void setVFlowMeanAdapt(double vFlowMeanAdapt)  { mVFlowMeanAdapt = vFlowMeanAdapt; }

            /** @returns Mean flow velocity
              */
            inline double vFlowMeanAdapt()  const { return mVFlowMeanAdapt; }

            /** @brief set Shock wave velocity due to body movement
              * @param vImpactAdapt new Shock wave velocity due to body movement
              */
            inline void setVImpactAdapt(double vImpactAdapt)  { mVImpactAdapt = vImpactAdapt; }

            /** @returns Shock wave velocity due to body movement
              */
            inline double vImpactAdapt()  const { return mVImpactAdapt; }

        private:
//            // Simulation state variables
//            UCircAdaptSignal mVolume;       // Volume of compliance

//            UCircAdaptSignal mVDot;    // d/dt Volume

//            UCircAdaptSignal mA;       // Cross-section of compliance

//            UCircAdaptSignal mP;       // Pressure of compliance

//            UCircAdaptSignal mZ;       // Impedance

//            UCircAdaptSignal mPin;     // Pressure at entrance

//            UCircAdaptSignal mQRemod;  // Through flow used for diameter adaptation


            double mK;          // Stiffness exponent of wall material

            double mQ0;         // Mean flow at rest (not used)

            double mP0;         // Mean pressure at rest

            double mVesselLength;// Effective length of vessel (compliance)

            double mPmax;       // Maximum pressure

            double mAWall;      // Cross-sectional area of wall

            double mA0;         // Mean cross-section at rest

            // Specification parameters for adaptation
            double mWallStressAdapt; // Maximum wall stress

            double mVFlowMeanAdapt;  // Mean flow velocity

            double mVImpactAdapt;    // Shock wave velocity due to body movement

        };
    }
}


QDebug & operator << ( QDebug & dbg, const Peacs::CircAdapt::UTube & state );



#endif
